/*
   This file is part of TALER
   Copyright (C) 2024 Taler Systems SA

   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 3, or (at your option) any later version.

   TALER is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with
   TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */
/**
 * @file backenddb/pg_account_kyc_set_failed.h
 * @brief implementation of the account_kyc_set_failed function for Postgres
 * @author Christian Grothoff
 */
#ifndef PG_ACCOUNT_KYC_SET_FAILED_H
#define PG_ACCOUNT_KYC_SET_FAILED_H

#include <taler/taler_util.h>
#include <taler/taler_json_lib.h>
#include "taler_merchantdb_plugin.h"


/**
 * Set an instance's account's KYC status to failed.
 *
 * @param cls closure
 * @param merchant_id merchant backend instance ID
 * @param h_wire hash of the wire account to check
 * @param exchange_url base URL of the exchange to check
 * @param timestamp timestamp to store
 * @param exchange_http_status HTTP status code returned last by the exchange
 * @param kyc_ok current KYC status (should be false)
 * @return database result code
 */
enum GNUNET_DB_QueryStatus
TMH_PG_account_kyc_set_failed (
  void *cls,
  const char *merchant_id,
  const struct TALER_MerchantWireHashP *h_wire,
  const char *exchange_url,
  struct GNUNET_TIME_Timestamp timestamp,
  unsigned int exchange_http_status,
  bool kyc_ok);

#endif
