/*
  This file is part of TALER
  Copyright (C) 2025 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
*/
/**
 * @file backenddb/pg_lookup_product_image.h
 * @brief declaration of lookup_product_image_by_hash for Postgres
 * @author Bohdan Potuzhnyi
 */
#ifndef PG_LOOKUP_PRODUCT_IMAGE_H
#define PG_LOOKUP_PRODUCT_IMAGE_H

#include "pg_helper.h"

enum GNUNET_DB_QueryStatus
TMH_PG_lookup_product_image_by_hash (void *cls,
                                     const char *instance_id,
                                     const char *image_hash,
                                     char **image);

#endif