/*
  This file is part of TALER
  Copyright (C) 2025 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU Affero General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.

  You should have received a copy of the GNU Affero General Public License along with
  TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
*/
/**
 * @file taler-merchant-httpd_get-products-image.h
 * @brief implement GET /products/$HASH/image
 * @author Bohdan Potuzhnyi
 */
#ifndef TALER_MERCHANT_HTTPD_GET_PRODUCTS_IMAGE_H
#define TALER_MERCHANT_HTTPD_GET_PRODUCTS_IMAGE_H

#include "taler-merchant-httpd.h"

MHD_RESULT
TMH_get_products_image (const struct TMH_RequestHandler *rh,
                        struct MHD_Connection *connection,
                        struct TMH_HandlerContext *hc);

#endif