// src/i18n/strings.ts
var strings = {};
strings["uk"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=3; plural=n%10==1 && n%100!=11 ? 0 : n%10>=2 && n%10<=4 && (n%100<10 || n%100>=20) ? 1 : 2;",
        "lang": "uk"
      },
      "Cancel": [
        "\u0421\u043A\u0430\u0441\u0443\u0432\u0430\u0442\u0438"
      ],
      "%1$s": [
        "%1$s"
      ],
      "Close": [
        "\u0417\u0430\u043A\u0440\u0438\u0442\u0438"
      ],
      "Continue": [
        "\u041F\u0440\u043E\u0434\u043E\u0432\u0436\u0438\u0442\u0438"
      ],
      "Clear": [
        "\u041E\u0447\u0438\u0441\u0442\u0438\u0442\u0438"
      ],
      "Confirm": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0438\u0442\u0438"
      ],
      "Correct form": [
        ""
      ],
      "Comparing account details": [
        ""
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        ""
      ],
      "Field": [
        ""
      ],
      "In the form": [
        ""
      ],
      "Reported": [
        ""
      ],
      "Type": [
        ""
      ],
      "IBAN": [
        ""
      ],
      "Address": [
        "\u0410\u0434\u0440\u0435\u0441\u0430"
      ],
      "Host": [
        "\u0425\u043E\u0441\u0442"
      ],
      "Account ID": [
        "\u0420\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "The account owner's legal name": [
        "\u0406\u043C'\u044F \u0432\u043B\u0430\u0441\u043D\u0438\u043A\u0430 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u0440\u0430\u0445\u0443\u043D\u043A\u0443."
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        "\u041C\u0456\u0441\u0442\u043E"
      ],
      "Bank host": [
        "\u0425\u043E\u0441\u0442 \u0431\u0430\u043D\u043A\u0443"
      ],
      "Bank account": [
        "\u0411\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "BIC": [
        ""
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        "\u041F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0430 \u0434\u0456\u0439\u0441\u043D\u043E\u0441\u0442\u0456 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u0440\u0430\u0445\u0443\u043D\u043A\u0443: %1$s"
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        ""
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        ""
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        ""
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        ""
      ],
      "Subject": [
        "\u041F\u0440\u0438\u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F"
      ],
      "Receiver postal code": [
        "\u0414\u0430\u0442\u0430 \u0434\u043E\u0441\u0442\u0430\u0432\u043A\u0438"
      ],
      "Receiver town": [
        "\u0414\u0430\u0442\u0430 \u0434\u043E\u0441\u0442\u0430\u0432\u043A\u0438"
      ],
      "Step 3:": [
        ""
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        ""
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        ""
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        ""
      ],
      "Operation in progress...": [
        "\u041E\u043F\u0435\u0440\u0430\u0446\u0456\u044F \u0432\u0438\u043A\u043E\u043D\u0443\u0454\u0442\u044C\u0441\u044F..."
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        "\u041E\u043F\u0435\u0440\u0430\u0446\u0456\u044F \u0431\u0443\u0434\u0435 \u0430\u0432\u0442\u043E\u043C\u0430\u0442\u0438\u0447\u043D\u043E \u0441\u043A\u0430\u0441\u043E\u0432\u0430\u043D\u0430 \u0447\u0435\u0440\u0435\u0437 %1$s \u0441\u0435\u043A\u0443\u043D\u0434"
      ],
      "days": [
        "\u0434\u043D\u0456"
      ],
      "hours": [
        "\u0433\u043E\u0434\u0438\u043D\u0438"
      ],
      "minutes": [
        "\u0445\u0432\u0438\u043B\u0438\u043D\u0438"
      ],
      "seconds": [
        "\u0441\u0435\u043A\u0443\u043D\u0434\u0438"
      ],
      "Forever": [
        "\u043D\u0430\u0437\u0430\u0432\u0436\u0434\u0438"
      ],
      "%1$sM": [
        "%1$s\u041C"
      ],
      "%1$sY": [
        "%1$s\u0420"
      ],
      "%1$sd": [
        "%1$s\u0434\u043D."
      ],
      "%1$sh": [
        "%1$s\u0433"
      ],
      "%1$smin": [
        "%1$s\u0445\u0432"
      ],
      "%1$ssec": [
        "%1$s\u0441\u0435\u043A"
      ],
      "Change value to never": [
        "\u0437\u043C\u0456\u043D\u0438\u0442\u0438 \u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F \u043D\u0430 \u043D\u0456\u043A\u043E\u043B\u0438"
      ],
      "Change value to empty": [
        "\u0437\u043C\u0456\u043D\u0438\u0442\u0438 \u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F \u043D\u0430 \u043F\u043E\u0440\u043E\u0436\u043D\u0454"
      ],
      "Image must be smaller than 1 MB": [
        "\u0417\u043E\u0431\u0440\u0430\u0436\u0435\u043D\u043D\u044F \u043F\u043E\u0432\u0438\u043D\u043D\u043E \u0431\u0443\u0442\u0438 \u043C\u0435\u043D\u0448\u0438\u043C \u0437\u0430 1 \u041C\u0411"
      ],
      "Add": [
        "\u0414\u043E\u0434\u0430\u0442\u0438"
      ],
      "Remove": [
        "\u0412\u0438\u0434\u0430\u043B\u0438\u0442\u0438"
      ],
      "Country": [
        "\u041A\u0440\u0430\u0457\u043D\u0430"
      ],
      "Building number": [
        "\u041D\u043E\u043C\u0435\u0440 \u0431\u0443\u0434\u0438\u043D\u043A\u0443"
      ],
      "Building name": [
        "\u041D\u0430\u0437\u0432\u0430 \u0431\u0443\u0434\u0438\u043D\u043A\u0443"
      ],
      "Street": [
        "\u0412\u0443\u043B\u0438\u0446\u044F"
      ],
      "Post code": [
        "\u041F\u043E\u0448\u0442\u043E\u0432\u0438\u0439 \u0456\u043D\u0434\u0435\u043A\u0441"
      ],
      "Town location": [
        "\u041E\u0431\u043B\u0430\u0441\u0442\u044C \u043C\u0456\u0441\u0442\u0430"
      ],
      "District": [
        "\u0420\u0430\u0439\u043E\u043D"
      ],
      "Country subdivision": [
        "\u0420\u0435\u0433\u0456\u043E\u043D \u043A\u0440\u0430\u0457\u043D\u0438"
      ],
      "Additional address lines": [
        "\u0414\u043E\u0434\u0430\u0442\u043A\u043E\u0432\u0430 \u0456\u043D\u0444\u043E\u0440\u043C\u0430\u0446\u0456\u044F"
      ],
      "Orders": [
        "\u0417\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Inventory": [
        ""
      ],
      "Categories": [
        ""
      ],
      "Wire transfers": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "Templates": [
        "\u0428\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u0434\u0456\u0457"
      ],
      "OTP Devices": [
        ""
      ],
      "Webhooks": [
        "\u0412\u0435\u0431\u0445\u0443\u043A\u0438"
      ],
      "Settings": [
        "\u041D\u0430\u043B\u0430\u0448\u0442\u0443\u0432\u0430\u043D\u043D\u044F"
      ],
      "Password": [
        ""
      ],
      "Access tokens": [
        "\u0422\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Connection": [
        "\u0417'\u0454\u0434\u043D\u0430\u043D\u043D\u044F"
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        "\u0406\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "New": [
        "\u041D\u043E\u0432\u0438\u0439"
      ],
      "List": [
        "\u0421\u043F\u0438\u0441\u043E\u043A"
      ],
      "Log out": [
        "\u0412\u0438\u0439\u0442\u0438"
      ],
      "Username": [
        ""
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        "\u041D\u0430\u0437\u0432\u0430 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457 \u0432 URL. \u0406\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044F 'default' \u0454 \u043E\u0441\u043E\u0431\u043B\u0438\u0432\u043E\u044E, \u043E\u0441\u043A\u0456\u043B\u044C\u043A\u0438 \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u043E\u0432\u0443\u0454\u0442\u044C\u0441\u044F \u0434\u043B\u044F \u0430\u0434\u043C\u0456\u043D\u0456\u0441\u0442\u0440\u0443\u0432\u0430\u043D\u043D\u044F \u0456\u043D\u0448\u0438\u0445 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0439."
      ],
      "Business name": [
        "\u041D\u0430\u0437\u0432\u0430 \u0431\u0456\u0437\u043D\u0435\u0441\u0443"
      ],
      "Legal name of the business represented by this instance.": [
        "\u042E\u0440\u0438\u0434\u0438\u0447\u043D\u0430 \u043D\u0430\u0437\u0432\u0430 \u0431\u0456\u0437\u043D\u0435\u0441\u0443, \u044F\u043A\u0438\u0439 \u043F\u0440\u0435\u0434\u0441\u0442\u0430\u0432\u043B\u044F\u0454 \u0446\u044F \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044F."
      ],
      "Email": [
        "Email"
      ],
      "Contact email": [
        "\u041A\u043E\u043D\u0442\u0430\u043A\u0442\u043D\u0438\u0439 email"
      ],
      "Phone number": [
        "\u041D\u043E\u043C\u0435\u0440 \u043C\u0430\u0440\u0448\u0440\u0443\u0442\u0438\u0437\u0430\u0446\u0456\u0457."
      ],
      "Contact phone": [
        "\u041A\u043E\u043D\u0442\u0430\u043A\u0442\u043D\u0438\u0439 email"
      ],
      "Website URL": [
        "URL \u0432\u0435\u0431\u0441\u0430\u0439\u0442\u0443"
      ],
      "URL.": [
        "URL."
      ],
      "Logo": [
        "\u041B\u043E\u0433\u043E\u0442\u0438\u043F"
      ],
      "Logo image.": [
        "\u0417\u043E\u0431\u0440\u0430\u0436\u0435\u043D\u043D\u044F \u043B\u043E\u0433\u043E\u0442\u0438\u043F\u0443."
      ],
      "Physical location of the merchant.": [
        "\u0424\u0456\u0437\u0438\u0447\u043D\u0435 \u0440\u043E\u0437\u0442\u0430\u0448\u0443\u0432\u0430\u043D\u043D\u044F \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u044F."
      ],
      "Jurisdiction": [
        "\u042E\u0440\u0438\u0441\u0434\u0438\u043A\u0446\u0456\u044F"
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        "\u042E\u0440\u0438\u0441\u0434\u0438\u043A\u0446\u0456\u044F \u0434\u043B\u044F \u043F\u0440\u0430\u0432\u043E\u0432\u0438\u0445 \u0441\u043F\u043E\u0440\u0456\u0432 \u0437 \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u0435\u043C."
      ],
      "Pay transaction fee": [
        ""
      ],
      "Cover the transaction cost or pass it on to the user.": [
        ""
      ],
      "Default payment delay": [
        "\u0417\u0430\u0442\u0440\u0438\u043C\u043A\u0430 \u043E\u043F\u043B\u0430\u0442\u0438 \u0437\u0430 \u0437\u0430\u043C\u043E\u0432\u0447\u0443\u0432\u0430\u043D\u043D\u044F\u043C"
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        "\u0427\u0430\u0441, \u044F\u043A\u0438\u0439 \u043C\u0430\u044E\u0442\u044C \u043A\u043B\u0456\u0454\u043D\u0442\u0438 \u0434\u043B\u044F \u043E\u043F\u043B\u0430\u0442\u0438 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0434\u043E \u0437\u0430\u043A\u0456\u043D\u0447\u0435\u043D\u043D\u044F \u0442\u0435\u0440\u043C\u0456\u043D\u0443 \u0434\u0456\u0457 \u043F\u0440\u043E\u043F\u043E\u0437\u0438\u0446\u0456\u0457 \u0437\u0430 \u0437\u0430\u043C\u043E\u0432\u0447\u0443\u0432\u0430\u043D\u043D\u044F\u043C."
      ],
      "Default refund delay": [
        "\u0417\u0430\u0442\u0440\u0438\u043C\u043A\u0430 \u043E\u043F\u043B\u0430\u0442\u0438 \u0437\u0430 \u0437\u0430\u043C\u043E\u0432\u0447\u0443\u0432\u0430\u043D\u043D\u044F\u043C"
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        "\u0417\u0430\u0442\u0440\u0438\u043C\u043A\u0430 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443 \u0437\u0430 \u0437\u0430\u043C\u043E\u0432\u0447\u0443\u0432\u0430\u043D\u043D\u044F\u043C"
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        "\u041C\u0430\u043A\u0441\u0438\u043C\u0430\u043B\u044C\u043D\u0438\u0439 \u0447\u0430\u0441, \u043D\u0430 \u044F\u043A\u0438\u0439 \u043E\u0431\u043C\u0456\u043D\u043D\u0438\u043A \u043C\u043E\u0436\u0435 \u0437\u0430\u0442\u0440\u0438\u043C\u0430\u0442\u0438 \u043F\u0435\u0440\u0435\u043A\u0430\u0437 \u043A\u043E\u0448\u0442\u0456\u0432 \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u044E, \u0434\u043E\u0437\u0432\u043E\u043B\u044F\u044E\u0447\u0438 \u0439\u043E\u043C\u0443 \u043E\u0431'\u0454\u0434\u043D\u0443\u0432\u0430\u0442\u0438 \u043C\u0435\u043D\u0448\u0456 \u043F\u043B\u0430\u0442\u0435\u0436\u0456 \u0443 \u0431\u0456\u043B\u044C\u0448\u0456 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0456 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0438 \u0442\u0430 \u0437\u043D\u0438\u0436\u0443\u044E\u0447\u0438 \u043A\u043E\u043C\u0456\u0441\u0456\u0457 \u0437\u0430 \u043F\u0435\u0440\u0435\u043A\u0430\u0437."
      ],
      "Default wire transfer rounding interval": [
        "\u0417\u0430\u0442\u0440\u0438\u043C\u043A\u0430 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443 \u0437\u0430 \u0437\u0430\u043C\u043E\u0432\u0447\u0443\u0432\u0430\u043D\u043D\u044F\u043C"
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        "\u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "To second": [
        "\u0441\u0435\u043A\u0443\u043D\u0434\u0438"
      ],
      "To minute": [
        "\u0445\u0432\u0438\u043B\u0438\u043D\u0438"
      ],
      "To hour": [
        "\u0433\u043E\u0434\u0438\u043D\u0438"
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        "\u0417\u0430\u043A\u0456\u043D\u0447\u0443\u0454\u0442\u044C\u0441\u044F \u043E"
      ],
      "Required": [
        "\u043E\u0431\u043E\u0432\u02BC\u044F\u0437\u043A\u043E\u0432\u043E"
      ],
      "Unauthorized": [
        "\u0410\u0432\u0442\u043E\u0440\u0438\u0437\u043E\u0432\u0430\u043D\u043E"
      ],
      "Sending the code failed": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u043F\u0456\u0434\u043A\u043B\u044E\u0447\u0438\u0442\u0438\u0441\u044F \u0434\u043E \u0441\u0435\u0440\u0432\u0435\u0440\u0430."
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        "\u0410\u043B\u0433\u043E\u0440\u0438\u0442\u043C \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0438"
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        "\u041D\u0430\u0437\u0430\u0434"
      ],
      "Verify": [
        "\u041F\u0435\u0440\u0435\u0432\u0456\u0440\u0435\u043D\u043E"
      ],
      "Failed to send the verification code.": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044E"
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        ""
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        ""
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        "\u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "Login required": [
        "\u041F\u043E\u0442\u0440\u0456\u0431\u0435\u043D \u0432\u0445\u0456\u0434"
      ],
      "instance name": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "current password": [
        "\u0412\u0441\u0442\u0430\u043D\u043E\u0432\u0438\u0442\u0438 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Forgot password": [
        ""
      ],
      "Create new account": [
        "\u0417\u0430\u0440\u0430\u0445\u043E\u0432\u0430\u043D\u0438\u0439 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "Invalid": [
        "\u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u0439"
      ],
      "Doesn't have the pattern of an email": [
        "URL \u043C\u0430\u0454 \u043D\u0435\u043F\u0440\u0430\u0432\u0438\u043B\u044C\u043D\u0438\u0439 \u0444\u043E\u0440\u043C\u0430\u0442"
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        "\u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u0439 \u0444\u043E\u0440\u043C\u0430\u0442"
      ],
      "Max 7 lines": [
        "\u043C\u0430\u043A\u0441\u0438\u043C\u0443\u043C 7 \u0440\u044F\u0434\u043A\u0456\u0432"
      ],
      "Doesn't match": [
        ""
      ],
      "Instance created": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Unauthorized.": [
        "\u0410\u0432\u0442\u043E\u0440\u0438\u0437\u043E\u0432\u0430\u043D\u043E"
      ],
      "Conflict.": [
        "\u041A\u043E\u043D\u0444\u043B\u0456\u043A\u0442."
      ],
      "Not found.": [
        "\u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "New password": [
        ""
      ],
      "Next password to be used": [
        "\u043D\u0430\u0441\u0442\u0443\u043F\u043D\u0438\u0439 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443, \u044F\u043A\u0438\u0439 \u0431\u0443\u0434\u0435 \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043E"
      ],
      "Repeat password": [
        "\u041F\u043E\u0432\u0442\u043E\u0440\u0456\u0442\u044C \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Confirm the same password": [
        "\u043F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0438\u0442\u0438 \u0442\u043E\u0439 \u0441\u0430\u043C\u0438\u0439 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Please complete the marked fields and choose authorization method": [
        "\u041D\u0435\u043E\u0431\u0445\u0456\u0434\u043D\u043E \u0437\u0430\u043F\u043E\u0432\u043D\u0438\u0442\u0438 \u043F\u043E\u0437\u043D\u0430\u0447\u0435\u043D\u0456 \u043F\u043E\u043B\u044F \u0442\u0430 \u0432\u0438\u0431\u0440\u0430\u0442\u0438 \u043C\u0435\u0442\u043E\u0434 \u0430\u0432\u0442\u043E\u0440\u0438\u0437\u0430\u0446\u0456\u0457"
      ],
      "Confirm operation": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0436\u0435\u043D\u043E"
      ],
      "Check the password.": [
        ""
      ],
      "Instance not found.": [
        "\u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "Description": [
        "\u041E\u043F\u0438\u0441"
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u0434\u0456\u0457"
      ],
      "Time the access token will be valid.": [
        "\u041D\u0430\u0434\u0430\u043D\u0438\u0439 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443 \u0454 \u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u043C."
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        "\u0412\u0438\u0431\u0435\u0440\u0456\u0442\u044C \u043E\u0434\u043D\u0443..."
      ],
      "Read only": [
        ""
      ],
      "All": [
        "\u0412\u0441\u0456"
      ],
      "Order simple": [
        "\u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Order Point-of-Sale": [
        "\u041A\u043B\u044E\u0447 \u0442\u043E\u0447\u043A\u0438 \u043F\u0440\u043E\u0434\u0430\u0436\u0443"
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        "\u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        "\u0412\u0441\u0442\u0430\u043D\u043E\u0432\u0438\u0442\u0438 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Please complete the marked fields": [
        "\u041D\u0435\u043E\u0431\u0445\u0456\u0434\u043D\u043E \u0437\u0430\u043F\u043E\u0432\u043D\u0438\u0442\u0438 \u043F\u043E\u0437\u043D\u0430\u0447\u0435\u043D\u0456 \u043F\u043E\u043B\u044F"
      ],
      "Access token created": [
        "\u0422\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        ""
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        "\u0406\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044F 'default' \u0449\u0435 \u043D\u0435 \u043D\u0430\u043B\u0430\u0448\u0442\u043E\u0432\u0430\u043D\u0430."
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        "\u0421\u0442\u0432\u043E\u0440\u0456\u0442\u044C \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044E 'default', \u0449\u043E\u0431 \u043F\u043E\u0447\u0430\u0442\u0438 \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u043E\u0432\u0443\u0432\u0430\u0442\u0438 \u0431\u0435\u043A\u043E\u0444\u0456\u0441 \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u044F."
      ],
      "Create access token": [
        "\u0412\u0441\u0442\u0430\u043D\u043E\u0432\u0438\u0442\u0438 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Load more devices before the first one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0448\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u0434\u043E \u043F\u0435\u0440\u0448\u043E\u0433\u043E"
      ],
      "Load first page": [
        ""
      ],
      "Created at": [
        "\u0421\u0442\u0432\u043E\u0440\u0435\u043D\u043E \u043E"
      ],
      "Expires at": [
        "\u0417\u0430\u043A\u0456\u043D\u0447\u0443\u0454\u0442\u044C\u0441\u044F \u043E"
      ],
      "Never": [
        "\u043D\u0456\u043A\u043E\u043B\u0438"
      ],
      "Remove this access token": [
        "\u041F\u043E\u0432\u0442\u043E\u0440\u0456\u0442\u044C \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Delete": [
        "\u0412\u0438\u0434\u0430\u043B\u0438\u0442\u0438"
      ],
      "Load more devices after the last one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0448\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u043F\u0456\u0441\u043B\u044F \u043E\u0441\u0442\u0430\u043D\u043D\u044C\u043E\u0433\u043E"
      ],
      "Load next page": [
        ""
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        "\u041F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        "\u0412\u0441\u0442\u0430\u043D\u043E\u0432\u0438\u0442\u0438 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "Deleting an access token cannot be undone.": [
        "\u043D\u0430\u0441\u0442\u0443\u043F\u043D\u0438\u0439 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443, \u044F\u043A\u0438\u0439 \u0431\u0443\u0434\u0435 \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043E"
      ],
      "This is not a valid Bitcoin address.": [
        "\u0426\u0435 \u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0430 \u0430\u0434\u0440\u0435\u0441\u0430 \u0431\u0456\u0442\u043A\u043E\u0439\u043D."
      ],
      "This is not a valid Ethereum address.": [
        "\u0426\u0435 \u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0430 \u0430\u0434\u0440\u0435\u0441\u0430 Ethereum."
      ],
      "This is not a valid host.": [
        "\u0426\u0435 \u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0430 \u0430\u0434\u0440\u0435\u0441\u0430 \u0431\u0456\u0442\u043A\u043E\u0439\u043D."
      ],
      "IBANs usually have more than 4 digits.": [
        "\u041D\u043E\u043C\u0435\u0440\u0430 IBAN \u0437\u0430\u0437\u0432\u0438\u0447\u0430\u0439 \u043C\u0430\u044E\u0442\u044C \u0431\u0456\u043B\u044C\u0448\u0435 4-\u044C\u043E\u0445 \u0446\u0438\u0444\u0440"
      ],
      "IBANs usually have fewer than 34 digits.": [
        "\u041D\u043E\u043C\u0435\u0440\u0430 IBAN \u0437\u0430\u0437\u0432\u0438\u0447\u0430\u0439 \u043C\u0430\u044E\u0442\u044C \u043C\u0435\u043D\u0448\u0435 34-\u044C\u043E\u0445 \u0446\u0438\u0444\u0440"
      ],
      "The IBAN's country code could not be retrieved.": [
        "\u041A\u043E\u0434 \u043A\u0440\u0430\u0457\u043D\u0438 IBAN \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        "\u041D\u043E\u043C\u0435\u0440 IBAN \u043D\u0435 \u043A\u043E\u0440\u0435\u043A\u0442\u043D\u0438\u0439, \u043A\u043E\u043D\u0442\u0440\u043E\u043B\u044C\u043D\u0430 \u0441\u0443\u043C\u0430 \u043D\u0435 \u0441\u0445\u043E\u0434\u0438\u0442\u044C\u0441\u044F"
      ],
      "This account is not allowed.": [
        "\u0426\u0435 \u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0430 \u0430\u0434\u0440\u0435\u0441\u0430 \u0431\u0456\u0442\u043A\u043E\u0439\u043D."
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        "\u041C\u0435\u0442\u043E\u0434 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        ""
      ],
      "Select a wire method...": [
        "\u0412\u0438\u0431\u0435\u0440\u0456\u0442\u044C \u043E\u0434\u0438\u043D \u043C\u0435\u0442\u043E\u0434 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "Routing": [
        "\u041C\u0430\u0440\u0448\u0440\u0443\u0442\u0438\u0437\u0430\u0446\u0456\u044F"
      ],
      "Routing number": [
        "\u041D\u043E\u043C\u0435\u0440 \u043C\u0430\u0440\u0448\u0440\u0443\u0442\u0438\u0437\u0430\u0446\u0456\u0457."
      ],
      "Account": [
        "\u0420\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "Account number": [
        "\u041D\u043E\u043C\u0435\u0440 \u0440\u0430\u0445\u0443\u043D\u043A\u0443."
      ],
      "Code": [
        "\u041A\u043E\u0434"
      ],
      "Business Identifier Code": [
        "\u041A\u043E\u0434 \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0446\u0456\u0457 \u0431\u0456\u0437\u043D\u0435\u0441\u0443."
      ],
      "International Bank Account Number": [
        "\u041D\u043E\u043C\u0435\u0440 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u0440\u0430\u0445\u0443\u043D\u043A\u0443."
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        "\u0423\u043D\u0456\u0444\u0456\u043A\u043E\u0432\u0430\u043D\u0438\u0439 \u0456\u043D\u0442\u0435\u0440\u0444\u0435\u0439\u0441 \u043F\u043B\u0430\u0442\u0435\u0436\u0456\u0432."
      ],
      "Bitcoin protocol": [
        "\u041F\u0440\u043E\u0442\u043E\u043A\u043E\u043B \u0431\u0456\u0442\u043A\u043E\u0439\u043D."
      ],
      "Ethereum protocol": [
        "\u041F\u0440\u043E\u0442\u043E\u043A\u043E\u043B Ethereum."
      ],
      "Interledger protocol": [
        "\u041F\u0440\u043E\u0442\u043E\u043A\u043E\u043B Interledger."
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        "\u041D\u0430\u0437\u0432\u0430 \u043A\u0430\u0442\u0435\u0433\u043E\u0440\u0456\u0457"
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        "\u041E\u0444\u0456\u0446\u0456\u0439\u043D\u0435 \u0456\u043C'\u044F \u0432\u043B\u0430\u0441\u043D\u0438\u043A\u0430 \u0440\u0430\u0445\u0443\u043D\u043A\u0443."
      ],
      "Invalid url": [
        "\u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u0439 \u0444\u043E\u0440\u043C\u0430\u0442"
      ],
      "URL must end with a '/'": [
        ""
      ],
      "URL must not contain params": [
        ""
      ],
      "URL must not hash param": [
        ""
      ],
      "Conflict": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0438\u0442\u0438"
      ],
      'Server replied with "bad request".': [
        "\u0412\u0456\u0434\u043F\u043E\u0432\u0456\u0434\u044C \u0441\u0435\u0440\u0432\u0435\u0440\u0430 \u0437 \u043A\u043E\u0434\u043E\u043C \u043F\u043E\u043C\u0438\u043B\u043A\u0438."
      ],
      "Unauthorized, check credentials.": [
        ""
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        ""
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        ""
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        "\u0420\u0430\u0445\u0443\u043D\u043E\u043A:"
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        ""
      ],
      "Endpoint URL": [
        ""
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        ""
      ],
      "Auth type": [
        "\u0422\u0438\u043F \u043F\u043E\u0434\u0456\u0457"
      ],
      "Choose the authentication type for the account info URL": [
        ""
      ],
      "Without authentication": [
        ""
      ],
      "With username and password": [
        ""
      ],
      "With token": [
        "\u0421\u0442\u0432\u043E\u0440\u0438\u0442\u0438 \u0442\u043E\u043A\u0435\u043D"
      ],
      "Do not change": [
        "URL \u043E\u0431\u043C\u0456\u043D\u043D\u0438\u043A\u0430"
      ],
      "Username to access the account information.": [
        ""
      ],
      "Password to access the account information.": [
        ""
      ],
      "Access token to access the account information.": [
        ""
      ],
      "Match": [
        ""
      ],
      "Check where the information match against the server info.": [
        ""
      ],
      "Compare info from server with account form": [
        ""
      ],
      "Test": [
        ""
      ],
      "Server didn't like the request we made.": [
        "\u0412\u0456\u0434\u043F\u043E\u0432\u0456\u0434\u044C \u0441\u0435\u0440\u0432\u0435\u0440\u0430 \u0437 \u043A\u043E\u0434\u043E\u043C \u043F\u043E\u043C\u0438\u043B\u043A\u0438."
      ],
      "Account details": [
        "\u0410\u0434\u0440\u0435\u0441\u0430 \u0440\u0430\u0445\u0443\u043D\u043A\u0443"
      ],
      "Check if the information matches the server info.": [
        ""
      ],
      "The bank account has been deleted.": [
        "\u0411\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A \u0443\u0441\u043F\u0456\u0448\u043D\u043E \u0432\u0438\u0434\u0430\u043B\u0435\u043D\u043E."
      ],
      "Bank accounts": [
        "\u0411\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "Add new account": [
        "\u0417\u0430\u0440\u0430\u0445\u043E\u0432\u0430\u043D\u0438\u0439 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "Owner's name": [
        "\u041D\u0430\u0437\u0432\u0430 \u0431\u0456\u0437\u043D\u0435\u0441\u0443"
      ],
      "Delete selected accounts from the database": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0438\u0431\u0440\u0430\u043D\u0438\u0439 \u043F\u0435\u0440\u0435\u043A\u0430\u0437 \u0437 \u0431\u0430\u0437\u0438 \u0434\u0430\u043D\u0438\u0445"
      ],
      "There are no accounts yet, add more pressing the + sign": [
        "\u041F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "You must provide a bank account to receive payments.": [
        "URI, \u0449\u043E \u0432\u043A\u0430\u0437\u0443\u0454 \u043D\u0430 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A \u0434\u043B\u044F \u0437\u0430\u0440\u0430\u0445\u0443\u0432\u0430\u043D\u043D\u044F \u0434\u043E\u0445\u043E\u0434\u0443."
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        ""
      ],
      "Name": [
        "\u041D\u0430\u0437\u0432\u0430"
      ],
      "Category name": [
        ""
      ],
      "Category deleted": [
        "\u0448\u0430\u0431\u043B\u043E\u043D \u0443\u0441\u043F\u0456\u0448\u043D\u043E \u0432\u0438\u0434\u0430\u043B\u0435\u043D\u043E"
      ],
      "Add new devices": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0438\u0439 \u0440\u0435\u0437\u0435\u0440\u0432"
      ],
      "ID": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440"
      ],
      "Total products": [
        "\u0417\u0430\u0433\u0430\u043B\u044C\u043D\u0430 \u0446\u0456\u043D\u0430"
      ],
      "Delete selected category from the database": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0438\u0431\u0440\u0430\u043D\u0438\u0439 \u043F\u0435\u0440\u0435\u043A\u0430\u0437 \u0437 \u0431\u0430\u0437\u0438 \u0434\u0430\u043D\u0438\u0445"
      ],
      "There are no categories yet, add more pressing the + sign": [
        "\u0428\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "Id:": [
        ""
      ],
      "Name of the category": [
        "\u041D\u0430\u0437\u0432\u0430 \u043A\u0430\u0442\u0435\u0433\u043E\u0440\u0456\u0457"
      ],
      "Products": [
        "\u0422\u043E\u0432\u0430\u0440\u0438"
      ],
      "Image": [
        "\u0417\u043E\u0431\u0440\u0430\u0436\u0435\u043D\u043D\u044F"
      ],
      "Load more products after the last one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0448\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u043F\u0456\u0441\u043B\u044F \u043E\u0441\u0442\u0430\u043D\u043D\u044C\u043E\u0433\u043E"
      ],
      "There are no products in this category.": [
        "\u041A\u0435\u0440\u0443\u0432\u0430\u043D\u043D\u044F \u0441\u043F\u0438\u0441\u043A\u043E\u043C \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0443 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u0456."
      ],
      "Account's KYC status": [
        "\u0410\u0434\u0440\u0435\u0441\u0430 \u0440\u0430\u0445\u0443\u043D\u043A\u0443"
      ],
      "Exchange": [
        "Exchange"
      ],
      "Status": [
        "URL \u0441\u0442\u0430\u0442\u0443\u0441\u0443"
      ],
      "Ok": [
        ""
      ],
      "Action required": [
        "\u041F\u043E\u0442\u0440\u0456\u0431\u0435\u043D \u0432\u0445\u0456\u0434"
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        "\u0421\u0443\u043C\u0430 \u043E\u0431\u043E\u0432'\u044F\u0437\u043A\u043E\u0432\u0430"
      ],
      "More information required.": [
        "\u0434\u043E\u0434\u0430\u0442\u043A\u043E\u0432\u0430 \u0456\u043D\u0444\u043E\u0440\u043C\u0430\u0446\u0456\u044F \u0434\u043B\u044F \u043D\u0430\u0434\u0430\u043D\u043D\u044F \u043A\u043E\u043D\u0442\u0435\u043A\u0441\u0442\u0443"
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        "\u0417\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0435\u043D\u043D\u044F..."
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        "HTTP \u0441\u0442\u0430\u0442\u0443\u0441"
      ],
      "No pending kyc verification!": [
        "\u041D\u0435\u043C\u0430\u0454 \u043E\u0447\u0456\u043A\u0443\u0432\u0430\u043D\u0438\u0445 \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u043E\u043A KYC!"
      ],
      "Wire account is invalid": [
        "\u041D\u0430\u0434\u0430\u043D\u0438\u0439 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443 \u0454 \u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u043C"
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        ""
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        "\u043E\u0431\u043E\u0432\u02BC\u044F\u0437\u043A\u043E\u0432\u043E"
      ],
      "The payment provider requires more information.": [
        ""
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        "\u0412\u0438 \u0432\u0441\u0442\u0430\u043D\u043E\u0432\u043B\u044E\u0454\u0442\u0435 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443 \u0434\u043B\u044F \u043D\u043E\u0432\u043E\u0457 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F, \u0434\u0435 \u043A\u043B\u0456\u0454\u043D\u0442\u0438 \u0437\u0430\u043F\u043B\u0430\u0442\u0438\u043B\u0438, \u0430\u043B\u0435 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0456 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0438 \u0432\u0456\u0434 \u043F\u043E\u0441\u0442\u0430\u0447\u0430\u043B\u044C\u043D\u0438\u043A\u0430 \u043F\u043B\u0430\u0442\u0435\u0436\u0456\u0432 \u0449\u0435 \u043D\u0435 \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u0456"
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        ""
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        "\u0437\u043C\u0456\u043D\u0438\u0442\u0438 \u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F \u043D\u0430 \u043D\u0435\u0432\u0456\u0434\u043E\u043C\u0443 \u0434\u0430\u0442\u0443"
      ],
      "Enter description or id": [
        "\u0412\u0432\u0435\u0434\u0456\u0442\u044C \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "No match found for that description or ID.": [
        "\u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0438 \u0437 \u0442\u0430\u043A\u0438\u043C \u043E\u043F\u0438\u0441\u043E\u043C \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "You must enter a valid product identifier.": [
        "\u0412\u0438 \u043F\u043E\u0432\u0438\u043D\u043D\u0456 \u0432\u0432\u0435\u0441\u0442\u0438 \u0434\u0456\u0439\u0441\u043D\u0438\u0439 \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443."
      ],
      "Quantity must be greater than 0!": [
        "\u041A\u0456\u043B\u044C\u043A\u0456\u0441\u0442\u044C \u043C\u0430\u0454 \u0431\u0443\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 0!"
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        "\u0426\u044F \u043A\u0456\u043B\u044C\u043A\u0456\u0441\u0442\u044C \u043F\u0435\u0440\u0435\u0432\u0438\u0449\u0443\u0454 \u0437\u0430\u043B\u0438\u0448\u043E\u043A \u043D\u0430 \u0441\u043A\u043B\u0430\u0434\u0456. \u041D\u0430\u0440\u0430\u0437\u0456 \u043D\u0430 \u0441\u043A\u043B\u0430\u0434\u0456 \u0437\u0430\u043B\u0438\u0448\u0438\u043B\u043E\u0441\u044F \u043B\u0438\u0448\u0435 %1$s \u043E\u0434\u0438\u043D\u0438\u0446\u044C, \u044F\u043A\u0456 \u043D\u0435 \u0437\u0430\u0440\u0435\u0437\u0435\u0440\u0432\u043E\u0432\u0430\u043D\u0456."
      ],
      "Search product": [
        "\u043D\u043E\u0432\u0430 \u0446\u0456\u043D\u0430 \u0434\u043B\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443"
      ],
      "Quantity": [
        "\u041A\u0456\u043B\u044C\u043A\u0456\u0441\u0442\u044C"
      ],
      "How many products will be added": [
        "\u0441\u043A\u0456\u043B\u044C\u043A\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0431\u0443\u0434\u0435 \u0434\u043E\u0434\u0430\u043D\u043E"
      ],
      "Add from inventory": [
        "\u0414\u043E\u0434\u0430\u0442\u0438 \u0437\u0456 \u0441\u043A\u043B\u0430\u0434\u0443"
      ],
      "This product has %1$s applicable taxes configured.": [
        ""
      ],
      "No taxes configured for this product.": [
        "\u041F\u043E\u0434\u0430\u0442\u043A\u0438 \u0434\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 \u043D\u0435 \u043D\u0430\u043B\u0430\u0448\u0442\u043E\u0432\u0430\u043D\u0456."
      ],
      "Amount": [
        "\u0421\u0443\u043C\u0430"
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        "\u041F\u043E\u0434\u0430\u0442\u043A\u0438 \u043C\u043E\u0436\u0443\u0442\u044C \u0431\u0443\u0442\u0438 \u0432 \u0432\u0430\u043B\u044E\u0442\u0430\u0445, \u0449\u043E \u0432\u0456\u0434\u0440\u0456\u0437\u043D\u044F\u044E\u0442\u044C\u0441\u044F \u0432\u0456\u0434 \u043E\u0441\u043D\u043E\u0432\u043D\u043E\u0457 \u0432\u0430\u043B\u044E\u0442\u0438, \u044F\u043A\u0443 \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u043E\u0432\u0443\u0454 \u043F\u0440\u043E\u0434\u0430\u0432\u0435\u0446\u044C."
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        "\u0412\u0432\u0435\u0434\u0456\u0442\u044C \u0432\u0430\u043B\u044E\u0442\u0443 \u0442\u0430 \u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F, \u0440\u043E\u0437\u0434\u0456\u043B\u0435\u043D\u0456 \u0434\u0432\u043E\u043A\u0440\u0430\u043F\u043A\u043E\u044E, \u043D\u0430\u043F\u0440\u0438\u043A\u043B\u0430\u0434, &quot;USD:2.3&quot;."
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        "\u041E\u0444\u0456\u0446\u0456\u0439\u043D\u0430 \u043D\u0430\u0437\u0432\u0430 \u043F\u043E\u0434\u0430\u0442\u043A\u0443, \u043D\u0430\u043F\u0440\u0438\u043A\u043B\u0430\u0434, \u041F\u0414\u0412 \u0430\u0431\u043E \u0456\u043C\u043F\u043E\u0440\u0442\u043D\u0456 \u043C\u0438\u0442\u0430."
      ],
      "Add tax to the tax list": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043F\u043E\u0434\u0430\u0442\u043E\u043A \u0434\u043E \u0441\u043F\u0438\u0441\u043A\u0443 \u043F\u043E\u0434\u0430\u0442\u043A\u0456\u0432"
      ],
      "Describe and add a product that is not in the inventory list": [
        "\u043E\u043F\u0438\u0448\u0456\u0442\u044C \u0456 \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u043F\u0440\u043E\u0434\u0443\u043A\u0442, \u044F\u043A\u043E\u0433\u043E \u043D\u0435\u043C\u0430\u0454 \u0432 \u0441\u043F\u0438\u0441\u043A\u0443 \u0456\u043D\u0432\u0435\u043D\u0442\u0430\u0440\u044E"
      ],
      "Add custom product": [
        "\u0414\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0438\u0439 \u043F\u0440\u043E\u0434\u0443\u043A\u0442"
      ],
      "Complete information of the product": [
        "\u041F\u043E\u0432\u043D\u0430 \u0456\u043D\u0444\u043E\u0440\u043C\u0430\u0446\u0456\u044F \u043F\u0440\u043E \u043F\u0440\u043E\u0434\u0443\u043A\u0442"
      ],
      "Must be a number.": [
        "\u043D\u0435 \u0454 \u0447\u0438\u0441\u043B\u043E\u043C"
      ],
      "Must be greater than 0.": [
        "\u043C\u0430\u0454 \u0431\u0443\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 0"
      ],
      "Photo of the product.": [
        "\u0444\u043E\u0442\u043E \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443."
      ],
      "Full product description.": [
        "\u043F\u043E\u0432\u043D\u0438\u0439 \u043E\u043F\u0438\u0441 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443."
      ],
      "Unit name": [
        "\u041E\u0434\u0438\u043D\u0438\u0446\u044F"
      ],
      "Name of the product unit.": [
        "\u043D\u0430\u0437\u0432\u0430 \u043E\u0434\u0438\u043D\u0438\u0446\u0456 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443."
      ],
      "Price per unit": [
        ""
      ],
      "Amount in the current currency.": [
        "\u0441\u0443\u043C\u0430 \u0432 \u043F\u043E\u0442\u043E\u0447\u043D\u0456\u0439 \u0432\u0430\u043B\u044E\u0442\u0456."
      ],
      "How many products will be added.": [
        "\u0441\u043A\u0456\u043B\u044C\u043A\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0431\u0443\u0434\u0435 \u0434\u043E\u0434\u0430\u043D\u043E."
      ],
      "Taxes": [
        "\u041F\u043E\u0434\u0430\u0442\u043A\u0438"
      ],
      "Total price": [
        "\u0417\u0430\u0433\u0430\u043B\u044C\u043D\u0430 \u0446\u0456\u043D\u0430"
      ],
      "Must be greater than 0": [
        "\u043C\u0430\u0454 \u0431\u0443\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 0"
      ],
      "Must have a refund deadline": [
        "\u043F\u043E\u0432\u0438\u043D\u0435\u043D \u0431\u0443\u0442\u0438 \u0432\u0441\u0442\u0430\u043D\u043E\u0432\u043B\u0435\u043D\u0438\u0439 \u0442\u0435\u0440\u043C\u0456\u043D \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Auto refund can't be after refund deadline": [
        "\u0430\u0432\u0442\u043E\u043C\u0430\u0442\u0438\u0447\u043D\u0435 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u043D\u0435 \u043C\u043E\u0436\u0435 \u0431\u0443\u0442\u0438 \u043F\u0456\u0441\u043B\u044F \u0442\u0435\u0440\u043C\u0456\u043D\u0443 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Must be in the future": [
        "\u043F\u043E\u0432\u0438\u043D\u043D\u043E \u0431\u0443\u0442\u0438 \u0432 \u043C\u0430\u0439\u0431\u0443\u0442\u043D\u044C\u043E\u043C\u0443"
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        "\u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u0439"
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        ""
      ],
      "Manage products in order": [
        "\u041A\u0435\u0440\u0443\u0432\u0430\u043D\u043D\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0430\u043C\u0438 \u0432 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u0456"
      ],
      "%1$s products with a total price of %2$s.": [
        "%1$s \u0442\u043E\u0432\u0430\u0440\u0456\u0432 \u0456\u0437 \u0437\u0430\u0433\u0430\u043B\u044C\u043D\u043E\u044E \u0446\u0456\u043D\u043E\u044E %2$s."
      ],
      "Manage list of products in the order.": [
        "\u041A\u0435\u0440\u0443\u0432\u0430\u043D\u043D\u044F \u0441\u043F\u0438\u0441\u043A\u043E\u043C \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0443 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u0456."
      ],
      "Remove this product from the order.": [
        "\u0412\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0446\u0435\u0439 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u0456\u0437 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F."
      ],
      "Products price sum": [
        "\u0421\u043F\u0438\u0441\u043E\u043A \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432"
      ],
      "Total product price added up": [
        "\u0437\u0430\u0433\u0430\u043B\u044C\u043D\u0430 \u0441\u0443\u043C\u0430 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443"
      ],
      "Order price": [
        "\u0426\u0456\u043D\u0430 \u0437\u0430 \u043E\u0434\u0438\u043D\u0438\u0446\u044E"
      ],
      "Amount to be paid by the customer": [
        "\u0421\u0443\u043C\u0430, \u044F\u043A\u0443 \u043C\u0430\u0454 \u0441\u043F\u043B\u0430\u0442\u0438\u0442\u0438 \u043A\u043B\u0456\u0454\u043D\u0442"
      ],
      "Final order price": [
        "\u043A\u0456\u043D\u0446\u0435\u0432\u0430 \u0446\u0456\u043D\u0430 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Summary": [
        "\u041F\u0456\u0434\u0441\u0443\u043C\u043E\u043A"
      ],
      "Title of the order to be shown to the customer": [
        "\u041D\u0430\u0437\u0432\u0430 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F, \u044F\u043A\u0443 \u0431\u0443\u0434\u0435 \u043F\u043E\u043A\u0430\u0437\u0430\u043D\u043E \u043A\u043B\u0456\u0454\u043D\u0442\u0443"
      ],
      "Shipping and fulfillment": [
        "\u0414\u043E\u0441\u0442\u0430\u0432\u043A\u0430 \u0442\u0430 \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u043D\u044F"
      ],
      "Delivery date": [
        "\u0414\u0430\u0442\u0430 \u0434\u043E\u0441\u0442\u0430\u0432\u043A\u0438"
      ],
      "Deadline for physical delivery assured by the merchant.": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u0444\u0456\u0437\u0438\u0447\u043D\u043E\u0457 \u0434\u043E\u0441\u0442\u0430\u0432\u043A\u0438, \u0433\u0430\u0440\u0430\u043D\u0442\u043E\u0432\u0430\u043D\u0438\u0439 \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u0435\u043C."
      ],
      "Delivery location": [
        "\u0414\u0430\u0442\u0430 \u0434\u043E\u0441\u0442\u0430\u0432\u043A\u0438"
      ],
      "Address where the products will be delivered": [
        "\u0430\u0434\u0440\u0435\u0441\u0430, \u0437\u0430 \u044F\u043A\u043E\u044E \u0431\u0443\u0434\u0443\u0442\u044C \u0434\u043E\u0441\u0442\u0430\u0432\u043B\u0435\u043D\u0456 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0438"
      ],
      "Fulfillment URL": [
        "URL \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u043D\u044F"
      ],
      "URL to which the user will be redirected after successful payment.": [
        "URL, \u043D\u0430 \u044F\u043A\u0438\u0439 \u043A\u043E\u0440\u0438\u0441\u0442\u0443\u0432\u0430\u0447\u0430 \u0431\u0443\u0434\u0435 \u043F\u0435\u0440\u0435\u043D\u0430\u043F\u0440\u0430\u0432\u043B\u0435\u043D\u043E \u043F\u0456\u0441\u043B\u044F \u0443\u0441\u043F\u0456\u0448\u043D\u043E\u0457 \u043E\u043F\u043B\u0430\u0442\u0438."
      ],
      "Fulfillment message": [
        "URL \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u043D\u044F"
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        "\u041E\u043F\u0446\u0456\u0457 \u043E\u043F\u043B\u0430\u0442\u0438 Taler"
      ],
      "Override default Taler payment settings for this order": [
        "\u041F\u0435\u0440\u0435\u0432\u0438\u0437\u043D\u0430\u0447\u0438\u0442\u0438 \u0441\u0442\u0430\u043D\u0434\u0430\u0440\u0442\u043D\u0456 \u043D\u0430\u043B\u0430\u0448\u0442\u0443\u0432\u0430\u043D\u043D\u044F \u043E\u043F\u043B\u0430\u0442\u0438 Taler \u0434\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Payment time": [
        "\u0422\u0430\u0439\u043C-\u0430\u0443\u0442 \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        "\u0422\u0435\u0440\u043C\u0456\u043D, \u0434\u043E \u044F\u043A\u043E\u0433\u043E \u043A\u043B\u0456\u0454\u043D\u0442 \u043F\u043E\u0432\u0438\u043D\u0435\u043D \u043E\u043F\u043B\u0430\u0442\u0438\u0442\u0438 \u043F\u0440\u043E\u043F\u043E\u0437\u0438\u0446\u0456\u044E, \u043F\u0435\u0440\u0448 \u043D\u0456\u0436 \u0432\u043E\u043D\u0430 \u0437\u0430\u043A\u0456\u043D\u0447\u0438\u0442\u044C\u0441\u044F. \u041F\u0440\u043E\u0434\u0443\u043A\u0442\u0438 \u0437 \u0456\u043D\u0432\u0435\u043D\u0442\u0430\u0440\u044E \u0431\u0443\u0434\u0443\u0442\u044C \u0437\u0430\u0440\u0435\u0437\u0435\u0440\u0432\u043E\u0432\u0430\u043D\u0456 \u0434\u043E \u0446\u044C\u043E\u0433\u043E \u0442\u0435\u0440\u043C\u0456\u043D\u0443."
      ],
      "Default": [
        "\u0421\u0443\u043C\u0430 \u0437\u0430 \u0437\u0430\u043C\u043E\u0432\u0447\u0443\u0432\u0430\u043D\u043D\u044F\u043C"
      ],
      "Refund time": [
        "\u041F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u0437\u0434\u0456\u0439\u0441\u043D\u0435\u043D\u043E"
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        "\u0427\u0430\u0441, \u0434\u043E \u044F\u043A\u043E\u0433\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043C\u043E\u0436\u0435 \u0431\u0443\u0442\u0438 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u0435 \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u0435\u043C."
      ],
      "Wire transfer time": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        "\u0422\u0435\u0440\u043C\u0456\u043D, \u0434\u043E \u044F\u043A\u043E\u0433\u043E \u043E\u0431\u043C\u0456\u043D\u043D\u0438\u043A \u043F\u043E\u0432\u0438\u043D\u0435\u043D \u0437\u0434\u0456\u0439\u0441\u043D\u0438\u0442\u0438 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u043F\u0435\u0440\u0435\u043A\u0430\u0437."
      ],
      "Auto-refund time": [
        "\u0417\u0430\u0442\u0440\u0438\u043C\u043A\u0430 \u0430\u0432\u0442\u043E\u043C\u0430\u0442\u0438\u0447\u043D\u043E\u0433\u043E \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        "\u0427\u0430\u0441, \u0434\u043E \u044F\u043A\u043E\u0433\u043E \u0433\u0430\u043C\u0430\u043D\u0435\u0446\u044C \u0430\u0432\u0442\u043E\u043C\u0430\u0442\u0438\u0447\u043D\u043E \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u044F\u0442\u0438\u043C\u0435 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u043A\u043E\u0448\u0442\u0456\u0432 \u0431\u0435\u0437 \u0432\u0437\u0430\u0454\u043C\u043E\u0434\u0456\u0457 \u0437 \u043A\u043E\u0440\u0438\u0441\u0442\u0443\u0432\u0430\u0447\u0435\u043C."
      ],
      "Maximum fee": [
        "\u041C\u0430\u043A\u0441\u0438\u043C\u0430\u043B\u044C\u043D\u0430 \u043A\u043E\u043C\u0456\u0441\u0456\u044F \u0437\u0430 \u043F\u0435\u0440\u0435\u043A\u0430\u0437"
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        "\u041C\u0430\u043A\u0441\u0438\u043C\u0430\u043B\u044C\u043D\u0430 \u043A\u043E\u043C\u0456\u0441\u0456\u044F \u0437\u0430 \u0434\u0435\u043F\u043E\u0437\u0438\u0442, \u044F\u043A\u0443 \u043F\u0440\u043E\u0434\u0430\u0432\u0435\u0446\u044C \u0433\u043E\u0442\u043E\u0432\u0438\u0439 \u043F\u043E\u043A\u0440\u0438\u0442\u0438 \u0434\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F. \u0412\u0438\u0449\u0456 \u043A\u043E\u043C\u0456\u0441\u0456\u0457 \u0437\u0430 \u0434\u0435\u043F\u043E\u0437\u0438\u0442 \u043F\u043E\u0432\u0438\u043D\u043D\u0456 \u0431\u0443\u0442\u0438 \u043F\u043E\u0432\u043D\u0456\u0441\u0442\u044E \u043F\u043E\u043A\u0440\u0438\u0442\u0456 \u0441\u043F\u043E\u0436\u0438\u0432\u0430\u0447\u0435\u043C."
      ],
      "Create token": [
        "\u0421\u0442\u0432\u043E\u0440\u0438\u0442\u0438 \u0442\u043E\u043A\u0435\u043D"
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        ""
      ],
      "Minimum age required": [
        "\u041C\u0456\u043D\u0456\u043C\u0430\u043B\u044C\u043D\u0438\u0439 \u0432\u0456\u043A"
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        "\u0411\u0443\u0434\u044C-\u044F\u043A\u0435 \u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F \u0431\u0456\u043B\u044C\u0448\u0435 0 \u043E\u0431\u043C\u0435\u0436\u0443\u0432\u0430\u0442\u0438\u043C\u0435 \u043C\u043E\u043D\u0435\u0442\u0438, \u044F\u043A\u0456 \u043C\u043E\u0436\u0443\u0442\u044C \u0431\u0443\u0442\u0438 \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u0456 \u0434\u043B\u044F \u043E\u043F\u043B\u0430\u0442\u0438 \u0446\u044C\u043E\u0433\u043E \u043A\u043E\u043D\u0442\u0440\u0430\u043A\u0442\u0443. \u042F\u043A\u0449\u043E \u043F\u043E\u0440\u043E\u0436\u043D\u0454, \u0432\u0456\u043A\u043E\u0432\u0435 \u043E\u0431\u043C\u0435\u0436\u0435\u043D\u043D\u044F \u0432\u0438\u0437\u043D\u0430\u0447\u0430\u0442\u0438\u043C\u0435\u0442\u044C\u0441\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0430\u043C\u0438"
      ],
      "Minimum age defined by the products is %1$s": [
        "\u041C\u0456\u043D\u0456\u043C\u0430\u043B\u044C\u043D\u0438\u0439 \u0432\u0456\u043A, \u0432\u0438\u0437\u043D\u0430\u0447\u0435\u043D\u0438\u0439 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0430\u043C\u0438, \u0441\u0442\u0430\u043D\u043E\u0432\u0438\u0442\u044C %1$s"
      ],
      "No product with age restriction in this order": [
        ""
      ],
      "Additional information": [
        "\u0414\u043E\u0434\u0430\u0442\u043A\u043E\u0432\u0430 \u0456\u043D\u0444\u043E\u0440\u043C\u0430\u0446\u0456\u044F"
      ],
      "Custom information to be included in the contract for this order.": [
        "\u0421\u043F\u0435\u0446\u0456\u0430\u043B\u044C\u043D\u0430 \u0456\u043D\u0444\u043E\u0440\u043C\u0430\u0446\u0456\u044F, \u044F\u043A\u0430 \u0431\u0443\u0434\u0435 \u0432\u043A\u043B\u044E\u0447\u0435\u043D\u0430 \u0432 \u043A\u043E\u043D\u0442\u0440\u0430\u043A\u0442 \u0434\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F."
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        "\u0412\u0438 \u043F\u043E\u0432\u0438\u043D\u043D\u0456 \u0432\u0432\u0435\u0441\u0442\u0438 \u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F \u0443 \u0444\u043E\u0440\u043C\u0430\u0442\u0456 JavaScript Object Notation (JSON)."
      ],
      "remove": [
        "\u0412\u0438\u0434\u0430\u043B\u0438\u0442\u0438"
      ],
      "Custom field name": [
        "\u041D\u0430\u0437\u0432\u0430 \u0431\u0443\u0434\u0438\u043D\u043A\u0443"
      ],
      "Disabled": [
        ""
      ],
      "No deadline": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Deadline at %1$s": [
        ""
      ],
      "Select date to show nearby orders": [
        "\u0432\u0438\u0431\u0435\u0440\u0456\u0442\u044C \u0434\u0430\u0442\u0443, \u0449\u043E\u0431 \u043F\u043E\u043A\u0430\u0437\u0430\u0442\u0438 \u043D\u0430\u0439\u0431\u043B\u0438\u0436\u0447\u0456 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Only show unpaid orders": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u043E\u043F\u043B\u0430\u0447\u0435\u043D\u0456 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Only show paid orders": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u043E\u043F\u043B\u0430\u0447\u0435\u043D\u0456 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Paid": [
        "\u041E\u043F\u043B\u0430\u0447\u0435\u043D\u043E"
      ],
      "Only show orders with refunds": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F\u043C\u0438"
      ],
      "Refunded": [
        "\u041F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043E"
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        "\u041D\u0435 \u043F\u0435\u0440\u0435\u0440\u0430\u0445\u043E\u0432\u0430\u043D\u043E"
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F, \u0434\u0435 \u043A\u043B\u0456\u0454\u043D\u0442\u0438 \u0437\u0430\u043F\u043B\u0430\u0442\u0438\u043B\u0438, \u0430\u043B\u0435 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0456 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0438 \u0432\u0456\u0434 \u043F\u043E\u0441\u0442\u0430\u0447\u0430\u043B\u044C\u043D\u0438\u043A\u0430 \u043F\u043B\u0430\u0442\u0435\u0436\u0456\u0432 \u0449\u0435 \u043D\u0435 \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u0456"
      ],
      "Completed": [
        "\u0413\u0440\u043E\u0448\u0456 \u043F\u0435\u0440\u0435\u0440\u0430\u0445\u043E\u0432\u0430\u043D\u0456"
      ],
      "Remove all filters": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0441\u0456 \u0444\u0456\u043B\u044C\u0442\u0440\u0438"
      ],
      "Clear date filter": [
        "\u043E\u0447\u0438\u0441\u0442\u0438\u0442\u0438 \u0444\u0456\u043B\u044C\u0442\u0440 \u0434\u0430\u0442\u0438"
      ],
      "Jump to date (%1$s)": [
        ""
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        "\u043F\u0435\u0440\u0435\u0439\u0442\u0438 \u0434\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u0437\u0430\u0437\u043D\u0430\u0447\u0435\u043D\u0438\u043C \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440\u043E\u043C"
      ],
      "Order id": [
        "\u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Instance unknown": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Order unknown": [
        "\u043D\u0435\u0432\u0456\u0434\u043E\u043C\u043E"
      ],
      "Create order": [
        "\u0441\u0442\u0432\u043E\u0440\u0438\u0442\u0438 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        "\u0414\u0430\u0442\u0430"
      ],
      "Refund": [
        "\u041F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "copy url": [
        "\u0441\u043A\u043E\u043F\u0456\u044E\u0432\u0430\u0442\u0438 url"
      ],
      "Load more orders after the last one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0456\u0432 \u043F\u0456\u0441\u043B\u044F \u043E\u0441\u0442\u0430\u043D\u043D\u044C\u043E\u0433\u043E"
      ],
      "No orders have been found matching your query!": [
        "\u0417\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F, \u0449\u043E \u0432\u0456\u0434\u043F\u043E\u0432\u0456\u0434\u0430\u044E\u0442\u044C \u0432\u0430\u0448\u043E\u043C\u0443 \u0437\u0430\u043F\u0438\u0442\u0443, \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E!"
      ],
      "Duplicated": [
        "\u0434\u0443\u0431\u043B\u044C\u043E\u0432\u0430\u043D\u043E"
      ],
      "This value exceeds the refundable amount.": [
        "\u0446\u044F \u0441\u0443\u043C\u0430 \u043F\u0435\u0440\u0435\u0432\u0438\u0449\u0443\u0454 \u0441\u0443\u043C\u0443, \u0449\u043E \u043F\u0456\u0434\u043B\u044F\u0433\u0430\u0454 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044E"
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        ""
      ],
      "refund": [
        "\u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Reason": [
        "\u041F\u0440\u0438\u0447\u0438\u043D\u0430"
      ],
      "Amount to be refunded": [
        "\u0441\u0443\u043C\u0430 \u0434\u043E \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Max refundable:": [
        "\u041C\u0430\u043A\u0441. \u0441\u0443\u043C\u0430 \u0434\u043B\u044F \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F:"
      ],
      "Choose one...": [
        "\u0412\u0438\u0431\u0435\u0440\u0456\u0442\u044C \u043E\u0434\u043D\u0443..."
      ],
      "Requested by the customer": [
        "\u0437\u0430\u043F\u0438\u0442\u0430\u043D\u043E \u043A\u043B\u0456\u0454\u043D\u0442\u043E\u043C"
      ],
      "Other": [
        "\u0456\u043D\u0448\u0435"
      ],
      "Why this order is being refunded": [
        "\u0447\u043E\u043C\u0443 \u0446\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043F\u043E\u0432\u0435\u0440\u0442\u0430\u0454\u0442\u044C\u0441\u044F"
      ],
      "More information to give context": [
        "\u0434\u043E\u0434\u0430\u0442\u043A\u043E\u0432\u0430 \u0456\u043D\u0444\u043E\u0440\u043C\u0430\u0446\u0456\u044F \u0434\u043B\u044F \u043D\u0430\u0434\u0430\u043D\u043D\u044F \u043A\u043E\u043D\u0442\u0435\u043A\u0441\u0442\u0443"
      ],
      "now": [
        "\u043D\u0456"
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F, \u0434\u0435 \u043A\u043B\u0456\u0454\u043D\u0442\u0438 \u0437\u0430\u043F\u043B\u0430\u0442\u0438\u043B\u0438, \u0430\u043B\u0435 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0456 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0438 \u0432\u0456\u0434 \u043F\u043E\u0441\u0442\u0430\u0447\u0430\u043B\u044C\u043D\u0438\u043A\u0430 \u043F\u043B\u0430\u0442\u0435\u0436\u0456\u0432 \u0449\u0435 \u043D\u0435 \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u0456"
      ],
      "This wire transfer has been notified by the payment service provider.": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F, \u0434\u0435 \u043A\u043B\u0456\u0454\u043D\u0442\u0438 \u0437\u0430\u043F\u043B\u0430\u0442\u0438\u043B\u0438, \u0430\u043B\u0435 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0456 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0438 \u0432\u0456\u0434 \u043F\u043E\u0441\u0442\u0430\u0447\u0430\u043B\u044C\u043D\u0438\u043A\u0430 \u043F\u043B\u0430\u0442\u0435\u0436\u0456\u0432 \u0449\u0435 \u043D\u0435 \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u0456"
      ],
      "This refund is waiting to be claimed by the customer.": [
        "\u0421\u0443\u043C\u0430, \u044F\u043A\u0443 \u043C\u0430\u0454 \u0441\u043F\u043B\u0430\u0442\u0438\u0442\u0438 \u043A\u043B\u0456\u0454\u043D\u0442"
      ],
      "This refund has been claimed by the customer.": [
        "\u0421\u0443\u043C\u0430, \u044F\u043A\u0443 \u043C\u0430\u0454 \u0441\u043F\u043B\u0430\u0442\u0438\u0442\u0438 \u043A\u043B\u0456\u0454\u043D\u0442"
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        "\u0423\u043C\u043E\u0432\u0438 \u043A\u043E\u043D\u0442\u0440\u0430\u043A\u0442\u0443"
      ],
      "Human-readable description of the whole purchase": [
        "\u0447\u0438\u0442\u0430\u0431\u0435\u043B\u044C\u043D\u0438\u0439 \u043E\u043F\u0438\u0441 \u0432\u0441\u0456\u0454\u0457 \u043F\u043E\u043A\u0443\u043F\u043A\u0438"
      ],
      "Total price for the transaction": [
        "\u0437\u0430\u0433\u0430\u043B\u044C\u043D\u0430 \u0446\u0456\u043D\u0430 \u0437\u0430 \u0442\u0440\u0430\u043D\u0437\u0430\u043A\u0446\u0456\u044E"
      ],
      "URL for this purchase": [
        "URL \u0434\u043B\u044F \u0446\u0456\u0454\u0457 \u043F\u043E\u043A\u0443\u043F\u043A\u0438"
      ],
      "Max fee": [
        "\u041C\u0430\u043A\u0441\u0438\u043C\u0430\u043B\u044C\u043D\u0430 \u043A\u043E\u043C\u0456\u0441\u0456\u044F"
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        "\u043C\u0430\u043A\u0441\u0438\u043C\u0430\u043B\u044C\u043D\u0430 \u0437\u0430\u0433\u0430\u043B\u044C\u043D\u0430 \u043A\u043E\u043C\u0456\u0441\u0456\u044F \u0437\u0430 \u0434\u0435\u043F\u043E\u0437\u0438\u0442, \u043F\u0440\u0438\u0439\u043D\u044F\u0442\u0430 \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u0435\u043C \u0434\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u043A\u043E\u043D\u0442\u0440\u0430\u043A\u0442\u0443"
      ],
      "Time when this contract was generated": [
        "\u0447\u0430\u0441, \u043A\u043E\u043B\u0438 \u0446\u0435\u0439 \u043A\u043E\u043D\u0442\u0440\u0430\u043A\u0442 \u0431\u0443\u043B\u043E \u0437\u0433\u0435\u043D\u0435\u0440\u043E\u0432\u0430\u043D\u043E"
      ],
      "Refund deadline": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        "\u043F\u0456\u0441\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u0442\u0435\u0440\u043C\u0456\u043D\u0443 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u043D\u0435 \u043F\u0440\u0438\u0439\u043C\u0430\u044E\u0442\u044C\u0441\u044F"
      ],
      "Payment deadline": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        "\u043F\u0456\u0441\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u0442\u0435\u0440\u043C\u0456\u043D\u0443 \u043F\u0440\u043E\u0434\u0430\u0432\u0435\u0446\u044C \u043D\u0435 \u043F\u0440\u0438\u0439\u043C\u0430\u0442\u0438\u043C\u0435 \u043F\u043B\u0430\u0442\u0435\u0436\u0456 \u0437\u0430 \u043A\u043E\u043D\u0442\u0440\u0430\u043A\u0442\u043E\u043C"
      ],
      "Wire transfer deadline": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "Transfer deadline for the exchange": [
        "\u0442\u0435\u0440\u043C\u0456\u043D \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443 \u0434\u043B\u044F \u043E\u0431\u043C\u0456\u043D\u043D\u0438\u043A\u0430"
      ],
      "Time indicating when the order should be delivered": [
        "\u0447\u0430\u0441, \u0449\u043E \u0432\u043A\u0430\u0437\u0443\u0454, \u043A\u043E\u043B\u0438 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043C\u0430\u0454 \u0431\u0443\u0442\u0438 \u0434\u043E\u0441\u0442\u0430\u0432\u043B\u0435\u043D\u0435"
      ],
      "Where the order will be delivered": [
        "\u043A\u0443\u0434\u0438 \u0431\u0443\u0434\u0435 \u0434\u043E\u0441\u0442\u0430\u0432\u043B\u0435\u043D\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Auto-refund delay": [
        "\u0417\u0430\u0442\u0440\u0438\u043C\u043A\u0430 \u0430\u0432\u0442\u043E\u043C\u0430\u0442\u0438\u0447\u043D\u043E\u0433\u043E \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        "\u0441\u043A\u0456\u043B\u044C\u043A\u0438 \u0447\u0430\u0441\u0443 \u0433\u0430\u043C\u0430\u043D\u0435\u0446\u044C \u043F\u043E\u0432\u0438\u043D\u0435\u043D \u043D\u0430\u043C\u0430\u0433\u0430\u0442\u0438\u0441\u044F \u043E\u0442\u0440\u0438\u043C\u0430\u0442\u0438 \u0430\u0432\u0442\u043E\u043C\u0430\u0442\u0438\u0447\u043D\u0435 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u0437\u0430 \u043F\u043E\u043A\u0443\u043F\u043A\u0443"
      ],
      "Extra info": [
        "\u0414\u043E\u0434\u0430\u0442\u043A\u043E\u0432\u0430 \u0456\u043D\u0444\u043E\u0440\u043C\u0430\u0446\u0456\u044F"
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        "\u0434\u043E\u0434\u0430\u0442\u043A\u043E\u0432\u0456 \u0434\u0430\u043D\u0456, \u044F\u043A\u0456 \u0456\u043D\u0442\u0435\u0440\u043F\u0440\u0435\u0442\u0443\u044E\u0442\u044C\u0441\u044F \u043B\u0438\u0448\u0435 \u0444\u0440\u043E\u043D\u0442\u0435\u043D\u0434\u043E\u043C \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u044F"
      ],
      "order created": [
        "\u0421\u0442\u0430\u0442\u0443\u0441 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "pay deadline": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "refund deadline": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "delivery": [
        "\u0414\u0430\u0442\u0430 \u0434\u043E\u0441\u0442\u0430\u0432\u043A\u0438"
      ],
      "Order": [
        "\u0417\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Claimed": [
        "\u043E\u0442\u0440\u0438\u043C\u0430\u043D\u043E"
      ],
      "Claimed at": [
        "\u043E\u0442\u0440\u0438\u043C\u0430\u043D\u043E \u043E"
      ],
      "Timeline": [
        "\u0425\u0440\u043E\u043D\u043E\u043B\u043E\u0433\u0456\u044F"
      ],
      "Payment details": [
        "\u0414\u0435\u0442\u0430\u043B\u0456 \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "Order status": [
        "\u0421\u0442\u0430\u0442\u0443\u0441 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Unpaid": [
        "\u043D\u0435\u043E\u043F\u043B\u0430\u0447\u0435\u043D\u043E"
      ],
      "Product list": [
        "\u0421\u043F\u0438\u0441\u043E\u043A \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432"
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        ""
      ],
      "refund created: %1$s": [
        "\u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u0443\u0441\u043F\u0456\u0448\u043D\u043E \u0441\u0442\u0432\u043E\u0440\u0435\u043D\u043E"
      ],
      "refund created: %1$s: %2$s": [
        "\u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u0443\u0441\u043F\u0456\u0448\u043D\u043E \u0441\u0442\u0432\u043E\u0440\u0435\u043D\u043E"
      ],
      "refund taken: %1$s": [
        "\u041F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u0437\u0434\u0456\u0439\u0441\u043D\u0435\u043D\u043E"
      ],
      "refund taken: %1$s: %2$s": [
        "\u041F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u0437\u0434\u0456\u0439\u0441\u043D\u0435\u043D\u043E"
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Wired": [
        "\u043F\u0435\u0440\u0435\u0440\u0430\u0445\u043E\u0432\u0430\u043D\u043E"
      ],
      "Refund order": [
        "\u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043D\u0430 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Not refundable": [
        "\u043D\u0435 \u043F\u0456\u0434\u043B\u044F\u0433\u0430\u0454 \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044E"
      ],
      "Next event in": [
        ""
      ],
      "The order was wired.": [
        ""
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        "\u041F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u0430 \u0441\u0443\u043C\u0430"
      ],
      "Refund taken": [
        "\u041F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F \u0437\u0434\u0456\u0439\u0441\u043D\u0435\u043D\u043E"
      ],
      "Status URL": [
        "URL \u0441\u0442\u0430\u0442\u0443\u0441\u0443"
      ],
      "Refund URI": [
        "URI \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "Pay at": [
        "URL \u0434\u043B\u044F \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "Order status URL": [
        "URL \u0441\u0442\u0430\u0442\u0443\u0441\u0443 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Payment URI": [
        "URI \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        "\u041D\u0435\u0432\u0456\u0434\u043E\u043C\u0438\u0439 \u0441\u0442\u0430\u0442\u0443\u0441 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F. \u0426\u0435 \u043F\u043E\u043C\u0438\u043B\u043A\u0430, \u0431\u0443\u0434\u044C \u043B\u0430\u0441\u043A\u0430, \u0437\u0432'\u044F\u0436\u0456\u0442\u044C\u0441\u044F \u0437 \u0430\u0434\u043C\u0456\u043D\u0456\u0441\u0442\u0440\u0430\u0442\u043E\u0440\u043E\u043C."
      ],
      "Invalid. Please enter letters and numbers only.": [
        ""
      ],
      "Just letters and numbers from 2 to 7": [
        "\u043B\u0438\u0448\u0435 \u043B\u0456\u0442\u0435\u0440\u0438 \u0442\u0430 \u0446\u0438\u0444\u0440\u0438 \u0432\u0456\u0434 2 \u0434\u043E 7"
      ],
      "The size of the key must be 32 characters": [
        "\u0440\u043E\u0437\u043C\u0456\u0440 \u043A\u043B\u044E\u0447\u0430 \u043F\u043E\u0432\u0438\u043D\u0435\u043D \u0431\u0443\u0442\u0438 32"
      ],
      "Internal ID on the system": [
        ""
      ],
      "Useful to identify the device physically": [
        ""
      ],
      "Verification algorithm": [
        "\u0410\u043B\u0433\u043E\u0440\u0438\u0442\u043C \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0438"
      ],
      "Algorithm used to verify transactions in offline mode": [
        "\u0410\u043B\u0433\u043E\u0440\u0438\u0442\u043C \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F \u0434\u043B\u044F \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0438 \u0442\u0440\u0430\u043D\u0437\u0430\u043A\u0446\u0456\u0457 \u0432 \u043E\u0444\u043B\u0430\u0439\u043D-\u0440\u0435\u0436\u0438\u043C\u0456"
      ],
      "Device key": [
        ""
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        ""
      ],
      "Your device needs to match exactly the same value": [
        ""
      ],
      "Generate random secret key": [
        "\u0437\u0433\u0435\u043D\u0435\u0440\u0443\u0432\u0430\u0442\u0438 \u0432\u0438\u043F\u0430\u0434\u043A\u043E\u0432\u0438\u0439 \u0441\u0435\u043A\u0440\u0435\u0442\u043D\u0438\u0439 \u043A\u043B\u044E\u0447"
      ],
      "Random": [
        "\u0432\u0438\u043F\u0430\u0434\u043A\u043E\u0432\u0438\u0439"
      ],
      "Create another": [
        "\u0441\u0442\u0432\u043E\u0440\u0438\u0442\u0438 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        ""
      ],
      "OTP devices": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0438\u0439 \u0440\u0435\u0437\u0435\u0440\u0432"
      ],
      "Delete selected devices from the database": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0438\u0431\u0440\u0430\u043D\u0438\u0439 \u0440\u0435\u0437\u0435\u0440\u0432 \u0437 \u0431\u0430\u0437\u0438 \u0434\u0430\u043D\u0438\u0445"
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        "\u0429\u0435 \u043D\u0435\u043C\u0430\u0454 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0439, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "Template id is unknown": [
        ""
      ],
      "The provided information is inconsistent with the current state of the template": [
        ""
      ],
      "Device:": [
        "\u041F\u0440\u0438\u0441\u0442\u0440\u0456\u0439:"
      ],
      "Algorithm to use to verify transactions in offline mode": [
        "\u0410\u043B\u0433\u043E\u0440\u0438\u0442\u043C \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F \u0434\u043B\u044F \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0438 \u0442\u0440\u0430\u043D\u0437\u0430\u043A\u0446\u0456\u0457 \u0432 \u043E\u0444\u043B\u0430\u0439\u043D-\u0440\u0435\u0436\u0438\u043C\u0456"
      ],
      "Not modified": [
        "\u041D\u0435 \u043F\u0435\u0440\u0435\u0440\u0430\u0445\u043E\u0432\u0430\u043D\u043E"
      ],
      "Be sure to be very hard to guess or use the random generator": [
        ""
      ],
      "Your device need to have exactly the same value": [
        ""
      ],
      "Change key": [
        ""
      ],
      "Is not the same": [
        "\u043D\u0435 \u0441\u043F\u0456\u0432\u043F\u0430\u0434\u0430\u0454"
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        "\u0412\u0438 \u043E\u043D\u043E\u0432\u043B\u044E\u0454\u0442\u0435 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443 \u0437 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457 \u0437 \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440\u043E\u043C %1$s"
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0438\u0442\u0438 \u0437\u043C\u0456\u043D\u0443"
      ],
      "Testing password change": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0438\u0442\u0438 \u0437\u043C\u0456\u043D\u0443"
      ],
      "Password changed": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0438\u0442\u0438 \u0437\u043C\u0456\u043D\u0443"
      ],
      "The current password is wrong.": [
        "\u0412\u0441\u0442\u0430\u043D\u043E\u0432\u0438\u0442\u0438 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        "\u0421\u0435\u0440\u0432\u0435\u0440 \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "Add element to the list": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u0435\u043B\u0435\u043C\u0435\u043D\u0442 \u0434\u043E \u0441\u043F\u0438\u0441\u043A\u0443"
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        "\u043D\u0430\u0442\u0438\u0441\u043D\u0456\u0442\u044C \u0442\u0443\u0442, \u0449\u043E\u0431 \u043D\u0430\u043B\u0430\u0448\u0442\u0443\u0432\u0430\u0442\u0438 \u0437\u0430\u043F\u0430\u0441 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443, \u0437\u0430\u043B\u0438\u0448\u0442\u0435 \u044F\u043A \u0454, \u0456 \u0431\u0435\u043A\u0435\u043D\u0434 \u043D\u0435 \u0431\u0443\u0434\u0435 \u043A\u043E\u043D\u0442\u0440\u043E\u043B\u044E\u0432\u0430\u0442\u0438 \u0437\u0430\u043F\u0430\u0441."
      ],
      "Manage stock": [
        "\u041A\u0435\u0440\u0443\u0432\u0430\u043D\u043D\u044F \u0437\u0430\u043F\u0430\u0441\u0430\u043C\u0438"
      ],
      "This product has been configured without stock control": [
        "\u0446\u0435\u0439 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u043D\u0430\u043B\u0430\u0448\u0442\u043E\u0432\u0430\u043D\u0438\u0439 \u0431\u0435\u0437 \u043A\u043E\u043D\u0442\u0440\u043E\u043B\u044E \u0437\u0430\u043F\u0430\u0441\u0456\u0432"
      ],
      "Infinite": [
        "\u041D\u0435\u0441\u043A\u0456\u043D\u0447\u0435\u043D\u043D\u0438\u0439"
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        "\u0432\u0442\u0440\u0430\u0442\u0438 \u043D\u0435 \u043C\u043E\u0436\u0443\u0442\u044C \u0431\u0443\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0438\u043C\u0438 \u0437\u0430 \u043F\u043E\u0442\u043E\u0447\u043D\u0456 \u0442\u0430 \u043F\u0440\u0438\u0431\u0443\u0432\u0430\u044E\u0447\u0456 (\u043C\u0430\u043A\u0441 %1$s)"
      ],
      "Incoming": [
        "\u041F\u0440\u0438\u0431\u0443\u0442\u0442\u044F"
      ],
      "Lost": [
        "\u0412\u0442\u0440\u0430\u0447\u0435\u043D\u043E"
      ],
      "Current": [
        "\u041F\u043E\u0442\u043E\u0447\u043D\u0438\u0439"
      ],
      "Remove stock control for this product": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u043A\u043E\u043D\u0442\u0440\u043E\u043B\u044C \u0437\u0430\u043F\u0430\u0441\u0456\u0432 \u0434\u043B\u044F \u0446\u044C\u043E\u0433\u043E \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443"
      ],
      "without stock": [
        "\u0431\u0435\u0437 \u0437\u0430\u043F\u0430\u0441\u0443"
      ],
      "Next restock": [
        "\u041D\u0430\u0441\u0442\u0443\u043F\u043D\u0435 \u043F\u043E\u043F\u043E\u0432\u043D\u0435\u043D\u043D\u044F"
      ],
      "Warehouse address": [
        "\u0410\u0434\u0440\u0435\u0441\u0430 \u0440\u0430\u0445\u0443\u043D\u043A\u0443"
      ],
      "Invalid amount": [
        "\u0424\u0456\u043A\u0441\u043E\u0432\u0430\u043D\u0430 \u0441\u0443\u043C\u0430"
      ],
      "Product identification to use in URLs (for internal use only).": [
        "\u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0446\u0456\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F \u0432 URL (\u0442\u0456\u043B\u044C\u043A\u0438 \u0434\u043B\u044F \u0432\u043D\u0443\u0442\u0440\u0456\u0448\u043D\u044C\u043E\u0433\u043E \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F)."
      ],
      "Illustration of the product for customers.": [
        "\u0456\u043B\u044E\u0441\u0442\u0440\u0430\u0446\u0456\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 \u0434\u043B\u044F \u043A\u043B\u0456\u0454\u043D\u0442\u0456\u0432."
      ],
      "Product name.": [
        "\u041F\u0440\u043E\u0434\u0443\u043A\u0442"
      ],
      "Product description for customers.": [
        "\u043E\u043F\u0438\u0441 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 \u0434\u043B\u044F \u043A\u043B\u0456\u0454\u043D\u0442\u0456\u0432."
      ],
      "Age restriction": [
        "\u041E\u0431\u043C\u0435\u0436\u0435\u043D\u0438\u0439 \u0437\u0430 \u0432\u0456\u043A\u043E\u043C"
      ],
      "Is this product restricted for customer below certain age?": [
        "\u0446\u0435\u0439 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u043E\u0431\u043C\u0435\u0436\u0435\u043D\u0438\u0439 \u0434\u043B\u044F \u043A\u043B\u0456\u0454\u043D\u0442\u0456\u0432 \u043C\u043E\u043B\u043E\u0434\u0448\u0435 \u043F\u0435\u0432\u043D\u043E\u0433\u043E \u0432\u0456\u043A\u0443?"
      ],
      "Minimum age of the customer": [
        "\u041C\u0456\u043D\u0456\u043C\u0430\u043B\u044C\u043D\u0438\u0439 \u0432\u0456\u043A"
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        "\u043E\u0434\u0438\u043D\u0438\u0446\u044F, \u0449\u043E \u043E\u043F\u0438\u0441\u0443\u0454 \u043A\u0456\u043B\u044C\u043A\u0456\u0441\u0442\u044C \u043F\u0440\u043E\u0434\u0430\u043D\u043E\u0433\u043E \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 (\u043D\u0430\u043F\u0440\u0438\u043A\u043B\u0430\u0434, 2 \u043A\u0456\u043B\u043E\u0433\u0440\u0430\u043C\u0438, 5 \u043B\u0456\u0442\u0440\u0456\u0432, 3 \u043F\u0440\u0435\u0434\u043C\u0435\u0442\u0438, 5 \u043C\u0435\u0442\u0440\u0456\u0432) \u0434\u043B\u044F \u043A\u043B\u0456\u0454\u043D\u0442\u0456\u0432."
      ],
      "Example: kg, items or liters": [
        ""
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        "\u0446\u0456\u043D\u0430 \u043F\u0440\u043E\u0434\u0430\u0436\u0443 \u0434\u043B\u044F \u043A\u043B\u0456\u0454\u043D\u0442\u0456\u0432, \u0432\u043A\u043B\u044E\u0447\u0430\u044E\u0447\u0438 \u043F\u043E\u0434\u0430\u0442\u043A\u0438, \u0437\u0430 \u0432\u0438\u0449\u0435\u0437\u0430\u0437\u043D\u0430\u0447\u0435\u043D\u0456 \u043E\u0434\u0438\u043D\u0438\u0446\u0456 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443."
      ],
      "Stock": [
        "\u0417\u0430\u043F\u0430\u0441"
      ],
      "Inventory for products with finite supply (for internal use only).": [
        "\u0456\u043D\u0432\u0435\u043D\u0442\u0430\u0440\u0438\u0437\u0430\u0446\u0456\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 \u0434\u043B\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0437 \u043E\u0431\u043C\u0435\u0436\u0435\u043D\u0438\u043C \u0437\u0430\u043F\u0430\u0441\u043E\u043C (\u0442\u0456\u043B\u044C\u043A\u0438 \u0434\u043B\u044F \u0432\u043D\u0443\u0442\u0440\u0456\u0448\u043D\u044C\u043E\u0433\u043E \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F)"
      ],
      "Taxes included in the product price, exposed to customers.": [
        "\u043F\u043E\u0434\u0430\u0442\u043A\u0438, \u0432\u043A\u043B\u044E\u0447\u0435\u043D\u0456 \u0432 \u0446\u0456\u043D\u0443 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443, \u043F\u043E\u043A\u0430\u0437\u0430\u043D\u0456 \u043A\u043B\u0456\u0454\u043D\u0442\u0430\u043C"
      ],
      "Search by category description or id": [
        "\u0448\u0443\u043A\u0430\u0442\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0438 \u0437\u0430 \u0457\u0445\u043D\u0456\u043C \u043E\u043F\u0438\u0441\u043E\u043C \u0430\u0431\u043E \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440\u043E\u043C"
      ],
      "Categories where this product will be listed on.": [
        "\u0430\u0434\u0440\u0435\u0441\u0430, \u0437\u0430 \u044F\u043A\u043E\u044E \u0431\u0443\u0434\u0443\u0442\u044C \u0434\u043E\u0441\u0442\u0430\u0432\u043B\u0435\u043D\u0456 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0438."
      ],
      "Add product to inventory": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u0434\u043E \u0456\u043D\u0432\u0435\u043D\u0442\u0430\u0440\u044E"
      ],
      "Product not found.": [
        "\u0421\u0435\u0440\u0432\u0435\u0440 \u043D\u0435 \u0437\u043D\u0430\u0439\u0434\u0435\u043D\u043E"
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        ""
      ],
      "Sold": [
        "\u041F\u0440\u043E\u0434\u0430\u043D\u043E"
      ],
      "Free": [
        ""
      ],
      "Go to product update page": [
        "\u043F\u0435\u0440\u0435\u0439\u0442\u0438 \u043D\u0430 \u0441\u0442\u043E\u0440\u0456\u043D\u043A\u0443 \u043E\u043D\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443"
      ],
      "Update": [
        "\u041E\u043D\u043E\u0432\u0438\u0442\u0438"
      ],
      "Remove this product from the database": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0446\u0435\u0439 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u0437 \u0431\u0430\u0437\u0438 \u0434\u0430\u043D\u0438\u0445"
      ],
      "Update the product with new price": [
        "\u043E\u043D\u043E\u0432\u0438\u0442\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u0437 \u043D\u043E\u0432\u043E\u044E \u0446\u0456\u043D\u043E\u044E"
      ],
      "Update product with new price": [
        "\u043E\u043D\u043E\u0432\u0438\u0442\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u0437 \u043D\u043E\u0432\u043E\u044E \u0446\u0456\u043D\u043E\u044E"
      ],
      "Confirm update": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0436\u0435\u043D\u043E"
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        "\u0432\u0442\u0440\u0430\u0442\u0438 \u043D\u0435 \u043C\u043E\u0436\u0443\u0442\u044C \u0431\u0443\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0438\u043C\u0438 \u0437\u0430 \u043F\u043E\u0442\u043E\u0447\u043D\u0456 \u0442\u0430 \u043F\u0440\u0438\u0431\u0443\u0432\u0430\u044E\u0447\u0456 (\u043C\u0430\u043A\u0441 %1$s)"
      ],
      "Add more elements to the inventory": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0435\u043B\u0435\u043C\u0435\u043D\u0442\u0456\u0432 \u0434\u043E \u0456\u043D\u0432\u0435\u043D\u0442\u0430\u0440\u044E"
      ],
      "Report elements lost in the inventory": [
        "\u043F\u043E\u0432\u0456\u0434\u043E\u043C\u0438\u0442\u0438 \u043F\u0440\u043E \u0432\u0442\u0440\u0430\u0442\u0443 \u0435\u043B\u0435\u043C\u0435\u043D\u0442\u0456\u0432 \u0432 \u0456\u043D\u0432\u0435\u043D\u0442\u0430\u0440\u0456"
      ],
      "New price for the product": [
        "\u043D\u043E\u0432\u0430 \u0446\u0456\u043D\u0430 \u0434\u043B\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443"
      ],
      "There are values with errors.": [
        "\u0454 \u0437\u043D\u0430\u0447\u0435\u043D\u043D\u044F \u0437 \u043F\u043E\u043C\u0438\u043B\u043A\u0430\u043C\u0438"
      ],
      "Update product with new stock and price": [
        "\u043E\u043D\u043E\u0432\u0438\u0442\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u0437 \u043D\u043E\u0432\u0438\u043C \u0437\u0430\u043F\u0430\u0441\u043E\u043C \u0456 \u0446\u0456\u043D\u043E\u044E"
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        "\u041F\u0440\u043E\u0434\u0443\u043A\u0442\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "Product (ID: %1$s) has been deleted": [
        "\u041F\u0440\u043E\u0434\u0443\u043A\u0442 (ID: %1$s) \u043E\u043D\u043E\u0432\u043B\u0435\u043D\u043E"
      ],
      "Jump to product with the given product ID": [
        "\u043F\u0435\u0440\u0435\u0439\u0442\u0438 \u0434\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u0437\u0430\u0437\u043D\u0430\u0447\u0435\u043D\u0438\u043C \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440\u043E\u043C"
      ],
      "Product id": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443"
      ],
      "Delete product": [
        "\u0412\u0438\u0434\u0430\u043B\u0435\u043D\u043E"
      ],
      'Delete the product "%1$s"': [
        "\u043D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442"
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        ""
      ],
      "Deleting a product cannot be undone.": [
        ""
      ],
      "Product id:": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443:"
      ],
      "Must be greater that 0": [
        "\u043C\u0430\u0454 \u0431\u0443\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 0"
      ],
      "Too short": [
        "\u0437\u0430\u043D\u0430\u0434\u0442\u043E \u043A\u043E\u0440\u043E\u0442\u043A\u0438\u0439"
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440"
      ],
      "Name of the template in URLs.": [
        "\u041D\u0430\u0437\u0432\u0430 \u0448\u0430\u0431\u043B\u043E\u043D\u0443 \u0432 URL."
      ],
      "Template name": [
        "\u0428\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "Describe what this template stands for": [
        "\u041E\u043F\u0438\u0448\u0456\u0442\u044C, \u0449\u043E \u043F\u0440\u0435\u0434\u0441\u0442\u0430\u0432\u043B\u044F\u0454 \u0446\u0435\u0439 \u0448\u0430\u0431\u043B\u043E\u043D"
      ],
      "Order summary": [
        "\u041F\u0456\u0434\u0441\u0443\u043C\u043E\u043A \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "If specified here, this template will create orders with the same summary": [
        "\u042F\u043A\u0449\u043E \u0432\u043A\u0430\u0437\u0430\u043D\u043E, \u0446\u0435\u0439 \u0448\u0430\u0431\u043B\u043E\u043D \u0441\u0442\u0432\u043E\u0440\u0438\u0442\u044C \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u043E\u0434\u043D\u0430\u043A\u043E\u0432\u0438\u043C \u043F\u0456\u0434\u0441\u0443\u043C\u043A\u043E\u043C"
      ],
      "Summary is editable": [
        ""
      ],
      "Allow the user to change the summary.": [
        ""
      ],
      "If specified here, this template will create orders with the same price": [
        "\u042F\u043A\u0449\u043E \u0432\u043A\u0430\u0437\u0430\u043D\u043E, \u0446\u0435\u0439 \u0448\u0430\u0431\u043B\u043E\u043D \u0441\u0442\u0432\u043E\u0440\u0438\u0442\u044C \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u043E\u0434\u043D\u0430\u043A\u043E\u0432\u043E\u044E \u0446\u0456\u043D\u043E\u044E"
      ],
      "Amount is editable": [
        "\u0417\u0430\u0440\u0430\u0445\u043E\u0432\u0430\u043D\u0430 \u0441\u0443\u043C\u0430"
      ],
      "Allow the user to select the amount to pay.": [
        ""
      ],
      "Currency is editable": [
        ""
      ],
      "Allow the user to change currency.": [
        ""
      ],
      "Supported currencies": [
        ""
      ],
      "Supported currencies: %1$s": [
        ""
      ],
      "Minimum age": [
        "\u041C\u0456\u043D\u0456\u043C\u0430\u043B\u044C\u043D\u0438\u0439 \u0432\u0456\u043A"
      ],
      "Is this contract restricted to some age?": [
        "\u0427\u0438 \u043E\u0431\u043C\u0435\u0436\u0435\u043D\u0438\u0439 \u0446\u0435\u0439 \u043A\u043E\u043D\u0442\u0440\u0430\u043A\u0442 \u0437\u0430 \u0432\u0456\u043A\u043E\u043C?"
      ],
      "Payment timeout": [
        "\u0422\u0430\u0439\u043C-\u0430\u0443\u0442 \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        "\u0421\u043A\u0456\u043B\u044C\u043A\u0438 \u0447\u0430\u0441\u0443 \u0443 \u043A\u043B\u0456\u0454\u043D\u0442\u0430 \u0434\u043B\u044F \u0437\u0430\u0432\u0435\u0440\u0448\u0435\u043D\u043D\u044F \u043E\u043F\u043B\u0430\u0442\u0438 \u043F\u0456\u0441\u043B\u044F \u0441\u0442\u0432\u043E\u0440\u0435\u043D\u043D\u044F \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F."
      ],
      "OTP device": [
        ""
      ],
      "Use to verify transactions in offline mode.": [
        "\u0410\u043B\u0433\u043E\u0440\u0438\u0442\u043C \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F \u0434\u043B\u044F \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0438 \u0442\u0440\u0430\u043D\u0437\u0430\u043A\u0446\u0456\u0457 \u0432 \u043E\u0444\u043B\u0430\u0439\u043D-\u0440\u0435\u0436\u0438\u043C\u0456."
      ],
      "No OTP device.": [
        ""
      ],
      "Add one first": [
        ""
      ],
      "No device": [
        ""
      ],
      "Add new templates": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "Load more templates before the first one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0448\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u0434\u043E \u043F\u0435\u0440\u0448\u043E\u0433\u043E"
      ],
      "Delete selected templates from the database": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0438\u0431\u0440\u0430\u043D\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438 \u0437 \u0431\u0430\u0437\u0438 \u0434\u0430\u043D\u0438\u0445"
      ],
      "Use template to create new order": [
        "\u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u043E\u0432\u0443\u0432\u0430\u0442\u0438 \u0448\u0430\u0431\u043B\u043E\u043D \u0434\u043B\u044F \u0441\u0442\u0432\u043E\u0440\u0435\u043D\u043D\u044F \u043D\u043E\u0432\u043E\u0433\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Generate a QR code for the template.": [
        "\u0441\u0442\u0432\u043E\u0440\u0438\u0442\u0438 QR-\u043A\u043E\u0434 \u0434\u043B\u044F \u0448\u0430\u0431\u043B\u043E\u043D\u0443."
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0448\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u043F\u0456\u0441\u043B\u044F \u043E\u0441\u0442\u0430\u043D\u043D\u044C\u043E\u0433\u043E"
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        "\u0428\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "Jump to template with the given template ID": [
        "\u043F\u0435\u0440\u0435\u0439\u0442\u0438 \u0434\u043E \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u0437\u0430\u0437\u043D\u0430\u0447\u0435\u043D\u0438\u043C \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440\u043E\u043C"
      ],
      "Template identification": [
        ""
      ],
      "Delete template": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      'Delete the template "%1$s"': [
        ""
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        ""
      ],
      "Deleting a template cannot be undone.": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u043D\u043E\u0432\u0456\u0448\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "Print": [
        "\u0414\u0440\u0443\u043A"
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        "\u042F\u043A\u0449\u043E \u0432\u043A\u0430\u0437\u0430\u043D\u043E, \u0446\u0435\u0439 \u0448\u0430\u0431\u043B\u043E\u043D \u0441\u0442\u0432\u043E\u0440\u0438\u0442\u044C \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u043E\u0434\u043D\u0430\u043A\u043E\u0432\u0438\u043C \u043F\u0456\u0434\u0441\u0443\u043C\u043A\u043E\u043C"
      ],
      "If specified, this template will create orders with the same price": [
        "\u042F\u043A\u0449\u043E \u0432\u043A\u0430\u0437\u0430\u043D\u043E, \u0446\u0435\u0439 \u0448\u0430\u0431\u043B\u043E\u043D \u0441\u0442\u0432\u043E\u0440\u0438\u0442\u044C \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0437 \u043E\u0434\u043D\u0430\u043A\u043E\u0432\u043E\u044E \u0446\u0456\u043D\u043E\u044E"
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        "\u0421\u043A\u0456\u043B\u044C\u043A\u0438 \u0447\u0430\u0441\u0443 \u0443 \u043A\u043B\u0456\u0454\u043D\u0442\u0430 \u0434\u043B\u044F \u0437\u0430\u0432\u0435\u0440\u0448\u0435\u043D\u043D\u044F \u043E\u043F\u043B\u0430\u0442\u0438 \u043F\u0456\u0441\u043B\u044F \u0441\u0442\u0432\u043E\u0440\u0435\u043D\u043D\u044F \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F."
      ],
      "An amount is required": [
        "\u0421\u0443\u043C\u0430 \u043E\u0431\u043E\u0432'\u044F\u0437\u043A\u043E\u0432\u0430"
      ],
      "An order summary is required": [
        "\u041F\u0456\u0434\u0441\u0443\u043C\u043E\u043A \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043E\u0431\u043E\u0432'\u044F\u0437\u043A\u043E\u0432\u0438\u0439"
      ],
      'No more stock for product with ID "%1$s".': [
        ""
      ],
      "New order from template": [
        "\u041D\u043E\u0432\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u0434\u043B\u044F \u0448\u0430\u0431\u043B\u043E\u043D\u0443"
      ],
      "Amount of the order": [
        "\u0421\u0443\u043C\u0430 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Slug": [
        ""
      ],
      "Token family slug to use in URLs (for internal use only)": [
        "\u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0446\u0456\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F \u0432 URL (\u0442\u0456\u043B\u044C\u043A\u0438 \u0434\u043B\u044F \u0432\u043D\u0443\u0442\u0440\u0456\u0448\u043D\u044C\u043E\u0433\u043E \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F)"
      ],
      "Kind": [
        ""
      ],
      "Token family kind": [
        ""
      ],
      "User-readable token family name": [
        ""
      ],
      "Token family description for customers": [
        "\u043E\u043F\u0438\u0441 \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443 \u0434\u043B\u044F \u043A\u043B\u0456\u0454\u043D\u0442\u0456\u0432"
      ],
      "Valid After": [
        "\u0414\u0456\u0439\u0441\u043D\u0438\u0439 \u0434\u043E"
      ],
      "Token family can issue tokens after this date": [
        ""
      ],
      "Valid Before": [
        "\u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u0439 \u0444\u043E\u0440\u043C\u0430\u0442"
      ],
      "Token family can issue tokens until this date": [
        ""
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        ""
      ],
      "Token Families": [
        ""
      ],
      "Add token family": [
        ""
      ],
      "Go to token family update page": [
        "\u043F\u0435\u0440\u0435\u0439\u0442\u0438 \u043D\u0430 \u0441\u0442\u043E\u0440\u0456\u043D\u043A\u0443 \u043E\u043D\u043E\u0432\u043B\u0435\u043D\u043D\u044F \u043F\u0440\u043E\u0434\u0443\u043A\u0442\u0443"
      ],
      "Remove this token family from the database": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0446\u0435\u0439 \u043F\u0440\u043E\u0434\u0443\u043A\u0442 \u0437 \u0431\u0430\u0437\u0438 \u0434\u0430\u043D\u0438\u0445"
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        "\u0428\u0430\u0431\u043B\u043E\u043D\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +."
      ],
      "Token family has been deleted.": [
        '\u0421\u0456\u043C\u0435\u0439\u0441\u0442\u0432\u043E \u0442\u043E\u043A\u0435\u043D\u0456\u0432 "%1$s" (SLUG: %2$s) \u0432\u0438\u0434\u0430\u043B\u0435\u043D\u043E'
      ],
      "Delete token family": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0441\u0456\u043C\u0435\u0439\u0441\u0442\u0432\u043E \u0442\u043E\u043A\u0435\u043D\u0456\u0432"
      ],
      'Delete the token family "%1$s"': [
        "\u0412\u0438\u0434\u0430\u043B\u0435\u043D\u043D\u044F \u0441\u0456\u043C\u0435\u0439\u0441\u0442\u0432\u0430 \u0442\u043E\u043A\u0435\u043D\u0456\u0432 %1$s ."
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        "\u042F\u043A\u0449\u043E \u0432\u0438 \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0435 \u0441\u0456\u043C\u0435\u0439\u0441\u0442\u0432\u043E \u0442\u043E\u043A\u0435\u043D\u0456\u0432 %1$s (Slug: %2$s), \u0432\u0441\u0456 \u0432\u0438\u043F\u0443\u0449\u0435\u043D\u0456 \u0442\u043E\u043A\u0435\u043D\u0438 \u0441\u0442\u0430\u043D\u0443\u0442\u044C \u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u043C\u0438."
      ],
      "Deleting a token family %1$s .": [
        "\u0412\u0438\u0434\u0430\u043B\u0435\u043D\u043D\u044F \u0441\u0456\u043C\u0435\u0439\u0441\u0442\u0432\u0430 \u0442\u043E\u043A\u0435\u043D\u0456\u0432 %1$s ."
      ],
      "can't be undone": [
        "\u043D\u0435 \u043C\u043E\u0436\u0435 \u0431\u0443\u0442\u0438 \u043F\u043E\u0440\u043E\u0436\u043D\u0456\u043C"
      ],
      "Token Family: %1$s": [
        "\u0421\u0456\u043C\u0435\u0439\u0441\u0442\u0432\u043E \u0442\u043E\u043A\u0435\u043D\u0456\u0432: %1$s"
      ],
      "Incoming wire transfers": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "Load more wire transfers preceding the first one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0456\u0432 \u0434\u043E \u043F\u0435\u0440\u0448\u043E\u0433\u043E"
      ],
      "Expected credit": [
        "\u041E\u0447\u0456\u043A\u0443\u0432\u0430\u043D\u0438\u0439 \u0431\u0430\u043B\u0430\u043D\u0441"
      ],
      "Confirmed": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0436\u0435\u043D\u043E"
      ],
      "Validated": [
        "\u0414\u0456\u0439\u0441\u043D\u0438\u0439 \u0434\u043E"
      ],
      "Executed on": [
        "\u0412\u0438\u043A\u043E\u043D\u0430\u043D\u043E \u043E"
      ],
      "yes": [
        "\u0442\u0430\u043A"
      ],
      "no": [
        "\u043D\u0456"
      ],
      "never": [
        "\u043D\u0456\u043A\u043E\u043B\u0438"
      ],
      "unknown": [
        "\u043D\u0435\u0432\u0456\u0434\u043E\u043C\u043E"
      ],
      "Load more transfers after the last one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0456\u0432 \u043F\u0456\u0441\u043B\u044F \u043E\u0441\u0442\u0430\u043D\u043D\u044C\u043E\u0433\u043E"
      ],
      "Verified wire transfers": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "Verified": [
        "\u041F\u0435\u0440\u0435\u0432\u0456\u0440\u0435\u043D\u043E"
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        "\u041F\u0435\u0440\u0435\u043A\u0430\u0437\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "All accounts": [
        "\u0420\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "Filter by account address": [
        "\u0444\u0456\u043B\u044C\u0442\u0440\u0443\u0432\u0430\u0442\u0438 \u0437\u0430 \u0430\u0434\u0440\u0435\u0441\u043E\u044E \u0440\u0430\u0445\u0443\u043D\u043A\u0443"
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F, \u0434\u0435 \u043A\u043B\u0456\u0454\u043D\u0442\u0438 \u0437\u0430\u043F\u043B\u0430\u0442\u0438\u043B\u0438, \u0430\u043B\u0435 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0456 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0438 \u0432\u0456\u0434 \u043F\u043E\u0441\u0442\u0430\u0447\u0430\u043B\u044C\u043D\u0438\u043A\u0430 \u043F\u043B\u0430\u0442\u0435\u0436\u0456\u0432 \u0449\u0435 \u043D\u0435 \u0432\u0438\u043A\u043E\u043D\u0430\u043D\u0456"
      ],
      "Only show wire transfers confirmed by the merchant": [
        "\u043F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u043B\u0438\u0448\u0435 \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0438, \u043F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0436\u0435\u043D\u0456 \u043F\u0440\u043E\u0434\u0430\u0432\u0446\u0435\u043C"
      ],
      "Wire transfer already confirmed.": [
        "\u0422\u0435\u0440\u043C\u0456\u043D \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        "\u043F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0438\u0442\u0438 \u0442\u043E\u0439 \u0441\u0430\u043C\u0438\u0439 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        "\u0425\u0440\u043E\u043D\u043E\u043B\u043E\u0433\u0456\u044F"
      ],
      "Transfer ID": [
        "\u041F\u0435\u0440\u0435\u043A\u0430\u0437\u0438"
      ],
      "It's not the same.": [
        "\u043D\u0435 \u0441\u043F\u0456\u0432\u043F\u0430\u0434\u0430\u0454"
      ],
      'You are deleting the instance with ID "%1$s"': [
        "\u0412\u0438 \u043E\u043D\u043E\u0432\u043B\u044E\u0454\u0442\u0435 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443 \u0437 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457 \u0437 \u0456\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440\u043E\u043C %1$s"
      ],
      "Instance": [
        "\u0406\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044F"
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        "\u041E\u0447\u0438\u0441\u0442\u0438\u0442\u0438"
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        ""
      ],
      "Instance id": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Delete this instance": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044E"
      ],
      "Must be one of '%1$s'": [
        "\u043F\u043E\u0432\u0438\u043D\u043D\u043E \u0431\u0443\u0442\u0438 \u043E\u0434\u043D\u0438\u043C \u0456\u0437 '%1$s'"
      ],
      "URL is invalid": [
        "\u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u0439"
      ],
      "Webhook ID to use": [
        "ID \u0432\u0435\u0431\u0445\u0443\u043A\u0430 \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F"
      ],
      "Event": [
        "\u041F\u043E\u0434\u0456\u044F"
      ],
      "Payment": [
        "URI \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "Order settled": [
        "\u0421\u0442\u0430\u0442\u0443\u0441 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Category added": [
        "\u0448\u0430\u0431\u043B\u043E\u043D \u0443\u0441\u043F\u0456\u0448\u043D\u043E \u0432\u0438\u0434\u0430\u043B\u0435\u043D\u043E"
      ],
      "Category updated": [
        "\u041F\u0456\u0434\u0442\u0432\u0435\u0440\u0434\u0436\u0435\u043D\u043E"
      ],
      "Inventory added": [
        ""
      ],
      "Inventory updated": [
        "ID \u0432\u0435\u0431\u0445\u0443\u043A\u0430 \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F"
      ],
      "Inventory deleted": [
        ""
      ],
      "The event of the webhook: why the webhook is used": [
        "\u041F\u043E\u0434\u0456\u044F \u0432\u0435\u0431\u0445\u0443\u043A\u0430: \u0447\u043E\u043C\u0443 \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u043E\u0432\u0443\u0454\u0442\u044C\u0441\u044F \u0432\u0435\u0431\u0445\u0443\u043A"
      ],
      "Method": [
        "\u041C\u0435\u0442\u043E\u0434"
      ],
      "GET": [
        ""
      ],
      "POST": [
        ""
      ],
      "PUT": [
        ""
      ],
      "PATCH": [
        ""
      ],
      "HEAD": [
        ""
      ],
      "Method used by the webhook": [
        "\u041C\u0435\u0442\u043E\u0434, \u0449\u043E \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u043E\u0432\u0443\u0454\u0442\u044C\u0441\u044F \u0432\u0435\u0431\u0445\u0443\u043A\u043E\u043C"
      ],
      "URL": [
        "URL"
      ],
      "URL of the webhook where the customer will be redirected": [
        "URL \u0432\u0435\u0431\u0445\u0443\u043A\u0430, \u043A\u0443\u0434\u0438 \u0431\u0443\u0434\u0435 \u043F\u0435\u0440\u0435\u043D\u0430\u043F\u0440\u0430\u0432\u043B\u0435\u043D\u0438\u0439 \u043A\u043B\u0456\u0454\u043D\u0442"
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        ""
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        ""
      ],
      "The short list of variables are:": [
        ""
      ],
      "order's description": [
        "\u043E\u043F\u0438\u0441"
      ],
      "order's price": [
        "\u0426\u0456\u043D\u0430 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "order's unique identification": [
        ""
      ],
      "the amount that was being refunded": [
        "\u0441\u0443\u043C\u0430 \u0434\u043E \u043F\u043E\u0432\u0435\u0440\u043D\u0435\u043D\u043D\u044F"
      ],
      "the reason entered by the merchant staff for granting the refund": [
        ""
      ],
      "time of the refund in nanoseconds since 1970": [
        ""
      ],
      "Http header": [
        "\u0417\u0430\u0433\u043E\u043B\u043E\u0432\u043E\u043A"
      ],
      "Header template of the webhook": [
        "\u0428\u0430\u0431\u043B\u043E\u043D \u0437\u0430\u0433\u043E\u043B\u043E\u0432\u043A\u0430 \u0432\u0435\u0431\u0445\u0443\u043A\u0430"
      ],
      "Http body": [
        ""
      ],
      "Body template used by the webhook.": [
        "\u0428\u0430\u0431\u043B\u043E\u043D \u0442\u0456\u043B\u0430 \u0432\u0435\u0431\u0445\u0443\u043A\u0430"
      ],
      "Webhook deleted successfully": [
        "\u0432\u0435\u0431\u0445\u0443\u043A \u0443\u0441\u043F\u0456\u0448\u043D\u043E \u0432\u0438\u0434\u0430\u043B\u0435\u043D\u043E"
      ],
      "Add new webhooks": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0456 \u0432\u0435\u0431\u0445\u0443\u043A\u0438"
      ],
      "Load more webhooks before the first one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0432\u0435\u0431\u0445\u0443\u043A\u0456\u0432 \u0434\u043E \u043F\u0435\u0440\u0448\u043E\u0433\u043E"
      ],
      "Event type": [
        "\u0422\u0438\u043F \u043F\u043E\u0434\u0456\u0457"
      ],
      "Delete selected webhook from the database": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0438\u0431\u0440\u0430\u043D\u0438\u0439 \u0432\u0435\u0431\u0445\u0443\u043A \u0437 \u0431\u0430\u0437\u0438 \u0434\u0430\u043D\u0438\u0445"
      ],
      "Load more webhooks after the last one": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u0431\u0456\u043B\u044C\u0448\u0435 \u0432\u0435\u0431\u0445\u0443\u043A\u0456\u0432 \u043F\u0456\u0441\u043B\u044F \u043E\u0441\u0442\u0430\u043D\u043D\u044C\u043E\u0433\u043E"
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        "\u0412\u0435\u0431\u0445\u0443\u043A\u0456\u0432 \u0449\u0435 \u043D\u0435\u043C\u0430\u0454, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      "Webhook updated": [
        "ID \u0432\u0435\u0431\u0445\u0443\u043A\u0430 \u0434\u043B\u044F \u0432\u0438\u043A\u043E\u0440\u0438\u0441\u0442\u0430\u043D\u043D\u044F"
      ],
      "Header": [
        "\u0417\u0430\u0433\u043E\u043B\u043E\u0432\u043E\u043A"
      ],
      "Body": [
        "\u0422\u0456\u043B\u043E"
      ],
      "Body template used by the webhook": [
        "\u0428\u0430\u0431\u043B\u043E\u043D \u0442\u0456\u043B\u0430 \u0432\u0435\u0431\u0445\u0443\u043A\u0430"
      ],
      "There is another instance with this username.": [
        ""
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        "\u041A\u043E\u043D\u0442\u0430\u043A\u0442\u043D\u0438\u0439 email"
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        "\u0421\u0442\u0432\u043E\u0440\u0435\u043D\u043E \u043E"
      ],
      "Password reset": [
        ""
      ],
      'Resetting access to the instance "%1$s"': [
        "\u0412\u0438 \u0432\u0441\u0442\u0430\u043D\u043E\u0432\u043B\u044E\u0454\u0442\u0435 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443 \u0434\u043B\u044F \u043D\u043E\u0432\u043E\u0457 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Reset": [
        ""
      ],
      "Language": [
        ""
      ],
      "Date format": [
        "\u043D\u0435\u0434\u0456\u0439\u0441\u043D\u0438\u0439 \u0444\u043E\u0440\u043C\u0430\u0442"
      ],
      "How the date is going to be displayed": [
        ""
      ],
      "Merchant type": [
        "\u0422\u0438\u043F \u043F\u043E\u0434\u0456\u0457"
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        ""
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        "\u0417\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        "\u0421\u0442\u0430\u0442\u0443\u0441 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        "\u0421\u0442\u0430\u0442\u0443\u0441 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        "\u0426\u0456\u043D\u0430 \u0437\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "hour": [
        "\u0433\u043E\u0434\u0438\u043D\u0438"
      ],
      "day": [
        "\u0434\u043D\u0456"
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        "\u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0441\u0456 \u0444\u0456\u043B\u044C\u0442\u0440\u0438"
      ],
      "Time range": [
        "\u0425\u0440\u043E\u043D\u043E\u043B\u043E\u0433\u0456\u044F"
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        "\u041F\u043E\u0442\u043E\u0447\u043D\u0438\u0439"
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        "\u0422\u0430\u0439\u043C-\u0430\u0443\u0442 \u043E\u043F\u043B\u0430\u0442\u0438"
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        "URI, \u0449\u043E \u0432\u043A\u0430\u0437\u0443\u0454 \u043D\u0430 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A \u0434\u043B\u044F \u0437\u0430\u0440\u0430\u0445\u0443\u0432\u0430\u043D\u043D\u044F \u0434\u043E\u0445\u043E\u0434\u0443."
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        ""
      ],
      "Hide for today": [
        "\u0421\u0445\u043E\u0432\u0430\u0442\u0438 \u043D\u0430 \u0441\u044C\u043E\u0433\u043E\u0434\u043D\u0456"
      ],
      "KYC verification needed": [
        "\u041E\u0447\u0456\u043A\u0443\u0432\u0430\u043D\u043D\u044F \u043F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0438 KYC"
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        ""
      ],
      "The request reached a timeout, check your connection.": [
        ""
      ],
      "The request was cancelled.": [
        ""
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        ""
      ],
      "The server's response was malformed.": [
        ""
      ],
      "Could not complete the request due to a network problem.": [
        ""
      ],
      "Unexpected request error.": [
        "\u041D\u0435\u0441\u043F\u043E\u0434\u0456\u0432\u0430\u043D\u0430 \u043F\u043E\u043C\u0438\u043B\u043A\u0430"
      ],
      "Unexpected error.": [
        "\u041D\u0435\u0441\u043F\u043E\u0434\u0456\u0432\u0430\u043D\u0430 \u043F\u043E\u043C\u0438\u043B\u043A\u0430"
      ],
      "Add new instance": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0443 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044E"
      ],
      "Edit": [
        "\u0420\u0435\u0434\u0430\u0433\u0443\u0432\u0430\u0442\u0438"
      ],
      "Change password": [
        ""
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        "\u0429\u0435 \u043D\u0435\u043C\u0430\u0454 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0439, \u0434\u043E\u0434\u0430\u0439\u0442\u0435 \u0431\u0456\u043B\u044C\u0448\u0435, \u043D\u0430\u0442\u0438\u0441\u043D\u0443\u0432\u0448\u0438 \u0437\u043D\u0430\u043A +"
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        '\u0406\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044F "%1$s" (ID: %2$s) \u0431\u0443\u043B\u0430 \u0432\u0438\u0434\u0430\u043B\u0435\u043D\u0430'
      ],
      "Only show active instances": [
        "\u041F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u0442\u0456\u043B\u044C\u043A\u0438 \u0430\u043A\u0442\u0438\u0432\u043D\u0456 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Active": [
        "\u0410\u043A\u0442\u0438\u0432\u043D\u0456"
      ],
      "Only show deleted instances": [
        "\u041F\u043E\u043A\u0430\u0437\u0443\u0432\u0430\u0442\u0438 \u0442\u0456\u043B\u044C\u043A\u0438 \u0432\u0438\u0434\u0430\u043B\u0435\u043D\u0456 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Deleted": [
        "\u0412\u0438\u0434\u0430\u043B\u0435\u043D\u043E"
      ],
      "Show all instances": [
        "\u041F\u043E\u043A\u0430\u0437\u0430\u0442\u0438 \u0432\u0441\u0456 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "Delete instance": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044E"
      ],
      'Delete the instance "%1$s"': [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u043D\u043E\u0432\u0456\u0448\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        ""
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        ""
      ],
      "Deleting an instance %1$s": [
        "\u0437\u0430\u0432\u0430\u043D\u0442\u0430\u0436\u0438\u0442\u0438 \u043D\u043E\u0432\u0456\u0448\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "This cannot be undone!": [
        "\u043D\u0435 \u043C\u043E\u0436\u0435 \u0431\u0443\u0442\u0438 \u043F\u043E\u0440\u043E\u0436\u043D\u0456\u043C"
      ],
      "Purge the instance": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044E"
      ],
      'Purge the instance "%1$s"': [
        "\u0412\u0438 \u0432\u0441\u0442\u0430\u043D\u043E\u0432\u043B\u044E\u0454\u0442\u0435 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443 \u0434\u043B\u044F \u043D\u043E\u0432\u043E\u0457 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        ""
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        ""
      ],
      "Purging an instance %1$s": [
        ""
      ],
      "checking compatibility with server...": [
        ""
      ],
      "Contacting the server failed": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u043F\u0456\u0434\u043A\u043B\u044E\u0447\u0438\u0442\u0438\u0441\u044F \u0434\u043E \u0441\u0435\u0440\u0432\u0435\u0440\u0430."
      ],
      "The server version is not supported": [
        ""
      ],
      'Supported version "%1$s", server version "%2$s".': [
        ""
      ],
      "Deleting": [
        "\u0412\u0438\u0434\u0430\u043B\u0435\u043D\u043D\u044F"
      ],
      "Changing": [
        "\u0417\u043C\u0456\u043D\u0430"
      ],
      "Manage password": [
        "\u041A\u0435\u0440\u0443\u0432\u0430\u043D\u043D\u044F \u0437\u0430\u043F\u0430\u0441\u0430\u043C\u0438"
      ],
      "Business Name": [
        "\u041D\u0430\u0437\u0432\u0430 \u0431\u0456\u0437\u043D\u0435\u0441\u0443"
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        "\u041D\u0430\u043B\u0430\u0448\u0442\u0443\u0432\u0430\u043D\u043D\u044F"
      ],
      "title%1$s: New bank account": [
        "\u0411\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "title%1$s: Bank accounts": [
        "\u0411\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u0438\u0439 \u0440\u0430\u0445\u0443\u043D\u043E\u043A"
      ],
      "title%1$s: Update bank Account": [
        "\u041F\u0435\u0440\u0435\u0432\u0456\u0440\u043A\u0430 \u0434\u0456\u0439\u0441\u043D\u043E\u0441\u0442\u0456 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u0440\u0430\u0445\u0443\u043D\u043A\u0443: %1$s"
      ],
      "title%1$s: Orders": [
        "\u0417\u0430\u043C\u043E\u0432\u043B\u0435\u043D\u043D\u044F"
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        ""
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        "\u043D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u043E\u043D\u043E\u0432\u0438\u0442\u0438 \u043F\u0440\u043E\u0434\u0443\u043A\u0442"
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        "\u043D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u043E\u043D\u043E\u0432\u0438\u0442\u0438 \u0448\u0430\u0431\u043B\u043E\u043D"
      ],
      "title%1$s: Wire transfers": [
        "\u0406\u0434\u0435\u043D\u0442\u0438\u0444\u0456\u043A\u0430\u0442\u043E\u0440 \u0431\u0430\u043D\u043A\u0456\u0432\u0441\u044C\u043A\u043E\u0433\u043E \u043F\u0435\u0440\u0435\u043A\u0430\u0437\u0443"
      ],
      "title%1$s: Webhooks": [
        "\u0412\u0435\u0431\u0445\u0443\u043A\u0438"
      ],
      "title%1$s: New webhook": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0456 \u0432\u0435\u0431\u0445\u0443\u043A\u0438"
      ],
      "title%1$s: Update webhook": [
        "\u043D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u0432\u0438\u0434\u0430\u043B\u0438\u0442\u0438 \u0432\u0435\u0431\u0445\u0443\u043A"
      ],
      "title%1$s: OTP devices": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0438\u0439 \u0440\u0435\u0437\u0435\u0440\u0432"
      ],
      "title%1$s: New OTP device": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0438\u0439 \u0440\u0435\u0437\u0435\u0440\u0432"
      ],
      "title%1$s: Update OTP device": [
        ""
      ],
      "title%1$s: New template": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "title%1$s: Update template": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "title%1$s: Templates": [
        "\u0428\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "title%1$s: Use template": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0456 \u0448\u0430\u0431\u043B\u043E\u043D\u0438"
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        ""
      ],
      "title%1$s: New token family": [
        ""
      ],
      "title%1$s: Update token family": [
        "\u041D\u0435 \u0432\u0434\u0430\u043B\u043E\u0441\u044F \u043E\u043D\u043E\u0432\u0438\u0442\u0438 \u0441\u0456\u043C\u0435\u0439\u0441\u0442\u0432\u043E \u0442\u043E\u043A\u0435\u043D\u0456\u0432"
      ],
      "title%1$s: Access tokens": [
        "\u0422\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "title%1$s: New access token": [
        "\u041D\u043E\u0432\u0438\u0439 \u0442\u043E\u043A\u0435\u043D \u0434\u043E\u0441\u0442\u0443\u043F\u0443"
      ],
      "titleNew instance": [
        "\u0434\u043E\u0434\u0430\u0442\u0438 \u043D\u043E\u0432\u0443 \u0456\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u044E"
      ],
      "titleInstances": [
        "\u0406\u043D\u0441\u0442\u0430\u043D\u0446\u0456\u0457"
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=3; plural=n%10==1 && n%100!=11 ? 0 : n%10>=2 && n%10<=4 && (n%100<10 || n%100>=20) ? 1 : 2;",
  "lang": "uk",
  "completeness": 64
};
strings["tr"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=2; plural=n != 1;",
        "lang": "tr"
      },
      "Cancel": [
        "\u0130ptal"
      ],
      "%1$s": [
        "%1$s"
      ],
      "Close": [
        ""
      ],
      "Continue": [
        ""
      ],
      "Clear": [
        ""
      ],
      "Confirm": [
        ""
      ],
      "Correct form": [
        ""
      ],
      "Comparing account details": [
        ""
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        ""
      ],
      "Field": [
        ""
      ],
      "In the form": [
        ""
      ],
      "Reported": [
        ""
      ],
      "Type": [
        ""
      ],
      "IBAN": [
        ""
      ],
      "Address": [
        ""
      ],
      "Host": [
        ""
      ],
      "Account ID": [
        "Hesap"
      ],
      "The account owner's legal name": [
        ""
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        "Kasaba"
      ],
      "Bank host": [
        "Banka hesab\u0131"
      ],
      "Bank account": [
        "Banka hesab\u0131"
      ],
      "BIC": [
        ""
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        ""
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        ""
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        ""
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        ""
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        ""
      ],
      "Subject": [
        "Konu"
      ],
      "Receiver postal code": [
        "Teslim tarihi"
      ],
      "Receiver town": [
        "Teslim tarihi"
      ],
      "Step 3:": [
        ""
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        ""
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        ""
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        ""
      ],
      "Operation in progress...": [
        ""
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        ""
      ],
      "days": [
        ""
      ],
      "hours": [
        ""
      ],
      "minutes": [
        ""
      ],
      "seconds": [
        ""
      ],
      "Forever": [
        ""
      ],
      "%1$sM": [
        ""
      ],
      "%1$sY": [
        ""
      ],
      "%1$sd": [
        ""
      ],
      "%1$sh": [
        ""
      ],
      "%1$smin": [
        ""
      ],
      "%1$ssec": [
        ""
      ],
      "Change value to never": [
        ""
      ],
      "Change value to empty": [
        ""
      ],
      "Image must be smaller than 1 MB": [
        ""
      ],
      "Add": [
        ""
      ],
      "Remove": [
        ""
      ],
      "Country": [
        "\xDClke"
      ],
      "Building number": [
        "Bina numaras\u0131"
      ],
      "Building name": [
        "Bina ad\u0131"
      ],
      "Street": [
        "Sokak"
      ],
      "Post code": [
        "Posta kodu"
      ],
      "Town location": [
        "Kasaba konumu"
      ],
      "District": [
        "Semt"
      ],
      "Country subdivision": [
        ""
      ],
      "Additional address lines": [
        ""
      ],
      "Orders": [
        ""
      ],
      "Inventory": [
        ""
      ],
      "Categories": [
        ""
      ],
      "Wire transfers": [
        ""
      ],
      "Templates": [
        ""
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        ""
      ],
      "OTP Devices": [
        ""
      ],
      "Webhooks": [
        ""
      ],
      "Settings": [
        ""
      ],
      "Password": [
        ""
      ],
      "Access tokens": [
        ""
      ],
      "Connection": [
        ""
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        ""
      ],
      "New": [
        ""
      ],
      "List": [
        ""
      ],
      "Log out": [
        ""
      ],
      "Username": [
        ""
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        ""
      ],
      "Business name": [
        ""
      ],
      "Legal name of the business represented by this instance.": [
        ""
      ],
      "Email": [
        ""
      ],
      "Contact email": [
        ""
      ],
      "Phone number": [
        ""
      ],
      "Contact phone": [
        ""
      ],
      "Website URL": [
        ""
      ],
      "URL.": [
        ""
      ],
      "Logo": [
        ""
      ],
      "Logo image.": [
        ""
      ],
      "Physical location of the merchant.": [
        ""
      ],
      "Jurisdiction": [
        ""
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        ""
      ],
      "Pay transaction fee": [
        ""
      ],
      "Cover the transaction cost or pass it on to the user.": [
        ""
      ],
      "Default payment delay": [
        ""
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        ""
      ],
      "Default refund delay": [
        ""
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        ""
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        ""
      ],
      "Default wire transfer rounding interval": [
        ""
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        ""
      ],
      "To second": [
        ""
      ],
      "To minute": [
        ""
      ],
      "To hour": [
        ""
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        ""
      ],
      "Required": [
        ""
      ],
      "Unauthorized": [
        ""
      ],
      "Sending the code failed": [
        ""
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        ""
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        ""
      ],
      "Verify": [
        ""
      ],
      "Failed to send the verification code.": [
        ""
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        ""
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        ""
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        ""
      ],
      "Login required": [
        ""
      ],
      "instance name": [
        ""
      ],
      "current password": [
        ""
      ],
      "Forgot password": [
        ""
      ],
      "Create new account": [
        "Banka hesab\u0131"
      ],
      "Invalid": [
        ""
      ],
      "Doesn't have the pattern of an email": [
        ""
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        "Ge\xE7erlilik"
      ],
      "Max 7 lines": [
        ""
      ],
      "Doesn't match": [
        ""
      ],
      "Instance created": [
        ""
      ],
      "Unauthorized.": [
        ""
      ],
      "Conflict.": [
        ""
      ],
      "Not found.": [
        ""
      ],
      "New password": [
        ""
      ],
      "Next password to be used": [
        ""
      ],
      "Repeat password": [
        ""
      ],
      "Confirm the same password": [
        ""
      ],
      "Please complete the marked fields and choose authorization method": [
        ""
      ],
      "Confirm operation": [
        "Kasaba konumu"
      ],
      "Check the password.": [
        ""
      ],
      "Instance not found.": [
        ""
      ],
      "Description": [
        ""
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        ""
      ],
      "Time the access token will be valid.": [
        ""
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        ""
      ],
      "Read only": [
        ""
      ],
      "All": [
        ""
      ],
      "Order simple": [
        ""
      ],
      "Order Point-of-Sale": [
        ""
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        ""
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        ""
      ],
      "Please complete the marked fields": [
        ""
      ],
      "Access token created": [
        ""
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        ""
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        ""
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        ""
      ],
      "Create access token": [
        "Olu\u015Fturulma"
      ],
      "Load more devices before the first one": [
        ""
      ],
      "Load first page": [
        ""
      ],
      "Created at": [
        "Olu\u015Fturulma"
      ],
      "Expires at": [
        ""
      ],
      "Never": [
        ""
      ],
      "Remove this access token": [
        ""
      ],
      "Delete": [
        ""
      ],
      "Load more devices after the last one": [
        ""
      ],
      "Load next page": [
        ""
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        ""
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        "Olu\u015Fturulma"
      ],
      "Deleting an access token cannot be undone.": [
        ""
      ],
      "This is not a valid Bitcoin address.": [
        ""
      ],
      "This is not a valid Ethereum address.": [
        ""
      ],
      "This is not a valid host.": [
        ""
      ],
      "IBANs usually have more than 4 digits.": [
        ""
      ],
      "IBANs usually have fewer than 34 digits.": [
        ""
      ],
      "The IBAN's country code could not be retrieved.": [
        ""
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        ""
      ],
      "This account is not allowed.": [
        ""
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        ""
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        ""
      ],
      "Select a wire method...": [
        ""
      ],
      "Routing": [
        ""
      ],
      "Routing number": [
        "Bina numaras\u0131"
      ],
      "Account": [
        "Hesap"
      ],
      "Account number": [
        "Hesap"
      ],
      "Code": [
        ""
      ],
      "Business Identifier Code": [
        ""
      ],
      "International Bank Account Number": [
        ""
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        ""
      ],
      "Bitcoin protocol": [
        ""
      ],
      "Ethereum protocol": [
        ""
      ],
      "Interledger protocol": [
        ""
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        ""
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        ""
      ],
      "Invalid url": [
        "Ge\xE7erlilik"
      ],
      "URL must end with a '/'": [
        ""
      ],
      "URL must not contain params": [
        ""
      ],
      "URL must not hash param": [
        ""
      ],
      "Conflict": [
        ""
      ],
      'Server replied with "bad request".': [
        ""
      ],
      "Unauthorized, check credentials.": [
        ""
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        ""
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        ""
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        "Hesap:"
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        ""
      ],
      "Endpoint URL": [
        ""
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        ""
      ],
      "Auth type": [
        ""
      ],
      "Choose the authentication type for the account info URL": [
        ""
      ],
      "Without authentication": [
        ""
      ],
      "With username and password": [
        ""
      ],
      "With token": [
        ""
      ],
      "Do not change": [
        ""
      ],
      "Username to access the account information.": [
        ""
      ],
      "Password to access the account information.": [
        ""
      ],
      "Access token to access the account information.": [
        ""
      ],
      "Match": [
        ""
      ],
      "Check where the information match against the server info.": [
        ""
      ],
      "Compare info from server with account form": [
        ""
      ],
      "Test": [
        ""
      ],
      "Server didn't like the request we made.": [
        ""
      ],
      "Account details": [
        "Hesap"
      ],
      "Check if the information matches the server info.": [
        ""
      ],
      "The bank account has been deleted.": [
        ""
      ],
      "Bank accounts": [
        "Banka hesab\u0131"
      ],
      "Add new account": [
        "Banka hesab\u0131"
      ],
      "Owner's name": [
        ""
      ],
      "Delete selected accounts from the database": [
        ""
      ],
      "There are no accounts yet, add more pressing the + sign": [
        ""
      ],
      "You must provide a bank account to receive payments.": [
        ""
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        ""
      ],
      "Name": [
        ""
      ],
      "Category name": [
        ""
      ],
      "Category deleted": [
        ""
      ],
      "Add new devices": [
        ""
      ],
      "ID": [
        ""
      ],
      "Total products": [
        "\xDCr\xFCnler"
      ],
      "Delete selected category from the database": [
        ""
      ],
      "There are no categories yet, add more pressing the + sign": [
        ""
      ],
      "Id:": [
        ""
      ],
      "Name of the category": [
        ""
      ],
      "Products": [
        "\xDCr\xFCnler"
      ],
      "Image": [
        ""
      ],
      "Load more products after the last one": [
        ""
      ],
      "There are no products in this category.": [
        ""
      ],
      "Account's KYC status": [
        "Hesap"
      ],
      "Exchange": [
        ""
      ],
      "Status": [
        ""
      ],
      "Ok": [
        ""
      ],
      "Action required": [
        ""
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        ""
      ],
      "More information required.": [
        ""
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        ""
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        ""
      ],
      "No pending kyc verification!": [
        ""
      ],
      "Wire account is invalid": [
        ""
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        ""
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        ""
      ],
      "The payment provider requires more information.": [
        ""
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        ""
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        ""
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        ""
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        ""
      ],
      "Enter description or id": [
        ""
      ],
      "No match found for that description or ID.": [
        ""
      ],
      "You must enter a valid product identifier.": [
        ""
      ],
      "Quantity must be greater than 0!": [
        ""
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        ""
      ],
      "Search product": [
        ""
      ],
      "Quantity": [
        ""
      ],
      "How many products will be added": [
        ""
      ],
      "Add from inventory": [
        ""
      ],
      "This product has %1$s applicable taxes configured.": [
        ""
      ],
      "No taxes configured for this product.": [
        ""
      ],
      "Amount": [
        "Miktar"
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        ""
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        ""
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        ""
      ],
      "Add tax to the tax list": [
        ""
      ],
      "Describe and add a product that is not in the inventory list": [
        ""
      ],
      "Add custom product": [
        ""
      ],
      "Complete information of the product": [
        ""
      ],
      "Must be a number.": [
        ""
      ],
      "Must be greater than 0.": [
        ""
      ],
      "Photo of the product.": [
        ""
      ],
      "Full product description.": [
        ""
      ],
      "Unit name": [
        ""
      ],
      "Name of the product unit.": [
        ""
      ],
      "Price per unit": [
        ""
      ],
      "Amount in the current currency.": [
        ""
      ],
      "How many products will be added.": [
        ""
      ],
      "Taxes": [
        ""
      ],
      "Total price": [
        "\xDCr\xFCnler"
      ],
      "Must be greater than 0": [
        ""
      ],
      "Must have a refund deadline": [
        ""
      ],
      "Auto refund can't be after refund deadline": [
        ""
      ],
      "Must be in the future": [
        ""
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        ""
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        ""
      ],
      "Manage products in order": [
        ""
      ],
      "%1$s products with a total price of %2$s.": [
        ""
      ],
      "Manage list of products in the order.": [
        ""
      ],
      "Remove this product from the order.": [
        ""
      ],
      "Products price sum": [
        "\xDCr\xFCnler"
      ],
      "Total product price added up": [
        "\xDCr\xFCnler"
      ],
      "Order price": [
        ""
      ],
      "Amount to be paid by the customer": [
        ""
      ],
      "Final order price": [
        ""
      ],
      "Summary": [
        ""
      ],
      "Title of the order to be shown to the customer": [
        ""
      ],
      "Shipping and fulfillment": [
        ""
      ],
      "Delivery date": [
        ""
      ],
      "Deadline for physical delivery assured by the merchant.": [
        ""
      ],
      "Delivery location": [
        "Kasaba konumu"
      ],
      "Address where the products will be delivered": [
        ""
      ],
      "Fulfillment URL": [
        ""
      ],
      "URL to which the user will be redirected after successful payment.": [
        ""
      ],
      "Fulfillment message": [
        ""
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        ""
      ],
      "Override default Taler payment settings for this order": [
        ""
      ],
      "Payment time": [
        ""
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        ""
      ],
      "Default": [
        ""
      ],
      "Refund time": [
        ""
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        ""
      ],
      "Wire transfer time": [
        ""
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        ""
      ],
      "Auto-refund time": [
        ""
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        ""
      ],
      "Maximum fee": [
        ""
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        ""
      ],
      "Create token": [
        "Olu\u015Fturulma"
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        ""
      ],
      "Minimum age required": [
        ""
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        ""
      ],
      "Minimum age defined by the products is %1$s": [
        ""
      ],
      "No product with age restriction in this order": [
        ""
      ],
      "Additional information": [
        ""
      ],
      "Custom information to be included in the contract for this order.": [
        ""
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        ""
      ],
      "remove": [
        ""
      ],
      "Custom field name": [
        "Bina ad\u0131"
      ],
      "Disabled": [
        ""
      ],
      "No deadline": [
        ""
      ],
      "Deadline at %1$s": [
        ""
      ],
      "Select date to show nearby orders": [
        ""
      ],
      "Only show unpaid orders": [
        ""
      ],
      "Only show paid orders": [
        ""
      ],
      "Paid": [
        ""
      ],
      "Only show orders with refunds": [
        ""
      ],
      "Refunded": [
        ""
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        ""
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        ""
      ],
      "Completed": [
        ""
      ],
      "Remove all filters": [
        ""
      ],
      "Clear date filter": [
        ""
      ],
      "Jump to date (%1$s)": [
        ""
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        ""
      ],
      "Order id": [
        ""
      ],
      "Instance unknown": [
        ""
      ],
      "Order unknown": [
        ""
      ],
      "Create order": [
        "Olu\u015Fturulma"
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        ""
      ],
      "Refund": [
        ""
      ],
      "copy url": [
        ""
      ],
      "Load more orders after the last one": [
        ""
      ],
      "No orders have been found matching your query!": [
        ""
      ],
      "Duplicated": [
        ""
      ],
      "This value exceeds the refundable amount.": [
        ""
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        ""
      ],
      "refund": [
        ""
      ],
      "Reason": [
        ""
      ],
      "Amount to be refunded": [
        ""
      ],
      "Max refundable:": [
        ""
      ],
      "Choose one...": [
        ""
      ],
      "Requested by the customer": [
        ""
      ],
      "Other": [
        ""
      ],
      "Why this order is being refunded": [
        ""
      ],
      "More information to give context": [
        ""
      ],
      "now": [
        ""
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        ""
      ],
      "This wire transfer has been notified by the payment service provider.": [
        ""
      ],
      "This refund is waiting to be claimed by the customer.": [
        ""
      ],
      "This refund has been claimed by the customer.": [
        ""
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        ""
      ],
      "Human-readable description of the whole purchase": [
        ""
      ],
      "Total price for the transaction": [
        ""
      ],
      "URL for this purchase": [
        ""
      ],
      "Max fee": [
        ""
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        ""
      ],
      "Time when this contract was generated": [
        ""
      ],
      "Refund deadline": [
        ""
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        ""
      ],
      "Payment deadline": [
        ""
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        ""
      ],
      "Wire transfer deadline": [
        ""
      ],
      "Transfer deadline for the exchange": [
        ""
      ],
      "Time indicating when the order should be delivered": [
        ""
      ],
      "Where the order will be delivered": [
        ""
      ],
      "Auto-refund delay": [
        ""
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        ""
      ],
      "Extra info": [
        ""
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        ""
      ],
      "order created": [
        ""
      ],
      "pay deadline": [
        ""
      ],
      "refund deadline": [
        ""
      ],
      "delivery": [
        ""
      ],
      "Order": [
        ""
      ],
      "Claimed": [
        ""
      ],
      "Claimed at": [
        "Olu\u015Fturulma"
      ],
      "Timeline": [
        ""
      ],
      "Payment details": [
        "Hesap"
      ],
      "Order status": [
        ""
      ],
      "Unpaid": [
        ""
      ],
      "Product list": [
        "\xDCr\xFCnler"
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        ""
      ],
      "refund created: %1$s": [
        ""
      ],
      "refund created: %1$s: %2$s": [
        ""
      ],
      "refund taken: %1$s": [
        ""
      ],
      "refund taken: %1$s: %2$s": [
        ""
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        ""
      ],
      "Wired": [
        ""
      ],
      "Refund order": [
        ""
      ],
      "Not refundable": [
        ""
      ],
      "Next event in": [
        ""
      ],
      "The order was wired.": [
        ""
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        ""
      ],
      "Refund taken": [
        ""
      ],
      "Status URL": [
        ""
      ],
      "Refund URI": [
        ""
      ],
      "Pay at": [
        ""
      ],
      "Order status URL": [
        ""
      ],
      "Payment URI": [
        ""
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        ""
      ],
      "Invalid. Please enter letters and numbers only.": [
        ""
      ],
      "Just letters and numbers from 2 to 7": [
        ""
      ],
      "The size of the key must be 32 characters": [
        ""
      ],
      "Internal ID on the system": [
        ""
      ],
      "Useful to identify the device physically": [
        ""
      ],
      "Verification algorithm": [
        ""
      ],
      "Algorithm used to verify transactions in offline mode": [
        ""
      ],
      "Device key": [
        ""
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        ""
      ],
      "Your device needs to match exactly the same value": [
        ""
      ],
      "Generate random secret key": [
        ""
      ],
      "Random": [
        ""
      ],
      "Create another": [
        "Olu\u015Fturulma"
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        ""
      ],
      "OTP devices": [
        ""
      ],
      "Delete selected devices from the database": [
        ""
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        ""
      ],
      "Template id is unknown": [
        ""
      ],
      "The provided information is inconsistent with the current state of the template": [
        ""
      ],
      "Device:": [
        ""
      ],
      "Algorithm to use to verify transactions in offline mode": [
        ""
      ],
      "Not modified": [
        ""
      ],
      "Be sure to be very hard to guess or use the random generator": [
        ""
      ],
      "Your device need to have exactly the same value": [
        ""
      ],
      "Change key": [
        ""
      ],
      "Is not the same": [
        ""
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        ""
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        "Kasaba konumu"
      ],
      "Testing password change": [
        ""
      ],
      "Password changed": [
        ""
      ],
      "The current password is wrong.": [
        ""
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        "Hesap"
      ],
      "Add element to the list": [
        ""
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        ""
      ],
      "Manage stock": [
        ""
      ],
      "This product has been configured without stock control": [
        ""
      ],
      "Infinite": [
        ""
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        ""
      ],
      "Incoming": [
        ""
      ],
      "Lost": [
        ""
      ],
      "Current": [
        ""
      ],
      "Remove stock control for this product": [
        ""
      ],
      "without stock": [
        ""
      ],
      "Next restock": [
        ""
      ],
      "Warehouse address": [
        ""
      ],
      "Invalid amount": [
        "Ge\xE7erlilik"
      ],
      "Product identification to use in URLs (for internal use only).": [
        ""
      ],
      "Illustration of the product for customers.": [
        ""
      ],
      "Product name.": [
        "\xDCr\xFCnler"
      ],
      "Product description for customers.": [
        ""
      ],
      "Age restriction": [
        ""
      ],
      "Is this product restricted for customer below certain age?": [
        ""
      ],
      "Minimum age of the customer": [
        ""
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        ""
      ],
      "Example: kg, items or liters": [
        ""
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        ""
      ],
      "Stock": [
        ""
      ],
      "Inventory for products with finite supply (for internal use only).": [
        ""
      ],
      "Taxes included in the product price, exposed to customers.": [
        ""
      ],
      "Search by category description or id": [
        ""
      ],
      "Categories where this product will be listed on.": [
        ""
      ],
      "Add product to inventory": [
        ""
      ],
      "Product not found.": [
        ""
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        ""
      ],
      "Sold": [
        ""
      ],
      "Free": [
        ""
      ],
      "Go to product update page": [
        ""
      ],
      "Update": [
        ""
      ],
      "Remove this product from the database": [
        ""
      ],
      "Update the product with new price": [
        ""
      ],
      "Update product with new price": [
        ""
      ],
      "Confirm update": [
        "Kasaba konumu"
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        ""
      ],
      "Add more elements to the inventory": [
        ""
      ],
      "Report elements lost in the inventory": [
        ""
      ],
      "New price for the product": [
        ""
      ],
      "There are values with errors.": [
        ""
      ],
      "Update product with new stock and price": [
        ""
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        ""
      ],
      "Product (ID: %1$s) has been deleted": [
        ""
      ],
      "Jump to product with the given product ID": [
        ""
      ],
      "Product id": [
        "\xDCr\xFCnler"
      ],
      "Delete product": [
        "\xDCr\xFCnler"
      ],
      'Delete the product "%1$s"': [
        ""
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        ""
      ],
      "Deleting a product cannot be undone.": [
        ""
      ],
      "Product id:": [
        "\xDCr\xFCnler"
      ],
      "Must be greater that 0": [
        ""
      ],
      "Too short": [
        ""
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        ""
      ],
      "Name of the template in URLs.": [
        ""
      ],
      "Template name": [
        ""
      ],
      "Describe what this template stands for": [
        ""
      ],
      "Order summary": [
        ""
      ],
      "If specified here, this template will create orders with the same summary": [
        ""
      ],
      "Summary is editable": [
        ""
      ],
      "Allow the user to change the summary.": [
        ""
      ],
      "If specified here, this template will create orders with the same price": [
        ""
      ],
      "Amount is editable": [
        "Hesap"
      ],
      "Allow the user to select the amount to pay.": [
        ""
      ],
      "Currency is editable": [
        ""
      ],
      "Allow the user to change currency.": [
        ""
      ],
      "Supported currencies": [
        ""
      ],
      "Supported currencies: %1$s": [
        ""
      ],
      "Minimum age": [
        ""
      ],
      "Is this contract restricted to some age?": [
        ""
      ],
      "Payment timeout": [
        ""
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        ""
      ],
      "OTP device": [
        ""
      ],
      "Use to verify transactions in offline mode.": [
        ""
      ],
      "No OTP device.": [
        ""
      ],
      "Add one first": [
        ""
      ],
      "No device": [
        ""
      ],
      "Add new templates": [
        "Banka hesab\u0131"
      ],
      "Load more templates before the first one": [
        ""
      ],
      "Delete selected templates from the database": [
        ""
      ],
      "Use template to create new order": [
        ""
      ],
      "Generate a QR code for the template.": [
        ""
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        ""
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        ""
      ],
      "Jump to template with the given template ID": [
        ""
      ],
      "Template identification": [
        ""
      ],
      "Delete template": [
        ""
      ],
      'Delete the template "%1$s"': [
        ""
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        ""
      ],
      "Deleting a template cannot be undone.": [
        ""
      ],
      "Print": [
        ""
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        ""
      ],
      "If specified, this template will create orders with the same price": [
        ""
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        ""
      ],
      "An amount is required": [
        ""
      ],
      "An order summary is required": [
        ""
      ],
      'No more stock for product with ID "%1$s".': [
        ""
      ],
      "New order from template": [
        ""
      ],
      "Amount of the order": [
        ""
      ],
      "Slug": [
        ""
      ],
      "Token family slug to use in URLs (for internal use only)": [
        ""
      ],
      "Kind": [
        ""
      ],
      "Token family kind": [
        ""
      ],
      "User-readable token family name": [
        ""
      ],
      "Token family description for customers": [
        ""
      ],
      "Valid After": [
        ""
      ],
      "Token family can issue tokens after this date": [
        ""
      ],
      "Valid Before": [
        ""
      ],
      "Token family can issue tokens until this date": [
        ""
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        ""
      ],
      "Token Families": [
        ""
      ],
      "Add token family": [
        ""
      ],
      "Go to token family update page": [
        ""
      ],
      "Remove this token family from the database": [
        ""
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        ""
      ],
      "Token family has been deleted.": [
        ""
      ],
      "Delete token family": [
        "Olu\u015Fturulma"
      ],
      'Delete the token family "%1$s"': [
        ""
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        ""
      ],
      "Deleting a token family %1$s .": [
        ""
      ],
      "can't be undone": [
        ""
      ],
      "Token Family: %1$s": [
        ""
      ],
      "Incoming wire transfers": [
        ""
      ],
      "Load more wire transfers preceding the first one": [
        ""
      ],
      "Expected credit": [
        ""
      ],
      "Confirmed": [
        "Kasaba konumu"
      ],
      "Validated": [
        ""
      ],
      "Executed on": [
        ""
      ],
      "yes": [
        ""
      ],
      "no": [
        ""
      ],
      "never": [
        ""
      ],
      "unknown": [
        ""
      ],
      "Load more transfers after the last one": [
        ""
      ],
      "Verified wire transfers": [
        ""
      ],
      "Verified": [
        ""
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        ""
      ],
      "All accounts": [
        "Hesap"
      ],
      "Filter by account address": [
        ""
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        ""
      ],
      "Only show wire transfers confirmed by the merchant": [
        ""
      ],
      "Wire transfer already confirmed.": [
        ""
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        ""
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        ""
      ],
      "Transfer ID": [
        ""
      ],
      "It's not the same.": [
        ""
      ],
      'You are deleting the instance with ID "%1$s"': [
        ""
      ],
      "Instance": [
        ""
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        ""
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        ""
      ],
      "Instance id": [
        ""
      ],
      "Delete this instance": [
        ""
      ],
      "Must be one of '%1$s'": [
        ""
      ],
      "URL is invalid": [
        ""
      ],
      "Webhook ID to use": [
        ""
      ],
      "Event": [
        ""
      ],
      "Payment": [
        ""
      ],
      "Order settled": [
        ""
      ],
      "Category added": [
        ""
      ],
      "Category updated": [
        ""
      ],
      "Inventory added": [
        ""
      ],
      "Inventory updated": [
        ""
      ],
      "Inventory deleted": [
        ""
      ],
      "The event of the webhook: why the webhook is used": [
        ""
      ],
      "Method": [
        ""
      ],
      "GET": [
        ""
      ],
      "POST": [
        ""
      ],
      "PUT": [
        ""
      ],
      "PATCH": [
        ""
      ],
      "HEAD": [
        ""
      ],
      "Method used by the webhook": [
        ""
      ],
      "URL": [
        ""
      ],
      "URL of the webhook where the customer will be redirected": [
        ""
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        ""
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        ""
      ],
      "The short list of variables are:": [
        ""
      ],
      "order's description": [
        ""
      ],
      "order's price": [
        ""
      ],
      "order's unique identification": [
        ""
      ],
      "the amount that was being refunded": [
        ""
      ],
      "the reason entered by the merchant staff for granting the refund": [
        ""
      ],
      "time of the refund in nanoseconds since 1970": [
        ""
      ],
      "Http header": [
        ""
      ],
      "Header template of the webhook": [
        ""
      ],
      "Http body": [
        ""
      ],
      "Body template used by the webhook.": [
        ""
      ],
      "Webhook deleted successfully": [
        ""
      ],
      "Add new webhooks": [
        ""
      ],
      "Load more webhooks before the first one": [
        ""
      ],
      "Event type": [
        ""
      ],
      "Delete selected webhook from the database": [
        ""
      ],
      "Load more webhooks after the last one": [
        ""
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        ""
      ],
      "Webhook updated": [
        ""
      ],
      "Header": [
        ""
      ],
      "Body": [
        ""
      ],
      "Body template used by the webhook": [
        ""
      ],
      "There is another instance with this username.": [
        ""
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        ""
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        "Olu\u015Fturulma"
      ],
      "Password reset": [
        ""
      ],
      'Resetting access to the instance "%1$s"': [
        ""
      ],
      "Reset": [
        ""
      ],
      "Language": [
        ""
      ],
      "Date format": [
        ""
      ],
      "How the date is going to be displayed": [
        ""
      ],
      "Merchant type": [
        ""
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        ""
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        ""
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        ""
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        ""
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        ""
      ],
      "hour": [
        ""
      ],
      "day": [
        ""
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        ""
      ],
      "Time range": [
        ""
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        ""
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        ""
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        ""
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        ""
      ],
      "Hide for today": [
        ""
      ],
      "KYC verification needed": [
        ""
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        ""
      ],
      "The request reached a timeout, check your connection.": [
        ""
      ],
      "The request was cancelled.": [
        ""
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        ""
      ],
      "The server's response was malformed.": [
        ""
      ],
      "Could not complete the request due to a network problem.": [
        ""
      ],
      "Unexpected request error.": [
        ""
      ],
      "Unexpected error.": [
        ""
      ],
      "Add new instance": [
        "Banka hesab\u0131"
      ],
      "Edit": [
        ""
      ],
      "Change password": [
        ""
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        ""
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        ""
      ],
      "Only show active instances": [
        ""
      ],
      "Active": [
        ""
      ],
      "Only show deleted instances": [
        ""
      ],
      "Deleted": [
        ""
      ],
      "Show all instances": [
        ""
      ],
      "Delete instance": [
        ""
      ],
      'Delete the instance "%1$s"': [
        ""
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        ""
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        ""
      ],
      "Deleting an instance %1$s": [
        ""
      ],
      "This cannot be undone!": [
        ""
      ],
      "Purge the instance": [
        ""
      ],
      'Purge the instance "%1$s"': [
        ""
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        ""
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        ""
      ],
      "Purging an instance %1$s": [
        ""
      ],
      "checking compatibility with server...": [
        ""
      ],
      "Contacting the server failed": [
        ""
      ],
      "The server version is not supported": [
        ""
      ],
      'Supported version "%1$s", server version "%2$s".': [
        ""
      ],
      "Deleting": [
        ""
      ],
      "Changing": [
        ""
      ],
      "Manage password": [
        ""
      ],
      "Business Name": [
        ""
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        ""
      ],
      "title%1$s: New bank account": [
        "Banka hesab\u0131"
      ],
      "title%1$s: Bank accounts": [
        "Banka hesab\u0131"
      ],
      "title%1$s: Update bank Account": [
        ""
      ],
      "title%1$s: Orders": [
        ""
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        ""
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        ""
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        ""
      ],
      "title%1$s: Wire transfers": [
        ""
      ],
      "title%1$s: Webhooks": [
        ""
      ],
      "title%1$s: New webhook": [
        ""
      ],
      "title%1$s: Update webhook": [
        ""
      ],
      "title%1$s: OTP devices": [
        ""
      ],
      "title%1$s: New OTP device": [
        ""
      ],
      "title%1$s: Update OTP device": [
        ""
      ],
      "title%1$s: New template": [
        ""
      ],
      "title%1$s: Update template": [
        ""
      ],
      "title%1$s: Templates": [
        ""
      ],
      "title%1$s: Use template": [
        ""
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        ""
      ],
      "title%1$s: New token family": [
        ""
      ],
      "title%1$s: Update token family": [
        ""
      ],
      "title%1$s: Access tokens": [
        "Olu\u015Fturulma"
      ],
      "title%1$s: New access token": [
        "Olu\u015Fturulma"
      ],
      "titleNew instance": [
        ""
      ],
      "titleInstances": [
        ""
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=2; plural=n != 1;",
  "lang": "tr",
  "completeness": 6
};
strings["sv"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=2; plural=n != 1;",
        "lang": "sv"
      },
      "Cancel": [
        "Avbryt"
      ],
      "%1$s": [
        "%1$s"
      ],
      "Close": [
        ""
      ],
      "Continue": [
        ""
      ],
      "Clear": [
        ""
      ],
      "Confirm": [
        "Bekr\xE4fta"
      ],
      "Correct form": [
        ""
      ],
      "Comparing account details": [
        ""
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        ""
      ],
      "Field": [
        ""
      ],
      "In the form": [
        ""
      ],
      "Reported": [
        ""
      ],
      "Type": [
        ""
      ],
      "IBAN": [
        ""
      ],
      "Address": [
        ""
      ],
      "Host": [
        ""
      ],
      "Account ID": [
        ""
      ],
      "The account owner's legal name": [
        ""
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        ""
      ],
      "Bank host": [
        ""
      ],
      "Bank account": [
        ""
      ],
      "BIC": [
        ""
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        ""
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        ""
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        ""
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        ""
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        ""
      ],
      "Subject": [
        ""
      ],
      "Receiver postal code": [
        ""
      ],
      "Receiver town": [
        ""
      ],
      "Step 3:": [
        ""
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        ""
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        ""
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        ""
      ],
      "Operation in progress...": [
        ""
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        ""
      ],
      "days": [
        ""
      ],
      "hours": [
        ""
      ],
      "minutes": [
        ""
      ],
      "seconds": [
        ""
      ],
      "Forever": [
        ""
      ],
      "%1$sM": [
        ""
      ],
      "%1$sY": [
        ""
      ],
      "%1$sd": [
        ""
      ],
      "%1$sh": [
        ""
      ],
      "%1$smin": [
        ""
      ],
      "%1$ssec": [
        ""
      ],
      "Change value to never": [
        ""
      ],
      "Change value to empty": [
        ""
      ],
      "Image must be smaller than 1 MB": [
        ""
      ],
      "Add": [
        ""
      ],
      "Remove": [
        ""
      ],
      "Country": [
        ""
      ],
      "Building number": [
        ""
      ],
      "Building name": [
        ""
      ],
      "Street": [
        ""
      ],
      "Post code": [
        ""
      ],
      "Town location": [
        ""
      ],
      "District": [
        ""
      ],
      "Country subdivision": [
        ""
      ],
      "Additional address lines": [
        ""
      ],
      "Orders": [
        ""
      ],
      "Inventory": [
        ""
      ],
      "Categories": [
        ""
      ],
      "Wire transfers": [
        ""
      ],
      "Templates": [
        ""
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        ""
      ],
      "OTP Devices": [
        ""
      ],
      "Webhooks": [
        ""
      ],
      "Settings": [
        ""
      ],
      "Password": [
        ""
      ],
      "Access tokens": [
        ""
      ],
      "Connection": [
        ""
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        ""
      ],
      "New": [
        ""
      ],
      "List": [
        ""
      ],
      "Log out": [
        ""
      ],
      "Username": [
        ""
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        ""
      ],
      "Business name": [
        ""
      ],
      "Legal name of the business represented by this instance.": [
        ""
      ],
      "Email": [
        ""
      ],
      "Contact email": [
        ""
      ],
      "Phone number": [
        ""
      ],
      "Contact phone": [
        ""
      ],
      "Website URL": [
        ""
      ],
      "URL.": [
        ""
      ],
      "Logo": [
        ""
      ],
      "Logo image.": [
        ""
      ],
      "Physical location of the merchant.": [
        ""
      ],
      "Jurisdiction": [
        ""
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        ""
      ],
      "Pay transaction fee": [
        ""
      ],
      "Cover the transaction cost or pass it on to the user.": [
        ""
      ],
      "Default payment delay": [
        ""
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        ""
      ],
      "Default refund delay": [
        ""
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        ""
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        ""
      ],
      "Default wire transfer rounding interval": [
        ""
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        ""
      ],
      "To second": [
        ""
      ],
      "To minute": [
        ""
      ],
      "To hour": [
        ""
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        ""
      ],
      "Required": [
        "Obligatoriskt"
      ],
      "Unauthorized": [
        ""
      ],
      "Sending the code failed": [
        ""
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        ""
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        ""
      ],
      "Verify": [
        ""
      ],
      "Failed to send the verification code.": [
        ""
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        ""
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        ""
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        ""
      ],
      "Login required": [
        ""
      ],
      "instance name": [
        ""
      ],
      "current password": [
        ""
      ],
      "Forgot password": [
        ""
      ],
      "Create new account": [
        ""
      ],
      "Invalid": [
        ""
      ],
      "Doesn't have the pattern of an email": [
        ""
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        ""
      ],
      "Max 7 lines": [
        ""
      ],
      "Doesn't match": [
        ""
      ],
      "Instance created": [
        ""
      ],
      "Unauthorized.": [
        ""
      ],
      "Conflict.": [
        "Bekr\xE4fta"
      ],
      "Not found.": [
        ""
      ],
      "New password": [
        ""
      ],
      "Next password to be used": [
        ""
      ],
      "Repeat password": [
        ""
      ],
      "Confirm the same password": [
        ""
      ],
      "Please complete the marked fields and choose authorization method": [
        ""
      ],
      "Confirm operation": [
        ""
      ],
      "Check the password.": [
        ""
      ],
      "Instance not found.": [
        ""
      ],
      "Description": [
        ""
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        ""
      ],
      "Time the access token will be valid.": [
        ""
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        ""
      ],
      "Read only": [
        ""
      ],
      "All": [
        ""
      ],
      "Order simple": [
        ""
      ],
      "Order Point-of-Sale": [
        ""
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        ""
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        ""
      ],
      "Please complete the marked fields": [
        ""
      ],
      "Access token created": [
        ""
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        ""
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        ""
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        ""
      ],
      "Create access token": [
        ""
      ],
      "Load more devices before the first one": [
        ""
      ],
      "Load first page": [
        ""
      ],
      "Created at": [
        ""
      ],
      "Expires at": [
        ""
      ],
      "Never": [
        ""
      ],
      "Remove this access token": [
        ""
      ],
      "Delete": [
        ""
      ],
      "Load more devices after the last one": [
        ""
      ],
      "Load next page": [
        ""
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        ""
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        ""
      ],
      "Deleting an access token cannot be undone.": [
        ""
      ],
      "This is not a valid Bitcoin address.": [
        ""
      ],
      "This is not a valid Ethereum address.": [
        ""
      ],
      "This is not a valid host.": [
        ""
      ],
      "IBANs usually have more than 4 digits.": [
        ""
      ],
      "IBANs usually have fewer than 34 digits.": [
        ""
      ],
      "The IBAN's country code could not be retrieved.": [
        ""
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        ""
      ],
      "This account is not allowed.": [
        ""
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        ""
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        ""
      ],
      "Select a wire method...": [
        ""
      ],
      "Routing": [
        ""
      ],
      "Routing number": [
        ""
      ],
      "Account": [
        ""
      ],
      "Account number": [
        ""
      ],
      "Code": [
        ""
      ],
      "Business Identifier Code": [
        ""
      ],
      "International Bank Account Number": [
        ""
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        ""
      ],
      "Bitcoin protocol": [
        ""
      ],
      "Ethereum protocol": [
        ""
      ],
      "Interledger protocol": [
        ""
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        ""
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        ""
      ],
      "Invalid url": [
        ""
      ],
      "URL must end with a '/'": [
        ""
      ],
      "URL must not contain params": [
        ""
      ],
      "URL must not hash param": [
        ""
      ],
      "Conflict": [
        "Bekr\xE4fta"
      ],
      'Server replied with "bad request".': [
        ""
      ],
      "Unauthorized, check credentials.": [
        ""
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        ""
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        ""
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        ""
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        ""
      ],
      "Endpoint URL": [
        ""
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        ""
      ],
      "Auth type": [
        ""
      ],
      "Choose the authentication type for the account info URL": [
        ""
      ],
      "Without authentication": [
        ""
      ],
      "With username and password": [
        ""
      ],
      "With token": [
        ""
      ],
      "Do not change": [
        ""
      ],
      "Username to access the account information.": [
        ""
      ],
      "Password to access the account information.": [
        ""
      ],
      "Access token to access the account information.": [
        ""
      ],
      "Match": [
        ""
      ],
      "Check where the information match against the server info.": [
        ""
      ],
      "Compare info from server with account form": [
        ""
      ],
      "Test": [
        ""
      ],
      "Server didn't like the request we made.": [
        ""
      ],
      "Account details": [
        ""
      ],
      "Check if the information matches the server info.": [
        ""
      ],
      "The bank account has been deleted.": [
        ""
      ],
      "Bank accounts": [
        ""
      ],
      "Add new account": [
        ""
      ],
      "Owner's name": [
        ""
      ],
      "Delete selected accounts from the database": [
        ""
      ],
      "There are no accounts yet, add more pressing the + sign": [
        ""
      ],
      "You must provide a bank account to receive payments.": [
        ""
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        ""
      ],
      "Name": [
        ""
      ],
      "Category name": [
        ""
      ],
      "Category deleted": [
        ""
      ],
      "Add new devices": [
        ""
      ],
      "ID": [
        ""
      ],
      "Total products": [
        ""
      ],
      "Delete selected category from the database": [
        ""
      ],
      "There are no categories yet, add more pressing the + sign": [
        ""
      ],
      "Id:": [
        ""
      ],
      "Name of the category": [
        ""
      ],
      "Products": [
        ""
      ],
      "Image": [
        ""
      ],
      "Load more products after the last one": [
        ""
      ],
      "There are no products in this category.": [
        ""
      ],
      "Account's KYC status": [
        ""
      ],
      "Exchange": [
        ""
      ],
      "Status": [
        ""
      ],
      "Ok": [
        ""
      ],
      "Action required": [
        "Obligatoriskt"
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        ""
      ],
      "More information required.": [
        ""
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        ""
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        ""
      ],
      "No pending kyc verification!": [
        ""
      ],
      "Wire account is invalid": [
        ""
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        ""
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        "Obligatoriskt"
      ],
      "The payment provider requires more information.": [
        ""
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        ""
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        ""
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        ""
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        ""
      ],
      "Enter description or id": [
        ""
      ],
      "No match found for that description or ID.": [
        ""
      ],
      "You must enter a valid product identifier.": [
        ""
      ],
      "Quantity must be greater than 0!": [
        ""
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        ""
      ],
      "Search product": [
        ""
      ],
      "Quantity": [
        ""
      ],
      "How many products will be added": [
        ""
      ],
      "Add from inventory": [
        ""
      ],
      "This product has %1$s applicable taxes configured.": [
        ""
      ],
      "No taxes configured for this product.": [
        ""
      ],
      "Amount": [
        ""
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        ""
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        ""
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        ""
      ],
      "Add tax to the tax list": [
        ""
      ],
      "Describe and add a product that is not in the inventory list": [
        ""
      ],
      "Add custom product": [
        ""
      ],
      "Complete information of the product": [
        ""
      ],
      "Must be a number.": [
        ""
      ],
      "Must be greater than 0.": [
        ""
      ],
      "Photo of the product.": [
        ""
      ],
      "Full product description.": [
        ""
      ],
      "Unit name": [
        ""
      ],
      "Name of the product unit.": [
        ""
      ],
      "Price per unit": [
        ""
      ],
      "Amount in the current currency.": [
        ""
      ],
      "How many products will be added.": [
        ""
      ],
      "Taxes": [
        ""
      ],
      "Total price": [
        ""
      ],
      "Must be greater than 0": [
        ""
      ],
      "Must have a refund deadline": [
        ""
      ],
      "Auto refund can't be after refund deadline": [
        ""
      ],
      "Must be in the future": [
        ""
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        ""
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        ""
      ],
      "Manage products in order": [
        ""
      ],
      "%1$s products with a total price of %2$s.": [
        ""
      ],
      "Manage list of products in the order.": [
        ""
      ],
      "Remove this product from the order.": [
        ""
      ],
      "Products price sum": [
        ""
      ],
      "Total product price added up": [
        ""
      ],
      "Order price": [
        ""
      ],
      "Amount to be paid by the customer": [
        ""
      ],
      "Final order price": [
        ""
      ],
      "Summary": [
        ""
      ],
      "Title of the order to be shown to the customer": [
        ""
      ],
      "Shipping and fulfillment": [
        ""
      ],
      "Delivery date": [
        ""
      ],
      "Deadline for physical delivery assured by the merchant.": [
        ""
      ],
      "Delivery location": [
        ""
      ],
      "Address where the products will be delivered": [
        ""
      ],
      "Fulfillment URL": [
        ""
      ],
      "URL to which the user will be redirected after successful payment.": [
        ""
      ],
      "Fulfillment message": [
        ""
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        ""
      ],
      "Override default Taler payment settings for this order": [
        ""
      ],
      "Payment time": [
        ""
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        ""
      ],
      "Default": [
        ""
      ],
      "Refund time": [
        ""
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        ""
      ],
      "Wire transfer time": [
        ""
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        ""
      ],
      "Auto-refund time": [
        ""
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        ""
      ],
      "Maximum fee": [
        ""
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        ""
      ],
      "Create token": [
        ""
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        ""
      ],
      "Minimum age required": [
        ""
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        ""
      ],
      "Minimum age defined by the products is %1$s": [
        ""
      ],
      "No product with age restriction in this order": [
        ""
      ],
      "Additional information": [
        ""
      ],
      "Custom information to be included in the contract for this order.": [
        ""
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        ""
      ],
      "remove": [
        ""
      ],
      "Custom field name": [
        ""
      ],
      "Disabled": [
        ""
      ],
      "No deadline": [
        ""
      ],
      "Deadline at %1$s": [
        ""
      ],
      "Select date to show nearby orders": [
        ""
      ],
      "Only show unpaid orders": [
        ""
      ],
      "Only show paid orders": [
        ""
      ],
      "Paid": [
        ""
      ],
      "Only show orders with refunds": [
        ""
      ],
      "Refunded": [
        ""
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        ""
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        ""
      ],
      "Completed": [
        ""
      ],
      "Remove all filters": [
        ""
      ],
      "Clear date filter": [
        ""
      ],
      "Jump to date (%1$s)": [
        ""
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        ""
      ],
      "Order id": [
        ""
      ],
      "Instance unknown": [
        ""
      ],
      "Order unknown": [
        ""
      ],
      "Create order": [
        ""
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        ""
      ],
      "Refund": [
        ""
      ],
      "copy url": [
        ""
      ],
      "Load more orders after the last one": [
        ""
      ],
      "No orders have been found matching your query!": [
        ""
      ],
      "Duplicated": [
        ""
      ],
      "This value exceeds the refundable amount.": [
        ""
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        ""
      ],
      "refund": [
        ""
      ],
      "Reason": [
        ""
      ],
      "Amount to be refunded": [
        ""
      ],
      "Max refundable:": [
        ""
      ],
      "Choose one...": [
        ""
      ],
      "Requested by the customer": [
        ""
      ],
      "Other": [
        ""
      ],
      "Why this order is being refunded": [
        ""
      ],
      "More information to give context": [
        ""
      ],
      "now": [
        ""
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        ""
      ],
      "This wire transfer has been notified by the payment service provider.": [
        ""
      ],
      "This refund is waiting to be claimed by the customer.": [
        ""
      ],
      "This refund has been claimed by the customer.": [
        ""
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        ""
      ],
      "Human-readable description of the whole purchase": [
        ""
      ],
      "Total price for the transaction": [
        ""
      ],
      "URL for this purchase": [
        ""
      ],
      "Max fee": [
        ""
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        ""
      ],
      "Time when this contract was generated": [
        ""
      ],
      "Refund deadline": [
        ""
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        ""
      ],
      "Payment deadline": [
        ""
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        ""
      ],
      "Wire transfer deadline": [
        ""
      ],
      "Transfer deadline for the exchange": [
        ""
      ],
      "Time indicating when the order should be delivered": [
        ""
      ],
      "Where the order will be delivered": [
        ""
      ],
      "Auto-refund delay": [
        ""
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        ""
      ],
      "Extra info": [
        ""
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        ""
      ],
      "order created": [
        ""
      ],
      "pay deadline": [
        ""
      ],
      "refund deadline": [
        ""
      ],
      "delivery": [
        ""
      ],
      "Order": [
        ""
      ],
      "Claimed": [
        ""
      ],
      "Claimed at": [
        ""
      ],
      "Timeline": [
        ""
      ],
      "Payment details": [
        ""
      ],
      "Order status": [
        ""
      ],
      "Unpaid": [
        ""
      ],
      "Product list": [
        ""
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        ""
      ],
      "refund created: %1$s": [
        ""
      ],
      "refund created: %1$s: %2$s": [
        ""
      ],
      "refund taken: %1$s": [
        ""
      ],
      "refund taken: %1$s: %2$s": [
        ""
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        ""
      ],
      "Wired": [
        ""
      ],
      "Refund order": [
        ""
      ],
      "Not refundable": [
        ""
      ],
      "Next event in": [
        ""
      ],
      "The order was wired.": [
        ""
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        ""
      ],
      "Refund taken": [
        ""
      ],
      "Status URL": [
        ""
      ],
      "Refund URI": [
        ""
      ],
      "Pay at": [
        ""
      ],
      "Order status URL": [
        ""
      ],
      "Payment URI": [
        ""
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        ""
      ],
      "Invalid. Please enter letters and numbers only.": [
        ""
      ],
      "Just letters and numbers from 2 to 7": [
        ""
      ],
      "The size of the key must be 32 characters": [
        ""
      ],
      "Internal ID on the system": [
        ""
      ],
      "Useful to identify the device physically": [
        ""
      ],
      "Verification algorithm": [
        ""
      ],
      "Algorithm used to verify transactions in offline mode": [
        ""
      ],
      "Device key": [
        ""
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        ""
      ],
      "Your device needs to match exactly the same value": [
        ""
      ],
      "Generate random secret key": [
        ""
      ],
      "Random": [
        ""
      ],
      "Create another": [
        ""
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        ""
      ],
      "OTP devices": [
        ""
      ],
      "Delete selected devices from the database": [
        ""
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        ""
      ],
      "Template id is unknown": [
        ""
      ],
      "The provided information is inconsistent with the current state of the template": [
        ""
      ],
      "Device:": [
        ""
      ],
      "Algorithm to use to verify transactions in offline mode": [
        ""
      ],
      "Not modified": [
        ""
      ],
      "Be sure to be very hard to guess or use the random generator": [
        ""
      ],
      "Your device need to have exactly the same value": [
        ""
      ],
      "Change key": [
        ""
      ],
      "Is not the same": [
        ""
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        ""
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        ""
      ],
      "Testing password change": [
        ""
      ],
      "Password changed": [
        ""
      ],
      "The current password is wrong.": [
        ""
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        ""
      ],
      "Add element to the list": [
        ""
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        ""
      ],
      "Manage stock": [
        ""
      ],
      "This product has been configured without stock control": [
        ""
      ],
      "Infinite": [
        ""
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        ""
      ],
      "Incoming": [
        ""
      ],
      "Lost": [
        ""
      ],
      "Current": [
        ""
      ],
      "Remove stock control for this product": [
        ""
      ],
      "without stock": [
        ""
      ],
      "Next restock": [
        ""
      ],
      "Warehouse address": [
        ""
      ],
      "Invalid amount": [
        ""
      ],
      "Product identification to use in URLs (for internal use only).": [
        ""
      ],
      "Illustration of the product for customers.": [
        ""
      ],
      "Product name.": [
        ""
      ],
      "Product description for customers.": [
        ""
      ],
      "Age restriction": [
        ""
      ],
      "Is this product restricted for customer below certain age?": [
        ""
      ],
      "Minimum age of the customer": [
        ""
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        ""
      ],
      "Example: kg, items or liters": [
        ""
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        ""
      ],
      "Stock": [
        ""
      ],
      "Inventory for products with finite supply (for internal use only).": [
        ""
      ],
      "Taxes included in the product price, exposed to customers.": [
        ""
      ],
      "Search by category description or id": [
        ""
      ],
      "Categories where this product will be listed on.": [
        ""
      ],
      "Add product to inventory": [
        ""
      ],
      "Product not found.": [
        ""
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        ""
      ],
      "Sold": [
        ""
      ],
      "Free": [
        ""
      ],
      "Go to product update page": [
        ""
      ],
      "Update": [
        ""
      ],
      "Remove this product from the database": [
        ""
      ],
      "Update the product with new price": [
        ""
      ],
      "Update product with new price": [
        ""
      ],
      "Confirm update": [
        ""
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        ""
      ],
      "Add more elements to the inventory": [
        ""
      ],
      "Report elements lost in the inventory": [
        ""
      ],
      "New price for the product": [
        ""
      ],
      "There are values with errors.": [
        ""
      ],
      "Update product with new stock and price": [
        ""
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        ""
      ],
      "Product (ID: %1$s) has been deleted": [
        ""
      ],
      "Jump to product with the given product ID": [
        ""
      ],
      "Product id": [
        ""
      ],
      "Delete product": [
        ""
      ],
      'Delete the product "%1$s"': [
        ""
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        ""
      ],
      "Deleting a product cannot be undone.": [
        ""
      ],
      "Product id:": [
        ""
      ],
      "Must be greater that 0": [
        ""
      ],
      "Too short": [
        ""
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        ""
      ],
      "Name of the template in URLs.": [
        ""
      ],
      "Template name": [
        ""
      ],
      "Describe what this template stands for": [
        ""
      ],
      "Order summary": [
        ""
      ],
      "If specified here, this template will create orders with the same summary": [
        ""
      ],
      "Summary is editable": [
        ""
      ],
      "Allow the user to change the summary.": [
        ""
      ],
      "If specified here, this template will create orders with the same price": [
        ""
      ],
      "Amount is editable": [
        ""
      ],
      "Allow the user to select the amount to pay.": [
        ""
      ],
      "Currency is editable": [
        ""
      ],
      "Allow the user to change currency.": [
        ""
      ],
      "Supported currencies": [
        ""
      ],
      "Supported currencies: %1$s": [
        ""
      ],
      "Minimum age": [
        ""
      ],
      "Is this contract restricted to some age?": [
        ""
      ],
      "Payment timeout": [
        ""
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        ""
      ],
      "OTP device": [
        ""
      ],
      "Use to verify transactions in offline mode.": [
        ""
      ],
      "No OTP device.": [
        ""
      ],
      "Add one first": [
        ""
      ],
      "No device": [
        ""
      ],
      "Add new templates": [
        ""
      ],
      "Load more templates before the first one": [
        ""
      ],
      "Delete selected templates from the database": [
        ""
      ],
      "Use template to create new order": [
        ""
      ],
      "Generate a QR code for the template.": [
        ""
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        ""
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        ""
      ],
      "Jump to template with the given template ID": [
        ""
      ],
      "Template identification": [
        ""
      ],
      "Delete template": [
        ""
      ],
      'Delete the template "%1$s"': [
        ""
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        ""
      ],
      "Deleting a template cannot be undone.": [
        ""
      ],
      "Print": [
        ""
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        ""
      ],
      "If specified, this template will create orders with the same price": [
        ""
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        ""
      ],
      "An amount is required": [
        ""
      ],
      "An order summary is required": [
        ""
      ],
      'No more stock for product with ID "%1$s".': [
        ""
      ],
      "New order from template": [
        ""
      ],
      "Amount of the order": [
        ""
      ],
      "Slug": [
        ""
      ],
      "Token family slug to use in URLs (for internal use only)": [
        ""
      ],
      "Kind": [
        ""
      ],
      "Token family kind": [
        ""
      ],
      "User-readable token family name": [
        ""
      ],
      "Token family description for customers": [
        ""
      ],
      "Valid After": [
        ""
      ],
      "Token family can issue tokens after this date": [
        ""
      ],
      "Valid Before": [
        ""
      ],
      "Token family can issue tokens until this date": [
        ""
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        ""
      ],
      "Token Families": [
        ""
      ],
      "Add token family": [
        ""
      ],
      "Go to token family update page": [
        ""
      ],
      "Remove this token family from the database": [
        ""
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        ""
      ],
      "Token family has been deleted.": [
        ""
      ],
      "Delete token family": [
        ""
      ],
      'Delete the token family "%1$s"': [
        ""
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        ""
      ],
      "Deleting a token family %1$s .": [
        ""
      ],
      "can't be undone": [
        ""
      ],
      "Token Family: %1$s": [
        ""
      ],
      "Incoming wire transfers": [
        ""
      ],
      "Load more wire transfers preceding the first one": [
        ""
      ],
      "Expected credit": [
        ""
      ],
      "Confirmed": [
        ""
      ],
      "Validated": [
        ""
      ],
      "Executed on": [
        ""
      ],
      "yes": [
        ""
      ],
      "no": [
        ""
      ],
      "never": [
        ""
      ],
      "unknown": [
        ""
      ],
      "Load more transfers after the last one": [
        ""
      ],
      "Verified wire transfers": [
        ""
      ],
      "Verified": [
        ""
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        ""
      ],
      "All accounts": [
        ""
      ],
      "Filter by account address": [
        ""
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        ""
      ],
      "Only show wire transfers confirmed by the merchant": [
        ""
      ],
      "Wire transfer already confirmed.": [
        ""
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        ""
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        ""
      ],
      "Transfer ID": [
        ""
      ],
      "It's not the same.": [
        ""
      ],
      'You are deleting the instance with ID "%1$s"': [
        ""
      ],
      "Instance": [
        ""
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        ""
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        ""
      ],
      "Instance id": [
        ""
      ],
      "Delete this instance": [
        ""
      ],
      "Must be one of '%1$s'": [
        ""
      ],
      "URL is invalid": [
        ""
      ],
      "Webhook ID to use": [
        ""
      ],
      "Event": [
        ""
      ],
      "Payment": [
        ""
      ],
      "Order settled": [
        ""
      ],
      "Category added": [
        ""
      ],
      "Category updated": [
        ""
      ],
      "Inventory added": [
        ""
      ],
      "Inventory updated": [
        ""
      ],
      "Inventory deleted": [
        ""
      ],
      "The event of the webhook: why the webhook is used": [
        ""
      ],
      "Method": [
        ""
      ],
      "GET": [
        ""
      ],
      "POST": [
        ""
      ],
      "PUT": [
        ""
      ],
      "PATCH": [
        ""
      ],
      "HEAD": [
        ""
      ],
      "Method used by the webhook": [
        ""
      ],
      "URL": [
        ""
      ],
      "URL of the webhook where the customer will be redirected": [
        ""
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        ""
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        ""
      ],
      "The short list of variables are:": [
        ""
      ],
      "order's description": [
        ""
      ],
      "order's price": [
        ""
      ],
      "order's unique identification": [
        ""
      ],
      "the amount that was being refunded": [
        ""
      ],
      "the reason entered by the merchant staff for granting the refund": [
        ""
      ],
      "time of the refund in nanoseconds since 1970": [
        ""
      ],
      "Http header": [
        ""
      ],
      "Header template of the webhook": [
        ""
      ],
      "Http body": [
        ""
      ],
      "Body template used by the webhook.": [
        ""
      ],
      "Webhook deleted successfully": [
        ""
      ],
      "Add new webhooks": [
        ""
      ],
      "Load more webhooks before the first one": [
        ""
      ],
      "Event type": [
        ""
      ],
      "Delete selected webhook from the database": [
        ""
      ],
      "Load more webhooks after the last one": [
        ""
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        ""
      ],
      "Webhook updated": [
        ""
      ],
      "Header": [
        ""
      ],
      "Body": [
        ""
      ],
      "Body template used by the webhook": [
        ""
      ],
      "There is another instance with this username.": [
        ""
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        ""
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        ""
      ],
      "Password reset": [
        ""
      ],
      'Resetting access to the instance "%1$s"': [
        ""
      ],
      "Reset": [
        ""
      ],
      "Language": [
        ""
      ],
      "Date format": [
        ""
      ],
      "How the date is going to be displayed": [
        ""
      ],
      "Merchant type": [
        ""
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        ""
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        ""
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        ""
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        ""
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        ""
      ],
      "hour": [
        ""
      ],
      "day": [
        ""
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        ""
      ],
      "Time range": [
        ""
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        ""
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        ""
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        ""
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        ""
      ],
      "Hide for today": [
        ""
      ],
      "KYC verification needed": [
        ""
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        ""
      ],
      "The request reached a timeout, check your connection.": [
        ""
      ],
      "The request was cancelled.": [
        ""
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        ""
      ],
      "The server's response was malformed.": [
        ""
      ],
      "Could not complete the request due to a network problem.": [
        ""
      ],
      "Unexpected request error.": [
        ""
      ],
      "Unexpected error.": [
        ""
      ],
      "Add new instance": [
        ""
      ],
      "Edit": [
        ""
      ],
      "Change password": [
        ""
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        ""
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        ""
      ],
      "Only show active instances": [
        ""
      ],
      "Active": [
        ""
      ],
      "Only show deleted instances": [
        ""
      ],
      "Deleted": [
        ""
      ],
      "Show all instances": [
        ""
      ],
      "Delete instance": [
        ""
      ],
      'Delete the instance "%1$s"': [
        ""
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        ""
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        ""
      ],
      "Deleting an instance %1$s": [
        ""
      ],
      "This cannot be undone!": [
        ""
      ],
      "Purge the instance": [
        ""
      ],
      'Purge the instance "%1$s"': [
        ""
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        ""
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        ""
      ],
      "Purging an instance %1$s": [
        ""
      ],
      "checking compatibility with server...": [
        ""
      ],
      "Contacting the server failed": [
        ""
      ],
      "The server version is not supported": [
        ""
      ],
      'Supported version "%1$s", server version "%2$s".': [
        ""
      ],
      "Deleting": [
        ""
      ],
      "Changing": [
        ""
      ],
      "Manage password": [
        ""
      ],
      "Business Name": [
        ""
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        ""
      ],
      "title%1$s: New bank account": [
        ""
      ],
      "title%1$s: Bank accounts": [
        ""
      ],
      "title%1$s: Update bank Account": [
        ""
      ],
      "title%1$s: Orders": [
        ""
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        ""
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        ""
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        ""
      ],
      "title%1$s: Wire transfers": [
        ""
      ],
      "title%1$s: Webhooks": [
        ""
      ],
      "title%1$s: New webhook": [
        ""
      ],
      "title%1$s: Update webhook": [
        ""
      ],
      "title%1$s: OTP devices": [
        ""
      ],
      "title%1$s: New OTP device": [
        ""
      ],
      "title%1$s: Update OTP device": [
        ""
      ],
      "title%1$s: New template": [
        ""
      ],
      "title%1$s: Update template": [
        ""
      ],
      "title%1$s: Templates": [
        ""
      ],
      "title%1$s: Use template": [
        ""
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        ""
      ],
      "title%1$s: New token family": [
        ""
      ],
      "title%1$s: Update token family": [
        ""
      ],
      "title%1$s: Access tokens": [
        ""
      ],
      "title%1$s: New access token": [
        ""
      ],
      "titleNew instance": [
        ""
      ],
      "titleInstances": [
        ""
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=2; plural=n != 1;",
  "lang": "sv",
  "completeness": 0
};
strings["it"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=2; plural=n != 1;",
        "lang": "it"
      },
      "Cancel": [
        ""
      ],
      "%1$s": [
        ""
      ],
      "Close": [
        ""
      ],
      "Continue": [
        ""
      ],
      "Clear": [
        ""
      ],
      "Confirm": [
        "Confermare"
      ],
      "Correct form": [
        ""
      ],
      "Comparing account details": [
        ""
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        ""
      ],
      "Field": [
        ""
      ],
      "In the form": [
        ""
      ],
      "Reported": [
        ""
      ],
      "Type": [
        ""
      ],
      "IBAN": [
        ""
      ],
      "Address": [
        ""
      ],
      "Host": [
        ""
      ],
      "Account ID": [
        "Importo"
      ],
      "The account owner's legal name": [
        ""
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        ""
      ],
      "Bank host": [
        ""
      ],
      "Bank account": [
        ""
      ],
      "BIC": [
        ""
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        ""
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        ""
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        ""
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        ""
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        ""
      ],
      "Subject": [
        "Soggetto"
      ],
      "Receiver postal code": [
        ""
      ],
      "Receiver town": [
        ""
      ],
      "Step 3:": [
        ""
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        ""
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        ""
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        ""
      ],
      "Operation in progress...": [
        ""
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        ""
      ],
      "days": [
        ""
      ],
      "hours": [
        ""
      ],
      "minutes": [
        ""
      ],
      "seconds": [
        ""
      ],
      "Forever": [
        ""
      ],
      "%1$sM": [
        ""
      ],
      "%1$sY": [
        ""
      ],
      "%1$sd": [
        ""
      ],
      "%1$sh": [
        ""
      ],
      "%1$smin": [
        ""
      ],
      "%1$ssec": [
        ""
      ],
      "Change value to never": [
        ""
      ],
      "Change value to empty": [
        ""
      ],
      "Image must be smaller than 1 MB": [
        ""
      ],
      "Add": [
        ""
      ],
      "Remove": [
        ""
      ],
      "Country": [
        ""
      ],
      "Building number": [
        ""
      ],
      "Building name": [
        ""
      ],
      "Street": [
        ""
      ],
      "Post code": [
        ""
      ],
      "Town location": [
        ""
      ],
      "District": [
        ""
      ],
      "Country subdivision": [
        ""
      ],
      "Additional address lines": [
        ""
      ],
      "Orders": [
        ""
      ],
      "Inventory": [
        ""
      ],
      "Categories": [
        ""
      ],
      "Wire transfers": [
        ""
      ],
      "Templates": [
        ""
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        ""
      ],
      "OTP Devices": [
        ""
      ],
      "Webhooks": [
        ""
      ],
      "Settings": [
        "Impostazioni"
      ],
      "Password": [
        ""
      ],
      "Access tokens": [
        ""
      ],
      "Connection": [
        ""
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        ""
      ],
      "New": [
        ""
      ],
      "List": [
        ""
      ],
      "Log out": [
        ""
      ],
      "Username": [
        ""
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        ""
      ],
      "Business name": [
        ""
      ],
      "Legal name of the business represented by this instance.": [
        ""
      ],
      "Email": [
        ""
      ],
      "Contact email": [
        ""
      ],
      "Phone number": [
        ""
      ],
      "Contact phone": [
        ""
      ],
      "Website URL": [
        ""
      ],
      "URL.": [
        ""
      ],
      "Logo": [
        ""
      ],
      "Logo image.": [
        ""
      ],
      "Physical location of the merchant.": [
        ""
      ],
      "Jurisdiction": [
        ""
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        ""
      ],
      "Pay transaction fee": [
        ""
      ],
      "Cover the transaction cost or pass it on to the user.": [
        ""
      ],
      "Default payment delay": [
        ""
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        ""
      ],
      "Default refund delay": [
        ""
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        ""
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        ""
      ],
      "Default wire transfer rounding interval": [
        ""
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        ""
      ],
      "To second": [
        ""
      ],
      "To minute": [
        ""
      ],
      "To hour": [
        ""
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        ""
      ],
      "Required": [
        "Obbligatorio"
      ],
      "Unauthorized": [
        ""
      ],
      "Sending the code failed": [
        ""
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        ""
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        "Indietro"
      ],
      "Verify": [
        ""
      ],
      "Failed to send the verification code.": [
        ""
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        ""
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        ""
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        ""
      ],
      "Login required": [
        ""
      ],
      "instance name": [
        ""
      ],
      "current password": [
        ""
      ],
      "Forgot password": [
        ""
      ],
      "Create new account": [
        ""
      ],
      "Invalid": [
        ""
      ],
      "Doesn't have the pattern of an email": [
        ""
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        ""
      ],
      "Max 7 lines": [
        ""
      ],
      "Doesn't match": [
        ""
      ],
      "Instance created": [
        ""
      ],
      "Unauthorized.": [
        ""
      ],
      "Conflict.": [
        "Confermare"
      ],
      "Not found.": [
        ""
      ],
      "New password": [
        ""
      ],
      "Next password to be used": [
        ""
      ],
      "Repeat password": [
        ""
      ],
      "Confirm the same password": [
        ""
      ],
      "Please complete the marked fields and choose authorization method": [
        ""
      ],
      "Confirm operation": [
        ""
      ],
      "Check the password.": [
        ""
      ],
      "Instance not found.": [
        ""
      ],
      "Description": [
        ""
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        ""
      ],
      "Time the access token will be valid.": [
        ""
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        ""
      ],
      "Read only": [
        ""
      ],
      "All": [
        ""
      ],
      "Order simple": [
        ""
      ],
      "Order Point-of-Sale": [
        ""
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        ""
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        ""
      ],
      "Please complete the marked fields": [
        ""
      ],
      "Access token created": [
        ""
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        ""
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        ""
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        ""
      ],
      "Create access token": [
        ""
      ],
      "Load more devices before the first one": [
        ""
      ],
      "Load first page": [
        ""
      ],
      "Created at": [
        ""
      ],
      "Expires at": [
        ""
      ],
      "Never": [
        ""
      ],
      "Remove this access token": [
        ""
      ],
      "Delete": [
        ""
      ],
      "Load more devices after the last one": [
        ""
      ],
      "Load next page": [
        ""
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        ""
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        ""
      ],
      "Deleting an access token cannot be undone.": [
        ""
      ],
      "This is not a valid Bitcoin address.": [
        ""
      ],
      "This is not a valid Ethereum address.": [
        ""
      ],
      "This is not a valid host.": [
        ""
      ],
      "IBANs usually have more than 4 digits.": [
        ""
      ],
      "IBANs usually have fewer than 34 digits.": [
        ""
      ],
      "The IBAN's country code could not be retrieved.": [
        ""
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        ""
      ],
      "This account is not allowed.": [
        ""
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        ""
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        ""
      ],
      "Select a wire method...": [
        ""
      ],
      "Routing": [
        ""
      ],
      "Routing number": [
        ""
      ],
      "Account": [
        ""
      ],
      "Account number": [
        "Importo"
      ],
      "Code": [
        ""
      ],
      "Business Identifier Code": [
        ""
      ],
      "International Bank Account Number": [
        ""
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        ""
      ],
      "Bitcoin protocol": [
        ""
      ],
      "Ethereum protocol": [
        ""
      ],
      "Interledger protocol": [
        ""
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        ""
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        ""
      ],
      "Invalid url": [
        ""
      ],
      "URL must end with a '/'": [
        ""
      ],
      "URL must not contain params": [
        ""
      ],
      "URL must not hash param": [
        ""
      ],
      "Conflict": [
        "Confermare"
      ],
      'Server replied with "bad request".': [
        ""
      ],
      "Unauthorized, check credentials.": [
        ""
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        ""
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        ""
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        "Conto:"
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        ""
      ],
      "Endpoint URL": [
        ""
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        ""
      ],
      "Auth type": [
        ""
      ],
      "Choose the authentication type for the account info URL": [
        ""
      ],
      "Without authentication": [
        ""
      ],
      "With username and password": [
        ""
      ],
      "With token": [
        ""
      ],
      "Do not change": [
        ""
      ],
      "Username to access the account information.": [
        ""
      ],
      "Password to access the account information.": [
        ""
      ],
      "Access token to access the account information.": [
        ""
      ],
      "Match": [
        ""
      ],
      "Check where the information match against the server info.": [
        ""
      ],
      "Compare info from server with account form": [
        ""
      ],
      "Test": [
        ""
      ],
      "Server didn't like the request we made.": [
        ""
      ],
      "Account details": [
        ""
      ],
      "Check if the information matches the server info.": [
        ""
      ],
      "The bank account has been deleted.": [
        ""
      ],
      "Bank accounts": [
        ""
      ],
      "Add new account": [
        ""
      ],
      "Owner's name": [
        ""
      ],
      "Delete selected accounts from the database": [
        ""
      ],
      "There are no accounts yet, add more pressing the + sign": [
        ""
      ],
      "You must provide a bank account to receive payments.": [
        ""
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        ""
      ],
      "Name": [
        ""
      ],
      "Category name": [
        ""
      ],
      "Category deleted": [
        ""
      ],
      "Add new devices": [
        ""
      ],
      "ID": [
        ""
      ],
      "Total products": [
        ""
      ],
      "Delete selected category from the database": [
        ""
      ],
      "There are no categories yet, add more pressing the + sign": [
        ""
      ],
      "Id:": [
        ""
      ],
      "Name of the category": [
        ""
      ],
      "Products": [
        ""
      ],
      "Image": [
        ""
      ],
      "Load more products after the last one": [
        ""
      ],
      "There are no products in this category.": [
        ""
      ],
      "Account's KYC status": [
        ""
      ],
      "Exchange": [
        ""
      ],
      "Status": [
        ""
      ],
      "Ok": [
        ""
      ],
      "Action required": [
        "Obbligatorio"
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        ""
      ],
      "More information required.": [
        ""
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        ""
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        ""
      ],
      "No pending kyc verification!": [
        ""
      ],
      "Wire account is invalid": [
        ""
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        ""
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        "Obbligatorio"
      ],
      "The payment provider requires more information.": [
        ""
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        ""
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        ""
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        ""
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        ""
      ],
      "Enter description or id": [
        ""
      ],
      "No match found for that description or ID.": [
        ""
      ],
      "You must enter a valid product identifier.": [
        ""
      ],
      "Quantity must be greater than 0!": [
        ""
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        ""
      ],
      "Search product": [
        ""
      ],
      "Quantity": [
        ""
      ],
      "How many products will be added": [
        ""
      ],
      "Add from inventory": [
        ""
      ],
      "This product has %1$s applicable taxes configured.": [
        ""
      ],
      "No taxes configured for this product.": [
        ""
      ],
      "Amount": [
        "Importo"
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        ""
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        ""
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        ""
      ],
      "Add tax to the tax list": [
        ""
      ],
      "Describe and add a product that is not in the inventory list": [
        ""
      ],
      "Add custom product": [
        ""
      ],
      "Complete information of the product": [
        ""
      ],
      "Must be a number.": [
        ""
      ],
      "Must be greater than 0.": [
        ""
      ],
      "Photo of the product.": [
        ""
      ],
      "Full product description.": [
        ""
      ],
      "Unit name": [
        ""
      ],
      "Name of the product unit.": [
        ""
      ],
      "Price per unit": [
        ""
      ],
      "Amount in the current currency.": [
        ""
      ],
      "How many products will be added.": [
        ""
      ],
      "Taxes": [
        ""
      ],
      "Total price": [
        ""
      ],
      "Must be greater than 0": [
        ""
      ],
      "Must have a refund deadline": [
        ""
      ],
      "Auto refund can't be after refund deadline": [
        ""
      ],
      "Must be in the future": [
        ""
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        ""
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        ""
      ],
      "Manage products in order": [
        ""
      ],
      "%1$s products with a total price of %2$s.": [
        ""
      ],
      "Manage list of products in the order.": [
        ""
      ],
      "Remove this product from the order.": [
        ""
      ],
      "Products price sum": [
        ""
      ],
      "Total product price added up": [
        ""
      ],
      "Order price": [
        ""
      ],
      "Amount to be paid by the customer": [
        ""
      ],
      "Final order price": [
        ""
      ],
      "Summary": [
        ""
      ],
      "Title of the order to be shown to the customer": [
        ""
      ],
      "Shipping and fulfillment": [
        ""
      ],
      "Delivery date": [
        ""
      ],
      "Deadline for physical delivery assured by the merchant.": [
        ""
      ],
      "Delivery location": [
        ""
      ],
      "Address where the products will be delivered": [
        ""
      ],
      "Fulfillment URL": [
        ""
      ],
      "URL to which the user will be redirected after successful payment.": [
        ""
      ],
      "Fulfillment message": [
        ""
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        ""
      ],
      "Override default Taler payment settings for this order": [
        ""
      ],
      "Payment time": [
        ""
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        ""
      ],
      "Default": [
        ""
      ],
      "Refund time": [
        "Rimborsato"
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        ""
      ],
      "Wire transfer time": [
        ""
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        ""
      ],
      "Auto-refund time": [
        ""
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        ""
      ],
      "Maximum fee": [
        ""
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        ""
      ],
      "Create token": [
        ""
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        ""
      ],
      "Minimum age required": [
        ""
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        ""
      ],
      "Minimum age defined by the products is %1$s": [
        ""
      ],
      "No product with age restriction in this order": [
        ""
      ],
      "Additional information": [
        ""
      ],
      "Custom information to be included in the contract for this order.": [
        ""
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        ""
      ],
      "remove": [
        ""
      ],
      "Custom field name": [
        ""
      ],
      "Disabled": [
        ""
      ],
      "No deadline": [
        ""
      ],
      "Deadline at %1$s": [
        ""
      ],
      "Select date to show nearby orders": [
        ""
      ],
      "Only show unpaid orders": [
        ""
      ],
      "Only show paid orders": [
        ""
      ],
      "Paid": [
        ""
      ],
      "Only show orders with refunds": [
        ""
      ],
      "Refunded": [
        "Rimborsato"
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        ""
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        ""
      ],
      "Completed": [
        ""
      ],
      "Remove all filters": [
        ""
      ],
      "Clear date filter": [
        ""
      ],
      "Jump to date (%1$s)": [
        ""
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        ""
      ],
      "Order id": [
        ""
      ],
      "Instance unknown": [
        ""
      ],
      "Order unknown": [
        ""
      ],
      "Create order": [
        ""
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        "Data"
      ],
      "Refund": [
        ""
      ],
      "copy url": [
        ""
      ],
      "Load more orders after the last one": [
        ""
      ],
      "No orders have been found matching your query!": [
        ""
      ],
      "Duplicated": [
        ""
      ],
      "This value exceeds the refundable amount.": [
        ""
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        ""
      ],
      "refund": [
        "Rimborsato"
      ],
      "Reason": [
        ""
      ],
      "Amount to be refunded": [
        ""
      ],
      "Max refundable:": [
        ""
      ],
      "Choose one...": [
        ""
      ],
      "Requested by the customer": [
        ""
      ],
      "Other": [
        ""
      ],
      "Why this order is being refunded": [
        ""
      ],
      "More information to give context": [
        ""
      ],
      "now": [
        ""
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        ""
      ],
      "This wire transfer has been notified by the payment service provider.": [
        ""
      ],
      "This refund is waiting to be claimed by the customer.": [
        ""
      ],
      "This refund has been claimed by the customer.": [
        ""
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        ""
      ],
      "Human-readable description of the whole purchase": [
        ""
      ],
      "Total price for the transaction": [
        ""
      ],
      "URL for this purchase": [
        ""
      ],
      "Max fee": [
        ""
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        ""
      ],
      "Time when this contract was generated": [
        ""
      ],
      "Refund deadline": [
        ""
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        ""
      ],
      "Payment deadline": [
        ""
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        ""
      ],
      "Wire transfer deadline": [
        ""
      ],
      "Transfer deadline for the exchange": [
        ""
      ],
      "Time indicating when the order should be delivered": [
        ""
      ],
      "Where the order will be delivered": [
        ""
      ],
      "Auto-refund delay": [
        ""
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        ""
      ],
      "Extra info": [
        ""
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        ""
      ],
      "order created": [
        ""
      ],
      "pay deadline": [
        ""
      ],
      "refund deadline": [
        "Rimborsato"
      ],
      "delivery": [
        ""
      ],
      "Order": [
        ""
      ],
      "Claimed": [
        ""
      ],
      "Claimed at": [
        ""
      ],
      "Timeline": [
        ""
      ],
      "Payment details": [
        ""
      ],
      "Order status": [
        ""
      ],
      "Unpaid": [
        ""
      ],
      "Product list": [
        ""
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        ""
      ],
      "refund created: %1$s": [
        ""
      ],
      "refund created: %1$s: %2$s": [
        ""
      ],
      "refund taken: %1$s": [
        ""
      ],
      "refund taken: %1$s: %2$s": [
        ""
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        ""
      ],
      "Wired": [
        ""
      ],
      "Refund order": [
        "Rimborsato"
      ],
      "Not refundable": [
        ""
      ],
      "Next event in": [
        ""
      ],
      "The order was wired.": [
        ""
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        ""
      ],
      "Refund taken": [
        ""
      ],
      "Status URL": [
        ""
      ],
      "Refund URI": [
        ""
      ],
      "Pay at": [
        ""
      ],
      "Order status URL": [
        ""
      ],
      "Payment URI": [
        ""
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        ""
      ],
      "Invalid. Please enter letters and numbers only.": [
        ""
      ],
      "Just letters and numbers from 2 to 7": [
        ""
      ],
      "The size of the key must be 32 characters": [
        ""
      ],
      "Internal ID on the system": [
        ""
      ],
      "Useful to identify the device physically": [
        ""
      ],
      "Verification algorithm": [
        ""
      ],
      "Algorithm used to verify transactions in offline mode": [
        ""
      ],
      "Device key": [
        ""
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        ""
      ],
      "Your device needs to match exactly the same value": [
        ""
      ],
      "Generate random secret key": [
        ""
      ],
      "Random": [
        ""
      ],
      "Create another": [
        ""
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        ""
      ],
      "OTP devices": [
        ""
      ],
      "Delete selected devices from the database": [
        ""
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        ""
      ],
      "Template id is unknown": [
        ""
      ],
      "The provided information is inconsistent with the current state of the template": [
        ""
      ],
      "Device:": [
        ""
      ],
      "Algorithm to use to verify transactions in offline mode": [
        ""
      ],
      "Not modified": [
        ""
      ],
      "Be sure to be very hard to guess or use the random generator": [
        ""
      ],
      "Your device need to have exactly the same value": [
        ""
      ],
      "Change key": [
        ""
      ],
      "Is not the same": [
        ""
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        ""
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        ""
      ],
      "Testing password change": [
        ""
      ],
      "Password changed": [
        ""
      ],
      "The current password is wrong.": [
        ""
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        "Importo"
      ],
      "Add element to the list": [
        ""
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        ""
      ],
      "Manage stock": [
        ""
      ],
      "This product has been configured without stock control": [
        ""
      ],
      "Infinite": [
        ""
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        ""
      ],
      "Incoming": [
        ""
      ],
      "Lost": [
        ""
      ],
      "Current": [
        ""
      ],
      "Remove stock control for this product": [
        ""
      ],
      "without stock": [
        ""
      ],
      "Next restock": [
        ""
      ],
      "Warehouse address": [
        ""
      ],
      "Invalid amount": [
        ""
      ],
      "Product identification to use in URLs (for internal use only).": [
        ""
      ],
      "Illustration of the product for customers.": [
        ""
      ],
      "Product name.": [
        ""
      ],
      "Product description for customers.": [
        ""
      ],
      "Age restriction": [
        ""
      ],
      "Is this product restricted for customer below certain age?": [
        ""
      ],
      "Minimum age of the customer": [
        ""
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        ""
      ],
      "Example: kg, items or liters": [
        ""
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        ""
      ],
      "Stock": [
        ""
      ],
      "Inventory for products with finite supply (for internal use only).": [
        ""
      ],
      "Taxes included in the product price, exposed to customers.": [
        ""
      ],
      "Search by category description or id": [
        ""
      ],
      "Categories where this product will be listed on.": [
        ""
      ],
      "Add product to inventory": [
        ""
      ],
      "Product not found.": [
        ""
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        ""
      ],
      "Sold": [
        ""
      ],
      "Free": [
        ""
      ],
      "Go to product update page": [
        ""
      ],
      "Update": [
        ""
      ],
      "Remove this product from the database": [
        ""
      ],
      "Update the product with new price": [
        ""
      ],
      "Update product with new price": [
        ""
      ],
      "Confirm update": [
        ""
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        ""
      ],
      "Add more elements to the inventory": [
        ""
      ],
      "Report elements lost in the inventory": [
        ""
      ],
      "New price for the product": [
        ""
      ],
      "There are values with errors.": [
        ""
      ],
      "Update product with new stock and price": [
        ""
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        ""
      ],
      "Product (ID: %1$s) has been deleted": [
        ""
      ],
      "Jump to product with the given product ID": [
        ""
      ],
      "Product id": [
        ""
      ],
      "Delete product": [
        ""
      ],
      'Delete the product "%1$s"': [
        ""
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        ""
      ],
      "Deleting a product cannot be undone.": [
        ""
      ],
      "Product id:": [
        ""
      ],
      "Must be greater that 0": [
        ""
      ],
      "Too short": [
        ""
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        ""
      ],
      "Name of the template in URLs.": [
        ""
      ],
      "Template name": [
        ""
      ],
      "Describe what this template stands for": [
        ""
      ],
      "Order summary": [
        ""
      ],
      "If specified here, this template will create orders with the same summary": [
        ""
      ],
      "Summary is editable": [
        ""
      ],
      "Allow the user to change the summary.": [
        ""
      ],
      "If specified here, this template will create orders with the same price": [
        ""
      ],
      "Amount is editable": [
        ""
      ],
      "Allow the user to select the amount to pay.": [
        ""
      ],
      "Currency is editable": [
        ""
      ],
      "Allow the user to change currency.": [
        ""
      ],
      "Supported currencies": [
        ""
      ],
      "Supported currencies: %1$s": [
        ""
      ],
      "Minimum age": [
        ""
      ],
      "Is this contract restricted to some age?": [
        ""
      ],
      "Payment timeout": [
        ""
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        ""
      ],
      "OTP device": [
        ""
      ],
      "Use to verify transactions in offline mode.": [
        ""
      ],
      "No OTP device.": [
        ""
      ],
      "Add one first": [
        ""
      ],
      "No device": [
        ""
      ],
      "Add new templates": [
        ""
      ],
      "Load more templates before the first one": [
        ""
      ],
      "Delete selected templates from the database": [
        ""
      ],
      "Use template to create new order": [
        ""
      ],
      "Generate a QR code for the template.": [
        ""
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        ""
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        ""
      ],
      "Jump to template with the given template ID": [
        ""
      ],
      "Template identification": [
        ""
      ],
      "Delete template": [
        ""
      ],
      'Delete the template "%1$s"': [
        ""
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        ""
      ],
      "Deleting a template cannot be undone.": [
        ""
      ],
      "Print": [
        ""
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        ""
      ],
      "If specified, this template will create orders with the same price": [
        ""
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        ""
      ],
      "An amount is required": [
        ""
      ],
      "An order summary is required": [
        ""
      ],
      'No more stock for product with ID "%1$s".': [
        ""
      ],
      "New order from template": [
        ""
      ],
      "Amount of the order": [
        ""
      ],
      "Slug": [
        ""
      ],
      "Token family slug to use in URLs (for internal use only)": [
        ""
      ],
      "Kind": [
        ""
      ],
      "Token family kind": [
        ""
      ],
      "User-readable token family name": [
        ""
      ],
      "Token family description for customers": [
        ""
      ],
      "Valid After": [
        ""
      ],
      "Token family can issue tokens after this date": [
        ""
      ],
      "Valid Before": [
        ""
      ],
      "Token family can issue tokens until this date": [
        ""
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        ""
      ],
      "Token Families": [
        ""
      ],
      "Add token family": [
        ""
      ],
      "Go to token family update page": [
        ""
      ],
      "Remove this token family from the database": [
        ""
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        ""
      ],
      "Token family has been deleted.": [
        ""
      ],
      "Delete token family": [
        ""
      ],
      'Delete the token family "%1$s"': [
        ""
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        ""
      ],
      "Deleting a token family %1$s .": [
        ""
      ],
      "can't be undone": [
        ""
      ],
      "Token Family: %1$s": [
        ""
      ],
      "Incoming wire transfers": [
        ""
      ],
      "Load more wire transfers preceding the first one": [
        ""
      ],
      "Expected credit": [
        ""
      ],
      "Confirmed": [
        ""
      ],
      "Validated": [
        ""
      ],
      "Executed on": [
        ""
      ],
      "yes": [
        ""
      ],
      "no": [
        ""
      ],
      "never": [
        ""
      ],
      "unknown": [
        ""
      ],
      "Load more transfers after the last one": [
        ""
      ],
      "Verified wire transfers": [
        ""
      ],
      "Verified": [
        ""
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        ""
      ],
      "All accounts": [
        ""
      ],
      "Filter by account address": [
        ""
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        ""
      ],
      "Only show wire transfers confirmed by the merchant": [
        ""
      ],
      "Wire transfer already confirmed.": [
        ""
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        ""
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        ""
      ],
      "Transfer ID": [
        ""
      ],
      "It's not the same.": [
        ""
      ],
      'You are deleting the instance with ID "%1$s"': [
        ""
      ],
      "Instance": [
        ""
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        ""
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        ""
      ],
      "Instance id": [
        ""
      ],
      "Delete this instance": [
        ""
      ],
      "Must be one of '%1$s'": [
        ""
      ],
      "URL is invalid": [
        ""
      ],
      "Webhook ID to use": [
        ""
      ],
      "Event": [
        ""
      ],
      "Payment": [
        ""
      ],
      "Order settled": [
        ""
      ],
      "Category added": [
        ""
      ],
      "Category updated": [
        ""
      ],
      "Inventory added": [
        ""
      ],
      "Inventory updated": [
        ""
      ],
      "Inventory deleted": [
        ""
      ],
      "The event of the webhook: why the webhook is used": [
        ""
      ],
      "Method": [
        ""
      ],
      "GET": [
        ""
      ],
      "POST": [
        ""
      ],
      "PUT": [
        ""
      ],
      "PATCH": [
        ""
      ],
      "HEAD": [
        ""
      ],
      "Method used by the webhook": [
        ""
      ],
      "URL": [
        ""
      ],
      "URL of the webhook where the customer will be redirected": [
        ""
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        ""
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        ""
      ],
      "The short list of variables are:": [
        ""
      ],
      "order's description": [
        ""
      ],
      "order's price": [
        ""
      ],
      "order's unique identification": [
        ""
      ],
      "the amount that was being refunded": [
        ""
      ],
      "the reason entered by the merchant staff for granting the refund": [
        ""
      ],
      "time of the refund in nanoseconds since 1970": [
        ""
      ],
      "Http header": [
        ""
      ],
      "Header template of the webhook": [
        ""
      ],
      "Http body": [
        ""
      ],
      "Body template used by the webhook.": [
        ""
      ],
      "Webhook deleted successfully": [
        ""
      ],
      "Add new webhooks": [
        ""
      ],
      "Load more webhooks before the first one": [
        ""
      ],
      "Event type": [
        ""
      ],
      "Delete selected webhook from the database": [
        ""
      ],
      "Load more webhooks after the last one": [
        ""
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        ""
      ],
      "Webhook updated": [
        ""
      ],
      "Header": [
        ""
      ],
      "Body": [
        ""
      ],
      "Body template used by the webhook": [
        ""
      ],
      "There is another instance with this username.": [
        ""
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        ""
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        ""
      ],
      "Password reset": [
        ""
      ],
      'Resetting access to the instance "%1$s"': [
        ""
      ],
      "Reset": [
        ""
      ],
      "Language": [
        ""
      ],
      "Date format": [
        ""
      ],
      "How the date is going to be displayed": [
        ""
      ],
      "Merchant type": [
        ""
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        ""
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        ""
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        ""
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        ""
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        ""
      ],
      "hour": [
        ""
      ],
      "day": [
        ""
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        ""
      ],
      "Time range": [
        ""
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        ""
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        ""
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        ""
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        ""
      ],
      "Hide for today": [
        ""
      ],
      "KYC verification needed": [
        ""
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        ""
      ],
      "The request reached a timeout, check your connection.": [
        ""
      ],
      "The request was cancelled.": [
        ""
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        ""
      ],
      "The server's response was malformed.": [
        ""
      ],
      "Could not complete the request due to a network problem.": [
        ""
      ],
      "Unexpected request error.": [
        ""
      ],
      "Unexpected error.": [
        ""
      ],
      "Add new instance": [
        ""
      ],
      "Edit": [
        ""
      ],
      "Change password": [
        ""
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        ""
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        ""
      ],
      "Only show active instances": [
        ""
      ],
      "Active": [
        ""
      ],
      "Only show deleted instances": [
        ""
      ],
      "Deleted": [
        ""
      ],
      "Show all instances": [
        ""
      ],
      "Delete instance": [
        ""
      ],
      'Delete the instance "%1$s"': [
        ""
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        ""
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        ""
      ],
      "Deleting an instance %1$s": [
        ""
      ],
      "This cannot be undone!": [
        ""
      ],
      "Purge the instance": [
        ""
      ],
      'Purge the instance "%1$s"': [
        ""
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        ""
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        ""
      ],
      "Purging an instance %1$s": [
        ""
      ],
      "checking compatibility with server...": [
        ""
      ],
      "Contacting the server failed": [
        ""
      ],
      "The server version is not supported": [
        ""
      ],
      'Supported version "%1$s", server version "%2$s".': [
        ""
      ],
      "Deleting": [
        ""
      ],
      "Changing": [
        ""
      ],
      "Manage password": [
        ""
      ],
      "Business Name": [
        ""
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        "Impostazioni"
      ],
      "title%1$s: New bank account": [
        ""
      ],
      "title%1$s: Bank accounts": [
        ""
      ],
      "title%1$s: Update bank Account": [
        ""
      ],
      "title%1$s: Orders": [
        ""
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        ""
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        ""
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        ""
      ],
      "title%1$s: Wire transfers": [
        ""
      ],
      "title%1$s: Webhooks": [
        ""
      ],
      "title%1$s: New webhook": [
        ""
      ],
      "title%1$s: Update webhook": [
        ""
      ],
      "title%1$s: OTP devices": [
        ""
      ],
      "title%1$s: New OTP device": [
        ""
      ],
      "title%1$s: Update OTP device": [
        ""
      ],
      "title%1$s: New template": [
        ""
      ],
      "title%1$s: Update template": [
        ""
      ],
      "title%1$s: Templates": [
        ""
      ],
      "title%1$s: Use template": [
        ""
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        ""
      ],
      "title%1$s: New token family": [
        ""
      ],
      "title%1$s: Update token family": [
        ""
      ],
      "title%1$s: Access tokens": [
        ""
      ],
      "title%1$s: New access token": [
        ""
      ],
      "titleNew instance": [
        ""
      ],
      "titleInstances": [
        ""
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=2; plural=n != 1;",
  "lang": "it",
  "completeness": 2
};
strings["he"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=4; plural=(n == 1) ? 0 : ((n == 2) ? 1 : ((n > 10 && n % 10 == 0) ? 2 : 3));",
        "lang": "he"
      },
      "Cancel": [
        ""
      ],
      "%1$s": [
        ""
      ],
      "Close": [
        ""
      ],
      "Continue": [
        ""
      ],
      "Clear": [
        ""
      ],
      "Confirm": [
        ""
      ],
      "Correct form": [
        ""
      ],
      "Comparing account details": [
        ""
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        ""
      ],
      "Field": [
        ""
      ],
      "In the form": [
        ""
      ],
      "Reported": [
        ""
      ],
      "Type": [
        ""
      ],
      "IBAN": [
        ""
      ],
      "Address": [
        ""
      ],
      "Host": [
        ""
      ],
      "Account ID": [
        ""
      ],
      "The account owner's legal name": [
        ""
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        ""
      ],
      "Bank host": [
        ""
      ],
      "Bank account": [
        ""
      ],
      "BIC": [
        ""
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        ""
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        ""
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        ""
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        ""
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        ""
      ],
      "Subject": [
        ""
      ],
      "Receiver postal code": [
        ""
      ],
      "Receiver town": [
        ""
      ],
      "Step 3:": [
        ""
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        ""
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        ""
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        ""
      ],
      "Operation in progress...": [
        ""
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        ""
      ],
      "days": [
        ""
      ],
      "hours": [
        ""
      ],
      "minutes": [
        ""
      ],
      "seconds": [
        ""
      ],
      "Forever": [
        ""
      ],
      "%1$sM": [
        ""
      ],
      "%1$sY": [
        ""
      ],
      "%1$sd": [
        ""
      ],
      "%1$sh": [
        ""
      ],
      "%1$smin": [
        ""
      ],
      "%1$ssec": [
        ""
      ],
      "Change value to never": [
        ""
      ],
      "Change value to empty": [
        ""
      ],
      "Image must be smaller than 1 MB": [
        ""
      ],
      "Add": [
        ""
      ],
      "Remove": [
        ""
      ],
      "Country": [
        ""
      ],
      "Building number": [
        ""
      ],
      "Building name": [
        ""
      ],
      "Street": [
        ""
      ],
      "Post code": [
        ""
      ],
      "Town location": [
        ""
      ],
      "District": [
        ""
      ],
      "Country subdivision": [
        ""
      ],
      "Additional address lines": [
        ""
      ],
      "Orders": [
        ""
      ],
      "Inventory": [
        ""
      ],
      "Categories": [
        ""
      ],
      "Wire transfers": [
        ""
      ],
      "Templates": [
        ""
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        ""
      ],
      "OTP Devices": [
        ""
      ],
      "Webhooks": [
        ""
      ],
      "Settings": [
        ""
      ],
      "Password": [
        ""
      ],
      "Access tokens": [
        ""
      ],
      "Connection": [
        ""
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        ""
      ],
      "New": [
        ""
      ],
      "List": [
        ""
      ],
      "Log out": [
        ""
      ],
      "Username": [
        ""
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        ""
      ],
      "Business name": [
        ""
      ],
      "Legal name of the business represented by this instance.": [
        ""
      ],
      "Email": [
        ""
      ],
      "Contact email": [
        ""
      ],
      "Phone number": [
        ""
      ],
      "Contact phone": [
        ""
      ],
      "Website URL": [
        ""
      ],
      "URL.": [
        ""
      ],
      "Logo": [
        ""
      ],
      "Logo image.": [
        ""
      ],
      "Physical location of the merchant.": [
        ""
      ],
      "Jurisdiction": [
        ""
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        ""
      ],
      "Pay transaction fee": [
        ""
      ],
      "Cover the transaction cost or pass it on to the user.": [
        ""
      ],
      "Default payment delay": [
        ""
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        ""
      ],
      "Default refund delay": [
        ""
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        ""
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        ""
      ],
      "Default wire transfer rounding interval": [
        ""
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        ""
      ],
      "To second": [
        ""
      ],
      "To minute": [
        ""
      ],
      "To hour": [
        ""
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        ""
      ],
      "Required": [
        ""
      ],
      "Unauthorized": [
        ""
      ],
      "Sending the code failed": [
        ""
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        ""
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        ""
      ],
      "Verify": [
        ""
      ],
      "Failed to send the verification code.": [
        ""
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        ""
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        ""
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        ""
      ],
      "Login required": [
        ""
      ],
      "instance name": [
        ""
      ],
      "current password": [
        ""
      ],
      "Forgot password": [
        ""
      ],
      "Create new account": [
        ""
      ],
      "Invalid": [
        ""
      ],
      "Doesn't have the pattern of an email": [
        ""
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        ""
      ],
      "Max 7 lines": [
        ""
      ],
      "Doesn't match": [
        ""
      ],
      "Instance created": [
        ""
      ],
      "Unauthorized.": [
        ""
      ],
      "Conflict.": [
        ""
      ],
      "Not found.": [
        ""
      ],
      "New password": [
        ""
      ],
      "Next password to be used": [
        ""
      ],
      "Repeat password": [
        ""
      ],
      "Confirm the same password": [
        ""
      ],
      "Please complete the marked fields and choose authorization method": [
        ""
      ],
      "Confirm operation": [
        ""
      ],
      "Check the password.": [
        ""
      ],
      "Instance not found.": [
        ""
      ],
      "Description": [
        ""
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        ""
      ],
      "Time the access token will be valid.": [
        ""
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        ""
      ],
      "Read only": [
        ""
      ],
      "All": [
        ""
      ],
      "Order simple": [
        ""
      ],
      "Order Point-of-Sale": [
        ""
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        ""
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        ""
      ],
      "Please complete the marked fields": [
        ""
      ],
      "Access token created": [
        ""
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        ""
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        ""
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        ""
      ],
      "Create access token": [
        ""
      ],
      "Load more devices before the first one": [
        ""
      ],
      "Load first page": [
        ""
      ],
      "Created at": [
        ""
      ],
      "Expires at": [
        ""
      ],
      "Never": [
        ""
      ],
      "Remove this access token": [
        ""
      ],
      "Delete": [
        ""
      ],
      "Load more devices after the last one": [
        ""
      ],
      "Load next page": [
        ""
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        ""
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        ""
      ],
      "Deleting an access token cannot be undone.": [
        ""
      ],
      "This is not a valid Bitcoin address.": [
        ""
      ],
      "This is not a valid Ethereum address.": [
        ""
      ],
      "This is not a valid host.": [
        ""
      ],
      "IBANs usually have more than 4 digits.": [
        ""
      ],
      "IBANs usually have fewer than 34 digits.": [
        ""
      ],
      "The IBAN's country code could not be retrieved.": [
        ""
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        ""
      ],
      "This account is not allowed.": [
        ""
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        ""
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        ""
      ],
      "Select a wire method...": [
        ""
      ],
      "Routing": [
        ""
      ],
      "Routing number": [
        ""
      ],
      "Account": [
        ""
      ],
      "Account number": [
        ""
      ],
      "Code": [
        ""
      ],
      "Business Identifier Code": [
        ""
      ],
      "International Bank Account Number": [
        ""
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        ""
      ],
      "Bitcoin protocol": [
        ""
      ],
      "Ethereum protocol": [
        ""
      ],
      "Interledger protocol": [
        ""
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        ""
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        ""
      ],
      "Invalid url": [
        ""
      ],
      "URL must end with a '/'": [
        ""
      ],
      "URL must not contain params": [
        ""
      ],
      "URL must not hash param": [
        ""
      ],
      "Conflict": [
        ""
      ],
      'Server replied with "bad request".': [
        ""
      ],
      "Unauthorized, check credentials.": [
        ""
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        ""
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        ""
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        ""
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        ""
      ],
      "Endpoint URL": [
        ""
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        ""
      ],
      "Auth type": [
        ""
      ],
      "Choose the authentication type for the account info URL": [
        ""
      ],
      "Without authentication": [
        ""
      ],
      "With username and password": [
        ""
      ],
      "With token": [
        ""
      ],
      "Do not change": [
        ""
      ],
      "Username to access the account information.": [
        ""
      ],
      "Password to access the account information.": [
        ""
      ],
      "Access token to access the account information.": [
        ""
      ],
      "Match": [
        ""
      ],
      "Check where the information match against the server info.": [
        ""
      ],
      "Compare info from server with account form": [
        ""
      ],
      "Test": [
        ""
      ],
      "Server didn't like the request we made.": [
        ""
      ],
      "Account details": [
        ""
      ],
      "Check if the information matches the server info.": [
        ""
      ],
      "The bank account has been deleted.": [
        ""
      ],
      "Bank accounts": [
        ""
      ],
      "Add new account": [
        ""
      ],
      "Owner's name": [
        ""
      ],
      "Delete selected accounts from the database": [
        ""
      ],
      "There are no accounts yet, add more pressing the + sign": [
        ""
      ],
      "You must provide a bank account to receive payments.": [
        ""
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        ""
      ],
      "Name": [
        ""
      ],
      "Category name": [
        ""
      ],
      "Category deleted": [
        ""
      ],
      "Add new devices": [
        ""
      ],
      "ID": [
        ""
      ],
      "Total products": [
        ""
      ],
      "Delete selected category from the database": [
        ""
      ],
      "There are no categories yet, add more pressing the + sign": [
        ""
      ],
      "Id:": [
        ""
      ],
      "Name of the category": [
        ""
      ],
      "Products": [
        ""
      ],
      "Image": [
        ""
      ],
      "Load more products after the last one": [
        ""
      ],
      "There are no products in this category.": [
        ""
      ],
      "Account's KYC status": [
        ""
      ],
      "Exchange": [
        ""
      ],
      "Status": [
        ""
      ],
      "Ok": [
        ""
      ],
      "Action required": [
        ""
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        ""
      ],
      "More information required.": [
        ""
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        ""
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        ""
      ],
      "No pending kyc verification!": [
        ""
      ],
      "Wire account is invalid": [
        ""
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        ""
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        ""
      ],
      "The payment provider requires more information.": [
        ""
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        ""
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        ""
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        ""
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        ""
      ],
      "Enter description or id": [
        ""
      ],
      "No match found for that description or ID.": [
        ""
      ],
      "You must enter a valid product identifier.": [
        ""
      ],
      "Quantity must be greater than 0!": [
        ""
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        ""
      ],
      "Search product": [
        ""
      ],
      "Quantity": [
        ""
      ],
      "How many products will be added": [
        ""
      ],
      "Add from inventory": [
        ""
      ],
      "This product has %1$s applicable taxes configured.": [
        ""
      ],
      "No taxes configured for this product.": [
        ""
      ],
      "Amount": [
        ""
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        ""
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        ""
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        ""
      ],
      "Add tax to the tax list": [
        ""
      ],
      "Describe and add a product that is not in the inventory list": [
        ""
      ],
      "Add custom product": [
        ""
      ],
      "Complete information of the product": [
        ""
      ],
      "Must be a number.": [
        ""
      ],
      "Must be greater than 0.": [
        ""
      ],
      "Photo of the product.": [
        ""
      ],
      "Full product description.": [
        ""
      ],
      "Unit name": [
        ""
      ],
      "Name of the product unit.": [
        ""
      ],
      "Price per unit": [
        ""
      ],
      "Amount in the current currency.": [
        ""
      ],
      "How many products will be added.": [
        ""
      ],
      "Taxes": [
        ""
      ],
      "Total price": [
        ""
      ],
      "Must be greater than 0": [
        ""
      ],
      "Must have a refund deadline": [
        ""
      ],
      "Auto refund can't be after refund deadline": [
        ""
      ],
      "Must be in the future": [
        ""
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        ""
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        ""
      ],
      "Manage products in order": [
        ""
      ],
      "%1$s products with a total price of %2$s.": [
        ""
      ],
      "Manage list of products in the order.": [
        ""
      ],
      "Remove this product from the order.": [
        ""
      ],
      "Products price sum": [
        ""
      ],
      "Total product price added up": [
        ""
      ],
      "Order price": [
        ""
      ],
      "Amount to be paid by the customer": [
        ""
      ],
      "Final order price": [
        ""
      ],
      "Summary": [
        ""
      ],
      "Title of the order to be shown to the customer": [
        ""
      ],
      "Shipping and fulfillment": [
        ""
      ],
      "Delivery date": [
        ""
      ],
      "Deadline for physical delivery assured by the merchant.": [
        ""
      ],
      "Delivery location": [
        ""
      ],
      "Address where the products will be delivered": [
        ""
      ],
      "Fulfillment URL": [
        ""
      ],
      "URL to which the user will be redirected after successful payment.": [
        ""
      ],
      "Fulfillment message": [
        ""
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        ""
      ],
      "Override default Taler payment settings for this order": [
        ""
      ],
      "Payment time": [
        ""
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        ""
      ],
      "Default": [
        ""
      ],
      "Refund time": [
        ""
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        ""
      ],
      "Wire transfer time": [
        ""
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        ""
      ],
      "Auto-refund time": [
        ""
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        ""
      ],
      "Maximum fee": [
        ""
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        ""
      ],
      "Create token": [
        ""
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        ""
      ],
      "Minimum age required": [
        ""
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        ""
      ],
      "Minimum age defined by the products is %1$s": [
        ""
      ],
      "No product with age restriction in this order": [
        ""
      ],
      "Additional information": [
        ""
      ],
      "Custom information to be included in the contract for this order.": [
        ""
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        ""
      ],
      "remove": [
        ""
      ],
      "Custom field name": [
        ""
      ],
      "Disabled": [
        ""
      ],
      "No deadline": [
        ""
      ],
      "Deadline at %1$s": [
        ""
      ],
      "Select date to show nearby orders": [
        ""
      ],
      "Only show unpaid orders": [
        ""
      ],
      "Only show paid orders": [
        ""
      ],
      "Paid": [
        ""
      ],
      "Only show orders with refunds": [
        ""
      ],
      "Refunded": [
        ""
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        ""
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        ""
      ],
      "Completed": [
        ""
      ],
      "Remove all filters": [
        ""
      ],
      "Clear date filter": [
        ""
      ],
      "Jump to date (%1$s)": [
        ""
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        ""
      ],
      "Order id": [
        ""
      ],
      "Instance unknown": [
        ""
      ],
      "Order unknown": [
        ""
      ],
      "Create order": [
        ""
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        ""
      ],
      "Refund": [
        ""
      ],
      "copy url": [
        ""
      ],
      "Load more orders after the last one": [
        ""
      ],
      "No orders have been found matching your query!": [
        ""
      ],
      "Duplicated": [
        ""
      ],
      "This value exceeds the refundable amount.": [
        ""
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        ""
      ],
      "refund": [
        ""
      ],
      "Reason": [
        ""
      ],
      "Amount to be refunded": [
        ""
      ],
      "Max refundable:": [
        ""
      ],
      "Choose one...": [
        ""
      ],
      "Requested by the customer": [
        ""
      ],
      "Other": [
        ""
      ],
      "Why this order is being refunded": [
        ""
      ],
      "More information to give context": [
        ""
      ],
      "now": [
        ""
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        ""
      ],
      "This wire transfer has been notified by the payment service provider.": [
        ""
      ],
      "This refund is waiting to be claimed by the customer.": [
        ""
      ],
      "This refund has been claimed by the customer.": [
        ""
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        ""
      ],
      "Human-readable description of the whole purchase": [
        ""
      ],
      "Total price for the transaction": [
        ""
      ],
      "URL for this purchase": [
        ""
      ],
      "Max fee": [
        ""
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        ""
      ],
      "Time when this contract was generated": [
        ""
      ],
      "Refund deadline": [
        ""
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        ""
      ],
      "Payment deadline": [
        ""
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        ""
      ],
      "Wire transfer deadline": [
        ""
      ],
      "Transfer deadline for the exchange": [
        ""
      ],
      "Time indicating when the order should be delivered": [
        ""
      ],
      "Where the order will be delivered": [
        ""
      ],
      "Auto-refund delay": [
        ""
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        ""
      ],
      "Extra info": [
        ""
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        ""
      ],
      "order created": [
        ""
      ],
      "pay deadline": [
        ""
      ],
      "refund deadline": [
        ""
      ],
      "delivery": [
        ""
      ],
      "Order": [
        ""
      ],
      "Claimed": [
        ""
      ],
      "Claimed at": [
        ""
      ],
      "Timeline": [
        ""
      ],
      "Payment details": [
        ""
      ],
      "Order status": [
        ""
      ],
      "Unpaid": [
        ""
      ],
      "Product list": [
        ""
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        ""
      ],
      "refund created: %1$s": [
        ""
      ],
      "refund created: %1$s: %2$s": [
        ""
      ],
      "refund taken: %1$s": [
        ""
      ],
      "refund taken: %1$s: %2$s": [
        ""
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        ""
      ],
      "Wired": [
        ""
      ],
      "Refund order": [
        ""
      ],
      "Not refundable": [
        ""
      ],
      "Next event in": [
        ""
      ],
      "The order was wired.": [
        ""
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        ""
      ],
      "Refund taken": [
        ""
      ],
      "Status URL": [
        ""
      ],
      "Refund URI": [
        ""
      ],
      "Pay at": [
        ""
      ],
      "Order status URL": [
        ""
      ],
      "Payment URI": [
        ""
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        ""
      ],
      "Invalid. Please enter letters and numbers only.": [
        ""
      ],
      "Just letters and numbers from 2 to 7": [
        ""
      ],
      "The size of the key must be 32 characters": [
        ""
      ],
      "Internal ID on the system": [
        ""
      ],
      "Useful to identify the device physically": [
        ""
      ],
      "Verification algorithm": [
        ""
      ],
      "Algorithm used to verify transactions in offline mode": [
        ""
      ],
      "Device key": [
        ""
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        ""
      ],
      "Your device needs to match exactly the same value": [
        ""
      ],
      "Generate random secret key": [
        ""
      ],
      "Random": [
        ""
      ],
      "Create another": [
        ""
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        ""
      ],
      "OTP devices": [
        ""
      ],
      "Delete selected devices from the database": [
        ""
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        ""
      ],
      "Template id is unknown": [
        ""
      ],
      "The provided information is inconsistent with the current state of the template": [
        ""
      ],
      "Device:": [
        ""
      ],
      "Algorithm to use to verify transactions in offline mode": [
        ""
      ],
      "Not modified": [
        ""
      ],
      "Be sure to be very hard to guess or use the random generator": [
        ""
      ],
      "Your device need to have exactly the same value": [
        ""
      ],
      "Change key": [
        ""
      ],
      "Is not the same": [
        ""
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        ""
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        ""
      ],
      "Testing password change": [
        ""
      ],
      "Password changed": [
        ""
      ],
      "The current password is wrong.": [
        ""
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        ""
      ],
      "Add element to the list": [
        ""
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        ""
      ],
      "Manage stock": [
        ""
      ],
      "This product has been configured without stock control": [
        ""
      ],
      "Infinite": [
        ""
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        ""
      ],
      "Incoming": [
        ""
      ],
      "Lost": [
        ""
      ],
      "Current": [
        ""
      ],
      "Remove stock control for this product": [
        ""
      ],
      "without stock": [
        ""
      ],
      "Next restock": [
        ""
      ],
      "Warehouse address": [
        ""
      ],
      "Invalid amount": [
        ""
      ],
      "Product identification to use in URLs (for internal use only).": [
        ""
      ],
      "Illustration of the product for customers.": [
        ""
      ],
      "Product name.": [
        ""
      ],
      "Product description for customers.": [
        ""
      ],
      "Age restriction": [
        ""
      ],
      "Is this product restricted for customer below certain age?": [
        ""
      ],
      "Minimum age of the customer": [
        ""
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        ""
      ],
      "Example: kg, items or liters": [
        ""
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        ""
      ],
      "Stock": [
        ""
      ],
      "Inventory for products with finite supply (for internal use only).": [
        ""
      ],
      "Taxes included in the product price, exposed to customers.": [
        ""
      ],
      "Search by category description or id": [
        ""
      ],
      "Categories where this product will be listed on.": [
        ""
      ],
      "Add product to inventory": [
        ""
      ],
      "Product not found.": [
        ""
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        ""
      ],
      "Sold": [
        ""
      ],
      "Free": [
        ""
      ],
      "Go to product update page": [
        ""
      ],
      "Update": [
        ""
      ],
      "Remove this product from the database": [
        ""
      ],
      "Update the product with new price": [
        ""
      ],
      "Update product with new price": [
        ""
      ],
      "Confirm update": [
        ""
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        ""
      ],
      "Add more elements to the inventory": [
        ""
      ],
      "Report elements lost in the inventory": [
        ""
      ],
      "New price for the product": [
        ""
      ],
      "There are values with errors.": [
        ""
      ],
      "Update product with new stock and price": [
        ""
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        ""
      ],
      "Product (ID: %1$s) has been deleted": [
        ""
      ],
      "Jump to product with the given product ID": [
        ""
      ],
      "Product id": [
        ""
      ],
      "Delete product": [
        ""
      ],
      'Delete the product "%1$s"': [
        ""
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        ""
      ],
      "Deleting a product cannot be undone.": [
        ""
      ],
      "Product id:": [
        ""
      ],
      "Must be greater that 0": [
        ""
      ],
      "Too short": [
        ""
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        ""
      ],
      "Name of the template in URLs.": [
        ""
      ],
      "Template name": [
        ""
      ],
      "Describe what this template stands for": [
        ""
      ],
      "Order summary": [
        ""
      ],
      "If specified here, this template will create orders with the same summary": [
        ""
      ],
      "Summary is editable": [
        ""
      ],
      "Allow the user to change the summary.": [
        ""
      ],
      "If specified here, this template will create orders with the same price": [
        ""
      ],
      "Amount is editable": [
        ""
      ],
      "Allow the user to select the amount to pay.": [
        ""
      ],
      "Currency is editable": [
        ""
      ],
      "Allow the user to change currency.": [
        ""
      ],
      "Supported currencies": [
        ""
      ],
      "Supported currencies: %1$s": [
        ""
      ],
      "Minimum age": [
        ""
      ],
      "Is this contract restricted to some age?": [
        ""
      ],
      "Payment timeout": [
        ""
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        ""
      ],
      "OTP device": [
        ""
      ],
      "Use to verify transactions in offline mode.": [
        ""
      ],
      "No OTP device.": [
        ""
      ],
      "Add one first": [
        ""
      ],
      "No device": [
        ""
      ],
      "Add new templates": [
        ""
      ],
      "Load more templates before the first one": [
        ""
      ],
      "Delete selected templates from the database": [
        ""
      ],
      "Use template to create new order": [
        ""
      ],
      "Generate a QR code for the template.": [
        ""
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        ""
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        ""
      ],
      "Jump to template with the given template ID": [
        ""
      ],
      "Template identification": [
        ""
      ],
      "Delete template": [
        ""
      ],
      'Delete the template "%1$s"': [
        ""
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        ""
      ],
      "Deleting a template cannot be undone.": [
        ""
      ],
      "Print": [
        ""
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        ""
      ],
      "If specified, this template will create orders with the same price": [
        ""
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        ""
      ],
      "An amount is required": [
        ""
      ],
      "An order summary is required": [
        ""
      ],
      'No more stock for product with ID "%1$s".': [
        ""
      ],
      "New order from template": [
        ""
      ],
      "Amount of the order": [
        ""
      ],
      "Slug": [
        ""
      ],
      "Token family slug to use in URLs (for internal use only)": [
        ""
      ],
      "Kind": [
        ""
      ],
      "Token family kind": [
        ""
      ],
      "User-readable token family name": [
        ""
      ],
      "Token family description for customers": [
        ""
      ],
      "Valid After": [
        ""
      ],
      "Token family can issue tokens after this date": [
        ""
      ],
      "Valid Before": [
        ""
      ],
      "Token family can issue tokens until this date": [
        ""
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        ""
      ],
      "Token Families": [
        ""
      ],
      "Add token family": [
        ""
      ],
      "Go to token family update page": [
        ""
      ],
      "Remove this token family from the database": [
        ""
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        ""
      ],
      "Token family has been deleted.": [
        ""
      ],
      "Delete token family": [
        ""
      ],
      'Delete the token family "%1$s"': [
        ""
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        ""
      ],
      "Deleting a token family %1$s .": [
        ""
      ],
      "can't be undone": [
        ""
      ],
      "Token Family: %1$s": [
        ""
      ],
      "Incoming wire transfers": [
        ""
      ],
      "Load more wire transfers preceding the first one": [
        ""
      ],
      "Expected credit": [
        ""
      ],
      "Confirmed": [
        ""
      ],
      "Validated": [
        ""
      ],
      "Executed on": [
        ""
      ],
      "yes": [
        ""
      ],
      "no": [
        ""
      ],
      "never": [
        ""
      ],
      "unknown": [
        ""
      ],
      "Load more transfers after the last one": [
        ""
      ],
      "Verified wire transfers": [
        ""
      ],
      "Verified": [
        ""
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        ""
      ],
      "All accounts": [
        ""
      ],
      "Filter by account address": [
        ""
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        ""
      ],
      "Only show wire transfers confirmed by the merchant": [
        ""
      ],
      "Wire transfer already confirmed.": [
        ""
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        ""
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        ""
      ],
      "Transfer ID": [
        ""
      ],
      "It's not the same.": [
        ""
      ],
      'You are deleting the instance with ID "%1$s"': [
        ""
      ],
      "Instance": [
        ""
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        ""
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        ""
      ],
      "Instance id": [
        ""
      ],
      "Delete this instance": [
        ""
      ],
      "Must be one of '%1$s'": [
        ""
      ],
      "URL is invalid": [
        ""
      ],
      "Webhook ID to use": [
        ""
      ],
      "Event": [
        ""
      ],
      "Payment": [
        ""
      ],
      "Order settled": [
        ""
      ],
      "Category added": [
        ""
      ],
      "Category updated": [
        ""
      ],
      "Inventory added": [
        ""
      ],
      "Inventory updated": [
        ""
      ],
      "Inventory deleted": [
        ""
      ],
      "The event of the webhook: why the webhook is used": [
        ""
      ],
      "Method": [
        ""
      ],
      "GET": [
        ""
      ],
      "POST": [
        ""
      ],
      "PUT": [
        ""
      ],
      "PATCH": [
        ""
      ],
      "HEAD": [
        ""
      ],
      "Method used by the webhook": [
        ""
      ],
      "URL": [
        ""
      ],
      "URL of the webhook where the customer will be redirected": [
        ""
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        ""
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        ""
      ],
      "The short list of variables are:": [
        ""
      ],
      "order's description": [
        ""
      ],
      "order's price": [
        ""
      ],
      "order's unique identification": [
        ""
      ],
      "the amount that was being refunded": [
        ""
      ],
      "the reason entered by the merchant staff for granting the refund": [
        ""
      ],
      "time of the refund in nanoseconds since 1970": [
        ""
      ],
      "Http header": [
        ""
      ],
      "Header template of the webhook": [
        ""
      ],
      "Http body": [
        ""
      ],
      "Body template used by the webhook.": [
        ""
      ],
      "Webhook deleted successfully": [
        ""
      ],
      "Add new webhooks": [
        ""
      ],
      "Load more webhooks before the first one": [
        ""
      ],
      "Event type": [
        ""
      ],
      "Delete selected webhook from the database": [
        ""
      ],
      "Load more webhooks after the last one": [
        ""
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        ""
      ],
      "Webhook updated": [
        ""
      ],
      "Header": [
        ""
      ],
      "Body": [
        ""
      ],
      "Body template used by the webhook": [
        ""
      ],
      "There is another instance with this username.": [
        ""
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        ""
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        ""
      ],
      "Password reset": [
        ""
      ],
      'Resetting access to the instance "%1$s"': [
        ""
      ],
      "Reset": [
        ""
      ],
      "Language": [
        ""
      ],
      "Date format": [
        ""
      ],
      "How the date is going to be displayed": [
        ""
      ],
      "Merchant type": [
        ""
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        ""
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        ""
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        ""
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        ""
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        ""
      ],
      "hour": [
        ""
      ],
      "day": [
        ""
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        ""
      ],
      "Time range": [
        ""
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        ""
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        ""
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        ""
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        ""
      ],
      "Hide for today": [
        ""
      ],
      "KYC verification needed": [
        ""
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        ""
      ],
      "The request reached a timeout, check your connection.": [
        ""
      ],
      "The request was cancelled.": [
        ""
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        ""
      ],
      "The server's response was malformed.": [
        ""
      ],
      "Could not complete the request due to a network problem.": [
        ""
      ],
      "Unexpected request error.": [
        ""
      ],
      "Unexpected error.": [
        ""
      ],
      "Add new instance": [
        ""
      ],
      "Edit": [
        ""
      ],
      "Change password": [
        ""
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        ""
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        ""
      ],
      "Only show active instances": [
        ""
      ],
      "Active": [
        ""
      ],
      "Only show deleted instances": [
        ""
      ],
      "Deleted": [
        ""
      ],
      "Show all instances": [
        ""
      ],
      "Delete instance": [
        ""
      ],
      'Delete the instance "%1$s"': [
        ""
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        ""
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        ""
      ],
      "Deleting an instance %1$s": [
        ""
      ],
      "This cannot be undone!": [
        ""
      ],
      "Purge the instance": [
        ""
      ],
      'Purge the instance "%1$s"': [
        ""
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        ""
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        ""
      ],
      "Purging an instance %1$s": [
        ""
      ],
      "checking compatibility with server...": [
        ""
      ],
      "Contacting the server failed": [
        ""
      ],
      "The server version is not supported": [
        ""
      ],
      'Supported version "%1$s", server version "%2$s".': [
        ""
      ],
      "Deleting": [
        ""
      ],
      "Changing": [
        ""
      ],
      "Manage password": [
        ""
      ],
      "Business Name": [
        ""
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        ""
      ],
      "title%1$s: New bank account": [
        ""
      ],
      "title%1$s: Bank accounts": [
        ""
      ],
      "title%1$s: Update bank Account": [
        ""
      ],
      "title%1$s: Orders": [
        ""
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        ""
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        ""
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        ""
      ],
      "title%1$s: Wire transfers": [
        ""
      ],
      "title%1$s: Webhooks": [
        ""
      ],
      "title%1$s: New webhook": [
        ""
      ],
      "title%1$s: Update webhook": [
        ""
      ],
      "title%1$s: OTP devices": [
        ""
      ],
      "title%1$s: New OTP device": [
        ""
      ],
      "title%1$s: Update OTP device": [
        ""
      ],
      "title%1$s: New template": [
        ""
      ],
      "title%1$s: Update template": [
        ""
      ],
      "title%1$s: Templates": [
        ""
      ],
      "title%1$s: Use template": [
        ""
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        ""
      ],
      "title%1$s: New token family": [
        ""
      ],
      "title%1$s: Update token family": [
        ""
      ],
      "title%1$s: Access tokens": [
        ""
      ],
      "title%1$s: New access token": [
        ""
      ],
      "titleNew instance": [
        ""
      ],
      "titleInstances": [
        ""
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=4; plural=(n == 1) ? 0 : ((n == 2) ? 1 : ((n > 10 && n % 10 == 0) ? 2 : 3));",
  "lang": "he",
  "completeness": 0
};
strings["fr"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=2; plural=(n!=1);",
        "lang": "fr"
      },
      "Cancel": [
        "Annuler"
      ],
      "%1$s": [
        "%1$s"
      ],
      "Close": [
        "Fermer"
      ],
      "Continue": [
        "Continuer"
      ],
      "Clear": [
        "Effacer"
      ],
      "Confirm": [
        "Confirmer"
      ],
      "Correct form": [
        "Formulaire correct"
      ],
      "Comparing account details": [
        "Comparaison des d\xE9tails du compte"
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        "Le test de l'URL des informations du compte a r\xE9ussi mais les informations sur le compte retourn\xE9es sont diff\xE9rentes de celles du formulaire des d\xE9tails du compte."
      ],
      "Field": [
        "Champs"
      ],
      "In the form": [
        "Dans le formulaire"
      ],
      "Reported": [
        "Rapport\xE9"
      ],
      "Type": [
        "Type"
      ],
      "IBAN": [
        "IBAN"
      ],
      "Address": [
        "Adresse"
      ],
      "Host": [
        "H\xF4te"
      ],
      "Account ID": [
        "Compte"
      ],
      "The account owner's legal name": [
        ""
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        "Ville"
      ],
      "Bank host": [
        "Serveur de la banque"
      ],
      "Bank account": [
        "Compte bancaire"
      ],
      "BIC": [
        "Code BIC"
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        "Valider le compte bancaire\u202F: %1$s"
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        "\xC9tape 1\u202F:"
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        "Copiez et collez l'IBAN et le nom dans les champs destinataires de votre application bancaire ou de votre site web"
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        "\xC9tape 2\u202F:"
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        "Copiez ce code et collez-le dans le champ 'r\xE9f\xE9rence/objet' de votre application bancaire ou du site web de votre banque"
      ],
      "Subject": [
        "R\xE9f\xE9rence"
      ],
      "Receiver postal code": [
        "Nom du b\xE9n\xE9ficiaire"
      ],
      "Receiver town": [
        "Nom du b\xE9n\xE9ficiaire"
      ],
      "Step 3:": [
        "\xC9tape 3\u202F:"
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        "S\xE9lectionnez le montant le plus petit possible pour le virement bancaire dans votre application bancaire ou sur votre site Web de banque en ligne."
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        "Assurez-vous que TOUTES les donn\xE9es sont correctes, y compris la r\xE9f\xE9rence, et que vous utilisez le compte bancaire s\xE9lectionn\xE9. Vous pouvez utiliser les boutons de copie (%1$s) pour \xE9viter les erreurs de frappe ou l'URI \xAB payto:// \xBB ci-dessous pour copier une seule valeur."
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        "Alternativement, si votre banque prend en charge les URI de type PayTO, vous pouvez utiliser ce %1$s lien \xE0 la place"
      ],
      "Operation in progress...": [
        "Op\xE9ration en cours..."
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        "L'op\xE9ration sera automatiquement annul\xE9e apr\xE8s %1$s secondes"
      ],
      "days": [
        "jours"
      ],
      "hours": [
        "heures"
      ],
      "minutes": [
        "minutes"
      ],
      "seconds": [
        "secondes"
      ],
      "Forever": [
        "Pour toujours"
      ],
      "%1$sM": [
        "%1$sM"
      ],
      "%1$sY": [
        "%1$sY"
      ],
      "%1$sd": [
        "%1$sd"
      ],
      "%1$sh": [
        "%1$sh"
      ],
      "%1$smin": [
        "%1$smin"
      ],
      "%1$ssec": [
        "%1$ss"
      ],
      "Change value to never": [
        "Changer la valeur pour jamais"
      ],
      "Change value to empty": [
        "Changer la valeur pour vide"
      ],
      "Image must be smaller than 1 MB": [
        "La taille de l'image doit \xEAtre inf\xE9rieure \xE0 1Mo"
      ],
      "Add": [
        "Ajouter"
      ],
      "Remove": [
        "Effacer"
      ],
      "Country": [
        "Pays"
      ],
      "Building number": [
        "Num\xE9ro de b\xE2timent"
      ],
      "Building name": [
        "Nom du b\xE2timent"
      ],
      "Street": [
        "Rue"
      ],
      "Post code": [
        "Code postal"
      ],
      "Town location": [
        "Localisation de la ville"
      ],
      "District": [
        "District"
      ],
      "Country subdivision": [
        "Subdivision du pays"
      ],
      "Additional address lines": [
        "Informations compl\xE9mentaires"
      ],
      "Orders": [
        "Commandes"
      ],
      "Inventory": [
        "Inventaire"
      ],
      "Categories": [
        "Cat\xE9gories"
      ],
      "Wire transfers": [
        "Virements bancaires"
      ],
      "Templates": [
        "Mod\xE8les"
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        "Configuration"
      ],
      "OTP Devices": [
        "Appareils OTP"
      ],
      "Webhooks": [
        "Webhooks"
      ],
      "Settings": [
        "Configuration"
      ],
      "Password": [
        "Mot de passe"
      ],
      "Access tokens": [
        "Jeton d'acc\xE8s"
      ],
      "Connection": [
        "Connexion"
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        "Instances"
      ],
      "New": [
        "Nouveau"
      ],
      "List": [
        "Liste"
      ],
      "Log out": [
        "Logout"
      ],
      "Username": [
        "Nom d'utilisateur"
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        "Nom de l'instance dans les URL. L'instance 'default' est particuli\xE8re en ce sens qu'elle est utilis\xE9e pour administrer les autres instances."
      ],
      "Business name": [
        "Nom de l'entreprise"
      ],
      "Legal name of the business represented by this instance.": [
        "Nom l\xE9gal de l'entreprise repr\xE9sent\xE9e par cette instance."
      ],
      "Email": [
        "Email"
      ],
      "Contact email": [
        "Email de contact"
      ],
      "Phone number": [
        "Code bancaire."
      ],
      "Contact phone": [
        "Email de contact"
      ],
      "Website URL": [
        "URL du site web"
      ],
      "URL.": [
        "URL."
      ],
      "Logo": [
        "Logo"
      ],
      "Logo image.": [
        "Image du Logo."
      ],
      "Physical location of the merchant.": [
        "Localisation physique du commer\xE7ant."
      ],
      "Jurisdiction": [
        "Juridiction"
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        "Juridiction comp\xE9tente pour les litiges avec le commer\xE7ant."
      ],
      "Pay transaction fee": [
        "Payer les frais de transaction"
      ],
      "Cover the transaction cost or pass it on to the user.": [
        "Prendre en charge le co\xFBt de la transaction ou le r\xE9percuter sur l'utilisateur."
      ],
      "Default payment delay": [
        "D\xE9lai de paiement par d\xE9faut"
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        "Temps par d\xE9faut qu'ont les clients pour payer une commande avant que celle ci n'expire."
      ],
      "Default refund delay": [
        "D\xE9lai de paiement par d\xE9faut"
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        "D\xE9lai par d\xE9faut pour les virements bancaires"
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        "Dur\xE9e maximale pendant laquelle un exchange est autoris\xE9 \xE0 retarder le transfert de fonds au commer\xE7ant, ce qui lui permet de regrouper les petits paiements en virements plus importants et de r\xE9duire les frais de virement."
      ],
      "Default wire transfer rounding interval": [
        "D\xE9lai par d\xE9faut pour les virements bancaires"
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        "Non trouv\xE9"
      ],
      "To second": [
        "secondes"
      ],
      "To minute": [
        "minutes"
      ],
      "To hour": [
        "heures"
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        "Virement effectu\xE9"
      ],
      "Required": [
        "Obligatoire"
      ],
      "Unauthorized": [
        ""
      ],
      "Sending the code failed": [
        "Impossible de contacter le serveur"
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        "Algorithme de v\xE9rification"
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        "Retour"
      ],
      "Verify": [
        "V\xE9rifi\xE9"
      ],
      "Failed to send the verification code.": [
        "Echec de la suppression de l'instance"
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        "Aucune authentification"
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        "Aucun exchange ne pourra accepter le paiement en raison des exigences en mati\xE8re de connaissance du client (proc\xE9dure KYC)."
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        "Non trouv\xE9"
      ],
      "Login required": [
        "Login obligatoire"
      ],
      "instance name": [
        "Identifiant d'instance"
      ],
      "current password": [
        "Avec mot de passe"
      ],
      "Forgot password": [
        "Avec mot de passe"
      ],
      "Create new account": [
        "Compte bancaire cr\xE9dit\xE9"
      ],
      "Invalid": [
        "Invalide"
      ],
      "Doesn't have the pattern of an email": [
        "L'URL n'a pas le format correct"
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        "URL invalide"
      ],
      "Max 7 lines": [
        "7 lignes maximum"
      ],
      "Doesn't match": [
        "Ne correspond pas"
      ],
      "Instance created": [
        "Identifiant d'instance"
      ],
      "Unauthorized.": [
        ""
      ],
      "Conflict.": [
        "Confirmer"
      ],
      "Not found.": [
        "Non trouv\xE9"
      ],
      "New password": [
        "Avec mot de passe"
      ],
      "Next password to be used": [
        "Prochain jeton d'acc\xE8s utilis\xE9"
      ],
      "Repeat password": [
        "Avec mot de passe"
      ],
      "Confirm the same password": [
        "Confirmez le m\xEAme jeton d'acc\xE8s"
      ],
      "Please complete the marked fields and choose authorization method": [
        "Il faut remplir les champs marqu\xE9s et choisir la m\xE9thode d'autorisation"
      ],
      "Confirm operation": [
        "Confirmez l'op\xE9ration"
      ],
      "Check the password.": [
        "Avec mot de passe"
      ],
      "Instance not found.": [
        "Votre instance n'a pas \xE9t\xE9 trouv\xE9e"
      ],
      "Description": [
        "Description"
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        "Dur\xE9e"
      ],
      "Time the access token will be valid.": [
        "Prochain jeton d'acc\xE8s utilis\xE9"
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        "Choisissez un ..."
      ],
      "Read only": [
        ""
      ],
      "All": [
        "Toutes"
      ],
      "Order simple": [
        "Identifiant de commande"
      ],
      "Order Point-of-Sale": [
        ""
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        "Identifiant de commande"
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        "Avec mot de passe"
      ],
      "Please complete the marked fields": [
        "Besoin de compl\xE9ter les champs marqu\xE9s"
      ],
      "Access token created": [
        "Jeton d'acc\xE8s"
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        "Jeton"
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        "Aucune instance 'par d\xE9faut' n'a encore \xE9t\xE9 configur\xE9e."
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        "Cr\xE9er une instance 'par d\xE9faut' pour commencer \xE0 utiliser le backoffice du commer\xE7ant."
      ],
      "Create access token": [
        "Jeton d'acc\xE8s actuel"
      ],
      "Load more devices before the first one": [
        "Ajouter d'autres appareils avant le premier"
      ],
      "Load first page": [
        "Charger la premi\xE8re page"
      ],
      "Created at": [
        "Cr\xE9\xE9 \xE0"
      ],
      "Expires at": [
        ""
      ],
      "Never": [
        "Jamais"
      ],
      "Remove this access token": [
        "R\xE9p\xE9tez le jeton d'acc\xE8s"
      ],
      "Delete": [
        "Effacer"
      ],
      "Load more devices after the last one": [
        "Ajouter d'autres appareils apr\xE8s le dernier"
      ],
      "Load next page": [
        "Charger la page suivante"
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        "Il n'y a pas encore de compte, ajoutez-en un en appuyant sur le signe +"
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        "D\xE9finir un jeton d'acc\xE8s"
      ],
      "Deleting an access token cannot be undone.": [
        "La suppression d'un produit est irr\xE9versible."
      ],
      "This is not a valid Bitcoin address.": [
        "Ceci n'est pas une adresse Bitcoin valide."
      ],
      "This is not a valid Ethereum address.": [
        "Ceci n'est pas une adresse Ethereum valide."
      ],
      "This is not a valid host.": [
        "Ceci n'est pas un serveur valide."
      ],
      "IBANs usually have more than 4 digits.": [
        "Les num\xE9ros IBAN comportent g\xE9n\xE9ralement plus de 4 chiffres"
      ],
      "IBANs usually have fewer than 34 digits.": [
        "Les num\xE9ros IBAN comportent g\xE9n\xE9ralement moins de 34 chiffres"
      ],
      "The IBAN's country code could not be retrieved.": [
        "Le code pays de l'IBAN n'a pas \xE9t\xE9 trouv\xE9"
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        "Le num\xE9ro IBAN n'est pas valide, la somme de contr\xF4le est erron\xE9e"
      ],
      "This account is not allowed.": [
        "Ceci n'est pas un serveur valide."
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        "M\xE9thode de virement\u202F: IBAN"
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        "S\xE9lectionnez le montant le plus petit possible pour le virement bancaire dans votre application bancaire ou sur votre site Web de banque en ligne."
      ],
      "Select a wire method...": [
        ""
      ],
      "Routing": [
        "Code bancaire"
      ],
      "Routing number": [
        "Code bancaire."
      ],
      "Account": [
        "Compte"
      ],
      "Account number": [
        "Num\xE9ro de compte."
      ],
      "Code": [
        "Code"
      ],
      "Business Identifier Code": [
        "Code BIC de la banque."
      ],
      "International Bank Account Number": [
        "Code IBAN."
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        "Interface de paiement unifi\xE9e (Unified Payment Interface)."
      ],
      "Bitcoin protocol": [
        "Protocole Bitcoin."
      ],
      "Ethereum protocol": [
        "Protocole Ethereum."
      ],
      "Interledger protocol": [
        "Protocole Interledger."
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        "Nom de la cat\xE9gorie"
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        "Nom l\xE9gal de la personne titulaire du compte."
      ],
      "Invalid url": [
        "URL invalide"
      ],
      "URL must end with a '/'": [
        "L'URL doit terminer avec un '/'"
      ],
      "URL must not contain params": [
        "L'URL ne doit pas contenir de parametres"
      ],
      "URL must not hash param": [
        "L'URL ne doit pas comporter de param\xE8tre de hachage"
      ],
      "Conflict": [
        "Confirmer"
      ],
      'Server replied with "bad request".': [
        'Le serveur a r\xE9pondu "Mauvaise requ\xEAte".'
      ],
      "Unauthorized, check credentials.": [
        "Non autoris\xE9, v\xE9rifiez les informations d'identification."
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        "Ce endpoint ne semble par ressembler \xE0 une URL Taler Revenue."
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        "La requ\xEAte a abouti mais le serveur n'a pas r\xE9pondu avec la valeur appropri\xE9e pour le champ 'credit_account' et nous ne pouvons donc pas v\xE9rifier que le compte est le m\xEAme."
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        "Compte\u202F:"
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        "Si la banque supporte l'API Taler Revenue, vous pouvez ajouter l'URL du endpoint ci-dessous pour garder les informations sur les revenus synchronis\xE9es."
      ],
      "Endpoint URL": [
        "URL du endpoint"
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        "Endroit d'o\xF9 le commer\xE7ant peut t\xE9l\xE9charger des informations sur les virements entrants sur ce compte"
      ],
      "Auth type": [
        "Type d'authentification"
      ],
      "Choose the authentication type for the account info URL": [
        "Choisir le type d'authentification pour l'URL des info sur le compte"
      ],
      "Without authentication": [
        "Aucune authentification"
      ],
      "With username and password": [
        "Avec mot de passe"
      ],
      "With token": [
        "Avec jeton"
      ],
      "Do not change": [
        "Ne pas changer"
      ],
      "Username to access the account information.": [
        "Nom d'utilisateur pour acc\xE9der aux informations sur le compte."
      ],
      "Password to access the account information.": [
        "Mot de passe pour acc\xE9der aux informations du compte."
      ],
      "Access token to access the account information.": [
        "Jeton d'acc\xE8s pour acc\xE9der aux informations du compte."
      ],
      "Match": [
        "Correspondance"
      ],
      "Check where the information match against the server info.": [
        "V\xE9rifier si les informations correspondent \xE0 celles du serveur."
      ],
      "Compare info from server with account form": [
        "Comparer les informations du serveur avec le formulaire de compte"
      ],
      "Test": [
        "Test"
      ],
      "Server didn't like the request we made.": [
        'Le serveur a r\xE9pondu "Mauvaise requ\xEAte".'
      ],
      "Account details": [
        "D\xE9tails du compte"
      ],
      "Check if the information matches the server info.": [
        "V\xE9rifier si les informations correspondent \xE0 celles du serveur."
      ],
      "The bank account has been deleted.": [
        "Le compte bancaire a \xE9t\xE9 supprim\xE9 avec succ\xE8s."
      ],
      "Bank accounts": [
        "Comptes bancaires"
      ],
      "Add new account": [
        "Ajouter un nouveau compte"
      ],
      "Owner's name": [
        "Nom du propri\xE9taire"
      ],
      "Delete selected accounts from the database": [
        "Supprimer les comptes s\xE9lectionn\xE9s de la base de donn\xE9es"
      ],
      "There are no accounts yet, add more pressing the + sign": [
        "Il n'y a pas encore de compte, ajoutez-en un en appuyant sur le signe +"
      ],
      "You must provide a bank account to receive payments.": [
        "Vous devez associer un compte bancaire pour recevoir des fonds."
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        "Invalide. Veuillez ins\xE9rer uniquement des caract\xE8res et des chiffres"
      ],
      "Name": [
        "Nom"
      ],
      "Category name": [
        "Nom de la cat\xE9gorie"
      ],
      "Category deleted": [
        "Nom de la cat\xE9gorie"
      ],
      "Add new devices": [
        "Ajouter de nouveaux appareils"
      ],
      "ID": [
        "Identifiant"
      ],
      "Total products": [
        "Total des produits"
      ],
      "Delete selected category from the database": [
        "Supprimer la cat\xE9gorie s\xE9lectionn\xE9e de la base de donn\xE9es"
      ],
      "There are no categories yet, add more pressing the + sign": [
        "Il n'y a pas encore de cat\xE9gorie, ajoutez-en en appuyant sur le signe +"
      ],
      "Id:": [
        "Identifiant\u202F:"
      ],
      "Name of the category": [
        "Nom de la cat\xE9gorie"
      ],
      "Products": [
        "Produits"
      ],
      "Image": [
        "Image"
      ],
      "Load more products after the last one": [
        "Ajouter des produits apr\xE8s le dernier"
      ],
      "There are no products in this category.": [
        "G\xE9rer la liste des produits dans la commande."
      ],
      "Account's KYC status": [
        "D\xE9tails du compte"
      ],
      "Exchange": [
        "Exchange"
      ],
      "Status": [
        "URL de l'\xE9tat"
      ],
      "Ok": [
        "OK"
      ],
      "Action required": [
        "Login obligatoire"
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        "Montant obligatoire"
      ],
      "More information required.": [
        "Plus d'informations pour situer le contexte"
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        "Chargement ..."
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        "Status HTTP"
      ],
      "No pending kyc verification!": [
        "Aucune v\xE9rification kyc en cours\u202F!"
      ],
      "Wire account is invalid": [
        ""
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        "Changer la cl\xE9"
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        "Obligatoire"
      ],
      "The payment provider requires more information.": [
        "Le prestataire du service de paiement requi\xE8re une v\xE9rification du compte."
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        "Vous d\xE9finissez le jeton d'acc\xE8s pour la nouvelle instance"
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        "Afficher uniquement les commandes qui ont d\xE9j\xE0 \xE9t\xE9 transf\xE9r\xE9es par le prestataire de services de paiement"
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        "Interface"
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        "Changer la valeur pour une date inconnue"
      ],
      "Enter description or id": [
        "Entrez soit une description soit un identifiant"
      ],
      "No match found for that description or ID.": [
        "aucune correspondance de trouv\xE9e pour cette description ou cet identifiant"
      ],
      "You must enter a valid product identifier.": [
        "Vous devez entrer un identifiant de produit valide."
      ],
      "Quantity must be greater than 0!": [
        "La quantit\xE9 doit \xEAtre sup\xE9rieure \xE0 0\u202F!"
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        "Cette quantit\xE9 d\xE9passe le stock restant. Actuellement, il ne reste que %1$s unit\xE9s non r\xE9serv\xE9es en stock."
      ],
      "Search product": [
        "Recherche produit"
      ],
      "Quantity": [
        "Quantit\xE9"
      ],
      "How many products will be added": [
        "Combien de produits seront ajout\xE9s"
      ],
      "Add from inventory": [
        "Ajouter depuis l'inventaire"
      ],
      "This product has %1$s applicable taxes configured.": [
        "Ce produit a %1$s taxes applicables configur\xE9es."
      ],
      "No taxes configured for this product.": [
        "Aucune taxe n'est configur\xE9e pour ce produit."
      ],
      "Amount": [
        "Montant"
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        "Les taxes peuvent \xEAtre libell\xE9es dans des monnaies diff\xE9rentes de la monnaie principale utilis\xE9e par le commer\xE7ant."
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        "Saisissez la devise et la valeur s\xE9par\xE9es par deux points, par exemple &quot;USD:2.3&quot\u202F;."
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        "Nom l\xE9gal de la taxe, par exemple TVA ou droits d'importation."
      ],
      "Add tax to the tax list": [
        "Ajouter une taxe \xE0 la liste des taxes"
      ],
      "Describe and add a product that is not in the inventory list": [
        "D\xE9crire et ajouter un produit qui ne figure pas dans la liste d'inventaire"
      ],
      "Add custom product": [
        "Ajouter un produit personnalis\xE9"
      ],
      "Complete information of the product": [
        "Informations compl\xE8tes sur le produit"
      ],
      "Must be a number.": [
        "Doit \xEAtre un nombre"
      ],
      "Must be greater than 0.": [
        "Doit \xEAtre sup\xE9rieur \xE0 0"
      ],
      "Photo of the product.": [
        "Photo du produit."
      ],
      "Full product description.": [
        "Description d\xE9taill\xE9e du produit."
      ],
      "Unit name": [
        "Nom de l'unit\xE9"
      ],
      "Name of the product unit.": [
        "Nom de l'unit\xE9 du produit."
      ],
      "Price per unit": [
        "Prix \xE0 l'unit\xE9"
      ],
      "Amount in the current currency.": [
        "Montant dans la monnaie courante."
      ],
      "How many products will be added.": [
        "Combien de produits seront ajout\xE9s."
      ],
      "Taxes": [
        "Taxes"
      ],
      "Total price": [
        "Prix total"
      ],
      "Must be greater than 0": [
        "Doit \xEAtre sup\xE9rieur \xE0 0"
      ],
      "Must have a refund deadline": [
        "Le d\xE9lai de remboursement doit \xEAtre fix\xE9"
      ],
      "Auto refund can't be after refund deadline": [
        "Le remboursement automatique ne peut pas \xEAtre effectu\xE9 apr\xE8s la date limite de remboursement"
      ],
      "Must be in the future": [
        "Doit \xEAtre dans le future"
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        "Ceci n'est pas un serveur valide."
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        "Aucun exchange ne pourra accepter le paiement en raison des exigences en mati\xE8re de connaissance du client (proc\xE9dure KYC)."
      ],
      "Manage products in order": [
        "G\xE9rer les produits dans l'ordre"
      ],
      "%1$s products with a total price of %2$s.": [
        "%1$s produits pour un prix total de %2$s."
      ],
      "Manage list of products in the order.": [
        "G\xE9rer la liste des produits dans la commande."
      ],
      "Remove this product from the order.": [
        "Retirer ce produit de la commande."
      ],
      "Products price sum": [
        "Liste des produits"
      ],
      "Total product price added up": [
        "Prix total des produits additionn\xE9s"
      ],
      "Order price": [
        "Prix par article"
      ],
      "Amount to be paid by the customer": [
        "Montant \xE0 payer par le client"
      ],
      "Final order price": [
        "Prix final de la commande"
      ],
      "Summary": [
        "R\xE9sum\xE9"
      ],
      "Title of the order to be shown to the customer": [
        "Titre de la commande \xE0 afficher au client"
      ],
      "Shipping and fulfillment": [
        "Exp\xE9dition et traitement des commandes"
      ],
      "Delivery date": [
        "Date de livraison"
      ],
      "Deadline for physical delivery assured by the merchant.": [
        "Date limite de livraison physique assur\xE9e par le commer\xE7ant."
      ],
      "Delivery location": [
        "Date de livraison"
      ],
      "Address where the products will be delivered": [
        "Adresse de livraison des produits"
      ],
      "Fulfillment URL": [
        "URL du service de traitement des commandes"
      ],
      "URL to which the user will be redirected after successful payment.": [
        "URL vers laquelle l'utilisateur sera redirig\xE9 apr\xE8s un paiement r\xE9ussi."
      ],
      "Fulfillment message": [
        "URL du service de traitement des commandes"
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        "Options de paiement Taler"
      ],
      "Override default Taler payment settings for this order": [
        "Remplacer les param\xE8tres de paiement par d\xE9faut de Taler pour cette commande"
      ],
      "Payment time": [
        "D\xE9lai de paiement"
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        "D\xE9lai accord\xE9 au client pour payer l'offre avant qu'elle n'expire. Les produits en stock seront r\xE9serv\xE9s jusqu'\xE0 cette date. Le temps commence \xE0 courir apr\xE8s la cr\xE9ation de la commande."
      ],
      "Default": [
        "Par d\xE9faut"
      ],
      "Refund time": [
        "D\xE9lai de remboursement"
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        "D\xE9lai pendant lequel la commande peut \xEAtre rembours\xE9e par le commer\xE7ant. Le temps commence \xE0 la cr\xE9ation de la commande."
      ],
      "Wire transfer time": [
        "D\xE9lai de virement"
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        "Temps n\xE9cessaire \xE0 l'exchange pour effectuer le virement. Le temps commence \xE0 la cr\xE9ation de la commande."
      ],
      "Auto-refund time": [
        "D\xE9lai de remboursement automatique"
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        "D\xE9lai jusqu'auquel le portefeuille v\xE9rifiera automatiquement les remboursements sans intervention de l'utilisateur."
      ],
      "Maximum fee": [
        "Frais maximums"
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        "Frais maximums que le commer\xE7ant est pr\xEAt \xE0 couvrir pour cette commande. Les frais de d\xE9p\xF4t plus \xE9lev\xE9s doivent \xEAtre int\xE9gralement pris en charge par le consommateur."
      ],
      "Create token": [
        "Cr\xE9er un jeton"
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        "Si l'identifiant de la commande est facile \xE0 deviner, le jeton emp\xEAchera les utilisateurs de voler les commandes d'autres personnes."
      ],
      "Minimum age required": [
        "\xC2ge minimum requis"
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        "Toute valeur sup\xE9rieure \xE0 0 limitera le type de pi\xE8ces pouvant \xEAtre utilis\xE9es pour payer ce contrat. Si la valeur est vide, la restriction d'\xE2ge sera d\xE9finie par les produits"
      ],
      "Minimum age defined by the products is %1$s": [
        "L'\xE2ge minimum d\xE9fini par les produits est de %1$s"
      ],
      "No product with age restriction in this order": [
        "Aucun produit avec restriction d'\xE2ge dans cette commande"
      ],
      "Additional information": [
        "Informations compl\xE9mentaires"
      ],
      "Custom information to be included in the contract for this order.": [
        "Informations personnalis\xE9es \xE0 inclure dans le contrat pour cette commande."
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        "Vous devez saisir une valeur en JavaScript Object Notation (JSON)."
      ],
      "remove": [
        "Effacer"
      ],
      "Custom field name": [
        "Nom du champ personnalis\xE9"
      ],
      "Disabled": [
        "D\xE9sactiv\xE9"
      ],
      "No deadline": [
        "Aucune date limite"
      ],
      "Deadline at %1$s": [
        "Date limite \xE0 %1$s"
      ],
      "Select date to show nearby orders": [
        "S\xE9lectionnez la date pour afficher les commandes proches"
      ],
      "Only show unpaid orders": [
        "Afficher uniquement les commandes pay\xE9es"
      ],
      "Only show paid orders": [
        "Afficher uniquement les commandes pay\xE9es"
      ],
      "Paid": [
        "Pay\xE9e"
      ],
      "Only show orders with refunds": [
        "Afficher uniquement les commandes avec remboursement"
      ],
      "Refunded": [
        "Rembours\xE9e"
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        "Virement non effectu\xE9"
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        "Afficher uniquement les commandes qui ont d\xE9j\xE0 \xE9t\xE9 transf\xE9r\xE9es par le prestataire de services de paiement"
      ],
      "Completed": [
        "Transfer\xE9"
      ],
      "Remove all filters": [
        "Effacer tous les filtres"
      ],
      "Clear date filter": [
        "Effacer le filtre de date"
      ],
      "Jump to date (%1$s)": [
        "Aller \xE0 la date (%1$s)"
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        "Aller \xE0 la commande en fonction d'un identifiant de produit"
      ],
      "Order id": [
        "Identifiant de commande"
      ],
      "Instance unknown": [
        "Identifiant d'instance"
      ],
      "Order unknown": [
        "inconnu"
      ],
      "Create order": [
        "Cr\xE9er une commande"
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        "Date"
      ],
      "Refund": [
        "Remboursement"
      ],
      "copy url": [
        "copiez l'URL"
      ],
      "Load more orders after the last one": [
        "Charger d'autres commandes apr\xE8s la derni\xE8re"
      ],
      "No orders have been found matching your query!": [
        "Aucune commande n'a \xE9t\xE9 trouv\xE9e correspondant \xE0 votre requ\xEAte\u202F!"
      ],
      "Duplicated": [
        "Dupliqu\xE9"
      ],
      "This value exceeds the refundable amount.": [
        "Cette valeur d\xE9passe le montant remboursable"
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        "Des exigences en mati\xE8re de connaissance du client (KYC) sont en suspens."
      ],
      "refund": [
        "Remboursement"
      ],
      "Reason": [
        "R\xE9f\xE9rence"
      ],
      "Amount to be refunded": [
        "Montant \xE0 rembourser"
      ],
      "Max refundable:": [
        "Remboursement maximum\u202F:"
      ],
      "Choose one...": [
        "Choisissez un ..."
      ],
      "Requested by the customer": [
        "A la demande du client"
      ],
      "Other": [
        "Autre"
      ],
      "Why this order is being refunded": [
        "Pourquoi cette commande est-elle rembours\xE9e"
      ],
      "More information to give context": [
        "Plus d'informations pour situer le contexte"
      ],
      "now": [
        "non"
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        "Afficher uniquement les commandes qui ont d\xE9j\xE0 \xE9t\xE9 transf\xE9r\xE9es par le prestataire de services de paiement"
      ],
      "This wire transfer has been notified by the payment service provider.": [
        "Afficher uniquement les commandes qui ont d\xE9j\xE0 \xE9t\xE9 transf\xE9r\xE9es par le prestataire de services de paiement"
      ],
      "This refund is waiting to be claimed by the customer.": [
        "Montant \xE0 payer par le client"
      ],
      "This refund has been claimed by the customer.": [
        "Montant \xE0 payer par le client"
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        "Conditions du contrat"
      ],
      "Human-readable description of the whole purchase": [
        "Description en langage courant de l'ensemble de la transaction"
      ],
      "Total price for the transaction": [
        "Prix total de la transaction"
      ],
      "URL for this purchase": [
        "URL de cet achat"
      ],
      "Max fee": [
        "Frais maximums"
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        "Montant total maximum des frais de d\xE9p\xF4t accept\xE9s par le commer\xE7ant pour ce contrat"
      ],
      "Time when this contract was generated": [
        "Heure \xE0 laquelle ce contrat a \xE9t\xE9 g\xE9n\xE9r\xE9"
      ],
      "Refund deadline": [
        "Date limite de remboursement"
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        "Pass\xE9 ce d\xE9lai, aucun remboursement ne sera accept\xE9"
      ],
      "Payment deadline": [
        "Date limite de paiement"
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        "Pass\xE9 ce d\xE9lai, le commer\xE7ant n'acceptera plus de paiements pour le contrat"
      ],
      "Wire transfer deadline": [
        "Date limite pour les virements"
      ],
      "Transfer deadline for the exchange": [
        "Date limite de transfert pour l'exchange"
      ],
      "Time indicating when the order should be delivered": [
        "Date \xE0 laquelle la commande doit \xEAtre livr\xE9e"
      ],
      "Where the order will be delivered": [
        "Lieu de livraison de la commande"
      ],
      "Auto-refund delay": [
        "D\xE9lai de remboursement automatique"
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        "Combien de temps le portefeuille doit-il essayer d'obtenir un remboursement automatique de l'achat"
      ],
      "Extra info": [
        "Infos suppl\xE9mentaires"
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        "Donn\xE9es suppl\xE9mentaires qui ne sont interpr\xE9t\xE9es que par le frontend du commer\xE7ant"
      ],
      "order created": [
        "Statut de la commande"
      ],
      "pay deadline": [
        "Date limite de paiement"
      ],
      "refund deadline": [
        "Date limite de remboursement"
      ],
      "delivery": [
        "Date de livraison"
      ],
      "Order": [
        "Commande"
      ],
      "Claimed": [
        "Id\xE9es pour les projets r\xE9clam\xE9es"
      ],
      "Claimed at": [
        "R\xE9clam\xE9e \xE0"
      ],
      "Timeline": [
        "Calendrier"
      ],
      "Payment details": [
        "Modalit\xE9s de paiement"
      ],
      "Order status": [
        "Statut de la commande"
      ],
      "Unpaid": [
        "Non pay\xE9"
      ],
      "Product list": [
        "Liste des produits"
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        ""
      ],
      "refund created: %1$s": [
        "Cr\xE9ation r\xE9ussie du remboursement"
      ],
      "refund created: %1$s: %2$s": [
        "Cr\xE9ation r\xE9ussie du remboursement"
      ],
      "refund taken: %1$s": [
        "Remboursement pris"
      ],
      "refund taken: %1$s: %2$s": [
        "Remboursement pris"
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        "Aucune date limite"
      ],
      "Wired": [
        "Virement effectu\xE9"
      ],
      "Refund order": [
        "Rembourser la commande"
      ],
      "Not refundable": [
        "Non remboursable"
      ],
      "Next event in": [
        "Nouvel \xE9v\xE8nement dans"
      ],
      "The order was wired.": [
        "La demande a \xE9t\xE9 annul\xE9e."
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        "Montant rembours\xE9"
      ],
      "Refund taken": [
        "Remboursement pris"
      ],
      "Status URL": [
        "URL de l'\xE9tat"
      ],
      "Refund URI": [
        "URI de remboursement"
      ],
      "Pay at": [
        "URL de paiement"
      ],
      "Order status URL": [
        "URL de l'\xE9tat de la commande"
      ],
      "Payment URI": [
        "URI de payment"
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        "Statut de la commande inconnu. Il s'agit d'une erreur, veuillez contacter l'administrateur."
      ],
      "Invalid. Please enter letters and numbers only.": [
        "Invalide. Veuillez n'ins\xE9rer que des caract\xE8res et des chiffres"
      ],
      "Just letters and numbers from 2 to 7": [
        "Juste des lettres et des chiffres de 2 \xE0 7"
      ],
      "The size of the key must be 32 characters": [
        "La cl\xE9 doit comporter exactement 32 caract\xE8res"
      ],
      "Internal ID on the system": [
        "Identifiant interne au syst\xE8me"
      ],
      "Useful to identify the device physically": [
        "Utile pour identifier physiquement l'appareil"
      ],
      "Verification algorithm": [
        "Algorithme de v\xE9rification"
      ],
      "Algorithm used to verify transactions in offline mode": [
        "Algorithme utilis\xE9 pour v\xE9rifier la transaction en mode hors ligne"
      ],
      "Device key": [
        "Cl\xE9 de l'appareil"
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        "Veillez \xE0 ce que votre mot de passe soit tr\xE8s difficile \xE0 deviner ou utilisez le g\xE9n\xE9rateur al\xE9atoire"
      ],
      "Your device needs to match exactly the same value": [
        "Votre appareil doit afficher exactement la m\xEAme valeur"
      ],
      "Generate random secret key": [
        "G\xE9n\xE9rer une cl\xE9 secr\xE8te al\xE9atoire"
      ],
      "Random": [
        "Al\xE9atoire"
      ],
      "Create another": [
        "Cr\xE9er une commande"
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        "Vous pouvez scanner le code QR suivant avec votre appareil ou enregistrer la cl\xE9 avant de continuer."
      ],
      "OTP devices": [
        "Appareils OTP"
      ],
      "Delete selected devices from the database": [
        "Supprimer les appareils s\xE9lectionn\xE9s de la base de donn\xE9es"
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        "Il n'y a pas encore d'appareil, ajoutez-en un en appuyant sur le signe +"
      ],
      "Template id is unknown": [
        "Identifiant du mod\xE8le inconnu"
      ],
      "The provided information is inconsistent with the current state of the template": [
        "Les informations fournies ne correspondent pas \xE0 l'\xE9tat actuel du mod\xE8le"
      ],
      "Device:": [
        "Appareil\u202F:"
      ],
      "Algorithm to use to verify transactions in offline mode": [
        "Algorithme \xE0 utiliser pour v\xE9rifier les transactions en mode hors ligne"
      ],
      "Not modified": [
        "Pas de modification"
      ],
      "Be sure to be very hard to guess or use the random generator": [
        "Veillez \xE0 ce qu'il soit tr\xE8s difficile \xE0 deviner ou utilisez le g\xE9n\xE9rateur al\xE9atoire"
      ],
      "Your device need to have exactly the same value": [
        "Votre appareil doit avoir exactement la m\xEAme valeur"
      ],
      "Change key": [
        "Changer la cl\xE9"
      ],
      "Is not the same": [
        "N'est pas le m\xEAme"
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        "Vous mettez \xE0 jour le jeton d'acc\xE8s de l'instance dont l'identifiant est %1$s"
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        "Confirmer le changement"
      ],
      "Testing password change": [
        "Mot de passe"
      ],
      "Password changed": [
        "Mot de passe"
      ],
      "The current password is wrong.": [
        "Avec mot de passe"
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        "Identifiant du produit"
      ],
      "Add element to the list": [
        "Ajouter l'\xE9l\xE9ment \xE0 la liste"
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        "Cliquez ici pour configurer le stock du produit, laissez-le tel quel et le backend ne contr\xF4lera pas le stock."
      ],
      "Manage stock": [
        "Gestion du stock"
      ],
      "This product has been configured without stock control": [
        "Ce produit a \xE9t\xE9 configur\xE9 sans contr\xF4le du stock"
      ],
      "Infinite": [
        "Infini"
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        "Le retrait ne peut pas d\xE9passer le stock actuel plus les quantit\xE9es entrantes (maximum %1$s)"
      ],
      "Incoming": [
        "Arriv\xE9es"
      ],
      "Lost": [
        "Perte"
      ],
      "Current": [
        "Courant"
      ],
      "Remove stock control for this product": [
        "Supprimer le contr\xF4le des stocks pour ce produit"
      ],
      "without stock": [
        "sans stock"
      ],
      "Next restock": [
        "Prochain r\xE9approvisionnement"
      ],
      "Warehouse address": [
        "Adresse de l'entrep\xF4t"
      ],
      "Invalid amount": [
        "Montant invalide"
      ],
      "Product identification to use in URLs (for internal use only).": [
        "Identification du produit \xE0 utiliser dans les URL (\xE0 usage interne uniquement)."
      ],
      "Illustration of the product for customers.": [
        "Illustration du produit pour les clients."
      ],
      "Product name.": [
        "Identifiant du produit"
      ],
      "Product description for customers.": [
        "Description du produit pour les clients."
      ],
      "Age restriction": [
        "Restriction d'\xE2ge"
      ],
      "Is this product restricted for customer below certain age?": [
        "Ce produit est-il interdit aux personnes en dessous d'un certain \xE2ge\u202F?"
      ],
      "Minimum age of the customer": [
        "\xC2ge minimum du client"
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        "Unit\xE9 d\xE9crivant la quantit\xE9 de produits vendus (par exemple, 2 kilogrammes, 5 litres, 3 articles, 5 m\xE8tres) aux clients."
      ],
      "Example: kg, items or liters": [
        "Exemple\u202F: kg, pi\xE8ces ou litres"
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        "Prix de vente TTC des produits pour les clients, correspondant aux unit\xE9s de produits ci-dessus."
      ],
      "Stock": [
        "Stock"
      ],
      "Inventory for products with finite supply (for internal use only).": [
        "Inventaire des produits dont l'offre est limit\xE9e (\xE0 usage interne uniquement)."
      ],
      "Taxes included in the product price, exposed to customers.": [
        "Taxes incluses dans le prix du produit, communiqu\xE9es aux clients."
      ],
      "Search by category description or id": [
        "Recherche par description de cat\xE9gorie ou par identifiant"
      ],
      "Categories where this product will be listed on.": [
        "Cat\xE9gories dans lesquelles ce produit sera rang\xE9."
      ],
      "Add product to inventory": [
        "Ajouter le produit \xE0 l'inventaire"
      ],
      "Product not found.": [
        "Identifiant du produit"
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        "Ventes"
      ],
      "Sold": [
        "Vendu"
      ],
      "Free": [
        "Libre"
      ],
      "Go to product update page": [
        "Aller \xE0 la page de modification du produit"
      ],
      "Update": [
        "Modification"
      ],
      "Remove this product from the database": [
        "Effacer ce produit de la base de donn\xE9es"
      ],
      "Update the product with new price": [
        "Modifier le prix du produit"
      ],
      "Update product with new price": [
        "Modifier le prix du produit"
      ],
      "Confirm update": [
        "Confirmer la modification"
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        "La perte ne peut pas \xEAtre sup\xE9rieure au courant + entr\xE9es (max %1$s)"
      ],
      "Add more elements to the inventory": [
        "Ajouter d'autres \xE9l\xE9ments \xE0 l'inventaire"
      ],
      "Report elements lost in the inventory": [
        "Signaler les \xE9l\xE9ments perdus dans l'inventaire"
      ],
      "New price for the product": [
        "Nouveau prix pour le produit"
      ],
      "There are values with errors.": [
        "Certaines valeurs ont des erreurs"
      ],
      "Update product with new stock and price": [
        "Changer le stock et le prix du produit"
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        "Il n'y a pas encore de produits, ajoutez-en en appuyant sur le signe +"
      ],
      "Product (ID: %1$s) has been deleted": [
        "Le produit (identifiant\u202F: %1$s) a \xE9t\xE9 modifi\xE9"
      ],
      "Jump to product with the given product ID": [
        "Aller directement au produit avec son identifiant"
      ],
      "Product id": [
        "Identifiant du produit"
      ],
      "Delete product": [
        "Effac\xE9es"
      ],
      'Delete the product "%1$s"': [
        "Impossible d'effacer le produit"
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        "Si vous effacez le produit nomm\xE9 %1$s (identifiant\u202F: %2$s), le stock et toutes les informations li\xE9es seront perdus"
      ],
      "Deleting a product cannot be undone.": [
        "La suppression d'un produit est irr\xE9versible."
      ],
      "Product id:": [
        "Identifiant du produit\u202F:"
      ],
      "Must be greater that 0": [
        "Doit \xEAtre sup\xE9rieur \xE0 0"
      ],
      "Too short": [
        "Trop court"
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        "Identifiant"
      ],
      "Name of the template in URLs.": [
        "Nom du mod\xE8le dans les URLs."
      ],
      "Template name": [
        "Mod\xE8les"
      ],
      "Describe what this template stands for": [
        "D\xE9crire ce que ce mod\xE8le repr\xE9sente"
      ],
      "Order summary": [
        "R\xE9sum\xE9 de la commande"
      ],
      "If specified here, this template will create orders with the same summary": [
        "S'il est utilis\xE9, ce mod\xE8le cr\xE9era des commandes avec le m\xEAme r\xE9sum\xE9"
      ],
      "Summary is editable": [
        "Le r\xE9sum\xE9 est \xE9ditable"
      ],
      "Allow the user to change the summary.": [
        "Autoriser l'utilisateur \xE0 modifier le r\xE9sum\xE9."
      ],
      "If specified here, this template will create orders with the same price": [
        "S'il est utilis\xE9, ce mod\xE8le cr\xE9era des commandes avec le m\xEAme prix"
      ],
      "Amount is editable": [
        "Le montant est modifiable"
      ],
      "Allow the user to select the amount to pay.": [
        "Permettre \xE0 l'utilisateur de s\xE9lectionner le montant \xE0 payer."
      ],
      "Currency is editable": [
        "La devise est modifiable"
      ],
      "Allow the user to change currency.": [
        "Permettre \xE0 l'utilisateur de changer de devise."
      ],
      "Supported currencies": [
        "Devises prises en charge"
      ],
      "Supported currencies: %1$s": [
        "Devises prises en charge\u202F: %1$s"
      ],
      "Minimum age": [
        "Age minimum"
      ],
      "Is this contract restricted to some age?": [
        "Ce contrat doit-il contenir une limite d'\xE2ge\u202F?"
      ],
      "Payment timeout": [
        "D\xE9lai de paiement"
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        "De combien de temps le client dispose pour le paiement apr\xE8s la cr\xE9ation de la commande."
      ],
      "OTP device": [
        "Appareil OTP"
      ],
      "Use to verify transactions in offline mode.": [
        "A utiliser pour v\xE9rifier la transaction en mode hors ligne."
      ],
      "No OTP device.": [
        "Aucun appareil OTP."
      ],
      "Add one first": [
        "Ajouter un premier \xE9l\xE9ment"
      ],
      "No device": [
        "Aucun appareil"
      ],
      "Add new templates": [
        "Ajouter de nouveaux mod\xE8les"
      ],
      "Load more templates before the first one": [
        "Ajouter de nouveaux mod\xE8les avant le premier"
      ],
      "Delete selected templates from the database": [
        "Effacer les mod\xE8les s\xE9lectionn\xE9s de la base de donn\xE9es"
      ],
      "Use template to create new order": [
        "Utiliser le mod\xE8le pour cr\xE9er une nouvelle commande"
      ],
      "Generate a QR code for the template.": [
        "Cr\xE9er un code QR pour le mod\xE8le."
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        "Ajouter des mod\xE8les apr\xE8s le dernier"
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        "Il n'y a pas encore de mod\xE8les, ajoutez-en en appuyant sur le signe +"
      ],
      "Jump to template with the given template ID": [
        "Aller directement au mod\xE8le avec cet identifiant"
      ],
      "Template identification": [
        "Identification du mod\xE8le"
      ],
      "Delete template": [
        "Suppression d'un mod\xE8le"
      ],
      'Delete the template "%1$s"': [
        "Suppression d'un mod\xE8le"
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        "Si vous effacez le mod\xE8le %1$s (identifiant\u202F: %2$s) vous pourriez perdre des informations"
      ],
      "Deleting a template cannot be undone.": [
        "La suppression d'un produit est irr\xE9versible."
      ],
      "Print": [
        "Imprimer"
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        "S'il est utilis\xE9, ce mod\xE8le cr\xE9era des commandes avec le m\xEAme r\xE9sum\xE9."
      ],
      "If specified, this template will create orders with the same price": [
        "S'il est utilis\xE9, ce mod\xE8le cr\xE9era des commandes avec le m\xEAme prix"
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        "De combien de temps le client dispose pour le paiement apr\xE8s la cr\xE9ation de la commande."
      ],
      "An amount is required": [
        "Montant obligatoire"
      ],
      "An order summary is required": [
        "Un r\xE9sum\xE9 de la commande est obligatoire"
      ],
      'No more stock for product with ID "%1$s".': [
        `Il n'y a plus de stock pour le produit dont l'identifiant est "%1$s".`
      ],
      "New order from template": [
        "Nouvelle commande \xE0 partir d'un mod\xE8le"
      ],
      "Amount of the order": [
        "Montant de la commande"
      ],
      "Slug": [
        "Type"
      ],
      "Token family slug to use in URLs (for internal use only)": [
        "Type de famille de jetons \xE0 utiliser dans les URL (\xE0 usage interne uniquement)"
      ],
      "Kind": [
        "Type"
      ],
      "Token family kind": [
        "Type de famille de jeton"
      ],
      "User-readable token family name": [
        "Nom de famille du jeton compr\xE9hensible par l'utilisateur"
      ],
      "Token family description for customers": [
        "Description de la famille de jetons pour les clients"
      ],
      "Valid After": [
        "Valide \xE0 partir de"
      ],
      "Token family can issue tokens after this date": [
        "La famille de jetons peut \xE9mettre des jetons apr\xE8s cette date"
      ],
      "Valid Before": [
        "Valide jusqu'\xE0"
      ],
      "Token family can issue tokens until this date": [
        "La famille de jetons peut \xE9mettre des jetons jusqu'\xE0 cette date"
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        "Dur\xE9e de validit\xE9 d'un jeton \xE9mis"
      ],
      "Token Families": [
        "Familles de jetons"
      ],
      "Add token family": [
        "Ajouter une famille de jetons"
      ],
      "Go to token family update page": [
        "Aller \xE0 la page de modification des familles de jetons"
      ],
      "Remove this token family from the database": [
        "Effacer cette famille de jetons de la base de donn\xE9es"
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        "Il n'y a pas encore de familles de jetons, ajoutez la premi\xE8re en appuyant sur le signe +."
      ],
      "Token family has been deleted.": [
        'Famille de jetons "%1$s" (type\u202F: %2$s) a \xE9t\xE9 effac\xE9e'
      ],
      "Delete token family": [
        "\xC9chec de la suppression de la famille de jetons"
      ],
      'Delete the token family "%1$s"': [
        "Suppression de la famille de jetons %1$s ."
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        "Si vous supprimez la famille de jetons %1$s ( Cat\xE9gorie\u202F: %2$s), tous les jetons \xE9mis deviendront invalides."
      ],
      "Deleting a token family %1$s .": [
        "Suppression de la famille de jetons %1$s ."
      ],
      "can't be undone": [
        "ne peut \xEAtre annul\xE9"
      ],
      "Token Family: %1$s": [
        "Famille de jetons\u202F: %1$s"
      ],
      "Incoming wire transfers": [
        "Virements bancaires"
      ],
      "Load more wire transfers preceding the first one": [
        "Charger des virements avant le premier"
      ],
      "Expected credit": [
        "Ex\xE9cut\xE9 le"
      ],
      "Confirmed": [
        "Confirm\xE9"
      ],
      "Validated": [
        "Valide \xE0 partir de"
      ],
      "Executed on": [
        "Ex\xE9cut\xE9 le"
      ],
      "yes": [
        "oui"
      ],
      "no": [
        "non"
      ],
      "never": [
        "jamais"
      ],
      "unknown": [
        "inconnu"
      ],
      "Load more transfers after the last one": [
        "Ajouter des virements apr\xE8s le dernier"
      ],
      "Verified wire transfers": [
        "Virements bancaires"
      ],
      "Verified": [
        "V\xE9rifi\xE9"
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        "Il n'y a pas encore de virements, ajoutez-en en appuyant sur le signe +"
      ],
      "All accounts": [
        "Tous les comptes"
      ],
      "Filter by account address": [
        "Filtre sur l'adresse du compte"
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        "Afficher uniquement les commandes qui ont d\xE9j\xE0 \xE9t\xE9 transf\xE9r\xE9es par le prestataire de services de paiement"
      ],
      "Only show wire transfers confirmed by the merchant": [
        "Afficher uniquement les virements bancaires confirm\xE9s par le commer\xE7ant"
      ],
      "Wire transfer already confirmed.": [
        "Date limite pour les virements"
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        "Confirmez le m\xEAme jeton d'acc\xE8s"
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        "Calendrier"
      ],
      "Transfer ID": [
        "Virements"
      ],
      "It's not the same.": [
        "N'est pas le m\xEAme"
      ],
      'You are deleting the instance with ID "%1$s"': [
        `Vous modifiez le jeton d'acc\xE8s de l'instance dont l'identifiant est "%1$s"`
      ],
      "Instance": [
        "Instances"
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        "Purger"
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        "Le d\xE9lai de paiement ne peut \xEAtre sup\xE9rieur au d\xE9lai de virement"
      ],
      "Instance id": [
        "Identifiant d'instance"
      ],
      "Delete this instance": [
        "Suppression d'une instance %1$s ."
      ],
      "Must be one of '%1$s'": [
        "Doit \xEAtre l'un des \xE9l\xE9ments suivants '%1$s'"
      ],
      "URL is invalid": [
        "La chaine JSON est invalide"
      ],
      "Webhook ID to use": [
        "Identifiant du Webhook \xE0 utiliser"
      ],
      "Event": [
        "Ev\xE8nement"
      ],
      "Payment": [
        "URI de payment"
      ],
      "Order settled": [
        "Statut de la commande"
      ],
      "Category added": [
        "Nom de la cat\xE9gorie"
      ],
      "Category updated": [
        "Nom de la cat\xE9gorie"
      ],
      "Inventory added": [
        "Inventaire"
      ],
      "Inventory updated": [
        "Inventaire"
      ],
      "Inventory deleted": [
        "Inventaire"
      ],
      "The event of the webhook: why the webhook is used": [
        "L'\xE9v\xE9nement du webhook\u202F: pourquoi le webhook est-il utilis\xE9"
      ],
      "Method": [
        "M\xE9thode"
      ],
      "GET": [
        "GET"
      ],
      "POST": [
        "POST"
      ],
      "PUT": [
        "PUT"
      ],
      "PATCH": [
        "PATCH"
      ],
      "HEAD": [
        "HEAD"
      ],
      "Method used by the webhook": [
        "M\xE9thode utilis\xE9e par le webhook"
      ],
      "URL": [
        "URL"
      ],
      "URL of the webhook where the customer will be redirected": [
        "URL du webhook o\xF9 le client sera redirig\xE9"
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        "Le texte ci-dessous prend en charge le moteur de mod\xE8le %1$s. Toute cha\xEEne comprise entre %2$s et %3$s sera remplac\xE9e par la valeur de la variable correspondante."
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        "Par exemple, %1$s sera remplac\xE9 par le prix de la commande"
      ],
      "The short list of variables are:": [
        "La liste restreinte des variables est la suivante\u202F:"
      ],
      "order's description": [
        "description de la commande"
      ],
      "order's price": [
        "prix de la commande"
      ],
      "order's unique identification": [
        "identification unique de la commande"
      ],
      "the amount that was being refunded": [
        "le montant rembours\xE9"
      ],
      "the reason entered by the merchant staff for granting the refund": [
        "la raison invoqu\xE9e par le personnel du commer\xE7ant pour accorder le remboursement"
      ],
      "time of the refund in nanoseconds since 1970": [
        "instant du remboursement en nanosecondes depuis 1970"
      ],
      "Http header": [
        "Ent\xEAte"
      ],
      "Header template of the webhook": [
        "Mod\xE8le d'ent\xEAte pour le webhook"
      ],
      "Http body": [
        "body HTTP"
      ],
      "Body template used by the webhook.": [
        "Mod\xE8le de body pour le webhook"
      ],
      "Webhook deleted successfully": [
        "Suppression r\xE9ussie du webhook"
      ],
      "Add new webhooks": [
        "Ajouter de nouveaux webhooks"
      ],
      "Load more webhooks before the first one": [
        "Ajouter des webhooks avant le premier"
      ],
      "Event type": [
        "Type d'\xE9v\xE8nement"
      ],
      "Delete selected webhook from the database": [
        "Supprimer le webhook s\xE9lectionn\xE9 de la base de donn\xE9es"
      ],
      "Load more webhooks after the last one": [
        "Ajouter des webhooks apr\xE8s le dernier"
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        "Il n'y a pas encore de webhook, ajoutez-en en appuyant sur le signe +"
      ],
      "Webhook updated": [
        "Webhook modifi\xE9"
      ],
      "Header": [
        "Ent\xEAte"
      ],
      "Body": [
        "Corps"
      ],
      "Body template used by the webhook": [
        "Mod\xE8le de body pour le webhook"
      ],
      "There is another instance with this username.": [
        "Affiche plus d'options dans le formulaire de configuration de l'instance"
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        "Email de contact"
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        "Cr\xE9\xE9 \xE0"
      ],
      "Password reset": [
        "Mot de passe"
      ],
      'Resetting access to the instance "%1$s"': [
        "Vous d\xE9finissez le jeton d'acc\xE8s pour la nouvelle instance"
      ],
      "Reset": [
        ""
      ],
      "Language": [
        "Langue"
      ],
      "Date format": [
        "Format de la date"
      ],
      "How the date is going to be displayed": [
        "Comment la date sera affich\xE9e"
      ],
      "Merchant type": [
        "Type d'\xE9v\xE8nement"
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        "Mode d\xE9velopeur"
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        "Commandes"
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        "Statut de la commande"
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        "Statut de la commande"
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        "prix de la commande"
      ],
      "hour": [
        "heures"
      ],
      "day": [
        "jours"
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        "Effacer tous les filtres"
      ],
      "Time range": [
        "Calendrier"
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        "Courant"
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        "D\xE9lai de paiement"
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        "Vous devez associer un compte bancaire pour recevoir des fonds."
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        "Sans cela, le backend du commer\xE7ant refusera de cr\xE9er de nouvelles commandes."
      ],
      "Hide for today": [
        "Cacher pour aujourd'hui"
      ],
      "KYC verification needed": [
        "V\xE9rification de l'identit\xE9 du client (KYC) n\xE9cessaire"
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        "Certains transferts sont suspendus jusqu'\xE0 ce que la proc\xE9dure de v\xE9rification de l'identit\xE9 du client (KYC) soit achev\xE9e. Pour plus d'informations, consultez la section KYC dans le panneau de gauche"
      ],
      "The request reached a timeout, check your connection.": [
        "La demande a d\xE9pass\xE9 le temps de traitement maximum, v\xE9rifiez votre connexion."
      ],
      "The request was cancelled.": [
        "La demande a \xE9t\xE9 annul\xE9e."
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        "Un grand nombre de demandes ont \xE9t\xE9 adress\xE9es au m\xEAme serveur et cette action a \xE9t\xE9 ralentie."
      ],
      "The server's response was malformed.": [
        "La r\xE9ponse de la demande est malform\xE9e."
      ],
      "Could not complete the request due to a network problem.": [
        "La demande n'a pas pu \xEAtre trait\xE9e en raison d'un probl\xE8me de r\xE9seau."
      ],
      "Unexpected request error.": [
        "Erreur inattendue sur la requ\xEAte."
      ],
      "Unexpected error.": [
        "Erreur inattendue."
      ],
      "Add new instance": [
        "Ajouter une nouvelle instance"
      ],
      "Edit": [
        "Editer"
      ],
      "Change password": [
        "Avec mot de passe"
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        "Il n'y a pas encore d'instance, ajoutez-en une en appuyant sur le signe +"
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        `L'instance "%1$s" (identifiant\u202F: %2$s) a \xE9t\xE9 effac\xE9e`
      ],
      "Only show active instances": [
        "N'afficher que les instances actives"
      ],
      "Active": [
        "Actives"
      ],
      "Only show deleted instances": [
        "N'afficher que les instances effac\xE9es"
      ],
      "Deleted": [
        "Effac\xE9es"
      ],
      "Show all instances": [
        "Afficher toutes les instances"
      ],
      "Delete instance": [
        "Suppression d'une instance %1$s ."
      ],
      'Delete the instance "%1$s"': [
        "Suppression d'une instance %1$s ."
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        "Si vous supprimez l'instance nomm\xE9e %1$s (ID\u202F: %2$s), le commer\xE7ant ne sera plus en mesure de traiter les commandes ou les remboursements"
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        "Cette action supprime la cl\xE9 priv\xE9e de l'instance, mais pr\xE9serve toutes les donn\xE9es de transaction. Vous pouvez toujours acc\xE9der \xE0 ces donn\xE9es apr\xE8s avoir supprim\xE9 l'instance."
      ],
      "Deleting an instance %1$s": [
        "Suppression d'une instance %1$s ."
      ],
      "This cannot be undone!": [
        "ne peut \xEAtre annul\xE9"
      ],
      "Purge the instance": [
        "Suppression d'une instance %1$s ."
      ],
      'Purge the instance "%1$s"': [
        "Purge d'une instance %1$s ."
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        "Si vous purgez l'instance nomm\xE9e %1$s (ID\u202F: %2$s), vous supprimerez \xE9galement toutes ses donn\xE9es de transaction."
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        "L'instance dispara\xEEtra de votre liste et vous ne pourrez plus acc\xE9der \xE0 ses donn\xE9es."
      ],
      "Purging an instance %1$s": [
        "Purge d'une instance %1$s ."
      ],
      "checking compatibility with server...": [
        "v\xE9rification de la compatibilit\xE9 avec le serveur..."
      ],
      "Contacting the server failed": [
        "Impossible de contacter le serveur"
      ],
      "The server version is not supported": [
        "La version du serveur n'est pas support\xE9e"
      ],
      'Supported version "%1$s", server version "%2$s".': [
        'Version support\xE9e "%1$s", version du serveur "%2$s".'
      ],
      "Deleting": [
        "Suppression"
      ],
      "Changing": [
        "Modification"
      ],
      "Manage password": [
        "Gestion du stock"
      ],
      "Business Name": [
        "Nom de l'entreprise"
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        "Configuration"
      ],
      "title%1$s: New bank account": [
        "Compte bancaire"
      ],
      "title%1$s: Bank accounts": [
        "Comptes bancaires"
      ],
      "title%1$s: Update bank Account": [
        "Valider le compte bancaire\u202F: %1$s"
      ],
      "title%1$s: Orders": [
        "Commandes"
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        "Inventaire"
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        "Modification du produit impossible"
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        "Impossible de modifier la cat\xE9gorie"
      ],
      "title%1$s: Wire transfers": [
        "Virements bancaires"
      ],
      "title%1$s: Webhooks": [
        "Webhooks"
      ],
      "title%1$s: New webhook": [
        "Ajouter de nouveaux webhooks"
      ],
      "title%1$s: Update webhook": [
        "Impossible de modifier le webhook"
      ],
      "title%1$s: OTP devices": [
        "Appareils OTP"
      ],
      "title%1$s: New OTP device": [
        "Aucun appareil OTP."
      ],
      "title%1$s: Update OTP device": [
        "Appareil OTP"
      ],
      "title%1$s: New template": [
        "Utiliser le mod\xE8le"
      ],
      "title%1$s: Update template": [
        "Utiliser le mod\xE8le"
      ],
      "title%1$s: Templates": [
        "Mod\xE8les"
      ],
      "title%1$s: Use template": [
        "Utiliser le mod\xE8le"
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        "Familles de jetons"
      ],
      "title%1$s: New token family": [
        "Ajouter une famille de jetons"
      ],
      "title%1$s: Update token family": [
        "Impossible de mettre \xE0 jour la famille de jetons"
      ],
      "title%1$s: Access tokens": [
        "Jeton d'acc\xE8s"
      ],
      "title%1$s: New access token": [
        "Nouveau jeton d'acc\xE8s"
      ],
      "titleNew instance": [
        "Ajouter une nouvelle instance"
      ],
      "titleInstances": [
        "Instances"
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=2; plural=(n!=1);",
  "lang": "fr",
  "completeness": 79
};
strings["es"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=2; plural=n != 1;",
        "lang": "es"
      },
      "Cancel": [
        "Cancelar"
      ],
      "%1$s": [
        "%1$s"
      ],
      "Close": [
        "Cerrar"
      ],
      "Continue": [
        "Continuar"
      ],
      "Clear": [
        "Limpiar"
      ],
      "Confirm": [
        "Confirmar"
      ],
      "Correct form": [
        "Formulario correcto"
      ],
      "Comparing account details": [
        "Comparando detalle de cuentas"
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        "Verificando contra la URL de informaci\xF3n de cuenta exitoso pero la informaci\xF3n de la cuenta reportada es diferente a los detalles de cuenta en el formulario."
      ],
      "Field": [
        "Campo"
      ],
      "In the form": [
        "En el formulario"
      ],
      "Reported": [
        "Reportado"
      ],
      "Type": [
        "Tipo"
      ],
      "IBAN": [
        "IBAN"
      ],
      "Address": [
        "Direcci\xF3n"
      ],
      "Host": [
        "Host"
      ],
      "Account ID": [
        "Cuenta"
      ],
      "The account owner's legal name": [
        "Nombre del titular de la cuenta bancaria."
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        "Ciudad"
      ],
      "Bank host": [
        "Host del banco"
      ],
      "Bank account": [
        "Cuenta bancaria"
      ],
      "BIC": [
        "BIC"
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        "Validar cuenta bancaria: %1$s"
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        "Paso 1:"
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        "Copia y pega este IBAN y el nombre en el campo destinatario en to aplicaion bancaria o sitio web"
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        "Paso 2:"
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        "Copia este codigo y pegalo en el campo asunto de to aplicacion bancaria o sitio de banco"
      ],
      "Subject": [
        "Asunto"
      ],
      "Receiver postal code": [
        "Fecha de entrega"
      ],
      "Receiver town": [
        "Fecha de entrega"
      ],
      "Step 3:": [
        "Paso 3:"
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        ""
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        'Asegurate que TODA la informacion es correcta, incuyendo el asunto y que estas usando la cuenta indicada. Puedes usar el boton copiar (%1$s) para prevenir errores de tipeo o el "payto://" URI debajo para copiar un solo valor.'
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        "Alternativamente si tu banco soporta PayTo URI, puedes usar este %1$s link"
      ],
      "Operation in progress...": [
        "Operaci\xF3n en progreso..."
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        "La operaci\xF3n ser\xE1 autom\xE1ticamente cancelada luego de %1$s segundos"
      ],
      "days": [
        "d\xEDas"
      ],
      "hours": [
        "horas"
      ],
      "minutes": [
        "minutos"
      ],
      "seconds": [
        "segundos"
      ],
      "Forever": [
        "por siempre"
      ],
      "%1$sM": [
        "%1$sM"
      ],
      "%1$sY": [
        "%1$sA"
      ],
      "%1$sd": [
        "%1$sd"
      ],
      "%1$sh": [
        "%1$sh"
      ],
      "%1$smin": [
        "%1$smin"
      ],
      "%1$ssec": [
        "%1$sseg"
      ],
      "Change value to never": [
        "cambiar valor a nunca"
      ],
      "Change value to empty": [
        "cambiar valor a vac\xEDo"
      ],
      "Image must be smaller than 1 MB": [
        "La imagen debe ser mas chica que 1 MB"
      ],
      "Add": [
        "Agregar"
      ],
      "Remove": [
        "Eliminar"
      ],
      "Country": [
        "Pa\xEDs"
      ],
      "Building number": [
        "N\xFAmero de edificio"
      ],
      "Building name": [
        "Nombre de edificio"
      ],
      "Street": [
        "Calle"
      ],
      "Post code": [
        "C\xF3digo postal"
      ],
      "Town location": [
        "Ubicaci\xF3n de ciudad"
      ],
      "District": [
        "Distrito"
      ],
      "Country subdivision": [
        "Subdivisi\xF3n de pa\xEDs"
      ],
      "Additional address lines": [
        "Informaci\xF3n adicional"
      ],
      "Orders": [
        "\xD3rdenes"
      ],
      "Inventory": [
        "Inventario"
      ],
      "Categories": [
        "Categorias"
      ],
      "Wire transfers": [
        "ID de la transferencia"
      ],
      "Templates": [
        "Plantillas"
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        "Expiraci\xF3n"
      ],
      "OTP Devices": [
        "Dispositivos OTP"
      ],
      "Webhooks": [
        "Webhooks"
      ],
      "Settings": [
        "Configuraci\xF3n"
      ],
      "Password": [
        "Contrasena"
      ],
      "Access tokens": [
        "Token de acceso"
      ],
      "Connection": [
        "Conexi\xF3n"
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        "Instancias"
      ],
      "New": [
        "Nuevo"
      ],
      "List": [
        "Lista"
      ],
      "Log out": [
        "Salir"
      ],
      "Username": [
        "Usuario"
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        'Nombre de la instancia en URL. La instancia "por defecto" es especial, ya que se utiliza para administrar otras instancias.'
      ],
      "Business name": [
        "Nombre del negocio"
      ],
      "Legal name of the business represented by this instance.": [
        "Nombre legal de la empresa representada por esta instancia."
      ],
      "Email": [
        "Correo eletr\xF3nico"
      ],
      "Contact email": [
        "Correo electr\xF3nico del contacto"
      ],
      "Phone number": [
        "N\xFAmero de enrutamiento."
      ],
      "Contact phone": [
        "Correo electr\xF3nico del contacto"
      ],
      "Website URL": [
        "URL de sitio web"
      ],
      "URL.": [
        "URL."
      ],
      "Logo": [
        "Logotipo"
      ],
      "Logo image.": [
        "Imagen del logotipo."
      ],
      "Physical location of the merchant.": [
        "Ubicaci\xF3n f\xEDsica del comerciante."
      ],
      "Jurisdiction": [
        "Jurisdicci\xF3n"
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        "Jurisdicci\xF3n para disputas legales con el comerciante."
      ],
      "Pay transaction fee": [
        ""
      ],
      "Cover the transaction cost or pass it on to the user.": [
        ""
      ],
      "Default payment delay": [
        "Retraso del pago por defecto"
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        "Tiempo que los clientes tienen para pagar un pedido antes de que caduque la oferta de forma predeterminada."
      ],
      "Default refund delay": [
        "Retraso del pago por defecto"
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        "Retrazo de transferencia por omisi\xF3n"
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        "Tiempo m\xE1ximo que un proveedor puede retrasar la transferencia de fondos al comerciante, lo que le permite agrupar pagos m\xE1s peque\xF1os en transferencias m\xE1s grandes y reducir las comisiones por transferencia."
      ],
      "Default wire transfer rounding interval": [
        "Retrazo de transferencia por omisi\xF3n"
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        "Orden no encontrada"
      ],
      "To second": [
        "segundos"
      ],
      "To minute": [
        "minutos"
      ],
      "To hour": [
        "horas"
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        "Vence en"
      ],
      "Required": [
        "Requerido"
      ],
      "Unauthorized": [
        "Autorizado"
      ],
      "Sending the code failed": [
        "No se pudo aceder al servidor"
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        "Algoritmo de verificaci\xF3n"
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        "Regresar"
      ],
      "Verify": [
        "Verificado"
      ],
      "Failed to send the verification code.": [
        "Fallo al eliminar instancia"
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        "Sin autenticacion"
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        "Ningun exchange aceptara pagos debido a requerimientos de KYC."
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        "Orden no encontrada"
      ],
      "Login required": [
        "Login necesario"
      ],
      "instance name": [
        "ID de instancia"
      ],
      "current password": [
        "Con contrasena"
      ],
      "Forgot password": [
        "Con contrasena"
      ],
      "Create new account": [
        "Abono en cuenta bancaria"
      ],
      "Invalid": [
        "no v\xE1lido"
      ],
      "Doesn't have the pattern of an email": [
        "La URL no tiene el formato correcto"
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        "formato inv\xE1lido"
      ],
      "Max 7 lines": [
        "m\xE1ximo 7 l\xEDneas"
      ],
      "Doesn't match": [
        ""
      ],
      "Instance created": [
        "ID de instancia"
      ],
      "Unauthorized.": [
        "Autorizado"
      ],
      "Conflict.": [
        "Conflicto."
      ],
      "Not found.": [
        "Orden no encontrada"
      ],
      "New password": [
        "Nueva contrase\xF1a"
      ],
      "Next password to be used": [
        "siguiente token de acceso a usar"
      ],
      "Repeat password": [
        "Repetir contrase\xF1a"
      ],
      "Confirm the same password": [
        "confirmar el mismo token de acceso"
      ],
      "Please complete the marked fields and choose authorization method": [
        "Necesita completar campos marcados y escoger un m\xE9todo de autorizaci\xF3n"
      ],
      "Confirm operation": [
        "Confirmado"
      ],
      "Check the password.": [
        "Con contrasena"
      ],
      "Instance not found.": [
        "Orden no encontrada"
      ],
      "Description": [
        "Descripci\xF3n"
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        "Expiraci\xF3n"
      ],
      "Time the access token will be valid.": [
        "El token de acceso proporcionado no es v\xE1lido."
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        "Elija uno..."
      ],
      "Read only": [
        ""
      ],
      "All": [
        "Todo"
      ],
      "Order simple": [
        "ID de la orden"
      ],
      "Order Point-of-Sale": [
        "Clave punto de venta"
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        "ID de la orden"
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        "Con contrasena"
      ],
      "Please complete the marked fields": [
        "Necesita completar los campos marcados"
      ],
      "Access token created": [
        "Token de acceso"
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        "Token"
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        "No se ha configurado una instancia por 'defecto' todav\xEDa."
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        'Cree una instancia "por defecto" para empezar a utilizar el backoffice comerciante.'
      ],
      "Create access token": [
        "Establecer token de acceso"
      ],
      "Load more devices before the first one": [
        "cargar m\xE1s plantillas antes de la primera"
      ],
      "Load first page": [
        "Cargar primera pagina"
      ],
      "Created at": [
        "Creado en"
      ],
      "Expires at": [
        "Vence en"
      ],
      "Never": [
        "nunca"
      ],
      "Remove this access token": [
        "Repetir token de acceso"
      ],
      "Delete": [
        "Borrar"
      ],
      "Load more devices after the last one": [
        "cargar m\xE1s plantillas despu\xE9s de la \xFAltima"
      ],
      "Load next page": [
        "Cargar siguiente pagina"
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        "No existen productos todav\xEDa, a\xF1adir m\xE1s pulsando el s\xEDmbolo +"
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        "Establecer token de acceso"
      ],
      "Deleting an access token cannot be undone.": [
        "El borrado de un producto no se puede restaurar."
      ],
      "This is not a valid Bitcoin address.": [
        "Esta no es una direcci\xF3n de bitcoin v\xE1lida."
      ],
      "This is not a valid Ethereum address.": [
        "Esta no es una direcci\xF3n de Ethereum v\xE1lida."
      ],
      "This is not a valid host.": [
        "Esta no es una direcci\xF3n de bitcoin v\xE1lida."
      ],
      "IBANs usually have more than 4 digits.": [
        "Los n\xFAmeros IBAN usualmente tienen mas de 4 digitos"
      ],
      "IBANs usually have fewer than 34 digits.": [
        "Los n\xFAmeros IBAN usualmente tienen menos de 34 digitos"
      ],
      "The IBAN's country code could not be retrieved.": [
        "C\xF3digo de pais de IBAN no encontrado"
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        "El n\xFAmero IBAN no es v\xE1lido, fall\xF3 la verificaci\xF3n"
      ],
      "This account is not allowed.": [
        "Esta no es una direcci\xF3n de bitcoin v\xE1lida."
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        "M\xE9todo de transferencia"
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        ""
      ],
      "Select a wire method...": [
        "Selecciona un m\xE9todo de transferencia"
      ],
      "Routing": [
        "Enrutamiento"
      ],
      "Routing number": [
        "N\xFAmero de enrutamiento."
      ],
      "Account": [
        "Cuenta"
      ],
      "Account number": [
        "Numero de cuenta"
      ],
      "Code": [
        "C\xF3digo"
      ],
      "Business Identifier Code": [
        "C\xF3digo de identificaci\xF3n de la empresa."
      ],
      "International Bank Account Number": [
        "N\xFAmero de cuenta bancaria."
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        "Interfaz de pago unificado."
      ],
      "Bitcoin protocol": [
        "Protocolo Bitcoin."
      ],
      "Ethereum protocol": [
        "Protocolo Ethereum."
      ],
      "Interledger protocol": [
        "Protocolo Interledger."
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        "El nombre de la categor\xEDa"
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        "Nombre legal de la persona titular de la cuenta."
      ],
      "Invalid url": [
        "formato inv\xE1lido"
      ],
      "URL must end with a '/'": [
        "La URL que debe terminar con '/'"
      ],
      "URL must not contain params": [
        "La URL no debe contener parametros"
      ],
      "URL must not hash param": [
        "La URL no debe contener hash"
      ],
      "Conflict": [
        "Confirmar"
      ],
      'Server replied with "bad request".': [
        "El servidor responde con un c\xF3digo de error."
      ],
      "Unauthorized, check credentials.": [
        "Sin autorizacion, revisar credenciales."
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        "El endpoint no parece pertenecer al API Taler Revenue."
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        'El request tuvo exito pero el server no respondio con el campo "credit_account" asi que no podemos verificar que sea la misma cuenta.'
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        "Cuenta:"
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        "Si el banco soporta el API Taler Revenue entonces puedes agregar el endpoint URL debajo para mantener la informacion de ganancias sincronizada."
      ],
      "Endpoint URL": [
        "URL endpoint"
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        "Desde donde el comerciante puede descargar informacion acerca de las transferencias bancarias entrantes hacia esta cuenta"
      ],
      "Auth type": [
        "Tipo de evento"
      ],
      "Choose the authentication type for the account info URL": [
        "Verifica el tipo de autenticacion para la URL de infomacion de cuenta"
      ],
      "Without authentication": [
        "Sin autenticacion"
      ],
      "With username and password": [
        "Con contrasena"
      ],
      "With token": [
        "Crear token"
      ],
      "Do not change": [
        "URL del proveedor"
      ],
      "Username to access the account information.": [
        "Usuario para cceder a la informacion de la cuenta."
      ],
      "Password to access the account information.": [
        "Contrasena para acceder a la informacion de la cuenta."
      ],
      "Access token to access the account information.": [
        "Token de acceso para acceder a la informacion de la cuenta."
      ],
      "Match": [
        "Coincidencia"
      ],
      "Check where the information match against the server info.": [
        "Verifica si la informacion con el servidor coincide."
      ],
      "Compare info from server with account form": [
        "Comparar la informacion del servidor con el formulario de la cuenta"
      ],
      "Test": [
        "Probar"
      ],
      "Server didn't like the request we made.": [
        "El servidor responde con un c\xF3digo de error."
      ],
      "Account details": [
        "Direcci\xF3n de cuenta"
      ],
      "Check if the information matches the server info.": [
        "Verifica si la informacion con el servidor coincide."
      ],
      "The bank account has been deleted.": [
        "La cuenta bancaria se ha eliminado correctamente."
      ],
      "Bank accounts": [
        "Cuenta bancaria"
      ],
      "Add new account": [
        "Abono en cuenta bancaria"
      ],
      "Owner's name": [
        "Nombre del dueno"
      ],
      "Delete selected accounts from the database": [
        "eliminar transferencia seleccionada de la base de datos"
      ],
      "There are no accounts yet, add more pressing the + sign": [
        "No existen productos todav\xEDa, a\xF1adir m\xE1s pulsando el s\xEDmbolo +"
      ],
      "You must provide a bank account to receive payments.": [
        "URI que especifica la cuenta bancaria para acreditar los ingresos."
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        "Invalido. Solo caracteres y numero"
      ],
      "Name": [
        "Nombre"
      ],
      "Category name": [
        ""
      ],
      "Category deleted": [
        "producto fue eliminado correctamente"
      ],
      "Add new devices": [
        "cargar nuevas transferencias"
      ],
      "ID": [
        "ID"
      ],
      "Total products": [
        "Precio total"
      ],
      "Delete selected category from the database": [
        "eliminar transferencia seleccionada de la base de datos"
      ],
      "There are no categories yet, add more pressing the + sign": [
        "No hay propinas todav\xEDa, agregar mas presionando el signo +"
      ],
      "Id:": [
        ""
      ],
      "Name of the category": [
        "El nombre de la categor\xEDa"
      ],
      "Products": [
        "Productos"
      ],
      "Image": [
        "Imagen"
      ],
      "Load more products after the last one": [
        "cargar m\xE1s plantillas despu\xE9s de la \xFAltima"
      ],
      "There are no products in this category.": [
        "Manejar lista de productos en la orden."
      ],
      "Account's KYC status": [
        "Direcci\xF3n de cuenta"
      ],
      "Exchange": [
        "Exchange"
      ],
      "Status": [
        "Estado de la URL"
      ],
      "Ok": [
        "Ok"
      ],
      "Action required": [
        "Login necesario"
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        "Se requiere una cantidad"
      ],
      "More information required.": [
        "m\xE1s informaci\xF3n para dar contexto"
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        "Cargando..."
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        "Estado http"
      ],
      "No pending kyc verification!": [
        "\xA1No hay verificaci\xF3n kyc pendiente!"
      ],
      "Wire account is invalid": [
        "El token de acceso proporcionado no es v\xE1lido"
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        "Cambiar clave"
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        "Requerido"
      ],
      "The payment provider requires more information.": [
        "El exchange require una verificacion de la cuenta."
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        "Est\xE1 estableciendo el token de acceso para la nueva instancia"
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        "mostrar s\xF3lo los pedidos en los que los clientes han pagado, pero los pagos por transferencia del proveedor de pago siguen pendientes"
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        ""
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        "cambiar valor a fecha desconocida"
      ],
      "Enter description or id": [
        "Insertar un ID para el pedido"
      ],
      "No match found for that description or ID.": [
        "no se encontraron productos con esa descripci\xF3n"
      ],
      "You must enter a valid product identifier.": [
        "Debe ingresar un identificador de producto v\xE1lido."
      ],
      "Quantity must be greater than 0!": [
        "\xA1Cantidad debe ser mayor que 0!"
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        "Esta cantidad excede las existencias restantes. Actualmente, solo quedan %1$s unidades sin reservar en las existencias."
      ],
      "Search product": [
        "Productos de inventario"
      ],
      "Quantity": [
        "Cantidad"
      ],
      "How many products will be added": [
        "cu\xE1ntos productos ser\xE1n agregados"
      ],
      "Add from inventory": [
        "Agregar del inventario"
      ],
      "This product has %1$s applicable taxes configured.": [
        "Este producto tiene %1$s impuestos aplicables configurados."
      ],
      "No taxes configured for this product.": [
        "Ningun impuesto configurado para este producto."
      ],
      "Amount": [
        "Monto"
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        "Impuestos pueden estar en divisas que difieren de la principal divisa usada por el comerciante."
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        "Ingrese divisa y valor separado por dos puntos, e.g. &quot;USD:2.3&quot;."
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        "Nombre legal del impuesto, e.g. IVA o arancel."
      ],
      "Add tax to the tax list": [
        "agregar impuesto a la lista de impuestos"
      ],
      "Describe and add a product that is not in the inventory list": [
        "describa y agregue un producto que no est\xE1 en la lista de inventarios"
      ],
      "Add custom product": [
        "Agregue un producto personalizado"
      ],
      "Complete information of the product": [
        "Complete informaci\xF3n del producto"
      ],
      "Must be a number.": [
        "no es un n\xFAmero"
      ],
      "Must be greater than 0.": [
        "debe ser mayor que 0"
      ],
      "Photo of the product.": [
        "foto del producto."
      ],
      "Full product description.": [
        "descripci\xF3n completa del producto"
      ],
      "Unit name": [
        "Unidad"
      ],
      "Name of the product unit.": [
        "nombre de la unidad del producto."
      ],
      "Price per unit": [
        "Precio por unidad"
      ],
      "Amount in the current currency.": [
        "Monto de la divisa actual."
      ],
      "How many products will be added.": [
        "Cu\xE1ntos productos ser\xE1n agregados."
      ],
      "Taxes": [
        "Impuestos"
      ],
      "Total price": [
        "Precio total"
      ],
      "Must be greater than 0": [
        "debe ser mayor que 0"
      ],
      "Must have a refund deadline": [
        "deber\xEDa tener un plazo de reembolso"
      ],
      "Auto refund can't be after refund deadline": [
        "reembolso autom\xE1tico no puede ser despu\xE9s qu el plazo de reembolso"
      ],
      "Must be in the future": [
        "deber\xEDan ser en el futuro"
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        "no es v\xE1lido"
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        "Ningun exchange aceptara pagos debido a requerimientos de KYC."
      ],
      "Manage products in order": [
        "Manejar productos en orden"
      ],
      "%1$s products with a total price of %2$s.": [
        "%1$s productos con un precio total de %2$s."
      ],
      "Manage list of products in the order.": [
        "Manejar lista de productos en la orden."
      ],
      "Remove this product from the order.": [
        "Remover este producto de la orden."
      ],
      "Products price sum": [
        "Lista de producto"
      ],
      "Total product price added up": [
        "precio total de producto agregado"
      ],
      "Order price": [
        "Precio por art\xEDculo"
      ],
      "Amount to be paid by the customer": [
        "Monto a ser pagado por el cliente"
      ],
      "Final order price": [
        "Precio final de la orden"
      ],
      "Summary": [
        "Resumen"
      ],
      "Title of the order to be shown to the customer": [
        "T\xEDtulo de la orden a ser mostrado al cliente"
      ],
      "Shipping and fulfillment": [
        "Env\xEDo y cumplimiento"
      ],
      "Delivery date": [
        "Fecha de entrega"
      ],
      "Deadline for physical delivery assured by the merchant.": [
        "Plazo para la entrega f\xEDsica asegurado por el comerciante."
      ],
      "Delivery location": [
        "Fecha de entrega"
      ],
      "Address where the products will be delivered": [
        "direcci\xF3n a donde los productos ser\xE1n entregados"
      ],
      "Fulfillment URL": [
        "URL de cumplimiento"
      ],
      "URL to which the user will be redirected after successful payment.": [
        "URL al cual el usuario ser\xE1 redirigido luego de pago exitoso."
      ],
      "Fulfillment message": [
        "URL de cumplimiento"
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        "Opciones de pago de Taler"
      ],
      "Override default Taler payment settings for this order": [
        "Sobrescribir pagos por omisi\xF3n de Taler para esta orden"
      ],
      "Payment time": [
        "Opciones de pago"
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        "Plazo l\xEDmite para que el cliente pague por la oferta antes de que expire. Productos del inventario ser\xE1n reservados hasta este plazo l\xEDmite."
      ],
      "Default": [
        "Importe por defecto"
      ],
      "Refund time": [
        "Reembolsado"
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        "Tiempo hasta el cual la orden puede ser reembolsada por el comerciante."
      ],
      "Wire transfer time": [
        "ID de la transferencia"
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        "Plazo para que el proveedor haga la transferencia bancaria."
      ],
      "Auto-refund time": [
        "Plazo de reembolso autom\xE1tico"
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        "Tiempo hasta el cual la cartera ser\xE1 autom\xE1ticamente revisada por reembolsos sin interacci\xF3n por parte del usuario."
      ],
      "Maximum fee": [
        "M\xE1xima tarifa de transferencia"
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        "M\xE1xima tarifa de dep\xF3sito que el comerciante esta dispuesto a cubir para esta orden. Mayores tarifas de dep\xF3sito deben ser cubiertas completamente por el consumidor."
      ],
      "Create token": [
        "Crear token"
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        "Si el ID de orden es facil de adivinar el token previene que los usuarios puedan robar ordenes de otras personas."
      ],
      "Minimum age required": [
        "Edad m\xEDnima requerida"
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        "Cualquier valor superior a 0 limitar\xE1 las monedas que se pueden utilizar para pagar este contrato. Si est\xE1 vac\xEDo, la restricci\xF3n de edad vendr\xE1 definida por los productos"
      ],
      "Minimum age defined by the products is %1$s": [
        "La edad m\xEDnima definida por el producto es%1$s"
      ],
      "No product with age restriction in this order": [
        "No hay productos con restriccion de edad en esta orden"
      ],
      "Additional information": [
        "Informaci\xF3n adicional"
      ],
      "Custom information to be included in the contract for this order.": [
        "Informaci\xF3n personalizada que debe incluirse en el contrato para este pedido."
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        "Debes introducir un valor en JavaScript Object Notation (JSON)."
      ],
      "remove": [
        "Eliminar"
      ],
      "Custom field name": [
        "Nombre de edificio"
      ],
      "Disabled": [
        "Inhabilitado"
      ],
      "No deadline": [
        "Plazo de reembolso"
      ],
      "Deadline at %1$s": [
        "Expiracion en %1$s"
      ],
      "Select date to show nearby orders": [
        "seleccione la fecha para mostrar pedidos cercanos"
      ],
      "Only show unpaid orders": [
        "mostrar s\xF3lo pedidos pagados"
      ],
      "Only show paid orders": [
        "mostrar s\xF3lo pedidos pagados"
      ],
      "Paid": [
        "Pagado"
      ],
      "Only show orders with refunds": [
        "mostrar solo pedidos con reembolso"
      ],
      "Refunded": [
        "Reembolsado"
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        "No transferido"
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        "mostrar s\xF3lo los pedidos en los que los clientes han pagado, pero los pagos por transferencia del proveedor de pago siguen pendientes"
      ],
      "Completed": [
        "Transferido"
      ],
      "Remove all filters": [
        "Eliminar todos los filtros"
      ],
      "Clear date filter": [
        "borrar filtro de fechas"
      ],
      "Jump to date (%1$s)": [
        "Saltar a la fecha (%1$s)"
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        "saltar al pedido con el ID de pedido proporcionado"
      ],
      "Order id": [
        "ID de la orden"
      ],
      "Instance unknown": [
        "ID de instancia"
      ],
      "Order unknown": [
        "desconocido"
      ],
      "Create order": [
        "crear orden"
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        "Fecha"
      ],
      "Refund": [
        "Devoluci\xF3n"
      ],
      "copy url": [
        "copiar url"
      ],
      "Load more orders after the last one": [
        "cargue m\xE1s transferencia luego de la \xFAltima"
      ],
      "No orders have been found matching your query!": [
        "\xA1No se encontraron \xF3rdenes que emparejen su b\xFAsqueda!"
      ],
      "Duplicated": [
        "duplicado"
      ],
      "This value exceeds the refundable amount.": [
        "este monto excede el monto reembolsable"
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        "Hay requerimientos de KYC pendientes."
      ],
      "refund": [
        "reembolzar"
      ],
      "Reason": [
        "Raz\xF3n"
      ],
      "Amount to be refunded": [
        "monto a ser reembolsado"
      ],
      "Max refundable:": [
        "M\xE1ximo reembolzable:"
      ],
      "Choose one...": [
        "Elija uno..."
      ],
      "Requested by the customer": [
        "pedido por el consumidor"
      ],
      "Other": [
        "otro"
      ],
      "Why this order is being refunded": [
        "por qu\xE9 esta orden est\xE1 siendo reembolsada"
      ],
      "More information to give context": [
        "m\xE1s informaci\xF3n para dar contexto"
      ],
      "now": [
        "no"
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        "mostrar s\xF3lo los pedidos en los que los clientes han pagado, pero los pagos por transferencia del proveedor de pago siguen pendientes"
      ],
      "This wire transfer has been notified by the payment service provider.": [
        "mostrar s\xF3lo los pedidos en los que los clientes han pagado, pero los pagos por transferencia del proveedor de pago siguen pendientes"
      ],
      "This refund is waiting to be claimed by the customer.": [
        "Monto a ser pagado por el cliente"
      ],
      "This refund has been claimed by the customer.": [
        "Monto a ser pagado por el cliente"
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        "T\xE9rminos de contrato"
      ],
      "Human-readable description of the whole purchase": [
        "descripci\xF3n legible de toda la compra"
      ],
      "Total price for the transaction": [
        "precio total de la transacci\xF3n"
      ],
      "URL for this purchase": [
        "URL para esta compra"
      ],
      "Max fee": [
        "M\xE1xima comisi\xF3n"
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        "tasa m\xE1xima total de dep\xF3sito aceptada por el comerciante para este contrato"
      ],
      "Time when this contract was generated": [
        "momento en que se gener\xF3 este contrato"
      ],
      "Refund deadline": [
        "Plazo de reembolso"
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        "pasado este plazo no se aceptar\xE1n devoluciones"
      ],
      "Payment deadline": [
        "Plazo de pago"
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        "pasado este plazo, el comerciante no aceptar\xE1 pagos por el contrato"
      ],
      "Wire transfer deadline": [
        "Plazo de la transferencia bancaria"
      ],
      "Transfer deadline for the exchange": [
        "plazo de transferencia para el proveedor"
      ],
      "Time indicating when the order should be delivered": [
        "fecha en la que debe entregarse el pedido"
      ],
      "Where the order will be delivered": [
        "d\xF3nde se entregar\xE1 el pedido"
      ],
      "Auto-refund delay": [
        "Plazo de reembolso autom\xE1tico"
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        "cu\xE1nto tiempo debe intentar la cartera obtener el reembolso autom\xE1tico de la compra"
      ],
      "Extra info": [
        "Informaci\xF3n adicional"
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        "datos adicionales que solo son interpretados por la interfaz del comerciante"
      ],
      "order created": [
        "Estado de orden"
      ],
      "pay deadline": [
        "Plazo de pago"
      ],
      "refund deadline": [
        "Plazo de reembolso"
      ],
      "delivery": [
        "Fecha de entrega"
      ],
      "Order": [
        "Orden"
      ],
      "Claimed": [
        "reclamado"
      ],
      "Claimed at": [
        "reclamado en"
      ],
      "Timeline": [
        "Cronolog\xEDa"
      ],
      "Payment details": [
        "Detalles de pago"
      ],
      "Order status": [
        "Estado de orden"
      ],
      "Unpaid": [
        "impago"
      ],
      "Product list": [
        "Lista de producto"
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        "Recibimos el mensaje %1$s desde %2$s"
      ],
      "refund created: %1$s": [
        "reembolzo creado satisfactoriamente"
      ],
      "refund created: %1$s: %2$s": [
        "reembolzo creado satisfactoriamente"
      ],
      "refund taken: %1$s": [
        "Reembolsado"
      ],
      "refund taken: %1$s: %2$s": [
        "Reembolsado"
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        "Plazo de reembolso"
      ],
      "Wired": [
        "transferido"
      ],
      "Refund order": [
        "reembolsado"
      ],
      "Not refundable": [
        "No reembolsable"
      ],
      "Next event in": [
        "Proximo evento en"
      ],
      "The order was wired.": [
        "El pedido a sido cancelado."
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        "Monto reembolsado"
      ],
      "Refund taken": [
        "Reembolsado"
      ],
      "Status URL": [
        "Estado de la URL"
      ],
      "Refund URI": [
        "URI de devoluci\xF3n"
      ],
      "Pay at": [
        "URL de pago"
      ],
      "Order status URL": [
        "URL de estado de orden"
      ],
      "Payment URI": [
        "URI de pago"
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        "Estado de orden desconocido. Esto es un error, por favor contacte a su administrador."
      ],
      "Invalid. Please enter letters and numbers only.": [
        "Invalido. Solo caracteres y numero"
      ],
      "Just letters and numbers from 2 to 7": [
        "s\xF3lo letras y n\xFAmeros del 2 al 7"
      ],
      "The size of the key must be 32 characters": [
        "el tama\xF1o de la clave debe ser 32"
      ],
      "Internal ID on the system": [
        "Id interno en el sistema"
      ],
      "Useful to identify the device physically": [
        "Util para identificar dispositivos fisicos"
      ],
      "Verification algorithm": [
        "Algoritmo de verificaci\xF3n"
      ],
      "Algorithm used to verify transactions in offline mode": [
        "Algoritmo a utilizar para verificar la transacci\xF3n en modo offline"
      ],
      "Device key": [
        "Clave del dispositivo"
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        "Asegurese de que sea muy dificil de adivinar o use el generado aleatorio"
      ],
      "Your device needs to match exactly the same value": [
        "Su dispositivo necesita tener exactamente el mismo valor"
      ],
      "Generate random secret key": [
        "generar clave secreta aleatoria"
      ],
      "Random": [
        "aleatorio"
      ],
      "Create another": [
        "crear orden"
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        "Puede escanear el siguiente codigo QR con su dispositivo o guardar la clave antes de continuar."
      ],
      "OTP devices": [
        "Dispositivo OTP"
      ],
      "Delete selected devices from the database": [
        "eliminar la reserva seleccionada de la base de datos"
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        "Todav\xEDa no hay instancias, agregue m\xE1s presionando el signo +"
      ],
      "Template id is unknown": [
        "El id de la plantilla es desconocido"
      ],
      "The provided information is inconsistent with the current state of the template": [
        "La informacion provista es inconsistente con el estado actual de la plantilla"
      ],
      "Device:": [
        "Dispositivo:"
      ],
      "Algorithm to use to verify transactions in offline mode": [
        "Algoritmo a utilizar para verificar la transacci\xF3n en modo offline"
      ],
      "Not modified": [
        "No transferido"
      ],
      "Be sure to be very hard to guess or use the random generator": [
        "Asegurese de que sea muy dificil de adivinar o use el generado aleatorio"
      ],
      "Your device need to have exactly the same value": [
        "Su dispositivo necesita tener exactamente el mismo valor"
      ],
      "Change key": [
        "Cambiar clave"
      ],
      "Is not the same": [
        "no son iguales"
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        "Est\xE1 actualizando el token de acceso para la instancia con id %1$s"
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        "Confirmar modificaci\xF3n"
      ],
      "Testing password change": [
        "Contrasena"
      ],
      "Password changed": [
        "Contrasena"
      ],
      "The current password is wrong.": [
        "Con contrasena"
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        "Servidor no encontrado"
      ],
      "Add element to the list": [
        "agregar elemento a la lista"
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        "Pulse aqu\xED para configurar el stock del producto, d\xE9jelo como est\xE1 y el backend no controlar\xE1 el stock."
      ],
      "Manage stock": [
        "Administrar stock"
      ],
      "This product has been configured without stock control": [
        "este producto se ha configurado sin control de existencias"
      ],
      "Infinite": [
        "Inifinito"
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        "la p\xE9rdida no puede ser mayor que la cantidad entrante actual (max %1$s )"
      ],
      "Incoming": [
        "Ingresando"
      ],
      "Lost": [
        "Perdido"
      ],
      "Current": [
        "Actual"
      ],
      "Remove stock control for this product": [
        "eliminar el control de existencias de este producto"
      ],
      "without stock": [
        "sin stock"
      ],
      "Next restock": [
        "Pr\xF3ximo reabastecimiento"
      ],
      "Warehouse address": [
        "Direcci\xF3n de cuenta"
      ],
      "Invalid amount": [
        "Importe fijo"
      ],
      "Product identification to use in URLs (for internal use only).": [
        "Identificaci\xF3n del producto para usar en las URL (solo para uso interno)."
      ],
      "Illustration of the product for customers.": [
        "ilustraci\xF3n del producto para los clientes."
      ],
      "Product name.": [
        "Producto"
      ],
      "Product description for customers.": [
        "Descripci\xF3n del producto para los clientes."
      ],
      "Age restriction": [
        "Restricci\xF3n de edad"
      ],
      "Is this product restricted for customer below certain age?": [
        "\xBFeste producto est\xE1 restringido para clientes menores de cierta edad?"
      ],
      "Minimum age of the customer": [
        "Edad m\xEDnima requerida"
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        "Unidad que describe la cantidad de producto vendido (por ejemplo, 2 kilogramos, 5 litros, 3 art\xEDculos, 5 metros) para los clientes."
      ],
      "Example: kg, items or liters": [
        "Ejemplo: kg, item o litros"
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        "Precio de venta para los clientes, impuestos incluidos, por encima de las unidades del producto."
      ],
      "Stock": [
        "Existencias"
      ],
      "Inventory for products with finite supply (for internal use only).": [
        "Inventario de productos para productos con suministro finito (s\xF3lo para uso interno)."
      ],
      "Taxes included in the product price, exposed to customers.": [
        "Impuestos incluidos en el precio del producto, expuestos a los clientes."
      ],
      "Search by category description or id": [
        "buscar productos por su descripci\xF3n o ID"
      ],
      "Categories where this product will be listed on.": [
        "Direcci\xF3n a donde los productos ser\xE1n entregados."
      ],
      "Add product to inventory": [
        "A\xF1adir producto al inventario"
      ],
      "Product not found.": [
        "Servidor no encontrado"
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        "Ventas"
      ],
      "Sold": [
        "Vendido"
      ],
      "Free": [
        "Gratuito"
      ],
      "Go to product update page": [
        "ir a la p\xE1gina de actualizaci\xF3n del producto"
      ],
      "Update": [
        "Actualizar"
      ],
      "Remove this product from the database": [
        "eliminar este producto de la base de datos"
      ],
      "Update the product with new price": [
        "actualizar el producto con el nuevo precio"
      ],
      "Update product with new price": [
        "actualizar producto con nuevo precio"
      ],
      "Confirm update": [
        "Confirmado"
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        "la p\xE9rdida no puede ser mayor que la cantidad entrante actual (max %1$s )"
      ],
      "Add more elements to the inventory": [
        "a\xF1adir m\xE1s elementos al inventario"
      ],
      "Report elements lost in the inventory": [
        "informar de elementos perdidos en el inventario"
      ],
      "New price for the product": [
        "nuevo precio para el producto"
      ],
      "There are values with errors.": [
        "hay valores con errores"
      ],
      "Update product with new stock and price": [
        "actualizar el producto con nuevas existencias y precio"
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        "No existen productos todav\xEDa, a\xF1adir m\xE1s pulsando el s\xEDmbolo +"
      ],
      "Product (ID: %1$s) has been deleted": [
        "El producto (ID: %1$s) se ha actualizado"
      ],
      "Jump to product with the given product ID": [
        "saltar al pedido con el ID de pedido proporcionado"
      ],
      "Product id": [
        "Id de producto"
      ],
      "Delete product": [
        "seleccione un producto primero"
      ],
      'Delete the product "%1$s"': [
        "no se pudo eliminar el producto"
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        "Si elimina el producto con nombre %1$s (ID: %2$s ), el stock y la informacion relacionada se perdera"
      ],
      "Deleting a product cannot be undone.": [
        "El borrado de un producto no se puede restaurar."
      ],
      "Product id:": [
        "ID de producto:"
      ],
      "Must be greater that 0": [
        "debe ser mayor que 0"
      ],
      "Too short": [
        "demasiado corta"
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        "Identificador"
      ],
      "Name of the template in URLs.": [
        "Nombre de la plantilla en las URL."
      ],
      "Template name": [
        "Plantillas"
      ],
      "Describe what this template stands for": [
        "Describa lo que representa esta plantilla"
      ],
      "Order summary": [
        "Estado de orden"
      ],
      "If specified here, this template will create orders with the same summary": [
        "Si se especifica, esta plantilla crear\xE1 pedidos con el mismo resumen"
      ],
      "Summary is editable": [
        "El detalle es editable"
      ],
      "Allow the user to change the summary.": [
        "Permitir el usuario cambiar el detalle."
      ],
      "If specified here, this template will create orders with the same price": [
        "Si se especifica, esta plantilla crear\xE1 pedidos con el mismo precio"
      ],
      "Amount is editable": [
        "Monto abonado"
      ],
      "Allow the user to select the amount to pay.": [
        "Permitir al usuario seleccionar el monto a pagar."
      ],
      "Currency is editable": [
        "La moneda es editable"
      ],
      "Allow the user to change currency.": [
        "Permitir al usuario cambiar la moneda."
      ],
      "Supported currencies": [
        "Monedas soportadas"
      ],
      "Supported currencies: %1$s": [
        "Monedas soportadas: %1$s"
      ],
      "Minimum age": [
        "Edad m\xEDnima"
      ],
      "Is this contract restricted to some age?": [
        "\xBFEste contrato est\xE1 restringido a alguna edad?"
      ],
      "Payment timeout": [
        "Opciones de pago"
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        "Cu\xE1nto tiempo tiene el cliente para completar el pago una vez creado el pedido."
      ],
      "OTP device": [
        "Dispositivo OTP"
      ],
      "Use to verify transactions in offline mode.": [
        "Algoritmo a utilizar para verificar la transacci\xF3n en modo offline."
      ],
      "No OTP device.": [
        "Sin dispositivo OTP."
      ],
      "Add one first": [
        "Agregar uno primero"
      ],
      "No device": [
        "Sin dispotivo"
      ],
      "Add new templates": [
        "a\xF1adir nuevas plantillas"
      ],
      "Load more templates before the first one": [
        "cargar m\xE1s plantillas antes de la primera"
      ],
      "Delete selected templates from the database": [
        "eliminar las plantillas seleccionadas de la base de datos"
      ],
      "Use template to create new order": [
        "utilizar la plantilla para crear un nuevo pedido"
      ],
      "Generate a QR code for the template.": [
        "No se pudo create el reembolso"
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        "cargar m\xE1s plantillas despu\xE9s de la \xFAltima"
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        "No hay propinas todav\xEDa, agregar mas presionando el signo +"
      ],
      "Jump to template with the given template ID": [
        "saltar al pedido con el ID de pedido proporcionado"
      ],
      "Template identification": [
        "Identificacion de plantilla"
      ],
      "Delete template": [
        "cargar nuevas transferencias"
      ],
      'Delete the template "%1$s"': [
        "cargar nuevas transferencias"
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        "Si elimina la plantilla %1$s (ID: %2$s) perdera informacion"
      ],
      "Deleting a template cannot be undone.": [
        "El borrado de un producto no se puede restaurar."
      ],
      "Print": [
        "Imprimir"
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        "Si se especifica, esta plantilla crear\xE1 pedidos con el mismo resumen"
      ],
      "If specified, this template will create orders with the same price": [
        "Si se especifica, esta plantilla crear\xE1 pedidos con el mismo precio"
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        "Cu\xE1nto tiempo tiene el cliente para completar el pago una vez creado el pedido."
      ],
      "An amount is required": [
        "Se requiere una cantidad"
      ],
      "An order summary is required": [
        "Se requiere resumen del pedido"
      ],
      'No more stock for product with ID "%1$s".': [
        'No hay mas stock del producto con id "%1$s".'
      ],
      "New order from template": [
        "cargar viejas transferencias"
      ],
      "Amount of the order": [
        "Importe del pedido"
      ],
      "Slug": [
        ""
      ],
      "Token family slug to use in URLs (for internal use only)": [
        "Identificaci\xF3n del producto para usar en las URL (solo para uso interno)"
      ],
      "Kind": [
        ""
      ],
      "Token family kind": [
        ""
      ],
      "User-readable token family name": [
        ""
      ],
      "Token family description for customers": [
        "descripci\xF3n del producto para los clientes"
      ],
      "Valid After": [
        "V\xE1lido hasta"
      ],
      "Token family can issue tokens after this date": [
        ""
      ],
      "Valid Before": [
        "formato inv\xE1lido"
      ],
      "Token family can issue tokens until this date": [
        ""
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        ""
      ],
      "Token Families": [
        "Familias de tokens"
      ],
      "Add token family": [
        ""
      ],
      "Go to token family update page": [
        "ir a la p\xE1gina de actualizaci\xF3n del producto"
      ],
      "Remove this token family from the database": [
        "eliminar este producto de la base de datos"
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        "No hay propinas todav\xEDa, agregar mas presionando el signo +."
      ],
      "Token family has been deleted.": [
        'Se ha eliminado la familia de tokens "%1$s" (SLUG: %2$s)'
      ],
      "Delete token family": [
        "No se ha podido eliminar la familia de tokens"
      ],
      'Delete the token family "%1$s"': [
        "Eliminando una familia de tokens %1$s ."
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        "Si elimina la familia de tokens %1$s (Slug: %2$s), todos los tokens emitidos no ser\xE1n v\xE1lidos."
      ],
      "Deleting a token family %1$s .": [
        "Eliminando una familia de tokens %1$s ."
      ],
      "can't be undone": [
        "no puede ser vac\xEDo"
      ],
      "Token Family: %1$s": [
        "Familia de tokens: %1$s"
      ],
      "Incoming wire transfers": [
        "ID de la transferencia"
      ],
      "Load more wire transfers preceding the first one": [
        "cargar m\xE1s transferencias antes de la primera"
      ],
      "Expected credit": [
        "Ejecutado en"
      ],
      "Confirmed": [
        "Confirmado"
      ],
      "Validated": [
        "V\xE1lido hasta"
      ],
      "Executed on": [
        "Ejecutado en"
      ],
      "yes": [
        "si"
      ],
      "no": [
        "no"
      ],
      "never": [
        "nunca"
      ],
      "unknown": [
        "desconocido"
      ],
      "Load more transfers after the last one": [
        "cargue m\xE1s transferencia luego de la \xFAltima"
      ],
      "Verified wire transfers": [
        "ID de la transferencia"
      ],
      "Verified": [
        "Verificado"
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        "No hay transferencias todav\xEDa, agregar mas presionando el signo +"
      ],
      "All accounts": [
        "Cuenta"
      ],
      "Filter by account address": [
        "filtrar por direcci\xF3n de cuenta"
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        "mostrar s\xF3lo los pedidos en los que los clientes han pagado, pero los pagos por transferencia del proveedor de pago siguen pendientes"
      ],
      "Only show wire transfers confirmed by the merchant": [
        "mostrar s\xF3lo las transferencias confirmadas por el comerciante"
      ],
      "Wire transfer already confirmed.": [
        "Plazo de la transferencia bancaria"
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        "confirmar el mismo token de acceso"
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        "Cronolog\xEDa"
      ],
      "Transfer ID": [
        "Transferencias"
      ],
      "It's not the same.": [
        "no son iguales"
      ],
      'You are deleting the instance with ID "%1$s"': [
        "Est\xE1 actualizando el token de acceso para la instancia con id %1$s"
      ],
      "Instance": [
        "Instancia"
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        "Purgar"
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        ""
      ],
      "Instance id": [
        "ID de instancia"
      ],
      "Delete this instance": [
        "Borrando la instancia %1$s."
      ],
      "Must be one of '%1$s'": [
        "deber\xEDan ser iguales"
      ],
      "URL is invalid": [
        "no es v\xE1lido"
      ],
      "Webhook ID to use": [
        "ID de webhook a utilizar"
      ],
      "Event": [
        "Evento"
      ],
      "Payment": [
        "URI de pago"
      ],
      "Order settled": [
        "Estado de orden"
      ],
      "Category added": [
        "producto fue eliminado correctamente"
      ],
      "Category updated": [
        "Confirmado"
      ],
      "Inventory added": [
        "Inventario"
      ],
      "Inventory updated": [
        "Inventario"
      ],
      "Inventory deleted": [
        "Inventario"
      ],
      "The event of the webhook: why the webhook is used": [
        "El evento del webhook: por qu\xE9 se utiliza el webhook"
      ],
      "Method": [
        "M\xE9todo"
      ],
      "GET": [
        ""
      ],
      "POST": [
        ""
      ],
      "PUT": [
        ""
      ],
      "PATCH": [
        ""
      ],
      "HEAD": [
        ""
      ],
      "Method used by the webhook": [
        "M\xE9todo utilizado por el webhook"
      ],
      "URL": [
        "URL"
      ],
      "URL of the webhook where the customer will be redirected": [
        "URL del webhook al que se redirigir\xE1 al cliente"
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        ""
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        ""
      ],
      "The short list of variables are:": [
        ""
      ],
      "order's description": [
        "descripci\xF3n"
      ],
      "order's price": [
        "Precio de la orden"
      ],
      "order's unique identification": [
        ""
      ],
      "the amount that was being refunded": [
        "monto a ser reembolsado"
      ],
      "the reason entered by the merchant staff for granting the refund": [
        ""
      ],
      "time of the refund in nanoseconds since 1970": [
        ""
      ],
      "Http header": [
        "Cabecera"
      ],
      "Header template of the webhook": [
        "Plantilla de cabecera del webhook"
      ],
      "Http body": [
        ""
      ],
      "Body template used by the webhook.": [
        "Plantilla del cuerpo del webhook"
      ],
      "Webhook deleted successfully": [
        "el webhook ha sido borrado correctamente"
      ],
      "Add new webhooks": [
        "a\xF1adir nuevos webhooks"
      ],
      "Load more webhooks before the first one": [
        "cargar m\xE1s webhooks antes del primero"
      ],
      "Event type": [
        "Tipo de evento"
      ],
      "Delete selected webhook from the database": [
        "eliminar el webhook seleccionado de la base de datos"
      ],
      "Load more webhooks after the last one": [
        "cargar m\xE1s webhooks despu\xE9s del \xFAltimo"
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        "No hay webhooks todav\xEDa, a\xF1ade m\xE1s pulsando sobre el s\xEDmbolo +"
      ],
      "Webhook updated": [
        "ID de webhook a utilizar"
      ],
      "Header": [
        "Cabecera"
      ],
      "Body": [
        "Cuerpo"
      ],
      "Body template used by the webhook": [
        "Plantilla del cuerpo del webhook"
      ],
      "There is another instance with this username.": [
        ""
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        "Correo electr\xF3nico del contacto"
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        "Creado en"
      ],
      "Password reset": [
        "Contrasena"
      ],
      'Resetting access to the instance "%1$s"': [
        "Est\xE1 estableciendo el token de acceso para la nueva instancia"
      ],
      "Reset": [
        ""
      ],
      "Language": [
        ""
      ],
      "Date format": [
        "formato inv\xE1lido"
      ],
      "How the date is going to be displayed": [
        ""
      ],
      "Merchant type": [
        "Tipo de evento"
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        ""
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        "\xD3rdenes"
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        "Estado de orden"
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        "Estado de orden"
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        "Precio de la orden"
      ],
      "hour": [
        "horas"
      ],
      "day": [
        "d\xEDas"
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        "Eliminar todos los filtros"
      ],
      "Time range": [
        "Cronolog\xEDa"
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        "Actual"
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        "Opciones de pago"
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        "URI que especifica la cuenta bancaria para acreditar los ingresos."
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        ""
      ],
      "Hide for today": [
        "Ocultar por hoy"
      ],
      "KYC verification needed": [
        "Verificaci\xF3n KYC pendiente"
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        ""
      ],
      "The request reached a timeout, check your connection.": [
        "El pedido a terminado en tiempo agotado, verifique su conexi\xF3n."
      ],
      "The request was cancelled.": [
        "El pedido a sido cancelado."
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        "Se hicieron muchos pedidos al mismo servidor y esta acci\xF3n a sido limitada."
      ],
      "The server's response was malformed.": [
        "La respuesta del pedido no esta bien formada."
      ],
      "Could not complete the request due to a network problem.": [
        "No se pudo completar el pedido por problemas de red."
      ],
      "Unexpected request error.": [
        "Error inesperado en el pedido."
      ],
      "Unexpected error.": [
        "Error inesperado."
      ],
      "Add new instance": [
        "agregar nueva instancia"
      ],
      "Edit": [
        "Editar"
      ],
      "Change password": [
        "Con contrasena"
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        "Todav\xEDa no hay instancias, agregue m\xE1s presionando el signo +"
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        "La instancia '%1$s' (ID: %2$s) fue eliminada"
      ],
      "Only show active instances": [
        "Solo mostrar instancias activas"
      ],
      "Active": [
        "Activo"
      ],
      "Only show deleted instances": [
        "Mostrar solo instancias eliminadas"
      ],
      "Deleted": [
        "Eliminado"
      ],
      "Show all instances": [
        "Mostrar todas las instancias"
      ],
      "Delete instance": [
        "Borrando la instancia %1$s."
      ],
      'Delete the instance "%1$s"': [
        "Borrando la instancia %1$s."
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        "Si eliminas la instancia con nombre %1$s (ID: %2$s), el comerciante no podr\xE1 procesar ordenes o rembolsos"
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        "Esta accion elimina la clave privada de la instancia pero preserva toda la informaci\xF3n transaccional. Podr\xE1s acceder a esa informaci\xF3n despu\xE9s de borrar la instancia."
      ],
      "Deleting an instance %1$s": [
        "Borrando la instancia %1$s."
      ],
      "This cannot be undone!": [
        "no puede ser vac\xEDo"
      ],
      "Purge the instance": [
        "Borrando la instancia %1$s."
      ],
      'Purge the instance "%1$s"': [
        "Purgando la instancia %1$s."
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        "Si purgas la instancia con nombre %1$s (ID: %2$s), tambien borraras toda su informacion transaccional."
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        "La instancia desaparecera de tu lista, y no volveras a tener acceso a su informacion."
      ],
      "Purging an instance %1$s": [
        "Purgando la instancia %1$s."
      ],
      "checking compatibility with server...": [
        ""
      ],
      "Contacting the server failed": [
        "No se pudo aceder al servidor"
      ],
      "The server version is not supported": [
        ""
      ],
      'Supported version "%1$s", server version "%2$s".': [
        ""
      ],
      "Deleting": [
        "Borrando"
      ],
      "Changing": [
        "Cambiando"
      ],
      "Manage password": [
        "Administrar stock"
      ],
      "Business Name": [
        "Nombre del negocio"
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        "Configuraci\xF3n"
      ],
      "title%1$s: New bank account": [
        "Cuenta bancaria"
      ],
      "title%1$s: Bank accounts": [
        "Cuenta bancaria"
      ],
      "title%1$s: Update bank Account": [
        "Validar cuenta bancaria: %1$s"
      ],
      "title%1$s: Orders": [
        "\xD3rdenes"
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        "Inventario"
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        "No se pudo actualizar el producto"
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        "no se pudo actualizar el producto"
      ],
      "title%1$s: Wire transfers": [
        "ID de la transferencia"
      ],
      "title%1$s: Webhooks": [
        "Webhooks"
      ],
      "title%1$s: New webhook": [
        "a\xF1adir nuevos webhooks"
      ],
      "title%1$s: Update webhook": [
        "no se ha podido eliminar el webhook"
      ],
      "title%1$s: OTP devices": [
        "Dispositivo OTP"
      ],
      "title%1$s: New OTP device": [
        "Sin dispositivo OTP."
      ],
      "title%1$s: Update OTP device": [
        "Dispositivo OTP"
      ],
      "title%1$s: New template": [
        "a\xF1adir nuevas plantillas"
      ],
      "title%1$s: Update template": [
        "a\xF1adir nuevas plantillas"
      ],
      "title%1$s: Templates": [
        "Plantillas"
      ],
      "title%1$s: Use template": [
        "a\xF1adir nuevas plantillas"
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        "Familias de tokens"
      ],
      "title%1$s: New token family": [
        ""
      ],
      "title%1$s: Update token family": [
        "No se ha podido actualizar la familia de tokens"
      ],
      "title%1$s: Access tokens": [
        "Token de acceso"
      ],
      "title%1$s: New access token": [
        "Nuevo token de acceso"
      ],
      "titleNew instance": [
        "agregar nueva instancia"
      ],
      "titleInstances": [
        "Instancias"
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=2; plural=n != 1;",
  "lang": "es",
  "completeness": 76
};
strings["en"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=2; plural=(n != 1);",
        "lang": "en"
      },
      "Cancel": [
        ""
      ],
      "%1$s": [
        ""
      ],
      "Close": [
        ""
      ],
      "Continue": [
        ""
      ],
      "Clear": [
        ""
      ],
      "Confirm": [
        ""
      ],
      "Correct form": [
        ""
      ],
      "Comparing account details": [
        ""
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        ""
      ],
      "Field": [
        ""
      ],
      "In the form": [
        ""
      ],
      "Reported": [
        ""
      ],
      "Type": [
        ""
      ],
      "IBAN": [
        ""
      ],
      "Address": [
        ""
      ],
      "Host": [
        ""
      ],
      "Account ID": [
        ""
      ],
      "The account owner's legal name": [
        ""
      ],
      "The postal code of the account owner's address": [
        ""
      ],
      "Town": [
        ""
      ],
      "Bank host": [
        ""
      ],
      "Bank account": [
        ""
      ],
      "BIC": [
        ""
      ],
      "OK": [
        ""
      ],
      "Validate bank account: %1$s": [
        ""
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        ""
      ],
      "Step 1:": [
        ""
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        ""
      ],
      "Beneficiary": [
        ""
      ],
      "Step 2:": [
        ""
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        ""
      ],
      "Subject": [
        ""
      ],
      "Receiver postal code": [
        ""
      ],
      "Receiver town": [
        ""
      ],
      "Step 3:": [
        ""
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        ""
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        ""
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        ""
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        ""
      ],
      "Operation in progress...": [
        ""
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        ""
      ],
      "days": [
        ""
      ],
      "hours": [
        ""
      ],
      "minutes": [
        ""
      ],
      "seconds": [
        ""
      ],
      "Forever": [
        ""
      ],
      "%1$sM": [
        ""
      ],
      "%1$sY": [
        ""
      ],
      "%1$sd": [
        ""
      ],
      "%1$sh": [
        ""
      ],
      "%1$smin": [
        ""
      ],
      "%1$ssec": [
        ""
      ],
      "Change value to never": [
        ""
      ],
      "Change value to empty": [
        ""
      ],
      "Image must be smaller than 1 MB": [
        ""
      ],
      "Add": [
        ""
      ],
      "Remove": [
        ""
      ],
      "Country": [
        ""
      ],
      "Building number": [
        ""
      ],
      "Building name": [
        ""
      ],
      "Street": [
        ""
      ],
      "Post code": [
        ""
      ],
      "Town location": [
        ""
      ],
      "District": [
        ""
      ],
      "Country subdivision": [
        ""
      ],
      "Additional address lines": [
        ""
      ],
      "Orders": [
        ""
      ],
      "Inventory": [
        ""
      ],
      "Categories": [
        ""
      ],
      "Wire transfers": [
        ""
      ],
      "Templates": [
        ""
      ],
      "Statistics": [
        ""
      ],
      "Subscriptions and Discounts": [
        ""
      ],
      "Configuration": [
        ""
      ],
      "OTP Devices": [
        ""
      ],
      "Webhooks": [
        ""
      ],
      "Settings": [
        ""
      ],
      "Password": [
        ""
      ],
      "Access tokens": [
        ""
      ],
      "Connection": [
        ""
      ],
      "Personalization": [
        ""
      ],
      "Instances": [
        ""
      ],
      "New": [
        ""
      ],
      "List": [
        ""
      ],
      "Log out": [
        ""
      ],
      "Username": [
        ""
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        ""
      ],
      "Business name": [
        ""
      ],
      "Legal name of the business represented by this instance.": [
        ""
      ],
      "Email": [
        ""
      ],
      "Contact email": [
        ""
      ],
      "Phone number": [
        ""
      ],
      "Contact phone": [
        ""
      ],
      "Website URL": [
        ""
      ],
      "URL.": [
        ""
      ],
      "Logo": [
        ""
      ],
      "Logo image.": [
        ""
      ],
      "Physical location of the merchant.": [
        ""
      ],
      "Jurisdiction": [
        ""
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        ""
      ],
      "Pay transaction fee": [
        ""
      ],
      "Cover the transaction cost or pass it on to the user.": [
        ""
      ],
      "Default payment delay": [
        ""
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        ""
      ],
      "Default refund delay": [
        ""
      ],
      "Time merchants have to refund an order.": [
        ""
      ],
      "Default wire transfer delay": [
        ""
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        ""
      ],
      "Default wire transfer rounding interval": [
        ""
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        ""
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        ""
      ],
      "No rounding": [
        ""
      ],
      "To second": [
        ""
      ],
      "To minute": [
        ""
      ],
      "To hour": [
        ""
      ],
      "To day": [
        ""
      ],
      "To week": [
        ""
      ],
      "To month": [
        ""
      ],
      "To quarter": [
        ""
      ],
      "To year": [
        ""
      ],
      "Expired": [
        ""
      ],
      "Required": [
        ""
      ],
      "Unauthorized": [
        ""
      ],
      "Sending the code failed": [
        ""
      ],
      "Challenge expired": [
        ""
      ],
      "Too many attempts": [
        ""
      ],
      "Validation code sent.": [
        ""
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        ""
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        ""
      ],
      "Verification code": [
        ""
      ],
      "It will expire at %1$s": [
        ""
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        ""
      ],
      "Back": [
        ""
      ],
      "Verify": [
        ""
      ],
      "Failed to send the verification code.": [
        ""
      ],
      "The request was valid, but the server is refusing action.": [
        ""
      ],
      "The backend is not aware of the specified MFA challenge.": [
        ""
      ],
      "Code transmission failed.": [
        ""
      ],
      "Already solved.": [
        ""
      ],
      "It is too early to request another transmission of the challenge.": [
        ""
      ],
      "Multi-factor authentication required.": [
        ""
      ],
      "You need to complete all of this requirements.": [
        ""
      ],
      "You need to complete at least one of this requirements.": [
        ""
      ],
      "An SMS to the phone number starting with %1$s": [
        ""
      ],
      "An email to the address starting with %1$s": [
        ""
      ],
      "You have to wait until %1$s to send a new code.": [
        ""
      ],
      "I have a code": [
        ""
      ],
      "Send me a message": [
        ""
      ],
      "Logged in": [
        ""
      ],
      "Not found": [
        ""
      ],
      "Login required": [
        ""
      ],
      "instance name": [
        ""
      ],
      "current password": [
        ""
      ],
      "Forgot password": [
        ""
      ],
      "Create new account": [
        ""
      ],
      "Invalid": [
        ""
      ],
      "Doesn't have the pattern of an email": [
        ""
      ],
      "Should start with +": [
        ""
      ],
      "A phone number consists of numbers only": [
        ""
      ],
      "Invalid value": [
        ""
      ],
      "Max 7 lines": [
        ""
      ],
      "Doesn't match": [
        ""
      ],
      "Instance created": [
        ""
      ],
      "Unauthorized.": [
        ""
      ],
      "Conflict.": [
        ""
      ],
      "Not found.": [
        ""
      ],
      "New password": [
        ""
      ],
      "Next password to be used": [
        ""
      ],
      "Repeat password": [
        ""
      ],
      "Confirm the same password": [
        ""
      ],
      "Please complete the marked fields and choose authorization method": [
        ""
      ],
      "Confirm operation": [
        ""
      ],
      "Check the password.": [
        ""
      ],
      "Instance not found.": [
        ""
      ],
      "Description": [
        ""
      ],
      "Helps you remember where this token is being used before deleting it.": [
        ""
      ],
      "Duration": [
        ""
      ],
      "Time the access token will be valid.": [
        ""
      ],
      "Refreshable tokens can pose a security risk!": [
        ""
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        ""
      ],
      "Scope": [
        ""
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        ""
      ],
      "Allows all operations without limit.": [
        ""
      ],
      "Allows all operations to read information.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        ""
      ],
      "Allows all operations to read information with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        ""
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        ""
      ],
      "All (refreshable)": [
        ""
      ],
      "Spa": [
        ""
      ],
      "Spa (refreshable)": [
        ""
      ],
      "Choose one": [
        ""
      ],
      "Read only": [
        ""
      ],
      "All": [
        ""
      ],
      "Order simple": [
        ""
      ],
      "Order Point-of-Sale": [
        ""
      ],
      "Order management": [
        ""
      ],
      "Order full": [
        ""
      ],
      "Read only (refreshable)": [
        ""
      ],
      "Order simple (refreshable)": [
        ""
      ],
      "Order Point-of-Sale (refreshable)": [
        ""
      ],
      "Order management (refreshable)": [
        ""
      ],
      "Order full (refreshable)": [
        ""
      ],
      "Current password": [
        ""
      ],
      "Please complete the marked fields": [
        ""
      ],
      "Access token created": [
        ""
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        ""
      ],
      "Token": [
        ""
      ],
      "This token will never expire": [
        ""
      ],
      "This token will be available until %1$s": [
        ""
      ],
      "No 'admin' instance configured yet.": [
        ""
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        ""
      ],
      "Create access token": [
        ""
      ],
      "Load more devices before the first one": [
        ""
      ],
      "Load first page": [
        ""
      ],
      "Created at": [
        ""
      ],
      "Expires at": [
        ""
      ],
      "Never": [
        ""
      ],
      "Remove this access token": [
        ""
      ],
      "Delete": [
        ""
      ],
      "Load more devices after the last one": [
        ""
      ],
      "Load next page": [
        ""
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        ""
      ],
      "Forbidden.": [
        ""
      ],
      "Delete access token": [
        ""
      ],
      "Deleting an access token cannot be undone.": [
        ""
      ],
      "This is not a valid Bitcoin address.": [
        ""
      ],
      "This is not a valid Ethereum address.": [
        ""
      ],
      "This is not a valid host.": [
        ""
      ],
      "IBANs usually have more than 4 digits.": [
        ""
      ],
      "IBANs usually have fewer than 34 digits.": [
        ""
      ],
      "The IBAN's country code could not be retrieved.": [
        ""
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        ""
      ],
      "This account is not allowed.": [
        ""
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        ""
      ],
      "Wire method": [
        ""
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        ""
      ],
      "Select a wire method...": [
        ""
      ],
      "Routing": [
        ""
      ],
      "Routing number": [
        ""
      ],
      "Account": [
        ""
      ],
      "Account number": [
        ""
      ],
      "Code": [
        ""
      ],
      "Business Identifier Code": [
        ""
      ],
      "International Bank Account Number": [
        ""
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        ""
      ],
      "Unified Payment Interface": [
        ""
      ],
      "Bitcoin protocol": [
        ""
      ],
      "Ethereum protocol": [
        ""
      ],
      "Interledger protocol": [
        ""
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        ""
      ],
      "Name of the account holder": [
        ""
      ],
      "John Doe": [
        ""
      ],
      "Legal name of the person holding the account.": [
        ""
      ],
      "Invalid url": [
        ""
      ],
      "URL must end with a '/'": [
        ""
      ],
      "URL must not contain params": [
        ""
      ],
      "URL must not hash param": [
        ""
      ],
      "Conflict": [
        ""
      ],
      'Server replied with "bad request".': [
        ""
      ],
      "Unauthorized, check credentials.": [
        ""
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        ""
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        ""
      ],
      "Unsupported type of account": [
        ""
      ],
      "Account:": [
        ""
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        ""
      ],
      "Endpoint URL": [
        ""
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        ""
      ],
      "Auth type": [
        ""
      ],
      "Choose the authentication type for the account info URL": [
        ""
      ],
      "Without authentication": [
        ""
      ],
      "With username and password": [
        ""
      ],
      "With token": [
        ""
      ],
      "Do not change": [
        ""
      ],
      "Username to access the account information.": [
        ""
      ],
      "Password to access the account information.": [
        ""
      ],
      "Access token to access the account information.": [
        ""
      ],
      "Match": [
        ""
      ],
      "Check where the information match against the server info.": [
        ""
      ],
      "Compare info from server with account form": [
        ""
      ],
      "Test": [
        ""
      ],
      "Server didn't like the request we made.": [
        ""
      ],
      "Account details": [
        ""
      ],
      "Check if the information matches the server info.": [
        ""
      ],
      "The bank account has been deleted.": [
        ""
      ],
      "Bank accounts": [
        ""
      ],
      "Add new account": [
        ""
      ],
      "Owner's name": [
        ""
      ],
      "Delete selected accounts from the database": [
        ""
      ],
      "There are no accounts yet, add more pressing the + sign": [
        ""
      ],
      "You must provide a bank account to receive payments.": [
        ""
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        ""
      ],
      "Invalid. Please use only letters and numbers.": [
        ""
      ],
      "Name": [
        ""
      ],
      "Category name": [
        ""
      ],
      "Category deleted": [
        ""
      ],
      "Add new devices": [
        ""
      ],
      "ID": [
        ""
      ],
      "Total products": [
        ""
      ],
      "Delete selected category from the database": [
        ""
      ],
      "There are no categories yet, add more pressing the + sign": [
        ""
      ],
      "Id:": [
        ""
      ],
      "Name of the category": [
        ""
      ],
      "Products": [
        ""
      ],
      "Image": [
        ""
      ],
      "Load more products after the last one": [
        ""
      ],
      "There are no products in this category.": [
        ""
      ],
      "Account's KYC status": [
        ""
      ],
      "Exchange": [
        ""
      ],
      "Status": [
        ""
      ],
      "Ok": [
        ""
      ],
      "Action required": [
        ""
      ],
      "Warning": [
        ""
      ],
      "Error": [
        ""
      ],
      "Bank account verification required.": [
        ""
      ],
      "More information required.": [
        ""
      ],
      "Awaiting for account review.": [
        ""
      ],
      "Ready": [
        ""
      ],
      "Syncing...": [
        ""
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        ""
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        ""
      ],
      "Incompatible core banking system.": [
        ""
      ],
      "Backend internal error. Contact administrator or check again later.": [
        ""
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        ""
      ],
      "Http Status": [
        ""
      ],
      "No pending kyc verification!": [
        ""
      ],
      "Wire account is invalid": [
        ""
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        ""
      ],
      "No exchange keys": [
        ""
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        ""
      ],
      "Can't wire from this account": [
        ""
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        ""
      ],
      "KYC Required": [
        ""
      ],
      "The payment provider requires more information.": [
        ""
      ],
      "Click here to complete this step.": [
        ""
      ],
      "We are waiting for the access token to be present. Check again later.": [
        ""
      ],
      "Awaiting AML review": [
        ""
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        ""
      ],
      "This account is ready to be used with the payment service provider.": [
        ""
      ],
      "Logic bug": [
        ""
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "Internal error": [
        ""
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        ""
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        ""
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        ""
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        ""
      ],
      "Change value to unknown date": [
        ""
      ],
      "Enter description or id": [
        ""
      ],
      "No match found for that description or ID.": [
        ""
      ],
      "You must enter a valid product identifier.": [
        ""
      ],
      "Quantity must be greater than 0!": [
        ""
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        ""
      ],
      "Search product": [
        ""
      ],
      "Quantity": [
        ""
      ],
      "How many products will be added": [
        ""
      ],
      "Add from inventory": [
        ""
      ],
      "This product has %1$s applicable taxes configured.": [
        ""
      ],
      "No taxes configured for this product.": [
        ""
      ],
      "Amount": [
        ""
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        ""
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        ""
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        ""
      ],
      "Add tax to the tax list": [
        ""
      ],
      "Describe and add a product that is not in the inventory list": [
        ""
      ],
      "Add custom product": [
        ""
      ],
      "Complete information of the product": [
        ""
      ],
      "Must be a number.": [
        ""
      ],
      "Must be greater than 0.": [
        ""
      ],
      "Photo of the product.": [
        ""
      ],
      "Full product description.": [
        ""
      ],
      "Unit name": [
        ""
      ],
      "Name of the product unit.": [
        ""
      ],
      "Price per unit": [
        ""
      ],
      "Amount in the current currency.": [
        ""
      ],
      "How many products will be added.": [
        ""
      ],
      "Taxes": [
        ""
      ],
      "Total price": [
        ""
      ],
      "Must be greater than 0": [
        ""
      ],
      "Must have a refund deadline": [
        ""
      ],
      "Auto refund can't be after refund deadline": [
        ""
      ],
      "Must be in the future": [
        ""
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        ""
      ],
      "is not a valid URL": [
        ""
      ],
      'Product with ID "%1$s" is out of stock.': [
        ""
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        ""
      ],
      "Manage products in order": [
        ""
      ],
      "%1$s products with a total price of %2$s.": [
        ""
      ],
      "Manage list of products in the order.": [
        ""
      ],
      "Remove this product from the order.": [
        ""
      ],
      "Products price sum": [
        ""
      ],
      "Total product price added up": [
        ""
      ],
      "Order price": [
        ""
      ],
      "Amount to be paid by the customer": [
        ""
      ],
      "Final order price": [
        ""
      ],
      "Summary": [
        ""
      ],
      "Title of the order to be shown to the customer": [
        ""
      ],
      "Shipping and fulfillment": [
        ""
      ],
      "Delivery date": [
        ""
      ],
      "Deadline for physical delivery assured by the merchant.": [
        ""
      ],
      "Delivery location": [
        ""
      ],
      "Address where the products will be delivered": [
        ""
      ],
      "Fulfillment URL": [
        ""
      ],
      "URL to which the user will be redirected after successful payment.": [
        ""
      ],
      "Fulfillment message": [
        ""
      ],
      "Message shown to the customer after paying for the order.": [
        ""
      ],
      "Taler payment options": [
        ""
      ],
      "Override default Taler payment settings for this order": [
        ""
      ],
      "Payment time": [
        ""
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        ""
      ],
      "Default": [
        ""
      ],
      "Refund time": [
        ""
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        ""
      ],
      "Wire transfer time": [
        ""
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        ""
      ],
      "Auto-refund time": [
        ""
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        ""
      ],
      "Maximum fee": [
        ""
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        ""
      ],
      "Create token": [
        ""
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        ""
      ],
      "Minimum age required": [
        ""
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        ""
      ],
      "Minimum age defined by the products is %1$s": [
        ""
      ],
      "No product with age restriction in this order": [
        ""
      ],
      "Additional information": [
        ""
      ],
      "Custom information to be included in the contract for this order.": [
        ""
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        ""
      ],
      "remove": [
        ""
      ],
      "Custom field name": [
        ""
      ],
      "Disabled": [
        ""
      ],
      "No deadline": [
        ""
      ],
      "Deadline at %1$s": [
        ""
      ],
      "Select date to show nearby orders": [
        ""
      ],
      "Only show unpaid orders": [
        ""
      ],
      "Only show paid orders": [
        ""
      ],
      "Paid": [
        ""
      ],
      "Only show orders with refunds": [
        ""
      ],
      "Refunded": [
        ""
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        ""
      ],
      "Not wired": [
        ""
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        ""
      ],
      "Completed": [
        ""
      ],
      "Remove all filters": [
        ""
      ],
      "Clear date filter": [
        ""
      ],
      "Jump to date (%1$s)": [
        ""
      ],
      "Gone.": [
        ""
      ],
      "UnavailableForLegalReasons.": [
        ""
      ],
      "Jump to order with the given product ID": [
        ""
      ],
      "Order id": [
        ""
      ],
      "Instance unknown": [
        ""
      ],
      "Order unknown": [
        ""
      ],
      "Create order": [
        ""
      ],
      "The instance doesn't exist": [
        ""
      ],
      "The order doesn't exist": [
        ""
      ],
      "Date": [
        ""
      ],
      "Refund": [
        ""
      ],
      "copy url": [
        ""
      ],
      "Load more orders after the last one": [
        ""
      ],
      "No orders have been found matching your query!": [
        ""
      ],
      "Duplicated": [
        ""
      ],
      "This value exceeds the refundable amount.": [
        ""
      ],
      "Forbidden": [
        ""
      ],
      "Gone": [
        ""
      ],
      "There are pending KYC requirements.": [
        ""
      ],
      "refund": [
        ""
      ],
      "Reason": [
        ""
      ],
      "Amount to be refunded": [
        ""
      ],
      "Max refundable:": [
        ""
      ],
      "Choose one...": [
        ""
      ],
      "Requested by the customer": [
        ""
      ],
      "Other": [
        ""
      ],
      "Why this order is being refunded": [
        ""
      ],
      "More information to give context": [
        ""
      ],
      "now": [
        ""
      ],
      "This is when the time for making refund has been expired.": [
        ""
      ],
      "This is when the time for making the payment has been expired.": [
        ""
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        ""
      ],
      "This wire transfer has been notified by the payment service provider.": [
        ""
      ],
      "This refund is waiting to be claimed by the customer.": [
        ""
      ],
      "This refund has been claimed by the customer.": [
        ""
      ],
      "The current moment in time.": [
        ""
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        ""
      ],
      "Contract terms": [
        ""
      ],
      "Human-readable description of the whole purchase": [
        ""
      ],
      "Total price for the transaction": [
        ""
      ],
      "URL for this purchase": [
        ""
      ],
      "Max fee": [
        ""
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        ""
      ],
      "Time when this contract was generated": [
        ""
      ],
      "Refund deadline": [
        ""
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        ""
      ],
      "Payment deadline": [
        ""
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        ""
      ],
      "Wire transfer deadline": [
        ""
      ],
      "Transfer deadline for the exchange": [
        ""
      ],
      "Time indicating when the order should be delivered": [
        ""
      ],
      "Where the order will be delivered": [
        ""
      ],
      "Auto-refund delay": [
        ""
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        ""
      ],
      "Extra info": [
        ""
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        ""
      ],
      "order created": [
        ""
      ],
      "pay deadline": [
        ""
      ],
      "refund deadline": [
        ""
      ],
      "delivery": [
        ""
      ],
      "Order": [
        ""
      ],
      "Claimed": [
        ""
      ],
      "Claimed at": [
        ""
      ],
      "Timeline": [
        ""
      ],
      "Payment details": [
        ""
      ],
      "Order status": [
        ""
      ],
      "Unpaid": [
        ""
      ],
      "Product list": [
        ""
      ],
      "refund missed: %1$s": [
        ""
      ],
      "refund missed: %1$s: %2$s": [
        ""
      ],
      "refund created: %1$s": [
        ""
      ],
      "refund created: %1$s: %2$s": [
        ""
      ],
      "refund taken: %1$s": [
        ""
      ],
      "refund taken: %1$s: %2$s": [
        ""
      ],
      "The contract terms has a v1 order without choices_index.": [
        ""
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        ""
      ],
      "wired %1$s": [
        ""
      ],
      "wire deadline": [
        ""
      ],
      "Wired": [
        ""
      ],
      "Refund order": [
        ""
      ],
      "Not refundable": [
        ""
      ],
      "Next event in": [
        ""
      ],
      "The order was wired.": [
        ""
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        ""
      ],
      "Refunded amount": [
        ""
      ],
      "Refund taken": [
        ""
      ],
      "Status URL": [
        ""
      ],
      "Refund URI": [
        ""
      ],
      "Pay at": [
        ""
      ],
      "Order status URL": [
        ""
      ],
      "Payment URI": [
        ""
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        ""
      ],
      "Invalid. Please enter letters and numbers only.": [
        ""
      ],
      "Just letters and numbers from 2 to 7": [
        ""
      ],
      "The size of the key must be 32 characters": [
        ""
      ],
      "Internal ID on the system": [
        ""
      ],
      "Useful to identify the device physically": [
        ""
      ],
      "Verification algorithm": [
        ""
      ],
      "Algorithm used to verify transactions in offline mode": [
        ""
      ],
      "Device key": [
        ""
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        ""
      ],
      "Your device needs to match exactly the same value": [
        ""
      ],
      "Generate random secret key": [
        ""
      ],
      "Random": [
        ""
      ],
      "Create another": [
        ""
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        ""
      ],
      "OTP devices": [
        ""
      ],
      "Delete selected devices from the database": [
        ""
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        ""
      ],
      "Template id is unknown": [
        ""
      ],
      "The provided information is inconsistent with the current state of the template": [
        ""
      ],
      "Device:": [
        ""
      ],
      "Algorithm to use to verify transactions in offline mode": [
        ""
      ],
      "Not modified": [
        ""
      ],
      "Be sure to be very hard to guess or use the random generator": [
        ""
      ],
      "Your device need to have exactly the same value": [
        ""
      ],
      "Change key": [
        ""
      ],
      "Is not the same": [
        ""
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        ""
      ],
      "In order to verify that you have access.": [
        ""
      ],
      "Confirm change": [
        ""
      ],
      "Testing password change": [
        ""
      ],
      "Password changed": [
        ""
      ],
      "The current password is wrong.": [
        ""
      ],
      "No enough rights to change the password.": [
        ""
      ],
      "Account not found.": [
        ""
      ],
      "Add element to the list": [
        ""
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        ""
      ],
      "Manage stock": [
        ""
      ],
      "This product has been configured without stock control": [
        ""
      ],
      "Infinite": [
        ""
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        ""
      ],
      "Incoming": [
        ""
      ],
      "Lost": [
        ""
      ],
      "Current": [
        ""
      ],
      "Remove stock control for this product": [
        ""
      ],
      "without stock": [
        ""
      ],
      "Next restock": [
        ""
      ],
      "Warehouse address": [
        ""
      ],
      "Invalid amount": [
        ""
      ],
      "Product identification to use in URLs (for internal use only).": [
        ""
      ],
      "Illustration of the product for customers.": [
        ""
      ],
      "Product name.": [
        ""
      ],
      "Product description for customers.": [
        ""
      ],
      "Age restriction": [
        ""
      ],
      "Is this product restricted for customer below certain age?": [
        ""
      ],
      "Minimum age of the customer": [
        ""
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        ""
      ],
      "Example: kg, items or liters": [
        ""
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        ""
      ],
      "Stock": [
        ""
      ],
      "Inventory for products with finite supply (for internal use only).": [
        ""
      ],
      "Taxes included in the product price, exposed to customers.": [
        ""
      ],
      "Search by category description or id": [
        ""
      ],
      "Categories where this product will be listed on.": [
        ""
      ],
      "Add product to inventory": [
        ""
      ],
      "Product not found.": [
        ""
      ],
      "This change was made without outdated info.": [
        ""
      ],
      "Sales": [
        ""
      ],
      "Sold": [
        ""
      ],
      "Free": [
        ""
      ],
      "Go to product update page": [
        ""
      ],
      "Update": [
        ""
      ],
      "Remove this product from the database": [
        ""
      ],
      "Update the product with new price": [
        ""
      ],
      "Update product with new price": [
        ""
      ],
      "Confirm update": [
        ""
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        ""
      ],
      "Add more elements to the inventory": [
        ""
      ],
      "Report elements lost in the inventory": [
        ""
      ],
      "New price for the product": [
        ""
      ],
      "There are values with errors.": [
        ""
      ],
      "Update product with new stock and price": [
        ""
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        ""
      ],
      "Product (ID: %1$s) has been deleted": [
        ""
      ],
      "Jump to product with the given product ID": [
        ""
      ],
      "Product id": [
        ""
      ],
      "Delete product": [
        ""
      ],
      'Delete the product "%1$s"': [
        ""
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        ""
      ],
      "Deleting a product cannot be undone.": [
        ""
      ],
      "Product id:": [
        ""
      ],
      "Must be greater that 0": [
        ""
      ],
      "Too short": [
        ""
      ],
      "You don't have enough permissions.": [
        ""
      ],
      "There is a template with this identifier, please choose another one.": [
        ""
      ],
      "Identifier": [
        ""
      ],
      "Name of the template in URLs.": [
        ""
      ],
      "Template name": [
        ""
      ],
      "Describe what this template stands for": [
        ""
      ],
      "Order summary": [
        ""
      ],
      "If specified here, this template will create orders with the same summary": [
        ""
      ],
      "Summary is editable": [
        ""
      ],
      "Allow the user to change the summary.": [
        ""
      ],
      "If specified here, this template will create orders with the same price": [
        ""
      ],
      "Amount is editable": [
        ""
      ],
      "Allow the user to select the amount to pay.": [
        ""
      ],
      "Currency is editable": [
        ""
      ],
      "Allow the user to change currency.": [
        ""
      ],
      "Supported currencies": [
        ""
      ],
      "Supported currencies: %1$s": [
        ""
      ],
      "Minimum age": [
        ""
      ],
      "Is this contract restricted to some age?": [
        ""
      ],
      "Payment timeout": [
        ""
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        ""
      ],
      "OTP device": [
        ""
      ],
      "Use to verify transactions in offline mode.": [
        ""
      ],
      "No OTP device.": [
        ""
      ],
      "Add one first": [
        ""
      ],
      "No device": [
        ""
      ],
      "Add new templates": [
        ""
      ],
      "Load more templates before the first one": [
        ""
      ],
      "Delete selected templates from the database": [
        ""
      ],
      "Use template to create new order": [
        ""
      ],
      "Generate a QR code for the template.": [
        ""
      ],
      "Show QR": [
        ""
      ],
      "Load more templates after the last one": [
        ""
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        ""
      ],
      "Jump to template with the given template ID": [
        ""
      ],
      "Template identification": [
        ""
      ],
      "Delete template": [
        ""
      ],
      'Delete the template "%1$s"': [
        ""
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        ""
      ],
      "Deleting a template cannot be undone.": [
        ""
      ],
      "Print": [
        ""
      ],
      "The template configuration needs to be fixed.": [
        ""
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        ""
      ],
      "If specified, this template will create order with the same summary": [
        ""
      ],
      "If specified, this template will create orders with the same price": [
        ""
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        ""
      ],
      "An amount is required": [
        ""
      ],
      "An order summary is required": [
        ""
      ],
      'No more stock for product with ID "%1$s".': [
        ""
      ],
      "New order from template": [
        ""
      ],
      "Amount of the order": [
        ""
      ],
      "Slug": [
        ""
      ],
      "Token family slug to use in URLs (for internal use only)": [
        ""
      ],
      "Kind": [
        ""
      ],
      "Token family kind": [
        ""
      ],
      "User-readable token family name": [
        ""
      ],
      "Token family description for customers": [
        ""
      ],
      "Valid After": [
        ""
      ],
      "Token family can issue tokens after this date": [
        ""
      ],
      "Valid Before": [
        ""
      ],
      "Token family can issue tokens until this date": [
        ""
      ],
      "Validity Granularity": [
        ""
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        ""
      ],
      "Validity duration of a issued token": [
        ""
      ],
      "Token Families": [
        ""
      ],
      "Add token family": [
        ""
      ],
      "Go to token family update page": [
        ""
      ],
      "Remove this token family from the database": [
        ""
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        ""
      ],
      "Token family has been deleted.": [
        ""
      ],
      "Delete token family": [
        ""
      ],
      'Delete the token family "%1$s"': [
        ""
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        ""
      ],
      "Deleting a token family %1$s .": [
        ""
      ],
      "can't be undone": [
        ""
      ],
      "Token Family: %1$s": [
        ""
      ],
      "Incoming wire transfers": [
        ""
      ],
      "Load more wire transfers preceding the first one": [
        ""
      ],
      "Expected credit": [
        ""
      ],
      "Confirmed": [
        ""
      ],
      "Validated": [
        ""
      ],
      "Executed on": [
        ""
      ],
      "yes": [
        ""
      ],
      "no": [
        ""
      ],
      "never": [
        ""
      ],
      "unknown": [
        ""
      ],
      "Load more transfers after the last one": [
        ""
      ],
      "Verified wire transfers": [
        ""
      ],
      "Verified": [
        ""
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        ""
      ],
      "All accounts": [
        ""
      ],
      "Filter by account address": [
        ""
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        ""
      ],
      "Only show wire transfers confirmed by the merchant": [
        ""
      ],
      "Wire transfer already confirmed.": [
        ""
      ],
      "I have received the wire transfer": [
        ""
      ],
      "Confirm the wire transfer": [
        ""
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        ""
      ],
      "Time": [
        ""
      ],
      "Transfer ID": [
        ""
      ],
      "It's not the same.": [
        ""
      ],
      'You are deleting the instance with ID "%1$s"': [
        ""
      ],
      "Instance": [
        ""
      ],
      "Write the instance name to confirm the deletion": [
        ""
      ],
      "Purge": [
        ""
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        ""
      ],
      "DELETE": [
        ""
      ],
      "Pay delay can't be greater than wire transfer delay": [
        ""
      ],
      "Instance id": [
        ""
      ],
      "Delete this instance": [
        ""
      ],
      "Must be one of '%1$s'": [
        ""
      ],
      "URL is invalid": [
        ""
      ],
      "Webhook ID to use": [
        ""
      ],
      "Event": [
        ""
      ],
      "Payment": [
        ""
      ],
      "Order settled": [
        ""
      ],
      "Category added": [
        ""
      ],
      "Category updated": [
        ""
      ],
      "Inventory added": [
        ""
      ],
      "Inventory updated": [
        ""
      ],
      "Inventory deleted": [
        ""
      ],
      "The event of the webhook: why the webhook is used": [
        ""
      ],
      "Method": [
        ""
      ],
      "GET": [
        ""
      ],
      "POST": [
        ""
      ],
      "PUT": [
        ""
      ],
      "PATCH": [
        ""
      ],
      "HEAD": [
        ""
      ],
      "Method used by the webhook": [
        ""
      ],
      "URL": [
        ""
      ],
      "URL of the webhook where the customer will be redirected": [
        ""
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        ""
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        ""
      ],
      "The short list of variables are:": [
        ""
      ],
      "order's description": [
        ""
      ],
      "order's price": [
        ""
      ],
      "order's unique identification": [
        ""
      ],
      "the amount that was being refunded": [
        ""
      ],
      "the reason entered by the merchant staff for granting the refund": [
        ""
      ],
      "time of the refund in nanoseconds since 1970": [
        ""
      ],
      "Http header": [
        ""
      ],
      "Header template of the webhook": [
        ""
      ],
      "Http body": [
        ""
      ],
      "Body template used by the webhook.": [
        ""
      ],
      "Webhook deleted successfully": [
        ""
      ],
      "Add new webhooks": [
        ""
      ],
      "Load more webhooks before the first one": [
        ""
      ],
      "Event type": [
        ""
      ],
      "Delete selected webhook from the database": [
        ""
      ],
      "Load more webhooks after the last one": [
        ""
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        ""
      ],
      "Webhook updated": [
        ""
      ],
      "Header": [
        ""
      ],
      "Body": [
        ""
      ],
      "Body template used by the webhook": [
        ""
      ],
      "There is another instance with this username.": [
        ""
      ],
      "Self provision": [
        ""
      ],
      "Phone": [
        ""
      ],
      "Contact phone number": [
        ""
      ],
      "I agree": [
        ""
      ],
      "Accept the Terms of service": [
        ""
      ],
      "You must accept the Terms of service to continue.": [
        ""
      ],
      "Create": [
        ""
      ],
      "Password reset": [
        ""
      ],
      'Resetting access to the instance "%1$s"': [
        ""
      ],
      "Reset": [
        ""
      ],
      "Language": [
        ""
      ],
      "Date format": [
        ""
      ],
      "How the date is going to be displayed": [
        ""
      ],
      "Merchant type": [
        ""
      ],
      "Simplify UI based on the user usage.": [
        ""
      ],
      "Expert user": [
        ""
      ],
      "Taler developer": [
        ""
      ],
      "Unattended in-person offline vending": [
        ""
      ],
      "In-person online point-of-sale with inventory": [
        ""
      ],
      "Digital publishing": [
        ""
      ],
      "E-commerce site": [
        ""
      ],
      "Developer mode": [
        ""
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        ""
      ],
      "Your browser does not support the canvas element.": [
        ""
      ],
      "Bad gateway": [
        ""
      ],
      "Service unavailable": [
        ""
      ],
      "Select date from which to show statistics": [
        ""
      ],
      "Show chart": [
        ""
      ],
      "Orders chart": [
        ""
      ],
      "Show table": [
        ""
      ],
      "Orders table": [
        ""
      ],
      "Start from (%1$s)": [
        ""
      ],
      "Orders statistics": [
        ""
      ],
      "Since": [
        ""
      ],
      "No order statistics yet.": [
        ""
      ],
      "# of orders since": [
        ""
      ],
      "hour": [
        ""
      ],
      "day": [
        ""
      ],
      "week": [
        ""
      ],
      "weeks": [
        ""
      ],
      "month": [
        ""
      ],
      "months": [
        ""
      ],
      "quarter": [
        ""
      ],
      "quarters": [
        ""
      ],
      "years": [
        ""
      ],
      "Revenue chart": [
        ""
      ],
      "Revenue table": [
        ""
      ],
      "Revenue statistics filter": [
        ""
      ],
      "Time range": [
        ""
      ],
      "Select time range to group dataset by": [
        ""
      ],
      "Select the number of ranges to include in data set": [
        ""
      ],
      "Currency": [
        ""
      ],
      "Select the currency to show statistics for": [
        ""
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        ""
      ],
      "Start time": [
        ""
      ],
      "No revenue statistics yet.": [
        ""
      ],
      "Revenue": [
        ""
      ],
      "Welcome!": [
        ""
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        ""
      ],
      "You need to associate a bank account to receive revenue.": [
        ""
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        ""
      ],
      "Hide for today": [
        ""
      ],
      "KYC verification needed": [
        ""
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        ""
      ],
      "The request reached a timeout, check your connection.": [
        ""
      ],
      "The request was cancelled.": [
        ""
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        ""
      ],
      "The server's response was malformed.": [
        ""
      ],
      "Could not complete the request due to a network problem.": [
        ""
      ],
      "Unexpected request error.": [
        ""
      ],
      "Unexpected error.": [
        ""
      ],
      "Add new instance": [
        ""
      ],
      "Edit": [
        ""
      ],
      "Change password": [
        ""
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        ""
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        ""
      ],
      "Only show active instances": [
        ""
      ],
      "Active": [
        ""
      ],
      "Only show deleted instances": [
        ""
      ],
      "Deleted": [
        ""
      ],
      "Show all instances": [
        ""
      ],
      "Delete instance": [
        ""
      ],
      'Delete the instance "%1$s"': [
        ""
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        ""
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        ""
      ],
      "Deleting an instance %1$s": [
        ""
      ],
      "This cannot be undone!": [
        ""
      ],
      "Purge the instance": [
        ""
      ],
      'Purge the instance "%1$s"': [
        ""
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        ""
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        ""
      ],
      "Purging an instance %1$s": [
        ""
      ],
      "checking compatibility with server...": [
        ""
      ],
      "Contacting the server failed": [
        ""
      ],
      "The server version is not supported": [
        ""
      ],
      'Supported version "%1$s", server version "%2$s".': [
        ""
      ],
      "Deleting": [
        ""
      ],
      "Changing": [
        ""
      ],
      "Manage password": [
        ""
      ],
      "Business Name": [
        ""
      ],
      "Available statistics": [
        ""
      ],
      "There are no statistics to list": [
        ""
      ],
      "title%1$s: Settings": [
        ""
      ],
      "title%1$s: New bank account": [
        ""
      ],
      "title%1$s: Bank accounts": [
        ""
      ],
      "title%1$s: Update bank Account": [
        ""
      ],
      "title%1$s: Orders": [
        ""
      ],
      "title%1$s: New order": [
        ""
      ],
      "title%1$s: Inventory": [
        ""
      ],
      "title%1$s: New product": [
        ""
      ],
      "title%1$s: Update product": [
        ""
      ],
      "title%1$s: Category": [
        ""
      ],
      "title%1$s: New category": [
        ""
      ],
      "title%1$s: Update category": [
        ""
      ],
      "title%1$s: Wire transfers": [
        ""
      ],
      "title%1$s: Webhooks": [
        ""
      ],
      "title%1$s: New webhook": [
        ""
      ],
      "title%1$s: Update webhook": [
        ""
      ],
      "title%1$s: OTP devices": [
        ""
      ],
      "title%1$s: New OTP device": [
        ""
      ],
      "title%1$s: Update OTP device": [
        ""
      ],
      "title%1$s: New template": [
        ""
      ],
      "title%1$s: Update template": [
        ""
      ],
      "title%1$s: Templates": [
        ""
      ],
      "title%1$s: Use template": [
        ""
      ],
      "title%1$s: Personalization": [
        ""
      ],
      "title%1$s: Token families": [
        ""
      ],
      "title%1$s: New token family": [
        ""
      ],
      "title%1$s: Update token family": [
        ""
      ],
      "title%1$s: Access tokens": [
        ""
      ],
      "title%1$s: New access token": [
        ""
      ],
      "titleNew instance": [
        ""
      ],
      "titleInstances": [
        ""
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=2; plural=(n != 1);",
  "lang": "en",
  "completeness": 100
};
strings["de"] = {
  "locale_data": {
    "messages": {
      "": {
        "domain": "messages",
        "plural_forms": "nplurals=2; plural=n != 1;",
        "lang": "de"
      },
      "Cancel": [
        "Abbrechen"
      ],
      "%1$s": [
        "%1$s"
      ],
      "Close": [
        "Schlie\xDFen"
      ],
      "Continue": [
        "Weiter"
      ],
      "Clear": [
        "Leeren"
      ],
      "Confirm": [
        "Eingeben"
      ],
      "Correct form": [
        "Korrektur der Angaben"
      ],
      "Comparing account details": [
        "Vergleich der Kontodetails"
      ],
      "The connection to the account info URL was successful, but the reported account information does not match the details in the form.": [
        "Die Verbindung zur Kontoinformationen-URL ist zustandegekommen, aber die zur\xFCckgemeldeten Kontendaten unterscheiden sich von den Kontodetails im Formular."
      ],
      "Field": [
        "Feld"
      ],
      "In the form": [
        "Im Formular"
      ],
      "Reported": [
        "Gemeldet"
      ],
      "Type": [
        "Typ"
      ],
      "IBAN": [
        "IBAN"
      ],
      "Address": [
        "Adresse"
      ],
      "Host": [
        "Host"
      ],
      "Account ID": [
        "Konto-ID"
      ],
      "The account owner's legal name": [
        "Der rechtsg\xFCltige Vor- und Zuname des Kontoinhabers"
      ],
      "The postal code of the account owner's address": [
        "Postleitzahl der Adresse des Kontoinhabers"
      ],
      "Town": [
        "Ort"
      ],
      "Bank host": [
        "Server der Bank (Bank host)"
      ],
      "Bank account": [
        "Bankkonto"
      ],
      "BIC": [
        "BIC"
      ],
      "OK": [
        "OK"
      ],
      "Validate bank account: %1$s": [
        "Pr\xFCfung der G\xFCltigkeit der Kontonummer: %1$s"
      ],
      "In order to prove that you are the beneficial owner of the bank account, you are required to wire a small amount to a specified bank account with the subject below.": [
        "Um nachzuweisen, dass Sie der wirtschaftlich Berechtigte des Bankkontos sind, m\xFCssen Sie einen kleinen Betrag auf das Bankkonto des Zahlungsdiensts mit dem folgenden Text im Mitteilungsfeld \xFCberweisen."
      ],
      "Step 1:": [
        "Schritt 1:"
      ],
      "Copy and paste this IBAN and the legal name of the beneficial owner into the respective fields in your preferred banking app or online banking website.": [
        "Kopieren Sie diese IBAN und f\xFCgen Sie diese mit dem Namen des Beg\xFCnstigten in die entsprechenden Felder Ihrer Banking-App oder Ihrer Online-Banking-Website ein."
      ],
      "Beneficiary": [
        "Beg\xFCnstigter"
      ],
      "Step 2:": [
        "Schritt 2:"
      ],
      "Copy this string and paste it into the 'Subject' or 'Purpose' field in your preferred banking app or online banking website.": [
        'Kopieren Sie diese Zeichenkette ins Feld \u201EBuchungsvermerk\u201C (bzw. \u201EVerwendungszweck\u201C oder "Mitteilung an Empf\xE4nger") Ihrer Banking-App oder Online-Banking-Webseite.'
      ],
      "Subject": [
        "Buchungsvermerk"
      ],
      "Receiver postal code": [
        "Postleitzahl der Adresse des Beg\xFCnstigten"
      ],
      "Receiver town": [
        "Stadt mit Sitz des Beg\xFCnstigten"
      ],
      "Step 3:": [
        "Schritt 3:"
      ],
      "Select the smallest possible amount for a wire transfer in your preferred banking app or online banking website.": [
        "W\xE4hlen Sie in Ihrer Banking-App oder Ihrem Online-Banking den kleinstm\xF6glichen Betrag f\xFCr eine \xDCberweisung aus."
      ],
      'Make sure ALL data is correct, especially the subject, and that you are choosing the bank account from which you definitely want to make the wire transfer. You can use the copy buttons (%1$s) to avoid typos or make use of the "payto://" URI below to copy just one value.': [
        'Vergewissern Sie sich, dass ALLE Daten und der Buchungsvermerk korrekt sind und Sie das richtige Bankkonto gew\xE4hlt haben. Sie k\xF6nnen die Daten durch Klicks auf die Kopiersymbole \xFCbernehmen (%1$s), um Fehler durch eigenh\xE4ndige Eingabe zu vermeiden, oder den "payto://"-URI verwenden, um mit einer Angabe alles zu kopieren.'
      ],
      "If your bank application allows you make a wire transfer using standard QR codes then try scanning the following:": [
        "Wenn Ihre Bank-App eine \xDCberweisung mit QR-Codes erm\xF6glicht, versuchen Sie, diesen Code zu scannen:"
      ],
      "As an alternative, in case that your bank already supports the 'PayTo URI' standard, you can use this %1$s link instead": [
        "Falls Ihre Bank den technischen Standard \u201Epayto:// URI\u201C bereits unterst\xFCtzt, k\xF6nnen Sie alternativ dazu den Link %1$s verwenden"
      ],
      "Operation in progress...": [
        "Die Verarbeitung ist im Gange..."
      ],
      "The operation will be automatically cancelled after %1$s seconds": [
        "Der Vorgang wird nach %1$s Sekunden automatisch abgebrochen"
      ],
      "days": [
        "Tage"
      ],
      "hours": [
        "Stunden"
      ],
      "minutes": [
        "Minuten"
      ],
      "seconds": [
        "Sekunden"
      ],
      "Forever": [
        "Unbegrenzt"
      ],
      "%1$sM": [
        "%1$sM"
      ],
      "%1$sY": [
        "%1$sY"
      ],
      "%1$sd": [
        "%1$sd"
      ],
      "%1$sh": [
        "%1$sh"
      ],
      "%1$smin": [
        "%1$smin"
      ],
      "%1$ssec": [
        "%1$ssec"
      ],
      "Change value to never": [
        "Wert auf \u201Enie\u201C setzen"
      ],
      "Change value to empty": [
        "Wert annullieren"
      ],
      "Image must be smaller than 1 MB": [
        "Die Bilddatei muss kleiner als 1 MB sein"
      ],
      "Add": [
        "Hinzuf\xFCgen"
      ],
      "Remove": [
        "Entfernen"
      ],
      "Country": [
        "Land"
      ],
      "Building number": [
        "Zus\xE4tzliche Adressangabe"
      ],
      "Building name": [
        "Name des Geb\xE4udes"
      ],
      "Street": [
        "Str."
      ],
      "Post code": [
        "Postleitzahl"
      ],
      "Town location": [
        "Ortsteil"
      ],
      "District": [
        "Bezirk/Kanton"
      ],
      "Country subdivision": [
        "Region"
      ],
      "Additional address lines": [
        "Zus\xE4tzliche Adresszeilen"
      ],
      "Orders": [
        "Bestellungen"
      ],
      "Inventory": [
        "Lagerbestand"
      ],
      "Categories": [
        "Kategorien"
      ],
      "Wire transfers": [
        "\xDCberweisungen"
      ],
      "Templates": [
        "Vorlagen"
      ],
      "Statistics": [
        "Statistiken"
      ],
      "Subscriptions and Discounts": [
        "Abos und Rabatte"
      ],
      "Configuration": [
        "Einstellungen"
      ],
      "OTP Devices": [
        "OTP-Ger\xE4te"
      ],
      "Webhooks": [
        "Webhooks"
      ],
      "Settings": [
        "Einstellungen"
      ],
      "Password": [
        "Passwort"
      ],
      "Access tokens": [
        "Zugangstoken"
      ],
      "Connection": [
        "Verbindung"
      ],
      "Personalization": [
        "Personalisierung"
      ],
      "Instances": [
        "Instanzen"
      ],
      "New": [
        "Neu"
      ],
      "List": [
        "Listen"
      ],
      "Log out": [
        "Abmelden"
      ],
      "Username": [
        "Benutzername"
      ],
      "Name of the instance in URLs. The 'admin' instance is special in that it is used to administer other instances.": [
        "K\xFCrzel Ihres Betriebsnamens, der auch f\xFCr eine Webadresse verwendet werden wird."
      ],
      "Business name": [
        "Firmenname"
      ],
      "Legal name of the business represented by this instance.": [
        "Offizieller Name des Gesch\xE4fts, der Firma oder des Verkaufsorts (z.B. wie im Handelsregister angegeben)."
      ],
      "Email": [
        "E-Mail"
      ],
      "Contact email": [
        "E-Mail-Adresse f\xFCr allgemeinen Kontakt"
      ],
      "Phone number": [
        "Telefonnummer"
      ],
      "Contact phone": [
        "Telefonnummer f\xFCr allgemeinen Kontakt"
      ],
      "Website URL": [
        "Webseiten-URL"
      ],
      "URL.": [
        "URL."
      ],
      "Logo": [
        "Logo (Signet)"
      ],
      "Logo image.": [
        "Abbildung des Firmenlogos."
      ],
      "Physical location of the merchant.": [
        "Niederlassung (Firmensitz) des H\xE4ndlers oder Verk\xE4ufers."
      ],
      "Jurisdiction": [
        "Gerichtsstand"
      ],
      "Jurisdiction for legal disputes with the merchant.": [
        "Gerichtsstand der Firma."
      ],
      "Pay transaction fee": [
        "Transaktionsgeb\xFChren bezahlen"
      ],
      "Cover the transaction cost or pass it on to the user.": [
        "Geb\xFChren aller Transaktionen \xFCbernehmen die Verk\xE4ufer (an) oder die Zahlenden (aus)."
      ],
      "Default payment delay": [
        "Normale Zahlungsfrist"
      ],
      "Time customers have to pay an order before the offer expires by default.": [
        "Zeitraum, in dem K\xE4ufer eine Bestellung bezahlen m\xFCssen, bevor das Angebot regul\xE4r verf\xE4llt."
      ],
      "Default refund delay": [
        "Zeitraum f\xFCr R\xFCckerstattungen"
      ],
      "Time merchants have to refund an order.": [
        "Dies ist der Zeitraum, in dem ein Verk\xE4ufer R\xFCckerstattungen f\xFCr bereits bezahlte Bestellungen gew\xE4hren kann."
      ],
      "Default wire transfer delay": [
        "Zeitlicher Aufschub der \xDCberweisungen"
      ],
      "Maximum time an exchange is allowed to delay wiring funds to the merchant, enabling it to aggregate smaller payments into larger wire transfers and reducing wire fees.": [
        "Zeitraum, in dem alle Kundenzahlungen zu einem Betrag zusammengefasst und vom Zahlungsdienstleister an das Bankkonto des Verk\xE4ufers \xFCberwiesen werden, um die \xDCberweisungsgeb\xFChren zu reduzieren."
      ],
      "Default wire transfer rounding interval": [
        "Rundung des zeitlichen Aufschubs der \xDCberweisungen"
      ],
      "Interval to which wire deadlines should be rounded up to.": [
        "Dies ist die Rundung des Zeitraums auf ganze Zeiteinheiten, nach deren Ablauf die zusammengefassten Ums\xE4tze an das Bankkonto \xFCberwiesen werden."
      ],
      "As an example if you set the interval to day the wire transfer deadline will be rounded to midnight.": [
        "Wenn Sie beispielsweise das Intervall auf einen Tag festlegen, wird die \xDCberweisungsfrist zu Mitternacht dieses Tages aufgerundet."
      ],
      "No rounding": [
        "Keine Rundung"
      ],
      "To second": [
        "Rundung auf eine Sekunde"
      ],
      "To minute": [
        "Rundung auf eine Minute"
      ],
      "To hour": [
        "Rundung auf eine Stunde"
      ],
      "To day": [
        "Rundung auf einen Tag"
      ],
      "To week": [
        "Rundung auf eine Woche"
      ],
      "To month": [
        "Rundung auf einen Monat"
      ],
      "To quarter": [
        "Rundung auf ein Vierteljahr"
      ],
      "To year": [
        "Rundung auf ein Jahr"
      ],
      "Expired": [
        "Abgelaufen"
      ],
      "Required": [
        "Erforderlich"
      ],
      "Unauthorized": [
        "Unberechtigter Zugriff"
      ],
      "Sending the code failed": [
        "Die \xDCbermittlung des Best\xE4tigungscodes ist fehlgeschlagen"
      ],
      "Challenge expired": [
        "Der Zeitraum f\xFCr die \xDCberpr\xFCfung ist abgelaufen"
      ],
      "Too many attempts": [
        "Es gab zu viele fehlgeschlagene Versuche"
      ],
      "Validation code sent.": [
        "Ein Best\xE4tigungscode wurde gerade zugesendet."
      ],
      'The verification code sent to the phone number starting with "%1$s"': [
        'Ein Best\xE4tigungscode wurde an die Telefonnummer gesendet, die mit "%1$s" beginnt'
      ],
      'The verification code sent to the email address starting with "%1$s"': [
        'Ein Best\xE4tigungscode wurde an die E-Mailadresse gesendet, die mit "%1$s" beginnt'
      ],
      "Verification code": [
        "Best\xE4tigungscode"
      ],
      "It will expire at %1$s": [
        "Ende der G\xFCltigkeit %1$s"
      ],
      "The challenge is expired and can't be solved but you can go back and create a new challenge.": [
        "Das Pr\xFCfverfahren zur gesicherten Anmeldung ist abgelaufen und kann nicht mehr verwendet werden, es ist jedoch m\xF6glich, ein neues Pr\xFCfverfahren anzufordern (bitte gehen Sie im Browser einen Schritt zur\xFCck)."
      ],
      "Back": [
        "Zur\xFCck"
      ],
      "Verify": [
        "Pr\xFCfen"
      ],
      "Failed to send the verification code.": [
        "Das Versenden des Best\xE4tigungscodes hat nicht funktioniert."
      ],
      "The request was valid, but the server is refusing action.": [
        "Die Anfrage war g\xFCltig, aber der Server verweigert die Bearbeitung."
      ],
      "The backend is not aware of the specified MFA challenge.": [
        "Dieser Anwendung ist die angegebene Multi-Faktor-\xDCberpr\xFCfung nicht bekannt."
      ],
      "Code transmission failed.": [
        "Die Code-\xDCbertragung ist fehlgeschlagen."
      ],
      "Already solved.": [
        "Bereits gel\xF6st."
      ],
      "It is too early to request another transmission of the challenge.": [
        "Es muss noch gewartet werden, um eine weitere \xDCbertragung des \xDCberpr\xFCfungscodes zu verlangen."
      ],
      "Multi-factor authentication required.": [
        "Es ist eine Multi-Faktor-Authentifizierung erforderlich."
      ],
      "You need to complete all of this requirements.": [
        "Sie m\xFCssen alle dieser Bedingungen erf\xFCllen."
      ],
      "You need to complete at least one of this requirements.": [
        "Sie m\xFCssen mindestens eine dieser Bedingungen erf\xFCllen."
      ],
      "An SMS to the phone number starting with %1$s": [
        "SMS erhalten an die Telefonnummer beginnend mit %1$s"
      ],
      "An email to the address starting with %1$s": [
        "Best\xE4tigungscode erhalten an die E-Mailadresse beginnend mit %1$s"
      ],
      "You have to wait until %1$s to send a new code.": [
        "Sie m\xFCssen bis %1$s warten, damit ein neuer Best\xE4tigungscode gesendet werden kann."
      ],
      "I have a code": [
        "Ich habe bereits einen g\xFCltigen Best\xE4tigungscode"
      ],
      "Send me a message": [
        "Best\xE4tigungscode zusenden"
      ],
      "Logged in": [
        "Sie sind angemeldet"
      ],
      "Not found": [
        "Nicht gefunden"
      ],
      "Login required": [
        "Anmeldung erforderlich"
      ],
      "instance name": [
        "Name zur Anmeldung im Taler Merchant Backoffice UND zur Verwendung in Webadressen"
      ],
      "current password": [
        "Ihr Passwort f\xFCr diesen Benutzernamen"
      ],
      "Forgot password": [
        "Ich habe mein Passwort vergessen"
      ],
      "Create new account": [
        "Neues Benutzerkonto erstellen"
      ],
      "Invalid": [
        "Ung\xFCltig"
      ],
      "Doesn't have the pattern of an email": [
        "Dies entspricht keiner g\xFCltigen E-Mailadresse"
      ],
      "Should start with +": [
        "Die Nummer sollte mit + beginnen"
      ],
      "A phone number consists of numbers only": [
        "Eine Telefonnummer besteht nur aus Ziffern"
      ],
      "Invalid value": [
        "Ung\xFCltiger Wert"
      ],
      "Max 7 lines": [
        "H\xF6chstens 7 Zeilen"
      ],
      "Doesn't match": [
        "Dies stimmt nicht \xFCberein"
      ],
      "Instance created": [
        "Die Instanz wurde angelegt"
      ],
      "Unauthorized.": [
        "Unberechtigter Zugriff."
      ],
      "Conflict.": [
        "Es besteht ein Konflikt zwischen erwarteter und gesendeter Information."
      ],
      "Not found.": [
        "Die Information wurde nicht gefunden."
      ],
      "New password": [
        "Neues Passwort"
      ],
      "Next password to be used": [
        "Geben Sie hier ein neues Passwort ein"
      ],
      "Repeat password": [
        "Passwort wiederholen"
      ],
      "Confirm the same password": [
        "Bitte geben Sie das Passwort hier wiederholt ein"
      ],
      "Please complete the marked fields and choose authorization method": [
        "Sie m\xFCssen die gekennzeichneten Felder ausf\xFCllen und eine Methode der Berechtigungspr\xFCfung w\xE4hlen"
      ],
      "Confirm operation": [
        "Vorgang best\xE4tigen"
      ],
      "Check the password.": [
        "\xDCberpr\xFCfen Sie das Passwort."
      ],
      "Instance not found.": [
        "Die Instanz wurde nicht gefunden."
      ],
      "Description": [
        "Beschreibung"
      ],
      "Helps you remember where this token is being used before deleting it.": [
        "Dies hilft Ihnen zur Erinnerung, wo dieser Zugangstoken verwendet wird, bevor Sie ihn l\xF6schen."
      ],
      "Duration": [
        "G\xFCltigkeitszeitraum"
      ],
      "Time the access token will be valid.": [
        "Zeitraum, in dem der Zugangstoken g\xFCltig ist."
      ],
      "Refreshable tokens can pose a security risk!": [
        "Aktualisierbare Token k\xF6nnen ein Sicherheitsrisiko darstellen!"
      ],
      "Refreshable tokens can be refreshed before their lifetime ends, effectively giving any bearer access without expiration. Only use this if you know what you are doing and you have evaluated associated risks especially in respect to the permissions granted by the scope.": [
        "Aktualisierbare Token k\xF6nnen vor Ablauf ihrer Lebensdauer aktualisiert werden, wodurch alle ihre Besitzer effektiv eine zeitlich unbegrenzte Zugangsberechtigung haben w\xFCrde. Verwenden Sie diese daher nur, wenn Sie wissen, was Sie tun und die damit verbundenen Risiken insbesondere im Hinblick auf den zugewiesenen Einsatzbereich der Berechtigung richtig einsch\xE4tzen k\xF6nnen."
      ],
      "Scope": [
        "Einsatzbereich des Zugangstokens"
      ],
      "The scope defines the set of permissions for the access token. Refreshable tokens has the permission to extend the expiration time.": [
        "Der Einsatzbereich definiert die Berechtigungen eines Zugangstokens. Bei aktualisierbaren Token k\xF6nnen ihre Besitzer die G\xFCltigkeitszeit unbegrenzt verl\xE4ngern."
      ],
      "Allows all operations without limit.": [
        "Dies erlaubt alle Vorg\xE4nge ohne jede Einschr\xE4nkung."
      ],
      "Allows all operations to read information.": [
        "Dies erm\xF6glicht allen Vorg\xE4ngen, Informationen zu lesen."
      ],
      "Allows the creation of orders and checking of payment status.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen sowie die \xDCberpr\xFCfung des Zahlungsstatus."
      ],
      "Allows the creation of orders, checking of payment status and inventory locking.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen, die \xDCberpr\xFCfung des Zahlungsstatus und das Sperren des Lagerbestands."
      ],
      "Allows the creation of orders, checking of payment status and refunds.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen, die \xDCberpr\xFCfung des Zahlungsstatus sowie R\xFCckerstattungen."
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen, das Sperren des Lagerbestands sowie R\xFCckerstattungen."
      ],
      "Allows all operations to read information with extendable expiration time.": [
        "Dies erm\xF6glicht allen Vorg\xE4ngen das Einlesen von Informationen mit einer verl\xE4ngerbaren Verfallszeit."
      ],
      "Allows the creation of orders and checking of payment status with extendable expiration time.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen sowie die \xDCberpr\xFCfung des Zahlungsstatus mit verl\xE4ngerbarer Verfallszeit."
      ],
      "Allows the creation of orders, checking of payment status and inventory locking with extendable expiration time.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen, die \xDCberpr\xFCfung des Zahlungsstatus sowie das Sperren des Lagerbestands mit verl\xE4ngerbarer Verfallszeit."
      ],
      "Allows the creation of orders, checking of payment status and refunds with extendable expiration time.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen, die \xDCberpr\xFCfung des Zahlungsstatus sowie R\xFCckerstattungen mit verl\xE4ngerbarer Verfallszeit."
      ],
      "Allows the creation of orders, checking of payment status, inventory locking and refunds with extendable expiration time.": [
        "Dies erm\xF6glicht die Erstellung von Bestellungen und Barrechnungen, die \xDCberpr\xFCfung des Zahlungsstatus sowie das Sperren des Lagerbestands und R\xFCckerstattungen mit verl\xE4ngerbarer Verfallszeit."
      ],
      "All (refreshable)": [
        "Alle (aktualisierbar)"
      ],
      "Spa": [
        "SPA"
      ],
      "Spa (refreshable)": [
        "SPA (aktualisierbar)"
      ],
      "Choose one": [
        "Bitte w\xE4hlen Sie eine Option aus"
      ],
      "Read only": [
        "Schreibgesch\xFCtzt"
      ],
      "All": [
        "Alle"
      ],
      "Order simple": [
        "Einfache Bestellungen"
      ],
      "Order Point-of-Sale": [
        "Bestellungen an der Verkaufsstelle (POS)"
      ],
      "Order management": [
        "Verwaltung der Bestellungen"
      ],
      "Order full": [
        "Komplexe Bestellungen"
      ],
      "Read only (refreshable)": [
        "Schreibgesch\xFCtzt (aktualisierbar)"
      ],
      "Order simple (refreshable)": [
        "Einfache Bestellungen (aktualisierbar)"
      ],
      "Order Point-of-Sale (refreshable)": [
        "Bestellungen an der Verkaufsstelle (aktualisierbar)"
      ],
      "Order management (refreshable)": [
        "Verwaltung der Bestellungen (aktualisierbar)"
      ],
      "Order full (refreshable)": [
        "Komplexe Bestellungen (aktualisierbar)"
      ],
      "Current password": [
        "Aktuelles Passwort dieser Instanz"
      ],
      "Please complete the marked fields": [
        "Bitte f\xFCllen Sie die Felder mit verpflichtenden Angaben aus"
      ],
      "Access token created": [
        "Zugangstoken wurde erzeugt"
      ],
      "Copy the value of the access token and save it, as you cannot retrieve it again.": [
        "Kopieren und speichern Sie den Zugangstoken sofort, da Sie ihn nicht erneut abrufen k\xF6nnen."
      ],
      "Token": [
        "Zugangstoken"
      ],
      "This token will never expire": [
        "Dieser Zugangstoken wird nie ung\xFCltig"
      ],
      "This token will be available until %1$s": [
        "Dieser Zugangstoken ist verf\xFCgbar bis %1$s"
      ],
      "No 'admin' instance configured yet.": [
        "Es wurde noch keine Instanz 'admin' konfiguriert."
      ],
      "Create an 'admin' instance to begin using the merchant backoffice.": [
        "Erstellen Sie eine Instanz 'admin', um Merchant Backoffice erstmalig zu verwenden."
      ],
      "Create access token": [
        "Zugangstoken erzeugen"
      ],
      "Load more devices before the first one": [
        "Mehr Ger\xE4te vor ein erstes Ger\xE4t laden"
      ],
      "Load first page": [
        "Erste Seite laden"
      ],
      "Created at": [
        "Erstellungsdatum"
      ],
      "Expires at": [
        "Wird ung\xFCltig zum Zeitpunkt"
      ],
      "Never": [
        "Nie"
      ],
      "Remove this access token": [
        "Diesen Zugangstoken entfernen"
      ],
      "Delete": [
        "L\xF6schen"
      ],
      "Load more devices after the last one": [
        "Mehr Ger\xE4te nach dem letzten Ger\xE4t laden"
      ],
      "Load next page": [
        "N\xE4chste Seite laden"
      ],
      "There are no active sessions yet, add one by pressing the + sign": [
        "Es bestehen noch keine aktiven Sitzungen, Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "Forbidden.": [
        "Unzul\xE4ssig."
      ],
      "Delete access token": [
        "Zugangstoken l\xF6schen"
      ],
      "Deleting an access token cannot be undone.": [
        "Das L\xF6schen eines Zugangstoken kann nicht r\xFCckg\xE4ngig gemacht werden."
      ],
      "This is not a valid Bitcoin address.": [
        "Dies ist keine g\xFCltige Bitcoin-Adresse."
      ],
      "This is not a valid Ethereum address.": [
        "Dies ist keine zul\xE4ssige Ethereum-Adresse."
      ],
      "This is not a valid host.": [
        "Dies ist kein g\xFCltiger Host."
      ],
      "IBANs usually have more than 4 digits.": [
        "Eine IBAN besteht normalerweise aus mehr als 4 Ziffern."
      ],
      "IBANs usually have fewer than 34 digits.": [
        "Eine IBAN besteht normalerweise aus weniger als 34 Ziffern."
      ],
      "The IBAN's country code could not be retrieved.": [
        "Der IBAN-L\xE4ndercode wurde nicht gefunden."
      ],
      "The IBAN is invalid because the checksum is wrong.": [
        "Die IBAN-Nummer ist ung\xFCltig, die Pr\xFCfsumme ist falsch."
      ],
      "This account is not allowed.": [
        "Dieses Konto ist nicht zul\xE4ssig."
      ],
      "None of the supported wire method of the server are currently supported by this app. Server settings: %1$s": [
        "Keine der g\xE4ngigen \xDCbertragungsmethoden des Servers wird zurzeit von dieser Anwendung unterst\xFCtzt. Server-Einstellungen: %1$s"
      ],
      "Wire method": [
        "\xDCberweisungsmethode"
      ],
      "Select the method you want to use to transfer your earnings to your business account.": [
        "W\xE4hlen Sie die Methode, mit der Sie Ihre Einnahmen an Ihr Gesch\xE4ftskonto \xFCberweisen m\xF6chten."
      ],
      "Select a wire method...": [
        "W\xE4hlen Sie eine \xDCberweisungsmethode..."
      ],
      "Routing": [
        "Bankleitzahlensystem"
      ],
      "Routing number": [
        "Bankleitzahl"
      ],
      "Account": [
        "Konto"
      ],
      "Account number": [
        "Kontonummer"
      ],
      "Code": [
        "Code"
      ],
      "Business Identifier Code": [
        "Business Identifier Code (BIC)"
      ],
      "International Bank Account Number": [
        "IBAN (Internationale Bankkontonummer)"
      ],
      "your bank account number, e.g. DE12 0000 1111 2222 3333 00": [
        "Ihre IBAN, z.B. CH12 0000 1111 2222 3333 00"
      ],
      "Unified Payment Interface": [
        "Unified Payment Interface (UPI)"
      ],
      "Bitcoin protocol": [
        "Bitcoin-Protokoll"
      ],
      "Ethereum protocol": [
        "Ethereum-Protokoll"
      ],
      "Interledger protocol": [
        "Interledger-Protokoll"
      ],
      "Enter the data without a scheme. A subpath may be included:": [
        "Geben Sie die Daten ohne Schema ein. Ein Pfad kann darin enthalten sein:"
      ],
      "Name of the account holder": [
        "Name des Kontoinhabers"
      ],
      "John Doe": [
        "Max Muster"
      ],
      "Legal name of the person holding the account.": [
        "Rechtsg\xFCltiger Name des Kontoinhabers."
      ],
      "Invalid url": [
        "ung\xFCltige URL"
      ],
      "URL must end with a '/'": [
        "Die URL muss mit '/' enden"
      ],
      "URL must not contain params": [
        "Die URL darf keine Parameter enthalten"
      ],
      "URL must not hash param": [
        "Die URL darf keinen Hash-Parameter enthalten"
      ],
      "Conflict": [
        "Konflikt"
      ],
      'Server replied with "bad request".': [
        'Der Server antwortete mit "bad request".'
      ],
      "Unauthorized, check credentials.": [
        "Unzul\xE4ssiger Vorgang, bitte die Zugangsdaten (Token, Anmeldename, Passwort) \xFCberpr\xFCfen."
      ],
      "The endpoint does not seem to be a Taler Revenue API.": [
        "Die Gegenstelle scheint keine Taler-Einzahlungsschnittstelle (Taler Revenue API) zu sein."
      ],
      "The request was made correctly, but the bank's server did not respond with the appropriate value for 'credit_account', so we cannot confirm that it is the same bank account.": [
        "Die Anfrage war richtig gestellt, doch der Server der Bank antwortete nicht mit dem passenden Wert (f\xFCr 'credit_account'), so dass wir nicht best\xE4tigen k\xF6nnen, dass es sich um dasselbe Bankkonto handelt."
      ],
      "Unsupported type of account": [
        "Nicht unterst\xFCtzter Kontotyp"
      ],
      "Account:": [
        "Konto:"
      ],
      "If the bank supports Taler Revenue API then you can add the endpoint URL below to keep the revenue information in sync.": [
        "Wenn die kontof\xFChrende Bank die Schnittstelle Taler Revenue API unterst\xFCtzt, k\xF6nnen Sie die untenstehende Gegenstelle (endpoint URL) hinzuf\xFCgen, um die Umsatzinformationen zwischen dem Backend und dem Bankkonto zu synchronisieren."
      ],
      "Endpoint URL": [
        "Gegenstelle (endpoint URL)"
      ],
      "From where the merchant can download information about incoming wire transfers to this account": [
        "Informationsquelle zu eingehenden \xDCberweisungen auf dieses Bankkonto"
      ],
      "Auth type": [
        "Authentifizierungsmethode"
      ],
      "Choose the authentication type for the account info URL": [
        "W\xE4hlen Sie die Authentifizierungsmethode zum Zugriff auf das Bankkonto, um die eingegangenen \xDCberweisungen abzufragen"
      ],
      "Without authentication": [
        "Ohne Authentifizierung"
      ],
      "With username and password": [
        "Mit Benutzername und Passwort"
      ],
      "With token": [
        "Mit Zugangstoken"
      ],
      "Do not change": [
        "Lassen Sie dies unver\xE4ndert"
      ],
      "Username to access the account information.": [
        "Benutzername f\xFCr den Zugriff auf die Kontoinformationen."
      ],
      "Password to access the account information.": [
        "Passwort f\xFCr den Zugriff auf die Kontoinformationen."
      ],
      "Access token to access the account information.": [
        "Zugangstoken f\xFCr den Zugriff auf die Kontoinformationen."
      ],
      "Match": [
        "\xDCbereinstimmung"
      ],
      "Check where the information match against the server info.": [
        "Pr\xFCfung der \xDCbereinstimmung dieser Information und der Daten auf dem Server."
      ],
      "Compare info from server with account form": [
        "Vergleich der Information des Servers mit den Bankkonto-Formulardaten"
      ],
      "Test": [
        "Test"
      ],
      "Server didn't like the request we made.": [
        "Der Server war mit unserer Anfrage nicht einverstanden."
      ],
      "Account details": [
        "Kontodetails"
      ],
      "Check if the information matches the server info.": [
        "Bitte pr\xFCfen Sie, ob die Information und die Daten auf dem Server \xFCbereinstimmen."
      ],
      "The bank account has been deleted.": [
        "Die Verkn\xFCpfung mit dem Bankkonto wurde erfolgreich gel\xF6scht."
      ],
      "Bank accounts": [
        "Bankkonten"
      ],
      "Add new account": [
        "Neues Konto hinzuf\xFCgen"
      ],
      "Owner's name": [
        "Name des Kontoinhabers"
      ],
      "Delete selected accounts from the database": [
        "L\xF6schen ausgew\xE4hlter Konten von der Datenbank"
      ],
      "There are no accounts yet, add more pressing the + sign": [
        "Es sind noch keine Konten angelegt, Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "You must provide a bank account to receive payments.": [
        "Sie m\xFCssen ein Bankkonto verkn\xFCpfen, um Einnahmen an dieses Konto zu \xFCberweisen."
      ],
      "Without this information, you cannot create new payment orders that are transferred to a bank account.": [
        "Ohne diese Informationen k\xF6nnen Sie keine neuen Zahlungsabwicklungen erstellen, die an ein Bankkonto \xFCberwiesen werden."
      ],
      "Invalid. Please use only letters and numbers.": [
        "Ung\xFCltig. Bitte geben Sie nur Buchstaben und Zahlen ein."
      ],
      "Name": [
        "Name"
      ],
      "Category name": [
        "Kategoriename"
      ],
      "Category deleted": [
        "Die Kategorie wurde gel\xF6scht"
      ],
      "Add new devices": [
        "Neue Ger\xE4te hinzuf\xFCgen"
      ],
      "ID": [
        "ID"
      ],
      "Total products": [
        "Gesamte Artikelmenge"
      ],
      "Delete selected category from the database": [
        "Ausgew\xE4hlte Kategorie aus der Datenbank l\xF6schen"
      ],
      "There are no categories yet, add more pressing the + sign": [
        "Es gibt noch keine Kategorien. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "Id:": [
        "ID:"
      ],
      "Name of the category": [
        "Benennung der Kategorie (Warengruppe)"
      ],
      "Products": [
        "Artikel"
      ],
      "Image": [
        "Bild"
      ],
      "Load more products after the last one": [
        "Weitere Artikel nach dem letzten laden"
      ],
      "There are no products in this category.": [
        "Es befinden sich keine Artikel in dieser Kategorie."
      ],
      "Account's KYC status": [
        "KYC-Status des Bankkontos"
      ],
      "Exchange": [
        "Zahlungsdienstleister"
      ],
      "Status": [
        "Status"
      ],
      "Ok": [
        "OK"
      ],
      "Action required": [
        "Sie m\xFCssen noch eine Aktion ausf\xFChren (eine \xDCberweisung oder eine KYC-Pr\xFCfung ist erforderlich)"
      ],
      "Warning": [
        "Warnung"
      ],
      "Error": [
        "Fehler"
      ],
      "Bank account verification required.": [
        "Eine Bankkonto-Verifizierung ist erforderlich."
      ],
      "More information required.": [
        "Weitere Informationen sind erforderlich."
      ],
      "Awaiting for account review.": [
        "Es wird auf die \xDCberpr\xFCfung des Kontos gewartet."
      ],
      "Ready": [
        "Bereit zum Einsatz"
      ],
      "Syncing...": [
        "Die Daten werden eingelesen..."
      ],
      "Payment service internal error. Contact administrator or check again later.": [
        "Interner Fehler des Zahlungsdiensts. Wenden Sie sich an den Administrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Payment service timeout. Contact administrator or check again later.": [
        "Zeit\xFCberschreitung der Antwort des Zahlungsdiensts. Wenden Sie sich an den Administrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Payment service unreachable. Contact administrator or check again later.": [
        "Der Zahlungsdienst ist nicht erreichbar. Wenden Sie sich an den Administrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Incompatible core banking system.": [
        "Inkompatibles Kernbankensystem."
      ],
      "Backend internal error. Contact administrator or check again later.": [
        "Interner Fehler im Backend. Wenden Sie sich an den Administrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Payment service response is invalid. Contact administrator or check again later.": [
        "Die Antwort des Zahlungsdiensts ist ung\xFCltig. Wenden Sie sich an den Administrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Http Status": [
        "HTTP-Status"
      ],
      "No pending kyc verification!": [
        "Eine KYC-Pr\xFCfung ist nicht (mehr) ausstehend!"
      ],
      "Wire account is invalid": [
        "Das zur \xDCberweisung gew\xE4hlte Bankkonto ist ung\xFCltig"
      ],
      'The backend service responded with "%1$s" which is invalid.': [
        'Das Backend hat mit "%1$s" geantwortet, was ung\xFCltig ist.'
      ],
      "No exchange keys": [
        "Es sind keine Schl\xFCssel des Exchange vorhanden"
      ],
      "The backend service is still synchronizing with the payment service provider.": [
        "Das Backend synchronisiert zurzeit noch mit dem Zahlungsdienst."
      ],
      "Can't wire from this account": [
        "Von diesem Bankkonto kann keine \xDCberweisung erfolgen"
      ],
      "The core banking system cannot perform wire transfers between the payment service provider's accounts and your bank accounts. Thus you cannot use this payment service provider. the exchange.": [
        "Das Kernbankensystem kann keine \xDCberweisungen zwischen dem Bankkonto des Zahlungsdiensts und Ihren Bankkonten durchf\xFChren. Daher k\xF6nnen Sie diesen Zahlungsdienst nicht nutzen. Der Taler Exchange."
      ],
      "KYC Required": [
        "Es ist eine KYC-Pr\xFCfung erforderlich"
      ],
      "The payment provider requires more information.": [
        "Der Zahlungsdienst verlangt mehr Informationen."
      ],
      "Click here to complete this step.": [
        "Klicken Sie hier, um diesen Schritt abzuschlie\xDFen."
      ],
      "We are waiting for the access token to be present. Check again later.": [
        "Bitte warten Sie auf die Erzeugung des Zugangstokens und pr\xFCfen diese zu einem sp\xE4teren Zeitpunkt erneut."
      ],
      "Awaiting AML review": [
        "Eine AML-\xDCberpr\xFCfung steht noch aus"
      ],
      "This account is can't be used. The payment service provider is doing manual review of the account information.": [
        "Dieses Konto kann nicht verwendet werden. Der Zahlungsdienst f\xFChrt eine manuelle \xDCberpr\xFCfung der Kontoinformationen durch."
      ],
      "This account is ready to be used with the payment service provider.": [
        "Dieses Konto ist bereit f\xFCr die Nutzung mit diesem Zahlungsdienst."
      ],
      "Logic bug": [
        "Logischer Fehler"
      ],
      "The backend service detected an internal error, contact the system administrator or check again later.": [
        "Das Backend hat einen internen Fehler festgestellt. Wenden Sie sich an den Systemadministrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Internal error": [
        "Interner Systemfehler"
      ],
      "The payment service provider detected an internal error, contact the system administrator or check again later.": [
        "Interner Fehler des Zahlungsdiensts. Wenden Sie sich an den Administrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "The backend service couldn't contact the payment service provider due to a timeout, contact the system administrator or check again later.": [
        "Das Backend konnte den Zahlungsdienst aufgrund einer Zeit\xFCberschreitung nicht kontaktieren. Wenden Sie sich an den Systemadministrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Unable to reach the payment service provider, contact the system administrator or check again later.": [
        "Der Zahlungsdienst ist zurzeit nicht erreichbar. Wenden Sie sich an den Administrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "The payment service provider replied with a invalid status, contact the system administrator or check again later.": [
        "Der Zahlungsdienst hat mit einem ung\xFCltigen Status geantwortet. Wenden Sie sich an den Systemadministrator oder pr\xFCfen Sie sp\xE4ter erneut."
      ],
      "Change value to unknown date": [
        "Wert auf unbekanntes Datum \xE4ndern"
      ],
      "Enter description or id": [
        "Geben Sie die Beschreibung oder eine ID ein"
      ],
      "No match found for that description or ID.": [
        "F\xFCr diese Beschreibung oder ID wurde keine \xDCbereinstimmung gefunden."
      ],
      "You must enter a valid product identifier.": [
        "Sie m\xFCssen eine g\xFCltige Artikelkennung eingeben."
      ],
      "Quantity must be greater than 0!": [
        "Die Menge muss gr\xF6\xDFer als 0 sein!"
      ],
      "This quantity exceeds remaining stock. Currently, only %1$s units remain unreserved in stock.": [
        "Diese Menge \xFCberschreitet den Bestand. Momentan sind nur %1$s St\xFCck im Inventar ohne Vorbestellung."
      ],
      "Search product": [
        "Artikel suchen"
      ],
      "Quantity": [
        "Menge"
      ],
      "How many products will be added": [
        "Wie viele Artikel hinzugef\xFCgt werden"
      ],
      "Add from inventory": [
        "Vom Bestand hinzuf\xFCgen"
      ],
      "This product has %1$s applicable taxes configured.": [
        "F\xFCr diesen Artikel ist ein Steuersatz von %1$s angelegt."
      ],
      "No taxes configured for this product.": [
        "F\xFCr diesen Artikel sind keine Steuern angelegt."
      ],
      "Amount": [
        "Betrag"
      ],
      "Taxes can be in currencies that differ from the main currency used by the merchant.": [
        "Steuern k\xF6nnen in einer anderen W\xE4hrung sein als die Hauptw\xE4hrung, welche f\xFCr Verk\xE4ufe normalerweise verwendet wird."
      ],
      "Enter the currency and value separated by a colon (e.g., &quot;USD:2.3&quot;).": [
        "Geben Sie W\xE4hrung und Betrag mit einem Doppelpunkt getrennt ein, z.B. &quot;CH:2.3&quot; oder &quot;EUR:2,3&quot;."
      ],
      "Legal name of the tax, e.g. VAT or import duties.": [
        "Offizielle Bezeichnung der Steuer, z.B. Mehrwertsteuer oder Einfuhrsteuer."
      ],
      "Add tax to the tax list": [
        "Steuersatz zur Liste der Steuern hinzuf\xFCgen"
      ],
      "Describe and add a product that is not in the inventory list": [
        "Bisher nicht inventarisierte Artikel beschreiben und hinzuf\xFCgen"
      ],
      "Add custom product": [
        "Individuellen Artikel hinzuf\xFCgen"
      ],
      "Complete information of the product": [
        "Informationen zum Artikel vervollst\xE4ndigen"
      ],
      "Must be a number.": [
        "Dieser Wert muss eine Nummer sein."
      ],
      "Must be greater than 0.": [
        "Dieser Wert muss gr\xF6\xDFer als Null sein."
      ],
      "Photo of the product.": [
        "Abbildung des Artikels."
      ],
      "Full product description.": [
        "Vollst\xE4ndige Artikelbeschreibung."
      ],
      "Unit name": [
        "Art der Mengeneinheit"
      ],
      "Name of the product unit.": [
        "Bezeichnung der Einheit des Artikels."
      ],
      "Price per unit": [
        "Preis pro Mengeneinheit"
      ],
      "Amount in the current currency.": [
        "Betrag in der aktuellen W\xE4hrung."
      ],
      "How many products will be added.": [
        "Wie viele Artikel hinzugef\xFCgt werden."
      ],
      "Taxes": [
        "Steuern"
      ],
      "Total price": [
        "Gesamtpreis"
      ],
      "Must be greater than 0": [
        "Dieser Wert muss gr\xF6\xDFer als Null sein"
      ],
      "Must have a refund deadline": [
        "Es sollte eine R\xFCckerstattungsfrist geben"
      ],
      "Auto refund can't be after refund deadline": [
        "Eine automatische R\xFCckerstattung kann nicht nach der R\xFCckerstattungsfrist erfolgen"
      ],
      "Must be in the future": [
        "Dieser Wert sollte in der Zukunft liegen"
      ],
      "Either fulfillment url or fulfillment message must be specified.": [
        "Es m\xFCssen entweder fulfillment-URL oder fulfillment message angegeben werden."
      ],
      "is not a valid URL": [
        "ist keine g\xFCltige URL"
      ],
      'Product with ID "%1$s" is out of stock.': [
        'Der Artikel mit der ID "%1$s" ist nicht auf Lager.'
      ],
      "No exchange would accept a payment because of KYC requirements.": [
        "Aufgrund vorgegebener KYC-Erfordernisse w\xFCrde kein Zahlungsdienstleister eine Zahlung annehmen."
      ],
      "Manage products in order": [
        "Artikel in der Bestellung verwalten"
      ],
      "%1$s products with a total price of %2$s.": [
        "%1$s Artikel zu einer Gesamtsumme von %2$s."
      ],
      "Manage list of products in the order.": [
        "Liste der Artikel in der Bestellung verwalten."
      ],
      "Remove this product from the order.": [
        "Diesen Artikel aus der Bestellung entfernen."
      ],
      "Products price sum": [
        "Summe der Preise der Artikel"
      ],
      "Total product price added up": [
        "Gesamtpreis des Artikels hinzugerechnet"
      ],
      "Order price": [
        "Preis eines Artikels"
      ],
      "Amount to be paid by the customer": [
        "Vom Kunden zu zahlender Betrag"
      ],
      "Final order price": [
        "Endg\xFCltiger Bestellwert"
      ],
      "Summary": [
        "Zusammenfassung"
      ],
      "Title of the order to be shown to the customer": [
        "Bezeichnung der Bestellung, die den Kunden angezeigt wird"
      ],
      "Shipping and fulfillment": [
        "Versand und Lieferung"
      ],
      "Delivery date": [
        "Lieferdatum"
      ],
      "Deadline for physical delivery assured by the merchant.": [
        "Vom H\xE4ndler zugesicherte Zustellfrist."
      ],
      "Delivery location": [
        "Lieferadresse"
      ],
      "Address where the products will be delivered": [
        "Zustelladresse der Artikel"
      ],
      "Fulfillment URL": [
        "Adresse digitaler Dienstleistung (Fulfillment-URL)"
      ],
      "URL to which the user will be redirected after successful payment.": [
        "URL der von Kunden zu besuchenden Adresse nach erfolgter Bezahlung."
      ],
      "Fulfillment message": [
        "Fulfillment message (Nachricht an Zahlende \xFCber bereitgestelltes digitales Gut)"
      ],
      "Message shown to the customer after paying for the order.": [
        "Die Nachricht, die den Kunden nach der Bezahlung der Bestellung angezeigt wird."
      ],
      "Taler payment options": [
        "Taler-Zahlungsoptionen"
      ],
      "Override default Taler payment settings for this order": [
        "\xDCberschreibe die vorgegebenen Zahlungseinstellungen im Taler-System f\xFCr diese Bestellung"
      ],
      "Payment time": [
        "Zahlungsdatum"
      ],
      "Time for the customer to pay before the offer expires. Inventory products will be reserved until this deadline. The time starts after the order is created.": [
        "Zeitraum, im dem die K\xE4ufer bezahlen m\xFCssen, bevor das Vertragsangebot verf\xE4llt. Inventarisierte Artikel f\xFCr ein aktuelles Angebot werden bis zum Ablauf dieses Zeitraums reserviert. Der Zeitablauf beginnt mit der Erstellzeit der Bestellung."
      ],
      "Default": [
        "Standardeinstellung"
      ],
      "Refund time": [
        "R\xFCckerstattungsdatum"
      ],
      "Time while the order can be refunded by the merchant. Time starts after the order is created.": [
        "Zeitdauer, in der die Bestellung vom Verk\xE4ufer erstattet werden kann. Die Zeitdauer beginnt mit der Erstellzeit der Bestellungen."
      ],
      "Wire transfer time": [
        "Zeitpunkt der \xDCberweisung"
      ],
      "Time for the exchange to make the wire transfer. Time starts after the order is created.": [
        "Zeitraum, nach dessen Ablauf der Zahlungsdienst \xDCberweisungen ausf\xFChrt. Die Zeitdauer beginnt mit der Erstellzeit der Bestellung."
      ],
      "Auto-refund time": [
        "Zeitdauer f\xFCr automatische R\xFCckerstattungen"
      ],
      "Time until which the wallet will automatically check for refunds without user interaction.": [
        "Zeit, bis zu der das Wallet automatisch und ohne Benutzerinteraktion auf R\xFCckerstattungen pr\xFCft."
      ],
      "Maximum fee": [
        "Maximale Geb\xFChrenh\xF6he"
      ],
      "Maximum fees the merchant is willing to cover for this order. Higher deposit fees must be covered in full by the consumer.": [
        "Die maximale Geb\xFChrenh\xF6he, die ein Verk\xE4ufer f\xFCr eine Bestellung zu \xFCbernehmen bereit ist. Einl\xF6segeb\xFChren, die \xFCber dieser liegen, sind komplett von den K\xE4ufern zu \xFCbernehmen."
      ],
      "Create token": [
        "Token erzeugen"
      ],
      "If the order ID is easy to guess, the token will prevent other users from claiming the order.": [
        "Selbst wenn die ID einer Bestellung leicht zu erraten w\xE4re, verhindert der Token, dass die Bestellung von Unberechtigten \xFCbernommen werden k\xF6nnte."
      ],
      "Minimum age required": [
        "Mindestalter erforderlich"
      ],
      "Any value greater than 0 will limit the coins able be used to pay this contract. If empty the age restriction will be defined by the products": [
        "Eine Zahl gr\xF6\xDFer als 0 schr\xE4nkt ein, welches e-Geld zur Bezahlung dieses Kaufvertrags verwendet werden kann. Bleibt das Feld leer, werden Altersbeschr\xE4nkungen durch die Artikel selbst festgelegt"
      ],
      "Minimum age defined by the products is %1$s": [
        "Das Mindestalter f\xFCr den Erwerb dieses Artikels ist %1$s"
      ],
      "No product with age restriction in this order": [
        "Es befindet sich kein Artikel mit einer Altersbeschr\xE4nkung in dieser Bestellung"
      ],
      "Additional information": [
        "Zus\xE4tzliche Informationen"
      ],
      "Custom information to be included in the contract for this order.": [
        "Individuelle Informationen, die in den Kaufvertrag f\xFCr diese Bestellung aufgenommen werden sollen."
      ],
      "You must enter a value in JavaScript Object Notation (JSON).": [
        "Sie m\xFCssen einen Wert im Format JavaScript Object Notation (JSON) eingeben."
      ],
      "remove": [
        "Entfernen"
      ],
      "Custom field name": [
        "Benutzerdefinierter Feldname"
      ],
      "Disabled": [
        "Deaktiviert"
      ],
      "No deadline": [
        "Keine Zahlungsfrist"
      ],
      "Deadline at %1$s": [
        "Frist bis %1$s"
      ],
      "Select date to show nearby orders": [
        "Bitte w\xE4hlen Sie ein Datum, um die Bestellungen dieses Zeitraums anzuzeigen"
      ],
      "Only show unpaid orders": [
        "Nur unbezahlte Bestellungen anzeigen"
      ],
      "Only show paid orders": [
        "Nur bezahlte Bestellungen anzeigen"
      ],
      "Paid": [
        "Bezahlt"
      ],
      "Only show orders with refunds": [
        "Nur Bestellungen mit erfolgter R\xFCckerstattung anzeigen"
      ],
      "Refunded": [
        "R\xFCckerstattet"
      ],
      "Show only paid orders for which the wire transfer is still pending.": [
        "Nur bezahlte Bestellungen anzeigen, f\xFCr welche die \xDCberweisung noch aussteht."
      ],
      "Not wired": [
        "Nicht \xFCberwiesen"
      ],
      "Only display orders that have already been transferred by the payment service provider": [
        "Nur Bestellungen anzeigen, die vom Zahlungsdienst bereits an das H\xE4ndlerbankkonto \xFCberwiesen wurden"
      ],
      "Completed": [
        "\xDCberwiesen"
      ],
      "Remove all filters": [
        "Alle Filter aufheben"
      ],
      "Clear date filter": [
        "Datumsfilter aufheben"
      ],
      "Jump to date (%1$s)": [
        "Springe zum Datum (%1$s)"
      ],
      "Gone.": [
        "Nicht mehr vorhanden."
      ],
      "UnavailableForLegalReasons.": [
        "Aus rechtlichen Gr\xFCnden nicht verf\xFCgbar."
      ],
      "Jump to order with the given product ID": [
        "Zur Bestellung mit angegebener Artikel-ID gehen"
      ],
      "Order id": [
        "Bestellnummer"
      ],
      "Instance unknown": [
        "Unbekannte Instanz"
      ],
      "Order unknown": [
        "Unbekannte Bestellung"
      ],
      "Create order": [
        "Bestellung erzeugen"
      ],
      "The instance doesn't exist": [
        "Die Instanz existiert nicht"
      ],
      "The order doesn't exist": [
        "Die Bestellung existiert nicht"
      ],
      "Date": [
        "Datum"
      ],
      "Refund": [
        "R\xFCckerstattung"
      ],
      "copy url": [
        "URL kopieren"
      ],
      "Load more orders after the last one": [
        "Nach der letzten Bestellung weitere laden"
      ],
      "No orders have been found matching your query!": [
        "Es wurden keine Bestellungen gefunden, die zu Ihrer Abfrage passen!"
      ],
      "Duplicated": [
        "Dupliziert"
      ],
      "This value exceeds the refundable amount.": [
        "Dieser Wert \xFCbersteigt den erstattungsf\xE4higen Betrag."
      ],
      "Forbidden": [
        "Verboten"
      ],
      "Gone": [
        "Nicht mehr vorhanden"
      ],
      "There are pending KYC requirements.": [
        "Es sind noch KYC-Erfordernisse zu erf\xFCllen (Pr\xFCfung der wirtschaftlich Berechtigten)."
      ],
      "refund": [
        "R\xFCckerstattung"
      ],
      "Reason": [
        "Buchungsvermerk"
      ],
      "Amount to be refunded": [
        "Zu erstattender Betrag"
      ],
      "Max refundable:": [
        "Maximale R\xFCckerstattung:"
      ],
      "Choose one...": [
        "Bitte w\xE4hlen Sie eine Option aus..."
      ],
      "Requested by the customer": [
        "Vom K\xE4ufer angefordert"
      ],
      "Other": [
        "Sonstiges"
      ],
      "Why this order is being refunded": [
        "Grund f\xFCr die Erstattung dieser Bestellung"
      ],
      "More information to give context": [
        "Weitere Informationen f\xFCr eine genauere Beschreibung"
      ],
      "now": [
        "jetzt"
      ],
      "This is when the time for making refund has been expired.": [
        "Zu diesem Zeitpunkt ist die R\xFCckerstattungsfrist abgelaufen."
      ],
      "This is when the time for making the payment has been expired.": [
        "Zu diesem Zeitpunkt ist die Zahlungsfrist abgelaufen."
      ],
      "This is when the wire transfer is going to be executed by the payment service provider.": [
        "Zu diesem Zeitpunkt ist der Aufschub f\xFCr die \xDCberweisung durch den Zahlungsdienst abgelaufen."
      ],
      "This wire transfer has been notified by the payment service provider.": [
        "Diese \xDCberweisung wurde vom Zahlungsdienst gemeldet."
      ],
      "This refund is waiting to be claimed by the customer.": [
        "Diese R\xFCckerstattung kann vom K\xE4ufer beantragt werden."
      ],
      "This refund has been claimed by the customer.": [
        "Diese R\xFCckerstattung wurde vom K\xE4ufer verlangt."
      ],
      "The current moment in time.": [
        "Der jetzige Zeitpunkt."
      ],
      "This refund can't be claimed because the wire transfer has already be made.": [
        "Diese R\xFCckerstattung kann nicht mehr beantragt werden, da die Zahlung des K\xE4ufers bereits \xFCberwiesen wurde."
      ],
      "Contract terms": [
        "Vertragsbedingungen"
      ],
      "Human-readable description of the whole purchase": [
        "F\xFCr Menschen lesbare Information \xFCber den gesamten Kaufvorgang"
      ],
      "Total price for the transaction": [
        "Gesamtpreis der Transaktion"
      ],
      "URL for this purchase": [
        "URL f\xFCr diesen Kauf"
      ],
      "Max fee": [
        "H\xF6chster Geb\xFChrenbetrag"
      ],
      "Maximum total deposit fee accepted by the merchant for this contract": [
        "Maximale Einl\xF6sege\xFChr, die ein Verk\xE4ufer f\xFCr diesen Kaufvertrag selbst tragen will"
      ],
      "Time when this contract was generated": [
        "Zeitpunkt, zu dem der Kaufvertrag erstellt wurde"
      ],
      "Refund deadline": [
        "R\xFCckerstattungsfrist"
      ],
      "After this deadline has passed, no refunds will be accepted.": [
        "Nach Ablauf dieser Frist werden keine R\xFCckerstattungen mehr akzeptiert."
      ],
      "Payment deadline": [
        "Zahlungsfrist"
      ],
      "After this deadline, the merchant won't accept payments for the contract": [
        "Nach Ablauf dieser Frist akzeptiert ein Verk\xE4ufer f\xFCr den Vertrag keinen Rechnungsausgleich (Bezahlung durch K\xE4ufer)"
      ],
      "Wire transfer deadline": [
        "\xDCberweisungsfrist"
      ],
      "Transfer deadline for the exchange": [
        "Frist der \xDCberweisung vom Zahlungsdienstleister an das Bankkonto"
      ],
      "Time indicating when the order should be delivered": [
        "Zeitangabe, wann die Bestellung geliefert werden soll"
      ],
      "Where the order will be delivered": [
        "Zustelladresse der Bestellung"
      ],
      "Auto-refund delay": [
        "Zeitliche Verz\xF6gerung der automatischen R\xFCckerstattung"
      ],
      "How long the wallet should try to get an automatic refund for the purchase": [
        "So lange sollte ein Wallet versuchen, eine automatische R\xFCckerstattung auf den Kaufvertrag zu erhalten"
      ],
      "Extra info": [
        "Weitere Informationen"
      ],
      "Extra data that is only interpreted by the merchant frontend": [
        "Zus\xE4tzliche Daten, die nur im Verk\xE4ufer-Frontend verarbeitet werden"
      ],
      "order created": [
        "Die Bestellung wurde angelegt"
      ],
      "pay deadline": [
        "Zahlungsfrist"
      ],
      "refund deadline": [
        "R\xFCckerstattungsfrist"
      ],
      "delivery": [
        "Lieferung"
      ],
      "Order": [
        "Bestellung"
      ],
      "Claimed": [
        "Schon f\xFCr jemanden reserviert"
      ],
      "Claimed at": [
        "Reserviert f\xFCr"
      ],
      "Timeline": [
        "Zeitleiste"
      ],
      "Payment details": [
        "Zahlungsdetails"
      ],
      "Order status": [
        "Bestellstatus"
      ],
      "Unpaid": [
        "Noch nicht bezahlt"
      ],
      "Product list": [
        "Artikelliste"
      ],
      "refund missed: %1$s": [
        "Nicht verlangte R\xFCckerstattung: %1$s"
      ],
      "refund missed: %1$s: %2$s": [
        "Nicht verlangte R\xFCckerstattung: %1$s: %2$s"
      ],
      "refund created: %1$s": [
        "R\xFCckerstattung gew\xE4hrt: %1$s"
      ],
      "refund created: %1$s: %2$s": [
        "R\xFCckerstattung gew\xE4hrt: %1$s: %2$s"
      ],
      "refund taken: %1$s": [
        "Dem K\xE4ufer zugegangene R\xFCckerstattung: %1$s"
      ],
      "refund taken: %1$s: %2$s": [
        "Dem K\xE4ufer zugegangene R\xFCckerstattung: %1$s: %2$s"
      ],
      "The contract terms has a v1 order without choices_index.": [
        "Die Vertragsbedingungen haben eine Bestellung der Version v1 (ohne choices_index)."
      ],
      "The contract terms has a v1 order with a bad choices_index.": [
        "Die Vertragsbedingungen haben eine Bestellung der Version v1 (mit einem schlechten choices_index)."
      ],
      "wired %1$s": [
        "\xFCberwiesen %1$s"
      ],
      "wire deadline": [
        "\xDCberweisungszeitpunkt"
      ],
      "Wired": [
        "\xDCberwiesen"
      ],
      "Refund order": [
        "Zahlbetrag der Bestellung zur\xFCckerstatten"
      ],
      "Not refundable": [
        "Nicht erstattungsf\xE4hig"
      ],
      "Next event in": [
        "N\xE4chster Event in"
      ],
      "The order was wired.": [
        "Die Zahlung wurde \xFCberwiesen."
      ],
      "Bank processing can take a few business days, depending on your bank.": [
        "Die Verarbeitung der \xDCberweisung kann je nach Finanzinstitut einige Bankarbeitstage dauern."
      ],
      "Refunded amount": [
        "R\xFCckerstattungsbetrag"
      ],
      "Refund taken": [
        "Erhaltener R\xFCckerstattungsbetrag"
      ],
      "Status URL": [
        "Status-URL"
      ],
      "Refund URI": [
        "R\xFCckerstattungs-URI"
      ],
      "Pay at": [
        "Zahlungs-URL"
      ],
      "Order status URL": [
        "Bestellstatus-URL"
      ],
      "Payment URI": [
        "Zahlungs-URI"
      ],
      "Unknown order status. This is an error, please contact the administrator.": [
        "Unbekannter Bestellstatus. Dies ist ein Fehler, bitte kontaktieren Sie den Administrator."
      ],
      "Invalid. Please enter letters and numbers only.": [
        "Ung\xFCltig. Bitte geben Sie nur Buchstaben und Zahlen ein."
      ],
      "Just letters and numbers from 2 to 7": [
        "Nur Buchstaben und Zahlen von 2 bis 7"
      ],
      "The size of the key must be 32 characters": [
        "Die Schl\xFCssell\xE4nge muss 32 Zeichen betragen"
      ],
      "Internal ID on the system": [
        "Interne ID auf dem System"
      ],
      "Useful to identify the device physically": [
        "Dies ist n\xFCtzlich, um das Ger\xE4t physisch zu identifizieren"
      ],
      "Verification algorithm": [
        "\xDCberpr\xFCfungsalgorithmus"
      ],
      "Algorithm used to verify transactions in offline mode": [
        "Algorithmus zum \xDCberpr\xFCfen von Transaktionen im Offline-Modus"
      ],
      "Device key": [
        "Ger\xE4teschl\xFCssel (Device key)"
      ],
      "Be sure to choose a password that is hard to guess, or use the random generator.": [
        "Das Passwort muss schwer zu erraten sein - andernfalls verwenden Sie den Zufallszahlengenerator."
      ],
      "Your device needs to match exactly the same value": [
        "Ihr Ger\xE4t muss genau das gleiche Ergebnis anzeigen"
      ],
      "Generate random secret key": [
        "Zuf\xE4lligen privaten Schl\xFCssel erzeugen"
      ],
      "Random": [
        "Zufallsbasiert"
      ],
      "Create another": [
        "Weitere erzeugen"
      ],
      "You can scan the next QR code with your device or save the key before continuing.": [
        "Sie k\xF6nnen den n\xE4chsten QR-Code mit Ihrem Ger\xE4t scannen oder den Schl\xFCssel speichern, bevor Sie fortfahren."
      ],
      "OTP devices": [
        "OTP-Ger\xE4te"
      ],
      "Delete selected devices from the database": [
        "Ausgew\xE4hlte Ger\xE4te aus der Datenbank entfernen"
      ],
      "There are no devices to list yet, add more by pressing the + sign": [
        "Es gibt noch keine Ger\xE4te. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "Template id is unknown": [
        "Die Vorlagen-ID ist unbekannt"
      ],
      "The provided information is inconsistent with the current state of the template": [
        "Die bereitgestellten Informationen stimmen nicht mit dem aktuellen Stand der Vorlage \xFCberein"
      ],
      "Device:": [
        "Ger\xE4t:"
      ],
      "Algorithm to use to verify transactions in offline mode": [
        "Algorithmus zum \xDCberpr\xFCfen von Transaktionen im Offline-Modus"
      ],
      "Not modified": [
        "Unver\xE4ndert"
      ],
      "Be sure to be very hard to guess or use the random generator": [
        "Das Passwort muss schwer zu erraten sein oder aus einem Zufallszahlengenerator stammen"
      ],
      "Your device need to have exactly the same value": [
        "Ihr Ger\xE4t muss genau das gleiche Ergebnis anzeigen"
      ],
      "Change key": [
        "Schl\xFCssel \xE4ndern"
      ],
      "Is not the same": [
        "ungleich"
      ],
      'You are updating the password for the instance with ID "%1$s"': [
        'Sie erneuern nun das Passwort f\xFCr die Instanz mit der ID "%1$s"'
      ],
      "In order to verify that you have access.": [
        "Um zu \xFCberpr\xFCfen, ob der Zugang gew\xE4hrt wird."
      ],
      "Confirm change": [
        "\xC4nderung best\xE4tigen"
      ],
      "Testing password change": [
        "Die Passwort\xE4nderung wird \xFCberpr\xFCft"
      ],
      "Password changed": [
        "Passwort ge\xE4ndert"
      ],
      "The current password is wrong.": [
        "Das aktuelle Passwort ist falsch."
      ],
      "No enough rights to change the password.": [
        "Es bestehen keine ausreichenden Rechte, um das Passwort zu \xE4ndern."
      ],
      "Account not found.": [
        "Das Bankkonto wurde nicht gefunden."
      ],
      "Add element to the list": [
        "Ein Element zur Liste hinzuf\xFCgen"
      ],
      "Click here to configure the product's stock. If left as is, the backend will not control stock.": [
        "Klicken Sie hier, um den Lagerbestand des Artikels zu konfigurieren. Falls dieses Programm den Bestand nicht \xFCberwachen soll, lassen Sie seine Grundeinstellung unver\xE4ndert."
      ],
      "Manage stock": [
        "Bestand verwalten"
      ],
      "This product has been configured without stock control": [
        "Dieser Artikel wurde ohne Bestandskontrolle angelegt"
      ],
      "Infinite": [
        "Unbegrenzt"
      ],
      "Shrinkage cannot exceed the current stock and incoming supplies (maximum %1$s)": [
        "Der Schwund kann nicht den aktuellen Bestand und die Menge der Zulieferung \xFCbersteigen (maximal %1$s)"
      ],
      "Incoming": [
        "Erwartete \xDCberweisungen"
      ],
      "Lost": [
        "Schwund (Inventurdifferenz)"
      ],
      "Current": [
        "Aktuell"
      ],
      "Remove stock control for this product": [
        "Bestandskontrolle f\xFCr diesen Artikel entfernen"
      ],
      "without stock": [
        "ohne Bestand"
      ],
      "Next restock": [
        "N\xE4chstes Auff\xFCllen des Bestands"
      ],
      "Warehouse address": [
        "Adresse des Lagers"
      ],
      "Invalid amount": [
        "Kein g\xFCltiger Betrag"
      ],
      "Product identification to use in URLs (for internal use only).": [
        "Identifikation zur Nutzung in URLs (nur f\xFCr den internen Gebrauch)."
      ],
      "Illustration of the product for customers.": [
        "Abbildung des Artikels f\xFCr Kunden."
      ],
      "Product name.": [
        "Artikelbezeichnung."
      ],
      "Product description for customers.": [
        "Artikelbeschreibung, die Kunden zu sehen bekommen."
      ],
      "Age restriction": [
        "Altersbeschr\xE4nkung"
      ],
      "Is this product restricted for customer below certain age?": [
        "Besteht f\xFCr diesen Artikel eine Verkaufsbeschr\xE4nkung unterhalb eines bestimmten Alters?"
      ],
      "Minimum age of the customer": [
        "Mindestalter von K\xE4ufern"
      ],
      "Unit describing quantity of product sold (e.g. 2 kilograms, 5 liters, 3 items, 5 meters) for customers.": [
        "Die den K\xE4ufern angezeigte Mengeneinheit des Artikels im Verkauf (z.B. 2 Kilogramm, 5 Liter, 3 St\xFCck, 5 Meter)."
      ],
      "Example: kg, items or liters": [
        "Beispiel: kg, St\xFCck oder Liter"
      ],
      "Sale price for customers, including taxes, for above units of the product.": [
        "Verkaufspreis an Endkunden, inklusive Steuern, f\xFCr jede oben genannte Mengeneinheit des Artikels."
      ],
      "Stock": [
        "Bestand"
      ],
      "Inventory for products with finite supply (for internal use only).": [
        "Lagerbestand f\xFCr Artikel mit begrenztem Vorrat (nur f\xFCr den internen Gebrauch)."
      ],
      "Taxes included in the product price, exposed to customers.": [
        "Im Artikelpreis enthaltene Steuern, die den K\xE4ufern angezeigt werden."
      ],
      "Search by category description or id": [
        "Nach Kategoriebeschreibung oder -ID suchen"
      ],
      "Categories where this product will be listed on.": [
        "Kategorien, in denen dieser Artikel aufgef\xFChrt sein wird."
      ],
      "Add product to inventory": [
        "Artikel zum Lagerbestand hinzuf\xFCgen"
      ],
      "Product not found.": [
        "Der Artikel wurde nicht gefunden."
      ],
      "This change was made without outdated info.": [
        "Diese \xC4nderung wurde ohne bisher bekannte Daten durchgef\xFChrt."
      ],
      "Sales": [
        "Sales"
      ],
      "Sold": [
        "Verkauft"
      ],
      "Free": [
        "Kostenlos"
      ],
      "Go to product update page": [
        "Zur Seite der Artikelaktualisierung gehen"
      ],
      "Update": [
        "Aktualisieren"
      ],
      "Remove this product from the database": [
        "Diesen Artikel aus der Datenbank entfernen"
      ],
      "Update the product with new price": [
        "Artikel mit neuem Preis aktualisieren"
      ],
      "Update product with new price": [
        "Artikel mit neuem Verkaufspreis aktualisieren"
      ],
      "Confirm update": [
        "Aktualisierung best\xE4tigen"
      ],
      "lost can't be greater that current + incoming (max %1$s)": [
        "Der Schwund kann nicht gr\xF6\xDFer sein als der aktuelle Bestand und die Menge der Zulieferung (max %1$s)"
      ],
      "Add more elements to the inventory": [
        "Mehr Elemente zum Lagerbestand hinzuf\xFCgen"
      ],
      "Report elements lost in the inventory": [
        "Aus dem Lagerbestand verlorene Elemente melden"
      ],
      "New price for the product": [
        "Neuer Preis f\xFCr den Artikel"
      ],
      "There are values with errors.": [
        "Einige Werte sind fehlerhaft."
      ],
      "Update product with new stock and price": [
        "Artikel mit neuem Bestand und Preis aktualisieren"
      ],
      "There are no products to list yet, add more by pressing the + sign": [
        "Es gibt noch keine Artikel. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "Product (ID: %1$s) has been deleted": [
        "Der Artikel (ID: %1$s) wurde gel\xF6scht"
      ],
      "Jump to product with the given product ID": [
        "Zum Artikel mit der angegebenen Artikel-ID springen"
      ],
      "Product id": [
        "Artikel-ID"
      ],
      "Delete product": [
        "Artikel l\xF6schen"
      ],
      'Delete the product "%1$s"': [
        'Artikel "%1$s" l\xF6schen'
      ],
      "If you delete the product named %1$s (ID: %2$s ), the stock and related information will be lost": [
        "Wenn Sie den Artikel mit der Bezeichnung %1$s (ID: %2$s) l\xF6schen, gehen seine Bestandsdaten und alle zugeh\xF6rigen Informationen verloren"
      ],
      "Deleting a product cannot be undone.": [
        "Das L\xF6schen eines Artikels kann nicht r\xFCckg\xE4ngig gemacht werden."
      ],
      "Product id:": [
        "Artikel-ID:"
      ],
      "Must be greater that 0": [
        "Dieser Wert muss gr\xF6\xDFer als Null sein"
      ],
      "Too short": [
        "Zu kurz"
      ],
      "You don't have enough permissions.": [
        "Sie haben keine ausreichende Berechtigung."
      ],
      "There is a template with this identifier, please choose another one.": [
        "Es gibt bereits eine Vorlage mit dieser Kennung, bitte w\xE4hlen Sie eine andere."
      ],
      "Identifier": [
        "Eindeutige Kennung"
      ],
      "Name of the template in URLs.": [
        "Name der Vorlage in URLs."
      ],
      "Template name": [
        "Name der Vorlage"
      ],
      "Describe what this template stands for": [
        "Beschreiben Sie, wof\xFCr diese Vorlage stehen soll"
      ],
      "Order summary": [
        "Zusammenfassung der Bestellung"
      ],
      "If specified here, this template will create orders with the same summary": [
        "Falls angegeben, wird diese Vorlage Bestellungen mit der gleichen Zusammenfassung erzeugen"
      ],
      "Summary is editable": [
        "Die Zusammenfassung kann umgeschrieben werden"
      ],
      "Allow the user to change the summary.": [
        "Den Nutzern die Bearbeitung der Zusammenfassung erlauben."
      ],
      "If specified here, this template will create orders with the same price": [
        "Diese Vorlage wird Bestellungen zum stets gleichen Preis erzeugen"
      ],
      "Amount is editable": [
        "Der Betrag l\xE4sst sich ver\xE4ndern"
      ],
      "Allow the user to select the amount to pay.": [
        "Den Nutzern die Auswahl des Zahlbetrags erlauben."
      ],
      "Currency is editable": [
        "Die W\xE4hrung ist ver\xE4nderbar"
      ],
      "Allow the user to change currency.": [
        "Den Nutzern die \xC4nderung der W\xE4hrung erlauben."
      ],
      "Supported currencies": [
        "Unterst\xFCtzte W\xE4hrungen"
      ],
      "Supported currencies: %1$s": [
        "Unterst\xFCtzte W\xE4hrungen: %1$s"
      ],
      "Minimum age": [
        "Mindestalter"
      ],
      "Is this contract restricted to some age?": [
        "Verlangt dieser Kaufvertrag ein bestimmtes Mindestalter?"
      ],
      "Payment timeout": [
        "Die Zahlungsfrist ist verstrichen nach"
      ],
      "How much time the customer has to complete the payment once the order was created.": [
        "Wie lange der K\xE4ufer Zeit hat, um zu bezahlen, nachdem die Bestellung angelegt wurde."
      ],
      "OTP device": [
        "OTP-Ger\xE4t"
      ],
      "Use to verify transactions in offline mode.": [
        "Verwenden zum \xDCberpr\xFCfen von Transaktionen im Offline-Modus."
      ],
      "No OTP device.": [
        "Kein OTP-Ger\xE4t."
      ],
      "Add one first": [
        "F\xFCgen Sie zuerst ein erstes hinzu"
      ],
      "No device": [
        "Kein Ger\xE4t"
      ],
      "Add new templates": [
        "Neue Vorlagen hinzuf\xFCgen"
      ],
      "Load more templates before the first one": [
        "Mehrere Vorlagen vor der ersten laden"
      ],
      "Delete selected templates from the database": [
        "Ausgew\xE4hlte Vorlagen aus der Datenbank l\xF6schen"
      ],
      "Use template to create new order": [
        "Vorlage verwenden, um eine neue Bestellung anzulegen"
      ],
      "Generate a QR code for the template.": [
        "QR-Code f\xFCr die Vorlage erstellen."
      ],
      "Show QR": [
        "QR-Code anzeigen"
      ],
      "Load more templates after the last one": [
        "Weitere Vorlagen nach der letzten laden"
      ],
      "There are no templates to list yet, add more by pressing the + sign": [
        "Es gibt noch keine Vorlagen. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "Jump to template with the given template ID": [
        "Zur Vorlage mit der angegebenen Vorlagen-ID springen"
      ],
      "Template identification": [
        "Vorlagenidentifikation"
      ],
      "Delete template": [
        "Vorlage l\xF6schen"
      ],
      'Delete the template "%1$s"': [
        'Vorlage "%1$s" l\xF6schen'
      ],
      "If you delete the template %1$s (ID: %2$s) you may loose information": [
        "Wenn Sie die Vorlage %1$s (ID: %2$s) l\xF6schen, k\xF6nnen Information verlorengehen"
      ],
      "Deleting a template cannot be undone.": [
        "Das L\xF6schen einer Vorlage kann nicht r\xFCckg\xE4ngig gemacht werden."
      ],
      "Print": [
        "Drucken"
      ],
      "The template configuration needs to be fixed.": [
        "Die Detaileinstellung der Vorlage muss korrigiert werden."
      ],
      "The currency of the template is %1$s and is not in the list of supported currencies.": [
        "Die W\xE4hrung der Vorlage lautet %1$s, steht aber nicht in der Liste der unterst\xFCtzten W\xE4hrungen."
      ],
      "If specified, this template will create order with the same summary": [
        "Falls angegeben, wird diese Vorlage Bestellungen mit der gleichen Zusammenfassung erzeugen"
      ],
      "If specified, this template will create orders with the same price": [
        "Diese Vorlage wird Bestellungen zum stets gleichen Preis erzeugen"
      ],
      "How much time has the customer to complete the payment once the order was created.": [
        "Dies bestimmt die Zeitdauer zur Bezahlung nach einer erfolgten Bestellung."
      ],
      "An amount is required": [
        "Ein Betrag ist erforderlich"
      ],
      "An order summary is required": [
        "Eine Zusammenfassung der Bestellung ist erforderlich"
      ],
      'No more stock for product with ID "%1$s".': [
        'Es gibt keinen Bestand f\xFCr den Artikel mit der ID "%1$s".'
      ],
      "New order from template": [
        "Neue Bestellung aus einer Vorlage erzeugen"
      ],
      "Amount of the order": [
        "Summe der Bestellung"
      ],
      "Slug": [
        "Slug (benutzer- und suchmaschinenfreundlicher Text als Bestandteil einer URL)"
      ],
      "Token family slug to use in URLs (for internal use only)": [
        "Slug der Token-Familie zur Benutzung in URLs (nur f\xFCr interne Zwecke)"
      ],
      "Kind": [
        "Art"
      ],
      "Token family kind": [
        "Art der Tokenfamilie"
      ],
      "User-readable token family name": [
        "Benutzerfreundliche Bezeichnung der Tokenfamilie"
      ],
      "Token family description for customers": [
        "Tokenfamilien-Beschreibung f\xFCr Kunden"
      ],
      "Valid After": [
        "G\xFCltig nach"
      ],
      "Token family can issue tokens after this date": [
        "Die Tokenfamilie kann nach diesem Datum Token ausgeben"
      ],
      "Valid Before": [
        "G\xFCltig vor"
      ],
      "Token family can issue tokens until this date": [
        "Die Tokenfamilie kann bis zu diesem Datum Token ausgeben"
      ],
      "Validity Granularity": [
        "Rundungsgenauigkeit des G\xFCltigkeitszeitraums"
      ],
      "Rounding granularity for the start validity of keys, must be 1 minute, 1 hour, 1 day, 1 week, 30 days or 90 days": [
        "Rundungsgenauigkeit f\xFCr Schl\xFCssel, diese muss 1 Minute, 1 Stunde, 1 Tag, 1 Woche, 30 Tage oder 90 Tage betragen"
      ],
      "Validity duration of a issued token": [
        "Zeitraum der G\xFCltigkeit eines ausgestellten Tokens"
      ],
      "Token Families": [
        "Tokenfamilien"
      ],
      "Add token family": [
        "Tokenfamilie hinzuf\xFCgen"
      ],
      "Go to token family update page": [
        "Zur Aktualisierungsseite f\xFCr Tokenfamilien gehen"
      ],
      "Remove this token family from the database": [
        "Diese Token aus der Datenbank entfernen"
      ],
      "There are no token families yet, add the first one by pressing the + sign.": [
        "Es sind noch keine Token-Familien angelegt. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen."
      ],
      "Token family has been deleted.": [
        "Die Tokenfamilie wurde gel\xF6scht."
      ],
      "Delete token family": [
        "Tokenfamilie l\xF6schen"
      ],
      'Delete the token family "%1$s"': [
        'Tokenfamilie "%1$s" l\xF6schen'
      ],
      "If you delete the %1$s token family (Slug: %2$s), all issued tokens will become invalid.": [
        "Wenn Sie die Tokenfamilie %1$s (SLUG: %2$s) l\xF6schen, werden alle ausgestellten Token ung\xFCltig."
      ],
      "Deleting a token family %1$s .": [
        "Tokenfamilie %1$s wird gel\xF6scht."
      ],
      "can't be undone": [
        "Das kann nicht r\xFCckg\xE4ngig gemacht werden"
      ],
      "Token Family: %1$s": [
        "Tokenfamilie: %1$s"
      ],
      "Incoming wire transfers": [
        "Erwartete \xDCberweisungen"
      ],
      "Load more wire transfers preceding the first one": [
        "Weitere \xDCberweisungen vor der ersten laden"
      ],
      "Expected credit": [
        "Erwartete Gutschrift"
      ],
      "Confirmed": [
        "Best\xE4tigt"
      ],
      "Validated": [
        "\xDCberpr\xFCft (validiert)"
      ],
      "Executed on": [
        "Ausf\xFChrungsdatum"
      ],
      "yes": [
        "ja"
      ],
      "no": [
        "nein"
      ],
      "never": [
        "nie"
      ],
      "unknown": [
        "unbekannt"
      ],
      "Load more transfers after the last one": [
        "Mehr \xDCberweisungen nach der letzten laden"
      ],
      "Verified wire transfers": [
        "\xDCberpr\xFCfte und best\xE4tigte \xDCberweisungen"
      ],
      "Verified": [
        "Verifizierte (best\xE4tigte) \xDCberweisungen"
      ],
      "There are no transfers to list yet, add more by pressing the + sign": [
        "Es gibt noch keine \xDCberweisungen anzuzeigen. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "All accounts": [
        "Alle Konten"
      ],
      "Filter by account address": [
        "Filtern nach Kontoadressen"
      ],
      "Only display transfers that have already been transferred to your bank account by the payment service provider.": [
        "Nur \xDCberweisungen anzeigen, die vom Zahlungsdienst bereits an Ihr Bankkonto \xFCberwiesen wurden."
      ],
      "Only show wire transfers confirmed by the merchant": [
        "Nur \xDCberweisungen anzeigen, die vom Verk\xE4ufer als auf dem Bankkonto eingegangen best\xE4tigt sind"
      ],
      "Wire transfer already confirmed.": [
        "Die \xDCberweisung ist bereits best\xE4tigt."
      ],
      "I have received the wire transfer": [
        "Ich habe die \xDCberweisung erhalten"
      ],
      "Confirm the wire transfer": [
        "Best\xE4tigen der \xDCberweisung"
      ],
      "The wire transfer has been sent and should be in your bank account in any time. You can manually confirm the reception usign the information below.": [
        "Die \xDCberweisung wurde ausgef\xFChrt und sollte in K\xFCrze auf Ihrem Bankkonto eingehen. Mithilfe der unten stehenden Informationen k\xF6nnen Sie die Gutschrift manuell \xFCberpr\xFCfen und best\xE4tigen."
      ],
      "Time": [
        "Zeit"
      ],
      "Transfer ID": [
        "Transfer-ID"
      ],
      "It's not the same.": [
        "Dies stimmt nicht \xFCberein."
      ],
      'You are deleting the instance with ID "%1$s"': [
        'Sie l\xF6schen nun die Instanz mit der ID "%1$s"'
      ],
      "Instance": [
        "Instanz"
      ],
      "Write the instance name to confirm the deletion": [
        "Schreiben Sie den Namen der Instanz hier hinein, um die L\xF6schung zu best\xE4tigen"
      ],
      "Purge": [
        "Vollst\xE4ndig verwerfen"
      ],
      "All the data will be fully deleted, otherwise only the access will be removed.": [
        "Alle Daten werden vollst\xE4ndig gel\xF6scht, andernfalls wird nur der Zugang entfernt."
      ],
      "DELETE": [
        "L\xD6SCHEN"
      ],
      "Pay delay can't be greater than wire transfer delay": [
        "Die Zahlungsfrist f\xFCr K\xE4ufer darf nicht l\xE4nger sein als der zeitliche Aufschub der \xDCberweisung"
      ],
      "Instance id": [
        "Instanz-ID"
      ],
      "Delete this instance": [
        "L\xF6schung dieser Instanz"
      ],
      "Must be one of '%1$s'": [
        "Muss eines von '%1$s' sein"
      ],
      "URL is invalid": [
        "URL ist ung\xFCltig"
      ],
      "Webhook ID to use": [
        "zu verwendende Webhook-ID"
      ],
      "Event": [
        "Event"
      ],
      "Payment": [
        "Zahlung"
      ],
      "Order settled": [
        "Die Bestellung wurde bezahlt und dem Konto gutgeschrieben"
      ],
      "Category added": [
        "Kategorie hinzugef\xFCgt"
      ],
      "Category updated": [
        "Die Kategorie wurde aktualisiert"
      ],
      "Inventory added": [
        "Der Bestand wurde hinzugef\xFCgt"
      ],
      "Inventory updated": [
        "Der Bestand wurde aktualisiert"
      ],
      "Inventory deleted": [
        "Bestand gel\xF6scht"
      ],
      "The event of the webhook: why the webhook is used": [
        "Event (Anlass, Bedingung) f\xFCr einen Webhook, um durch diesen nachfolgende Prozesse auszul\xF6sen"
      ],
      "Method": [
        "Methode"
      ],
      "GET": [
        "GET"
      ],
      "POST": [
        "POST"
      ],
      "PUT": [
        "PUT"
      ],
      "PATCH": [
        "PATCH"
      ],
      "HEAD": [
        "HEAD"
      ],
      "Method used by the webhook": [
        "Methode, die der Webhook verwendet"
      ],
      "URL": [
        "URL"
      ],
      "URL of the webhook where the customer will be redirected": [
        "URL des Webhooks, zu der die K\xE4ufer nach der Bezahlung weitergeleitet werden"
      ],
      "The text below supports the %1$s template engine. Any string between %2$s and %3$s will be replaced with the value of the corresponding variable.": [
        "Der folgende Text unterst\xFCtzt %1$s Vorlagen-Ersteller. Jede Zeichenkette zwischen %2$s und %3$s wird durch den Wert der entsprechenden Variablen ersetzt."
      ],
      "For example, %1$s will be replaced with the the order's price.": [
        "Beispielsweise wird %1$s durch den Bestellwert ersetzt."
      ],
      "The short list of variables are:": [
        "Einige der Variablen sind:"
      ],
      "order's description": [
        "Beschreibung der Bestellung"
      ],
      "order's price": [
        "Gesamtpreis der Bestellung"
      ],
      "order's unique identification": [
        "Laufende Nummer der Bestellung"
      ],
      "the amount that was being refunded": [
        "Der erstattete Betrag"
      ],
      "the reason entered by the merchant staff for granting the refund": [
        "Der vom Personal des Verk\xE4ufers angegebene Grund f\xFCr die Gew\xE4hrung einer Erstattung"
      ],
      "time of the refund in nanoseconds since 1970": [
        "Zeitpunkt der Erstattung in Nanosekunden seit 1970"
      ],
      "Http header": [
        "HTTP-Header"
      ],
      "Header template of the webhook": [
        "Vorlage f\xFCr den Webhook-Header"
      ],
      "Http body": [
        "HTTP-Body"
      ],
      "Body template used by the webhook.": [
        "Vorlage f\xFCr den Webhook-Body."
      ],
      "Webhook deleted successfully": [
        "Webhook wurde erfolgreich gel\xF6scht"
      ],
      "Add new webhooks": [
        "Neue Webhooks hinzuf\xFCgen"
      ],
      "Load more webhooks before the first one": [
        "Mehr Webhooks vor dem ersten laden"
      ],
      "Event type": [
        "Event-Typ"
      ],
      "Delete selected webhook from the database": [
        "Ausgew\xE4hlten Webhook aus der Datenbank l\xF6schen"
      ],
      "Load more webhooks after the last one": [
        "Mehr Webhooks nach dem letzten laden"
      ],
      "There are no webhooks to list yet, add more by pressing the + sign": [
        "Es gibt noch keine Webhooks. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen"
      ],
      "Webhook updated": [
        "Webhook aktualisiert"
      ],
      "Header": [
        "Header"
      ],
      "Body": [
        "Body"
      ],
      "Body template used by the webhook": [
        "Vorlage, die f\xFCr den Webhook-Body verwendet wird"
      ],
      "There is another instance with this username.": [
        "Es besteht bereits eine andere Instanz mit diesem Benutzernamen."
      ],
      "Self provision": [
        "Erstellen Sie Ihr eigenes Taler-Portal"
      ],
      "Phone": [
        "Telefon"
      ],
      "Contact phone number": [
        "Telefonnummer f\xFCr allgemeinen Kontakt"
      ],
      "I agree": [
        "AGB annehmen"
      ],
      "Accept the Terms of service": [
        "Allgemeine Gesch\xE4ftsbedingungen annehmen"
      ],
      "You must accept the Terms of service to continue.": [
        "Um fortzufahren m\xFCssen Sie zuerst die Allgemeinen Gesch\xE4ftsbedingungen dieser Verwaltungssoftware akzeptieren."
      ],
      "Create": [
        "Anlegen"
      ],
      "Password reset": [
        "Passwort zur\xFCcksetzen"
      ],
      'Resetting access to the instance "%1$s"': [
        'Zur\xFCcksetzen des Zugangs zur Instanz "%1$s"'
      ],
      "Reset": [
        "Zur\xFCcksetzen"
      ],
      "Language": [
        "Sprache"
      ],
      "Date format": [
        "Datumsformat"
      ],
      "How the date is going to be displayed": [
        "Wie das Datum angezeigt wird"
      ],
      "Merchant type": [
        "Merchant-Typus"
      ],
      "Simplify UI based on the user usage.": [
        "Passen Sie hier den Komplexit\xE4tsgrad der Benutzeroberfl\xE4che f\xFCr verschiedene Benutzer an."
      ],
      "Expert user": [
        "Erfahrener Benutzer (expert user)"
      ],
      "Taler developer": [
        "Entwickler (Taler developer)"
      ],
      "Unattended in-person offline vending": [
        "Personalloser Verkauf, z.B. Hofl\xE4den mit Vertrauenskassen (Unattended in-person offline vending)"
      ],
      "In-person online point-of-sale with inventory": [
        "Verkaufsstelle mit Internetzugang und Bestandserfassung (In-person online point-of-sale with inventory)"
      ],
      "Digital publishing": [
        "Vertrieb digitaler G\xFCter (Digital publishing)"
      ],
      "E-commerce site": [
        "Onlineshops (E-commerce site)"
      ],
      "Developer mode": [
        "Entwicklermodus"
      ],
      "Only use developer mode if you know how the application works. Some features enabled in this mode are still under testing.": [
        "Verwenden Sie den Entwicklermodus nur, wenn Sie wissen, wie die Anwendung funktioniert. Einige der in diesem Modus aktivierten Funktionen befinden sich noch in der Testphase."
      ],
      "Your browser does not support the canvas element.": [
        "Ihr Browser unterst\xFCtzt nicht das Canvas-Element."
      ],
      "Bad gateway": [
        "502 Bad gateway"
      ],
      "Service unavailable": [
        "Der Dienst ist nicht verf\xFCgbar"
      ],
      "Select date from which to show statistics": [
        "Bitte w\xE4hlen Sie das Datum, ab dem die Statistik angezeigt werden soll"
      ],
      "Show chart": [
        "Diagramm anzeigen"
      ],
      "Orders chart": [
        "Bestellungen als Diagramm"
      ],
      "Show table": [
        "Tabelle anzeigen"
      ],
      "Orders table": [
        "Tabellarische \xDCbersicht der Bestellungen"
      ],
      "Start from (%1$s)": [
        "Start von (%1$s)"
      ],
      "Orders statistics": [
        "Statistik der Bestellungen"
      ],
      "Since": [
        "Seit"
      ],
      "No order statistics yet.": [
        "Es wurde noch keine Bestellstatistik erstellt."
      ],
      "# of orders since": [
        "Anzahl der Bestellungen seit"
      ],
      "hour": [
        "Stunde"
      ],
      "day": [
        "Tag"
      ],
      "week": [
        "Woche"
      ],
      "weeks": [
        "Wochen"
      ],
      "month": [
        "Monat"
      ],
      "months": [
        "Monate"
      ],
      "quarter": [
        "Quartal"
      ],
      "quarters": [
        "Quartale"
      ],
      "years": [
        "Jahre"
      ],
      "Revenue chart": [
        "Umsatzdiagramm"
      ],
      "Revenue table": [
        "Tabellarische \xDCbersicht der Ums\xE4tze"
      ],
      "Revenue statistics filter": [
        "Umsatzstatistik-Filter"
      ],
      "Time range": [
        "Zeitraum"
      ],
      "Select time range to group dataset by": [
        "W\xE4hlen Sie den Zeitraum f\xFCr die Darstellung der Daten"
      ],
      "Select the number of ranges to include in data set": [
        "W\xE4hlen Sie die Anzahl der Zeitr\xE4ume f\xFCr die Datens\xE4tze, die angezeigt werden sollen"
      ],
      "Currency": [
        "W\xE4hrung"
      ],
      "Select the currency to show statistics for": [
        "Bitte w\xE4hlen Sie die W\xE4hrung, f\xFCr die Statistiken angezeigt werden sollen"
      ],
      "Revenue statistics over the past %1$s %2$s for currency '%3$s'": [
        "Umsatzstatistik der letzten %1$s %2$s f\xFCr W\xE4hrung '%3$s'"
      ],
      "Start time": [
        "Zeitraum"
      ],
      "No revenue statistics yet.": [
        "Es wurde noch keine Umsatzstatistik erstellt."
      ],
      "Revenue": [
        "Ums\xE4tze (Einnahmen)"
      ],
      "Welcome!": [
        "Willkommen im Taler Merchant Backoffice!"
      ],
      "The application is in a unexpected state and can't recover from here. You can report the problem to the developers at %1$s": [
        "Die Anwendung befindet sich in einem unerwarteten Betriebszustand. Bitte seien Sie so nett und melden das Problem den Entwicklern unter %1$s"
      ],
      "You need to associate a bank account to receive revenue.": [
        "Sie m\xFCssen ein Bankkonto verkn\xFCpfen, um Einnahmen aus \xDCberweisungen auf dieses Konto zu erhalten."
      ],
      "Without this the merchant backend will refuse to create new orders.": [
        "Ohne dies wird Taler Merchant Backend keine neuen Bestellungen erzeugen."
      ],
      "Hide for today": [
        "F\xFCr heute nicht anzeigen"
      ],
      "KYC verification needed": [
        "Eine KYC-\xDCberpr\xFCfung ist notwendig (Know-Your-Customer)"
      ],
      "Some transfers are on hold until the KYC process is completed. Visit the KYC section in the left-hand menu for more information.": [
        "Einige \xDCberweisungen werden bis zum Abschluss des KYC-Prozesses zur\xFCckgehalten. Bitte gehen Sie zum KYC-Status im Hauptmen\xFC links f\xFCr weitere Informationen."
      ],
      "The request reached a timeout, check your connection.": [
        "Die Anfrage hat ihr Zeitlimit erreicht, \xFCberpr\xFCfen Sie bitte Ihre Internetverbindung."
      ],
      "The request was cancelled.": [
        "Die Anfrage wurde abgebrochen."
      ],
      "Too many requests were made to the server, and this action was throttled.": [
        "Es wurden gleichzeitig zu viele Anfragen an denselben Server gestellt, daher ist diese Aktion gedrosselt worden."
      ],
      "The server's response was malformed.": [
        "Die Antwort des Servers auf die Anfrage war fehlerhaft."
      ],
      "Could not complete the request due to a network problem.": [
        "Die Anfrage konnte aufgrund eines Netzwerkproblems nicht abgeschlossen werden."
      ],
      "Unexpected request error.": [
        "Unerwarteter Fehler bei der Anfrage."
      ],
      "Unexpected error.": [
        "Unerwarteter Fehler."
      ],
      "Add new instance": [
        "Neue Instanz hinzuf\xFCgen"
      ],
      "Edit": [
        "Bearbeiten"
      ],
      "Change password": [
        "Passwort \xE4ndern"
      ],
      "There are no instances yet. Add one by pressing the '+' sign.": [
        "Es gibt noch keine Instanzen. Sie k\xF6nnen diese mit dem +-Zeichen hinzuf\xFCgen."
      ],
      'Instance "%1$s" (ID: %2$s) has been deleted.': [
        'Instanz "%1$s" (ID: %2$s) wurde gel\xF6scht.'
      ],
      "Only show active instances": [
        "Zeige nur aktive Instanzen"
      ],
      "Active": [
        "Aktiv"
      ],
      "Only show deleted instances": [
        "Zeige nur gel\xF6schte Instanzen"
      ],
      "Deleted": [
        "Gel\xF6scht"
      ],
      "Show all instances": [
        "Zeige alle Instanzen"
      ],
      "Delete instance": [
        "Instanz l\xF6schen"
      ],
      'Delete the instance "%1$s"': [
        'Instanz "%1$s" l\xF6schen'
      ],
      "If you delete the instance named %1$s (ID: %2$s), the merchant will no longer be able to process orders and refunds": [
        "Wenn Sie die Instanz mit der Bezeichnung %1$s (ID: %2$s) l\xF6schen, wird sie keine weiteren Bestellungen oder R\xFCckerstattungen bearbeiten k\xF6nnen"
      ],
      "This action deletes the instance's private key, but preserves all transaction data. You can still access the transaction data after having deleted the instance.": [
        "Dieser Vorgang l\xF6scht den privaten Schl\xFCssel der Instanz, l\xE4sst jedoch alle Transaktionsdaten unber\xFChrt. Sie k\xF6nnen daher auch weiterhin nach der L\xF6schung der Instanz auf diese Daten zugreifen."
      ],
      "Deleting an instance %1$s": [
        "L\xF6schung der Instanz %1$s"
      ],
      "This cannot be undone!": [
        "Das kann nicht r\xFCckg\xE4ngig gemacht werden!"
      ],
      "Purge the instance": [
        "Instanz l\xF6schen"
      ],
      'Purge the instance "%1$s"': [
        'Instanz "%1$s" wird endg\xFCltig verworfen'
      ],
      "If you purge the instance named %1$s (ID: %2$s), you will also delete all of its transaction data!": [
        "Falls Sie die Instanz %1$s (ID: %2$s) vollst\xE4ndig verwerfen, l\xF6schen Sie damit auch unwiderruflich alle Transaktionsdaten!"
      ],
      "The instance will disappear from your list and you will no longer be able to access its data.": [
        "Die Instanz wird von Ihrer Liste gestrichen und alle ihre Daten werden nicht mehr zug\xE4nglich sein."
      ],
      "Purging an instance %1$s": [
        "Die Instanz %1$s wird endg\xFCltig verworfen"
      ],
      "checking compatibility with server...": [
        "Pr\xFCfe Kompatibilit\xE4t mit dem Server..."
      ],
      "Contacting the server failed": [
        "Der Kontaktversuch mit dem Server ist fehlgeschlagen"
      ],
      "The server version is not supported": [
        "Die Serverversion wird nicht unterst\xFCtzt"
      ],
      'Supported version "%1$s", server version "%2$s".': [
        'Unterst\xFCtzte Version "%1$s", Serverversion "%2$s".'
      ],
      "Deleting": [
        "l\xF6scht\u2026"
      ],
      "Changing": [
        "\xE4ndern\u2026"
      ],
      "Manage password": [
        "Passwort verwalten"
      ],
      "Business Name": [
        "Firmenname"
      ],
      "Available statistics": [
        "Verf\xFCgbare Statistiken"
      ],
      "There are no statistics to list": [
        "Es gibt noch keine Statistiken anzuzeigen"
      ],
      "title%1$s: Settings": [
        "%1$s: Einstellungen"
      ],
      "title%1$s: New bank account": [
        "%1$s: Neues Bankkonto"
      ],
      "title%1$s: Bank accounts": [
        "%1$s: Bankkonten"
      ],
      "title%1$s: Update bank Account": [
        "%1$s: Bankkonto aktualisieren"
      ],
      "title%1$s: Orders": [
        "%1$s: Bestellungen"
      ],
      "title%1$s: New order": [
        "%1$s: Neue Bestellung"
      ],
      "title%1$s: Inventory": [
        "%1$s: Lagerbestand"
      ],
      "title%1$s: New product": [
        "%1$s: Neuer Artikel"
      ],
      "title%1$s: Update product": [
        "%1$s: Artikel aktualisieren"
      ],
      "title%1$s: Category": [
        "%1$s: Kategorie"
      ],
      "title%1$s: New category": [
        "%1$s: Neue Kategorie"
      ],
      "title%1$s: Update category": [
        "%1$s: Kategorie aktualisieren"
      ],
      "title%1$s: Wire transfers": [
        "%1$s: \xDCberweisungen"
      ],
      "title%1$s: Webhooks": [
        "%1$s: Webhooks"
      ],
      "title%1$s: New webhook": [
        "%1$s: Neuer Webhook"
      ],
      "title%1$s: Update webhook": [
        "%1$s: Webhook aktualisieren"
      ],
      "title%1$s: OTP devices": [
        "%1$s: OTP-Ger\xE4te"
      ],
      "title%1$s: New OTP device": [
        "%1$s: Neues OTP-Ger\xE4t"
      ],
      "title%1$s: Update OTP device": [
        "%1$s: OTP-Ger\xE4t aktualisieren"
      ],
      "title%1$s: New template": [
        "%1$s: Neue Vorlage"
      ],
      "title%1$s: Update template": [
        "%1$s: Vorlage aktualisieren"
      ],
      "title%1$s: Templates": [
        "%1$s: Vorlagen"
      ],
      "title%1$s: Use template": [
        "%1$s: Vorlage verwenden"
      ],
      "title%1$s: Personalization": [
        "%1$s: Personalisierung"
      ],
      "title%1$s: Token families": [
        "%1$s: Tokenfamilien"
      ],
      "title%1$s: New token family": [
        "%1$s: Neue Tokenfamilie"
      ],
      "title%1$s: Update token family": [
        "%1$s: Tokenfamilie aktualisieren"
      ],
      "title%1$s: Access tokens": [
        "%1$s: Zugangstoken"
      ],
      "title%1$s: New access token": [
        "%1$s: Neuer Zugangstoken"
      ],
      "titleNew instance": [
        "Neue Instanz"
      ],
      "titleInstances": [
        "Instanzen"
      ]
    }
  },
  "domain": "messages",
  "plural_forms": "nplurals=2; plural=n != 1;",
  "lang": "de",
  "completeness": 100
};

// src/lang.ts
var keys = Object.keys(strings);
var options = process.argv.reduce(
  (prev, arg, idx, arr) => {
    if (arg === "--missing") {
      prev.ACTION = "SHOW_MISSING";
    }
    if (arg === "--lang") {
      prev.lang = idx + 1 < arr.length ? arr[idx + 1] : void 0;
    }
    if (arg === "--suggest") {
      prev.ACTION = "SUGGEST";
    }
    return prev;
  },
  {
    ACTION: "SHOW_STATS",
    lang: void 0
  }
);
switch (options.ACTION) {
  case "SHOW_MISSING": {
    keys.forEach((key) => {
      if (options.lang && options.lang !== key) {
        return;
      }
      const msgs = Object.entries(
        strings[key].locale_data.messages
      );
      let empty = msgs.filter(([key2, value]) => {
        if (!value || !value.length || !value[0])
          return true;
        return false;
      });
      empty.forEach((msg) => {
        console.log(msg[0]);
      });
    });
    break;
  }
  case "SUGGEST": {
    keys.forEach((key) => {
      if (options.lang && options.lang !== key) {
        return;
      }
      const content = `#: src/Application.tsx:208
#, c-format
msgid "checking compatibility with server..."
msgstr ""

#: src/Application.tsx:217
#, fuzzy, c-format
msgid "Contacting the server failed"
msgstr "No se pudo aceder al servidor"

#: src/Application.tsx:229
#, c-format
msgid "The server version is not supported"
msgstr ""

#: src/Application.tsx:230
#, c-format
msgid "Supported version "%1$s", server version "%2$s"."
msgstr ""
`;
      const msgs = Object.entries(
        strings[key].locale_data.messages
      );
      let empty = msgs.filter(([key2, value]) => {
        if (!value || !value.length || !value[0])
          return true;
        return false;
      });
      empty.forEach((msg) => {
        const search = new RegExp(`

(#.*
)+msgid "${msg[0]}"
msgstr .*

`, "gm");
        const res = search.exec(content);
        if (res) {
          console.log(res[0].trim());
        }
      });
    });
    break;
  }
  case "SHOW_STATS": {
    keys.forEach((key) => {
      const com = strings[key].completeness;
      const msgs = Object.entries(
        strings[key].locale_data.messages
      );
      const empty = msgs.filter(([key2, value]) => {
        if (!value || !value.length || !value[0])
          return true;
        return false;
      }).length;
      console.log(
        ` * ${key}: ${com.toLocaleString(void 0, {
          minimumIntegerDigits: 2
        })}% ${(msgs.length - empty).toLocaleString(void 0, {
          minimumIntegerDigits: 3
        })} of ${msgs.length}`
      );
    });
    break;
  }
  default: {
    throw Error(`unknown action: ${options.ACTION}`);
  }
}
//# sourceMappingURL=lang.js.map
