@charset "UTF-8";

/* src/scss/DurationPicker.scss */
.rdp-picker {
  display: flex;
  height: 175px;
}
@media (max-width: 400px) {
  .rdp-picker {
    width: 250px;
  }
}
.rdp-masked-div {
  overflow: hidden;
  height: 175px;
  position: relative;
}
.rdp-column-container {
  flex-grow: 1;
  display: inline-block;
}
.rdp-column {
  position: absolute;
  z-index: 0;
  width: 100%;
}
.rdp-reticule {
  border: 0;
  border-top: 2px solid rgb(109, 202, 236);
  height: 2px;
  position: absolute;
  width: 80%;
  margin: 0;
  z-index: 100;
  left: 50%;
  -webkit-transform: translateX(-50%);
  transform: translateX(-50%);
}
.rdp-text-overlay {
  position: absolute;
  display: flex;
  align-items: center;
  justify-content: center;
  height: 35px;
  font-size: 20px;
  left: 50%;
  -webkit-transform: translateX(-50%);
  transform: translateX(-50%);
}
.rdp-cell div {
  font-size: 17px;
  color: gray;
  font-style: italic;
}
.rdp-cell {
  display: flex;
  align-items: center;
  justify-content: center;
  height: 35px;
  font-size: 18px;
}
.rdp-center {
  font-size: 25px;
}

/* src/scss/fonts/nunito.css */
@font-face {
  font-family: "Nunito";
  font-style: normal;
  font-weight: 400;
  src: url("./XRXV3I6Li01BKofINeaE-34D53UZZ.ttf") format("truetype");
}

/* src/scss/icons/materialdesignicons-4.9.95.min.css */
@font-face {
  font-family: "Material Design Icons";
  src: url("./materialdesignicons-webfont-4.9.95-HENMUVWG.eot");
  src:
    url("./materialdesignicons-webfont-4.9.95-7W2GKO6O.woff2") format("woff2"),
    url("./materialdesignicons-webfont-4.9.95-WP2MXZKI.woff") format("woff"),
    url("./materialdesignicons-webfont-4.9.95-GELXKCZG.ttf") format("truetype");
  font-weight: normal;
  font-style: normal;
}
.mdi:before,
.mdi-set {
  display: inline-block;
  font: normal normal normal 24px/1 "Material Design Icons";
  font-size: inherit;
  text-rendering: auto;
  line-height: inherit;
  -webkit-font-smoothing: antialiased;
  -moz-osx-font-smoothing: grayscale;
}
.mdi-ab-testing::before {
  content: "\f001c";
}
.mdi-abjad-arabic::before {
  content: "\f0353";
}
.mdi-abjad-hebrew::before {
  content: "\f0354";
}
.mdi-abugida-devanagari::before {
  content: "\f0355";
}
.mdi-abugida-thai::before {
  content: "\f0356";
}
.mdi-access-point::before {
  content: "\f002";
}
.mdi-access-point-network::before {
  content: "\f003";
}
.mdi-access-point-network-off::before {
  content: "\fbbd";
}
.mdi-account::before {
  content: "\f004";
}
.mdi-account-alert::before {
  content: "\f005";
}
.mdi-account-alert-outline::before {
  content: "\fb2c";
}
.mdi-account-arrow-left::before {
  content: "\fb2d";
}
.mdi-account-arrow-left-outline::before {
  content: "\fb2e";
}
.mdi-account-arrow-right::before {
  content: "\fb2f";
}
.mdi-account-arrow-right-outline::before {
  content: "\fb30";
}
.mdi-account-badge::before {
  content: "\fd83";
}
.mdi-account-badge-alert::before {
  content: "\fd84";
}
.mdi-account-badge-alert-outline::before {
  content: "\fd85";
}
.mdi-account-badge-horizontal::before {
  content: "\fdf0";
}
.mdi-account-badge-horizontal-outline::before {
  content: "\fdf1";
}
.mdi-account-badge-outline::before {
  content: "\fd86";
}
.mdi-account-box::before {
  content: "\f006";
}
.mdi-account-box-multiple::before {
  content: "\f933";
}
.mdi-account-box-multiple-outline::before {
  content: "\f002c";
}
.mdi-account-box-outline::before {
  content: "\f007";
}
.mdi-account-cancel::before {
  content: "\f030a";
}
.mdi-account-cancel-outline::before {
  content: "\f030b";
}
.mdi-account-card-details::before {
  content: "\f5d2";
}
.mdi-account-card-details-outline::before {
  content: "\fd87";
}
.mdi-account-cash::before {
  content: "\f00c2";
}
.mdi-account-cash-outline::before {
  content: "\f00c3";
}
.mdi-account-check::before {
  content: "\f008";
}
.mdi-account-check-outline::before {
  content: "\fbbe";
}
.mdi-account-child::before {
  content: "\fa88";
}
.mdi-account-child-circle::before {
  content: "\fa89";
}
.mdi-account-child-outline::before {
  content: "\f00f3";
}
.mdi-account-circle::before {
  content: "\f009";
}
.mdi-account-circle-outline::before {
  content: "\fb31";
}
.mdi-account-clock::before {
  content: "\fb32";
}
.mdi-account-clock-outline::before {
  content: "\fb33";
}
.mdi-account-cog::before {
  content: "\f039b";
}
.mdi-account-cog-outline::before {
  content: "\f039c";
}
.mdi-account-convert::before {
  content: "\f00a";
}
.mdi-account-convert-outline::before {
  content: "\f032c";
}
.mdi-account-details::before {
  content: "\f631";
}
.mdi-account-details-outline::before {
  content: "\f039d";
}
.mdi-account-edit::before {
  content: "\f6bb";
}
.mdi-account-edit-outline::before {
  content: "\f001d";
}
.mdi-account-group::before {
  content: "\f848";
}
.mdi-account-group-outline::before {
  content: "\fb34";
}
.mdi-account-heart::before {
  content: "\f898";
}
.mdi-account-heart-outline::before {
  content: "\fbbf";
}
.mdi-account-key::before {
  content: "\f00b";
}
.mdi-account-key-outline::before {
  content: "\fbc0";
}
.mdi-account-lock::before {
  content: "\f0189";
}
.mdi-account-lock-outline::before {
  content: "\f018a";
}
.mdi-account-minus::before {
  content: "\f00d";
}
.mdi-account-minus-outline::before {
  content: "\faeb";
}
.mdi-account-multiple::before {
  content: "\f00e";
}
.mdi-account-multiple-check::before {
  content: "\f8c4";
}
.mdi-account-multiple-check-outline::before {
  content: "\f0229";
}
.mdi-account-multiple-minus::before {
  content: "\f5d3";
}
.mdi-account-multiple-minus-outline::before {
  content: "\fbc1";
}
.mdi-account-multiple-outline::before {
  content: "\f00f";
}
.mdi-account-multiple-plus::before {
  content: "\f010";
}
.mdi-account-multiple-plus-outline::before {
  content: "\f7ff";
}
.mdi-account-multiple-remove::before {
  content: "\f0235";
}
.mdi-account-multiple-remove-outline::before {
  content: "\f0236";
}
.mdi-account-network::before {
  content: "\f011";
}
.mdi-account-network-outline::before {
  content: "\fbc2";
}
.mdi-account-off::before {
  content: "\f012";
}
.mdi-account-off-outline::before {
  content: "\fbc3";
}
.mdi-account-outline::before {
  content: "\f013";
}
.mdi-account-plus::before {
  content: "\f014";
}
.mdi-account-plus-outline::before {
  content: "\f800";
}
.mdi-account-question::before {
  content: "\fb35";
}
.mdi-account-question-outline::before {
  content: "\fb36";
}
.mdi-account-remove::before {
  content: "\f015";
}
.mdi-account-remove-outline::before {
  content: "\faec";
}
.mdi-account-search::before {
  content: "\f016";
}
.mdi-account-search-outline::before {
  content: "\f934";
}
.mdi-account-settings::before {
  content: "\f630";
}
.mdi-account-settings-outline::before {
  content: "\f00f4";
}
.mdi-account-star::before {
  content: "\f017";
}
.mdi-account-star-outline::before {
  content: "\fbc4";
}
.mdi-account-supervisor::before {
  content: "\fa8a";
}
.mdi-account-supervisor-circle::before {
  content: "\fa8b";
}
.mdi-account-supervisor-outline::before {
  content: "\f0158";
}
.mdi-account-switch::before {
  content: "\f019";
}
.mdi-account-tie::before {
  content: "\fcbf";
}
.mdi-account-tie-outline::before {
  content: "\f00f5";
}
.mdi-account-tie-voice::before {
  content: "\f0333";
}
.mdi-account-tie-voice-off::before {
  content: "\f0335";
}
.mdi-account-tie-voice-off-outline::before {
  content: "\f0336";
}
.mdi-account-tie-voice-outline::before {
  content: "\f0334";
}
.mdi-accusoft::before {
  content: "\f849";
}
.mdi-adjust::before {
  content: "\f01a";
}
.mdi-adobe::before {
  content: "\f935";
}
.mdi-adobe-acrobat::before {
  content: "\ffbd";
}
.mdi-air-conditioner::before {
  content: "\f01b";
}
.mdi-air-filter::before {
  content: "\fd1f";
}
.mdi-air-horn::before {
  content: "\fd88";
}
.mdi-air-humidifier::before {
  content: "\f00c4";
}
.mdi-air-purifier::before {
  content: "\fd20";
}
.mdi-airbag::before {
  content: "\fbc5";
}
.mdi-airballoon::before {
  content: "\f01c";
}
.mdi-airballoon-outline::before {
  content: "\f002d";
}
.mdi-airplane::before {
  content: "\f01d";
}
.mdi-airplane-landing::before {
  content: "\f5d4";
}
.mdi-airplane-off::before {
  content: "\f01e";
}
.mdi-airplane-takeoff::before {
  content: "\f5d5";
}
.mdi-airplay::before {
  content: "\f01f";
}
.mdi-airport::before {
  content: "\f84a";
}
.mdi-alarm::before {
  content: "\f020";
}
.mdi-alarm-bell::before {
  content: "\f78d";
}
.mdi-alarm-check::before {
  content: "\f021";
}
.mdi-alarm-light::before {
  content: "\f78e";
}
.mdi-alarm-light-outline::before {
  content: "\fbc6";
}
.mdi-alarm-multiple::before {
  content: "\f022";
}
.mdi-alarm-note::before {
  content: "\fe8e";
}
.mdi-alarm-note-off::before {
  content: "\fe8f";
}
.mdi-alarm-off::before {
  content: "\f023";
}
.mdi-alarm-plus::before {
  content: "\f024";
}
.mdi-alarm-snooze::before {
  content: "\f68d";
}
.mdi-album::before {
  content: "\f025";
}
.mdi-alert::before {
  content: "\f026";
}
.mdi-alert-box::before {
  content: "\f027";
}
.mdi-alert-box-outline::before {
  content: "\fcc0";
}
.mdi-alert-circle::before {
  content: "\f028";
}
.mdi-alert-circle-check::before {
  content: "\f0218";
}
.mdi-alert-circle-check-outline::before {
  content: "\f0219";
}
.mdi-alert-circle-outline::before {
  content: "\f5d6";
}
.mdi-alert-decagram::before {
  content: "\f6bc";
}
.mdi-alert-decagram-outline::before {
  content: "\fcc1";
}
.mdi-alert-octagon::before {
  content: "\f029";
}
.mdi-alert-octagon-outline::before {
  content: "\fcc2";
}
.mdi-alert-octagram::before {
  content: "\f766";
}
.mdi-alert-octagram-outline::before {
  content: "\fcc3";
}
.mdi-alert-outline::before {
  content: "\f02a";
}
.mdi-alert-rhombus::before {
  content: "\f01f9";
}
.mdi-alert-rhombus-outline::before {
  content: "\f01fa";
}
.mdi-alien::before {
  content: "\f899";
}
.mdi-alien-outline::before {
  content: "\f00f6";
}
.mdi-align-horizontal-center::before {
  content: "\f01ee";
}
.mdi-align-horizontal-left::before {
  content: "\f01ed";
}
.mdi-align-horizontal-right::before {
  content: "\f01ef";
}
.mdi-align-vertical-bottom::before {
  content: "\f01f0";
}
.mdi-align-vertical-center::before {
  content: "\f01f1";
}
.mdi-align-vertical-top::before {
  content: "\f01f2";
}
.mdi-all-inclusive::before {
  content: "\f6bd";
}
.mdi-allergy::before {
  content: "\f0283";
}
.mdi-alpha::before {
  content: "\f02b";
}
.mdi-alpha-a::before {
  content: "A";
}
.mdi-alpha-a-box::before {
  content: "\faed";
}
.mdi-alpha-a-box-outline::before {
  content: "\fbc7";
}
.mdi-alpha-a-circle::before {
  content: "\fbc8";
}
.mdi-alpha-a-circle-outline::before {
  content: "\fbc9";
}
.mdi-alpha-b::before {
  content: "B";
}
.mdi-alpha-b-box::before {
  content: "\faee";
}
.mdi-alpha-b-box-outline::before {
  content: "\fbca";
}
.mdi-alpha-b-circle::before {
  content: "\fbcb";
}
.mdi-alpha-b-circle-outline::before {
  content: "\fbcc";
}
.mdi-alpha-c::before {
  content: "C";
}
.mdi-alpha-c-box::before {
  content: "\faef";
}
.mdi-alpha-c-box-outline::before {
  content: "\fbcd";
}
.mdi-alpha-c-circle::before {
  content: "\fbce";
}
.mdi-alpha-c-circle-outline::before {
  content: "\fbcf";
}
.mdi-alpha-d::before {
  content: "D";
}
.mdi-alpha-d-box::before {
  content: "\faf0";
}
.mdi-alpha-d-box-outline::before {
  content: "\fbd0";
}
.mdi-alpha-d-circle::before {
  content: "\fbd1";
}
.mdi-alpha-d-circle-outline::before {
  content: "\fbd2";
}
.mdi-alpha-e::before {
  content: "E";
}
.mdi-alpha-e-box::before {
  content: "\faf1";
}
.mdi-alpha-e-box-outline::before {
  content: "\fbd3";
}
.mdi-alpha-e-circle::before {
  content: "\fbd4";
}
.mdi-alpha-e-circle-outline::before {
  content: "\fbd5";
}
.mdi-alpha-f::before {
  content: "F";
}
.mdi-alpha-f-box::before {
  content: "\faf2";
}
.mdi-alpha-f-box-outline::before {
  content: "\fbd6";
}
.mdi-alpha-f-circle::before {
  content: "\fbd7";
}
.mdi-alpha-f-circle-outline::before {
  content: "\fbd8";
}
.mdi-alpha-g::before {
  content: "G";
}
.mdi-alpha-g-box::before {
  content: "\faf3";
}
.mdi-alpha-g-box-outline::before {
  content: "\fbd9";
}
.mdi-alpha-g-circle::before {
  content: "\fbda";
}
.mdi-alpha-g-circle-outline::before {
  content: "\fbdb";
}
.mdi-alpha-h::before {
  content: "H";
}
.mdi-alpha-h-box::before {
  content: "\faf4";
}
.mdi-alpha-h-box-outline::before {
  content: "\fbdc";
}
.mdi-alpha-h-circle::before {
  content: "\fbdd";
}
.mdi-alpha-h-circle-outline::before {
  content: "\fbde";
}
.mdi-alpha-i::before {
  content: "I";
}
.mdi-alpha-i-box::before {
  content: "\faf5";
}
.mdi-alpha-i-box-outline::before {
  content: "\fbdf";
}
.mdi-alpha-i-circle::before {
  content: "\fbe0";
}
.mdi-alpha-i-circle-outline::before {
  content: "\fbe1";
}
.mdi-alpha-j::before {
  content: "J";
}
.mdi-alpha-j-box::before {
  content: "\faf6";
}
.mdi-alpha-j-box-outline::before {
  content: "\fbe2";
}
.mdi-alpha-j-circle::before {
  content: "\fbe3";
}
.mdi-alpha-j-circle-outline::before {
  content: "\fbe4";
}
.mdi-alpha-k::before {
  content: "K";
}
.mdi-alpha-k-box::before {
  content: "\faf7";
}
.mdi-alpha-k-box-outline::before {
  content: "\fbe5";
}
.mdi-alpha-k-circle::before {
  content: "\fbe6";
}
.mdi-alpha-k-circle-outline::before {
  content: "\fbe7";
}
.mdi-alpha-l::before {
  content: "L";
}
.mdi-alpha-l-box::before {
  content: "\faf8";
}
.mdi-alpha-l-box-outline::before {
  content: "\fbe8";
}
.mdi-alpha-l-circle::before {
  content: "\fbe9";
}
.mdi-alpha-l-circle-outline::before {
  content: "\fbea";
}
.mdi-alpha-m::before {
  content: "M";
}
.mdi-alpha-m-box::before {
  content: "\faf9";
}
.mdi-alpha-m-box-outline::before {
  content: "\fbeb";
}
.mdi-alpha-m-circle::before {
  content: "\fbec";
}
.mdi-alpha-m-circle-outline::before {
  content: "\fbed";
}
.mdi-alpha-n::before {
  content: "N";
}
.mdi-alpha-n-box::before {
  content: "\fafa";
}
.mdi-alpha-n-box-outline::before {
  content: "\fbee";
}
.mdi-alpha-n-circle::before {
  content: "\fbef";
}
.mdi-alpha-n-circle-outline::before {
  content: "\fbf0";
}
.mdi-alpha-o::before {
  content: "O";
}
.mdi-alpha-o-box::before {
  content: "\fafb";
}
.mdi-alpha-o-box-outline::before {
  content: "\fbf1";
}
.mdi-alpha-o-circle::before {
  content: "\fbf2";
}
.mdi-alpha-o-circle-outline::before {
  content: "\fbf3";
}
.mdi-alpha-p::before {
  content: "P";
}
.mdi-alpha-p-box::before {
  content: "\fafc";
}
.mdi-alpha-p-box-outline::before {
  content: "\fbf4";
}
.mdi-alpha-p-circle::before {
  content: "\fbf5";
}
.mdi-alpha-p-circle-outline::before {
  content: "\fbf6";
}
.mdi-alpha-q::before {
  content: "Q";
}
.mdi-alpha-q-box::before {
  content: "\fafd";
}
.mdi-alpha-q-box-outline::before {
  content: "\fbf7";
}
.mdi-alpha-q-circle::before {
  content: "\fbf8";
}
.mdi-alpha-q-circle-outline::before {
  content: "\fbf9";
}
.mdi-alpha-r::before {
  content: "R";
}
.mdi-alpha-r-box::before {
  content: "\fafe";
}
.mdi-alpha-r-box-outline::before {
  content: "\fbfa";
}
.mdi-alpha-r-circle::before {
  content: "\fbfb";
}
.mdi-alpha-r-circle-outline::before {
  content: "\fbfc";
}
.mdi-alpha-s::before {
  content: "S";
}
.mdi-alpha-s-box::before {
  content: "\faff";
}
.mdi-alpha-s-box-outline::before {
  content: "\fbfd";
}
.mdi-alpha-s-circle::before {
  content: "\fbfe";
}
.mdi-alpha-s-circle-outline::before {
  content: "\fbff";
}
.mdi-alpha-t::before {
  content: "T";
}
.mdi-alpha-t-box::before {
  content: "\fb00";
}
.mdi-alpha-t-box-outline::before {
  content: "\fc00";
}
.mdi-alpha-t-circle::before {
  content: "\fc01";
}
.mdi-alpha-t-circle-outline::before {
  content: "\fc02";
}
.mdi-alpha-u::before {
  content: "U";
}
.mdi-alpha-u-box::before {
  content: "\fb01";
}
.mdi-alpha-u-box-outline::before {
  content: "\fc03";
}
.mdi-alpha-u-circle::before {
  content: "\fc04";
}
.mdi-alpha-u-circle-outline::before {
  content: "\fc05";
}
.mdi-alpha-v::before {
  content: "V";
}
.mdi-alpha-v-box::before {
  content: "\fb02";
}
.mdi-alpha-v-box-outline::before {
  content: "\fc06";
}
.mdi-alpha-v-circle::before {
  content: "\fc07";
}
.mdi-alpha-v-circle-outline::before {
  content: "\fc08";
}
.mdi-alpha-w::before {
  content: "W";
}
.mdi-alpha-w-box::before {
  content: "\fb03";
}
.mdi-alpha-w-box-outline::before {
  content: "\fc09";
}
.mdi-alpha-w-circle::before {
  content: "\fc0a";
}
.mdi-alpha-w-circle-outline::before {
  content: "\fc0b";
}
.mdi-alpha-x::before {
  content: "X";
}
.mdi-alpha-x-box::before {
  content: "\fb04";
}
.mdi-alpha-x-box-outline::before {
  content: "\fc0c";
}
.mdi-alpha-x-circle::before {
  content: "\fc0d";
}
.mdi-alpha-x-circle-outline::before {
  content: "\fc0e";
}
.mdi-alpha-y::before {
  content: "Y";
}
.mdi-alpha-y-box::before {
  content: "\fb05";
}
.mdi-alpha-y-box-outline::before {
  content: "\fc0f";
}
.mdi-alpha-y-circle::before {
  content: "\fc10";
}
.mdi-alpha-y-circle-outline::before {
  content: "\fc11";
}
.mdi-alpha-z::before {
  content: "Z";
}
.mdi-alpha-z-box::before {
  content: "\fb06";
}
.mdi-alpha-z-box-outline::before {
  content: "\fc12";
}
.mdi-alpha-z-circle::before {
  content: "\fc13";
}
.mdi-alpha-z-circle-outline::before {
  content: "\fc14";
}
.mdi-alphabet-aurebesh::before {
  content: "\f0357";
}
.mdi-alphabet-cyrillic::before {
  content: "\f0358";
}
.mdi-alphabet-greek::before {
  content: "\f0359";
}
.mdi-alphabet-latin::before {
  content: "\f035a";
}
.mdi-alphabet-piqad::before {
  content: "\f035b";
}
.mdi-alphabet-tengwar::before {
  content: "\f0362";
}
.mdi-alphabetical::before {
  content: "\f02c";
}
.mdi-alphabetical-off::before {
  content: "\f002e";
}
.mdi-alphabetical-variant::before {
  content: "\f002f";
}
.mdi-alphabetical-variant-off::before {
  content: "\f0030";
}
.mdi-altimeter::before {
  content: "\f5d7";
}
.mdi-amazon::before {
  content: "\f02d";
}
.mdi-amazon-alexa::before {
  content: "\f8c5";
}
.mdi-amazon-drive::before {
  content: "\f02e";
}
.mdi-ambulance::before {
  content: "\f02f";
}
.mdi-ammunition::before {
  content: "\fcc4";
}
.mdi-ampersand::before {
  content: "\fa8c";
}
.mdi-amplifier::before {
  content: "\f030";
}
.mdi-amplifier-off::before {
  content: "\f01e0";
}
.mdi-anchor::before {
  content: "\f031";
}
.mdi-android::before {
  content: "\f032";
}
.mdi-android-auto::before {
  content: "\fa8d";
}
.mdi-android-debug-bridge::before {
  content: "\f033";
}
.mdi-android-head::before {
  content: "\f78f";
}
.mdi-android-messages::before {
  content: "\fd21";
}
.mdi-android-studio::before {
  content: "\f034";
}
.mdi-angle-acute::before {
  content: "\f936";
}
.mdi-angle-obtuse::before {
  content: "\f937";
}
.mdi-angle-right::before {
  content: "\f938";
}
.mdi-angular::before {
  content: "\f6b1";
}
.mdi-angularjs::before {
  content: "\f6be";
}
.mdi-animation::before {
  content: "\f5d8";
}
.mdi-animation-outline::before {
  content: "\fa8e";
}
.mdi-animation-play::before {
  content: "\f939";
}
.mdi-animation-play-outline::before {
  content: "\fa8f";
}
.mdi-ansible::before {
  content: "\f00c5";
}
.mdi-antenna::before {
  content: "\f0144";
}
.mdi-anvil::before {
  content: "\f89a";
}
.mdi-apache-kafka::before {
  content: "\f0031";
}
.mdi-api::before {
  content: "\f00c6";
}
.mdi-api-off::before {
  content: "\f0282";
}
.mdi-apple::before {
  content: "\f035";
}
.mdi-apple-finder::before {
  content: "\f036";
}
.mdi-apple-icloud::before {
  content: "\f038";
}
.mdi-apple-ios::before {
  content: "\f037";
}
.mdi-apple-keyboard-caps::before {
  content: "\f632";
}
.mdi-apple-keyboard-command::before {
  content: "\f633";
}
.mdi-apple-keyboard-control::before {
  content: "\f634";
}
.mdi-apple-keyboard-option::before {
  content: "\f635";
}
.mdi-apple-keyboard-shift::before {
  content: "\f636";
}
.mdi-apple-safari::before {
  content: "\f039";
}
.mdi-application::before {
  content: "\f614";
}
.mdi-application-export::before {
  content: "\fd89";
}
.mdi-application-import::before {
  content: "\fd8a";
}
.mdi-approximately-equal::before {
  content: "\ffbe";
}
.mdi-approximately-equal-box::before {
  content: "\ffbf";
}
.mdi-apps::before {
  content: "\f03b";
}
.mdi-apps-box::before {
  content: "\fd22";
}
.mdi-arch::before {
  content: "\f8c6";
}
.mdi-archive::before {
  content: "\f03c";
}
.mdi-archive-arrow-down::before {
  content: "\f0284";
}
.mdi-archive-arrow-down-outline::before {
  content: "\f0285";
}
.mdi-archive-arrow-up::before {
  content: "\f0286";
}
.mdi-archive-arrow-up-outline::before {
  content: "\f0287";
}
.mdi-archive-outline::before {
  content: "\f0239";
}
.mdi-arm-flex::before {
  content: "\f008f";
}
.mdi-arm-flex-outline::before {
  content: "\f0090";
}
.mdi-arrange-bring-forward::before {
  content: "\f03d";
}
.mdi-arrange-bring-to-front::before {
  content: "\f03e";
}
.mdi-arrange-send-backward::before {
  content: "\f03f";
}
.mdi-arrange-send-to-back::before {
  content: "\f040";
}
.mdi-arrow-all::before {
  content: "\f041";
}
.mdi-arrow-bottom-left::before {
  content: "\f042";
}
.mdi-arrow-bottom-left-bold-outline::before {
  content: "\f9b6";
}
.mdi-arrow-bottom-left-thick::before {
  content: "\f9b7";
}
.mdi-arrow-bottom-right::before {
  content: "\f043";
}
.mdi-arrow-bottom-right-bold-outline::before {
  content: "\f9b8";
}
.mdi-arrow-bottom-right-thick::before {
  content: "\f9b9";
}
.mdi-arrow-collapse::before {
  content: "\f615";
}
.mdi-arrow-collapse-all::before {
  content: "\f044";
}
.mdi-arrow-collapse-down::before {
  content: "\f791";
}
.mdi-arrow-collapse-horizontal::before {
  content: "\f84b";
}
.mdi-arrow-collapse-left::before {
  content: "\f792";
}
.mdi-arrow-collapse-right::before {
  content: "\f793";
}
.mdi-arrow-collapse-up::before {
  content: "\f794";
}
.mdi-arrow-collapse-vertical::before {
  content: "\f84c";
}
.mdi-arrow-decision::before {
  content: "\f9ba";
}
.mdi-arrow-decision-auto::before {
  content: "\f9bb";
}
.mdi-arrow-decision-auto-outline::before {
  content: "\f9bc";
}
.mdi-arrow-decision-outline::before {
  content: "\f9bd";
}
.mdi-arrow-down::before {
  content: "\f045";
}
.mdi-arrow-down-bold::before {
  content: "\f72d";
}
.mdi-arrow-down-bold-box::before {
  content: "\f72e";
}
.mdi-arrow-down-bold-box-outline::before {
  content: "\f72f";
}
.mdi-arrow-down-bold-circle::before {
  content: "\f047";
}
.mdi-arrow-down-bold-circle-outline::before {
  content: "\f048";
}
.mdi-arrow-down-bold-hexagon-outline::before {
  content: "\f049";
}
.mdi-arrow-down-bold-outline::before {
  content: "\f9be";
}
.mdi-arrow-down-box::before {
  content: "\f6bf";
}
.mdi-arrow-down-circle::before {
  content: "\fcb7";
}
.mdi-arrow-down-circle-outline::before {
  content: "\fcb8";
}
.mdi-arrow-down-drop-circle::before {
  content: "\f04a";
}
.mdi-arrow-down-drop-circle-outline::before {
  content: "\f04b";
}
.mdi-arrow-down-thick::before {
  content: "\f046";
}
.mdi-arrow-expand::before {
  content: "\f616";
}
.mdi-arrow-expand-all::before {
  content: "\f04c";
}
.mdi-arrow-expand-down::before {
  content: "\f795";
}
.mdi-arrow-expand-horizontal::before {
  content: "\f84d";
}
.mdi-arrow-expand-left::before {
  content: "\f796";
}
.mdi-arrow-expand-right::before {
  content: "\f797";
}
.mdi-arrow-expand-up::before {
  content: "\f798";
}
.mdi-arrow-expand-vertical::before {
  content: "\f84e";
}
.mdi-arrow-horizontal-lock::before {
  content: "\f0186";
}
.mdi-arrow-left::before {
  content: "\f04d";
}
.mdi-arrow-left-bold::before {
  content: "\f730";
}
.mdi-arrow-left-bold-box::before {
  content: "\f731";
}
.mdi-arrow-left-bold-box-outline::before {
  content: "\f732";
}
.mdi-arrow-left-bold-circle::before {
  content: "\f04f";
}
.mdi-arrow-left-bold-circle-outline::before {
  content: "\f050";
}
.mdi-arrow-left-bold-hexagon-outline::before {
  content: "\f051";
}
.mdi-arrow-left-bold-outline::before {
  content: "\f9bf";
}
.mdi-arrow-left-box::before {
  content: "\f6c0";
}
.mdi-arrow-left-circle::before {
  content: "\fcb9";
}
.mdi-arrow-left-circle-outline::before {
  content: "\fcba";
}
.mdi-arrow-left-drop-circle::before {
  content: "\f052";
}
.mdi-arrow-left-drop-circle-outline::before {
  content: "\f053";
}
.mdi-arrow-left-right::before {
  content: "\fe90";
}
.mdi-arrow-left-right-bold::before {
  content: "\fe91";
}
.mdi-arrow-left-right-bold-outline::before {
  content: "\f9c0";
}
.mdi-arrow-left-thick::before {
  content: "\f04e";
}
.mdi-arrow-right::before {
  content: "\f054";
}
.mdi-arrow-right-bold::before {
  content: "\f733";
}
.mdi-arrow-right-bold-box::before {
  content: "\f734";
}
.mdi-arrow-right-bold-box-outline::before {
  content: "\f735";
}
.mdi-arrow-right-bold-circle::before {
  content: "\f056";
}
.mdi-arrow-right-bold-circle-outline::before {
  content: "\f057";
}
.mdi-arrow-right-bold-hexagon-outline::before {
  content: "\f058";
}
.mdi-arrow-right-bold-outline::before {
  content: "\f9c1";
}
.mdi-arrow-right-box::before {
  content: "\f6c1";
}
.mdi-arrow-right-circle::before {
  content: "\fcbb";
}
.mdi-arrow-right-circle-outline::before {
  content: "\fcbc";
}
.mdi-arrow-right-drop-circle::before {
  content: "\f059";
}
.mdi-arrow-right-drop-circle-outline::before {
  content: "\f05a";
}
.mdi-arrow-right-thick::before {
  content: "\f055";
}
.mdi-arrow-split-horizontal::before {
  content: "\f93a";
}
.mdi-arrow-split-vertical::before {
  content: "\f93b";
}
.mdi-arrow-top-left::before {
  content: "\f05b";
}
.mdi-arrow-top-left-bold-outline::before {
  content: "\f9c2";
}
.mdi-arrow-top-left-bottom-right::before {
  content: "\fe92";
}
.mdi-arrow-top-left-bottom-right-bold::before {
  content: "\fe93";
}
.mdi-arrow-top-left-thick::before {
  content: "\f9c3";
}
.mdi-arrow-top-right::before {
  content: "\f05c";
}
.mdi-arrow-top-right-bold-outline::before {
  content: "\f9c4";
}
.mdi-arrow-top-right-bottom-left::before {
  content: "\fe94";
}
.mdi-arrow-top-right-bottom-left-bold::before {
  content: "\fe95";
}
.mdi-arrow-top-right-thick::before {
  content: "\f9c5";
}
.mdi-arrow-up::before {
  content: "\f05d";
}
.mdi-arrow-up-bold::before {
  content: "\f736";
}
.mdi-arrow-up-bold-box::before {
  content: "\f737";
}
.mdi-arrow-up-bold-box-outline::before {
  content: "\f738";
}
.mdi-arrow-up-bold-circle::before {
  content: "\f05f";
}
.mdi-arrow-up-bold-circle-outline::before {
  content: "\f060";
}
.mdi-arrow-up-bold-hexagon-outline::before {
  content: "\f061";
}
.mdi-arrow-up-bold-outline::before {
  content: "\f9c6";
}
.mdi-arrow-up-box::before {
  content: "\f6c2";
}
.mdi-arrow-up-circle::before {
  content: "\fcbd";
}
.mdi-arrow-up-circle-outline::before {
  content: "\fcbe";
}
.mdi-arrow-up-down::before {
  content: "\fe96";
}
.mdi-arrow-up-down-bold::before {
  content: "\fe97";
}
.mdi-arrow-up-down-bold-outline::before {
  content: "\f9c7";
}
.mdi-arrow-up-drop-circle::before {
  content: "\f062";
}
.mdi-arrow-up-drop-circle-outline::before {
  content: "\f063";
}
.mdi-arrow-up-thick::before {
  content: "\f05e";
}
.mdi-arrow-vertical-lock::before {
  content: "\f0187";
}
.mdi-artist::before {
  content: "\f802";
}
.mdi-artist-outline::before {
  content: "\fcc5";
}
.mdi-artstation::before {
  content: "\fb37";
}
.mdi-aspect-ratio::before {
  content: "\fa23";
}
.mdi-assistant::before {
  content: "\f064";
}
.mdi-asterisk::before {
  content: "\f6c3";
}
.mdi-at::before {
  content: "\f065";
}
.mdi-atlassian::before {
  content: "\f803";
}
.mdi-atm::before {
  content: "\fd23";
}
.mdi-atom::before {
  content: "\f767";
}
.mdi-atom-variant::before {
  content: "\fe98";
}
.mdi-attachment::before {
  content: "\f066";
}
.mdi-audio-video::before {
  content: "\f93c";
}
.mdi-audio-video-off::before {
  content: "\f01e1";
}
.mdi-audiobook::before {
  content: "\f067";
}
.mdi-augmented-reality::before {
  content: "\f84f";
}
.mdi-auto-download::before {
  content: "\f03a9";
}
.mdi-auto-fix::before {
  content: "\f068";
}
.mdi-auto-upload::before {
  content: "\f069";
}
.mdi-autorenew::before {
  content: "\f06a";
}
.mdi-av-timer::before {
  content: "\f06b";
}
.mdi-aws::before {
  content: "\fdf2";
}
.mdi-axe::before {
  content: "\f8c7";
}
.mdi-axis::before {
  content: "\fd24";
}
.mdi-axis-arrow::before {
  content: "\fd25";
}
.mdi-axis-arrow-lock::before {
  content: "\fd26";
}
.mdi-axis-lock::before {
  content: "\fd27";
}
.mdi-axis-x-arrow::before {
  content: "\fd28";
}
.mdi-axis-x-arrow-lock::before {
  content: "\fd29";
}
.mdi-axis-x-rotate-clockwise::before {
  content: "\fd2a";
}
.mdi-axis-x-rotate-counterclockwise::before {
  content: "\fd2b";
}
.mdi-axis-x-y-arrow-lock::before {
  content: "\fd2c";
}
.mdi-axis-y-arrow::before {
  content: "\fd2d";
}
.mdi-axis-y-arrow-lock::before {
  content: "\fd2e";
}
.mdi-axis-y-rotate-clockwise::before {
  content: "\fd2f";
}
.mdi-axis-y-rotate-counterclockwise::before {
  content: "\fd30";
}
.mdi-axis-z-arrow::before {
  content: "\fd31";
}
.mdi-axis-z-arrow-lock::before {
  content: "\fd32";
}
.mdi-axis-z-rotate-clockwise::before {
  content: "\fd33";
}
.mdi-axis-z-rotate-counterclockwise::before {
  content: "\fd34";
}
.mdi-azure::before {
  content: "\f804";
}
.mdi-azure-devops::before {
  content: "\f0091";
}
.mdi-babel::before {
  content: "\fa24";
}
.mdi-baby::before {
  content: "\f06c";
}
.mdi-baby-bottle::before {
  content: "\ff56";
}
.mdi-baby-bottle-outline::before {
  content: "\ff57";
}
.mdi-baby-carriage::before {
  content: "\f68e";
}
.mdi-baby-carriage-off::before {
  content: "\ffc0";
}
.mdi-baby-face::before {
  content: "\fe99";
}
.mdi-baby-face-outline::before {
  content: "\fe9a";
}
.mdi-backburger::before {
  content: "\f06d";
}
.mdi-backspace::before {
  content: "\f06e";
}
.mdi-backspace-outline::before {
  content: "\fb38";
}
.mdi-backspace-reverse::before {
  content: "\fe9b";
}
.mdi-backspace-reverse-outline::before {
  content: "\fe9c";
}
.mdi-backup-restore::before {
  content: "\f06f";
}
.mdi-bacteria::before {
  content: "\fef2";
}
.mdi-bacteria-outline::before {
  content: "\fef3";
}
.mdi-badminton::before {
  content: "\f850";
}
.mdi-bag-carry-on::before {
  content: "\ff58";
}
.mdi-bag-carry-on-check::before {
  content: "\fd41";
}
.mdi-bag-carry-on-off::before {
  content: "\ff59";
}
.mdi-bag-checked::before {
  content: "\ff5a";
}
.mdi-bag-personal::before {
  content: "\fdf3";
}
.mdi-bag-personal-off::before {
  content: "\fdf4";
}
.mdi-bag-personal-off-outline::before {
  content: "\fdf5";
}
.mdi-bag-personal-outline::before {
  content: "\fdf6";
}
.mdi-baguette::before {
  content: "\ff5b";
}
.mdi-balloon::before {
  content: "\fa25";
}
.mdi-ballot::before {
  content: "\f9c8";
}
.mdi-ballot-outline::before {
  content: "\f9c9";
}
.mdi-ballot-recount::before {
  content: "\fc15";
}
.mdi-ballot-recount-outline::before {
  content: "\fc16";
}
.mdi-bandage::before {
  content: "\fd8b";
}
.mdi-bandcamp::before {
  content: "\f674";
}
.mdi-bank::before {
  content: "\f070";
}
.mdi-bank-minus::before {
  content: "\fd8c";
}
.mdi-bank-outline::before {
  content: "\fe9d";
}
.mdi-bank-plus::before {
  content: "\fd8d";
}
.mdi-bank-remove::before {
  content: "\fd8e";
}
.mdi-bank-transfer::before {
  content: "\fa26";
}
.mdi-bank-transfer-in::before {
  content: "\fa27";
}
.mdi-bank-transfer-out::before {
  content: "\fa28";
}
.mdi-barcode::before {
  content: "\f071";
}
.mdi-barcode-off::before {
  content: "\f0261";
}
.mdi-barcode-scan::before {
  content: "\f072";
}
.mdi-barley::before {
  content: "\f073";
}
.mdi-barley-off::before {
  content: "\fb39";
}
.mdi-barn::before {
  content: "\fb3a";
}
.mdi-barrel::before {
  content: "\f074";
}
.mdi-baseball::before {
  content: "\f851";
}
.mdi-baseball-bat::before {
  content: "\f852";
}
.mdi-basecamp::before {
  content: "\f075";
}
.mdi-bash::before {
  content: "\f01ae";
}
.mdi-basket::before {
  content: "\f076";
}
.mdi-basket-fill::before {
  content: "\f077";
}
.mdi-basket-outline::before {
  content: "\f01ac";
}
.mdi-basket-unfill::before {
  content: "\f078";
}
.mdi-basketball::before {
  content: "\f805";
}
.mdi-basketball-hoop::before {
  content: "\fc17";
}
.mdi-basketball-hoop-outline::before {
  content: "\fc18";
}
.mdi-bat::before {
  content: "\fb3b";
}
.mdi-battery::before {
  content: "\f079";
}
.mdi-battery-10::before {
  content: "\f07a";
}
.mdi-battery-10-bluetooth::before {
  content: "\f93d";
}
.mdi-battery-20::before {
  content: "\f07b";
}
.mdi-battery-20-bluetooth::before {
  content: "\f93e";
}
.mdi-battery-30::before {
  content: "\f07c";
}
.mdi-battery-30-bluetooth::before {
  content: "\f93f";
}
.mdi-battery-40::before {
  content: "\f07d";
}
.mdi-battery-40-bluetooth::before {
  content: "\f940";
}
.mdi-battery-50::before {
  content: "\f07e";
}
.mdi-battery-50-bluetooth::before {
  content: "\f941";
}
.mdi-battery-60::before {
  content: "\f07f";
}
.mdi-battery-60-bluetooth::before {
  content: "\f942";
}
.mdi-battery-70::before {
  content: "\f080";
}
.mdi-battery-70-bluetooth::before {
  content: "\f943";
}
.mdi-battery-80::before {
  content: "\f081";
}
.mdi-battery-80-bluetooth::before {
  content: "\f944";
}
.mdi-battery-90::before {
  content: "\f082";
}
.mdi-battery-90-bluetooth::before {
  content: "\f945";
}
.mdi-battery-alert::before {
  content: "\f083";
}
.mdi-battery-alert-bluetooth::before {
  content: "\f946";
}
.mdi-battery-alert-variant::before {
  content: "\f00f7";
}
.mdi-battery-alert-variant-outline::before {
  content: "\f00f8";
}
.mdi-battery-bluetooth::before {
  content: "\f947";
}
.mdi-battery-bluetooth-variant::before {
  content: "\f948";
}
.mdi-battery-charging::before {
  content: "\f084";
}
.mdi-battery-charging-10::before {
  content: "\f89b";
}
.mdi-battery-charging-100::before {
  content: "\f085";
}
.mdi-battery-charging-20::before {
  content: "\f086";
}
.mdi-battery-charging-30::before {
  content: "\f087";
}
.mdi-battery-charging-40::before {
  content: "\f088";
}
.mdi-battery-charging-50::before {
  content: "\f89c";
}
.mdi-battery-charging-60::before {
  content: "\f089";
}
.mdi-battery-charging-70::before {
  content: "\f89d";
}
.mdi-battery-charging-80::before {
  content: "\f08a";
}
.mdi-battery-charging-90::before {
  content: "\f08b";
}
.mdi-battery-charging-high::before {
  content: "\f02d1";
}
.mdi-battery-charging-low::before {
  content: "\f02cf";
}
.mdi-battery-charging-medium::before {
  content: "\f02d0";
}
.mdi-battery-charging-outline::before {
  content: "\f89e";
}
.mdi-battery-charging-wireless::before {
  content: "\f806";
}
.mdi-battery-charging-wireless-10::before {
  content: "\f807";
}
.mdi-battery-charging-wireless-20::before {
  content: "\f808";
}
.mdi-battery-charging-wireless-30::before {
  content: "\f809";
}
.mdi-battery-charging-wireless-40::before {
  content: "\f80a";
}
.mdi-battery-charging-wireless-50::before {
  content: "\f80b";
}
.mdi-battery-charging-wireless-60::before {
  content: "\f80c";
}
.mdi-battery-charging-wireless-70::before {
  content: "\f80d";
}
.mdi-battery-charging-wireless-80::before {
  content: "\f80e";
}
.mdi-battery-charging-wireless-90::before {
  content: "\f80f";
}
.mdi-battery-charging-wireless-alert::before {
  content: "\f810";
}
.mdi-battery-charging-wireless-outline::before {
  content: "\f811";
}
.mdi-battery-heart::before {
  content: "\f023a";
}
.mdi-battery-heart-outline::before {
  content: "\f023b";
}
.mdi-battery-heart-variant::before {
  content: "\f023c";
}
.mdi-battery-high::before {
  content: "\f02ce";
}
.mdi-battery-low::before {
  content: "\f02cc";
}
.mdi-battery-medium::before {
  content: "\f02cd";
}
.mdi-battery-minus::before {
  content: "\f08c";
}
.mdi-battery-negative::before {
  content: "\f08d";
}
.mdi-battery-off::before {
  content: "\f0288";
}
.mdi-battery-off-outline::before {
  content: "\f0289";
}
.mdi-battery-outline::before {
  content: "\f08e";
}
.mdi-battery-plus::before {
  content: "\f08f";
}
.mdi-battery-positive::before {
  content: "\f090";
}
.mdi-battery-unknown::before {
  content: "\f091";
}
.mdi-battery-unknown-bluetooth::before {
  content: "\f949";
}
.mdi-battlenet::before {
  content: "\fb3c";
}
.mdi-beach::before {
  content: "\f092";
}
.mdi-beaker::before {
  content: "\fcc6";
}
.mdi-beaker-alert::before {
  content: "\f0254";
}
.mdi-beaker-alert-outline::before {
  content: "\f0255";
}
.mdi-beaker-check::before {
  content: "\f0256";
}
.mdi-beaker-check-outline::before {
  content: "\f0257";
}
.mdi-beaker-minus::before {
  content: "\f0258";
}
.mdi-beaker-minus-outline::before {
  content: "\f0259";
}
.mdi-beaker-outline::before {
  content: "\f68f";
}
.mdi-beaker-plus::before {
  content: "\f025a";
}
.mdi-beaker-plus-outline::before {
  content: "\f025b";
}
.mdi-beaker-question::before {
  content: "\f025c";
}
.mdi-beaker-question-outline::before {
  content: "\f025d";
}
.mdi-beaker-remove::before {
  content: "\f025e";
}
.mdi-beaker-remove-outline::before {
  content: "\f025f";
}
.mdi-beats::before {
  content: "\f097";
}
.mdi-bed-double::before {
  content: "\f0092";
}
.mdi-bed-double-outline::before {
  content: "\f0093";
}
.mdi-bed-empty::before {
  content: "\f89f";
}
.mdi-bed-king::before {
  content: "\f0094";
}
.mdi-bed-king-outline::before {
  content: "\f0095";
}
.mdi-bed-queen::before {
  content: "\f0096";
}
.mdi-bed-queen-outline::before {
  content: "\f0097";
}
.mdi-bed-single::before {
  content: "\f0098";
}
.mdi-bed-single-outline::before {
  content: "\f0099";
}
.mdi-bee::before {
  content: "\ffc1";
}
.mdi-bee-flower::before {
  content: "\ffc2";
}
.mdi-beehive-outline::before {
  content: "\f00f9";
}
.mdi-beer::before {
  content: "\f098";
}
.mdi-beer-outline::before {
  content: "\f0337";
}
.mdi-behance::before {
  content: "\f099";
}
.mdi-bell::before {
  content: "\f09a";
}
.mdi-bell-alert::before {
  content: "\fd35";
}
.mdi-bell-alert-outline::before {
  content: "\fe9e";
}
.mdi-bell-check::before {
  content: "\f0210";
}
.mdi-bell-check-outline::before {
  content: "\f0211";
}
.mdi-bell-circle::before {
  content: "\fd36";
}
.mdi-bell-circle-outline::before {
  content: "\fd37";
}
.mdi-bell-off::before {
  content: "\f09b";
}
.mdi-bell-off-outline::before {
  content: "\fa90";
}
.mdi-bell-outline::before {
  content: "\f09c";
}
.mdi-bell-plus::before {
  content: "\f09d";
}
.mdi-bell-plus-outline::before {
  content: "\fa91";
}
.mdi-bell-ring::before {
  content: "\f09e";
}
.mdi-bell-ring-outline::before {
  content: "\f09f";
}
.mdi-bell-sleep::before {
  content: "\f0a0";
}
.mdi-bell-sleep-outline::before {
  content: "\fa92";
}
.mdi-beta::before {
  content: "\f0a1";
}
.mdi-betamax::before {
  content: "\f9ca";
}
.mdi-biathlon::before {
  content: "\fdf7";
}
.mdi-bible::before {
  content: "\f0a2";
}
.mdi-bicycle::before {
  content: "\f00c7";
}
.mdi-bicycle-basket::before {
  content: "\f0260";
}
.mdi-bike::before {
  content: "\f0a3";
}
.mdi-bike-fast::before {
  content: "\f014a";
}
.mdi-billboard::before {
  content: "\f0032";
}
.mdi-billiards::before {
  content: "\fb3d";
}
.mdi-billiards-rack::before {
  content: "\fb3e";
}
.mdi-bing::before {
  content: "\f0a4";
}
.mdi-binoculars::before {
  content: "\f0a5";
}
.mdi-bio::before {
  content: "\f0a6";
}
.mdi-biohazard::before {
  content: "\f0a7";
}
.mdi-bitbucket::before {
  content: "\f0a8";
}
.mdi-bitcoin::before {
  content: "\f812";
}
.mdi-black-mesa::before {
  content: "\f0a9";
}
.mdi-blackberry::before {
  content: "\f0aa";
}
.mdi-blender::before {
  content: "\fcc7";
}
.mdi-blender-software::before {
  content: "\f0ab";
}
.mdi-blinds::before {
  content: "\f0ac";
}
.mdi-blinds-open::before {
  content: "\f0033";
}
.mdi-block-helper::before {
  content: "\f0ad";
}
.mdi-blogger::before {
  content: "\f0ae";
}
.mdi-blood-bag::before {
  content: "\fcc8";
}
.mdi-bluetooth::before {
  content: "\f0af";
}
.mdi-bluetooth-audio::before {
  content: "\f0b0";
}
.mdi-bluetooth-connect::before {
  content: "\f0b1";
}
.mdi-bluetooth-off::before {
  content: "\f0b2";
}
.mdi-bluetooth-settings::before {
  content: "\f0b3";
}
.mdi-bluetooth-transfer::before {
  content: "\f0b4";
}
.mdi-blur::before {
  content: "\f0b5";
}
.mdi-blur-linear::before {
  content: "\f0b6";
}
.mdi-blur-off::before {
  content: "\f0b7";
}
.mdi-blur-radial::before {
  content: "\f0b8";
}
.mdi-bolnisi-cross::before {
  content: "\fcc9";
}
.mdi-bolt::before {
  content: "\fd8f";
}
.mdi-bomb::before {
  content: "\f690";
}
.mdi-bomb-off::before {
  content: "\f6c4";
}
.mdi-bone::before {
  content: "\f0b9";
}
.mdi-book::before {
  content: "\f0ba";
}
.mdi-book-information-variant::before {
  content: "\f009a";
}
.mdi-book-lock::before {
  content: "\f799";
}
.mdi-book-lock-open::before {
  content: "\f79a";
}
.mdi-book-minus::before {
  content: "\f5d9";
}
.mdi-book-minus-multiple::before {
  content: "\fa93";
}
.mdi-book-multiple::before {
  content: "\f0bb";
}
.mdi-book-open::before {
  content: "\f0bd";
}
.mdi-book-open-outline::before {
  content: "\fb3f";
}
.mdi-book-open-page-variant::before {
  content: "\f5da";
}
.mdi-book-open-variant::before {
  content: "\f0be";
}
.mdi-book-outline::before {
  content: "\fb40";
}
.mdi-book-play::before {
  content: "\fe9f";
}
.mdi-book-play-outline::before {
  content: "\fea0";
}
.mdi-book-plus::before {
  content: "\f5db";
}
.mdi-book-plus-multiple::before {
  content: "\fa94";
}
.mdi-book-remove::before {
  content: "\fa96";
}
.mdi-book-remove-multiple::before {
  content: "\fa95";
}
.mdi-book-search::before {
  content: "\fea1";
}
.mdi-book-search-outline::before {
  content: "\fea2";
}
.mdi-book-variant::before {
  content: "\f0bf";
}
.mdi-book-variant-multiple::before {
  content: "\f0bc";
}
.mdi-bookmark::before {
  content: "\f0c0";
}
.mdi-bookmark-check::before {
  content: "\f0c1";
}
.mdi-bookmark-check-outline::before {
  content: "\f03a6";
}
.mdi-bookmark-minus::before {
  content: "\f9cb";
}
.mdi-bookmark-minus-outline::before {
  content: "\f9cc";
}
.mdi-bookmark-multiple::before {
  content: "\fdf8";
}
.mdi-bookmark-multiple-outline::before {
  content: "\fdf9";
}
.mdi-bookmark-music::before {
  content: "\f0c2";
}
.mdi-bookmark-music-outline::before {
  content: "\f03a4";
}
.mdi-bookmark-off::before {
  content: "\f9cd";
}
.mdi-bookmark-off-outline::before {
  content: "\f9ce";
}
.mdi-bookmark-outline::before {
  content: "\f0c3";
}
.mdi-bookmark-plus::before {
  content: "\f0c5";
}
.mdi-bookmark-plus-outline::before {
  content: "\f0c4";
}
.mdi-bookmark-remove::before {
  content: "\f0c6";
}
.mdi-bookmark-remove-outline::before {
  content: "\f03a5";
}
.mdi-bookshelf::before {
  content: "\f028a";
}
.mdi-boom-gate::before {
  content: "\fea3";
}
.mdi-boom-gate-alert::before {
  content: "\fea4";
}
.mdi-boom-gate-alert-outline::before {
  content: "\fea5";
}
.mdi-boom-gate-down::before {
  content: "\fea6";
}
.mdi-boom-gate-down-outline::before {
  content: "\fea7";
}
.mdi-boom-gate-outline::before {
  content: "\fea8";
}
.mdi-boom-gate-up::before {
  content: "\fea9";
}
.mdi-boom-gate-up-outline::before {
  content: "\feaa";
}
.mdi-boombox::before {
  content: "\f5dc";
}
.mdi-boomerang::before {
  content: "\f00fa";
}
.mdi-bootstrap::before {
  content: "\f6c5";
}
.mdi-border-all::before {
  content: "\f0c7";
}
.mdi-border-all-variant::before {
  content: "\f8a0";
}
.mdi-border-bottom::before {
  content: "\f0c8";
}
.mdi-border-bottom-variant::before {
  content: "\f8a1";
}
.mdi-border-color::before {
  content: "\f0c9";
}
.mdi-border-horizontal::before {
  content: "\f0ca";
}
.mdi-border-inside::before {
  content: "\f0cb";
}
.mdi-border-left::before {
  content: "\f0cc";
}
.mdi-border-left-variant::before {
  content: "\f8a2";
}
.mdi-border-none::before {
  content: "\f0cd";
}
.mdi-border-none-variant::before {
  content: "\f8a3";
}
.mdi-border-outside::before {
  content: "\f0ce";
}
.mdi-border-right::before {
  content: "\f0cf";
}
.mdi-border-right-variant::before {
  content: "\f8a4";
}
.mdi-border-style::before {
  content: "\f0d0";
}
.mdi-border-top::before {
  content: "\f0d1";
}
.mdi-border-top-variant::before {
  content: "\f8a5";
}
.mdi-border-vertical::before {
  content: "\f0d2";
}
.mdi-bottle-soda::before {
  content: "\f009b";
}
.mdi-bottle-soda-classic::before {
  content: "\f009c";
}
.mdi-bottle-soda-classic-outline::before {
  content: "\f038e";
}
.mdi-bottle-soda-outline::before {
  content: "\f009d";
}
.mdi-bottle-tonic::before {
  content: "\f0159";
}
.mdi-bottle-tonic-outline::before {
  content: "\f015a";
}
.mdi-bottle-tonic-plus::before {
  content: "\f015b";
}
.mdi-bottle-tonic-plus-outline::before {
  content: "\f015c";
}
.mdi-bottle-tonic-skull::before {
  content: "\f015d";
}
.mdi-bottle-tonic-skull-outline::before {
  content: "\f015e";
}
.mdi-bottle-wine::before {
  content: "\f853";
}
.mdi-bottle-wine-outline::before {
  content: "\f033b";
}
.mdi-bow-tie::before {
  content: "\f677";
}
.mdi-bowl::before {
  content: "\f617";
}
.mdi-bowling::before {
  content: "\f0d3";
}
.mdi-box::before {
  content: "\f0d4";
}
.mdi-box-cutter::before {
  content: "\f0d5";
}
.mdi-box-shadow::before {
  content: "\f637";
}
.mdi-boxing-glove::before {
  content: "\fb41";
}
.mdi-braille::before {
  content: "\f9cf";
}
.mdi-brain::before {
  content: "\f9d0";
}
.mdi-bread-slice::before {
  content: "\fcca";
}
.mdi-bread-slice-outline::before {
  content: "\fccb";
}
.mdi-bridge::before {
  content: "\f618";
}
.mdi-briefcase::before {
  content: "\f0d6";
}
.mdi-briefcase-account::before {
  content: "\fccc";
}
.mdi-briefcase-account-outline::before {
  content: "\fccd";
}
.mdi-briefcase-check::before {
  content: "\f0d7";
}
.mdi-briefcase-check-outline::before {
  content: "\f0349";
}
.mdi-briefcase-clock::before {
  content: "\f00fb";
}
.mdi-briefcase-clock-outline::before {
  content: "\f00fc";
}
.mdi-briefcase-download::before {
  content: "\f0d8";
}
.mdi-briefcase-download-outline::before {
  content: "\fc19";
}
.mdi-briefcase-edit::before {
  content: "\fa97";
}
.mdi-briefcase-edit-outline::before {
  content: "\fc1a";
}
.mdi-briefcase-minus::before {
  content: "\fa29";
}
.mdi-briefcase-minus-outline::before {
  content: "\fc1b";
}
.mdi-briefcase-outline::before {
  content: "\f813";
}
.mdi-briefcase-plus::before {
  content: "\fa2a";
}
.mdi-briefcase-plus-outline::before {
  content: "\fc1c";
}
.mdi-briefcase-remove::before {
  content: "\fa2b";
}
.mdi-briefcase-remove-outline::before {
  content: "\fc1d";
}
.mdi-briefcase-search::before {
  content: "\fa2c";
}
.mdi-briefcase-search-outline::before {
  content: "\fc1e";
}
.mdi-briefcase-upload::before {
  content: "\f0d9";
}
.mdi-briefcase-upload-outline::before {
  content: "\fc1f";
}
.mdi-brightness-1::before {
  content: "\f0da";
}
.mdi-brightness-2::before {
  content: "\f0db";
}
.mdi-brightness-3::before {
  content: "\f0dc";
}
.mdi-brightness-4::before {
  content: "\f0dd";
}
.mdi-brightness-5::before {
  content: "\f0de";
}
.mdi-brightness-6::before {
  content: "\f0df";
}
.mdi-brightness-7::before {
  content: "\f0e0";
}
.mdi-brightness-auto::before {
  content: "\f0e1";
}
.mdi-brightness-percent::before {
  content: "\fcce";
}
.mdi-broom::before {
  content: "\f0e2";
}
.mdi-brush::before {
  content: "\f0e3";
}
.mdi-buddhism::before {
  content: "\f94a";
}
.mdi-buffer::before {
  content: "\f619";
}
.mdi-bug::before {
  content: "\f0e4";
}
.mdi-bug-check::before {
  content: "\fa2d";
}
.mdi-bug-check-outline::before {
  content: "\fa2e";
}
.mdi-bug-outline::before {
  content: "\fa2f";
}
.mdi-bugle::before {
  content: "\fd90";
}
.mdi-bulldozer::before {
  content: "\fb07";
}
.mdi-bullet::before {
  content: "\fccf";
}
.mdi-bulletin-board::before {
  content: "\f0e5";
}
.mdi-bullhorn::before {
  content: "\f0e6";
}
.mdi-bullhorn-outline::before {
  content: "\fb08";
}
.mdi-bullseye::before {
  content: "\f5dd";
}
.mdi-bullseye-arrow::before {
  content: "\f8c8";
}
.mdi-bulma::before {
  content: "\f0312";
}
.mdi-bunk-bed::before {
  content: "\f032d";
}
.mdi-bus::before {
  content: "\f0e7";
}
.mdi-bus-alert::before {
  content: "\fa98";
}
.mdi-bus-articulated-end::before {
  content: "\f79b";
}
.mdi-bus-articulated-front::before {
  content: "\f79c";
}
.mdi-bus-clock::before {
  content: "\f8c9";
}
.mdi-bus-double-decker::before {
  content: "\f79d";
}
.mdi-bus-marker::before {
  content: "\f023d";
}
.mdi-bus-multiple::before {
  content: "\ff5c";
}
.mdi-bus-school::before {
  content: "\f79e";
}
.mdi-bus-side::before {
  content: "\f79f";
}
.mdi-bus-stop::before {
  content: "\f0034";
}
.mdi-bus-stop-covered::before {
  content: "\f0035";
}
.mdi-bus-stop-uncovered::before {
  content: "\f0036";
}
.mdi-cached::before {
  content: "\f0e8";
}
.mdi-cactus::before {
  content: "\fd91";
}
.mdi-cake::before {
  content: "\f0e9";
}
.mdi-cake-layered::before {
  content: "\f0ea";
}
.mdi-cake-variant::before {
  content: "\f0eb";
}
.mdi-calculator::before {
  content: "\f0ec";
}
.mdi-calculator-variant::before {
  content: "\fa99";
}
.mdi-calendar::before {
  content: "\f0ed";
}
.mdi-calendar-account::before {
  content: "\fef4";
}
.mdi-calendar-account-outline::before {
  content: "\fef5";
}
.mdi-calendar-alert::before {
  content: "\fa30";
}
.mdi-calendar-arrow-left::before {
  content: "\f015f";
}
.mdi-calendar-arrow-right::before {
  content: "\f0160";
}
.mdi-calendar-blank::before {
  content: "\f0ee";
}
.mdi-calendar-blank-multiple::before {
  content: "\f009e";
}
.mdi-calendar-blank-outline::before {
  content: "\fb42";
}
.mdi-calendar-check::before {
  content: "\f0ef";
}
.mdi-calendar-check-outline::before {
  content: "\fc20";
}
.mdi-calendar-clock::before {
  content: "\f0f0";
}
.mdi-calendar-edit::before {
  content: "\f8a6";
}
.mdi-calendar-export::before {
  content: "\fb09";
}
.mdi-calendar-heart::before {
  content: "\f9d1";
}
.mdi-calendar-import::before {
  content: "\fb0a";
}
.mdi-calendar-minus::before {
  content: "\fd38";
}
.mdi-calendar-month::before {
  content: "\fdfa";
}
.mdi-calendar-month-outline::before {
  content: "\fdfb";
}
.mdi-calendar-multiple::before {
  content: "\f0f1";
}
.mdi-calendar-multiple-check::before {
  content: "\f0f2";
}
.mdi-calendar-multiselect::before {
  content: "\fa31";
}
.mdi-calendar-outline::before {
  content: "\fb43";
}
.mdi-calendar-plus::before {
  content: "\f0f3";
}
.mdi-calendar-question::before {
  content: "\f691";
}
.mdi-calendar-range::before {
  content: "\f678";
}
.mdi-calendar-range-outline::before {
  content: "\fb44";
}
.mdi-calendar-remove::before {
  content: "\f0f4";
}
.mdi-calendar-remove-outline::before {
  content: "\fc21";
}
.mdi-calendar-repeat::before {
  content: "\feab";
}
.mdi-calendar-repeat-outline::before {
  content: "\feac";
}
.mdi-calendar-search::before {
  content: "\f94b";
}
.mdi-calendar-star::before {
  content: "\f9d2";
}
.mdi-calendar-text::before {
  content: "\f0f5";
}
.mdi-calendar-text-outline::before {
  content: "\fc22";
}
.mdi-calendar-today::before {
  content: "\f0f6";
}
.mdi-calendar-week::before {
  content: "\fa32";
}
.mdi-calendar-week-begin::before {
  content: "\fa33";
}
.mdi-calendar-weekend::before {
  content: "\fef6";
}
.mdi-calendar-weekend-outline::before {
  content: "\fef7";
}
.mdi-call-made::before {
  content: "\f0f7";
}
.mdi-call-merge::before {
  content: "\f0f8";
}
.mdi-call-missed::before {
  content: "\f0f9";
}
.mdi-call-received::before {
  content: "\f0fa";
}
.mdi-call-split::before {
  content: "\f0fb";
}
.mdi-camcorder::before {
  content: "\f0fc";
}
.mdi-camcorder-box::before {
  content: "\f0fd";
}
.mdi-camcorder-box-off::before {
  content: "\f0fe";
}
.mdi-camcorder-off::before {
  content: "\f0ff";
}
.mdi-camera::before {
  content: "\f100";
}
.mdi-camera-account::before {
  content: "\f8ca";
}
.mdi-camera-burst::before {
  content: "\f692";
}
.mdi-camera-control::before {
  content: "\fb45";
}
.mdi-camera-enhance::before {
  content: "\f101";
}
.mdi-camera-enhance-outline::before {
  content: "\fb46";
}
.mdi-camera-front::before {
  content: "\f102";
}
.mdi-camera-front-variant::before {
  content: "\f103";
}
.mdi-camera-gopro::before {
  content: "\f7a0";
}
.mdi-camera-image::before {
  content: "\f8cb";
}
.mdi-camera-iris::before {
  content: "\f104";
}
.mdi-camera-metering-center::before {
  content: "\f7a1";
}
.mdi-camera-metering-matrix::before {
  content: "\f7a2";
}
.mdi-camera-metering-partial::before {
  content: "\f7a3";
}
.mdi-camera-metering-spot::before {
  content: "\f7a4";
}
.mdi-camera-off::before {
  content: "\f5df";
}
.mdi-camera-outline::before {
  content: "\fd39";
}
.mdi-camera-party-mode::before {
  content: "\f105";
}
.mdi-camera-plus::before {
  content: "\fef8";
}
.mdi-camera-plus-outline::before {
  content: "\fef9";
}
.mdi-camera-rear::before {
  content: "\f106";
}
.mdi-camera-rear-variant::before {
  content: "\f107";
}
.mdi-camera-retake::before {
  content: "\fdfc";
}
.mdi-camera-retake-outline::before {
  content: "\fdfd";
}
.mdi-camera-switch::before {
  content: "\f108";
}
.mdi-camera-timer::before {
  content: "\f109";
}
.mdi-camera-wireless::before {
  content: "\fd92";
}
.mdi-camera-wireless-outline::before {
  content: "\fd93";
}
.mdi-campfire::before {
  content: "\fefa";
}
.mdi-cancel::before {
  content: "\f739";
}
.mdi-candle::before {
  content: "\f5e2";
}
.mdi-candycane::before {
  content: "\f10a";
}
.mdi-cannabis::before {
  content: "\f7a5";
}
.mdi-caps-lock::before {
  content: "\fa9a";
}
.mdi-car::before {
  content: "\f10b";
}
.mdi-car-2-plus::before {
  content: "\f0037";
}
.mdi-car-3-plus::before {
  content: "\f0038";
}
.mdi-car-back::before {
  content: "\fdfe";
}
.mdi-car-battery::before {
  content: "\f10c";
}
.mdi-car-brake-abs::before {
  content: "\fc23";
}
.mdi-car-brake-alert::before {
  content: "\fc24";
}
.mdi-car-brake-hold::before {
  content: "\fd3a";
}
.mdi-car-brake-parking::before {
  content: "\fd3b";
}
.mdi-car-brake-retarder::before {
  content: "\f0039";
}
.mdi-car-child-seat::before {
  content: "\ffc3";
}
.mdi-car-clutch::before {
  content: "\f003a";
}
.mdi-car-connected::before {
  content: "\f10d";
}
.mdi-car-convertible::before {
  content: "\f7a6";
}
.mdi-car-coolant-level::before {
  content: "\f003b";
}
.mdi-car-cruise-control::before {
  content: "\fd3c";
}
.mdi-car-defrost-front::before {
  content: "\fd3d";
}
.mdi-car-defrost-rear::before {
  content: "\fd3e";
}
.mdi-car-door::before {
  content: "\fb47";
}
.mdi-car-door-lock::before {
  content: "\f00c8";
}
.mdi-car-electric::before {
  content: "\fb48";
}
.mdi-car-esp::before {
  content: "\fc25";
}
.mdi-car-estate::before {
  content: "\f7a7";
}
.mdi-car-hatchback::before {
  content: "\f7a8";
}
.mdi-car-info::before {
  content: "\f01e9";
}
.mdi-car-key::before {
  content: "\fb49";
}
.mdi-car-light-dimmed::before {
  content: "\fc26";
}
.mdi-car-light-fog::before {
  content: "\fc27";
}
.mdi-car-light-high::before {
  content: "\fc28";
}
.mdi-car-limousine::before {
  content: "\f8cc";
}
.mdi-car-multiple::before {
  content: "\fb4a";
}
.mdi-car-off::before {
  content: "\fdff";
}
.mdi-car-parking-lights::before {
  content: "\fd3f";
}
.mdi-car-pickup::before {
  content: "\f7a9";
}
.mdi-car-seat::before {
  content: "\ffc4";
}
.mdi-car-seat-cooler::before {
  content: "\ffc5";
}
.mdi-car-seat-heater::before {
  content: "\ffc6";
}
.mdi-car-shift-pattern::before {
  content: "\ff5d";
}
.mdi-car-side::before {
  content: "\f7aa";
}
.mdi-car-sports::before {
  content: "\f7ab";
}
.mdi-car-tire-alert::before {
  content: "\fc29";
}
.mdi-car-traction-control::before {
  content: "\fd40";
}
.mdi-car-turbocharger::before {
  content: "\f003c";
}
.mdi-car-wash::before {
  content: "\f10e";
}
.mdi-car-windshield::before {
  content: "\f003d";
}
.mdi-car-windshield-outline::before {
  content: "\f003e";
}
.mdi-caravan::before {
  content: "\f7ac";
}
.mdi-card::before {
  content: "\fb4b";
}
.mdi-card-bulleted::before {
  content: "\fb4c";
}
.mdi-card-bulleted-off::before {
  content: "\fb4d";
}
.mdi-card-bulleted-off-outline::before {
  content: "\fb4e";
}
.mdi-card-bulleted-outline::before {
  content: "\fb4f";
}
.mdi-card-bulleted-settings::before {
  content: "\fb50";
}
.mdi-card-bulleted-settings-outline::before {
  content: "\fb51";
}
.mdi-card-outline::before {
  content: "\fb52";
}
.mdi-card-plus::before {
  content: "\f022a";
}
.mdi-card-plus-outline::before {
  content: "\f022b";
}
.mdi-card-search::before {
  content: "\f009f";
}
.mdi-card-search-outline::before {
  content: "\f00a0";
}
.mdi-card-text::before {
  content: "\fb53";
}
.mdi-card-text-outline::before {
  content: "\fb54";
}
.mdi-cards::before {
  content: "\f638";
}
.mdi-cards-club::before {
  content: "\f8cd";
}
.mdi-cards-diamond::before {
  content: "\f8ce";
}
.mdi-cards-diamond-outline::before {
  content: "\f003f";
}
.mdi-cards-heart::before {
  content: "\f8cf";
}
.mdi-cards-outline::before {
  content: "\f639";
}
.mdi-cards-playing-outline::before {
  content: "\f63a";
}
.mdi-cards-spade::before {
  content: "\f8d0";
}
.mdi-cards-variant::before {
  content: "\f6c6";
}
.mdi-carrot::before {
  content: "\f10f";
}
.mdi-cart::before {
  content: "\f110";
}
.mdi-cart-arrow-down::before {
  content: "\fd42";
}
.mdi-cart-arrow-right::before {
  content: "\fc2a";
}
.mdi-cart-arrow-up::before {
  content: "\fd43";
}
.mdi-cart-minus::before {
  content: "\fd44";
}
.mdi-cart-off::before {
  content: "\f66b";
}
.mdi-cart-outline::before {
  content: "\f111";
}
.mdi-cart-plus::before {
  content: "\f112";
}
.mdi-cart-remove::before {
  content: "\fd45";
}
.mdi-case-sensitive-alt::before {
  content: "\f113";
}
.mdi-cash::before {
  content: "\f114";
}
.mdi-cash-100::before {
  content: "\f115";
}
.mdi-cash-marker::before {
  content: "\fd94";
}
.mdi-cash-minus::before {
  content: "\f028b";
}
.mdi-cash-multiple::before {
  content: "\f116";
}
.mdi-cash-plus::before {
  content: "\f028c";
}
.mdi-cash-refund::before {
  content: "\fa9b";
}
.mdi-cash-register::before {
  content: "\fcd0";
}
.mdi-cash-remove::before {
  content: "\f028d";
}
.mdi-cash-usd::before {
  content: "\f01a1";
}
.mdi-cash-usd-outline::before {
  content: "\f117";
}
.mdi-cassette::before {
  content: "\f9d3";
}
.mdi-cast::before {
  content: "\f118";
}
.mdi-cast-audio::before {
  content: "\f0040";
}
.mdi-cast-connected::before {
  content: "\f119";
}
.mdi-cast-education::before {
  content: "\fe6d";
}
.mdi-cast-off::before {
  content: "\f789";
}
.mdi-castle::before {
  content: "\f11a";
}
.mdi-cat::before {
  content: "\f11b";
}
.mdi-cctv::before {
  content: "\f7ad";
}
.mdi-ceiling-light::before {
  content: "\f768";
}
.mdi-cellphone::before {
  content: "\f11c";
}
.mdi-cellphone-android::before {
  content: "\f11d";
}
.mdi-cellphone-arrow-down::before {
  content: "\f9d4";
}
.mdi-cellphone-basic::before {
  content: "\f11e";
}
.mdi-cellphone-dock::before {
  content: "\f11f";
}
.mdi-cellphone-erase::before {
  content: "\f94c";
}
.mdi-cellphone-information::before {
  content: "\ff5e";
}
.mdi-cellphone-iphone::before {
  content: "\f120";
}
.mdi-cellphone-key::before {
  content: "\f94d";
}
.mdi-cellphone-link::before {
  content: "\f121";
}
.mdi-cellphone-link-off::before {
  content: "\f122";
}
.mdi-cellphone-lock::before {
  content: "\f94e";
}
.mdi-cellphone-message::before {
  content: "\f8d2";
}
.mdi-cellphone-message-off::before {
  content: "\f00fd";
}
.mdi-cellphone-nfc::before {
  content: "\fead";
}
.mdi-cellphone-nfc-off::before {
  content: "\f0303";
}
.mdi-cellphone-off::before {
  content: "\f94f";
}
.mdi-cellphone-play::before {
  content: "\f0041";
}
.mdi-cellphone-screenshot::before {
  content: "\fa34";
}
.mdi-cellphone-settings::before {
  content: "\f123";
}
.mdi-cellphone-settings-variant::before {
  content: "\f950";
}
.mdi-cellphone-sound::before {
  content: "\f951";
}
.mdi-cellphone-text::before {
  content: "\f8d1";
}
.mdi-cellphone-wireless::before {
  content: "\f814";
}
.mdi-celtic-cross::before {
  content: "\fcd1";
}
.mdi-centos::before {
  content: "\f0145";
}
.mdi-certificate::before {
  content: "\f124";
}
.mdi-certificate-outline::before {
  content: "\f01b3";
}
.mdi-chair-rolling::before {
  content: "\ffba";
}
.mdi-chair-school::before {
  content: "\f125";
}
.mdi-charity::before {
  content: "\fc2b";
}
.mdi-chart-arc::before {
  content: "\f126";
}
.mdi-chart-areaspline::before {
  content: "\f127";
}
.mdi-chart-areaspline-variant::before {
  content: "\feae";
}
.mdi-chart-bar::before {
  content: "\f128";
}
.mdi-chart-bar-stacked::before {
  content: "\f769";
}
.mdi-chart-bell-curve::before {
  content: "\fc2c";
}
.mdi-chart-bell-curve-cumulative::before {
  content: "\ffc7";
}
.mdi-chart-bubble::before {
  content: "\f5e3";
}
.mdi-chart-donut::before {
  content: "\f7ae";
}
.mdi-chart-donut-variant::before {
  content: "\f7af";
}
.mdi-chart-gantt::before {
  content: "\f66c";
}
.mdi-chart-histogram::before {
  content: "\f129";
}
.mdi-chart-line::before {
  content: "\f12a";
}
.mdi-chart-line-stacked::before {
  content: "\f76a";
}
.mdi-chart-line-variant::before {
  content: "\f7b0";
}
.mdi-chart-multiline::before {
  content: "\f8d3";
}
.mdi-chart-multiple::before {
  content: "\f023e";
}
.mdi-chart-pie::before {
  content: "\f12b";
}
.mdi-chart-ppf::before {
  content: "\f03ab";
}
.mdi-chart-scatter-plot::before {
  content: "\feaf";
}
.mdi-chart-scatter-plot-hexbin::before {
  content: "\f66d";
}
.mdi-chart-snakey::before {
  content: "\f020a";
}
.mdi-chart-snakey-variant::before {
  content: "\f020b";
}
.mdi-chart-timeline::before {
  content: "\f66e";
}
.mdi-chart-timeline-variant::before {
  content: "\feb0";
}
.mdi-chart-tree::before {
  content: "\feb1";
}
.mdi-chat::before {
  content: "\fb55";
}
.mdi-chat-alert::before {
  content: "\fb56";
}
.mdi-chat-alert-outline::before {
  content: "\f02f4";
}
.mdi-chat-outline::before {
  content: "\fefb";
}
.mdi-chat-processing::before {
  content: "\fb57";
}
.mdi-chat-processing-outline::before {
  content: "\f02f5";
}
.mdi-chat-sleep::before {
  content: "\f02fc";
}
.mdi-chat-sleep-outline::before {
  content: "\f02fd";
}
.mdi-check::before {
  content: "\f12c";
}
.mdi-check-all::before {
  content: "\f12d";
}
.mdi-check-bold::before {
  content: "\fe6e";
}
.mdi-check-box-multiple-outline::before {
  content: "\fc2d";
}
.mdi-check-box-outline::before {
  content: "\fc2e";
}
.mdi-check-circle::before {
  content: "\f5e0";
}
.mdi-check-circle-outline::before {
  content: "\f5e1";
}
.mdi-check-decagram::before {
  content: "\f790";
}
.mdi-check-network::before {
  content: "\fc2f";
}
.mdi-check-network-outline::before {
  content: "\fc30";
}
.mdi-check-outline::before {
  content: "\f854";
}
.mdi-check-underline::before {
  content: "\fe70";
}
.mdi-check-underline-circle::before {
  content: "\fe71";
}
.mdi-check-underline-circle-outline::before {
  content: "\fe72";
}
.mdi-checkbook::before {
  content: "\fa9c";
}
.mdi-checkbox-blank::before {
  content: "\f12e";
}
.mdi-checkbox-blank-circle::before {
  content: "\f12f";
}
.mdi-checkbox-blank-circle-outline::before {
  content: "\f130";
}
.mdi-checkbox-blank-off::before {
  content: "\f0317";
}
.mdi-checkbox-blank-off-outline::before {
  content: "\f0318";
}
.mdi-checkbox-blank-outline::before {
  content: "\f131";
}
.mdi-checkbox-intermediate::before {
  content: "\f855";
}
.mdi-checkbox-marked::before {
  content: "\f132";
}
.mdi-checkbox-marked-circle::before {
  content: "\f133";
}
.mdi-checkbox-marked-circle-outline::before {
  content: "\f134";
}
.mdi-checkbox-marked-outline::before {
  content: "\f135";
}
.mdi-checkbox-multiple-blank::before {
  content: "\f136";
}
.mdi-checkbox-multiple-blank-circle::before {
  content: "\f63b";
}
.mdi-checkbox-multiple-blank-circle-outline::before {
  content: "\f63c";
}
.mdi-checkbox-multiple-blank-outline::before {
  content: "\f137";
}
.mdi-checkbox-multiple-marked::before {
  content: "\f138";
}
.mdi-checkbox-multiple-marked-circle::before {
  content: "\f63d";
}
.mdi-checkbox-multiple-marked-circle-outline::before {
  content: "\f63e";
}
.mdi-checkbox-multiple-marked-outline::before {
  content: "\f139";
}
.mdi-checkerboard::before {
  content: "\f13a";
}
.mdi-checkerboard-minus::before {
  content: "\f022d";
}
.mdi-checkerboard-plus::before {
  content: "\f022c";
}
.mdi-checkerboard-remove::before {
  content: "\f022e";
}
.mdi-cheese::before {
  content: "\f02e4";
}
.mdi-chef-hat::before {
  content: "\fb58";
}
.mdi-chemical-weapon::before {
  content: "\f13b";
}
.mdi-chess-bishop::before {
  content: "\f85b";
}
.mdi-chess-king::before {
  content: "\f856";
}
.mdi-chess-knight::before {
  content: "\f857";
}
.mdi-chess-pawn::before {
  content: "\f858";
}
.mdi-chess-queen::before {
  content: "\f859";
}
.mdi-chess-rook::before {
  content: "\f85a";
}
.mdi-chevron-double-down::before {
  content: "\f13c";
}
.mdi-chevron-double-left::before {
  content: "\f13d";
}
.mdi-chevron-double-right::before {
  content: "\f13e";
}
.mdi-chevron-double-up::before {
  content: "\f13f";
}
.mdi-chevron-down::before {
  content: "\f140";
}
.mdi-chevron-down-box::before {
  content: "\f9d5";
}
.mdi-chevron-down-box-outline::before {
  content: "\f9d6";
}
.mdi-chevron-down-circle::before {
  content: "\fb0b";
}
.mdi-chevron-down-circle-outline::before {
  content: "\fb0c";
}
.mdi-chevron-left::before {
  content: "\f141";
}
.mdi-chevron-left-box::before {
  content: "\f9d7";
}
.mdi-chevron-left-box-outline::before {
  content: "\f9d8";
}
.mdi-chevron-left-circle::before {
  content: "\fb0d";
}
.mdi-chevron-left-circle-outline::before {
  content: "\fb0e";
}
.mdi-chevron-right::before {
  content: "\f142";
}
.mdi-chevron-right-box::before {
  content: "\f9d9";
}
.mdi-chevron-right-box-outline::before {
  content: "\f9da";
}
.mdi-chevron-right-circle::before {
  content: "\fb0f";
}
.mdi-chevron-right-circle-outline::before {
  content: "\fb10";
}
.mdi-chevron-triple-down::before {
  content: "\fd95";
}
.mdi-chevron-triple-left::before {
  content: "\fd96";
}
.mdi-chevron-triple-right::before {
  content: "\fd97";
}
.mdi-chevron-triple-up::before {
  content: "\fd98";
}
.mdi-chevron-up::before {
  content: "\f143";
}
.mdi-chevron-up-box::before {
  content: "\f9db";
}
.mdi-chevron-up-box-outline::before {
  content: "\f9dc";
}
.mdi-chevron-up-circle::before {
  content: "\fb11";
}
.mdi-chevron-up-circle-outline::before {
  content: "\fb12";
}
.mdi-chili-hot::before {
  content: "\f7b1";
}
.mdi-chili-medium::before {
  content: "\f7b2";
}
.mdi-chili-mild::before {
  content: "\f7b3";
}
.mdi-chip::before {
  content: "\f61a";
}
.mdi-christianity::before {
  content: "\f952";
}
.mdi-christianity-outline::before {
  content: "\fcd2";
}
.mdi-church::before {
  content: "\f144";
}
.mdi-cigar::before {
  content: "\f01b4";
}
.mdi-circle::before {
  content: "\f764";
}
.mdi-circle-double::before {
  content: "\feb2";
}
.mdi-circle-edit-outline::before {
  content: "\f8d4";
}
.mdi-circle-expand::before {
  content: "\feb3";
}
.mdi-circle-medium::before {
  content: "\f9dd";
}
.mdi-circle-off-outline::before {
  content: "\f00fe";
}
.mdi-circle-outline::before {
  content: "\f765";
}
.mdi-circle-slice-1::before {
  content: "\fa9d";
}
.mdi-circle-slice-2::before {
  content: "\fa9e";
}
.mdi-circle-slice-3::before {
  content: "\fa9f";
}
.mdi-circle-slice-4::before {
  content: "\faa0";
}
.mdi-circle-slice-5::before {
  content: "\faa1";
}
.mdi-circle-slice-6::before {
  content: "\faa2";
}
.mdi-circle-slice-7::before {
  content: "\faa3";
}
.mdi-circle-slice-8::before {
  content: "\faa4";
}
.mdi-circle-small::before {
  content: "\f9de";
}
.mdi-circular-saw::before {
  content: "\fe73";
}
.mdi-cisco-webex::before {
  content: "\f145";
}
.mdi-city::before {
  content: "\f146";
}
.mdi-city-variant::before {
  content: "\fa35";
}
.mdi-city-variant-outline::before {
  content: "\fa36";
}
.mdi-clipboard::before {
  content: "\f147";
}
.mdi-clipboard-account::before {
  content: "\f148";
}
.mdi-clipboard-account-outline::before {
  content: "\fc31";
}
.mdi-clipboard-alert::before {
  content: "\f149";
}
.mdi-clipboard-alert-outline::before {
  content: "\fcd3";
}
.mdi-clipboard-arrow-down::before {
  content: "\f14a";
}
.mdi-clipboard-arrow-down-outline::before {
  content: "\fc32";
}
.mdi-clipboard-arrow-left::before {
  content: "\f14b";
}
.mdi-clipboard-arrow-left-outline::before {
  content: "\fcd4";
}
.mdi-clipboard-arrow-right::before {
  content: "\fcd5";
}
.mdi-clipboard-arrow-right-outline::before {
  content: "\fcd6";
}
.mdi-clipboard-arrow-up::before {
  content: "\fc33";
}
.mdi-clipboard-arrow-up-outline::before {
  content: "\fc34";
}
.mdi-clipboard-check::before {
  content: "\f14c";
}
.mdi-clipboard-check-multiple::before {
  content: "\f028e";
}
.mdi-clipboard-check-multiple-outline::before {
  content: "\f028f";
}
.mdi-clipboard-check-outline::before {
  content: "\f8a7";
}
.mdi-clipboard-file::before {
  content: "\f0290";
}
.mdi-clipboard-file-outline::before {
  content: "\f0291";
}
.mdi-clipboard-flow::before {
  content: "\f6c7";
}
.mdi-clipboard-flow-outline::before {
  content: "\f0142";
}
.mdi-clipboard-list::before {
  content: "\f00ff";
}
.mdi-clipboard-list-outline::before {
  content: "\f0100";
}
.mdi-clipboard-multiple::before {
  content: "\f0292";
}
.mdi-clipboard-multiple-outline::before {
  content: "\f0293";
}
.mdi-clipboard-outline::before {
  content: "\f14d";
}
.mdi-clipboard-play::before {
  content: "\fc35";
}
.mdi-clipboard-play-multiple::before {
  content: "\f0294";
}
.mdi-clipboard-play-multiple-outline::before {
  content: "\f0295";
}
.mdi-clipboard-play-outline::before {
  content: "\fc36";
}
.mdi-clipboard-plus::before {
  content: "\f750";
}
.mdi-clipboard-plus-outline::before {
  content: "\f034a";
}
.mdi-clipboard-pulse::before {
  content: "\f85c";
}
.mdi-clipboard-pulse-outline::before {
  content: "\f85d";
}
.mdi-clipboard-text::before {
  content: "\f14e";
}
.mdi-clipboard-text-multiple::before {
  content: "\f0296";
}
.mdi-clipboard-text-multiple-outline::before {
  content: "\f0297";
}
.mdi-clipboard-text-outline::before {
  content: "\fa37";
}
.mdi-clipboard-text-play::before {
  content: "\fc37";
}
.mdi-clipboard-text-play-outline::before {
  content: "\fc38";
}
.mdi-clippy::before {
  content: "\f14f";
}
.mdi-clock::before {
  content: "\f953";
}
.mdi-clock-alert::before {
  content: "\f954";
}
.mdi-clock-alert-outline::before {
  content: "\f5ce";
}
.mdi-clock-check::before {
  content: "\ffc8";
}
.mdi-clock-check-outline::before {
  content: "\ffc9";
}
.mdi-clock-digital::before {
  content: "\feb4";
}
.mdi-clock-end::before {
  content: "\f151";
}
.mdi-clock-fast::before {
  content: "\f152";
}
.mdi-clock-in::before {
  content: "\f153";
}
.mdi-clock-out::before {
  content: "\f154";
}
.mdi-clock-outline::before {
  content: "\f150";
}
.mdi-clock-start::before {
  content: "\f155";
}
.mdi-close::before {
  content: "\f156";
}
.mdi-close-box::before {
  content: "\f157";
}
.mdi-close-box-multiple::before {
  content: "\fc39";
}
.mdi-close-box-multiple-outline::before {
  content: "\fc3a";
}
.mdi-close-box-outline::before {
  content: "\f158";
}
.mdi-close-circle::before {
  content: "\f159";
}
.mdi-close-circle-outline::before {
  content: "\f15a";
}
.mdi-close-network::before {
  content: "\f15b";
}
.mdi-close-network-outline::before {
  content: "\fc3b";
}
.mdi-close-octagon::before {
  content: "\f15c";
}
.mdi-close-octagon-outline::before {
  content: "\f15d";
}
.mdi-close-outline::before {
  content: "\f6c8";
}
.mdi-closed-caption::before {
  content: "\f15e";
}
.mdi-closed-caption-outline::before {
  content: "\fd99";
}
.mdi-cloud::before {
  content: "\f15f";
}
.mdi-cloud-alert::before {
  content: "\f9df";
}
.mdi-cloud-braces::before {
  content: "\f7b4";
}
.mdi-cloud-check::before {
  content: "\f160";
}
.mdi-cloud-check-outline::before {
  content: "\f02f7";
}
.mdi-cloud-circle::before {
  content: "\f161";
}
.mdi-cloud-download::before {
  content: "\f162";
}
.mdi-cloud-download-outline::before {
  content: "\fb59";
}
.mdi-cloud-lock::before {
  content: "\f021c";
}
.mdi-cloud-lock-outline::before {
  content: "\f021d";
}
.mdi-cloud-off-outline::before {
  content: "\f164";
}
.mdi-cloud-outline::before {
  content: "\f163";
}
.mdi-cloud-print::before {
  content: "\f165";
}
.mdi-cloud-print-outline::before {
  content: "\f166";
}
.mdi-cloud-question::before {
  content: "\fa38";
}
.mdi-cloud-search::before {
  content: "\f955";
}
.mdi-cloud-search-outline::before {
  content: "\f956";
}
.mdi-cloud-sync::before {
  content: "\f63f";
}
.mdi-cloud-sync-outline::before {
  content: "\f0301";
}
.mdi-cloud-tags::before {
  content: "\f7b5";
}
.mdi-cloud-upload::before {
  content: "\f167";
}
.mdi-cloud-upload-outline::before {
  content: "\fb5a";
}
.mdi-clover::before {
  content: "\f815";
}
.mdi-coach-lamp::before {
  content: "\f0042";
}
.mdi-coat-rack::before {
  content: "\f00c9";
}
.mdi-code-array::before {
  content: "\f168";
}
.mdi-code-braces::before {
  content: "\f169";
}
.mdi-code-braces-box::before {
  content: "\f0101";
}
.mdi-code-brackets::before {
  content: "\f16a";
}
.mdi-code-equal::before {
  content: "\f16b";
}
.mdi-code-greater-than::before {
  content: "\f16c";
}
.mdi-code-greater-than-or-equal::before {
  content: "\f16d";
}
.mdi-code-less-than::before {
  content: "\f16e";
}
.mdi-code-less-than-or-equal::before {
  content: "\f16f";
}
.mdi-code-not-equal::before {
  content: "\f170";
}
.mdi-code-not-equal-variant::before {
  content: "\f171";
}
.mdi-code-parentheses::before {
  content: "\f172";
}
.mdi-code-parentheses-box::before {
  content: "\f0102";
}
.mdi-code-string::before {
  content: "\f173";
}
.mdi-code-tags::before {
  content: "\f174";
}
.mdi-code-tags-check::before {
  content: "\f693";
}
.mdi-codepen::before {
  content: "\f175";
}
.mdi-coffee::before {
  content: "\f176";
}
.mdi-coffee-maker::before {
  content: "\f00ca";
}
.mdi-coffee-off::before {
  content: "\ffca";
}
.mdi-coffee-off-outline::before {
  content: "\ffcb";
}
.mdi-coffee-outline::before {
  content: "\f6c9";
}
.mdi-coffee-to-go::before {
  content: "\f177";
}
.mdi-coffee-to-go-outline::before {
  content: "\f0339";
}
.mdi-coffin::before {
  content: "\fb5b";
}
.mdi-cog-clockwise::before {
  content: "\f0208";
}
.mdi-cog-counterclockwise::before {
  content: "\f0209";
}
.mdi-cogs::before {
  content: "\f8d5";
}
.mdi-coin::before {
  content: "\f0196";
}
.mdi-coin-outline::before {
  content: "\f178";
}
.mdi-coins::before {
  content: "\f694";
}
.mdi-collage::before {
  content: "\f640";
}
.mdi-collapse-all::before {
  content: "\faa5";
}
.mdi-collapse-all-outline::before {
  content: "\faa6";
}
.mdi-color-helper::before {
  content: "\f179";
}
.mdi-comma::before {
  content: "\fe74";
}
.mdi-comma-box::before {
  content: "\fe75";
}
.mdi-comma-box-outline::before {
  content: "\fe76";
}
.mdi-comma-circle::before {
  content: "\fe77";
}
.mdi-comma-circle-outline::before {
  content: "\fe78";
}
.mdi-comment::before {
  content: "\f17a";
}
.mdi-comment-account::before {
  content: "\f17b";
}
.mdi-comment-account-outline::before {
  content: "\f17c";
}
.mdi-comment-alert::before {
  content: "\f17d";
}
.mdi-comment-alert-outline::before {
  content: "\f17e";
}
.mdi-comment-arrow-left::before {
  content: "\f9e0";
}
.mdi-comment-arrow-left-outline::before {
  content: "\f9e1";
}
.mdi-comment-arrow-right::before {
  content: "\f9e2";
}
.mdi-comment-arrow-right-outline::before {
  content: "\f9e3";
}
.mdi-comment-check::before {
  content: "\f17f";
}
.mdi-comment-check-outline::before {
  content: "\f180";
}
.mdi-comment-edit::before {
  content: "\f01ea";
}
.mdi-comment-edit-outline::before {
  content: "\f02ef";
}
.mdi-comment-eye::before {
  content: "\fa39";
}
.mdi-comment-eye-outline::before {
  content: "\fa3a";
}
.mdi-comment-multiple::before {
  content: "\f85e";
}
.mdi-comment-multiple-outline::before {
  content: "\f181";
}
.mdi-comment-outline::before {
  content: "\f182";
}
.mdi-comment-plus::before {
  content: "\f9e4";
}
.mdi-comment-plus-outline::before {
  content: "\f183";
}
.mdi-comment-processing::before {
  content: "\f184";
}
.mdi-comment-processing-outline::before {
  content: "\f185";
}
.mdi-comment-question::before {
  content: "\f816";
}
.mdi-comment-question-outline::before {
  content: "\f186";
}
.mdi-comment-quote::before {
  content: "\f0043";
}
.mdi-comment-quote-outline::before {
  content: "\f0044";
}
.mdi-comment-remove::before {
  content: "\f5de";
}
.mdi-comment-remove-outline::before {
  content: "\f187";
}
.mdi-comment-search::before {
  content: "\fa3b";
}
.mdi-comment-search-outline::before {
  content: "\fa3c";
}
.mdi-comment-text::before {
  content: "\f188";
}
.mdi-comment-text-multiple::before {
  content: "\f85f";
}
.mdi-comment-text-multiple-outline::before {
  content: "\f860";
}
.mdi-comment-text-outline::before {
  content: "\f189";
}
.mdi-compare::before {
  content: "\f18a";
}
.mdi-compass::before {
  content: "\f18b";
}
.mdi-compass-off::before {
  content: "\fb5c";
}
.mdi-compass-off-outline::before {
  content: "\fb5d";
}
.mdi-compass-outline::before {
  content: "\f18c";
}
.mdi-compass-rose::before {
  content: "\f03ad";
}
.mdi-concourse-ci::before {
  content: "\f00cb";
}
.mdi-console::before {
  content: "\f18d";
}
.mdi-console-line::before {
  content: "\f7b6";
}
.mdi-console-network::before {
  content: "\f8a8";
}
.mdi-console-network-outline::before {
  content: "\fc3c";
}
.mdi-consolidate::before {
  content: "\f0103";
}
.mdi-contact-mail::before {
  content: "\f18e";
}
.mdi-contact-mail-outline::before {
  content: "\feb5";
}
.mdi-contact-phone::before {
  content: "\feb6";
}
.mdi-contact-phone-outline::before {
  content: "\feb7";
}
.mdi-contactless-payment::before {
  content: "\fd46";
}
.mdi-contacts::before {
  content: "\f6ca";
}
.mdi-contain::before {
  content: "\fa3d";
}
.mdi-contain-end::before {
  content: "\fa3e";
}
.mdi-contain-start::before {
  content: "\fa3f";
}
.mdi-content-copy::before {
  content: "\f18f";
}
.mdi-content-cut::before {
  content: "\f190";
}
.mdi-content-duplicate::before {
  content: "\f191";
}
.mdi-content-paste::before {
  content: "\f192";
}
.mdi-content-save::before {
  content: "\f193";
}
.mdi-content-save-alert::before {
  content: "\ff5f";
}
.mdi-content-save-alert-outline::before {
  content: "\ff60";
}
.mdi-content-save-all::before {
  content: "\f194";
}
.mdi-content-save-all-outline::before {
  content: "\ff61";
}
.mdi-content-save-edit::before {
  content: "\fcd7";
}
.mdi-content-save-edit-outline::before {
  content: "\fcd8";
}
.mdi-content-save-move::before {
  content: "\fe79";
}
.mdi-content-save-move-outline::before {
  content: "\fe7a";
}
.mdi-content-save-outline::before {
  content: "\f817";
}
.mdi-content-save-settings::before {
  content: "\f61b";
}
.mdi-content-save-settings-outline::before {
  content: "\fb13";
}
.mdi-contrast::before {
  content: "\f195";
}
.mdi-contrast-box::before {
  content: "\f196";
}
.mdi-contrast-circle::before {
  content: "\f197";
}
.mdi-controller-classic::before {
  content: "\fb5e";
}
.mdi-controller-classic-outline::before {
  content: "\fb5f";
}
.mdi-cookie::before {
  content: "\f198";
}
.mdi-coolant-temperature::before {
  content: "\f3c8";
}
.mdi-copyright::before {
  content: "\f5e6";
}
.mdi-cordova::before {
  content: "\f957";
}
.mdi-corn::before {
  content: "\f7b7";
}
.mdi-counter::before {
  content: "\f199";
}
.mdi-cow::before {
  content: "\f19a";
}
.mdi-cowboy::before {
  content: "\feb8";
}
.mdi-cpu-32-bit::before {
  content: "\fefc";
}
.mdi-cpu-64-bit::before {
  content: "\fefd";
}
.mdi-crane::before {
  content: "\f861";
}
.mdi-creation::before {
  content: "\f1c9";
}
.mdi-creative-commons::before {
  content: "\fd47";
}
.mdi-credit-card::before {
  content: "\f0010";
}
.mdi-credit-card-clock::before {
  content: "\fefe";
}
.mdi-credit-card-clock-outline::before {
  content: "\ffbc";
}
.mdi-credit-card-marker::before {
  content: "\f6a7";
}
.mdi-credit-card-marker-outline::before {
  content: "\fd9a";
}
.mdi-credit-card-minus::before {
  content: "\ffcc";
}
.mdi-credit-card-minus-outline::before {
  content: "\ffcd";
}
.mdi-credit-card-multiple::before {
  content: "\f0011";
}
.mdi-credit-card-multiple-outline::before {
  content: "\f19c";
}
.mdi-credit-card-off::before {
  content: "\f0012";
}
.mdi-credit-card-off-outline::before {
  content: "\f5e4";
}
.mdi-credit-card-outline::before {
  content: "\f19b";
}
.mdi-credit-card-plus::before {
  content: "\f0013";
}
.mdi-credit-card-plus-outline::before {
  content: "\f675";
}
.mdi-credit-card-refund::before {
  content: "\f0014";
}
.mdi-credit-card-refund-outline::before {
  content: "\faa7";
}
.mdi-credit-card-remove::before {
  content: "\ffce";
}
.mdi-credit-card-remove-outline::before {
  content: "\ffcf";
}
.mdi-credit-card-scan::before {
  content: "\f0015";
}
.mdi-credit-card-scan-outline::before {
  content: "\f19d";
}
.mdi-credit-card-settings::before {
  content: "\f0016";
}
.mdi-credit-card-settings-outline::before {
  content: "\f8d6";
}
.mdi-credit-card-wireless::before {
  content: "\f801";
}
.mdi-credit-card-wireless-outline::before {
  content: "\fd48";
}
.mdi-cricket::before {
  content: "\fd49";
}
.mdi-crop::before {
  content: "\f19e";
}
.mdi-crop-free::before {
  content: "\f19f";
}
.mdi-crop-landscape::before {
  content: "\f1a0";
}
.mdi-crop-portrait::before {
  content: "\f1a1";
}
.mdi-crop-rotate::before {
  content: "\f695";
}
.mdi-crop-square::before {
  content: "\f1a2";
}
.mdi-crosshairs::before {
  content: "\f1a3";
}
.mdi-crosshairs-gps::before {
  content: "\f1a4";
}
.mdi-crosshairs-off::before {
  content: "\ff62";
}
.mdi-crosshairs-question::before {
  content: "\f0161";
}
.mdi-crown::before {
  content: "\f1a5";
}
.mdi-crown-outline::before {
  content: "\f01fb";
}
.mdi-cryengine::before {
  content: "\f958";
}
.mdi-crystal-ball::before {
  content: "\fb14";
}
.mdi-cube::before {
  content: "\f1a6";
}
.mdi-cube-outline::before {
  content: "\f1a7";
}
.mdi-cube-scan::before {
  content: "\fb60";
}
.mdi-cube-send::before {
  content: "\f1a8";
}
.mdi-cube-unfolded::before {
  content: "\f1a9";
}
.mdi-cup::before {
  content: "\f1aa";
}
.mdi-cup-off::before {
  content: "\f5e5";
}
.mdi-cup-off-outline::before {
  content: "\f03a8";
}
.mdi-cup-outline::before {
  content: "\f033a";
}
.mdi-cup-water::before {
  content: "\f1ab";
}
.mdi-cupboard::before {
  content: "\ff63";
}
.mdi-cupboard-outline::before {
  content: "\ff64";
}
.mdi-cupcake::before {
  content: "\f959";
}
.mdi-curling::before {
  content: "\f862";
}
.mdi-currency-bdt::before {
  content: "\f863";
}
.mdi-currency-brl::before {
  content: "\fb61";
}
.mdi-currency-btc::before {
  content: "\f1ac";
}
.mdi-currency-cny::before {
  content: "\f7b9";
}
.mdi-currency-eth::before {
  content: "\f7ba";
}
.mdi-currency-eur::before {
  content: "\f1ad";
}
.mdi-currency-eur-off::before {
  content: "\f0340";
}
.mdi-currency-gbp::before {
  content: "\f1ae";
}
.mdi-currency-ils::before {
  content: "\fc3d";
}
.mdi-currency-inr::before {
  content: "\f1af";
}
.mdi-currency-jpy::before {
  content: "\f7bb";
}
.mdi-currency-krw::before {
  content: "\f7bc";
}
.mdi-currency-kzt::before {
  content: "\f864";
}
.mdi-currency-ngn::before {
  content: "\f1b0";
}
.mdi-currency-php::before {
  content: "\f9e5";
}
.mdi-currency-rial::before {
  content: "\feb9";
}
.mdi-currency-rub::before {
  content: "\f1b1";
}
.mdi-currency-sign::before {
  content: "\f7bd";
}
.mdi-currency-try::before {
  content: "\f1b2";
}
.mdi-currency-twd::before {
  content: "\f7be";
}
.mdi-currency-usd::before {
  content: "\f1b3";
}
.mdi-currency-usd-off::before {
  content: "\f679";
}
.mdi-current-ac::before {
  content: "\f95a";
}
.mdi-current-dc::before {
  content: "\f95b";
}
.mdi-cursor-default::before {
  content: "\f1b4";
}
.mdi-cursor-default-click::before {
  content: "\fcd9";
}
.mdi-cursor-default-click-outline::before {
  content: "\fcda";
}
.mdi-cursor-default-gesture::before {
  content: "\f0152";
}
.mdi-cursor-default-gesture-outline::before {
  content: "\f0153";
}
.mdi-cursor-default-outline::before {
  content: "\f1b5";
}
.mdi-cursor-move::before {
  content: "\f1b6";
}
.mdi-cursor-pointer::before {
  content: "\f1b7";
}
.mdi-cursor-text::before {
  content: "\f5e7";
}
.mdi-database::before {
  content: "\f1b8";
}
.mdi-database-check::before {
  content: "\faa8";
}
.mdi-database-edit::before {
  content: "\fb62";
}
.mdi-database-export::before {
  content: "\f95d";
}
.mdi-database-import::before {
  content: "\f95c";
}
.mdi-database-lock::before {
  content: "\faa9";
}
.mdi-database-marker::before {
  content: "\f0321";
}
.mdi-database-minus::before {
  content: "\f1b9";
}
.mdi-database-plus::before {
  content: "\f1ba";
}
.mdi-database-refresh::before {
  content: "\fcdb";
}
.mdi-database-remove::before {
  content: "\fcdc";
}
.mdi-database-search::before {
  content: "\f865";
}
.mdi-database-settings::before {
  content: "\fcdd";
}
.mdi-death-star::before {
  content: "\f8d7";
}
.mdi-death-star-variant::before {
  content: "\f8d8";
}
.mdi-deathly-hallows::before {
  content: "\fb63";
}
.mdi-debian::before {
  content: "\f8d9";
}
.mdi-debug-step-into::before {
  content: "\f1bb";
}
.mdi-debug-step-out::before {
  content: "\f1bc";
}
.mdi-debug-step-over::before {
  content: "\f1bd";
}
.mdi-decagram::before {
  content: "\f76b";
}
.mdi-decagram-outline::before {
  content: "\f76c";
}
.mdi-decimal::before {
  content: "\f00cc";
}
.mdi-decimal-comma::before {
  content: "\f00cd";
}
.mdi-decimal-comma-decrease::before {
  content: "\f00ce";
}
.mdi-decimal-comma-increase::before {
  content: "\f00cf";
}
.mdi-decimal-decrease::before {
  content: "\f1be";
}
.mdi-decimal-increase::before {
  content: "\f1bf";
}
.mdi-delete::before {
  content: "\f1c0";
}
.mdi-delete-alert::before {
  content: "\f00d0";
}
.mdi-delete-alert-outline::before {
  content: "\f00d1";
}
.mdi-delete-circle::before {
  content: "\f682";
}
.mdi-delete-circle-outline::before {
  content: "\fb64";
}
.mdi-delete-empty::before {
  content: "\f6cb";
}
.mdi-delete-empty-outline::before {
  content: "\feba";
}
.mdi-delete-forever::before {
  content: "\f5e8";
}
.mdi-delete-forever-outline::before {
  content: "\fb65";
}
.mdi-delete-off::before {
  content: "\f00d2";
}
.mdi-delete-off-outline::before {
  content: "\f00d3";
}
.mdi-delete-outline::before {
  content: "\f9e6";
}
.mdi-delete-restore::before {
  content: "\f818";
}
.mdi-delete-sweep::before {
  content: "\f5e9";
}
.mdi-delete-sweep-outline::before {
  content: "\fc3e";
}
.mdi-delete-variant::before {
  content: "\f1c1";
}
.mdi-delta::before {
  content: "\f1c2";
}
.mdi-desk::before {
  content: "\f0264";
}
.mdi-desk-lamp::before {
  content: "\f95e";
}
.mdi-deskphone::before {
  content: "\f1c3";
}
.mdi-desktop-classic::before {
  content: "\f7bf";
}
.mdi-desktop-mac::before {
  content: "\f1c4";
}
.mdi-desktop-mac-dashboard::before {
  content: "\f9e7";
}
.mdi-desktop-tower::before {
  content: "\f1c5";
}
.mdi-desktop-tower-monitor::before {
  content: "\faaa";
}
.mdi-details::before {
  content: "\f1c6";
}
.mdi-dev-to::before {
  content: "\fd4a";
}
.mdi-developer-board::before {
  content: "\f696";
}
.mdi-deviantart::before {
  content: "\f1c7";
}
.mdi-devices::before {
  content: "\ffd0";
}
.mdi-diabetes::before {
  content: "\f0151";
}
.mdi-dialpad::before {
  content: "\f61c";
}
.mdi-diameter::before {
  content: "\fc3f";
}
.mdi-diameter-outline::before {
  content: "\fc40";
}
.mdi-diameter-variant::before {
  content: "\fc41";
}
.mdi-diamond::before {
  content: "\fb66";
}
.mdi-diamond-outline::before {
  content: "\fb67";
}
.mdi-diamond-stone::before {
  content: "\f1c8";
}
.mdi-dice-1::before {
  content: "\f1ca";
}
.mdi-dice-1-outline::before {
  content: "\f0175";
}
.mdi-dice-2::before {
  content: "\f1cb";
}
.mdi-dice-2-outline::before {
  content: "\f0176";
}
.mdi-dice-3::before {
  content: "\f1cc";
}
.mdi-dice-3-outline::before {
  content: "\f0177";
}
.mdi-dice-4::before {
  content: "\f1cd";
}
.mdi-dice-4-outline::before {
  content: "\f0178";
}
.mdi-dice-5::before {
  content: "\f1ce";
}
.mdi-dice-5-outline::before {
  content: "\f0179";
}
.mdi-dice-6::before {
  content: "\f1cf";
}
.mdi-dice-6-outline::before {
  content: "\f017a";
}
.mdi-dice-d10::before {
  content: "\f017e";
}
.mdi-dice-d10-outline::before {
  content: "\f76e";
}
.mdi-dice-d12::before {
  content: "\f017f";
}
.mdi-dice-d12-outline::before {
  content: "\f866";
}
.mdi-dice-d20::before {
  content: "\f0180";
}
.mdi-dice-d20-outline::before {
  content: "\f5ea";
}
.mdi-dice-d4::before {
  content: "\f017b";
}
.mdi-dice-d4-outline::before {
  content: "\f5eb";
}
.mdi-dice-d6::before {
  content: "\f017c";
}
.mdi-dice-d6-outline::before {
  content: "\f5ec";
}
.mdi-dice-d8::before {
  content: "\f017d";
}
.mdi-dice-d8-outline::before {
  content: "\f5ed";
}
.mdi-dice-multiple::before {
  content: "\f76d";
}
.mdi-dice-multiple-outline::before {
  content: "\f0181";
}
.mdi-dictionary::before {
  content: "\f61d";
}
.mdi-digital-ocean::before {
  content: "\f0262";
}
.mdi-dip-switch::before {
  content: "\f7c0";
}
.mdi-directions::before {
  content: "\f1d0";
}
.mdi-directions-fork::before {
  content: "\f641";
}
.mdi-disc::before {
  content: "\f5ee";
}
.mdi-disc-alert::before {
  content: "\f1d1";
}
.mdi-disc-player::before {
  content: "\f95f";
}
.mdi-discord::before {
  content: "\f66f";
}
.mdi-dishwasher::before {
  content: "\faab";
}
.mdi-dishwasher-alert::before {
  content: "\f01e3";
}
.mdi-dishwasher-off::before {
  content: "\f01e4";
}
.mdi-disqus::before {
  content: "\f1d2";
}
.mdi-disqus-outline::before {
  content: "\f1d3";
}
.mdi-distribute-horizontal-center::before {
  content: "\f01f4";
}
.mdi-distribute-horizontal-left::before {
  content: "\f01f3";
}
.mdi-distribute-horizontal-right::before {
  content: "\f01f5";
}
.mdi-distribute-vertical-bottom::before {
  content: "\f01f6";
}
.mdi-distribute-vertical-center::before {
  content: "\f01f7";
}
.mdi-distribute-vertical-top::before {
  content: "\f01f8";
}
.mdi-diving-flippers::before {
  content: "\fd9b";
}
.mdi-diving-helmet::before {
  content: "\fd9c";
}
.mdi-diving-scuba::before {
  content: "\fd9d";
}
.mdi-diving-scuba-flag::before {
  content: "\fd9e";
}
.mdi-diving-scuba-tank::before {
  content: "\fd9f";
}
.mdi-diving-scuba-tank-multiple::before {
  content: "\fda0";
}
.mdi-diving-snorkel::before {
  content: "\fda1";
}
.mdi-division::before {
  content: "\f1d4";
}
.mdi-division-box::before {
  content: "\f1d5";
}
.mdi-dlna::before {
  content: "\fa40";
}
.mdi-dna::before {
  content: "\f683";
}
.mdi-dns::before {
  content: "\f1d6";
}
.mdi-dns-outline::before {
  content: "\fb68";
}
.mdi-do-not-disturb::before {
  content: "\f697";
}
.mdi-do-not-disturb-off::before {
  content: "\f698";
}
.mdi-dock-bottom::before {
  content: "\f00d4";
}
.mdi-dock-left::before {
  content: "\f00d5";
}
.mdi-dock-right::before {
  content: "\f00d6";
}
.mdi-dock-window::before {
  content: "\f00d7";
}
.mdi-docker::before {
  content: "\f867";
}
.mdi-doctor::before {
  content: "\fa41";
}
.mdi-dog::before {
  content: "\fa42";
}
.mdi-dog-service::before {
  content: "\faac";
}
.mdi-dog-side::before {
  content: "\fa43";
}
.mdi-dolby::before {
  content: "\f6b2";
}
.mdi-dolly::before {
  content: "\febb";
}
.mdi-domain::before {
  content: "\f1d7";
}
.mdi-domain-off::before {
  content: "\fd4b";
}
.mdi-domain-plus::before {
  content: "\f00d8";
}
.mdi-domain-remove::before {
  content: "\f00d9";
}
.mdi-domino-mask::before {
  content: "\f0045";
}
.mdi-donkey::before {
  content: "\f7c1";
}
.mdi-door::before {
  content: "\f819";
}
.mdi-door-closed::before {
  content: "\f81a";
}
.mdi-door-closed-lock::before {
  content: "\f00da";
}
.mdi-door-open::before {
  content: "\f81b";
}
.mdi-doorbell::before {
  content: "\f0311";
}
.mdi-doorbell-video::before {
  content: "\f868";
}
.mdi-dot-net::before {
  content: "\faad";
}
.mdi-dots-horizontal::before {
  content: "\f1d8";
}
.mdi-dots-horizontal-circle::before {
  content: "\f7c2";
}
.mdi-dots-horizontal-circle-outline::before {
  content: "\fb69";
}
.mdi-dots-vertical::before {
  content: "\f1d9";
}
.mdi-dots-vertical-circle::before {
  content: "\f7c3";
}
.mdi-dots-vertical-circle-outline::before {
  content: "\fb6a";
}
.mdi-douban::before {
  content: "\f699";
}
.mdi-download::before {
  content: "\f1da";
}
.mdi-download-lock::before {
  content: "\f034b";
}
.mdi-download-lock-outline::before {
  content: "\f034c";
}
.mdi-download-multiple::before {
  content: "\f9e8";
}
.mdi-download-network::before {
  content: "\f6f3";
}
.mdi-download-network-outline::before {
  content: "\fc42";
}
.mdi-download-off::before {
  content: "\f00db";
}
.mdi-download-off-outline::before {
  content: "\f00dc";
}
.mdi-download-outline::before {
  content: "\fb6b";
}
.mdi-drag::before {
  content: "\f1db";
}
.mdi-drag-horizontal::before {
  content: "\f1dc";
}
.mdi-drag-horizontal-variant::before {
  content: "\f031b";
}
.mdi-drag-variant::before {
  content: "\fb6c";
}
.mdi-drag-vertical::before {
  content: "\f1dd";
}
.mdi-drag-vertical-variant::before {
  content: "\f031c";
}
.mdi-drama-masks::before {
  content: "\fcde";
}
.mdi-draw::before {
  content: "\ff66";
}
.mdi-drawing::before {
  content: "\f1de";
}
.mdi-drawing-box::before {
  content: "\f1df";
}
.mdi-dresser::before {
  content: "\ff67";
}
.mdi-dresser-outline::before {
  content: "\ff68";
}
.mdi-dribbble::before {
  content: "\f1e0";
}
.mdi-dribbble-box::before {
  content: "\f1e1";
}
.mdi-drone::before {
  content: "\f1e2";
}
.mdi-dropbox::before {
  content: "\f1e3";
}
.mdi-drupal::before {
  content: "\f1e4";
}
.mdi-duck::before {
  content: "\f1e5";
}
.mdi-dumbbell::before {
  content: "\f1e6";
}
.mdi-dump-truck::before {
  content: "\fc43";
}
.mdi-ear-hearing::before {
  content: "\f7c4";
}
.mdi-ear-hearing-off::before {
  content: "\fa44";
}
.mdi-earth::before {
  content: "\f1e7";
}
.mdi-earth-arrow-right::before {
  content: "\f033c";
}
.mdi-earth-box::before {
  content: "\f6cc";
}
.mdi-earth-box-off::before {
  content: "\f6cd";
}
.mdi-earth-off::before {
  content: "\f1e8";
}
.mdi-edge::before {
  content: "\f1e9";
}
.mdi-edge-legacy::before {
  content: "\f027b";
}
.mdi-egg::before {
  content: "\faae";
}
.mdi-egg-easter::before {
  content: "\faaf";
}
.mdi-eight-track::before {
  content: "\f9e9";
}
.mdi-eject::before {
  content: "\f1ea";
}
.mdi-eject-outline::before {
  content: "\fb6d";
}
.mdi-electric-switch::before {
  content: "\febc";
}
.mdi-electric-switch-closed::before {
  content: "\f0104";
}
.mdi-electron-framework::before {
  content: "\f0046";
}
.mdi-elephant::before {
  content: "\f7c5";
}
.mdi-elevation-decline::before {
  content: "\f1eb";
}
.mdi-elevation-rise::before {
  content: "\f1ec";
}
.mdi-elevator::before {
  content: "\f1ed";
}
.mdi-elevator-down::before {
  content: "\f02ed";
}
.mdi-elevator-passenger::before {
  content: "\f03ac";
}
.mdi-elevator-up::before {
  content: "\f02ec";
}
.mdi-ellipse::before {
  content: "\febd";
}
.mdi-ellipse-outline::before {
  content: "\febe";
}
.mdi-email::before {
  content: "\f1ee";
}
.mdi-email-alert::before {
  content: "\f6ce";
}
.mdi-email-alert-outline::before {
  content: "\fd1e";
}
.mdi-email-box::before {
  content: "\fcdf";
}
.mdi-email-check::before {
  content: "\fab0";
}
.mdi-email-check-outline::before {
  content: "\fab1";
}
.mdi-email-edit::before {
  content: "\ff00";
}
.mdi-email-edit-outline::before {
  content: "\ff01";
}
.mdi-email-lock::before {
  content: "\f1f1";
}
.mdi-email-mark-as-unread::before {
  content: "\fb6e";
}
.mdi-email-minus::before {
  content: "\ff02";
}
.mdi-email-minus-outline::before {
  content: "\ff03";
}
.mdi-email-multiple::before {
  content: "\ff04";
}
.mdi-email-multiple-outline::before {
  content: "\ff05";
}
.mdi-email-newsletter::before {
  content: "\ffd1";
}
.mdi-email-open::before {
  content: "\f1ef";
}
.mdi-email-open-multiple::before {
  content: "\ff06";
}
.mdi-email-open-multiple-outline::before {
  content: "\ff07";
}
.mdi-email-open-outline::before {
  content: "\f5ef";
}
.mdi-email-outline::before {
  content: "\f1f0";
}
.mdi-email-plus::before {
  content: "\f9ea";
}
.mdi-email-plus-outline::before {
  content: "\f9eb";
}
.mdi-email-receive::before {
  content: "\f0105";
}
.mdi-email-receive-outline::before {
  content: "\f0106";
}
.mdi-email-search::before {
  content: "\f960";
}
.mdi-email-search-outline::before {
  content: "\f961";
}
.mdi-email-send::before {
  content: "\f0107";
}
.mdi-email-send-outline::before {
  content: "\f0108";
}
.mdi-email-sync::before {
  content: "\f02f2";
}
.mdi-email-sync-outline::before {
  content: "\f02f3";
}
.mdi-email-variant::before {
  content: "\f5f0";
}
.mdi-ember::before {
  content: "\fb15";
}
.mdi-emby::before {
  content: "\f6b3";
}
.mdi-emoticon::before {
  content: "\fc44";
}
.mdi-emoticon-angry::before {
  content: "\fc45";
}
.mdi-emoticon-angry-outline::before {
  content: "\fc46";
}
.mdi-emoticon-confused::before {
  content: "\f0109";
}
.mdi-emoticon-confused-outline::before {
  content: "\f010a";
}
.mdi-emoticon-cool::before {
  content: "\fc47";
}
.mdi-emoticon-cool-outline::before {
  content: "\f1f3";
}
.mdi-emoticon-cry::before {
  content: "\fc48";
}
.mdi-emoticon-cry-outline::before {
  content: "\fc49";
}
.mdi-emoticon-dead::before {
  content: "\fc4a";
}
.mdi-emoticon-dead-outline::before {
  content: "\f69a";
}
.mdi-emoticon-devil::before {
  content: "\fc4b";
}
.mdi-emoticon-devil-outline::before {
  content: "\f1f4";
}
.mdi-emoticon-excited::before {
  content: "\fc4c";
}
.mdi-emoticon-excited-outline::before {
  content: "\f69b";
}
.mdi-emoticon-frown::before {
  content: "\ff69";
}
.mdi-emoticon-frown-outline::before {
  content: "\ff6a";
}
.mdi-emoticon-happy::before {
  content: "\fc4d";
}
.mdi-emoticon-happy-outline::before {
  content: "\f1f5";
}
.mdi-emoticon-kiss::before {
  content: "\fc4e";
}
.mdi-emoticon-kiss-outline::before {
  content: "\fc4f";
}
.mdi-emoticon-lol::before {
  content: "\f023f";
}
.mdi-emoticon-lol-outline::before {
  content: "\f0240";
}
.mdi-emoticon-neutral::before {
  content: "\fc50";
}
.mdi-emoticon-neutral-outline::before {
  content: "\f1f6";
}
.mdi-emoticon-outline::before {
  content: "\f1f2";
}
.mdi-emoticon-poop::before {
  content: "\f1f7";
}
.mdi-emoticon-poop-outline::before {
  content: "\fc51";
}
.mdi-emoticon-sad::before {
  content: "\fc52";
}
.mdi-emoticon-sad-outline::before {
  content: "\f1f8";
}
.mdi-emoticon-tongue::before {
  content: "\f1f9";
}
.mdi-emoticon-tongue-outline::before {
  content: "\fc53";
}
.mdi-emoticon-wink::before {
  content: "\fc54";
}
.mdi-emoticon-wink-outline::before {
  content: "\fc55";
}
.mdi-engine::before {
  content: "\f1fa";
}
.mdi-engine-off::before {
  content: "\fa45";
}
.mdi-engine-off-outline::before {
  content: "\fa46";
}
.mdi-engine-outline::before {
  content: "\f1fb";
}
.mdi-epsilon::before {
  content: "\f010b";
}
.mdi-equal::before {
  content: "\f1fc";
}
.mdi-equal-box::before {
  content: "\f1fd";
}
.mdi-equalizer::before {
  content: "\febf";
}
.mdi-equalizer-outline::before {
  content: "\fec0";
}
.mdi-eraser::before {
  content: "\f1fe";
}
.mdi-eraser-variant::before {
  content: "\f642";
}
.mdi-escalator::before {
  content: "\f1ff";
}
.mdi-escalator-down::before {
  content: "\f02eb";
}
.mdi-escalator-up::before {
  content: "\f02ea";
}
.mdi-eslint::before {
  content: "\fc56";
}
.mdi-et::before {
  content: "\fab2";
}
.mdi-ethereum::before {
  content: "\f869";
}
.mdi-ethernet::before {
  content: "\f200";
}
.mdi-ethernet-cable::before {
  content: "\f201";
}
.mdi-ethernet-cable-off::before {
  content: "\f202";
}
.mdi-etsy::before {
  content: "\f203";
}
.mdi-ev-station::before {
  content: "\f5f1";
}
.mdi-eventbrite::before {
  content: "\f7c6";
}
.mdi-evernote::before {
  content: "\f204";
}
.mdi-excavator::before {
  content: "\f0047";
}
.mdi-exclamation::before {
  content: "\f205";
}
.mdi-exclamation-thick::before {
  content: "\f0263";
}
.mdi-exit-run::before {
  content: "\fa47";
}
.mdi-exit-to-app::before {
  content: "\f206";
}
.mdi-expand-all::before {
  content: "\fab3";
}
.mdi-expand-all-outline::before {
  content: "\fab4";
}
.mdi-expansion-card::before {
  content: "\f8ad";
}
.mdi-expansion-card-variant::before {
  content: "\ffd2";
}
.mdi-exponent::before {
  content: "\f962";
}
.mdi-exponent-box::before {
  content: "\f963";
}
.mdi-export::before {
  content: "\f207";
}
.mdi-export-variant::before {
  content: "\fb6f";
}
.mdi-eye::before {
  content: "\f208";
}
.mdi-eye-check::before {
  content: "\fce0";
}
.mdi-eye-check-outline::before {
  content: "\fce1";
}
.mdi-eye-circle::before {
  content: "\fb70";
}
.mdi-eye-circle-outline::before {
  content: "\fb71";
}
.mdi-eye-minus::before {
  content: "\f0048";
}
.mdi-eye-minus-outline::before {
  content: "\f0049";
}
.mdi-eye-off::before {
  content: "\f209";
}
.mdi-eye-off-outline::before {
  content: "\f6d0";
}
.mdi-eye-outline::before {
  content: "\f6cf";
}
.mdi-eye-plus::before {
  content: "\f86a";
}
.mdi-eye-plus-outline::before {
  content: "\f86b";
}
.mdi-eye-settings::before {
  content: "\f86c";
}
.mdi-eye-settings-outline::before {
  content: "\f86d";
}
.mdi-eyedropper::before {
  content: "\f20a";
}
.mdi-eyedropper-variant::before {
  content: "\f20b";
}
.mdi-face::before {
  content: "\f643";
}
.mdi-face-agent::before {
  content: "\fd4c";
}
.mdi-face-outline::before {
  content: "\fb72";
}
.mdi-face-profile::before {
  content: "\f644";
}
.mdi-face-profile-woman::before {
  content: "\f00a1";
}
.mdi-face-recognition::before {
  content: "\fc57";
}
.mdi-face-woman::before {
  content: "\f00a2";
}
.mdi-face-woman-outline::before {
  content: "\f00a3";
}
.mdi-facebook::before {
  content: "\f20c";
}
.mdi-facebook-box::before {
  content: "\f20d";
}
.mdi-facebook-messenger::before {
  content: "\f20e";
}
.mdi-facebook-workplace::before {
  content: "\fb16";
}
.mdi-factory::before {
  content: "\f20f";
}
.mdi-fan::before {
  content: "\f210";
}
.mdi-fan-off::before {
  content: "\f81c";
}
.mdi-fast-forward::before {
  content: "\f211";
}
.mdi-fast-forward-10::before {
  content: "\fd4d";
}
.mdi-fast-forward-30::before {
  content: "\fce2";
}
.mdi-fast-forward-5::before {
  content: "\f0223";
}
.mdi-fast-forward-outline::before {
  content: "\f6d1";
}
.mdi-fax::before {
  content: "\f212";
}
.mdi-feather::before {
  content: "\f6d2";
}
.mdi-feature-search::before {
  content: "\fa48";
}
.mdi-feature-search-outline::before {
  content: "\fa49";
}
.mdi-fedora::before {
  content: "\f8da";
}
.mdi-ferris-wheel::before {
  content: "\fec1";
}
.mdi-ferry::before {
  content: "\f213";
}
.mdi-file::before {
  content: "\f214";
}
.mdi-file-account::before {
  content: "\f73a";
}
.mdi-file-account-outline::before {
  content: "\f004a";
}
.mdi-file-alert::before {
  content: "\fa4a";
}
.mdi-file-alert-outline::before {
  content: "\fa4b";
}
.mdi-file-cabinet::before {
  content: "\fab5";
}
.mdi-file-cad::before {
  content: "\ff08";
}
.mdi-file-cad-box::before {
  content: "\ff09";
}
.mdi-file-cancel::before {
  content: "\fda2";
}
.mdi-file-cancel-outline::before {
  content: "\fda3";
}
.mdi-file-certificate::before {
  content: "\f01b1";
}
.mdi-file-certificate-outline::before {
  content: "\f01b2";
}
.mdi-file-chart::before {
  content: "\f215";
}
.mdi-file-chart-outline::before {
  content: "\f004b";
}
.mdi-file-check::before {
  content: "\f216";
}
.mdi-file-check-outline::before {
  content: "\fe7b";
}
.mdi-file-clock::before {
  content: "\f030c";
}
.mdi-file-clock-outline::before {
  content: "\f030d";
}
.mdi-file-cloud::before {
  content: "\f217";
}
.mdi-file-cloud-outline::before {
  content: "\f004c";
}
.mdi-file-code::before {
  content: "\f22e";
}
.mdi-file-code-outline::before {
  content: "\f004d";
}
.mdi-file-compare::before {
  content: "\f8a9";
}
.mdi-file-delimited::before {
  content: "\f218";
}
.mdi-file-delimited-outline::before {
  content: "\fec2";
}
.mdi-file-document::before {
  content: "\f219";
}
.mdi-file-document-box::before {
  content: "\f21a";
}
.mdi-file-document-box-check::before {
  content: "\fec3";
}
.mdi-file-document-box-check-outline::before {
  content: "\fec4";
}
.mdi-file-document-box-minus::before {
  content: "\fec5";
}
.mdi-file-document-box-minus-outline::before {
  content: "\fec6";
}
.mdi-file-document-box-multiple::before {
  content: "\fab6";
}
.mdi-file-document-box-multiple-outline::before {
  content: "\fab7";
}
.mdi-file-document-box-outline::before {
  content: "\f9ec";
}
.mdi-file-document-box-plus::before {
  content: "\fec7";
}
.mdi-file-document-box-plus-outline::before {
  content: "\fec8";
}
.mdi-file-document-box-remove::before {
  content: "\fec9";
}
.mdi-file-document-box-remove-outline::before {
  content: "\feca";
}
.mdi-file-document-box-search::before {
  content: "\fecb";
}
.mdi-file-document-box-search-outline::before {
  content: "\fecc";
}
.mdi-file-document-edit::before {
  content: "\fda4";
}
.mdi-file-document-edit-outline::before {
  content: "\fda5";
}
.mdi-file-document-outline::before {
  content: "\f9ed";
}
.mdi-file-download::before {
  content: "\f964";
}
.mdi-file-download-outline::before {
  content: "\f965";
}
.mdi-file-edit::before {
  content: "\f0212";
}
.mdi-file-edit-outline::before {
  content: "\f0213";
}
.mdi-file-excel::before {
  content: "\f21b";
}
.mdi-file-excel-box::before {
  content: "\f21c";
}
.mdi-file-excel-box-outline::before {
  content: "\f004e";
}
.mdi-file-excel-outline::before {
  content: "\f004f";
}
.mdi-file-export::before {
  content: "\f21d";
}
.mdi-file-export-outline::before {
  content: "\f0050";
}
.mdi-file-eye::before {
  content: "\fda6";
}
.mdi-file-eye-outline::before {
  content: "\fda7";
}
.mdi-file-find::before {
  content: "\f21e";
}
.mdi-file-find-outline::before {
  content: "\fb73";
}
.mdi-file-hidden::before {
  content: "\f613";
}
.mdi-file-image::before {
  content: "\f21f";
}
.mdi-file-image-outline::before {
  content: "\fecd";
}
.mdi-file-import::before {
  content: "\f220";
}
.mdi-file-import-outline::before {
  content: "\f0051";
}
.mdi-file-key::before {
  content: "\f01af";
}
.mdi-file-key-outline::before {
  content: "\f01b0";
}
.mdi-file-link::before {
  content: "\f01a2";
}
.mdi-file-link-outline::before {
  content: "\f01a3";
}
.mdi-file-lock::before {
  content: "\f221";
}
.mdi-file-lock-outline::before {
  content: "\f0052";
}
.mdi-file-move::before {
  content: "\fab8";
}
.mdi-file-move-outline::before {
  content: "\f0053";
}
.mdi-file-multiple::before {
  content: "\f222";
}
.mdi-file-multiple-outline::before {
  content: "\f0054";
}
.mdi-file-music::before {
  content: "\f223";
}
.mdi-file-music-outline::before {
  content: "\fe7c";
}
.mdi-file-outline::before {
  content: "\f224";
}
.mdi-file-pdf::before {
  content: "\f225";
}
.mdi-file-pdf-box::before {
  content: "\f226";
}
.mdi-file-pdf-box-outline::before {
  content: "\ffd3";
}
.mdi-file-pdf-outline::before {
  content: "\fe7d";
}
.mdi-file-percent::before {
  content: "\f81d";
}
.mdi-file-percent-outline::before {
  content: "\f0055";
}
.mdi-file-phone::before {
  content: "\f01a4";
}
.mdi-file-phone-outline::before {
  content: "\f01a5";
}
.mdi-file-plus::before {
  content: "\f751";
}
.mdi-file-plus-outline::before {
  content: "\ff0a";
}
.mdi-file-powerpoint::before {
  content: "\f227";
}
.mdi-file-powerpoint-box::before {
  content: "\f228";
}
.mdi-file-powerpoint-box-outline::before {
  content: "\f0056";
}
.mdi-file-powerpoint-outline::before {
  content: "\f0057";
}
.mdi-file-presentation-box::before {
  content: "\f229";
}
.mdi-file-question::before {
  content: "\f86e";
}
.mdi-file-question-outline::before {
  content: "\f0058";
}
.mdi-file-remove::before {
  content: "\fb74";
}
.mdi-file-remove-outline::before {
  content: "\f0059";
}
.mdi-file-replace::before {
  content: "\fb17";
}
.mdi-file-replace-outline::before {
  content: "\fb18";
}
.mdi-file-restore::before {
  content: "\f670";
}
.mdi-file-restore-outline::before {
  content: "\f005a";
}
.mdi-file-search::before {
  content: "\fc58";
}
.mdi-file-search-outline::before {
  content: "\fc59";
}
.mdi-file-send::before {
  content: "\f22a";
}
.mdi-file-send-outline::before {
  content: "\f005b";
}
.mdi-file-settings::before {
  content: "\f00a4";
}
.mdi-file-settings-outline::before {
  content: "\f00a5";
}
.mdi-file-settings-variant::before {
  content: "\f00a6";
}
.mdi-file-settings-variant-outline::before {
  content: "\f00a7";
}
.mdi-file-star::before {
  content: "\f005c";
}
.mdi-file-star-outline::before {
  content: "\f005d";
}
.mdi-file-swap::before {
  content: "\ffd4";
}
.mdi-file-swap-outline::before {
  content: "\ffd5";
}
.mdi-file-sync::before {
  content: "\f0241";
}
.mdi-file-sync-outline::before {
  content: "\f0242";
}
.mdi-file-table::before {
  content: "\fc5a";
}
.mdi-file-table-box::before {
  content: "\f010c";
}
.mdi-file-table-box-multiple::before {
  content: "\f010d";
}
.mdi-file-table-box-multiple-outline::before {
  content: "\f010e";
}
.mdi-file-table-box-outline::before {
  content: "\f010f";
}
.mdi-file-table-outline::before {
  content: "\fc5b";
}
.mdi-file-tree::before {
  content: "\f645";
}
.mdi-file-undo::before {
  content: "\f8db";
}
.mdi-file-undo-outline::before {
  content: "\f005e";
}
.mdi-file-upload::before {
  content: "\fa4c";
}
.mdi-file-upload-outline::before {
  content: "\fa4d";
}
.mdi-file-video::before {
  content: "\f22b";
}
.mdi-file-video-outline::before {
  content: "\fe10";
}
.mdi-file-word::before {
  content: "\f22c";
}
.mdi-file-word-box::before {
  content: "\f22d";
}
.mdi-file-word-box-outline::before {
  content: "\f005f";
}
.mdi-file-word-outline::before {
  content: "\f0060";
}
.mdi-film::before {
  content: "\f22f";
}
.mdi-filmstrip::before {
  content: "\f230";
}
.mdi-filmstrip-off::before {
  content: "\f231";
}
.mdi-filter::before {
  content: "\f232";
}
.mdi-filter-menu::before {
  content: "\f0110";
}
.mdi-filter-menu-outline::before {
  content: "\f0111";
}
.mdi-filter-minus::before {
  content: "\ff0b";
}
.mdi-filter-minus-outline::before {
  content: "\ff0c";
}
.mdi-filter-outline::before {
  content: "\f233";
}
.mdi-filter-plus::before {
  content: "\ff0d";
}
.mdi-filter-plus-outline::before {
  content: "\ff0e";
}
.mdi-filter-remove::before {
  content: "\f234";
}
.mdi-filter-remove-outline::before {
  content: "\f235";
}
.mdi-filter-variant::before {
  content: "\f236";
}
.mdi-filter-variant-minus::before {
  content: "\f013d";
}
.mdi-filter-variant-plus::before {
  content: "\f013e";
}
.mdi-filter-variant-remove::before {
  content: "\f0061";
}
.mdi-finance::before {
  content: "\f81e";
}
.mdi-find-replace::before {
  content: "\f6d3";
}
.mdi-fingerprint::before {
  content: "\f237";
}
.mdi-fingerprint-off::before {
  content: "\fece";
}
.mdi-fire::before {
  content: "\f238";
}
.mdi-fire-extinguisher::before {
  content: "\ff0f";
}
.mdi-fire-hydrant::before {
  content: "\f0162";
}
.mdi-fire-hydrant-alert::before {
  content: "\f0163";
}
.mdi-fire-hydrant-off::before {
  content: "\f0164";
}
.mdi-fire-truck::before {
  content: "\f8aa";
}
.mdi-firebase::before {
  content: "\f966";
}
.mdi-firefox::before {
  content: "\f239";
}
.mdi-fireplace::before {
  content: "\fe11";
}
.mdi-fireplace-off::before {
  content: "\fe12";
}
.mdi-firework::before {
  content: "\fe13";
}
.mdi-fish::before {
  content: "\f23a";
}
.mdi-fishbowl::before {
  content: "\ff10";
}
.mdi-fishbowl-outline::before {
  content: "\ff11";
}
.mdi-fit-to-page::before {
  content: "\ff12";
}
.mdi-fit-to-page-outline::before {
  content: "\ff13";
}
.mdi-flag::before {
  content: "\f23b";
}
.mdi-flag-checkered::before {
  content: "\f23c";
}
.mdi-flag-minus::before {
  content: "\fb75";
}
.mdi-flag-minus-outline::before {
  content: "\f00dd";
}
.mdi-flag-outline::before {
  content: "\f23d";
}
.mdi-flag-plus::before {
  content: "\fb76";
}
.mdi-flag-plus-outline::before {
  content: "\f00de";
}
.mdi-flag-remove::before {
  content: "\fb77";
}
.mdi-flag-remove-outline::before {
  content: "\f00df";
}
.mdi-flag-triangle::before {
  content: "\f23f";
}
.mdi-flag-variant::before {
  content: "\f240";
}
.mdi-flag-variant-outline::before {
  content: "\f23e";
}
.mdi-flare::before {
  content: "\fd4e";
}
.mdi-flash::before {
  content: "\f241";
}
.mdi-flash-alert::before {
  content: "\ff14";
}
.mdi-flash-alert-outline::before {
  content: "\ff15";
}
.mdi-flash-auto::before {
  content: "\f242";
}
.mdi-flash-circle::before {
  content: "\f81f";
}
.mdi-flash-off::before {
  content: "\f243";
}
.mdi-flash-outline::before {
  content: "\f6d4";
}
.mdi-flash-red-eye::before {
  content: "\f67a";
}
.mdi-flashlight::before {
  content: "\f244";
}
.mdi-flashlight-off::before {
  content: "\f245";
}
.mdi-flask::before {
  content: "\f093";
}
.mdi-flask-empty::before {
  content: "\f094";
}
.mdi-flask-empty-minus::before {
  content: "\f0265";
}
.mdi-flask-empty-minus-outline::before {
  content: "\f0266";
}
.mdi-flask-empty-outline::before {
  content: "\f095";
}
.mdi-flask-empty-plus::before {
  content: "\f0267";
}
.mdi-flask-empty-plus-outline::before {
  content: "\f0268";
}
.mdi-flask-empty-remove::before {
  content: "\f0269";
}
.mdi-flask-empty-remove-outline::before {
  content: "\f026a";
}
.mdi-flask-minus::before {
  content: "\f026b";
}
.mdi-flask-minus-outline::before {
  content: "\f026c";
}
.mdi-flask-outline::before {
  content: "\f096";
}
.mdi-flask-plus::before {
  content: "\f026d";
}
.mdi-flask-plus-outline::before {
  content: "\f026e";
}
.mdi-flask-remove::before {
  content: "\f026f";
}
.mdi-flask-remove-outline::before {
  content: "\f0270";
}
.mdi-flask-round-bottom::before {
  content: "\f0276";
}
.mdi-flask-round-bottom-empty::before {
  content: "\f0277";
}
.mdi-flask-round-bottom-empty-outline::before {
  content: "\f0278";
}
.mdi-flask-round-bottom-outline::before {
  content: "\f0279";
}
.mdi-flattr::before {
  content: "\f246";
}
.mdi-fleur-de-lis::before {
  content: "\f032e";
}
.mdi-flickr::before {
  content: "\fce3";
}
.mdi-flip-horizontal::before {
  content: "\f0112";
}
.mdi-flip-to-back::before {
  content: "\f247";
}
.mdi-flip-to-front::before {
  content: "\f248";
}
.mdi-flip-vertical::before {
  content: "\f0113";
}
.mdi-floor-lamp::before {
  content: "\f8dc";
}
.mdi-floor-lamp-dual::before {
  content: "\f0062";
}
.mdi-floor-lamp-variant::before {
  content: "\f0063";
}
.mdi-floor-plan::before {
  content: "\f820";
}
.mdi-floppy::before {
  content: "\f249";
}
.mdi-floppy-variant::before {
  content: "\f9ee";
}
.mdi-flower::before {
  content: "\f24a";
}
.mdi-flower-outline::before {
  content: "\f9ef";
}
.mdi-flower-poppy::before {
  content: "\fce4";
}
.mdi-flower-tulip::before {
  content: "\f9f0";
}
.mdi-flower-tulip-outline::before {
  content: "\f9f1";
}
.mdi-focus-auto::before {
  content: "\ff6b";
}
.mdi-focus-field::before {
  content: "\ff6c";
}
.mdi-focus-field-horizontal::before {
  content: "\ff6d";
}
.mdi-focus-field-vertical::before {
  content: "\ff6e";
}
.mdi-folder::before {
  content: "\f24b";
}
.mdi-folder-account::before {
  content: "\f24c";
}
.mdi-folder-account-outline::before {
  content: "\fb78";
}
.mdi-folder-alert::before {
  content: "\fda8";
}
.mdi-folder-alert-outline::before {
  content: "\fda9";
}
.mdi-folder-clock::before {
  content: "\fab9";
}
.mdi-folder-clock-outline::before {
  content: "\faba";
}
.mdi-folder-download::before {
  content: "\f24d";
}
.mdi-folder-download-outline::before {
  content: "\f0114";
}
.mdi-folder-edit::before {
  content: "\f8dd";
}
.mdi-folder-edit-outline::before {
  content: "\fdaa";
}
.mdi-folder-google-drive::before {
  content: "\f24e";
}
.mdi-folder-heart::before {
  content: "\f0115";
}
.mdi-folder-heart-outline::before {
  content: "\f0116";
}
.mdi-folder-home::before {
  content: "\f00e0";
}
.mdi-folder-home-outline::before {
  content: "\f00e1";
}
.mdi-folder-image::before {
  content: "\f24f";
}
.mdi-folder-information::before {
  content: "\f00e2";
}
.mdi-folder-information-outline::before {
  content: "\f00e3";
}
.mdi-folder-key::before {
  content: "\f8ab";
}
.mdi-folder-key-network::before {
  content: "\f8ac";
}
.mdi-folder-key-network-outline::before {
  content: "\fc5c";
}
.mdi-folder-key-outline::before {
  content: "\f0117";
}
.mdi-folder-lock::before {
  content: "\f250";
}
.mdi-folder-lock-open::before {
  content: "\f251";
}
.mdi-folder-marker::before {
  content: "\f0298";
}
.mdi-folder-marker-outline::before {
  content: "\f0299";
}
.mdi-folder-move::before {
  content: "\f252";
}
.mdi-folder-move-outline::before {
  content: "\f0271";
}
.mdi-folder-multiple::before {
  content: "\f253";
}
.mdi-folder-multiple-image::before {
  content: "\f254";
}
.mdi-folder-multiple-outline::before {
  content: "\f255";
}
.mdi-folder-music::before {
  content: "\f0384";
}
.mdi-folder-music-outline::before {
  content: "\f0385";
}
.mdi-folder-network::before {
  content: "\f86f";
}
.mdi-folder-network-outline::before {
  content: "\fc5d";
}
.mdi-folder-open::before {
  content: "\f76f";
}
.mdi-folder-open-outline::before {
  content: "\fdab";
}
.mdi-folder-outline::before {
  content: "\f256";
}
.mdi-folder-plus::before {
  content: "\f257";
}
.mdi-folder-plus-outline::before {
  content: "\fb79";
}
.mdi-folder-pound::before {
  content: "\fce5";
}
.mdi-folder-pound-outline::before {
  content: "\fce6";
}
.mdi-folder-remove::before {
  content: "\f258";
}
.mdi-folder-remove-outline::before {
  content: "\fb7a";
}
.mdi-folder-search::before {
  content: "\f967";
}
.mdi-folder-search-outline::before {
  content: "\f968";
}
.mdi-folder-settings::before {
  content: "\f00a8";
}
.mdi-folder-settings-outline::before {
  content: "\f00a9";
}
.mdi-folder-settings-variant::before {
  content: "\f00aa";
}
.mdi-folder-settings-variant-outline::before {
  content: "\f00ab";
}
.mdi-folder-star::before {
  content: "\f69c";
}
.mdi-folder-star-outline::before {
  content: "\fb7b";
}
.mdi-folder-swap::before {
  content: "\ffd6";
}
.mdi-folder-swap-outline::before {
  content: "\ffd7";
}
.mdi-folder-sync::before {
  content: "\fce7";
}
.mdi-folder-sync-outline::before {
  content: "\fce8";
}
.mdi-folder-table::before {
  content: "\f030e";
}
.mdi-folder-table-outline::before {
  content: "\f030f";
}
.mdi-folder-text::before {
  content: "\fc5e";
}
.mdi-folder-text-outline::before {
  content: "\fc5f";
}
.mdi-folder-upload::before {
  content: "\f259";
}
.mdi-folder-upload-outline::before {
  content: "\f0118";
}
.mdi-folder-zip::before {
  content: "\f6ea";
}
.mdi-folder-zip-outline::before {
  content: "\f7b8";
}
.mdi-font-awesome::before {
  content: "\f03a";
}
.mdi-food::before {
  content: "\f25a";
}
.mdi-food-apple::before {
  content: "\f25b";
}
.mdi-food-apple-outline::before {
  content: "\fc60";
}
.mdi-food-croissant::before {
  content: "\f7c7";
}
.mdi-food-fork-drink::before {
  content: "\f5f2";
}
.mdi-food-off::before {
  content: "\f5f3";
}
.mdi-food-variant::before {
  content: "\f25c";
}
.mdi-foot-print::before {
  content: "\ff6f";
}
.mdi-football::before {
  content: "\f25d";
}
.mdi-football-australian::before {
  content: "\f25e";
}
.mdi-football-helmet::before {
  content: "\f25f";
}
.mdi-forklift::before {
  content: "\f7c8";
}
.mdi-format-align-bottom::before {
  content: "\f752";
}
.mdi-format-align-center::before {
  content: "\f260";
}
.mdi-format-align-justify::before {
  content: "\f261";
}
.mdi-format-align-left::before {
  content: "\f262";
}
.mdi-format-align-middle::before {
  content: "\f753";
}
.mdi-format-align-right::before {
  content: "\f263";
}
.mdi-format-align-top::before {
  content: "\f754";
}
.mdi-format-annotation-minus::before {
  content: "\fabb";
}
.mdi-format-annotation-plus::before {
  content: "\f646";
}
.mdi-format-bold::before {
  content: "\f264";
}
.mdi-format-clear::before {
  content: "\f265";
}
.mdi-format-color-fill::before {
  content: "\f266";
}
.mdi-format-color-highlight::before {
  content: "\fe14";
}
.mdi-format-color-marker-cancel::before {
  content: "\f033e";
}
.mdi-format-color-text::before {
  content: "\f69d";
}
.mdi-format-columns::before {
  content: "\f8de";
}
.mdi-format-float-center::before {
  content: "\f267";
}
.mdi-format-float-left::before {
  content: "\f268";
}
.mdi-format-float-none::before {
  content: "\f269";
}
.mdi-format-float-right::before {
  content: "\f26a";
}
.mdi-format-font::before {
  content: "\f6d5";
}
.mdi-format-font-size-decrease::before {
  content: "\f9f2";
}
.mdi-format-font-size-increase::before {
  content: "\f9f3";
}
.mdi-format-header-1::before {
  content: "\f26b";
}
.mdi-format-header-2::before {
  content: "\f26c";
}
.mdi-format-header-3::before {
  content: "\f26d";
}
.mdi-format-header-4::before {
  content: "\f26e";
}
.mdi-format-header-5::before {
  content: "\f26f";
}
.mdi-format-header-6::before {
  content: "\f270";
}
.mdi-format-header-decrease::before {
  content: "\f271";
}
.mdi-format-header-equal::before {
  content: "\f272";
}
.mdi-format-header-increase::before {
  content: "\f273";
}
.mdi-format-header-pound::before {
  content: "\f274";
}
.mdi-format-horizontal-align-center::before {
  content: "\f61e";
}
.mdi-format-horizontal-align-left::before {
  content: "\f61f";
}
.mdi-format-horizontal-align-right::before {
  content: "\f620";
}
.mdi-format-indent-decrease::before {
  content: "\f275";
}
.mdi-format-indent-increase::before {
  content: "\f276";
}
.mdi-format-italic::before {
  content: "\f277";
}
.mdi-format-letter-case::before {
  content: "\fb19";
}
.mdi-format-letter-case-lower::before {
  content: "\fb1a";
}
.mdi-format-letter-case-upper::before {
  content: "\fb1b";
}
.mdi-format-letter-ends-with::before {
  content: "\ffd8";
}
.mdi-format-letter-matches::before {
  content: "\ffd9";
}
.mdi-format-letter-starts-with::before {
  content: "\ffda";
}
.mdi-format-line-spacing::before {
  content: "\f278";
}
.mdi-format-line-style::before {
  content: "\f5c8";
}
.mdi-format-line-weight::before {
  content: "\f5c9";
}
.mdi-format-list-bulleted::before {
  content: "\f279";
}
.mdi-format-list-bulleted-square::before {
  content: "\fdac";
}
.mdi-format-list-bulleted-triangle::before {
  content: "\fecf";
}
.mdi-format-list-bulleted-type::before {
  content: "\f27a";
}
.mdi-format-list-checkbox::before {
  content: "\f969";
}
.mdi-format-list-checks::before {
  content: "\f755";
}
.mdi-format-list-numbered::before {
  content: "\f27b";
}
.mdi-format-list-numbered-rtl::before {
  content: "\fce9";
}
.mdi-format-list-text::before {
  content: "\f029a";
}
.mdi-format-overline::before {
  content: "\fed0";
}
.mdi-format-page-break::before {
  content: "\f6d6";
}
.mdi-format-paint::before {
  content: "\f27c";
}
.mdi-format-paragraph::before {
  content: "\f27d";
}
.mdi-format-pilcrow::before {
  content: "\f6d7";
}
.mdi-format-quote-close::before {
  content: "\f27e";
}
.mdi-format-quote-close-outline::before {
  content: "\f01d3";
}
.mdi-format-quote-open::before {
  content: "\f756";
}
.mdi-format-quote-open-outline::before {
  content: "\f01d2";
}
.mdi-format-rotate-90::before {
  content: "\f6a9";
}
.mdi-format-section::before {
  content: "\f69e";
}
.mdi-format-size::before {
  content: "\f27f";
}
.mdi-format-strikethrough::before {
  content: "\f280";
}
.mdi-format-strikethrough-variant::before {
  content: "\f281";
}
.mdi-format-subscript::before {
  content: "\f282";
}
.mdi-format-superscript::before {
  content: "\f283";
}
.mdi-format-text::before {
  content: "\f284";
}
.mdi-format-text-rotation-angle-down::before {
  content: "\ffdb";
}
.mdi-format-text-rotation-angle-up::before {
  content: "\ffdc";
}
.mdi-format-text-rotation-down::before {
  content: "\fd4f";
}
.mdi-format-text-rotation-down-vertical::before {
  content: "\ffdd";
}
.mdi-format-text-rotation-none::before {
  content: "\fd50";
}
.mdi-format-text-rotation-up::before {
  content: "\ffde";
}
.mdi-format-text-rotation-vertical::before {
  content: "\ffdf";
}
.mdi-format-text-variant::before {
  content: "\fe15";
}
.mdi-format-text-wrapping-clip::before {
  content: "\fcea";
}
.mdi-format-text-wrapping-overflow::before {
  content: "\fceb";
}
.mdi-format-text-wrapping-wrap::before {
  content: "\fcec";
}
.mdi-format-textbox::before {
  content: "\fced";
}
.mdi-format-textdirection-l-to-r::before {
  content: "\f285";
}
.mdi-format-textdirection-r-to-l::before {
  content: "\f286";
}
.mdi-format-title::before {
  content: "\f5f4";
}
.mdi-format-underline::before {
  content: "\f287";
}
.mdi-format-vertical-align-bottom::before {
  content: "\f621";
}
.mdi-format-vertical-align-center::before {
  content: "\f622";
}
.mdi-format-vertical-align-top::before {
  content: "\f623";
}
.mdi-format-wrap-inline::before {
  content: "\f288";
}
.mdi-format-wrap-square::before {
  content: "\f289";
}
.mdi-format-wrap-tight::before {
  content: "\f28a";
}
.mdi-format-wrap-top-bottom::before {
  content: "\f28b";
}
.mdi-forum::before {
  content: "\f28c";
}
.mdi-forum-outline::before {
  content: "\f821";
}
.mdi-forward::before {
  content: "\f28d";
}
.mdi-forwardburger::before {
  content: "\fd51";
}
.mdi-fountain::before {
  content: "\f96a";
}
.mdi-fountain-pen::before {
  content: "\fcee";
}
.mdi-fountain-pen-tip::before {
  content: "\fcef";
}
.mdi-foursquare::before {
  content: "\f28e";
}
.mdi-freebsd::before {
  content: "\f8df";
}
.mdi-frequently-asked-questions::before {
  content: "\fed1";
}
.mdi-fridge::before {
  content: "\f290";
}
.mdi-fridge-alert::before {
  content: "\f01dc";
}
.mdi-fridge-alert-outline::before {
  content: "\f01dd";
}
.mdi-fridge-bottom::before {
  content: "\f292";
}
.mdi-fridge-off::before {
  content: "\f01da";
}
.mdi-fridge-off-outline::before {
  content: "\f01db";
}
.mdi-fridge-outline::before {
  content: "\f28f";
}
.mdi-fridge-top::before {
  content: "\f291";
}
.mdi-fruit-cherries::before {
  content: "\f0064";
}
.mdi-fruit-citrus::before {
  content: "\f0065";
}
.mdi-fruit-grapes::before {
  content: "\f0066";
}
.mdi-fruit-grapes-outline::before {
  content: "\f0067";
}
.mdi-fruit-pineapple::before {
  content: "\f0068";
}
.mdi-fruit-watermelon::before {
  content: "\f0069";
}
.mdi-fuel::before {
  content: "\f7c9";
}
.mdi-fullscreen::before {
  content: "\f293";
}
.mdi-fullscreen-exit::before {
  content: "\f294";
}
.mdi-function::before {
  content: "\f295";
}
.mdi-function-variant::before {
  content: "\f870";
}
.mdi-furigana-horizontal::before {
  content: "\f00ac";
}
.mdi-furigana-vertical::before {
  content: "\f00ad";
}
.mdi-fuse::before {
  content: "\fc61";
}
.mdi-fuse-blade::before {
  content: "\fc62";
}
.mdi-gamepad::before {
  content: "\f296";
}
.mdi-gamepad-circle::before {
  content: "\fe16";
}
.mdi-gamepad-circle-down::before {
  content: "\fe17";
}
.mdi-gamepad-circle-left::before {
  content: "\fe18";
}
.mdi-gamepad-circle-outline::before {
  content: "\fe19";
}
.mdi-gamepad-circle-right::before {
  content: "\fe1a";
}
.mdi-gamepad-circle-up::before {
  content: "\fe1b";
}
.mdi-gamepad-down::before {
  content: "\fe1c";
}
.mdi-gamepad-left::before {
  content: "\fe1d";
}
.mdi-gamepad-right::before {
  content: "\fe1e";
}
.mdi-gamepad-round::before {
  content: "\fe1f";
}
.mdi-gamepad-round-down::before {
  content: "\fe7e";
}
.mdi-gamepad-round-left::before {
  content: "\fe7f";
}
.mdi-gamepad-round-outline::before {
  content: "\fe80";
}
.mdi-gamepad-round-right::before {
  content: "\fe81";
}
.mdi-gamepad-round-up::before {
  content: "\fe82";
}
.mdi-gamepad-square::before {
  content: "\fed2";
}
.mdi-gamepad-square-outline::before {
  content: "\fed3";
}
.mdi-gamepad-up::before {
  content: "\fe83";
}
.mdi-gamepad-variant::before {
  content: "\f297";
}
.mdi-gamepad-variant-outline::before {
  content: "\fed4";
}
.mdi-gamma::before {
  content: "\f0119";
}
.mdi-gantry-crane::before {
  content: "\fdad";
}
.mdi-garage::before {
  content: "\f6d8";
}
.mdi-garage-alert::before {
  content: "\f871";
}
.mdi-garage-alert-variant::before {
  content: "\f0300";
}
.mdi-garage-open::before {
  content: "\f6d9";
}
.mdi-garage-open-variant::before {
  content: "\f02ff";
}
.mdi-garage-variant::before {
  content: "\f02fe";
}
.mdi-gas-cylinder::before {
  content: "\f647";
}
.mdi-gas-station::before {
  content: "\f298";
}
.mdi-gas-station-outline::before {
  content: "\fed5";
}
.mdi-gate::before {
  content: "\f299";
}
.mdi-gate-and::before {
  content: "\f8e0";
}
.mdi-gate-arrow-right::before {
  content: "\f0194";
}
.mdi-gate-nand::before {
  content: "\f8e1";
}
.mdi-gate-nor::before {
  content: "\f8e2";
}
.mdi-gate-not::before {
  content: "\f8e3";
}
.mdi-gate-open::before {
  content: "\f0195";
}
.mdi-gate-or::before {
  content: "\f8e4";
}
.mdi-gate-xnor::before {
  content: "\f8e5";
}
.mdi-gate-xor::before {
  content: "\f8e6";
}
.mdi-gatsby::before {
  content: "\fe84";
}
.mdi-gauge::before {
  content: "\f29a";
}
.mdi-gauge-empty::before {
  content: "\f872";
}
.mdi-gauge-full::before {
  content: "\f873";
}
.mdi-gauge-low::before {
  content: "\f874";
}
.mdi-gavel::before {
  content: "\f29b";
}
.mdi-gender-female::before {
  content: "\f29c";
}
.mdi-gender-male::before {
  content: "\f29d";
}
.mdi-gender-male-female::before {
  content: "\f29e";
}
.mdi-gender-male-female-variant::before {
  content: "\f016a";
}
.mdi-gender-non-binary::before {
  content: "\f016b";
}
.mdi-gender-transgender::before {
  content: "\f29f";
}
.mdi-gentoo::before {
  content: "\f8e7";
}
.mdi-gesture::before {
  content: "\f7ca";
}
.mdi-gesture-double-tap::before {
  content: "\f73b";
}
.mdi-gesture-pinch::before {
  content: "\fabc";
}
.mdi-gesture-spread::before {
  content: "\fabd";
}
.mdi-gesture-swipe::before {
  content: "\fd52";
}
.mdi-gesture-swipe-down::before {
  content: "\f73c";
}
.mdi-gesture-swipe-horizontal::before {
  content: "\fabe";
}
.mdi-gesture-swipe-left::before {
  content: "\f73d";
}
.mdi-gesture-swipe-right::before {
  content: "\f73e";
}
.mdi-gesture-swipe-up::before {
  content: "\f73f";
}
.mdi-gesture-swipe-vertical::before {
  content: "\fabf";
}
.mdi-gesture-tap::before {
  content: "\f740";
}
.mdi-gesture-tap-box::before {
  content: "\f02d4";
}
.mdi-gesture-tap-button::before {
  content: "\f02d3";
}
.mdi-gesture-tap-hold::before {
  content: "\fd53";
}
.mdi-gesture-two-double-tap::before {
  content: "\f741";
}
.mdi-gesture-two-tap::before {
  content: "\f742";
}
.mdi-ghost::before {
  content: "\f2a0";
}
.mdi-ghost-off::before {
  content: "\f9f4";
}
.mdi-gif::before {
  content: "\fd54";
}
.mdi-gift::before {
  content: "\fe85";
}
.mdi-gift-outline::before {
  content: "\f2a1";
}
.mdi-git::before {
  content: "\f2a2";
}
.mdi-github-box::before {
  content: "\f2a3";
}
.mdi-github-circle::before {
  content: "\f2a4";
}
.mdi-github-face::before {
  content: "\f6da";
}
.mdi-gitlab::before {
  content: "\fb7c";
}
.mdi-glass-cocktail::before {
  content: "\f356";
}
.mdi-glass-flute::before {
  content: "\f2a5";
}
.mdi-glass-mug::before {
  content: "\f2a6";
}
.mdi-glass-mug-variant::before {
  content: "\f0141";
}
.mdi-glass-pint-outline::before {
  content: "\f0338";
}
.mdi-glass-stange::before {
  content: "\f2a7";
}
.mdi-glass-tulip::before {
  content: "\f2a8";
}
.mdi-glass-wine::before {
  content: "\f875";
}
.mdi-glassdoor::before {
  content: "\f2a9";
}
.mdi-glasses::before {
  content: "\f2aa";
}
.mdi-globe-light::before {
  content: "\f0302";
}
.mdi-globe-model::before {
  content: "\f8e8";
}
.mdi-gmail::before {
  content: "\f2ab";
}
.mdi-gnome::before {
  content: "\f2ac";
}
.mdi-go-kart::before {
  content: "\fd55";
}
.mdi-go-kart-track::before {
  content: "\fd56";
}
.mdi-gog::before {
  content: "\fb7d";
}
.mdi-gold::before {
  content: "\f027a";
}
.mdi-golf::before {
  content: "\f822";
}
.mdi-golf-cart::before {
  content: "\f01cf";
}
.mdi-golf-tee::before {
  content: "\f00ae";
}
.mdi-gondola::before {
  content: "\f685";
}
.mdi-goodreads::before {
  content: "\fd57";
}
.mdi-google::before {
  content: "\f2ad";
}
.mdi-google-adwords::before {
  content: "\fc63";
}
.mdi-google-analytics::before {
  content: "\f7cb";
}
.mdi-google-assistant::before {
  content: "\f7cc";
}
.mdi-google-cardboard::before {
  content: "\f2ae";
}
.mdi-google-chrome::before {
  content: "\f2af";
}
.mdi-google-circles::before {
  content: "\f2b0";
}
.mdi-google-circles-communities::before {
  content: "\f2b1";
}
.mdi-google-circles-extended::before {
  content: "\f2b2";
}
.mdi-google-circles-group::before {
  content: "\f2b3";
}
.mdi-google-classroom::before {
  content: "\f2c0";
}
.mdi-google-cloud::before {
  content: "\f0221";
}
.mdi-google-controller::before {
  content: "\f2b4";
}
.mdi-google-controller-off::before {
  content: "\f2b5";
}
.mdi-google-downasaur::before {
  content: "\f038d";
}
.mdi-google-drive::before {
  content: "\f2b6";
}
.mdi-google-earth::before {
  content: "\f2b7";
}
.mdi-google-fit::before {
  content: "\f96b";
}
.mdi-google-glass::before {
  content: "\f2b8";
}
.mdi-google-hangouts::before {
  content: "\f2c9";
}
.mdi-google-home::before {
  content: "\f823";
}
.mdi-google-keep::before {
  content: "\f6db";
}
.mdi-google-lens::before {
  content: "\f9f5";
}
.mdi-google-maps::before {
  content: "\f5f5";
}
.mdi-google-my-business::before {
  content: "\f006a";
}
.mdi-google-nearby::before {
  content: "\f2b9";
}
.mdi-google-pages::before {
  content: "\f2ba";
}
.mdi-google-photos::before {
  content: "\f6dc";
}
.mdi-google-physical-web::before {
  content: "\f2bb";
}
.mdi-google-play::before {
  content: "\f2bc";
}
.mdi-google-plus::before {
  content: "\f2bd";
}
.mdi-google-plus-box::before {
  content: "\f2be";
}
.mdi-google-podcast::before {
  content: "\fed6";
}
.mdi-google-spreadsheet::before {
  content: "\f9f6";
}
.mdi-google-street-view::before {
  content: "\fc64";
}
.mdi-google-translate::before {
  content: "\f2bf";
}
.mdi-gradient::before {
  content: "\f69f";
}
.mdi-grain::before {
  content: "\fd58";
}
.mdi-graph::before {
  content: "\f006b";
}
.mdi-graph-outline::before {
  content: "\f006c";
}
.mdi-graphql::before {
  content: "\f876";
}
.mdi-grave-stone::before {
  content: "\fb7e";
}
.mdi-grease-pencil::before {
  content: "\f648";
}
.mdi-greater-than::before {
  content: "\f96c";
}
.mdi-greater-than-or-equal::before {
  content: "\f96d";
}
.mdi-grid::before {
  content: "\f2c1";
}
.mdi-grid-large::before {
  content: "\f757";
}
.mdi-grid-off::before {
  content: "\f2c2";
}
.mdi-grill::before {
  content: "\fe86";
}
.mdi-grill-outline::before {
  content: "\f01b5";
}
.mdi-group::before {
  content: "\f2c3";
}
.mdi-guitar-acoustic::before {
  content: "\f770";
}
.mdi-guitar-electric::before {
  content: "\f2c4";
}
.mdi-guitar-pick::before {
  content: "\f2c5";
}
.mdi-guitar-pick-outline::before {
  content: "\f2c6";
}
.mdi-guy-fawkes-mask::before {
  content: "\f824";
}
.mdi-hackernews::before {
  content: "\f624";
}
.mdi-hail::before {
  content: "\fac0";
}
.mdi-hair-dryer::before {
  content: "\f011a";
}
.mdi-hair-dryer-outline::before {
  content: "\f011b";
}
.mdi-halloween::before {
  content: "\fb7f";
}
.mdi-hamburger::before {
  content: "\f684";
}
.mdi-hammer::before {
  content: "\f8e9";
}
.mdi-hammer-screwdriver::before {
  content: "\f034d";
}
.mdi-hammer-wrench::before {
  content: "\f034e";
}
.mdi-hand::before {
  content: "\fa4e";
}
.mdi-hand-heart::before {
  content: "\f011c";
}
.mdi-hand-left::before {
  content: "\fe87";
}
.mdi-hand-okay::before {
  content: "\fa4f";
}
.mdi-hand-peace::before {
  content: "\fa50";
}
.mdi-hand-peace-variant::before {
  content: "\fa51";
}
.mdi-hand-pointing-down::before {
  content: "\fa52";
}
.mdi-hand-pointing-left::before {
  content: "\fa53";
}
.mdi-hand-pointing-right::before {
  content: "\f2c7";
}
.mdi-hand-pointing-up::before {
  content: "\fa54";
}
.mdi-hand-right::before {
  content: "\fe88";
}
.mdi-hand-saw::before {
  content: "\fe89";
}
.mdi-handball::before {
  content: "\ff70";
}
.mdi-handcuffs::before {
  content: "\f0169";
}
.mdi-handshake::before {
  content: "\f0243";
}
.mdi-hanger::before {
  content: "\f2c8";
}
.mdi-hard-hat::before {
  content: "\f96e";
}
.mdi-harddisk::before {
  content: "\f2ca";
}
.mdi-harddisk-plus::before {
  content: "\f006d";
}
.mdi-harddisk-remove::before {
  content: "\f006e";
}
.mdi-hat-fedora::before {
  content: "\fb80";
}
.mdi-hazard-lights::before {
  content: "\fc65";
}
.mdi-hdr::before {
  content: "\fd59";
}
.mdi-hdr-off::before {
  content: "\fd5a";
}
.mdi-head::before {
  content: "\f0389";
}
.mdi-head-alert::before {
  content: "\f0363";
}
.mdi-head-alert-outline::before {
  content: "\f0364";
}
.mdi-head-check::before {
  content: "\f0365";
}
.mdi-head-check-outline::before {
  content: "\f0366";
}
.mdi-head-cog::before {
  content: "\f0367";
}
.mdi-head-cog-outline::before {
  content: "\f0368";
}
.mdi-head-dots-horizontal::before {
  content: "\f0369";
}
.mdi-head-dots-horizontal-outline::before {
  content: "\f036a";
}
.mdi-head-flash::before {
  content: "\f036b";
}
.mdi-head-flash-outline::before {
  content: "\f036c";
}
.mdi-head-heart::before {
  content: "\f036d";
}
.mdi-head-heart-outline::before {
  content: "\f036e";
}
.mdi-head-lightbulb::before {
  content: "\f036f";
}
.mdi-head-lightbulb-outline::before {
  content: "\f0370";
}
.mdi-head-minus::before {
  content: "\f0371";
}
.mdi-head-minus-outline::before {
  content: "\f0372";
}
.mdi-head-outline::before {
  content: "\f038a";
}
.mdi-head-plus::before {
  content: "\f0373";
}
.mdi-head-plus-outline::before {
  content: "\f0374";
}
.mdi-head-question::before {
  content: "\f0375";
}
.mdi-head-question-outline::before {
  content: "\f0376";
}
.mdi-head-remove::before {
  content: "\f0377";
}
.mdi-head-remove-outline::before {
  content: "\f0378";
}
.mdi-head-snowflake::before {
  content: "\f0379";
}
.mdi-head-snowflake-outline::before {
  content: "\f037a";
}
.mdi-head-sync::before {
  content: "\f037b";
}
.mdi-head-sync-outline::before {
  content: "\f037c";
}
.mdi-headphones::before {
  content: "\f2cb";
}
.mdi-headphones-bluetooth::before {
  content: "\f96f";
}
.mdi-headphones-box::before {
  content: "\f2cc";
}
.mdi-headphones-off::before {
  content: "\f7cd";
}
.mdi-headphones-settings::before {
  content: "\f2cd";
}
.mdi-headset::before {
  content: "\f2ce";
}
.mdi-headset-dock::before {
  content: "\f2cf";
}
.mdi-headset-off::before {
  content: "\f2d0";
}
.mdi-heart::before {
  content: "\f2d1";
}
.mdi-heart-box::before {
  content: "\f2d2";
}
.mdi-heart-box-outline::before {
  content: "\f2d3";
}
.mdi-heart-broken::before {
  content: "\f2d4";
}
.mdi-heart-broken-outline::before {
  content: "\fcf0";
}
.mdi-heart-circle::before {
  content: "\f970";
}
.mdi-heart-circle-outline::before {
  content: "\f971";
}
.mdi-heart-flash::before {
  content: "\ff16";
}
.mdi-heart-half::before {
  content: "\f6de";
}
.mdi-heart-half-full::before {
  content: "\f6dd";
}
.mdi-heart-half-outline::before {
  content: "\f6df";
}
.mdi-heart-multiple::before {
  content: "\fa55";
}
.mdi-heart-multiple-outline::before {
  content: "\fa56";
}
.mdi-heart-off::before {
  content: "\f758";
}
.mdi-heart-outline::before {
  content: "\f2d5";
}
.mdi-heart-pulse::before {
  content: "\f5f6";
}
.mdi-helicopter::before {
  content: "\fac1";
}
.mdi-help::before {
  content: "\f2d6";
}
.mdi-help-box::before {
  content: "\f78a";
}
.mdi-help-circle::before {
  content: "\f2d7";
}
.mdi-help-circle-outline::before {
  content: "\f625";
}
.mdi-help-network::before {
  content: "\f6f4";
}
.mdi-help-network-outline::before {
  content: "\fc66";
}
.mdi-help-rhombus::before {
  content: "\fb81";
}
.mdi-help-rhombus-outline::before {
  content: "\fb82";
}
.mdi-hexadecimal::before {
  content: "\f02d2";
}
.mdi-hexagon::before {
  content: "\f2d8";
}
.mdi-hexagon-multiple::before {
  content: "\f6e0";
}
.mdi-hexagon-multiple-outline::before {
  content: "\f011d";
}
.mdi-hexagon-outline::before {
  content: "\f2d9";
}
.mdi-hexagon-slice-1::before {
  content: "\fac2";
}
.mdi-hexagon-slice-2::before {
  content: "\fac3";
}
.mdi-hexagon-slice-3::before {
  content: "\fac4";
}
.mdi-hexagon-slice-4::before {
  content: "\fac5";
}
.mdi-hexagon-slice-5::before {
  content: "\fac6";
}
.mdi-hexagon-slice-6::before {
  content: "\fac7";
}
.mdi-hexagram::before {
  content: "\fac8";
}
.mdi-hexagram-outline::before {
  content: "\fac9";
}
.mdi-high-definition::before {
  content: "\f7ce";
}
.mdi-high-definition-box::before {
  content: "\f877";
}
.mdi-highway::before {
  content: "\f5f7";
}
.mdi-hiking::before {
  content: "\fd5b";
}
.mdi-hinduism::before {
  content: "\f972";
}
.mdi-history::before {
  content: "\f2da";
}
.mdi-hockey-puck::before {
  content: "\f878";
}
.mdi-hockey-sticks::before {
  content: "\f879";
}
.mdi-hololens::before {
  content: "\f2db";
}
.mdi-home::before {
  content: "\f2dc";
}
.mdi-home-account::before {
  content: "\f825";
}
.mdi-home-alert::before {
  content: "\f87a";
}
.mdi-home-analytics::before {
  content: "\fed7";
}
.mdi-home-assistant::before {
  content: "\f7cf";
}
.mdi-home-automation::before {
  content: "\f7d0";
}
.mdi-home-circle::before {
  content: "\f7d1";
}
.mdi-home-circle-outline::before {
  content: "\f006f";
}
.mdi-home-city::before {
  content: "\fcf1";
}
.mdi-home-city-outline::before {
  content: "\fcf2";
}
.mdi-home-currency-usd::before {
  content: "\f8ae";
}
.mdi-home-edit::before {
  content: "\f0184";
}
.mdi-home-edit-outline::before {
  content: "\f0185";
}
.mdi-home-export-outline::before {
  content: "\ffb8";
}
.mdi-home-flood::before {
  content: "\ff17";
}
.mdi-home-floor-0::before {
  content: "\fdae";
}
.mdi-home-floor-1::before {
  content: "\fd5c";
}
.mdi-home-floor-2::before {
  content: "\fd5d";
}
.mdi-home-floor-3::before {
  content: "\fd5e";
}
.mdi-home-floor-a::before {
  content: "\fd5f";
}
.mdi-home-floor-b::before {
  content: "\fd60";
}
.mdi-home-floor-g::before {
  content: "\fd61";
}
.mdi-home-floor-l::before {
  content: "\fd62";
}
.mdi-home-floor-negative-1::before {
  content: "\fdaf";
}
.mdi-home-group::before {
  content: "\fdb0";
}
.mdi-home-heart::before {
  content: "\f826";
}
.mdi-home-import-outline::before {
  content: "\ffb9";
}
.mdi-home-lightbulb::before {
  content: "\f027c";
}
.mdi-home-lightbulb-outline::before {
  content: "\f027d";
}
.mdi-home-lock::before {
  content: "\f8ea";
}
.mdi-home-lock-open::before {
  content: "\f8eb";
}
.mdi-home-map-marker::before {
  content: "\f5f8";
}
.mdi-home-minus::before {
  content: "\f973";
}
.mdi-home-modern::before {
  content: "\f2dd";
}
.mdi-home-outline::before {
  content: "\f6a0";
}
.mdi-home-plus::before {
  content: "\f974";
}
.mdi-home-remove::before {
  content: "\f0272";
}
.mdi-home-roof::before {
  content: "\f0156";
}
.mdi-home-thermometer::before {
  content: "\ff71";
}
.mdi-home-thermometer-outline::before {
  content: "\ff72";
}
.mdi-home-variant::before {
  content: "\f2de";
}
.mdi-home-variant-outline::before {
  content: "\fb83";
}
.mdi-hook::before {
  content: "\f6e1";
}
.mdi-hook-off::before {
  content: "\f6e2";
}
.mdi-hops::before {
  content: "\f2df";
}
.mdi-horizontal-rotate-clockwise::before {
  content: "\f011e";
}
.mdi-horizontal-rotate-counterclockwise::before {
  content: "\f011f";
}
.mdi-horseshoe::before {
  content: "\fa57";
}
.mdi-hospital::before {
  content: "\f0017";
}
.mdi-hospital-box::before {
  content: "\f2e0";
}
.mdi-hospital-box-outline::before {
  content: "\f0018";
}
.mdi-hospital-building::before {
  content: "\f2e1";
}
.mdi-hospital-marker::before {
  content: "\f2e2";
}
.mdi-hot-tub::before {
  content: "\f827";
}
.mdi-hotel::before {
  content: "\f2e3";
}
.mdi-houzz::before {
  content: "\f2e4";
}
.mdi-houzz-box::before {
  content: "\f2e5";
}
.mdi-hubspot::before {
  content: "\fcf3";
}
.mdi-hulu::before {
  content: "\f828";
}
.mdi-human::before {
  content: "\f2e6";
}
.mdi-human-child::before {
  content: "\f2e7";
}
.mdi-human-female::before {
  content: "\f649";
}
.mdi-human-female-boy::before {
  content: "\fa58";
}
.mdi-human-female-female::before {
  content: "\fa59";
}
.mdi-human-female-girl::before {
  content: "\fa5a";
}
.mdi-human-greeting::before {
  content: "\f64a";
}
.mdi-human-handsdown::before {
  content: "\f64b";
}
.mdi-human-handsup::before {
  content: "\f64c";
}
.mdi-human-male::before {
  content: "\f64d";
}
.mdi-human-male-boy::before {
  content: "\fa5b";
}
.mdi-human-male-female::before {
  content: "\f2e8";
}
.mdi-human-male-girl::before {
  content: "\fa5c";
}
.mdi-human-male-height::before {
  content: "\ff18";
}
.mdi-human-male-height-variant::before {
  content: "\ff19";
}
.mdi-human-male-male::before {
  content: "\fa5d";
}
.mdi-human-pregnant::before {
  content: "\f5cf";
}
.mdi-humble-bundle::before {
  content: "\f743";
}
.mdi-hvac::before {
  content: "\f037d";
}
.mdi-hydraulic-oil-level::before {
  content: "\f034f";
}
.mdi-hydraulic-oil-temperature::before {
  content: "\f0350";
}
.mdi-hydro-power::before {
  content: "\f0310";
}
.mdi-ice-cream::before {
  content: "\f829";
}
.mdi-ice-pop::before {
  content: "\ff1a";
}
.mdi-id-card::before {
  content: "\ffe0";
}
.mdi-identifier::before {
  content: "\ff1b";
}
.mdi-ideogram-cjk::before {
  content: "\f035c";
}
.mdi-ideogram-cjk-variant::before {
  content: "\f035d";
}
.mdi-iframe::before {
  content: "\fc67";
}
.mdi-iframe-array::before {
  content: "\f0120";
}
.mdi-iframe-array-outline::before {
  content: "\f0121";
}
.mdi-iframe-braces::before {
  content: "\f0122";
}
.mdi-iframe-braces-outline::before {
  content: "\f0123";
}
.mdi-iframe-outline::before {
  content: "\fc68";
}
.mdi-iframe-parentheses::before {
  content: "\f0124";
}
.mdi-iframe-parentheses-outline::before {
  content: "\f0125";
}
.mdi-iframe-variable::before {
  content: "\f0126";
}
.mdi-iframe-variable-outline::before {
  content: "\f0127";
}
.mdi-image::before {
  content: "\f2e9";
}
.mdi-image-album::before {
  content: "\f2ea";
}
.mdi-image-area::before {
  content: "\f2eb";
}
.mdi-image-area-close::before {
  content: "\f2ec";
}
.mdi-image-auto-adjust::before {
  content: "\ffe1";
}
.mdi-image-broken::before {
  content: "\f2ed";
}
.mdi-image-broken-variant::before {
  content: "\f2ee";
}
.mdi-image-edit::before {
  content: "\f020e";
}
.mdi-image-edit-outline::before {
  content: "\f020f";
}
.mdi-image-filter::before {
  content: "\f2ef";
}
.mdi-image-filter-black-white::before {
  content: "\f2f0";
}
.mdi-image-filter-center-focus::before {
  content: "\f2f1";
}
.mdi-image-filter-center-focus-strong::before {
  content: "\ff1c";
}
.mdi-image-filter-center-focus-strong-outline::before {
  content: "\ff1d";
}
.mdi-image-filter-center-focus-weak::before {
  content: "\f2f2";
}
.mdi-image-filter-drama::before {
  content: "\f2f3";
}
.mdi-image-filter-frames::before {
  content: "\f2f4";
}
.mdi-image-filter-hdr::before {
  content: "\f2f5";
}
.mdi-image-filter-none::before {
  content: "\f2f6";
}
.mdi-image-filter-tilt-shift::before {
  content: "\f2f7";
}
.mdi-image-filter-vintage::before {
  content: "\f2f8";
}
.mdi-image-frame::before {
  content: "\fe8a";
}
.mdi-image-move::before {
  content: "\f9f7";
}
.mdi-image-multiple::before {
  content: "\f2f9";
}
.mdi-image-off::before {
  content: "\f82a";
}
.mdi-image-off-outline::before {
  content: "\f01fc";
}
.mdi-image-outline::before {
  content: "\f975";
}
.mdi-image-plus::before {
  content: "\f87b";
}
.mdi-image-search::before {
  content: "\f976";
}
.mdi-image-search-outline::before {
  content: "\f977";
}
.mdi-image-size-select-actual::before {
  content: "\fc69";
}
.mdi-image-size-select-large::before {
  content: "\fc6a";
}
.mdi-image-size-select-small::before {
  content: "\fc6b";
}
.mdi-import::before {
  content: "\f2fa";
}
.mdi-inbox::before {
  content: "\f686";
}
.mdi-inbox-arrow-down::before {
  content: "\f2fb";
}
.mdi-inbox-arrow-down-outline::before {
  content: "\f029b";
}
.mdi-inbox-arrow-up::before {
  content: "\f3d1";
}
.mdi-inbox-arrow-up-outline::before {
  content: "\f029c";
}
.mdi-inbox-full::before {
  content: "\f029d";
}
.mdi-inbox-full-outline::before {
  content: "\f029e";
}
.mdi-inbox-multiple::before {
  content: "\f8af";
}
.mdi-inbox-multiple-outline::before {
  content: "\fb84";
}
.mdi-inbox-outline::before {
  content: "\f029f";
}
.mdi-incognito::before {
  content: "\f5f9";
}
.mdi-infinity::before {
  content: "\f6e3";
}
.mdi-information::before {
  content: "\f2fc";
}
.mdi-information-outline::before {
  content: "\f2fd";
}
.mdi-information-variant::before {
  content: "\f64e";
}
.mdi-instagram::before {
  content: "\f2fe";
}
.mdi-instapaper::before {
  content: "\f2ff";
}
.mdi-instrument-triangle::before {
  content: "\f0070";
}
.mdi-internet-explorer::before {
  content: "\f300";
}
.mdi-invert-colors::before {
  content: "\f301";
}
.mdi-invert-colors-off::before {
  content: "\fe8b";
}
.mdi-iobroker::before {
  content: "\f0313";
}
.mdi-ip::before {
  content: "\fa5e";
}
.mdi-ip-network::before {
  content: "\fa5f";
}
.mdi-ip-network-outline::before {
  content: "\fc6c";
}
.mdi-ipod::before {
  content: "\fc6d";
}
.mdi-islam::before {
  content: "\f978";
}
.mdi-island::before {
  content: "\f0071";
}
.mdi-itunes::before {
  content: "\f676";
}
.mdi-iv-bag::before {
  content: "\f00e4";
}
.mdi-jabber::before {
  content: "\fdb1";
}
.mdi-jeepney::before {
  content: "\f302";
}
.mdi-jellyfish::before {
  content: "\ff1e";
}
.mdi-jellyfish-outline::before {
  content: "\ff1f";
}
.mdi-jira::before {
  content: "\f303";
}
.mdi-jquery::before {
  content: "\f87c";
}
.mdi-jsfiddle::before {
  content: "\f304";
}
.mdi-json::before {
  content: "\f626";
}
.mdi-judaism::before {
  content: "\f979";
}
.mdi-jump-rope::before {
  content: "\f032a";
}
.mdi-kabaddi::before {
  content: "\fd63";
}
.mdi-karate::before {
  content: "\f82b";
}
.mdi-keg::before {
  content: "\f305";
}
.mdi-kettle::before {
  content: "\f5fa";
}
.mdi-kettle-alert::before {
  content: "\f0342";
}
.mdi-kettle-alert-outline::before {
  content: "\f0343";
}
.mdi-kettle-off::before {
  content: "\f0346";
}
.mdi-kettle-off-outline::before {
  content: "\f0347";
}
.mdi-kettle-outline::before {
  content: "\ff73";
}
.mdi-kettle-steam::before {
  content: "\f0344";
}
.mdi-kettle-steam-outline::before {
  content: "\f0345";
}
.mdi-kettlebell::before {
  content: "\f032b";
}
.mdi-key::before {
  content: "\f306";
}
.mdi-key-arrow-right::before {
  content: "\f033d";
}
.mdi-key-change::before {
  content: "\f307";
}
.mdi-key-link::before {
  content: "\f01ca";
}
.mdi-key-minus::before {
  content: "\f308";
}
.mdi-key-outline::before {
  content: "\fdb2";
}
.mdi-key-plus::before {
  content: "\f309";
}
.mdi-key-remove::before {
  content: "\f30a";
}
.mdi-key-star::before {
  content: "\f01c9";
}
.mdi-key-variant::before {
  content: "\f30b";
}
.mdi-key-wireless::before {
  content: "\ffe2";
}
.mdi-keyboard::before {
  content: "\f30c";
}
.mdi-keyboard-backspace::before {
  content: "\f30d";
}
.mdi-keyboard-caps::before {
  content: "\f30e";
}
.mdi-keyboard-close::before {
  content: "\f30f";
}
.mdi-keyboard-esc::before {
  content: "\f02e2";
}
.mdi-keyboard-f1::before {
  content: "\f02d6";
}
.mdi-keyboard-f10::before {
  content: "\f02df";
}
.mdi-keyboard-f11::before {
  content: "\f02e0";
}
.mdi-keyboard-f12::before {
  content: "\f02e1";
}
.mdi-keyboard-f2::before {
  content: "\f02d7";
}
.mdi-keyboard-f3::before {
  content: "\f02d8";
}
.mdi-keyboard-f4::before {
  content: "\f02d9";
}
.mdi-keyboard-f5::before {
  content: "\f02da";
}
.mdi-keyboard-f6::before {
  content: "\f02db";
}
.mdi-keyboard-f7::before {
  content: "\f02dc";
}
.mdi-keyboard-f8::before {
  content: "\f02dd";
}
.mdi-keyboard-f9::before {
  content: "\f02de";
}
.mdi-keyboard-off::before {
  content: "\f310";
}
.mdi-keyboard-off-outline::before {
  content: "\fe8c";
}
.mdi-keyboard-outline::before {
  content: "\f97a";
}
.mdi-keyboard-return::before {
  content: "\f311";
}
.mdi-keyboard-settings::before {
  content: "\f9f8";
}
.mdi-keyboard-settings-outline::before {
  content: "\f9f9";
}
.mdi-keyboard-space::before {
  content: "\f0072";
}
.mdi-keyboard-tab::before {
  content: "\f312";
}
.mdi-keyboard-variant::before {
  content: "\f313";
}
.mdi-khanda::before {
  content: "\f0128";
}
.mdi-kickstarter::before {
  content: "\f744";
}
.mdi-klingon::before {
  content: "\f0386";
}
.mdi-knife::before {
  content: "\f9fa";
}
.mdi-knife-military::before {
  content: "\f9fb";
}
.mdi-kodi::before {
  content: "\f314";
}
.mdi-kotlin::before {
  content: "\f0244";
}
.mdi-kubernetes::before {
  content: "\f0129";
}
.mdi-label::before {
  content: "\f315";
}
.mdi-label-multiple::before {
  content: "\f03a0";
}
.mdi-label-multiple-outline::before {
  content: "\f03a1";
}
.mdi-label-off::before {
  content: "\faca";
}
.mdi-label-off-outline::before {
  content: "\facb";
}
.mdi-label-outline::before {
  content: "\f316";
}
.mdi-label-percent::before {
  content: "\f0315";
}
.mdi-label-percent-outline::before {
  content: "\f0316";
}
.mdi-label-variant::before {
  content: "\facc";
}
.mdi-label-variant-outline::before {
  content: "\facd";
}
.mdi-ladybug::before {
  content: "\f82c";
}
.mdi-lambda::before {
  content: "\f627";
}
.mdi-lamp::before {
  content: "\f6b4";
}
.mdi-lan::before {
  content: "\f317";
}
.mdi-lan-check::before {
  content: "\f02d5";
}
.mdi-lan-connect::before {
  content: "\f318";
}
.mdi-lan-disconnect::before {
  content: "\f319";
}
.mdi-lan-pending::before {
  content: "\f31a";
}
.mdi-language-c::before {
  content: "\f671";
}
.mdi-language-cpp::before {
  content: "\f672";
}
.mdi-language-csharp::before {
  content: "\f31b";
}
.mdi-language-css3::before {
  content: "\f31c";
}
.mdi-language-fortran::before {
  content: "\f0245";
}
.mdi-language-go::before {
  content: "\f7d2";
}
.mdi-language-haskell::before {
  content: "\fc6e";
}
.mdi-language-html5::before {
  content: "\f31d";
}
.mdi-language-java::before {
  content: "\fb1c";
}
.mdi-language-javascript::before {
  content: "\f31e";
}
.mdi-language-lua::before {
  content: "\f8b0";
}
.mdi-language-php::before {
  content: "\f31f";
}
.mdi-language-python::before {
  content: "\f320";
}
.mdi-language-python-text::before {
  content: "\f321";
}
.mdi-language-r::before {
  content: "\f7d3";
}
.mdi-language-ruby-on-rails::before {
  content: "\face";
}
.mdi-language-swift::before {
  content: "\f6e4";
}
.mdi-language-typescript::before {
  content: "\f6e5";
}
.mdi-laptop::before {
  content: "\f322";
}
.mdi-laptop-chromebook::before {
  content: "\f323";
}
.mdi-laptop-mac::before {
  content: "\f324";
}
.mdi-laptop-off::before {
  content: "\f6e6";
}
.mdi-laptop-windows::before {
  content: "\f325";
}
.mdi-laravel::before {
  content: "\facf";
}
.mdi-lasso::before {
  content: "\ff20";
}
.mdi-lastfm::before {
  content: "\f326";
}
.mdi-lastpass::before {
  content: "\f446";
}
.mdi-latitude::before {
  content: "\ff74";
}
.mdi-launch::before {
  content: "\f327";
}
.mdi-lava-lamp::before {
  content: "\f7d4";
}
.mdi-layers::before {
  content: "\f328";
}
.mdi-layers-minus::before {
  content: "\fe8d";
}
.mdi-layers-off::before {
  content: "\f329";
}
.mdi-layers-off-outline::before {
  content: "\f9fc";
}
.mdi-layers-outline::before {
  content: "\f9fd";
}
.mdi-layers-plus::before {
  content: "\fe30";
}
.mdi-layers-remove::before {
  content: "\fe31";
}
.mdi-layers-search::before {
  content: "\f0231";
}
.mdi-layers-search-outline::before {
  content: "\f0232";
}
.mdi-layers-triple::before {
  content: "\ff75";
}
.mdi-layers-triple-outline::before {
  content: "\ff76";
}
.mdi-lead-pencil::before {
  content: "\f64f";
}
.mdi-leaf::before {
  content: "\f32a";
}
.mdi-leaf-maple::before {
  content: "\fc6f";
}
.mdi-leaf-maple-off::before {
  content: "\f0305";
}
.mdi-leaf-off::before {
  content: "\f0304";
}
.mdi-leak::before {
  content: "\fdb3";
}
.mdi-leak-off::before {
  content: "\fdb4";
}
.mdi-led-off::before {
  content: "\f32b";
}
.mdi-led-on::before {
  content: "\f32c";
}
.mdi-led-outline::before {
  content: "\f32d";
}
.mdi-led-strip::before {
  content: "\f7d5";
}
.mdi-led-strip-variant::before {
  content: "\f0073";
}
.mdi-led-variant-off::before {
  content: "\f32e";
}
.mdi-led-variant-on::before {
  content: "\f32f";
}
.mdi-led-variant-outline::before {
  content: "\f330";
}
.mdi-leek::before {
  content: "\f01a8";
}
.mdi-less-than::before {
  content: "\f97b";
}
.mdi-less-than-or-equal::before {
  content: "\f97c";
}
.mdi-library::before {
  content: "\f331";
}
.mdi-library-books::before {
  content: "\f332";
}
.mdi-library-movie::before {
  content: "\fcf4";
}
.mdi-library-music::before {
  content: "\f333";
}
.mdi-library-music-outline::before {
  content: "\ff21";
}
.mdi-library-shelves::before {
  content: "\fb85";
}
.mdi-library-video::before {
  content: "\fcf5";
}
.mdi-license::before {
  content: "\ffe3";
}
.mdi-lifebuoy::before {
  content: "\f87d";
}
.mdi-light-switch::before {
  content: "\f97d";
}
.mdi-lightbulb::before {
  content: "\f335";
}
.mdi-lightbulb-cfl::before {
  content: "\f0233";
}
.mdi-lightbulb-cfl-off::before {
  content: "\f0234";
}
.mdi-lightbulb-cfl-spiral::before {
  content: "\f02a0";
}
.mdi-lightbulb-cfl-spiral-off::before {
  content: "\f02ee";
}
.mdi-lightbulb-group::before {
  content: "\f027e";
}
.mdi-lightbulb-group-off::before {
  content: "\f02f8";
}
.mdi-lightbulb-group-off-outline::before {
  content: "\f02f9";
}
.mdi-lightbulb-group-outline::before {
  content: "\f027f";
}
.mdi-lightbulb-multiple::before {
  content: "\f0280";
}
.mdi-lightbulb-multiple-off::before {
  content: "\f02fa";
}
.mdi-lightbulb-multiple-off-outline::before {
  content: "\f02fb";
}
.mdi-lightbulb-multiple-outline::before {
  content: "\f0281";
}
.mdi-lightbulb-off::before {
  content: "\fe32";
}
.mdi-lightbulb-off-outline::before {
  content: "\fe33";
}
.mdi-lightbulb-on::before {
  content: "\f6e7";
}
.mdi-lightbulb-on-outline::before {
  content: "\f6e8";
}
.mdi-lightbulb-outline::before {
  content: "\f336";
}
.mdi-lighthouse::before {
  content: "\f9fe";
}
.mdi-lighthouse-on::before {
  content: "\f9ff";
}
.mdi-link::before {
  content: "\f337";
}
.mdi-link-box::before {
  content: "\fcf6";
}
.mdi-link-box-outline::before {
  content: "\fcf7";
}
.mdi-link-box-variant::before {
  content: "\fcf8";
}
.mdi-link-box-variant-outline::before {
  content: "\fcf9";
}
.mdi-link-lock::before {
  content: "\f00e5";
}
.mdi-link-off::before {
  content: "\f338";
}
.mdi-link-plus::before {
  content: "\fc70";
}
.mdi-link-variant::before {
  content: "\f339";
}
.mdi-link-variant-minus::before {
  content: "\f012a";
}
.mdi-link-variant-off::before {
  content: "\f33a";
}
.mdi-link-variant-plus::before {
  content: "\f012b";
}
.mdi-link-variant-remove::before {
  content: "\f012c";
}
.mdi-linkedin::before {
  content: "\f33b";
}
.mdi-linkedin-box::before {
  content: "\f33c";
}
.mdi-linux::before {
  content: "\f33d";
}
.mdi-linux-mint::before {
  content: "\f8ec";
}
.mdi-litecoin::before {
  content: "\fa60";
}
.mdi-loading::before {
  content: "\f771";
}
.mdi-location-enter::before {
  content: "\ffe4";
}
.mdi-location-exit::before {
  content: "\ffe5";
}
.mdi-lock::before {
  content: "\f33e";
}
.mdi-lock-alert::before {
  content: "\f8ed";
}
.mdi-lock-clock::before {
  content: "\f97e";
}
.mdi-lock-open::before {
  content: "\f33f";
}
.mdi-lock-open-outline::before {
  content: "\f340";
}
.mdi-lock-open-variant::before {
  content: "\ffe6";
}
.mdi-lock-open-variant-outline::before {
  content: "\ffe7";
}
.mdi-lock-outline::before {
  content: "\f341";
}
.mdi-lock-pattern::before {
  content: "\f6e9";
}
.mdi-lock-plus::before {
  content: "\f5fb";
}
.mdi-lock-question::before {
  content: "\f8ee";
}
.mdi-lock-reset::before {
  content: "\f772";
}
.mdi-lock-smart::before {
  content: "\f8b1";
}
.mdi-locker::before {
  content: "\f7d6";
}
.mdi-locker-multiple::before {
  content: "\f7d7";
}
.mdi-login::before {
  content: "\f342";
}
.mdi-login-variant::before {
  content: "\f5fc";
}
.mdi-logout::before {
  content: "\f343";
}
.mdi-logout-variant::before {
  content: "\f5fd";
}
.mdi-longitude::before {
  content: "\ff77";
}
.mdi-looks::before {
  content: "\f344";
}
.mdi-loupe::before {
  content: "\f345";
}
.mdi-lumx::before {
  content: "\f346";
}
.mdi-lungs::before {
  content: "\f00af";
}
.mdi-lyft::before {
  content: "\fb1d";
}
.mdi-magnet::before {
  content: "\f347";
}
.mdi-magnet-on::before {
  content: "\f348";
}
.mdi-magnify::before {
  content: "\f349";
}
.mdi-magnify-close::before {
  content: "\f97f";
}
.mdi-magnify-minus::before {
  content: "\f34a";
}
.mdi-magnify-minus-cursor::before {
  content: "\fa61";
}
.mdi-magnify-minus-outline::before {
  content: "\f6eb";
}
.mdi-magnify-plus::before {
  content: "\f34b";
}
.mdi-magnify-plus-cursor::before {
  content: "\fa62";
}
.mdi-magnify-plus-outline::before {
  content: "\f6ec";
}
.mdi-magnify-remove-cursor::before {
  content: "\f0237";
}
.mdi-magnify-remove-outline::before {
  content: "\f0238";
}
.mdi-magnify-scan::before {
  content: "\f02a1";
}
.mdi-mail::before {
  content: "\fed8";
}
.mdi-mail-ru::before {
  content: "\f34c";
}
.mdi-mailbox::before {
  content: "\f6ed";
}
.mdi-mailbox-open::before {
  content: "\fd64";
}
.mdi-mailbox-open-outline::before {
  content: "\fd65";
}
.mdi-mailbox-open-up::before {
  content: "\fd66";
}
.mdi-mailbox-open-up-outline::before {
  content: "\fd67";
}
.mdi-mailbox-outline::before {
  content: "\fd68";
}
.mdi-mailbox-up::before {
  content: "\fd69";
}
.mdi-mailbox-up-outline::before {
  content: "\fd6a";
}
.mdi-map::before {
  content: "\f34d";
}
.mdi-map-check::before {
  content: "\fed9";
}
.mdi-map-check-outline::before {
  content: "\feda";
}
.mdi-map-clock::before {
  content: "\fcfa";
}
.mdi-map-clock-outline::before {
  content: "\fcfb";
}
.mdi-map-legend::before {
  content: "\fa00";
}
.mdi-map-marker::before {
  content: "\f34e";
}
.mdi-map-marker-alert::before {
  content: "\ff22";
}
.mdi-map-marker-alert-outline::before {
  content: "\ff23";
}
.mdi-map-marker-check::before {
  content: "\fc71";
}
.mdi-map-marker-check-outline::before {
  content: "\f0326";
}
.mdi-map-marker-circle::before {
  content: "\f34f";
}
.mdi-map-marker-distance::before {
  content: "\f8ef";
}
.mdi-map-marker-down::before {
  content: "\f012d";
}
.mdi-map-marker-left::before {
  content: "\f0306";
}
.mdi-map-marker-left-outline::before {
  content: "\f0308";
}
.mdi-map-marker-minus::before {
  content: "\f650";
}
.mdi-map-marker-minus-outline::before {
  content: "\f0324";
}
.mdi-map-marker-multiple::before {
  content: "\f350";
}
.mdi-map-marker-multiple-outline::before {
  content: "\f02a2";
}
.mdi-map-marker-off::before {
  content: "\f351";
}
.mdi-map-marker-off-outline::before {
  content: "\f0328";
}
.mdi-map-marker-outline::before {
  content: "\f7d8";
}
.mdi-map-marker-path::before {
  content: "\fcfc";
}
.mdi-map-marker-plus::before {
  content: "\f651";
}
.mdi-map-marker-plus-outline::before {
  content: "\f0323";
}
.mdi-map-marker-question::before {
  content: "\ff24";
}
.mdi-map-marker-question-outline::before {
  content: "\ff25";
}
.mdi-map-marker-radius::before {
  content: "\f352";
}
.mdi-map-marker-radius-outline::before {
  content: "\f0327";
}
.mdi-map-marker-remove::before {
  content: "\ff26";
}
.mdi-map-marker-remove-outline::before {
  content: "\f0325";
}
.mdi-map-marker-remove-variant::before {
  content: "\ff27";
}
.mdi-map-marker-right::before {
  content: "\f0307";
}
.mdi-map-marker-right-outline::before {
  content: "\f0309";
}
.mdi-map-marker-up::before {
  content: "\f012e";
}
.mdi-map-minus::before {
  content: "\f980";
}
.mdi-map-outline::before {
  content: "\f981";
}
.mdi-map-plus::before {
  content: "\f982";
}
.mdi-map-search::before {
  content: "\f983";
}
.mdi-map-search-outline::before {
  content: "\f984";
}
.mdi-mapbox::before {
  content: "\fb86";
}
.mdi-margin::before {
  content: "\f353";
}
.mdi-markdown::before {
  content: "\f354";
}
.mdi-markdown-outline::before {
  content: "\ff78";
}
.mdi-marker::before {
  content: "\f652";
}
.mdi-marker-cancel::before {
  content: "\fdb5";
}
.mdi-marker-check::before {
  content: "\f355";
}
.mdi-mastodon::before {
  content: "\fad0";
}
.mdi-mastodon-variant::before {
  content: "\fad1";
}
.mdi-material-design::before {
  content: "\f985";
}
.mdi-material-ui::before {
  content: "\f357";
}
.mdi-math-compass::before {
  content: "\f358";
}
.mdi-math-cos::before {
  content: "\fc72";
}
.mdi-math-integral::before {
  content: "\ffe8";
}
.mdi-math-integral-box::before {
  content: "\ffe9";
}
.mdi-math-log::before {
  content: "\f00b0";
}
.mdi-math-norm::before {
  content: "\ffea";
}
.mdi-math-norm-box::before {
  content: "\ffeb";
}
.mdi-math-sin::before {
  content: "\fc73";
}
.mdi-math-tan::before {
  content: "\fc74";
}
.mdi-matrix::before {
  content: "\f628";
}
.mdi-medal::before {
  content: "\f986";
}
.mdi-medal-outline::before {
  content: "\f0351";
}
.mdi-medical-bag::before {
  content: "\f6ee";
}
.mdi-meditation::before {
  content: "\f01a6";
}
.mdi-medium::before {
  content: "\f35a";
}
.mdi-meetup::before {
  content: "\fad2";
}
.mdi-memory::before {
  content: "\f35b";
}
.mdi-menu::before {
  content: "\f35c";
}
.mdi-menu-down::before {
  content: "\f35d";
}
.mdi-menu-down-outline::before {
  content: "\f6b5";
}
.mdi-menu-left::before {
  content: "\f35e";
}
.mdi-menu-left-outline::before {
  content: "\fa01";
}
.mdi-menu-open::before {
  content: "\fb87";
}
.mdi-menu-right::before {
  content: "\f35f";
}
.mdi-menu-right-outline::before {
  content: "\fa02";
}
.mdi-menu-swap::before {
  content: "\fa63";
}
.mdi-menu-swap-outline::before {
  content: "\fa64";
}
.mdi-menu-up::before {
  content: "\f360";
}
.mdi-menu-up-outline::before {
  content: "\f6b6";
}
.mdi-merge::before {
  content: "\ff79";
}
.mdi-message::before {
  content: "\f361";
}
.mdi-message-alert::before {
  content: "\f362";
}
.mdi-message-alert-outline::before {
  content: "\fa03";
}
.mdi-message-arrow-left::before {
  content: "\f031d";
}
.mdi-message-arrow-left-outline::before {
  content: "\f031e";
}
.mdi-message-arrow-right::before {
  content: "\f031f";
}
.mdi-message-arrow-right-outline::before {
  content: "\f0320";
}
.mdi-message-bulleted::before {
  content: "\f6a1";
}
.mdi-message-bulleted-off::before {
  content: "\f6a2";
}
.mdi-message-draw::before {
  content: "\f363";
}
.mdi-message-image::before {
  content: "\f364";
}
.mdi-message-image-outline::before {
  content: "\f0197";
}
.mdi-message-lock::before {
  content: "\ffec";
}
.mdi-message-lock-outline::before {
  content: "\f0198";
}
.mdi-message-minus::before {
  content: "\f0199";
}
.mdi-message-minus-outline::before {
  content: "\f019a";
}
.mdi-message-outline::before {
  content: "\f365";
}
.mdi-message-plus::before {
  content: "\f653";
}
.mdi-message-plus-outline::before {
  content: "\f00e6";
}
.mdi-message-processing::before {
  content: "\f366";
}
.mdi-message-processing-outline::before {
  content: "\f019b";
}
.mdi-message-reply::before {
  content: "\f367";
}
.mdi-message-reply-text::before {
  content: "\f368";
}
.mdi-message-settings::before {
  content: "\f6ef";
}
.mdi-message-settings-outline::before {
  content: "\f019c";
}
.mdi-message-settings-variant::before {
  content: "\f6f0";
}
.mdi-message-settings-variant-outline::before {
  content: "\f019d";
}
.mdi-message-text::before {
  content: "\f369";
}
.mdi-message-text-clock::before {
  content: "\f019e";
}
.mdi-message-text-clock-outline::before {
  content: "\f019f";
}
.mdi-message-text-lock::before {
  content: "\ffed";
}
.mdi-message-text-lock-outline::before {
  content: "\f01a0";
}
.mdi-message-text-outline::before {
  content: "\f36a";
}
.mdi-message-video::before {
  content: "\f36b";
}
.mdi-meteor::before {
  content: "\f629";
}
.mdi-metronome::before {
  content: "\f7d9";
}
.mdi-metronome-tick::before {
  content: "\f7da";
}
.mdi-micro-sd::before {
  content: "\f7db";
}
.mdi-microphone::before {
  content: "\f36c";
}
.mdi-microphone-minus::before {
  content: "\f8b2";
}
.mdi-microphone-off::before {
  content: "\f36d";
}
.mdi-microphone-outline::before {
  content: "\f36e";
}
.mdi-microphone-plus::before {
  content: "\f8b3";
}
.mdi-microphone-settings::before {
  content: "\f36f";
}
.mdi-microphone-variant::before {
  content: "\f370";
}
.mdi-microphone-variant-off::before {
  content: "\f371";
}
.mdi-microscope::before {
  content: "\f654";
}
.mdi-microsoft::before {
  content: "\f372";
}
.mdi-microsoft-dynamics::before {
  content: "\f987";
}
.mdi-microwave::before {
  content: "\fc75";
}
.mdi-middleware::before {
  content: "\ff7a";
}
.mdi-middleware-outline::before {
  content: "\ff7b";
}
.mdi-midi::before {
  content: "\f8f0";
}
.mdi-midi-port::before {
  content: "\f8f1";
}
.mdi-mine::before {
  content: "\fdb6";
}
.mdi-minecraft::before {
  content: "\f373";
}
.mdi-mini-sd::before {
  content: "\fa04";
}
.mdi-minidisc::before {
  content: "\fa05";
}
.mdi-minus::before {
  content: "\f374";
}
.mdi-minus-box::before {
  content: "\f375";
}
.mdi-minus-box-multiple::before {
  content: "\f016c";
}
.mdi-minus-box-multiple-outline::before {
  content: "\f016d";
}
.mdi-minus-box-outline::before {
  content: "\f6f1";
}
.mdi-minus-circle::before {
  content: "\f376";
}
.mdi-minus-circle-outline::before {
  content: "\f377";
}
.mdi-minus-network::before {
  content: "\f378";
}
.mdi-minus-network-outline::before {
  content: "\fc76";
}
.mdi-mirror::before {
  content: "\f0228";
}
.mdi-mixcloud::before {
  content: "\f62a";
}
.mdi-mixed-martial-arts::before {
  content: "\fd6b";
}
.mdi-mixed-reality::before {
  content: "\f87e";
}
.mdi-mixer::before {
  content: "\f7dc";
}
.mdi-molecule::before {
  content: "\fb88";
}
.mdi-monitor::before {
  content: "\f379";
}
.mdi-monitor-cellphone::before {
  content: "\f988";
}
.mdi-monitor-cellphone-star::before {
  content: "\f989";
}
.mdi-monitor-clean::before {
  content: "\f012f";
}
.mdi-monitor-dashboard::before {
  content: "\fa06";
}
.mdi-monitor-edit::before {
  content: "\f02f1";
}
.mdi-monitor-lock::before {
  content: "\fdb7";
}
.mdi-monitor-multiple::before {
  content: "\f37a";
}
.mdi-monitor-off::before {
  content: "\fd6c";
}
.mdi-monitor-screenshot::before {
  content: "\fe34";
}
.mdi-monitor-speaker::before {
  content: "\ff7c";
}
.mdi-monitor-speaker-off::before {
  content: "\ff7d";
}
.mdi-monitor-star::before {
  content: "\fdb8";
}
.mdi-moon-first-quarter::before {
  content: "\ff7e";
}
.mdi-moon-full::before {
  content: "\ff7f";
}
.mdi-moon-last-quarter::before {
  content: "\ff80";
}
.mdi-moon-new::before {
  content: "\ff81";
}
.mdi-moon-waning-crescent::before {
  content: "\ff82";
}
.mdi-moon-waning-gibbous::before {
  content: "\ff83";
}
.mdi-moon-waxing-crescent::before {
  content: "\ff84";
}
.mdi-moon-waxing-gibbous::before {
  content: "\ff85";
}
.mdi-moped::before {
  content: "\f00b1";
}
.mdi-more::before {
  content: "\f37b";
}
.mdi-mother-heart::before {
  content: "\f033f";
}
.mdi-mother-nurse::before {
  content: "\fcfd";
}
.mdi-motion-sensor::before {
  content: "\fd6d";
}
.mdi-motorbike::before {
  content: "\f37c";
}
.mdi-mouse::before {
  content: "\f37d";
}
.mdi-mouse-bluetooth::before {
  content: "\f98a";
}
.mdi-mouse-off::before {
  content: "\f37e";
}
.mdi-mouse-variant::before {
  content: "\f37f";
}
.mdi-mouse-variant-off::before {
  content: "\f380";
}
.mdi-move-resize::before {
  content: "\f655";
}
.mdi-move-resize-variant::before {
  content: "\f656";
}
.mdi-movie::before {
  content: "\f381";
}
.mdi-movie-edit::before {
  content: "\f014d";
}
.mdi-movie-edit-outline::before {
  content: "\f014e";
}
.mdi-movie-filter::before {
  content: "\f014f";
}
.mdi-movie-filter-outline::before {
  content: "\f0150";
}
.mdi-movie-open::before {
  content: "\ffee";
}
.mdi-movie-open-outline::before {
  content: "\ffef";
}
.mdi-movie-outline::before {
  content: "\fdb9";
}
.mdi-movie-roll::before {
  content: "\f7dd";
}
.mdi-movie-search::before {
  content: "\f01fd";
}
.mdi-movie-search-outline::before {
  content: "\f01fe";
}
.mdi-muffin::before {
  content: "\f98b";
}
.mdi-multiplication::before {
  content: "\f382";
}
.mdi-multiplication-box::before {
  content: "\f383";
}
.mdi-mushroom::before {
  content: "\f7de";
}
.mdi-mushroom-outline::before {
  content: "\f7df";
}
.mdi-music::before {
  content: "\f759";
}
.mdi-music-accidental-double-flat::before {
  content: "\ff86";
}
.mdi-music-accidental-double-sharp::before {
  content: "\ff87";
}
.mdi-music-accidental-flat::before {
  content: "\ff88";
}
.mdi-music-accidental-natural::before {
  content: "\ff89";
}
.mdi-music-accidental-sharp::before {
  content: "\ff8a";
}
.mdi-music-box::before {
  content: "\f384";
}
.mdi-music-box-outline::before {
  content: "\f385";
}
.mdi-music-circle::before {
  content: "\f386";
}
.mdi-music-circle-outline::before {
  content: "\fad3";
}
.mdi-music-clef-alto::before {
  content: "\ff8b";
}
.mdi-music-clef-bass::before {
  content: "\ff8c";
}
.mdi-music-clef-treble::before {
  content: "\ff8d";
}
.mdi-music-note::before {
  content: "\f387";
}
.mdi-music-note-bluetooth::before {
  content: "\f5fe";
}
.mdi-music-note-bluetooth-off::before {
  content: "\f5ff";
}
.mdi-music-note-eighth::before {
  content: "\f388";
}
.mdi-music-note-eighth-dotted::before {
  content: "\ff8e";
}
.mdi-music-note-half::before {
  content: "\f389";
}
.mdi-music-note-half-dotted::before {
  content: "\ff8f";
}
.mdi-music-note-off::before {
  content: "\f38a";
}
.mdi-music-note-off-outline::before {
  content: "\ff90";
}
.mdi-music-note-outline::before {
  content: "\ff91";
}
.mdi-music-note-plus::before {
  content: "\fdba";
}
.mdi-music-note-quarter::before {
  content: "\f38b";
}
.mdi-music-note-quarter-dotted::before {
  content: "\ff92";
}
.mdi-music-note-sixteenth::before {
  content: "\f38c";
}
.mdi-music-note-sixteenth-dotted::before {
  content: "\ff93";
}
.mdi-music-note-whole::before {
  content: "\f38d";
}
.mdi-music-note-whole-dotted::before {
  content: "\ff94";
}
.mdi-music-off::before {
  content: "\f75a";
}
.mdi-music-rest-eighth::before {
  content: "\ff95";
}
.mdi-music-rest-half::before {
  content: "\ff96";
}
.mdi-music-rest-quarter::before {
  content: "\ff97";
}
.mdi-music-rest-sixteenth::before {
  content: "\ff98";
}
.mdi-music-rest-whole::before {
  content: "\ff99";
}
.mdi-nail::before {
  content: "\fdbb";
}
.mdi-nas::before {
  content: "\f8f2";
}
.mdi-nativescript::before {
  content: "\f87f";
}
.mdi-nature::before {
  content: "\f38e";
}
.mdi-nature-people::before {
  content: "\f38f";
}
.mdi-navigation::before {
  content: "\f390";
}
.mdi-near-me::before {
  content: "\f5cd";
}
.mdi-necklace::before {
  content: "\ff28";
}
.mdi-needle::before {
  content: "\f391";
}
.mdi-netflix::before {
  content: "\f745";
}
.mdi-network::before {
  content: "\f6f2";
}
.mdi-network-off::before {
  content: "\fc77";
}
.mdi-network-off-outline::before {
  content: "\fc78";
}
.mdi-network-outline::before {
  content: "\fc79";
}
.mdi-network-router::before {
  content: "\f00b2";
}
.mdi-network-strength-1::before {
  content: "\f8f3";
}
.mdi-network-strength-1-alert::before {
  content: "\f8f4";
}
.mdi-network-strength-2::before {
  content: "\f8f5";
}
.mdi-network-strength-2-alert::before {
  content: "\f8f6";
}
.mdi-network-strength-3::before {
  content: "\f8f7";
}
.mdi-network-strength-3-alert::before {
  content: "\f8f8";
}
.mdi-network-strength-4::before {
  content: "\f8f9";
}
.mdi-network-strength-4-alert::before {
  content: "\f8fa";
}
.mdi-network-strength-off::before {
  content: "\f8fb";
}
.mdi-network-strength-off-outline::before {
  content: "\f8fc";
}
.mdi-network-strength-outline::before {
  content: "\f8fd";
}
.mdi-new-box::before {
  content: "\f394";
}
.mdi-newspaper::before {
  content: "\f395";
}
.mdi-newspaper-minus::before {
  content: "\ff29";
}
.mdi-newspaper-plus::before {
  content: "\ff2a";
}
.mdi-newspaper-variant::before {
  content: "\f0023";
}
.mdi-newspaper-variant-multiple::before {
  content: "\f0024";
}
.mdi-newspaper-variant-multiple-outline::before {
  content: "\f0025";
}
.mdi-newspaper-variant-outline::before {
  content: "\f0026";
}
.mdi-nfc::before {
  content: "\f396";
}
.mdi-nfc-off::before {
  content: "\fe35";
}
.mdi-nfc-search-variant::before {
  content: "\fe36";
}
.mdi-nfc-tap::before {
  content: "\f397";
}
.mdi-nfc-variant::before {
  content: "\f398";
}
.mdi-nfc-variant-off::before {
  content: "\fe37";
}
.mdi-ninja::before {
  content: "\f773";
}
.mdi-nintendo-switch::before {
  content: "\f7e0";
}
.mdi-nix::before {
  content: "\f0130";
}
.mdi-nodejs::before {
  content: "\f399";
}
.mdi-noodles::before {
  content: "\f01a9";
}
.mdi-not-equal::before {
  content: "\f98c";
}
.mdi-not-equal-variant::before {
  content: "\f98d";
}
.mdi-note::before {
  content: "\f39a";
}
.mdi-note-multiple::before {
  content: "\f6b7";
}
.mdi-note-multiple-outline::before {
  content: "\f6b8";
}
.mdi-note-outline::before {
  content: "\f39b";
}
.mdi-note-plus::before {
  content: "\f39c";
}
.mdi-note-plus-outline::before {
  content: "\f39d";
}
.mdi-note-text::before {
  content: "\f39e";
}
.mdi-note-text-outline::before {
  content: "\f0202";
}
.mdi-notebook::before {
  content: "\f82d";
}
.mdi-notebook-multiple::before {
  content: "\fe38";
}
.mdi-notebook-outline::before {
  content: "\fedc";
}
.mdi-notification-clear-all::before {
  content: "\f39f";
}
.mdi-npm::before {
  content: "\f6f6";
}
.mdi-npm-variant::before {
  content: "\f98e";
}
.mdi-npm-variant-outline::before {
  content: "\f98f";
}
.mdi-nuke::before {
  content: "\f6a3";
}
.mdi-null::before {
  content: "\f7e1";
}
.mdi-numeric::before {
  content: "\f3a0";
}
.mdi-numeric-0::before {
  content: "0";
}
.mdi-numeric-0-box::before {
  content: "\f3a1";
}
.mdi-numeric-0-box-multiple::before {
  content: "\ff2b";
}
.mdi-numeric-0-box-multiple-outline::before {
  content: "\f3a2";
}
.mdi-numeric-0-box-outline::before {
  content: "\f3a3";
}
.mdi-numeric-0-circle::before {
  content: "\fc7a";
}
.mdi-numeric-0-circle-outline::before {
  content: "\fc7b";
}
.mdi-numeric-1::before {
  content: "1";
}
.mdi-numeric-1-box::before {
  content: "\f3a4";
}
.mdi-numeric-1-box-multiple::before {
  content: "\ff2c";
}
.mdi-numeric-1-box-multiple-outline::before {
  content: "\f3a5";
}
.mdi-numeric-1-box-outline::before {
  content: "\f3a6";
}
.mdi-numeric-1-circle::before {
  content: "\fc7c";
}
.mdi-numeric-1-circle-outline::before {
  content: "\fc7d";
}
.mdi-numeric-10::before {
  content: "\f000a";
}
.mdi-numeric-10-box::before {
  content: "\ff9a";
}
.mdi-numeric-10-box-multiple::before {
  content: "\f000b";
}
.mdi-numeric-10-box-multiple-outline::before {
  content: "\f000c";
}
.mdi-numeric-10-box-outline::before {
  content: "\ff9b";
}
.mdi-numeric-10-circle::before {
  content: "\f000d";
}
.mdi-numeric-10-circle-outline::before {
  content: "\f000e";
}
.mdi-numeric-2::before {
  content: "2";
}
.mdi-numeric-2-box::before {
  content: "\f3a7";
}
.mdi-numeric-2-box-multiple::before {
  content: "\ff2d";
}
.mdi-numeric-2-box-multiple-outline::before {
  content: "\f3a8";
}
.mdi-numeric-2-box-outline::before {
  content: "\f3a9";
}
.mdi-numeric-2-circle::before {
  content: "\fc7e";
}
.mdi-numeric-2-circle-outline::before {
  content: "\fc7f";
}
.mdi-numeric-3::before {
  content: "3";
}
.mdi-numeric-3-box::before {
  content: "\f3aa";
}
.mdi-numeric-3-box-multiple::before {
  content: "\ff2e";
}
.mdi-numeric-3-box-multiple-outline::before {
  content: "\f3ab";
}
.mdi-numeric-3-box-outline::before {
  content: "\f3ac";
}
.mdi-numeric-3-circle::before {
  content: "\fc80";
}
.mdi-numeric-3-circle-outline::before {
  content: "\fc81";
}
.mdi-numeric-4::before {
  content: "4";
}
.mdi-numeric-4-box::before {
  content: "\f3ad";
}
.mdi-numeric-4-box-multiple::before {
  content: "\ff2f";
}
.mdi-numeric-4-box-multiple-outline::before {
  content: "\f3ae";
}
.mdi-numeric-4-box-outline::before {
  content: "\f3af";
}
.mdi-numeric-4-circle::before {
  content: "\fc82";
}
.mdi-numeric-4-circle-outline::before {
  content: "\fc83";
}
.mdi-numeric-5::before {
  content: "5";
}
.mdi-numeric-5-box::before {
  content: "\f3b0";
}
.mdi-numeric-5-box-multiple::before {
  content: "\ff30";
}
.mdi-numeric-5-box-multiple-outline::before {
  content: "\f3b1";
}
.mdi-numeric-5-box-outline::before {
  content: "\f3b2";
}
.mdi-numeric-5-circle::before {
  content: "\fc84";
}
.mdi-numeric-5-circle-outline::before {
  content: "\fc85";
}
.mdi-numeric-6::before {
  content: "6";
}
.mdi-numeric-6-box::before {
  content: "\f3b3";
}
.mdi-numeric-6-box-multiple::before {
  content: "\ff31";
}
.mdi-numeric-6-box-multiple-outline::before {
  content: "\f3b4";
}
.mdi-numeric-6-box-outline::before {
  content: "\f3b5";
}
.mdi-numeric-6-circle::before {
  content: "\fc86";
}
.mdi-numeric-6-circle-outline::before {
  content: "\fc87";
}
.mdi-numeric-7::before {
  content: "7";
}
.mdi-numeric-7-box::before {
  content: "\f3b6";
}
.mdi-numeric-7-box-multiple::before {
  content: "\ff32";
}
.mdi-numeric-7-box-multiple-outline::before {
  content: "\f3b7";
}
.mdi-numeric-7-box-outline::before {
  content: "\f3b8";
}
.mdi-numeric-7-circle::before {
  content: "\fc88";
}
.mdi-numeric-7-circle-outline::before {
  content: "\fc89";
}
.mdi-numeric-8::before {
  content: "8";
}
.mdi-numeric-8-box::before {
  content: "\f3b9";
}
.mdi-numeric-8-box-multiple::before {
  content: "\ff33";
}
.mdi-numeric-8-box-multiple-outline::before {
  content: "\f3ba";
}
.mdi-numeric-8-box-outline::before {
  content: "\f3bb";
}
.mdi-numeric-8-circle::before {
  content: "\fc8a";
}
.mdi-numeric-8-circle-outline::before {
  content: "\fc8b";
}
.mdi-numeric-9::before {
  content: "9";
}
.mdi-numeric-9-box::before {
  content: "\f3bc";
}
.mdi-numeric-9-box-multiple::before {
  content: "\ff34";
}
.mdi-numeric-9-box-multiple-outline::before {
  content: "\f3bd";
}
.mdi-numeric-9-box-outline::before {
  content: "\f3be";
}
.mdi-numeric-9-circle::before {
  content: "\fc8c";
}
.mdi-numeric-9-circle-outline::before {
  content: "\fc8d";
}
.mdi-numeric-9-plus::before {
  content: "\f000f";
}
.mdi-numeric-9-plus-box::before {
  content: "\f3bf";
}
.mdi-numeric-9-plus-box-multiple::before {
  content: "\ff35";
}
.mdi-numeric-9-plus-box-multiple-outline::before {
  content: "\f3c0";
}
.mdi-numeric-9-plus-box-outline::before {
  content: "\f3c1";
}
.mdi-numeric-9-plus-circle::before {
  content: "\fc8e";
}
.mdi-numeric-9-plus-circle-outline::before {
  content: "\fc8f";
}
.mdi-numeric-negative-1::before {
  content: "\f0074";
}
.mdi-nut::before {
  content: "\f6f7";
}
.mdi-nutrition::before {
  content: "\f3c2";
}
.mdi-nuxt::before {
  content: "\f0131";
}
.mdi-oar::before {
  content: "\f67b";
}
.mdi-ocarina::before {
  content: "\fdbc";
}
.mdi-oci::before {
  content: "\f0314";
}
.mdi-ocr::before {
  content: "\f0165";
}
.mdi-octagon::before {
  content: "\f3c3";
}
.mdi-octagon-outline::before {
  content: "\f3c4";
}
.mdi-octagram::before {
  content: "\f6f8";
}
.mdi-octagram-outline::before {
  content: "\f774";
}
.mdi-odnoklassniki::before {
  content: "\f3c5";
}
.mdi-offer::before {
  content: "\f0246";
}
.mdi-office::before {
  content: "\f3c6";
}
.mdi-office-building::before {
  content: "\f990";
}
.mdi-oil::before {
  content: "\f3c7";
}
.mdi-oil-lamp::before {
  content: "\ff36";
}
.mdi-oil-level::before {
  content: "\f0075";
}
.mdi-oil-temperature::before {
  content: "\f0019";
}
.mdi-omega::before {
  content: "\f3c9";
}
.mdi-one-up::before {
  content: "\fb89";
}
.mdi-onedrive::before {
  content: "\f3ca";
}
.mdi-onenote::before {
  content: "\f746";
}
.mdi-onepassword::before {
  content: "\f880";
}
.mdi-opacity::before {
  content: "\f5cc";
}
.mdi-open-in-app::before {
  content: "\f3cb";
}
.mdi-open-in-new::before {
  content: "\f3cc";
}
.mdi-open-source-initiative::before {
  content: "\fb8a";
}
.mdi-openid::before {
  content: "\f3cd";
}
.mdi-opera::before {
  content: "\f3ce";
}
.mdi-orbit::before {
  content: "\f018";
}
.mdi-origin::before {
  content: "\fb2b";
}
.mdi-ornament::before {
  content: "\f3cf";
}
.mdi-ornament-variant::before {
  content: "\f3d0";
}
.mdi-outdoor-lamp::before {
  content: "\f0076";
}
.mdi-outlook::before {
  content: "\fcfe";
}
.mdi-overscan::before {
  content: "\f0027";
}
.mdi-owl::before {
  content: "\f3d2";
}
.mdi-pac-man::before {
  content: "\fb8b";
}
.mdi-package::before {
  content: "\f3d3";
}
.mdi-package-down::before {
  content: "\f3d4";
}
.mdi-package-up::before {
  content: "\f3d5";
}
.mdi-package-variant::before {
  content: "\f3d6";
}
.mdi-package-variant-closed::before {
  content: "\f3d7";
}
.mdi-page-first::before {
  content: "\f600";
}
.mdi-page-last::before {
  content: "\f601";
}
.mdi-page-layout-body::before {
  content: "\f6f9";
}
.mdi-page-layout-footer::before {
  content: "\f6fa";
}
.mdi-page-layout-header::before {
  content: "\f6fb";
}
.mdi-page-layout-header-footer::before {
  content: "\ff9c";
}
.mdi-page-layout-sidebar-left::before {
  content: "\f6fc";
}
.mdi-page-layout-sidebar-right::before {
  content: "\f6fd";
}
.mdi-page-next::before {
  content: "\fb8c";
}
.mdi-page-next-outline::before {
  content: "\fb8d";
}
.mdi-page-previous::before {
  content: "\fb8e";
}
.mdi-page-previous-outline::before {
  content: "\fb8f";
}
.mdi-palette::before {
  content: "\f3d8";
}
.mdi-palette-advanced::before {
  content: "\f3d9";
}
.mdi-palette-outline::before {
  content: "\fe6c";
}
.mdi-palette-swatch::before {
  content: "\f8b4";
}
.mdi-palette-swatch-outline::before {
  content: "\f0387";
}
.mdi-palm-tree::before {
  content: "\f0077";
}
.mdi-pan::before {
  content: "\fb90";
}
.mdi-pan-bottom-left::before {
  content: "\fb91";
}
.mdi-pan-bottom-right::before {
  content: "\fb92";
}
.mdi-pan-down::before {
  content: "\fb93";
}
.mdi-pan-horizontal::before {
  content: "\fb94";
}
.mdi-pan-left::before {
  content: "\fb95";
}
.mdi-pan-right::before {
  content: "\fb96";
}
.mdi-pan-top-left::before {
  content: "\fb97";
}
.mdi-pan-top-right::before {
  content: "\fb98";
}
.mdi-pan-up::before {
  content: "\fb99";
}
.mdi-pan-vertical::before {
  content: "\fb9a";
}
.mdi-panda::before {
  content: "\f3da";
}
.mdi-pandora::before {
  content: "\f3db";
}
.mdi-panorama::before {
  content: "\f3dc";
}
.mdi-panorama-fisheye::before {
  content: "\f3dd";
}
.mdi-panorama-horizontal::before {
  content: "\f3de";
}
.mdi-panorama-vertical::before {
  content: "\f3df";
}
.mdi-panorama-wide-angle::before {
  content: "\f3e0";
}
.mdi-paper-cut-vertical::before {
  content: "\f3e1";
}
.mdi-paper-roll::before {
  content: "\f0182";
}
.mdi-paper-roll-outline::before {
  content: "\f0183";
}
.mdi-paperclip::before {
  content: "\f3e2";
}
.mdi-parachute::before {
  content: "\fc90";
}
.mdi-parachute-outline::before {
  content: "\fc91";
}
.mdi-parking::before {
  content: "\f3e3";
}
.mdi-party-popper::before {
  content: "\f0078";
}
.mdi-passport::before {
  content: "\f7e2";
}
.mdi-passport-biometric::before {
  content: "\fdbd";
}
.mdi-pasta::before {
  content: "\f018b";
}
.mdi-patio-heater::before {
  content: "\ff9d";
}
.mdi-patreon::before {
  content: "\f881";
}
.mdi-pause::before {
  content: "\f3e4";
}
.mdi-pause-circle::before {
  content: "\f3e5";
}
.mdi-pause-circle-outline::before {
  content: "\f3e6";
}
.mdi-pause-octagon::before {
  content: "\f3e7";
}
.mdi-pause-octagon-outline::before {
  content: "\f3e8";
}
.mdi-paw::before {
  content: "\f3e9";
}
.mdi-paw-off::before {
  content: "\f657";
}
.mdi-paypal::before {
  content: "\f882";
}
.mdi-pdf-box::before {
  content: "\fe39";
}
.mdi-peace::before {
  content: "\f883";
}
.mdi-peanut::before {
  content: "\f001e";
}
.mdi-peanut-off::before {
  content: "\f001f";
}
.mdi-peanut-off-outline::before {
  content: "\f0021";
}
.mdi-peanut-outline::before {
  content: "\f0020";
}
.mdi-pen::before {
  content: "\f3ea";
}
.mdi-pen-lock::before {
  content: "\fdbe";
}
.mdi-pen-minus::before {
  content: "\fdbf";
}
.mdi-pen-off::before {
  content: "\fdc0";
}
.mdi-pen-plus::before {
  content: "\fdc1";
}
.mdi-pen-remove::before {
  content: "\fdc2";
}
.mdi-pencil::before {
  content: "\f3eb";
}
.mdi-pencil-box::before {
  content: "\f3ec";
}
.mdi-pencil-box-multiple::before {
  content: "\f016f";
}
.mdi-pencil-box-multiple-outline::before {
  content: "\f0170";
}
.mdi-pencil-box-outline::before {
  content: "\f3ed";
}
.mdi-pencil-circle::before {
  content: "\f6fe";
}
.mdi-pencil-circle-outline::before {
  content: "\f775";
}
.mdi-pencil-lock::before {
  content: "\f3ee";
}
.mdi-pencil-lock-outline::before {
  content: "\fdc3";
}
.mdi-pencil-minus::before {
  content: "\fdc4";
}
.mdi-pencil-minus-outline::before {
  content: "\fdc5";
}
.mdi-pencil-off::before {
  content: "\f3ef";
}
.mdi-pencil-off-outline::before {
  content: "\fdc6";
}
.mdi-pencil-outline::before {
  content: "\fc92";
}
.mdi-pencil-plus::before {
  content: "\fdc7";
}
.mdi-pencil-plus-outline::before {
  content: "\fdc8";
}
.mdi-pencil-remove::before {
  content: "\fdc9";
}
.mdi-pencil-remove-outline::before {
  content: "\fdca";
}
.mdi-pencil-ruler::before {
  content: "\f037e";
}
.mdi-penguin::before {
  content: "\fedd";
}
.mdi-pentagon::before {
  content: "\f6ff";
}
.mdi-pentagon-outline::before {
  content: "\f700";
}
.mdi-percent::before {
  content: "\f3f0";
}
.mdi-percent-outline::before {
  content: "\f02a3";
}
.mdi-periodic-table::before {
  content: "\f8b5";
}
.mdi-periodic-table-co::before {
  content: "\f0329";
}
.mdi-periodic-table-co2::before {
  content: "\f7e3";
}
.mdi-periscope::before {
  content: "\f747";
}
.mdi-perspective-less::before {
  content: "\fcff";
}
.mdi-perspective-more::before {
  content: "\fd00";
}
.mdi-pharmacy::before {
  content: "\f3f1";
}
.mdi-phone::before {
  content: "\f3f2";
}
.mdi-phone-alert::before {
  content: "\ff37";
}
.mdi-phone-alert-outline::before {
  content: "\f01b9";
}
.mdi-phone-bluetooth::before {
  content: "\f3f3";
}
.mdi-phone-bluetooth-outline::before {
  content: "\f01ba";
}
.mdi-phone-cancel::before {
  content: "\f00e7";
}
.mdi-phone-cancel-outline::before {
  content: "\f01bb";
}
.mdi-phone-check::before {
  content: "\f01d4";
}
.mdi-phone-check-outline::before {
  content: "\f01d5";
}
.mdi-phone-classic::before {
  content: "\f602";
}
.mdi-phone-classic-off::before {
  content: "\f02a4";
}
.mdi-phone-forward::before {
  content: "\f3f4";
}
.mdi-phone-forward-outline::before {
  content: "\f01bc";
}
.mdi-phone-hangup::before {
  content: "\f3f5";
}
.mdi-phone-hangup-outline::before {
  content: "\f01bd";
}
.mdi-phone-in-talk::before {
  content: "\f3f6";
}
.mdi-phone-in-talk-outline::before {
  content: "\f01ad";
}
.mdi-phone-incoming::before {
  content: "\f3f7";
}
.mdi-phone-incoming-outline::before {
  content: "\f01be";
}
.mdi-phone-lock::before {
  content: "\f3f8";
}
.mdi-phone-lock-outline::before {
  content: "\f01bf";
}
.mdi-phone-log::before {
  content: "\f3f9";
}
.mdi-phone-log-outline::before {
  content: "\f01c0";
}
.mdi-phone-message::before {
  content: "\f01c1";
}
.mdi-phone-message-outline::before {
  content: "\f01c2";
}
.mdi-phone-minus::before {
  content: "\f658";
}
.mdi-phone-minus-outline::before {
  content: "\f01c3";
}
.mdi-phone-missed::before {
  content: "\f3fa";
}
.mdi-phone-missed-outline::before {
  content: "\f01d0";
}
.mdi-phone-off::before {
  content: "\fdcb";
}
.mdi-phone-off-outline::before {
  content: "\f01d1";
}
.mdi-phone-outgoing::before {
  content: "\f3fb";
}
.mdi-phone-outgoing-outline::before {
  content: "\f01c4";
}
.mdi-phone-outline::before {
  content: "\fdcc";
}
.mdi-phone-paused::before {
  content: "\f3fc";
}
.mdi-phone-paused-outline::before {
  content: "\f01c5";
}
.mdi-phone-plus::before {
  content: "\f659";
}
.mdi-phone-plus-outline::before {
  content: "\f01c6";
}
.mdi-phone-return::before {
  content: "\f82e";
}
.mdi-phone-return-outline::before {
  content: "\f01c7";
}
.mdi-phone-ring::before {
  content: "\f01d6";
}
.mdi-phone-ring-outline::before {
  content: "\f01d7";
}
.mdi-phone-rotate-landscape::before {
  content: "\f884";
}
.mdi-phone-rotate-portrait::before {
  content: "\f885";
}
.mdi-phone-settings::before {
  content: "\f3fd";
}
.mdi-phone-settings-outline::before {
  content: "\f01c8";
}
.mdi-phone-voip::before {
  content: "\f3fe";
}
.mdi-pi::before {
  content: "\f3ff";
}
.mdi-pi-box::before {
  content: "\f400";
}
.mdi-pi-hole::before {
  content: "\fdcd";
}
.mdi-piano::before {
  content: "\f67c";
}
.mdi-pickaxe::before {
  content: "\f8b6";
}
.mdi-picture-in-picture-bottom-right::before {
  content: "\fe3a";
}
.mdi-picture-in-picture-bottom-right-outline::before {
  content: "\fe3b";
}
.mdi-picture-in-picture-top-right::before {
  content: "\fe3c";
}
.mdi-picture-in-picture-top-right-outline::before {
  content: "\fe3d";
}
.mdi-pier::before {
  content: "\f886";
}
.mdi-pier-crane::before {
  content: "\f887";
}
.mdi-pig::before {
  content: "\f401";
}
.mdi-pig-variant::before {
  content: "\f0028";
}
.mdi-piggy-bank::before {
  content: "\f0029";
}
.mdi-pill::before {
  content: "\f402";
}
.mdi-pillar::before {
  content: "\f701";
}
.mdi-pin::before {
  content: "\f403";
}
.mdi-pin-off::before {
  content: "\f404";
}
.mdi-pin-off-outline::before {
  content: "\f92f";
}
.mdi-pin-outline::before {
  content: "\f930";
}
.mdi-pine-tree::before {
  content: "\f405";
}
.mdi-pine-tree-box::before {
  content: "\f406";
}
.mdi-pinterest::before {
  content: "\f407";
}
.mdi-pinterest-box::before {
  content: "\f408";
}
.mdi-pinwheel::before {
  content: "\fad4";
}
.mdi-pinwheel-outline::before {
  content: "\fad5";
}
.mdi-pipe::before {
  content: "\f7e4";
}
.mdi-pipe-disconnected::before {
  content: "\f7e5";
}
.mdi-pipe-leak::before {
  content: "\f888";
}
.mdi-pipe-wrench::before {
  content: "\f037f";
}
.mdi-pirate::before {
  content: "\fa07";
}
.mdi-pistol::before {
  content: "\f702";
}
.mdi-piston::before {
  content: "\f889";
}
.mdi-pizza::before {
  content: "\f409";
}
.mdi-play::before {
  content: "\f40a";
}
.mdi-play-box::before {
  content: "\f02a5";
}
.mdi-play-box-outline::before {
  content: "\f40b";
}
.mdi-play-circle::before {
  content: "\f40c";
}
.mdi-play-circle-outline::before {
  content: "\f40d";
}
.mdi-play-network::before {
  content: "\f88a";
}
.mdi-play-network-outline::before {
  content: "\fc93";
}
.mdi-play-outline::before {
  content: "\ff38";
}
.mdi-play-pause::before {
  content: "\f40e";
}
.mdi-play-protected-content::before {
  content: "\f40f";
}
.mdi-play-speed::before {
  content: "\f8fe";
}
.mdi-playlist-check::before {
  content: "\f5c7";
}
.mdi-playlist-edit::before {
  content: "\f8ff";
}
.mdi-playlist-minus::before {
  content: "\f410";
}
.mdi-playlist-music::before {
  content: "\fc94";
}
.mdi-playlist-music-outline::before {
  content: "\fc95";
}
.mdi-playlist-play::before {
  content: "\f411";
}
.mdi-playlist-plus::before {
  content: "\f412";
}
.mdi-playlist-remove::before {
  content: "\f413";
}
.mdi-playlist-star::before {
  content: "\fdce";
}
.mdi-playstation::before {
  content: "\f414";
}
.mdi-plex::before {
  content: "\f6b9";
}
.mdi-plus::before {
  content: "\f415";
}
.mdi-plus-box::before {
  content: "\f416";
}
.mdi-plus-box-multiple::before {
  content: "\f334";
}
.mdi-plus-box-multiple-outline::before {
  content: "\f016e";
}
.mdi-plus-box-outline::before {
  content: "\f703";
}
.mdi-plus-circle::before {
  content: "\f417";
}
.mdi-plus-circle-multiple-outline::before {
  content: "\f418";
}
.mdi-plus-circle-outline::before {
  content: "\f419";
}
.mdi-plus-minus::before {
  content: "\f991";
}
.mdi-plus-minus-box::before {
  content: "\f992";
}
.mdi-plus-network::before {
  content: "\f41a";
}
.mdi-plus-network-outline::before {
  content: "\fc96";
}
.mdi-plus-one::before {
  content: "\f41b";
}
.mdi-plus-outline::before {
  content: "\f704";
}
.mdi-plus-thick::before {
  content: "\f0217";
}
.mdi-pocket::before {
  content: "\f41c";
}
.mdi-podcast::before {
  content: "\f993";
}
.mdi-podium::before {
  content: "\fd01";
}
.mdi-podium-bronze::before {
  content: "\fd02";
}
.mdi-podium-gold::before {
  content: "\fd03";
}
.mdi-podium-silver::before {
  content: "\fd04";
}
.mdi-point-of-sale::before {
  content: "\fd6e";
}
.mdi-pokeball::before {
  content: "\f41d";
}
.mdi-pokemon-go::before {
  content: "\fa08";
}
.mdi-poker-chip::before {
  content: "\f82f";
}
.mdi-polaroid::before {
  content: "\f41e";
}
.mdi-police-badge::before {
  content: "\f0192";
}
.mdi-police-badge-outline::before {
  content: "\f0193";
}
.mdi-poll::before {
  content: "\f41f";
}
.mdi-poll-box::before {
  content: "\f420";
}
.mdi-poll-box-outline::before {
  content: "\f02a6";
}
.mdi-polymer::before {
  content: "\f421";
}
.mdi-pool::before {
  content: "\f606";
}
.mdi-popcorn::before {
  content: "\f422";
}
.mdi-post::before {
  content: "\f002a";
}
.mdi-post-outline::before {
  content: "\f002b";
}
.mdi-postage-stamp::before {
  content: "\fc97";
}
.mdi-pot::before {
  content: "\f65a";
}
.mdi-pot-mix::before {
  content: "\f65b";
}
.mdi-pound::before {
  content: "\f423";
}
.mdi-pound-box::before {
  content: "\f424";
}
.mdi-pound-box-outline::before {
  content: "\f01aa";
}
.mdi-power::before {
  content: "\f425";
}
.mdi-power-cycle::before {
  content: "\f900";
}
.mdi-power-off::before {
  content: "\f901";
}
.mdi-power-on::before {
  content: "\f902";
}
.mdi-power-plug::before {
  content: "\f6a4";
}
.mdi-power-plug-off::before {
  content: "\f6a5";
}
.mdi-power-settings::before {
  content: "\f426";
}
.mdi-power-sleep::before {
  content: "\f903";
}
.mdi-power-socket::before {
  content: "\f427";
}
.mdi-power-socket-au::before {
  content: "\f904";
}
.mdi-power-socket-de::before {
  content: "\f0132";
}
.mdi-power-socket-eu::before {
  content: "\f7e6";
}
.mdi-power-socket-fr::before {
  content: "\f0133";
}
.mdi-power-socket-jp::before {
  content: "\f0134";
}
.mdi-power-socket-uk::before {
  content: "\f7e7";
}
.mdi-power-socket-us::before {
  content: "\f7e8";
}
.mdi-power-standby::before {
  content: "\f905";
}
.mdi-powershell::before {
  content: "\fa09";
}
.mdi-prescription::before {
  content: "\f705";
}
.mdi-presentation::before {
  content: "\f428";
}
.mdi-presentation-play::before {
  content: "\f429";
}
.mdi-printer::before {
  content: "\f42a";
}
.mdi-printer-3d::before {
  content: "\f42b";
}
.mdi-printer-3d-nozzle::before {
  content: "\fe3e";
}
.mdi-printer-3d-nozzle-alert::before {
  content: "\f01eb";
}
.mdi-printer-3d-nozzle-alert-outline::before {
  content: "\f01ec";
}
.mdi-printer-3d-nozzle-outline::before {
  content: "\fe3f";
}
.mdi-printer-alert::before {
  content: "\f42c";
}
.mdi-printer-check::before {
  content: "\f0171";
}
.mdi-printer-off::before {
  content: "\fe40";
}
.mdi-printer-pos::before {
  content: "\f0079";
}
.mdi-printer-settings::before {
  content: "\f706";
}
.mdi-printer-wireless::before {
  content: "\fa0a";
}
.mdi-priority-high::before {
  content: "\f603";
}
.mdi-priority-low::before {
  content: "\f604";
}
.mdi-professional-hexagon::before {
  content: "\f42d";
}
.mdi-progress-alert::before {
  content: "\fc98";
}
.mdi-progress-check::before {
  content: "\f994";
}
.mdi-progress-clock::before {
  content: "\f995";
}
.mdi-progress-close::before {
  content: "\f0135";
}
.mdi-progress-download::before {
  content: "\f996";
}
.mdi-progress-upload::before {
  content: "\f997";
}
.mdi-progress-wrench::before {
  content: "\fc99";
}
.mdi-projector::before {
  content: "\f42e";
}
.mdi-projector-screen::before {
  content: "\f42f";
}
.mdi-propane-tank::before {
  content: "\f0382";
}
.mdi-propane-tank-outline::before {
  content: "\f0383";
}
.mdi-protocol::before {
  content: "\fff9";
}
.mdi-publish::before {
  content: "\f6a6";
}
.mdi-pulse::before {
  content: "\f430";
}
.mdi-pumpkin::before {
  content: "\fb9b";
}
.mdi-purse::before {
  content: "\ff39";
}
.mdi-purse-outline::before {
  content: "\ff3a";
}
.mdi-puzzle::before {
  content: "\f431";
}
.mdi-puzzle-outline::before {
  content: "\fa65";
}
.mdi-qi::before {
  content: "\f998";
}
.mdi-qqchat::before {
  content: "\f605";
}
.mdi-qrcode::before {
  content: "\f432";
}
.mdi-qrcode-edit::before {
  content: "\f8b7";
}
.mdi-qrcode-minus::before {
  content: "\f01b7";
}
.mdi-qrcode-plus::before {
  content: "\f01b6";
}
.mdi-qrcode-remove::before {
  content: "\f01b8";
}
.mdi-qrcode-scan::before {
  content: "\f433";
}
.mdi-quadcopter::before {
  content: "\f434";
}
.mdi-quality-high::before {
  content: "\f435";
}
.mdi-quality-low::before {
  content: "\fa0b";
}
.mdi-quality-medium::before {
  content: "\fa0c";
}
.mdi-quicktime::before {
  content: "\f436";
}
.mdi-quora::before {
  content: "\fd05";
}
.mdi-rabbit::before {
  content: "\f906";
}
.mdi-racing-helmet::before {
  content: "\fd6f";
}
.mdi-racquetball::before {
  content: "\fd70";
}
.mdi-radar::before {
  content: "\f437";
}
.mdi-radiator::before {
  content: "\f438";
}
.mdi-radiator-disabled::before {
  content: "\fad6";
}
.mdi-radiator-off::before {
  content: "\fad7";
}
.mdi-radio::before {
  content: "\f439";
}
.mdi-radio-am::before {
  content: "\fc9a";
}
.mdi-radio-fm::before {
  content: "\fc9b";
}
.mdi-radio-handheld::before {
  content: "\f43a";
}
.mdi-radio-off::before {
  content: "\f0247";
}
.mdi-radio-tower::before {
  content: "\f43b";
}
.mdi-radioactive::before {
  content: "\f43c";
}
.mdi-radioactive-off::before {
  content: "\fede";
}
.mdi-radiobox-blank::before {
  content: "\f43d";
}
.mdi-radiobox-marked::before {
  content: "\f43e";
}
.mdi-radius::before {
  content: "\fc9c";
}
.mdi-radius-outline::before {
  content: "\fc9d";
}
.mdi-railroad-light::before {
  content: "\ff3b";
}
.mdi-raspberry-pi::before {
  content: "\f43f";
}
.mdi-ray-end::before {
  content: "\f440";
}
.mdi-ray-end-arrow::before {
  content: "\f441";
}
.mdi-ray-start::before {
  content: "\f442";
}
.mdi-ray-start-arrow::before {
  content: "\f443";
}
.mdi-ray-start-end::before {
  content: "\f444";
}
.mdi-ray-vertex::before {
  content: "\f445";
}
.mdi-react::before {
  content: "\f707";
}
.mdi-read::before {
  content: "\f447";
}
.mdi-receipt::before {
  content: "\f449";
}
.mdi-record::before {
  content: "\f44a";
}
.mdi-record-circle::before {
  content: "\fedf";
}
.mdi-record-circle-outline::before {
  content: "\fee0";
}
.mdi-record-player::before {
  content: "\f999";
}
.mdi-record-rec::before {
  content: "\f44b";
}
.mdi-rectangle::before {
  content: "\fe41";
}
.mdi-rectangle-outline::before {
  content: "\fe42";
}
.mdi-recycle::before {
  content: "\f44c";
}
.mdi-reddit::before {
  content: "\f44d";
}
.mdi-redhat::before {
  content: "\f0146";
}
.mdi-redo::before {
  content: "\f44e";
}
.mdi-redo-variant::before {
  content: "\f44f";
}
.mdi-reflect-horizontal::before {
  content: "\fa0d";
}
.mdi-reflect-vertical::before {
  content: "\fa0e";
}
.mdi-refresh::before {
  content: "\f450";
}
.mdi-refresh-circle::before {
  content: "\f03a2";
}
.mdi-regex::before {
  content: "\f451";
}
.mdi-registered-trademark::before {
  content: "\fa66";
}
.mdi-relative-scale::before {
  content: "\f452";
}
.mdi-reload::before {
  content: "\f453";
}
.mdi-reload-alert::before {
  content: "\f0136";
}
.mdi-reminder::before {
  content: "\f88b";
}
.mdi-remote::before {
  content: "\f454";
}
.mdi-remote-desktop::before {
  content: "\f8b8";
}
.mdi-remote-off::before {
  content: "\fee1";
}
.mdi-remote-tv::before {
  content: "\fee2";
}
.mdi-remote-tv-off::before {
  content: "\fee3";
}
.mdi-rename-box::before {
  content: "\f455";
}
.mdi-reorder-horizontal::before {
  content: "\f687";
}
.mdi-reorder-vertical::before {
  content: "\f688";
}
.mdi-repeat::before {
  content: "\f456";
}
.mdi-repeat-off::before {
  content: "\f457";
}
.mdi-repeat-once::before {
  content: "\f458";
}
.mdi-replay::before {
  content: "\f459";
}
.mdi-reply::before {
  content: "\f45a";
}
.mdi-reply-all::before {
  content: "\f45b";
}
.mdi-reply-all-outline::before {
  content: "\ff3c";
}
.mdi-reply-circle::before {
  content: "\f01d9";
}
.mdi-reply-outline::before {
  content: "\ff3d";
}
.mdi-reproduction::before {
  content: "\f45c";
}
.mdi-resistor::before {
  content: "\fb1f";
}
.mdi-resistor-nodes::before {
  content: "\fb20";
}
.mdi-resize::before {
  content: "\fa67";
}
.mdi-resize-bottom-right::before {
  content: "\f45d";
}
.mdi-responsive::before {
  content: "\f45e";
}
.mdi-restart::before {
  content: "\f708";
}
.mdi-restart-alert::before {
  content: "\f0137";
}
.mdi-restart-off::before {
  content: "\fd71";
}
.mdi-restore::before {
  content: "\f99a";
}
.mdi-restore-alert::before {
  content: "\f0138";
}
.mdi-rewind::before {
  content: "\f45f";
}
.mdi-rewind-10::before {
  content: "\fd06";
}
.mdi-rewind-30::before {
  content: "\fd72";
}
.mdi-rewind-5::before {
  content: "\f0224";
}
.mdi-rewind-outline::before {
  content: "\f709";
}
.mdi-rhombus::before {
  content: "\f70a";
}
.mdi-rhombus-medium::before {
  content: "\fa0f";
}
.mdi-rhombus-outline::before {
  content: "\f70b";
}
.mdi-rhombus-split::before {
  content: "\fa10";
}
.mdi-ribbon::before {
  content: "\f460";
}
.mdi-rice::before {
  content: "\f7e9";
}
.mdi-ring::before {
  content: "\f7ea";
}
.mdi-rivet::before {
  content: "\fe43";
}
.mdi-road::before {
  content: "\f461";
}
.mdi-road-variant::before {
  content: "\f462";
}
.mdi-robber::before {
  content: "\f007a";
}
.mdi-robot::before {
  content: "\f6a8";
}
.mdi-robot-industrial::before {
  content: "\fb21";
}
.mdi-robot-mower::before {
  content: "\f0222";
}
.mdi-robot-mower-outline::before {
  content: "\f021e";
}
.mdi-robot-vacuum::before {
  content: "\f70c";
}
.mdi-robot-vacuum-variant::before {
  content: "\f907";
}
.mdi-rocket::before {
  content: "\f463";
}
.mdi-rodent::before {
  content: "\f0352";
}
.mdi-roller-skate::before {
  content: "\fd07";
}
.mdi-rollerblade::before {
  content: "\fd08";
}
.mdi-rollupjs::before {
  content: "\fb9c";
}
.mdi-roman-numeral-1::before {
  content: "\f00b3";
}
.mdi-roman-numeral-10::before {
  content: "\f00bc";
}
.mdi-roman-numeral-2::before {
  content: "\f00b4";
}
.mdi-roman-numeral-3::before {
  content: "\f00b5";
}
.mdi-roman-numeral-4::before {
  content: "\f00b6";
}
.mdi-roman-numeral-5::before {
  content: "\f00b7";
}
.mdi-roman-numeral-6::before {
  content: "\f00b8";
}
.mdi-roman-numeral-7::before {
  content: "\f00b9";
}
.mdi-roman-numeral-8::before {
  content: "\f00ba";
}
.mdi-roman-numeral-9::before {
  content: "\f00bb";
}
.mdi-room-service::before {
  content: "\f88c";
}
.mdi-room-service-outline::before {
  content: "\fd73";
}
.mdi-rotate-3d::before {
  content: "\fee4";
}
.mdi-rotate-3d-variant::before {
  content: "\f464";
}
.mdi-rotate-left::before {
  content: "\f465";
}
.mdi-rotate-left-variant::before {
  content: "\f466";
}
.mdi-rotate-orbit::before {
  content: "\fd74";
}
.mdi-rotate-right::before {
  content: "\f467";
}
.mdi-rotate-right-variant::before {
  content: "\f468";
}
.mdi-rounded-corner::before {
  content: "\f607";
}
.mdi-router::before {
  content: "\f020d";
}
.mdi-router-wireless::before {
  content: "\f469";
}
.mdi-router-wireless-settings::before {
  content: "\fa68";
}
.mdi-routes::before {
  content: "\f46a";
}
.mdi-routes-clock::before {
  content: "\f007b";
}
.mdi-rowing::before {
  content: "\f608";
}
.mdi-rss::before {
  content: "\f46b";
}
.mdi-rss-box::before {
  content: "\f46c";
}
.mdi-rss-off::before {
  content: "\ff3e";
}
.mdi-ruby::before {
  content: "\fd09";
}
.mdi-rugby::before {
  content: "\fd75";
}
.mdi-ruler::before {
  content: "\f46d";
}
.mdi-ruler-square::before {
  content: "\fc9e";
}
.mdi-ruler-square-compass::before {
  content: "\fedb";
}
.mdi-run::before {
  content: "\f70d";
}
.mdi-run-fast::before {
  content: "\f46e";
}
.mdi-rv-truck::before {
  content: "\f01ff";
}
.mdi-sack::before {
  content: "\fd0a";
}
.mdi-sack-percent::before {
  content: "\fd0b";
}
.mdi-safe::before {
  content: "\fa69";
}
.mdi-safe-square::before {
  content: "\f02a7";
}
.mdi-safe-square-outline::before {
  content: "\f02a8";
}
.mdi-safety-goggles::before {
  content: "\fd0c";
}
.mdi-sailing::before {
  content: "\fee5";
}
.mdi-sale::before {
  content: "\f46f";
}
.mdi-salesforce::before {
  content: "\f88d";
}
.mdi-sass::before {
  content: "\f7eb";
}
.mdi-satellite::before {
  content: "\f470";
}
.mdi-satellite-uplink::before {
  content: "\f908";
}
.mdi-satellite-variant::before {
  content: "\f471";
}
.mdi-sausage::before {
  content: "\f8b9";
}
.mdi-saw-blade::before {
  content: "\fe44";
}
.mdi-saxophone::before {
  content: "\f609";
}
.mdi-scale::before {
  content: "\f472";
}
.mdi-scale-balance::before {
  content: "\f5d1";
}
.mdi-scale-bathroom::before {
  content: "\f473";
}
.mdi-scale-off::before {
  content: "\f007c";
}
.mdi-scanner::before {
  content: "\f6aa";
}
.mdi-scanner-off::before {
  content: "\f909";
}
.mdi-scatter-plot::before {
  content: "\fee6";
}
.mdi-scatter-plot-outline::before {
  content: "\fee7";
}
.mdi-school::before {
  content: "\f474";
}
.mdi-school-outline::before {
  content: "\f01ab";
}
.mdi-scissors-cutting::before {
  content: "\fa6a";
}
.mdi-scooter::before {
  content: "\f0214";
}
.mdi-scoreboard::before {
  content: "\f02a9";
}
.mdi-scoreboard-outline::before {
  content: "\f02aa";
}
.mdi-screen-rotation::before {
  content: "\f475";
}
.mdi-screen-rotation-lock::before {
  content: "\f476";
}
.mdi-screw-flat-top::before {
  content: "\fdcf";
}
.mdi-screw-lag::before {
  content: "\fe54";
}
.mdi-screw-machine-flat-top::before {
  content: "\fe55";
}
.mdi-screw-machine-round-top::before {
  content: "\fe56";
}
.mdi-screw-round-top::before {
  content: "\fe57";
}
.mdi-screwdriver::before {
  content: "\f477";
}
.mdi-script::before {
  content: "\fb9d";
}
.mdi-script-outline::before {
  content: "\f478";
}
.mdi-script-text::before {
  content: "\fb9e";
}
.mdi-script-text-outline::before {
  content: "\fb9f";
}
.mdi-sd::before {
  content: "\f479";
}
.mdi-seal::before {
  content: "\f47a";
}
.mdi-seal-variant::before {
  content: "\fffa";
}
.mdi-search-web::before {
  content: "\f70e";
}
.mdi-seat::before {
  content: "\fc9f";
}
.mdi-seat-flat::before {
  content: "\f47b";
}
.mdi-seat-flat-angled::before {
  content: "\f47c";
}
.mdi-seat-individual-suite::before {
  content: "\f47d";
}
.mdi-seat-legroom-extra::before {
  content: "\f47e";
}
.mdi-seat-legroom-normal::before {
  content: "\f47f";
}
.mdi-seat-legroom-reduced::before {
  content: "\f480";
}
.mdi-seat-outline::before {
  content: "\fca0";
}
.mdi-seat-passenger::before {
  content: "\f0274";
}
.mdi-seat-recline-extra::before {
  content: "\f481";
}
.mdi-seat-recline-normal::before {
  content: "\f482";
}
.mdi-seatbelt::before {
  content: "\fca1";
}
.mdi-security::before {
  content: "\f483";
}
.mdi-security-network::before {
  content: "\f484";
}
.mdi-seed::before {
  content: "\fe45";
}
.mdi-seed-outline::before {
  content: "\fe46";
}
.mdi-segment::before {
  content: "\fee8";
}
.mdi-select::before {
  content: "\f485";
}
.mdi-select-all::before {
  content: "\f486";
}
.mdi-select-color::before {
  content: "\fd0d";
}
.mdi-select-compare::before {
  content: "\fad8";
}
.mdi-select-drag::before {
  content: "\fa6b";
}
.mdi-select-group::before {
  content: "\ff9f";
}
.mdi-select-inverse::before {
  content: "\f487";
}
.mdi-select-marker::before {
  content: "\f02ab";
}
.mdi-select-multiple::before {
  content: "\f02ac";
}
.mdi-select-multiple-marker::before {
  content: "\f02ad";
}
.mdi-select-off::before {
  content: "\f488";
}
.mdi-select-place::before {
  content: "\fffb";
}
.mdi-select-search::before {
  content: "\f022f";
}
.mdi-selection::before {
  content: "\f489";
}
.mdi-selection-drag::before {
  content: "\fa6c";
}
.mdi-selection-ellipse::before {
  content: "\fd0e";
}
.mdi-selection-ellipse-arrow-inside::before {
  content: "\ff3f";
}
.mdi-selection-marker::before {
  content: "\f02ae";
}
.mdi-selection-multiple-marker::before {
  content: "\f02af";
}
.mdi-selection-mutliple::before {
  content: "\f02b0";
}
.mdi-selection-off::before {
  content: "\f776";
}
.mdi-selection-search::before {
  content: "\f0230";
}
.mdi-semantic-web::before {
  content: "\f0341";
}
.mdi-send::before {
  content: "\f48a";
}
.mdi-send-check::before {
  content: "\f018c";
}
.mdi-send-check-outline::before {
  content: "\f018d";
}
.mdi-send-circle::before {
  content: "\fe58";
}
.mdi-send-circle-outline::before {
  content: "\fe59";
}
.mdi-send-clock::before {
  content: "\f018e";
}
.mdi-send-clock-outline::before {
  content: "\f018f";
}
.mdi-send-lock::before {
  content: "\f7ec";
}
.mdi-send-lock-outline::before {
  content: "\f0191";
}
.mdi-send-outline::before {
  content: "\f0190";
}
.mdi-serial-port::before {
  content: "\f65c";
}
.mdi-server::before {
  content: "\f48b";
}
.mdi-server-minus::before {
  content: "\f48c";
}
.mdi-server-network::before {
  content: "\f48d";
}
.mdi-server-network-off::before {
  content: "\f48e";
}
.mdi-server-off::before {
  content: "\f48f";
}
.mdi-server-plus::before {
  content: "\f490";
}
.mdi-server-remove::before {
  content: "\f491";
}
.mdi-server-security::before {
  content: "\f492";
}
.mdi-set-all::before {
  content: "\f777";
}
.mdi-set-center::before {
  content: "\f778";
}
.mdi-set-center-right::before {
  content: "\f779";
}
.mdi-set-left::before {
  content: "\f77a";
}
.mdi-set-left-center::before {
  content: "\f77b";
}
.mdi-set-left-right::before {
  content: "\f77c";
}
.mdi-set-none::before {
  content: "\f77d";
}
.mdi-set-right::before {
  content: "\f77e";
}
.mdi-set-top-box::before {
  content: "\f99e";
}
.mdi-settings::before {
  content: "\f493";
}
.mdi-settings-box::before {
  content: "\f494";
}
.mdi-settings-helper::before {
  content: "\fa6d";
}
.mdi-settings-outline::before {
  content: "\f8ba";
}
.mdi-settings-transfer::before {
  content: "\f007d";
}
.mdi-settings-transfer-outline::before {
  content: "\f007e";
}
.mdi-shaker::before {
  content: "\f0139";
}
.mdi-shaker-outline::before {
  content: "\f013a";
}
.mdi-shape::before {
  content: "\f830";
}
.mdi-shape-circle-plus::before {
  content: "\f65d";
}
.mdi-shape-outline::before {
  content: "\f831";
}
.mdi-shape-oval-plus::before {
  content: "\f0225";
}
.mdi-shape-plus::before {
  content: "\f495";
}
.mdi-shape-polygon-plus::before {
  content: "\f65e";
}
.mdi-shape-rectangle-plus::before {
  content: "\f65f";
}
.mdi-shape-square-plus::before {
  content: "\f660";
}
.mdi-share::before {
  content: "\f496";
}
.mdi-share-all::before {
  content: "\f021f";
}
.mdi-share-all-outline::before {
  content: "\f0220";
}
.mdi-share-circle::before {
  content: "\f01d8";
}
.mdi-share-off::before {
  content: "\ff40";
}
.mdi-share-off-outline::before {
  content: "\ff41";
}
.mdi-share-outline::before {
  content: "\f931";
}
.mdi-share-variant::before {
  content: "\f497";
}
.mdi-sheep::before {
  content: "\fca2";
}
.mdi-shield::before {
  content: "\f498";
}
.mdi-shield-account::before {
  content: "\f88e";
}
.mdi-shield-account-outline::before {
  content: "\fa11";
}
.mdi-shield-airplane::before {
  content: "\f6ba";
}
.mdi-shield-airplane-outline::before {
  content: "\fca3";
}
.mdi-shield-alert::before {
  content: "\fee9";
}
.mdi-shield-alert-outline::before {
  content: "\feea";
}
.mdi-shield-car::before {
  content: "\ffa0";
}
.mdi-shield-check::before {
  content: "\f565";
}
.mdi-shield-check-outline::before {
  content: "\fca4";
}
.mdi-shield-cross::before {
  content: "\fca5";
}
.mdi-shield-cross-outline::before {
  content: "\fca6";
}
.mdi-shield-edit::before {
  content: "\f01cb";
}
.mdi-shield-edit-outline::before {
  content: "\f01cc";
}
.mdi-shield-half::before {
  content: "\f038b";
}
.mdi-shield-half-full::before {
  content: "\f77f";
}
.mdi-shield-home::before {
  content: "\f689";
}
.mdi-shield-home-outline::before {
  content: "\fca7";
}
.mdi-shield-key::before {
  content: "\fba0";
}
.mdi-shield-key-outline::before {
  content: "\fba1";
}
.mdi-shield-link-variant::before {
  content: "\fd0f";
}
.mdi-shield-link-variant-outline::before {
  content: "\fd10";
}
.mdi-shield-lock::before {
  content: "\f99c";
}
.mdi-shield-lock-outline::before {
  content: "\fca8";
}
.mdi-shield-off::before {
  content: "\f99d";
}
.mdi-shield-off-outline::before {
  content: "\f99b";
}
.mdi-shield-outline::before {
  content: "\f499";
}
.mdi-shield-plus::before {
  content: "\fad9";
}
.mdi-shield-plus-outline::before {
  content: "\fada";
}
.mdi-shield-refresh::before {
  content: "\f01cd";
}
.mdi-shield-refresh-outline::before {
  content: "\f01ce";
}
.mdi-shield-remove::before {
  content: "\fadb";
}
.mdi-shield-remove-outline::before {
  content: "\fadc";
}
.mdi-shield-search::before {
  content: "\fd76";
}
.mdi-shield-star::before {
  content: "\f0166";
}
.mdi-shield-star-outline::before {
  content: "\f0167";
}
.mdi-shield-sun::before {
  content: "\f007f";
}
.mdi-shield-sun-outline::before {
  content: "\f0080";
}
.mdi-ship-wheel::before {
  content: "\f832";
}
.mdi-shoe-formal::before {
  content: "\fb22";
}
.mdi-shoe-heel::before {
  content: "\fb23";
}
.mdi-shoe-print::before {
  content: "\fe5a";
}
.mdi-shopify::before {
  content: "\fadd";
}
.mdi-shopping::before {
  content: "\f49a";
}
.mdi-shopping-music::before {
  content: "\f49b";
}
.mdi-shopping-outline::before {
  content: "\f0200";
}
.mdi-shopping-search::before {
  content: "\ffa1";
}
.mdi-shovel::before {
  content: "\f70f";
}
.mdi-shovel-off::before {
  content: "\f710";
}
.mdi-shower::before {
  content: "\f99f";
}
.mdi-shower-head::before {
  content: "\f9a0";
}
.mdi-shredder::before {
  content: "\f49c";
}
.mdi-shuffle::before {
  content: "\f49d";
}
.mdi-shuffle-disabled::before {
  content: "\f49e";
}
.mdi-shuffle-variant::before {
  content: "\f49f";
}
.mdi-shuriken::before {
  content: "\f03aa";
}
.mdi-sigma::before {
  content: "\f4a0";
}
.mdi-sigma-lower::before {
  content: "\f62b";
}
.mdi-sign-caution::before {
  content: "\f4a1";
}
.mdi-sign-direction::before {
  content: "\f780";
}
.mdi-sign-direction-minus::before {
  content: "\f0022";
}
.mdi-sign-direction-plus::before {
  content: "\fffd";
}
.mdi-sign-direction-remove::before {
  content: "\fffe";
}
.mdi-sign-real-estate::before {
  content: "\f0143";
}
.mdi-sign-text::before {
  content: "\f781";
}
.mdi-signal::before {
  content: "\f4a2";
}
.mdi-signal-2g::before {
  content: "\f711";
}
.mdi-signal-3g::before {
  content: "\f712";
}
.mdi-signal-4g::before {
  content: "\f713";
}
.mdi-signal-5g::before {
  content: "\fa6e";
}
.mdi-signal-cellular-1::before {
  content: "\f8bb";
}
.mdi-signal-cellular-2::before {
  content: "\f8bc";
}
.mdi-signal-cellular-3::before {
  content: "\f8bd";
}
.mdi-signal-cellular-outline::before {
  content: "\f8be";
}
.mdi-signal-distance-variant::before {
  content: "\fe47";
}
.mdi-signal-hspa::before {
  content: "\f714";
}
.mdi-signal-hspa-plus::before {
  content: "\f715";
}
.mdi-signal-off::before {
  content: "\f782";
}
.mdi-signal-variant::before {
  content: "\f60a";
}
.mdi-signature::before {
  content: "\fe5b";
}
.mdi-signature-freehand::before {
  content: "\fe5c";
}
.mdi-signature-image::before {
  content: "\fe5d";
}
.mdi-signature-text::before {
  content: "\fe5e";
}
.mdi-silo::before {
  content: "\fb24";
}
.mdi-silverware::before {
  content: "\f4a3";
}
.mdi-silverware-clean::before {
  content: "\ffff";
}
.mdi-silverware-fork::before {
  content: "\f4a4";
}
.mdi-silverware-fork-knife::before {
  content: "\fa6f";
}
.mdi-silverware-spoon::before {
  content: "\f4a5";
}
.mdi-silverware-variant::before {
  content: "\f4a6";
}
.mdi-sim::before {
  content: "\f4a7";
}
.mdi-sim-alert::before {
  content: "\f4a8";
}
.mdi-sim-off::before {
  content: "\f4a9";
}
.mdi-simple-icons::before {
  content: "\f0348";
}
.mdi-sina-weibo::before {
  content: "\fade";
}
.mdi-sitemap::before {
  content: "\f4aa";
}
.mdi-skate::before {
  content: "\fd11";
}
.mdi-skew-less::before {
  content: "\fd12";
}
.mdi-skew-more::before {
  content: "\fd13";
}
.mdi-ski::before {
  content: "\f032f";
}
.mdi-ski-cross-country::before {
  content: "\f0330";
}
.mdi-ski-water::before {
  content: "\f0331";
}
.mdi-skip-backward::before {
  content: "\f4ab";
}
.mdi-skip-backward-outline::before {
  content: "\ff42";
}
.mdi-skip-forward::before {
  content: "\f4ac";
}
.mdi-skip-forward-outline::before {
  content: "\ff43";
}
.mdi-skip-next::before {
  content: "\f4ad";
}
.mdi-skip-next-circle::before {
  content: "\f661";
}
.mdi-skip-next-circle-outline::before {
  content: "\f662";
}
.mdi-skip-next-outline::before {
  content: "\ff44";
}
.mdi-skip-previous::before {
  content: "\f4ae";
}
.mdi-skip-previous-circle::before {
  content: "\f663";
}
.mdi-skip-previous-circle-outline::before {
  content: "\f664";
}
.mdi-skip-previous-outline::before {
  content: "\ff45";
}
.mdi-skull::before {
  content: "\f68b";
}
.mdi-skull-crossbones::before {
  content: "\fba2";
}
.mdi-skull-crossbones-outline::before {
  content: "\fba3";
}
.mdi-skull-outline::before {
  content: "\fba4";
}
.mdi-skype::before {
  content: "\f4af";
}
.mdi-skype-business::before {
  content: "\f4b0";
}
.mdi-slack::before {
  content: "\f4b1";
}
.mdi-slackware::before {
  content: "\f90a";
}
.mdi-slash-forward::before {
  content: "\f0000";
}
.mdi-slash-forward-box::before {
  content: "\f0001";
}
.mdi-sleep::before {
  content: "\f4b2";
}
.mdi-sleep-off::before {
  content: "\f4b3";
}
.mdi-slope-downhill::before {
  content: "\fe5f";
}
.mdi-slope-uphill::before {
  content: "\fe60";
}
.mdi-slot-machine::before {
  content: "\f013f";
}
.mdi-slot-machine-outline::before {
  content: "\f0140";
}
.mdi-smart-card::before {
  content: "\f00e8";
}
.mdi-smart-card-outline::before {
  content: "\f00e9";
}
.mdi-smart-card-reader::before {
  content: "\f00ea";
}
.mdi-smart-card-reader-outline::before {
  content: "\f00eb";
}
.mdi-smog::before {
  content: "\fa70";
}
.mdi-smoke-detector::before {
  content: "\f392";
}
.mdi-smoking::before {
  content: "\f4b4";
}
.mdi-smoking-off::before {
  content: "\f4b5";
}
.mdi-snapchat::before {
  content: "\f4b6";
}
.mdi-snowboard::before {
  content: "\f0332";
}
.mdi-snowflake::before {
  content: "\f716";
}
.mdi-snowflake-alert::before {
  content: "\ff46";
}
.mdi-snowflake-melt::before {
  content: "\f02f6";
}
.mdi-snowflake-variant::before {
  content: "\ff47";
}
.mdi-snowman::before {
  content: "\f4b7";
}
.mdi-soccer::before {
  content: "\f4b8";
}
.mdi-soccer-field::before {
  content: "\f833";
}
.mdi-sofa::before {
  content: "\f4b9";
}
.mdi-solar-panel::before {
  content: "\fd77";
}
.mdi-solar-panel-large::before {
  content: "\fd78";
}
.mdi-solar-power::before {
  content: "\fa71";
}
.mdi-soldering-iron::before {
  content: "\f00bd";
}
.mdi-solid::before {
  content: "\f68c";
}
.mdi-sort::before {
  content: "\f4ba";
}
.mdi-sort-alphabetical::before {
  content: "\f4bb";
}
.mdi-sort-alphabetical-ascending::before {
  content: "\f0173";
}
.mdi-sort-alphabetical-descending::before {
  content: "\f0174";
}
.mdi-sort-ascending::before {
  content: "\f4bc";
}
.mdi-sort-descending::before {
  content: "\f4bd";
}
.mdi-sort-numeric::before {
  content: "\f4be";
}
.mdi-sort-variant::before {
  content: "\f4bf";
}
.mdi-sort-variant-lock::before {
  content: "\fca9";
}
.mdi-sort-variant-lock-open::before {
  content: "\fcaa";
}
.mdi-sort-variant-remove::before {
  content: "\f0172";
}
.mdi-soundcloud::before {
  content: "\f4c0";
}
.mdi-source-branch::before {
  content: "\f62c";
}
.mdi-source-commit::before {
  content: "\f717";
}
.mdi-source-commit-end::before {
  content: "\f718";
}
.mdi-source-commit-end-local::before {
  content: "\f719";
}
.mdi-source-commit-local::before {
  content: "\f71a";
}
.mdi-source-commit-next-local::before {
  content: "\f71b";
}
.mdi-source-commit-start::before {
  content: "\f71c";
}
.mdi-source-commit-start-next-local::before {
  content: "\f71d";
}
.mdi-source-fork::before {
  content: "\f4c1";
}
.mdi-source-merge::before {
  content: "\f62d";
}
.mdi-source-pull::before {
  content: "\f4c2";
}
.mdi-source-repository::before {
  content: "\fcab";
}
.mdi-source-repository-multiple::before {
  content: "\fcac";
}
.mdi-soy-sauce::before {
  content: "\f7ed";
}
.mdi-spa::before {
  content: "\fcad";
}
.mdi-spa-outline::before {
  content: "\fcae";
}
.mdi-space-invaders::before {
  content: "\fba5";
}
.mdi-space-station::before {
  content: "\f03ae";
}
.mdi-spade::before {
  content: "\fe48";
}
.mdi-speaker::before {
  content: "\f4c3";
}
.mdi-speaker-bluetooth::before {
  content: "\f9a1";
}
.mdi-speaker-multiple::before {
  content: "\fd14";
}
.mdi-speaker-off::before {
  content: "\f4c4";
}
.mdi-speaker-wireless::before {
  content: "\f71e";
}
.mdi-speedometer::before {
  content: "\f4c5";
}
.mdi-speedometer-medium::before {
  content: "\ffa2";
}
.mdi-speedometer-slow::before {
  content: "\ffa3";
}
.mdi-spellcheck::before {
  content: "\f4c6";
}
.mdi-spider::before {
  content: "\f0215";
}
.mdi-spider-thread::before {
  content: "\f0216";
}
.mdi-spider-web::before {
  content: "\fba6";
}
.mdi-spotify::before {
  content: "\f4c7";
}
.mdi-spotlight::before {
  content: "\f4c8";
}
.mdi-spotlight-beam::before {
  content: "\f4c9";
}
.mdi-spray::before {
  content: "\f665";
}
.mdi-spray-bottle::before {
  content: "\fadf";
}
.mdi-sprinkler::before {
  content: "\f0081";
}
.mdi-sprinkler-variant::before {
  content: "\f0082";
}
.mdi-sprout::before {
  content: "\fe49";
}
.mdi-sprout-outline::before {
  content: "\fe4a";
}
.mdi-square::before {
  content: "\f763";
}
.mdi-square-edit-outline::before {
  content: "\f90b";
}
.mdi-square-inc::before {
  content: "\f4ca";
}
.mdi-square-inc-cash::before {
  content: "\f4cb";
}
.mdi-square-medium::before {
  content: "\fa12";
}
.mdi-square-medium-outline::before {
  content: "\fa13";
}
.mdi-square-off::before {
  content: "\f0319";
}
.mdi-square-off-outline::before {
  content: "\f031a";
}
.mdi-square-outline::before {
  content: "\f762";
}
.mdi-square-root::before {
  content: "\f783";
}
.mdi-square-root-box::before {
  content: "\f9a2";
}
.mdi-square-small::before {
  content: "\fa14";
}
.mdi-squeegee::before {
  content: "\fae0";
}
.mdi-ssh::before {
  content: "\f8bf";
}
.mdi-stack-exchange::before {
  content: "\f60b";
}
.mdi-stack-overflow::before {
  content: "\f4cc";
}
.mdi-stackpath::before {
  content: "\f359";
}
.mdi-stadium::before {
  content: "\f001a";
}
.mdi-stadium-variant::before {
  content: "\f71f";
}
.mdi-stairs::before {
  content: "\f4cd";
}
.mdi-stairs-down::before {
  content: "\f02e9";
}
.mdi-stairs-up::before {
  content: "\f02e8";
}
.mdi-stamper::before {
  content: "\fd15";
}
.mdi-standard-definition::before {
  content: "\f7ee";
}
.mdi-star::before {
  content: "\f4ce";
}
.mdi-star-box::before {
  content: "\fa72";
}
.mdi-star-box-multiple::before {
  content: "\f02b1";
}
.mdi-star-box-multiple-outline::before {
  content: "\f02b2";
}
.mdi-star-box-outline::before {
  content: "\fa73";
}
.mdi-star-circle::before {
  content: "\f4cf";
}
.mdi-star-circle-outline::before {
  content: "\f9a3";
}
.mdi-star-face::before {
  content: "\f9a4";
}
.mdi-star-four-points::before {
  content: "\fae1";
}
.mdi-star-four-points-outline::before {
  content: "\fae2";
}
.mdi-star-half::before {
  content: "\f4d0";
}
.mdi-star-off::before {
  content: "\f4d1";
}
.mdi-star-outline::before {
  content: "\f4d2";
}
.mdi-star-three-points::before {
  content: "\fae3";
}
.mdi-star-three-points-outline::before {
  content: "\fae4";
}
.mdi-state-machine::before {
  content: "\f021a";
}
.mdi-steam::before {
  content: "\f4d3";
}
.mdi-steam-box::before {
  content: "\f90c";
}
.mdi-steering::before {
  content: "\f4d4";
}
.mdi-steering-off::before {
  content: "\f90d";
}
.mdi-step-backward::before {
  content: "\f4d5";
}
.mdi-step-backward-2::before {
  content: "\f4d6";
}
.mdi-step-forward::before {
  content: "\f4d7";
}
.mdi-step-forward-2::before {
  content: "\f4d8";
}
.mdi-stethoscope::before {
  content: "\f4d9";
}
.mdi-sticker::before {
  content: "\f038f";
}
.mdi-sticker-alert::before {
  content: "\f0390";
}
.mdi-sticker-alert-outline::before {
  content: "\f0391";
}
.mdi-sticker-check::before {
  content: "\f0392";
}
.mdi-sticker-check-outline::before {
  content: "\f0393";
}
.mdi-sticker-circle-outline::before {
  content: "\f5d0";
}
.mdi-sticker-emoji::before {
  content: "\f784";
}
.mdi-sticker-minus::before {
  content: "\f0394";
}
.mdi-sticker-minus-outline::before {
  content: "\f0395";
}
.mdi-sticker-outline::before {
  content: "\f0396";
}
.mdi-sticker-plus::before {
  content: "\f0397";
}
.mdi-sticker-plus-outline::before {
  content: "\f0398";
}
.mdi-sticker-remove::before {
  content: "\f0399";
}
.mdi-sticker-remove-outline::before {
  content: "\f039a";
}
.mdi-stocking::before {
  content: "\f4da";
}
.mdi-stomach::before {
  content: "\f00be";
}
.mdi-stop::before {
  content: "\f4db";
}
.mdi-stop-circle::before {
  content: "\f666";
}
.mdi-stop-circle-outline::before {
  content: "\f667";
}
.mdi-store::before {
  content: "\f4dc";
}
.mdi-store-24-hour::before {
  content: "\f4dd";
}
.mdi-store-outline::before {
  content: "\f038c";
}
.mdi-storefront::before {
  content: "\f00ec";
}
.mdi-stove::before {
  content: "\f4de";
}
.mdi-strategy::before {
  content: "\f0201";
}
.mdi-strava::before {
  content: "\fb25";
}
.mdi-stretch-to-page::before {
  content: "\ff48";
}
.mdi-stretch-to-page-outline::before {
  content: "\ff49";
}
.mdi-string-lights::before {
  content: "\f02e5";
}
.mdi-string-lights-off::before {
  content: "\f02e6";
}
.mdi-subdirectory-arrow-left::before {
  content: "\f60c";
}
.mdi-subdirectory-arrow-right::before {
  content: "\f60d";
}
.mdi-subtitles::before {
  content: "\fa15";
}
.mdi-subtitles-outline::before {
  content: "\fa16";
}
.mdi-subway::before {
  content: "\f6ab";
}
.mdi-subway-alert-variant::before {
  content: "\fd79";
}
.mdi-subway-variant::before {
  content: "\f4df";
}
.mdi-summit::before {
  content: "\f785";
}
.mdi-sunglasses::before {
  content: "\f4e0";
}
.mdi-surround-sound::before {
  content: "\f5c5";
}
.mdi-surround-sound-2-0::before {
  content: "\f7ef";
}
.mdi-surround-sound-3-1::before {
  content: "\f7f0";
}
.mdi-surround-sound-5-1::before {
  content: "\f7f1";
}
.mdi-surround-sound-7-1::before {
  content: "\f7f2";
}
.mdi-svg::before {
  content: "\f720";
}
.mdi-swap-horizontal::before {
  content: "\f4e1";
}
.mdi-swap-horizontal-bold::before {
  content: "\fba9";
}
.mdi-swap-horizontal-circle::before {
  content: "\f0002";
}
.mdi-swap-horizontal-circle-outline::before {
  content: "\f0003";
}
.mdi-swap-horizontal-variant::before {
  content: "\f8c0";
}
.mdi-swap-vertical::before {
  content: "\f4e2";
}
.mdi-swap-vertical-bold::before {
  content: "\fbaa";
}
.mdi-swap-vertical-circle::before {
  content: "\f0004";
}
.mdi-swap-vertical-circle-outline::before {
  content: "\f0005";
}
.mdi-swap-vertical-variant::before {
  content: "\f8c1";
}
.mdi-swim::before {
  content: "\f4e3";
}
.mdi-switch::before {
  content: "\f4e4";
}
.mdi-sword::before {
  content: "\f4e5";
}
.mdi-sword-cross::before {
  content: "\f786";
}
.mdi-syllabary-hangul::before {
  content: "\f035e";
}
.mdi-syllabary-hiragana::before {
  content: "\f035f";
}
.mdi-syllabary-katakana::before {
  content: "\f0360";
}
.mdi-syllabary-katakana-half-width::before {
  content: "\f0361";
}
.mdi-symfony::before {
  content: "\fae5";
}
.mdi-sync::before {
  content: "\f4e6";
}
.mdi-sync-alert::before {
  content: "\f4e7";
}
.mdi-sync-circle::before {
  content: "\f03a3";
}
.mdi-sync-off::before {
  content: "\f4e8";
}
.mdi-tab::before {
  content: "\f4e9";
}
.mdi-tab-minus::before {
  content: "\fb26";
}
.mdi-tab-plus::before {
  content: "\f75b";
}
.mdi-tab-remove::before {
  content: "\fb27";
}
.mdi-tab-unselected::before {
  content: "\f4ea";
}
.mdi-table::before {
  content: "\f4eb";
}
.mdi-table-border::before {
  content: "\fa17";
}
.mdi-table-chair::before {
  content: "\f0083";
}
.mdi-table-column::before {
  content: "\f834";
}
.mdi-table-column-plus-after::before {
  content: "\f4ec";
}
.mdi-table-column-plus-before::before {
  content: "\f4ed";
}
.mdi-table-column-remove::before {
  content: "\f4ee";
}
.mdi-table-column-width::before {
  content: "\f4ef";
}
.mdi-table-edit::before {
  content: "\f4f0";
}
.mdi-table-eye::before {
  content: "\f00bf";
}
.mdi-table-headers-eye::before {
  content: "\f0248";
}
.mdi-table-headers-eye-off::before {
  content: "\f0249";
}
.mdi-table-large::before {
  content: "\f4f1";
}
.mdi-table-large-plus::before {
  content: "\ffa4";
}
.mdi-table-large-remove::before {
  content: "\ffa5";
}
.mdi-table-merge-cells::before {
  content: "\f9a5";
}
.mdi-table-of-contents::before {
  content: "\f835";
}
.mdi-table-plus::before {
  content: "\fa74";
}
.mdi-table-remove::before {
  content: "\fa75";
}
.mdi-table-row::before {
  content: "\f836";
}
.mdi-table-row-height::before {
  content: "\f4f2";
}
.mdi-table-row-plus-after::before {
  content: "\f4f3";
}
.mdi-table-row-plus-before::before {
  content: "\f4f4";
}
.mdi-table-row-remove::before {
  content: "\f4f5";
}
.mdi-table-search::before {
  content: "\f90e";
}
.mdi-table-settings::before {
  content: "\f837";
}
.mdi-table-tennis::before {
  content: "\fe4b";
}
.mdi-tablet::before {
  content: "\f4f6";
}
.mdi-tablet-android::before {
  content: "\f4f7";
}
.mdi-tablet-cellphone::before {
  content: "\f9a6";
}
.mdi-tablet-dashboard::before {
  content: "\feeb";
}
.mdi-tablet-ipad::before {
  content: "\f4f8";
}
.mdi-taco::before {
  content: "\f761";
}
.mdi-tag::before {
  content: "\f4f9";
}
.mdi-tag-faces::before {
  content: "\f4fa";
}
.mdi-tag-heart::before {
  content: "\f68a";
}
.mdi-tag-heart-outline::before {
  content: "\fbab";
}
.mdi-tag-minus::before {
  content: "\f90f";
}
.mdi-tag-minus-outline::before {
  content: "\f024a";
}
.mdi-tag-multiple::before {
  content: "\f4fb";
}
.mdi-tag-multiple-outline::before {
  content: "\f0322";
}
.mdi-tag-off::before {
  content: "\f024b";
}
.mdi-tag-off-outline::before {
  content: "\f024c";
}
.mdi-tag-outline::before {
  content: "\f4fc";
}
.mdi-tag-plus::before {
  content: "\f721";
}
.mdi-tag-plus-outline::before {
  content: "\f024d";
}
.mdi-tag-remove::before {
  content: "\f722";
}
.mdi-tag-remove-outline::before {
  content: "\f024e";
}
.mdi-tag-text::before {
  content: "\f024f";
}
.mdi-tag-text-outline::before {
  content: "\f4fd";
}
.mdi-tank::before {
  content: "\fd16";
}
.mdi-tanker-truck::before {
  content: "\f0006";
}
.mdi-tape-measure::before {
  content: "\fb28";
}
.mdi-target::before {
  content: "\f4fe";
}
.mdi-target-account::before {
  content: "\fbac";
}
.mdi-target-variant::before {
  content: "\fa76";
}
.mdi-taxi::before {
  content: "\f4ff";
}
.mdi-tea::before {
  content: "\fd7a";
}
.mdi-tea-outline::before {
  content: "\fd7b";
}
.mdi-teach::before {
  content: "\f88f";
}
.mdi-teamviewer::before {
  content: "\f500";
}
.mdi-telegram::before {
  content: "\f501";
}
.mdi-telescope::before {
  content: "\fb29";
}
.mdi-television::before {
  content: "\f502";
}
.mdi-television-ambient-light::before {
  content: "\f0381";
}
.mdi-television-box::before {
  content: "\f838";
}
.mdi-television-classic::before {
  content: "\f7f3";
}
.mdi-television-classic-off::before {
  content: "\f839";
}
.mdi-television-clean::before {
  content: "\f013b";
}
.mdi-television-guide::before {
  content: "\f503";
}
.mdi-television-off::before {
  content: "\f83a";
}
.mdi-television-pause::before {
  content: "\ffa6";
}
.mdi-television-play::before {
  content: "\feec";
}
.mdi-television-stop::before {
  content: "\ffa7";
}
.mdi-temperature-celsius::before {
  content: "\f504";
}
.mdi-temperature-fahrenheit::before {
  content: "\f505";
}
.mdi-temperature-kelvin::before {
  content: "\f506";
}
.mdi-tennis::before {
  content: "\fd7c";
}
.mdi-tennis-ball::before {
  content: "\f507";
}
.mdi-tent::before {
  content: "\f508";
}
.mdi-terraform::before {
  content: "\f0084";
}
.mdi-terrain::before {
  content: "\f509";
}
.mdi-test-tube::before {
  content: "\f668";
}
.mdi-test-tube-empty::before {
  content: "\f910";
}
.mdi-test-tube-off::before {
  content: "\f911";
}
.mdi-text::before {
  content: "\f9a7";
}
.mdi-text-recognition::before {
  content: "\f0168";
}
.mdi-text-shadow::before {
  content: "\f669";
}
.mdi-text-short::before {
  content: "\f9a8";
}
.mdi-text-subject::before {
  content: "\f9a9";
}
.mdi-text-to-speech::before {
  content: "\f50a";
}
.mdi-text-to-speech-off::before {
  content: "\f50b";
}
.mdi-textarea::before {
  content: "\f00c0";
}
.mdi-textbox::before {
  content: "\f60e";
}
.mdi-textbox-lock::before {
  content: "\f0388";
}
.mdi-textbox-password::before {
  content: "\f7f4";
}
.mdi-texture::before {
  content: "\f50c";
}
.mdi-texture-box::before {
  content: "\f0007";
}
.mdi-theater::before {
  content: "\f50d";
}
.mdi-theme-light-dark::before {
  content: "\f50e";
}
.mdi-thermometer::before {
  content: "\f50f";
}
.mdi-thermometer-alert::before {
  content: "\fe61";
}
.mdi-thermometer-chevron-down::before {
  content: "\fe62";
}
.mdi-thermometer-chevron-up::before {
  content: "\fe63";
}
.mdi-thermometer-high::before {
  content: "\f00ed";
}
.mdi-thermometer-lines::before {
  content: "\f510";
}
.mdi-thermometer-low::before {
  content: "\f00ee";
}
.mdi-thermometer-minus::before {
  content: "\fe64";
}
.mdi-thermometer-plus::before {
  content: "\fe65";
}
.mdi-thermostat::before {
  content: "\f393";
}
.mdi-thermostat-box::before {
  content: "\f890";
}
.mdi-thought-bubble::before {
  content: "\f7f5";
}
.mdi-thought-bubble-outline::before {
  content: "\f7f6";
}
.mdi-thumb-down::before {
  content: "\f511";
}
.mdi-thumb-down-outline::before {
  content: "\f512";
}
.mdi-thumb-up::before {
  content: "\f513";
}
.mdi-thumb-up-outline::before {
  content: "\f514";
}
.mdi-thumbs-up-down::before {
  content: "\f515";
}
.mdi-ticket::before {
  content: "\f516";
}
.mdi-ticket-account::before {
  content: "\f517";
}
.mdi-ticket-confirmation::before {
  content: "\f518";
}
.mdi-ticket-outline::before {
  content: "\f912";
}
.mdi-ticket-percent::before {
  content: "\f723";
}
.mdi-tie::before {
  content: "\f519";
}
.mdi-tilde::before {
  content: "\f724";
}
.mdi-timelapse::before {
  content: "\f51a";
}
.mdi-timeline::before {
  content: "\fbad";
}
.mdi-timeline-alert::before {
  content: "\ffb2";
}
.mdi-timeline-alert-outline::before {
  content: "\ffb5";
}
.mdi-timeline-clock::before {
  content: "\f0226";
}
.mdi-timeline-clock-outline::before {
  content: "\f0227";
}
.mdi-timeline-help::before {
  content: "\ffb6";
}
.mdi-timeline-help-outline::before {
  content: "\ffb7";
}
.mdi-timeline-outline::before {
  content: "\fbae";
}
.mdi-timeline-plus::before {
  content: "\ffb3";
}
.mdi-timeline-plus-outline::before {
  content: "\ffb4";
}
.mdi-timeline-text::before {
  content: "\fbaf";
}
.mdi-timeline-text-outline::before {
  content: "\fbb0";
}
.mdi-timer::before {
  content: "\f51b";
}
.mdi-timer-10::before {
  content: "\f51c";
}
.mdi-timer-3::before {
  content: "\f51d";
}
.mdi-timer-off::before {
  content: "\f51e";
}
.mdi-timer-sand::before {
  content: "\f51f";
}
.mdi-timer-sand-empty::before {
  content: "\f6ac";
}
.mdi-timer-sand-full::before {
  content: "\f78b";
}
.mdi-timetable::before {
  content: "\f520";
}
.mdi-toaster::before {
  content: "\f0085";
}
.mdi-toaster-off::before {
  content: "\f01e2";
}
.mdi-toaster-oven::before {
  content: "\fcaf";
}
.mdi-toggle-switch::before {
  content: "\f521";
}
.mdi-toggle-switch-off::before {
  content: "\f522";
}
.mdi-toggle-switch-off-outline::before {
  content: "\fa18";
}
.mdi-toggle-switch-outline::before {
  content: "\fa19";
}
.mdi-toilet::before {
  content: "\f9aa";
}
.mdi-toolbox::before {
  content: "\f9ab";
}
.mdi-toolbox-outline::before {
  content: "\f9ac";
}
.mdi-tools::before {
  content: "\f0086";
}
.mdi-tooltip::before {
  content: "\f523";
}
.mdi-tooltip-account::before {
  content: "\f00c";
}
.mdi-tooltip-edit::before {
  content: "\f524";
}
.mdi-tooltip-edit-outline::before {
  content: "\f02f0";
}
.mdi-tooltip-image::before {
  content: "\f525";
}
.mdi-tooltip-image-outline::before {
  content: "\fbb1";
}
.mdi-tooltip-outline::before {
  content: "\f526";
}
.mdi-tooltip-plus::before {
  content: "\fbb2";
}
.mdi-tooltip-plus-outline::before {
  content: "\f527";
}
.mdi-tooltip-text::before {
  content: "\f528";
}
.mdi-tooltip-text-outline::before {
  content: "\fbb3";
}
.mdi-tooth::before {
  content: "\f8c2";
}
.mdi-tooth-outline::before {
  content: "\f529";
}
.mdi-toothbrush::before {
  content: "\f0154";
}
.mdi-toothbrush-electric::before {
  content: "\f0157";
}
.mdi-toothbrush-paste::before {
  content: "\f0155";
}
.mdi-tor::before {
  content: "\f52a";
}
.mdi-tortoise::before {
  content: "\fd17";
}
.mdi-toslink::before {
  content: "\f02e3";
}
.mdi-tournament::before {
  content: "\f9ad";
}
.mdi-tower-beach::before {
  content: "\f680";
}
.mdi-tower-fire::before {
  content: "\f681";
}
.mdi-towing::before {
  content: "\f83b";
}
.mdi-toy-brick::before {
  content: "\f02b3";
}
.mdi-toy-brick-marker::before {
  content: "\f02b4";
}
.mdi-toy-brick-marker-outline::before {
  content: "\f02b5";
}
.mdi-toy-brick-minus::before {
  content: "\f02b6";
}
.mdi-toy-brick-minus-outline::before {
  content: "\f02b7";
}
.mdi-toy-brick-outline::before {
  content: "\f02b8";
}
.mdi-toy-brick-plus::before {
  content: "\f02b9";
}
.mdi-toy-brick-plus-outline::before {
  content: "\f02ba";
}
.mdi-toy-brick-remove::before {
  content: "\f02bb";
}
.mdi-toy-brick-remove-outline::before {
  content: "\f02bc";
}
.mdi-toy-brick-search::before {
  content: "\f02bd";
}
.mdi-toy-brick-search-outline::before {
  content: "\f02be";
}
.mdi-track-light::before {
  content: "\f913";
}
.mdi-trackpad::before {
  content: "\f7f7";
}
.mdi-trackpad-lock::before {
  content: "\f932";
}
.mdi-tractor::before {
  content: "\f891";
}
.mdi-trademark::before {
  content: "\fa77";
}
.mdi-traffic-cone::before {
  content: "\f03a7";
}
.mdi-traffic-light::before {
  content: "\f52b";
}
.mdi-train::before {
  content: "\f52c";
}
.mdi-train-car::before {
  content: "\fbb4";
}
.mdi-train-variant::before {
  content: "\f8c3";
}
.mdi-tram::before {
  content: "\f52d";
}
.mdi-tram-side::before {
  content: "\f0008";
}
.mdi-transcribe::before {
  content: "\f52e";
}
.mdi-transcribe-close::before {
  content: "\f52f";
}
.mdi-transfer::before {
  content: "\f0087";
}
.mdi-transfer-down::before {
  content: "\fd7d";
}
.mdi-transfer-left::before {
  content: "\fd7e";
}
.mdi-transfer-right::before {
  content: "\f530";
}
.mdi-transfer-up::before {
  content: "\fd7f";
}
.mdi-transit-connection::before {
  content: "\fd18";
}
.mdi-transit-connection-variant::before {
  content: "\fd19";
}
.mdi-transit-detour::before {
  content: "\ffa8";
}
.mdi-transit-transfer::before {
  content: "\f6ad";
}
.mdi-transition::before {
  content: "\f914";
}
.mdi-transition-masked::before {
  content: "\f915";
}
.mdi-translate::before {
  content: "\f5ca";
}
.mdi-translate-off::before {
  content: "\fe66";
}
.mdi-transmission-tower::before {
  content: "\fd1a";
}
.mdi-trash-can::before {
  content: "\fa78";
}
.mdi-trash-can-outline::before {
  content: "\fa79";
}
.mdi-tray::before {
  content: "\f02bf";
}
.mdi-tray-alert::before {
  content: "\f02c0";
}
.mdi-tray-full::before {
  content: "\f02c1";
}
.mdi-tray-minus::before {
  content: "\f02c2";
}
.mdi-tray-plus::before {
  content: "\f02c3";
}
.mdi-tray-remove::before {
  content: "\f02c4";
}
.mdi-treasure-chest::before {
  content: "\f725";
}
.mdi-tree::before {
  content: "\f531";
}
.mdi-tree-outline::before {
  content: "\fe4c";
}
.mdi-trello::before {
  content: "\f532";
}
.mdi-trending-down::before {
  content: "\f533";
}
.mdi-trending-neutral::before {
  content: "\f534";
}
.mdi-trending-up::before {
  content: "\f535";
}
.mdi-triangle::before {
  content: "\f536";
}
.mdi-triangle-outline::before {
  content: "\f537";
}
.mdi-triforce::before {
  content: "\fbb5";
}
.mdi-trophy::before {
  content: "\f538";
}
.mdi-trophy-award::before {
  content: "\f539";
}
.mdi-trophy-broken::before {
  content: "\fd80";
}
.mdi-trophy-outline::before {
  content: "\f53a";
}
.mdi-trophy-variant::before {
  content: "\f53b";
}
.mdi-trophy-variant-outline::before {
  content: "\f53c";
}
.mdi-truck::before {
  content: "\f53d";
}
.mdi-truck-check::before {
  content: "\fcb0";
}
.mdi-truck-check-outline::before {
  content: "\f02c5";
}
.mdi-truck-delivery::before {
  content: "\f53e";
}
.mdi-truck-delivery-outline::before {
  content: "\f02c6";
}
.mdi-truck-fast::before {
  content: "\f787";
}
.mdi-truck-fast-outline::before {
  content: "\f02c7";
}
.mdi-truck-outline::before {
  content: "\f02c8";
}
.mdi-truck-trailer::before {
  content: "\f726";
}
.mdi-trumpet::before {
  content: "\f00c1";
}
.mdi-tshirt-crew::before {
  content: "\fa7a";
}
.mdi-tshirt-crew-outline::before {
  content: "\f53f";
}
.mdi-tshirt-v::before {
  content: "\fa7b";
}
.mdi-tshirt-v-outline::before {
  content: "\f540";
}
.mdi-tumble-dryer::before {
  content: "\f916";
}
.mdi-tumble-dryer-alert::before {
  content: "\f01e5";
}
.mdi-tumble-dryer-off::before {
  content: "\f01e6";
}
.mdi-tumblr::before {
  content: "\f541";
}
.mdi-tumblr-box::before {
  content: "\f917";
}
.mdi-tumblr-reblog::before {
  content: "\f542";
}
.mdi-tune::before {
  content: "\f62e";
}
.mdi-tune-vertical::before {
  content: "\f66a";
}
.mdi-turnstile::before {
  content: "\fcb1";
}
.mdi-turnstile-outline::before {
  content: "\fcb2";
}
.mdi-turtle::before {
  content: "\fcb3";
}
.mdi-twitch::before {
  content: "\f543";
}
.mdi-twitter::before {
  content: "\f544";
}
.mdi-twitter-box::before {
  content: "\f545";
}
.mdi-twitter-circle::before {
  content: "\f546";
}
.mdi-twitter-retweet::before {
  content: "\f547";
}
.mdi-two-factor-authentication::before {
  content: "\f9ae";
}
.mdi-typewriter::before {
  content: "\ff4a";
}
.mdi-uber::before {
  content: "\f748";
}
.mdi-ubisoft::before {
  content: "\fbb6";
}
.mdi-ubuntu::before {
  content: "\f548";
}
.mdi-ufo::before {
  content: "\f00ef";
}
.mdi-ufo-outline::before {
  content: "\f00f0";
}
.mdi-ultra-high-definition::before {
  content: "\f7f8";
}
.mdi-umbraco::before {
  content: "\f549";
}
.mdi-umbrella::before {
  content: "\f54a";
}
.mdi-umbrella-closed::before {
  content: "\f9af";
}
.mdi-umbrella-outline::before {
  content: "\f54b";
}
.mdi-undo::before {
  content: "\f54c";
}
.mdi-undo-variant::before {
  content: "\f54d";
}
.mdi-unfold-less-horizontal::before {
  content: "\f54e";
}
.mdi-unfold-less-vertical::before {
  content: "\f75f";
}
.mdi-unfold-more-horizontal::before {
  content: "\f54f";
}
.mdi-unfold-more-vertical::before {
  content: "\f760";
}
.mdi-ungroup::before {
  content: "\f550";
}
.mdi-unicode::before {
  content: "\feed";
}
.mdi-unity::before {
  content: "\f6ae";
}
.mdi-unreal::before {
  content: "\f9b0";
}
.mdi-untappd::before {
  content: "\f551";
}
.mdi-update::before {
  content: "\f6af";
}
.mdi-upload::before {
  content: "\f552";
}
.mdi-upload-lock::before {
  content: "\f039e";
}
.mdi-upload-lock-outline::before {
  content: "\f039f";
}
.mdi-upload-multiple::before {
  content: "\f83c";
}
.mdi-upload-network::before {
  content: "\f6f5";
}
.mdi-upload-network-outline::before {
  content: "\fcb4";
}
.mdi-upload-off::before {
  content: "\f00f1";
}
.mdi-upload-off-outline::before {
  content: "\f00f2";
}
.mdi-upload-outline::before {
  content: "\fe67";
}
.mdi-usb::before {
  content: "\f553";
}
.mdi-usb-flash-drive::before {
  content: "\f02c9";
}
.mdi-usb-flash-drive-outline::before {
  content: "\f02ca";
}
.mdi-usb-port::before {
  content: "\f021b";
}
.mdi-valve::before {
  content: "\f0088";
}
.mdi-valve-closed::before {
  content: "\f0089";
}
.mdi-valve-open::before {
  content: "\f008a";
}
.mdi-van-passenger::before {
  content: "\f7f9";
}
.mdi-van-utility::before {
  content: "\f7fa";
}
.mdi-vanish::before {
  content: "\f7fb";
}
.mdi-vanity-light::before {
  content: "\f020c";
}
.mdi-variable::before {
  content: "\fae6";
}
.mdi-variable-box::before {
  content: "\f013c";
}
.mdi-vector-arrange-above::before {
  content: "\f554";
}
.mdi-vector-arrange-below::before {
  content: "\f555";
}
.mdi-vector-bezier::before {
  content: "\fae7";
}
.mdi-vector-circle::before {
  content: "\f556";
}
.mdi-vector-circle-variant::before {
  content: "\f557";
}
.mdi-vector-combine::before {
  content: "\f558";
}
.mdi-vector-curve::before {
  content: "\f559";
}
.mdi-vector-difference::before {
  content: "\f55a";
}
.mdi-vector-difference-ab::before {
  content: "\f55b";
}
.mdi-vector-difference-ba::before {
  content: "\f55c";
}
.mdi-vector-ellipse::before {
  content: "\f892";
}
.mdi-vector-intersection::before {
  content: "\f55d";
}
.mdi-vector-line::before {
  content: "\f55e";
}
.mdi-vector-link::before {
  content: "\f0009";
}
.mdi-vector-point::before {
  content: "\f55f";
}
.mdi-vector-polygon::before {
  content: "\f560";
}
.mdi-vector-polyline::before {
  content: "\f561";
}
.mdi-vector-polyline-edit::before {
  content: "\f0250";
}
.mdi-vector-polyline-minus::before {
  content: "\f0251";
}
.mdi-vector-polyline-plus::before {
  content: "\f0252";
}
.mdi-vector-polyline-remove::before {
  content: "\f0253";
}
.mdi-vector-radius::before {
  content: "\f749";
}
.mdi-vector-rectangle::before {
  content: "\f5c6";
}
.mdi-vector-selection::before {
  content: "\f562";
}
.mdi-vector-square::before {
  content: "\f001";
}
.mdi-vector-triangle::before {
  content: "\f563";
}
.mdi-vector-union::before {
  content: "\f564";
}
.mdi-venmo::before {
  content: "\f578";
}
.mdi-vhs::before {
  content: "\fa1a";
}
.mdi-vibrate::before {
  content: "\f566";
}
.mdi-vibrate-off::before {
  content: "\fcb5";
}
.mdi-video::before {
  content: "\f567";
}
.mdi-video-3d::before {
  content: "\f7fc";
}
.mdi-video-3d-variant::before {
  content: "\feee";
}
.mdi-video-4k-box::before {
  content: "\f83d";
}
.mdi-video-account::before {
  content: "\f918";
}
.mdi-video-check::before {
  content: "\f008b";
}
.mdi-video-check-outline::before {
  content: "\f008c";
}
.mdi-video-image::before {
  content: "\f919";
}
.mdi-video-input-antenna::before {
  content: "\f83e";
}
.mdi-video-input-component::before {
  content: "\f83f";
}
.mdi-video-input-hdmi::before {
  content: "\f840";
}
.mdi-video-input-scart::before {
  content: "\ffa9";
}
.mdi-video-input-svideo::before {
  content: "\f841";
}
.mdi-video-minus::before {
  content: "\f9b1";
}
.mdi-video-off::before {
  content: "\f568";
}
.mdi-video-off-outline::before {
  content: "\fbb7";
}
.mdi-video-outline::before {
  content: "\fbb8";
}
.mdi-video-plus::before {
  content: "\f9b2";
}
.mdi-video-stabilization::before {
  content: "\f91a";
}
.mdi-video-switch::before {
  content: "\f569";
}
.mdi-video-vintage::before {
  content: "\fa1b";
}
.mdi-video-wireless::before {
  content: "\feef";
}
.mdi-video-wireless-outline::before {
  content: "\fef0";
}
.mdi-view-agenda::before {
  content: "\f56a";
}
.mdi-view-agenda-outline::before {
  content: "\f0203";
}
.mdi-view-array::before {
  content: "\f56b";
}
.mdi-view-carousel::before {
  content: "\f56c";
}
.mdi-view-column::before {
  content: "\f56d";
}
.mdi-view-comfy::before {
  content: "\fe4d";
}
.mdi-view-compact::before {
  content: "\fe4e";
}
.mdi-view-compact-outline::before {
  content: "\fe4f";
}
.mdi-view-dashboard::before {
  content: "\f56e";
}
.mdi-view-dashboard-outline::before {
  content: "\fa1c";
}
.mdi-view-dashboard-variant::before {
  content: "\f842";
}
.mdi-view-day::before {
  content: "\f56f";
}
.mdi-view-grid::before {
  content: "\f570";
}
.mdi-view-grid-outline::before {
  content: "\f0204";
}
.mdi-view-grid-plus::before {
  content: "\ffaa";
}
.mdi-view-grid-plus-outline::before {
  content: "\f0205";
}
.mdi-view-headline::before {
  content: "\f571";
}
.mdi-view-list::before {
  content: "\f572";
}
.mdi-view-module::before {
  content: "\f573";
}
.mdi-view-parallel::before {
  content: "\f727";
}
.mdi-view-quilt::before {
  content: "\f574";
}
.mdi-view-sequential::before {
  content: "\f728";
}
.mdi-view-split-horizontal::before {
  content: "\fba7";
}
.mdi-view-split-vertical::before {
  content: "\fba8";
}
.mdi-view-stream::before {
  content: "\f575";
}
.mdi-view-week::before {
  content: "\f576";
}
.mdi-vimeo::before {
  content: "\f577";
}
.mdi-violin::before {
  content: "\f60f";
}
.mdi-virtual-reality::before {
  content: "\f893";
}
.mdi-visual-studio::before {
  content: "\f610";
}
.mdi-visual-studio-code::before {
  content: "\fa1d";
}
.mdi-vk::before {
  content: "\f579";
}
.mdi-vk-box::before {
  content: "\f57a";
}
.mdi-vk-circle::before {
  content: "\f57b";
}
.mdi-vlc::before {
  content: "\f57c";
}
.mdi-voice::before {
  content: "\f5cb";
}
.mdi-voice-off::before {
  content: "\fef1";
}
.mdi-voicemail::before {
  content: "\f57d";
}
.mdi-volleyball::before {
  content: "\f9b3";
}
.mdi-volume-high::before {
  content: "\f57e";
}
.mdi-volume-low::before {
  content: "\f57f";
}
.mdi-volume-medium::before {
  content: "\f580";
}
.mdi-volume-minus::before {
  content: "\f75d";
}
.mdi-volume-mute::before {
  content: "\f75e";
}
.mdi-volume-off::before {
  content: "\f581";
}
.mdi-volume-plus::before {
  content: "\f75c";
}
.mdi-volume-source::before {
  content: "\f014b";
}
.mdi-volume-variant-off::before {
  content: "\fe68";
}
.mdi-volume-vibrate::before {
  content: "\f014c";
}
.mdi-vote::before {
  content: "\fa1e";
}
.mdi-vote-outline::before {
  content: "\fa1f";
}
.mdi-vpn::before {
  content: "\f582";
}
.mdi-vuejs::before {
  content: "\f843";
}
.mdi-vuetify::before {
  content: "\fe50";
}
.mdi-walk::before {
  content: "\f583";
}
.mdi-wall::before {
  content: "\f7fd";
}
.mdi-wall-sconce::before {
  content: "\f91b";
}
.mdi-wall-sconce-flat::before {
  content: "\f91c";
}
.mdi-wall-sconce-variant::before {
  content: "\f91d";
}
.mdi-wallet::before {
  content: "\f584";
}
.mdi-wallet-giftcard::before {
  content: "\f585";
}
.mdi-wallet-membership::before {
  content: "\f586";
}
.mdi-wallet-outline::before {
  content: "\fbb9";
}
.mdi-wallet-plus::before {
  content: "\ffab";
}
.mdi-wallet-plus-outline::before {
  content: "\ffac";
}
.mdi-wallet-travel::before {
  content: "\f587";
}
.mdi-wallpaper::before {
  content: "\fe69";
}
.mdi-wan::before {
  content: "\f588";
}
.mdi-wardrobe::before {
  content: "\ffad";
}
.mdi-wardrobe-outline::before {
  content: "\ffae";
}
.mdi-warehouse::before {
  content: "\ffbb";
}
.mdi-washing-machine::before {
  content: "\f729";
}
.mdi-washing-machine-alert::before {
  content: "\f01e7";
}
.mdi-washing-machine-off::before {
  content: "\f01e8";
}
.mdi-watch::before {
  content: "\f589";
}
.mdi-watch-export::before {
  content: "\f58a";
}
.mdi-watch-export-variant::before {
  content: "\f894";
}
.mdi-watch-import::before {
  content: "\f58b";
}
.mdi-watch-import-variant::before {
  content: "\f895";
}
.mdi-watch-variant::before {
  content: "\f896";
}
.mdi-watch-vibrate::before {
  content: "\f6b0";
}
.mdi-watch-vibrate-off::before {
  content: "\fcb6";
}
.mdi-water::before {
  content: "\f58c";
}
.mdi-water-boiler::before {
  content: "\ffaf";
}
.mdi-water-boiler-alert::before {
  content: "\f01de";
}
.mdi-water-boiler-off::before {
  content: "\f01df";
}
.mdi-water-off::before {
  content: "\f58d";
}
.mdi-water-outline::before {
  content: "\fe6a";
}
.mdi-water-percent::before {
  content: "\f58e";
}
.mdi-water-polo::before {
  content: "\f02cb";
}
.mdi-water-pump::before {
  content: "\f58f";
}
.mdi-water-pump-off::before {
  content: "\ffb0";
}
.mdi-water-well::before {
  content: "\f008d";
}
.mdi-water-well-outline::before {
  content: "\f008e";
}
.mdi-watermark::before {
  content: "\f612";
}
.mdi-wave::before {
  content: "\ff4b";
}
.mdi-waves::before {
  content: "\f78c";
}
.mdi-waze::before {
  content: "\fbba";
}
.mdi-weather-cloudy::before {
  content: "\f590";
}
.mdi-weather-cloudy-alert::before {
  content: "\ff4c";
}
.mdi-weather-cloudy-arrow-right::before {
  content: "\fe51";
}
.mdi-weather-fog::before {
  content: "\f591";
}
.mdi-weather-hail::before {
  content: "\f592";
}
.mdi-weather-hazy::before {
  content: "\ff4d";
}
.mdi-weather-hurricane::before {
  content: "\f897";
}
.mdi-weather-lightning::before {
  content: "\f593";
}
.mdi-weather-lightning-rainy::before {
  content: "\f67d";
}
.mdi-weather-night::before {
  content: "\f594";
}
.mdi-weather-night-partly-cloudy::before {
  content: "\ff4e";
}
.mdi-weather-partly-cloudy::before {
  content: "\f595";
}
.mdi-weather-partly-lightning::before {
  content: "\ff4f";
}
.mdi-weather-partly-rainy::before {
  content: "\ff50";
}
.mdi-weather-partly-snowy::before {
  content: "\ff51";
}
.mdi-weather-partly-snowy-rainy::before {
  content: "\ff52";
}
.mdi-weather-pouring::before {
  content: "\f596";
}
.mdi-weather-rainy::before {
  content: "\f597";
}
.mdi-weather-snowy::before {
  content: "\f598";
}
.mdi-weather-snowy-heavy::before {
  content: "\ff53";
}
.mdi-weather-snowy-rainy::before {
  content: "\f67e";
}
.mdi-weather-sunny::before {
  content: "\f599";
}
.mdi-weather-sunny-alert::before {
  content: "\ff54";
}
.mdi-weather-sunset::before {
  content: "\f59a";
}
.mdi-weather-sunset-down::before {
  content: "\f59b";
}
.mdi-weather-sunset-up::before {
  content: "\f59c";
}
.mdi-weather-tornado::before {
  content: "\ff55";
}
.mdi-weather-windy::before {
  content: "\f59d";
}
.mdi-weather-windy-variant::before {
  content: "\f59e";
}
.mdi-web::before {
  content: "\f59f";
}
.mdi-web-box::before {
  content: "\ffb1";
}
.mdi-web-clock::before {
  content: "\f0275";
}
.mdi-webcam::before {
  content: "\f5a0";
}
.mdi-webhook::before {
  content: "\f62f";
}
.mdi-webpack::before {
  content: "\f72a";
}
.mdi-webrtc::before {
  content: "\f0273";
}
.mdi-wechat::before {
  content: "\f611";
}
.mdi-weight::before {
  content: "\f5a1";
}
.mdi-weight-gram::before {
  content: "\fd1b";
}
.mdi-weight-kilogram::before {
  content: "\f5a2";
}
.mdi-weight-lifter::before {
  content: "\f0188";
}
.mdi-weight-pound::before {
  content: "\f9b4";
}
.mdi-whatsapp::before {
  content: "\f5a3";
}
.mdi-wheelchair-accessibility::before {
  content: "\f5a4";
}
.mdi-whistle::before {
  content: "\f9b5";
}
.mdi-whistle-outline::before {
  content: "\f02e7";
}
.mdi-white-balance-auto::before {
  content: "\f5a5";
}
.mdi-white-balance-incandescent::before {
  content: "\f5a6";
}
.mdi-white-balance-iridescent::before {
  content: "\f5a7";
}
.mdi-white-balance-sunny::before {
  content: "\f5a8";
}
.mdi-widgets::before {
  content: "\f72b";
}
.mdi-widgets-outline::before {
  content: "\f0380";
}
.mdi-wifi::before {
  content: "\f5a9";
}
.mdi-wifi-off::before {
  content: "\f5aa";
}
.mdi-wifi-star::before {
  content: "\fe6b";
}
.mdi-wifi-strength-1::before {
  content: "\f91e";
}
.mdi-wifi-strength-1-alert::before {
  content: "\f91f";
}
.mdi-wifi-strength-1-lock::before {
  content: "\f920";
}
.mdi-wifi-strength-2::before {
  content: "\f921";
}
.mdi-wifi-strength-2-alert::before {
  content: "\f922";
}
.mdi-wifi-strength-2-lock::before {
  content: "\f923";
}
.mdi-wifi-strength-3::before {
  content: "\f924";
}
.mdi-wifi-strength-3-alert::before {
  content: "\f925";
}
.mdi-wifi-strength-3-lock::before {
  content: "\f926";
}
.mdi-wifi-strength-4::before {
  content: "\f927";
}
.mdi-wifi-strength-4-alert::before {
  content: "\f928";
}
.mdi-wifi-strength-4-lock::before {
  content: "\f929";
}
.mdi-wifi-strength-alert-outline::before {
  content: "\f92a";
}
.mdi-wifi-strength-lock-outline::before {
  content: "\f92b";
}
.mdi-wifi-strength-off::before {
  content: "\f92c";
}
.mdi-wifi-strength-off-outline::before {
  content: "\f92d";
}
.mdi-wifi-strength-outline::before {
  content: "\f92e";
}
.mdi-wii::before {
  content: "\f5ab";
}
.mdi-wiiu::before {
  content: "\f72c";
}
.mdi-wikipedia::before {
  content: "\f5ac";
}
.mdi-wind-turbine::before {
  content: "\fd81";
}
.mdi-window-close::before {
  content: "\f5ad";
}
.mdi-window-closed::before {
  content: "\f5ae";
}
.mdi-window-closed-variant::before {
  content: "\f0206";
}
.mdi-window-maximize::before {
  content: "\f5af";
}
.mdi-window-minimize::before {
  content: "\f5b0";
}
.mdi-window-open::before {
  content: "\f5b1";
}
.mdi-window-open-variant::before {
  content: "\f0207";
}
.mdi-window-restore::before {
  content: "\f5b2";
}
.mdi-window-shutter::before {
  content: "\f0147";
}
.mdi-window-shutter-alert::before {
  content: "\f0148";
}
.mdi-window-shutter-open::before {
  content: "\f0149";
}
.mdi-windows::before {
  content: "\f5b3";
}
.mdi-windows-classic::before {
  content: "\fa20";
}
.mdi-wiper::before {
  content: "\fae8";
}
.mdi-wiper-wash::before {
  content: "\fd82";
}
.mdi-wordpress::before {
  content: "\f5b4";
}
.mdi-worker::before {
  content: "\f5b5";
}
.mdi-wrap::before {
  content: "\f5b6";
}
.mdi-wrap-disabled::before {
  content: "\fbbb";
}
.mdi-wrench::before {
  content: "\f5b7";
}
.mdi-wrench-outline::before {
  content: "\fbbc";
}
.mdi-wunderlist::before {
  content: "\f5b8";
}
.mdi-xamarin::before {
  content: "\f844";
}
.mdi-xamarin-outline::before {
  content: "\f845";
}
.mdi-xaml::before {
  content: "\f673";
}
.mdi-xbox::before {
  content: "\f5b9";
}
.mdi-xbox-controller::before {
  content: "\f5ba";
}
.mdi-xbox-controller-battery-alert::before {
  content: "\f74a";
}
.mdi-xbox-controller-battery-charging::before {
  content: "\fa21";
}
.mdi-xbox-controller-battery-empty::before {
  content: "\f74b";
}
.mdi-xbox-controller-battery-full::before {
  content: "\f74c";
}
.mdi-xbox-controller-battery-low::before {
  content: "\f74d";
}
.mdi-xbox-controller-battery-medium::before {
  content: "\f74e";
}
.mdi-xbox-controller-battery-unknown::before {
  content: "\f74f";
}
.mdi-xbox-controller-menu::before {
  content: "\fe52";
}
.mdi-xbox-controller-off::before {
  content: "\f5bb";
}
.mdi-xbox-controller-view::before {
  content: "\fe53";
}
.mdi-xda::before {
  content: "\f5bc";
}
.mdi-xing::before {
  content: "\f5bd";
}
.mdi-xing-box::before {
  content: "\f5be";
}
.mdi-xing-circle::before {
  content: "\f5bf";
}
.mdi-xml::before {
  content: "\f5c0";
}
.mdi-xmpp::before {
  content: "\f7fe";
}
.mdi-yahoo::before {
  content: "\fb2a";
}
.mdi-yammer::before {
  content: "\f788";
}
.mdi-yeast::before {
  content: "\f5c1";
}
.mdi-yelp::before {
  content: "\f5c2";
}
.mdi-yin-yang::before {
  content: "\f67f";
}
.mdi-yoga::before {
  content: "\f01a7";
}
.mdi-youtube::before {
  content: "\f5c3";
}
.mdi-youtube-creator-studio::before {
  content: "\f846";
}
.mdi-youtube-gaming::before {
  content: "\f847";
}
.mdi-youtube-subscription::before {
  content: "\fd1c";
}
.mdi-youtube-tv::before {
  content: "\f448";
}
.mdi-z-wave::before {
  content: "\fae9";
}
.mdi-zend::before {
  content: "\faea";
}
.mdi-zigbee::before {
  content: "\fd1d";
}
.mdi-zip-box::before {
  content: "\f5c4";
}
.mdi-zip-box-outline::before {
  content: "\f001b";
}
.mdi-zip-disk::before {
  content: "\fa22";
}
.mdi-zodiac-aquarius::before {
  content: "\fa7c";
}
.mdi-zodiac-aries::before {
  content: "\fa7d";
}
.mdi-zodiac-cancer::before {
  content: "\fa7e";
}
.mdi-zodiac-capricorn::before {
  content: "\fa7f";
}
.mdi-zodiac-gemini::before {
  content: "\fa80";
}
.mdi-zodiac-leo::before {
  content: "\fa81";
}
.mdi-zodiac-libra::before {
  content: "\fa82";
}
.mdi-zodiac-pisces::before {
  content: "\fa83";
}
.mdi-zodiac-sagittarius::before {
  content: "\fa84";
}
.mdi-zodiac-scorpio::before {
  content: "\fa85";
}
.mdi-zodiac-taurus::before {
  content: "\fa86";
}
.mdi-zodiac-virgo::before {
  content: "\fa87";
}
.mdi-blank::before {
  content: "\f68c";
  visibility: hidden;
}
.mdi-18px.mdi-set,
.mdi-18px.mdi:before {
  font-size: 18px;
}
.mdi-24px.mdi-set,
.mdi-24px.mdi:before {
  font-size: 24px;
}
.mdi-36px.mdi-set,
.mdi-36px.mdi:before {
  font-size: 36px;
}
.mdi-48px.mdi-set,
.mdi-48px.mdi:before {
  font-size: 48px;
}
.mdi-dark:before {
  color: rgba(0, 0, 0, 0.54);
}
.mdi-dark.mdi-inactive:before {
  color: rgba(0, 0, 0, 0.26);
}
.mdi-light:before {
  color: #fff;
}
.mdi-light.mdi-inactive:before {
  color: rgba(255, 255, 255, 0.3);
}
.mdi-rotate-45:before {
  -webkit-transform: rotate(45deg);
  -ms-transform: rotate(45deg);
  transform: rotate(45deg);
}
.mdi-rotate-90:before {
  -webkit-transform: rotate(90deg);
  -ms-transform: rotate(90deg);
  transform: rotate(90deg);
}
.mdi-rotate-135:before {
  -webkit-transform: rotate(135deg);
  -ms-transform: rotate(135deg);
  transform: rotate(135deg);
}
.mdi-rotate-180:before {
  -webkit-transform: rotate(180deg);
  -ms-transform: rotate(180deg);
  transform: rotate(180deg);
}
.mdi-rotate-225:before {
  -webkit-transform: rotate(225deg);
  -ms-transform: rotate(225deg);
  transform: rotate(225deg);
}
.mdi-rotate-270:before {
  -webkit-transform: rotate(270deg);
  -ms-transform: rotate(270deg);
  transform: rotate(270deg);
}
.mdi-rotate-315:before {
  -webkit-transform: rotate(315deg);
  -ms-transform: rotate(315deg);
  transform: rotate(315deg);
}
.mdi-flip-h:before {
  -webkit-transform: scaleX(-1);
  transform: scaleX(-1);
  filter: FlipH;
  -ms-filter: "FlipH";
}
.mdi-flip-v:before {
  -webkit-transform: scaleY(-1);
  transform: scaleY(-1);
  filter: FlipV;
  -ms-filter: "FlipV";
}
.mdi-spin:before {
  -webkit-animation: mdi-spin 2s infinite linear;
  animation: mdi-spin 2s infinite linear;
}
@-webkit-keyframes mdi-spin {
  0% {
    -webkit-transform: rotate(0deg);
    transform: rotate(0deg);
  }
  100% {
    -webkit-transform: rotate(359deg);
    transform: rotate(359deg);
  }
}
@keyframes mdi-spin {
  0% {
    -webkit-transform: rotate(0deg);
    transform: rotate(0deg);
  }
  100% {
    -webkit-transform: rotate(359deg);
    transform: rotate(359deg);
  }
}

/* ../../node_modules/.pnpm/@creativebulma+bulma-tooltip@1.2.0/node_modules/@creativebulma/bulma-tooltip/dist/bulma-tooltip.min.css */
[data-tooltip]:not(.is-disabled),
[data-tooltip]:not(.is-loading),
[data-tooltip]:not([disabled]) {
  cursor: pointer;
  overflow: visible;
  position: relative;
}
[data-tooltip]:not(.is-disabled):before,
[data-tooltip]:not(.is-loading):before,
[data-tooltip]:not([disabled]):before {
  background: rgba(74, 74, 74, 0.9);
  border-radius: 2px;
  content: attr(data-tooltip);
  padding: 0.5rem 1rem;
  text-overflow: ellipsis;
  white-space: pre;
  right: auto;
  bottom: auto;
  left: 50%;
  top: 0;
  margin-top: -5px;
  margin-bottom: auto;
  transform: translate(-50%, -100%);
}
[data-tooltip]:not(.is-disabled).has-tooltip-arrow:after,
[data-tooltip]:not(.is-disabled):before,
[data-tooltip]:not(.is-loading).has-tooltip-arrow:after,
[data-tooltip]:not(.is-loading):before,
[data-tooltip]:not([disabled]).has-tooltip-arrow:after,
[data-tooltip]:not([disabled]):before {
  box-sizing: border-box;
  color: #fff;
  display: inline-block;
  font-family:
    BlinkMacSystemFont,
    -apple-system,
    Segoe UI,
    Roboto,
    Oxygen,
    Ubuntu,
    Cantarell,
    Fira Sans,
    Droid Sans,
    Helvetica Neue,
    Helvetica,
    Arial,
    sans-serif;
  font-size: 0.75rem;
  -webkit-hyphens: auto;
  -ms-hyphens: auto;
  hyphens: auto;
  opacity: 0;
  overflow: hidden;
  pointer-events: none;
  position: absolute;
  visibility: hidden;
  z-index: 1;
}
[data-tooltip]:not(.is-disabled).has-tooltip-arrow:after,
[data-tooltip]:not(.is-loading).has-tooltip-arrow:after,
[data-tooltip]:not([disabled]).has-tooltip-arrow:after {
  content: "";
  border-style: solid;
  border-width: 6px;
  border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  margin-bottom: -5px;
}
[data-tooltip]:not(.is-disabled).has-tooltip-arrow.has-tooltip-arrow:after,
[data-tooltip]:not(.is-loading).has-tooltip-arrow.has-tooltip-arrow:after,
[data-tooltip]:not([disabled]).has-tooltip-arrow.has-tooltip-arrow:after {
  top: 0;
  right: auto;
  bottom: auto;
  left: 50%;
  margin: -5px auto auto -5px;
  border-color: rgba(74, 74, 74, 0.9) transparent transparent;
}
[data-tooltip]:not(.is-disabled).has-tooltip-bottom.has-tooltip-arrow:after,
[data-tooltip]:not(.is-loading).has-tooltip-bottom.has-tooltip-arrow:after,
[data-tooltip]:not([disabled]).has-tooltip-bottom.has-tooltip-arrow:after {
  top: auto;
  right: auto;
  bottom: -1px;
  left: 50%;
  margin: auto auto -5px -5px;
  border-color: transparent transparent rgba(74, 74, 74, 0.9);
}
[data-tooltip]:not(.is-disabled).has-tooltip-bottom:before,
[data-tooltip]:not(.is-loading).has-tooltip-bottom:before,
[data-tooltip]:not([disabled]).has-tooltip-bottom:before {
  top: auto;
  right: auto;
  bottom: 0;
  left: 50%;
  margin-top: auto;
  margin-bottom: -5px;
  transform: translate(-50%, 100%);
}
[data-tooltip]:not(.is-disabled).has-tooltip-left.has-tooltip-arrow:after,
[data-tooltip]:not(.is-loading).has-tooltip-left.has-tooltip-arrow:after,
[data-tooltip]:not([disabled]).has-tooltip-left.has-tooltip-arrow:after {
  top: auto;
  right: auto;
  bottom: 50%;
  left: 0;
  margin: auto auto -6px -5px;
  border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
}
[data-tooltip]:not(.is-disabled).has-tooltip-left:before,
[data-tooltip]:not(.is-loading).has-tooltip-left:before,
[data-tooltip]:not([disabled]).has-tooltip-left:before {
  top: auto;
  right: auto;
  bottom: 50%;
  left: -5px;
  transform: translate(-100%, 50%);
}
[data-tooltip]:not(.is-disabled).has-tooltip-right.has-tooltip-arrow:after,
[data-tooltip]:not(.is-loading).has-tooltip-right.has-tooltip-arrow:after,
[data-tooltip]:not([disabled]).has-tooltip-right.has-tooltip-arrow:after {
  top: auto;
  right: 0;
  bottom: 50%;
  left: auto;
  margin: auto -6px -6px auto;
  border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
}
[data-tooltip]:not(.is-disabled).has-tooltip-right:before,
[data-tooltip]:not(.is-loading).has-tooltip-right:before,
[data-tooltip]:not([disabled]).has-tooltip-right:before {
  top: auto;
  right: -5px;
  bottom: 50%;
  left: auto;
  margin-top: auto;
  transform: translate(100%, 50%);
}
[data-tooltip]:not(.is-disabled).has-tooltip-multiline:before,
[data-tooltip]:not(.is-loading).has-tooltip-multiline:before,
[data-tooltip]:not([disabled]).has-tooltip-multiline:before {
  height: auto;
  width: 15rem;
  max-width: 15rem;
  text-overflow: clip;
  white-space: normal;
  word-break: keep-all;
}
[data-tooltip]:not(.is-disabled).has-tooltip-text-left:before,
[data-tooltip]:not(.is-loading).has-tooltip-text-left:before,
[data-tooltip]:not([disabled]).has-tooltip-text-left:before {
  text-align: left;
}
[data-tooltip]:not(.is-disabled).has-tooltip-text-centered:before,
[data-tooltip]:not(.is-loading).has-tooltip-text-centered:before,
[data-tooltip]:not([disabled]).has-tooltip-text-centered:before {
  text-align: center;
}
[data-tooltip]:not(.is-disabled).has-tooltip-text-right:before,
[data-tooltip]:not(.is-loading).has-tooltip-text-right:before,
[data-tooltip]:not([disabled]).has-tooltip-text-right:before {
  text-align: right;
}
[data-tooltip]:not(.is-disabled).has-tooltip-white:after,
[data-tooltip]:not(.is-loading).has-tooltip-white:after,
[data-tooltip]:not([disabled]).has-tooltip-white:after {
  border-color: hsla(0, 0%, 100%, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-white.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-white.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-white.has-tooltip-bottom:after {
  border-color: transparent transparent hsla(0, 0%, 100%, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-white.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-white.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-white.has-tooltip-left:after {
  border-color: transparent transparent transparent hsla(0, 0%, 100%, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-white.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-white.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-white.has-tooltip-right:after {
  border-color: transparent hsla(0, 0%, 100%, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-white:before,
[data-tooltip]:not(.is-loading).has-tooltip-white:before,
[data-tooltip]:not([disabled]).has-tooltip-white:before {
  background-color: hsla(0, 0%, 100%, 0.9);
  color: #0a0a0a;
}
[data-tooltip]:not(.is-disabled).has-tooltip-black:after,
[data-tooltip]:not(.is-loading).has-tooltip-black:after,
[data-tooltip]:not([disabled]).has-tooltip-black:after {
  border-color: hsla(0, 0%, 4%, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-black.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-black.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-black.has-tooltip-bottom:after {
  border-color: transparent transparent hsla(0, 0%, 4%, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-black.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-black.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-black.has-tooltip-left:after {
  border-color: transparent transparent transparent hsla(0, 0%, 4%, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-black.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-black.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-black.has-tooltip-right:after {
  border-color: transparent hsla(0, 0%, 4%, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-black:before,
[data-tooltip]:not(.is-loading).has-tooltip-black:before,
[data-tooltip]:not([disabled]).has-tooltip-black:before {
  background-color: hsla(0, 0%, 4%, 0.9);
  color: #fff;
}
[data-tooltip]:not(.is-disabled).has-tooltip-light:after,
[data-tooltip]:not(.is-loading).has-tooltip-light:after,
[data-tooltip]:not([disabled]).has-tooltip-light:after {
  border-color: hsla(0, 0%, 96%, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-light.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-light.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-light.has-tooltip-bottom:after {
  border-color: transparent transparent hsla(0, 0%, 96%, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-light.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-light.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-light.has-tooltip-left:after {
  border-color: transparent transparent transparent hsla(0, 0%, 96%, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-light.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-light.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-light.has-tooltip-right:after {
  border-color: transparent hsla(0, 0%, 96%, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-light:before,
[data-tooltip]:not(.is-loading).has-tooltip-light:before,
[data-tooltip]:not([disabled]).has-tooltip-light:before {
  background-color: hsla(0, 0%, 96%, 0.9);
  color: rgba(0, 0, 0, 0.7);
}
[data-tooltip]:not(.is-disabled).has-tooltip-dark:after,
[data-tooltip]:not(.is-loading).has-tooltip-dark:after,
[data-tooltip]:not([disabled]).has-tooltip-dark:after {
  border-color: rgba(54, 54, 54, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-dark.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-dark.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-dark.has-tooltip-bottom:after {
  border-color: transparent transparent rgba(54, 54, 54, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-dark.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-dark.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-dark.has-tooltip-left:after {
  border-color: transparent transparent transparent rgba(54, 54, 54, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-dark.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-dark.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-dark.has-tooltip-right:after {
  border-color: transparent rgba(54, 54, 54, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-dark:before,
[data-tooltip]:not(.is-loading).has-tooltip-dark:before,
[data-tooltip]:not([disabled]).has-tooltip-dark:before {
  background-color: rgba(54, 54, 54, 0.9);
  color: #fff;
}
[data-tooltip]:not(.is-disabled).has-tooltip-primary:after,
[data-tooltip]:not(.is-loading).has-tooltip-primary:after,
[data-tooltip]:not([disabled]).has-tooltip-primary:after {
  border-color: rgba(0, 209, 178, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-primary.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-primary.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-primary.has-tooltip-bottom:after {
  border-color: transparent transparent rgba(0, 209, 178, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-primary.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-primary.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-primary.has-tooltip-left:after {
  border-color: transparent transparent transparent rgba(0, 209, 178, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-primary.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-primary.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-primary.has-tooltip-right:after {
  border-color: transparent rgba(0, 209, 178, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-primary:before,
[data-tooltip]:not(.is-loading).has-tooltip-primary:before,
[data-tooltip]:not([disabled]).has-tooltip-primary:before {
  background-color: rgba(0, 209, 178, 0.9);
  color: #fff;
}
[data-tooltip]:not(.is-disabled).has-tooltip-link:after,
[data-tooltip]:not(.is-loading).has-tooltip-link:after,
[data-tooltip]:not([disabled]).has-tooltip-link:after {
  border-color: rgba(50, 115, 220, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-link.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-link.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-link.has-tooltip-bottom:after {
  border-color: transparent transparent rgba(50, 115, 220, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-link.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-link.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-link.has-tooltip-left:after {
  border-color: transparent transparent transparent rgba(50, 115, 220, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-link.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-link.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-link.has-tooltip-right:after {
  border-color: transparent rgba(50, 115, 220, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-link:before,
[data-tooltip]:not(.is-loading).has-tooltip-link:before,
[data-tooltip]:not([disabled]).has-tooltip-link:before {
  background-color: rgba(50, 115, 220, 0.9);
  color: #fff;
}
[data-tooltip]:not(.is-disabled).has-tooltip-info:after,
[data-tooltip]:not(.is-loading).has-tooltip-info:after,
[data-tooltip]:not([disabled]).has-tooltip-info:after {
  border-color: rgba(50, 152, 220, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-info.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-info.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-info.has-tooltip-bottom:after {
  border-color: transparent transparent rgba(50, 152, 220, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-info.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-info.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-info.has-tooltip-left:after {
  border-color: transparent transparent transparent rgba(50, 152, 220, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-info.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-info.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-info.has-tooltip-right:after {
  border-color: transparent rgba(50, 152, 220, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-info:before,
[data-tooltip]:not(.is-loading).has-tooltip-info:before,
[data-tooltip]:not([disabled]).has-tooltip-info:before {
  background-color: rgba(50, 152, 220, 0.9);
  color: #fff;
}
[data-tooltip]:not(.is-disabled).has-tooltip-success:after,
[data-tooltip]:not(.is-loading).has-tooltip-success:after,
[data-tooltip]:not([disabled]).has-tooltip-success:after {
  border-color: rgba(72, 199, 116, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-success.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-success.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-success.has-tooltip-bottom:after {
  border-color: transparent transparent rgba(72, 199, 116, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-success.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-success.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-success.has-tooltip-left:after {
  border-color: transparent transparent transparent rgba(72, 199, 116, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-success.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-success.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-success.has-tooltip-right:after {
  border-color: transparent rgba(72, 199, 116, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-success:before,
[data-tooltip]:not(.is-loading).has-tooltip-success:before,
[data-tooltip]:not([disabled]).has-tooltip-success:before {
  background-color: rgba(72, 199, 116, 0.9);
  color: #fff;
}
[data-tooltip]:not(.is-disabled).has-tooltip-warning:after,
[data-tooltip]:not(.is-loading).has-tooltip-warning:after,
[data-tooltip]:not([disabled]).has-tooltip-warning:after {
  border-color: rgba(255, 221, 87, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-warning.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-warning.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-warning.has-tooltip-bottom:after {
  border-color: transparent transparent rgba(255, 221, 87, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-warning.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-warning.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-warning.has-tooltip-left:after {
  border-color: transparent transparent transparent rgba(255, 221, 87, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-warning.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-warning.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-warning.has-tooltip-right:after {
  border-color: transparent rgba(255, 221, 87, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-warning:before,
[data-tooltip]:not(.is-loading).has-tooltip-warning:before,
[data-tooltip]:not([disabled]).has-tooltip-warning:before {
  background-color: rgba(255, 221, 87, 0.9);
  color: rgba(0, 0, 0, 0.7);
}
[data-tooltip]:not(.is-disabled).has-tooltip-danger:after,
[data-tooltip]:not(.is-loading).has-tooltip-danger:after,
[data-tooltip]:not([disabled]).has-tooltip-danger:after {
  border-color: rgba(241, 70, 104, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-danger.has-tooltip-bottom:after,
[data-tooltip]:not(.is-loading).has-tooltip-danger.has-tooltip-bottom:after,
[data-tooltip]:not([disabled]).has-tooltip-danger.has-tooltip-bottom:after {
  border-color: transparent transparent rgba(241, 70, 104, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-danger.has-tooltip-left:after,
[data-tooltip]:not(.is-loading).has-tooltip-danger.has-tooltip-left:after,
[data-tooltip]:not([disabled]).has-tooltip-danger.has-tooltip-left:after {
  border-color: transparent transparent transparent rgba(241, 70, 104, 0.9) !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-danger.has-tooltip-right:after,
[data-tooltip]:not(.is-loading).has-tooltip-danger.has-tooltip-right:after,
[data-tooltip]:not([disabled]).has-tooltip-danger.has-tooltip-right:after {
  border-color: transparent rgba(241, 70, 104, 0.9) transparent transparent !important;
}
[data-tooltip]:not(.is-disabled).has-tooltip-danger:before,
[data-tooltip]:not(.is-loading).has-tooltip-danger:before,
[data-tooltip]:not([disabled]).has-tooltip-danger:before {
  background-color: rgba(241, 70, 104, 0.9);
  color: #fff;
}
[data-tooltip]:not(.is-disabled).has-tooltip-active:after,
[data-tooltip]:not(.is-disabled).has-tooltip-active:before,
[data-tooltip]:not(.is-disabled):hover:after,
[data-tooltip]:not(.is-disabled):hover:before,
[data-tooltip]:not(.is-loading).has-tooltip-active:after,
[data-tooltip]:not(.is-loading).has-tooltip-active:before,
[data-tooltip]:not(.is-loading):hover:after,
[data-tooltip]:not(.is-loading):hover:before,
[data-tooltip]:not([disabled]).has-tooltip-active:after,
[data-tooltip]:not([disabled]).has-tooltip-active:before,
[data-tooltip]:not([disabled]):hover:after,
[data-tooltip]:not([disabled]):hover:before {
  opacity: 1;
  visibility: visible;
}
[data-tooltip]:not(.is-disabled).has-tooltip-fade:after,
[data-tooltip]:not(.is-disabled).has-tooltip-fade:before,
[data-tooltip]:not(.is-loading).has-tooltip-fade:after,
[data-tooltip]:not(.is-loading).has-tooltip-fade:before,
[data-tooltip]:not([disabled]).has-tooltip-fade:after,
[data-tooltip]:not([disabled]).has-tooltip-fade:before {
  transition: opacity 0.3s linear, visibility 0.3s linear;
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-mobile.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-mobile:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-tablet.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-tablet:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-tablet-only.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-tablet-only:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-touch.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-touch.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-touch.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-touch:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-desktop.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-desktop:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-desktop-only.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-desktop-only:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-until-widescreen.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-until-widescreen:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-widescreen.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-widescreen:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-widescreen-only.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-widescreen-only:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-until-fullhd.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-until-fullhd:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-top-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-top-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-top-fullhd.has-tooltip-arrow:after {
    top: 0;
    right: auto;
    bottom: auto;
    left: 50%;
    margin: -5px auto auto -5px;
    border-color: rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-top-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-top-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-top-fullhd:before {
    right: auto;
    bottom: auto;
    left: 50%;
    top: 0;
    margin-top: -5px;
    margin-bottom: auto;
    transform: translate(-50%, -100%);
  }
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-mobile.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-mobile:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-tablet.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-tablet:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-tablet-only.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-tablet-only:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-touch.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-touch.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-touch.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-touch:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-desktop.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-desktop:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-desktop-only.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-desktop-only:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-until-widescreen.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-until-widescreen:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-widescreen.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-widescreen:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-widescreen-only.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-widescreen-only:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-until-fullhd.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-until-fullhd:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-right-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-right-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-right-fullhd.has-tooltip-arrow:after {
    top: auto;
    right: 0;
    bottom: 50%;
    left: auto;
    margin: auto -6px -6px auto;
    border-color: transparent rgba(74, 74, 74, 0.9) transparent transparent;
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-right-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-right-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-right-fullhd:before {
    top: auto;
    right: -5px;
    bottom: 50%;
    left: auto;
    margin-top: auto;
    transform: translate(100%, 50%);
  }
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-mobile.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-mobile:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-tablet.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-tablet:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-tablet-only.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-tablet-only:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-touch.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-touch.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-touch.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-touch:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-desktop.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-desktop:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-desktop-only.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-desktop-only:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-until-widescreen.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-until-widescreen:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-widescreen.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-widescreen:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-widescreen-only.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-widescreen-only:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-until-fullhd.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-until-fullhd:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-fullhd.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: -1px;
    left: 50%;
    margin: auto auto -5px -5px;
    border-color: transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-bottom-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-bottom-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-bottom-fullhd:before {
    top: auto;
    right: auto;
    bottom: 0;
    left: 50%;
    margin-top: auto;
    margin-bottom: -5px;
    transform: translate(-50%, 100%);
  }
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-mobile.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-mobile.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-mobile:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-tablet.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-tablet.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-tablet:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-tablet-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-tablet-only.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-tablet-only:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-touch.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-touch.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-touch.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-touch:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-desktop.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-desktop.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-desktop:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-desktop-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-desktop-only.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-desktop-only:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-until-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-until-widescreen.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-until-widescreen:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-widescreen.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-widescreen.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-widescreen:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-widescreen-only.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-widescreen-only.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-widescreen-only:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-until-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-until-fullhd.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-until-fullhd:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-left-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not(.is-loading).has-tooltip-left-fullhd.has-tooltip-arrow:after,
  [data-tooltip]:not([disabled]).has-tooltip-left-fullhd.has-tooltip-arrow:after {
    top: auto;
    right: auto;
    bottom: 50%;
    left: 0;
    margin: auto auto -6px -5px;
    border-color: transparent transparent transparent rgba(74, 74, 74, 0.9);
  }
  [data-tooltip]:not(.is-disabled).has-tooltip-left-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-left-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-left-fullhd:before {
    top: auto;
    right: auto;
    bottom: 50%;
    left: -5px;
    transform: translate(-100%, 50%);
  }
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-mobile:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-mobile:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-mobile:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-mobile:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-tablet:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-tablet:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-tablet:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-tablet:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-tablet-only:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-tablet-only:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-tablet-only:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-tablet-only:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-touch:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-touch:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-touch:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-touch:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-desktop:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-desktop:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-desktop:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-desktop:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-desktop-only:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-desktop-only:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-desktop-only:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-desktop-only:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-until-widescreen:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-until-widescreen:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-until-widescreen:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-until-widescreen:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-widescreen:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-widescreen:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-widescreen:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-widescreen:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-widescreen-only:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-widescreen-only:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-widescreen-only:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-widescreen-only:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-until-fullhd:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-until-fullhd:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-until-fullhd:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-until-fullhd:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-fullhd:after,
  [data-tooltip]:not(.is-disabled).has-tooltip-hidden-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-fullhd:after,
  [data-tooltip]:not(.is-loading).has-tooltip-hidden-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-fullhd:after,
  [data-tooltip]:not([disabled]).has-tooltip-hidden-fullhd:before {
    opacity: 0 !important;
    display: none !important;
  }
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-mobile:before {
    text-align: left;
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-tablet:before {
    text-align: left;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-tablet-only:before {
    text-align: left;
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-touch:before {
    text-align: left;
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-desktop:before {
    text-align: left;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-desktop-only:before {
    text-align: left;
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-until-widescreen:before {
    text-align: left;
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-widescreen:before {
    text-align: left;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-widescreen-only:before {
    text-align: left;
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-until-fullhd:before {
    text-align: left;
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-left-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-left-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-left-fullhd:before {
    text-align: left;
  }
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-mobile:before {
    text-align: center;
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-tablet:before {
    text-align: center;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-tablet-only:before {
    text-align: center;
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-touch:before {
    text-align: center;
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-desktop:before {
    text-align: center;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-desktop-only:before {
    text-align: center;
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-until-widescreen:before {
    text-align: center;
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-widescreen:before {
    text-align: center;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-widescreen-only:before {
    text-align: center;
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-until-fullhd:before {
    text-align: center;
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-centered-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-centered-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-centered-fullhd:before {
    text-align: center;
  }
}
@media screen and (max-width: 768px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-mobile:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-mobile:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-mobile:before {
    text-align: right;
  }
}
@media print, screen and (min-width: 769px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-tablet:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-tablet:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-tablet:before {
    text-align: right;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-tablet-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-tablet-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-tablet-only:before {
    text-align: right;
  }
}
@media screen and (max-width: 1023px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-touch:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-touch:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-touch:before {
    text-align: right;
  }
}
@media screen and (min-width: 1024px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-desktop:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-desktop:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-desktop:before {
    text-align: right;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-desktop-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-desktop-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-desktop-only:before {
    text-align: right;
  }
}
@media screen and (max-width: 1215px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-until-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-until-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-until-widescreen:before {
    text-align: right;
  }
}
@media screen and (min-width: 1216px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-widescreen:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-widescreen:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-widescreen:before {
    text-align: right;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-widescreen-only:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-widescreen-only:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-widescreen-only:before {
    text-align: right;
  }
}
@media screen and (max-width: 1407px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-until-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-until-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-until-fullhd:before {
    text-align: right;
  }
}
@media screen and (min-width: 1408px) {
  [data-tooltip]:not(.is-disabled).has-tooltip-text-right-fullhd:before,
  [data-tooltip]:not(.is-loading).has-tooltip-text-right-fullhd:before,
  [data-tooltip]:not([disabled]).has-tooltip-text-right-fullhd:before {
    text-align: right;
  }
}
span[data-tooltip] {
  border-bottom: 1px dashed #dbdbdb;
}
span[data-tooltip].has-tooltip-white {
  border-bottom-color: #fff;
}
span[data-tooltip].has-tooltip-black {
  border-bottom-color: #171717;
}
span[data-tooltip].has-tooltip-light {
  border-bottom-color: #fff;
}
span[data-tooltip].has-tooltip-dark {
  border-bottom-color: #424242;
}
span[data-tooltip].has-tooltip-primary {
  border-bottom-color: #00ebc7;
}
span[data-tooltip].has-tooltip-link {
  border-bottom-color: #4882e0;
}
span[data-tooltip].has-tooltip-info {
  border-bottom-color: #48a3e0;
}
span[data-tooltip].has-tooltip-success {
  border-bottom-color: #5bcd83;
}
span[data-tooltip].has-tooltip-warning {
  border-bottom-color: #ffe270;
}
span[data-tooltip].has-tooltip-danger {
  border-bottom-color: #f35e7c;
}
.control span[data-tooltip] {
  border-bottom: none;
}

/* ../../node_modules/.pnpm/bulma-timeline@3.0.5/node_modules/bulma-timeline/dist/css/bulma-timeline.min.css */
.timeline {
  display: flex;
  flex-direction: column;
}
.timeline .timeline-header {
  width: 4em;
  min-width: 4em;
  max-width: 8em;
  word-wrap: normal;
  text-align: center;
  display: flex;
  justify-content: center;
}
.timeline .timeline-item {
  display: flex;
  display: -webkit-flex;
  position: relative;
  margin-left: 2em;
  padding-bottom: 2em;
}
.timeline .timeline-item::before {
  content: "";
  background-color: #dbdbdb;
  display: block;
  width: 1px;
  height: 100%;
  position: absolute;
  left: -0.5px;
  top: 0;
}
.timeline .timeline-item .timeline-marker {
  position: absolute;
  background: #dbdbdb;
  border: 1px solid #dbdbdb;
  border-radius: 100%;
  content: "";
  display: block;
  height: 12px;
  transform: translateX(-50%);
  top: 1.2rem;
  width: 12px;
}
.timeline .timeline-item .timeline-marker.is-image {
  background: #dbdbdb;
  border: 1px solid #dbdbdb;
  border-radius: 100%;
  display: block;
  overflow: hidden;
}
.timeline .timeline-item .timeline-marker.is-image.is-16x16 {
  height: 16px;
  width: 16px;
}
.timeline .timeline-item .timeline-marker.is-image.is-24x24 {
  height: 24px;
  width: 24px;
}
.timeline .timeline-item .timeline-marker.is-image.is-32x32 {
  height: 32px;
  width: 32px;
}
.timeline .timeline-item .timeline-marker.is-image.is-48x48 {
  height: 48px;
  width: 48px;
}
.timeline .timeline-item .timeline-marker.is-image.is-64x64 {
  height: 64px;
  width: 64px;
}
.timeline .timeline-item .timeline-marker.is-image.is-96x96 {
  height: 96px;
  width: 96px;
}
.timeline .timeline-item .timeline-marker.is-image.is-128x128 {
  height: 128px;
  width: 128px;
}
.timeline .timeline-item .timeline-marker.is-icon {
  display: flex;
  align-items: center;
  justify-content: center;
  height: 24px;
  width: 24px;
  background: #dbdbdb;
  border: 1px solid #dbdbdb;
  border-radius: 100%;
  padding: 0.5rem;
}
.timeline .timeline-item .timeline-marker.is-icon > * {
  font-size: 0.75rem !important;
}
.timeline .timeline-item .timeline-marker.is-outlined .image {
  background: #fff;
}
.timeline .timeline-item .timeline-marker.is-outlined.is-icon {
  background: #fff;
}
.timeline .timeline-item .timeline-marker.is-outlined.is-icon > * {
  color: #dbdbdb;
}
.timeline .timeline-item .timeline-marker.is-white {
  background-color: #fff !important;
  border-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-white .image {
  border-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-white.is-icon {
  background-color: #fff !important;
  border-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-white.is-icon > * {
  color: #0a0a0a !important;
}
.timeline .timeline-item .timeline-marker.is-white.is-outlined {
  background-color: #fff !important;
  border-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-white.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-white.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-white.is-outlined.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-black {
  background-color: #0a0a0a !important;
  border-color: #0a0a0a !important;
}
.timeline .timeline-item .timeline-marker.is-black .image {
  border-color: #0a0a0a !important;
}
.timeline .timeline-item .timeline-marker.is-black.is-icon {
  background-color: #0a0a0a !important;
  border-color: #0a0a0a !important;
}
.timeline .timeline-item .timeline-marker.is-black.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-black.is-outlined {
  background-color: #fff !important;
  border-color: #0a0a0a !important;
}
.timeline .timeline-item .timeline-marker.is-black.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-black.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-black.is-outlined.is-icon > * {
  color: #0a0a0a !important;
}
.timeline .timeline-item .timeline-marker.is-light {
  background-color: #f5f5f5 !important;
  border-color: #f5f5f5 !important;
}
.timeline .timeline-item .timeline-marker.is-light .image {
  border-color: #f5f5f5 !important;
}
.timeline .timeline-item .timeline-marker.is-light.is-icon {
  background-color: #f5f5f5 !important;
  border-color: #f5f5f5 !important;
}
.timeline .timeline-item .timeline-marker.is-light.is-icon > * {
  color: rgba(0, 0, 0, 0.7) !important;
}
.timeline .timeline-item .timeline-marker.is-light.is-outlined {
  background-color: #fff !important;
  border-color: #f5f5f5 !important;
}
.timeline .timeline-item .timeline-marker.is-light.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-light.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-light.is-outlined.is-icon > * {
  color: #f5f5f5 !important;
}
.timeline .timeline-item .timeline-marker.is-dark {
  background-color: #363636 !important;
  border-color: #363636 !important;
}
.timeline .timeline-item .timeline-marker.is-dark .image {
  border-color: #363636 !important;
}
.timeline .timeline-item .timeline-marker.is-dark.is-icon {
  background-color: #363636 !important;
  border-color: #363636 !important;
}
.timeline .timeline-item .timeline-marker.is-dark.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-dark.is-outlined {
  background-color: #fff !important;
  border-color: #363636 !important;
}
.timeline .timeline-item .timeline-marker.is-dark.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-dark.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-dark.is-outlined.is-icon > * {
  color: #363636 !important;
}
.timeline .timeline-item .timeline-marker.is-primary {
  background-color: #00d1b2 !important;
  border-color: #00d1b2 !important;
}
.timeline .timeline-item .timeline-marker.is-primary .image {
  border-color: #00d1b2 !important;
}
.timeline .timeline-item .timeline-marker.is-primary.is-icon {
  background-color: #00d1b2 !important;
  border-color: #00d1b2 !important;
}
.timeline .timeline-item .timeline-marker.is-primary.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-primary.is-outlined {
  background-color: #fff !important;
  border-color: #00d1b2 !important;
}
.timeline .timeline-item .timeline-marker.is-primary.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-primary.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-primary.is-outlined.is-icon > * {
  color: #00d1b2 !important;
}
.timeline .timeline-item .timeline-marker.is-link {
  background-color: #485fc7 !important;
  border-color: #485fc7 !important;
}
.timeline .timeline-item .timeline-marker.is-link .image {
  border-color: #485fc7 !important;
}
.timeline .timeline-item .timeline-marker.is-link.is-icon {
  background-color: #485fc7 !important;
  border-color: #485fc7 !important;
}
.timeline .timeline-item .timeline-marker.is-link.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-link.is-outlined {
  background-color: #fff !important;
  border-color: #485fc7 !important;
}
.timeline .timeline-item .timeline-marker.is-link.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-link.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-link.is-outlined.is-icon > * {
  color: #485fc7 !important;
}
.timeline .timeline-item .timeline-marker.is-info {
  background-color: #3e8ed0 !important;
  border-color: #3e8ed0 !important;
}
.timeline .timeline-item .timeline-marker.is-info .image {
  border-color: #3e8ed0 !important;
}
.timeline .timeline-item .timeline-marker.is-info.is-icon {
  background-color: #3e8ed0 !important;
  border-color: #3e8ed0 !important;
}
.timeline .timeline-item .timeline-marker.is-info.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-info.is-outlined {
  background-color: #fff !important;
  border-color: #3e8ed0 !important;
}
.timeline .timeline-item .timeline-marker.is-info.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-info.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-info.is-outlined.is-icon > * {
  color: #3e8ed0 !important;
}
.timeline .timeline-item .timeline-marker.is-success {
  background-color: #48c78e !important;
  border-color: #48c78e !important;
}
.timeline .timeline-item .timeline-marker.is-success .image {
  border-color: #48c78e !important;
}
.timeline .timeline-item .timeline-marker.is-success.is-icon {
  background-color: #48c78e !important;
  border-color: #48c78e !important;
}
.timeline .timeline-item .timeline-marker.is-success.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-success.is-outlined {
  background-color: #fff !important;
  border-color: #48c78e !important;
}
.timeline .timeline-item .timeline-marker.is-success.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-success.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-success.is-outlined.is-icon > * {
  color: #48c78e !important;
}
.timeline .timeline-item .timeline-marker.is-warning {
  background-color: #ffe08a !important;
  border-color: #ffe08a !important;
}
.timeline .timeline-item .timeline-marker.is-warning .image {
  border-color: #ffe08a !important;
}
.timeline .timeline-item .timeline-marker.is-warning.is-icon {
  background-color: #ffe08a !important;
  border-color: #ffe08a !important;
}
.timeline .timeline-item .timeline-marker.is-warning.is-icon > * {
  color: rgba(0, 0, 0, 0.7) !important;
}
.timeline .timeline-item .timeline-marker.is-warning.is-outlined {
  background-color: #fff !important;
  border-color: #ffe08a !important;
}
.timeline .timeline-item .timeline-marker.is-warning.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-warning.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-warning.is-outlined.is-icon > * {
  color: #ffe08a !important;
}
.timeline .timeline-item .timeline-marker.is-danger {
  background-color: #f14668 !important;
  border-color: #f14668 !important;
}
.timeline .timeline-item .timeline-marker.is-danger .image {
  border-color: #f14668 !important;
}
.timeline .timeline-item .timeline-marker.is-danger.is-icon {
  background-color: #f14668 !important;
  border-color: #f14668 !important;
}
.timeline .timeline-item .timeline-marker.is-danger.is-icon > * {
  color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-danger.is-outlined {
  background-color: #fff !important;
  border-color: #f14668 !important;
}
.timeline .timeline-item .timeline-marker.is-danger.is-outlined .image {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-danger.is-outlined.is-icon {
  background-color: #fff !important;
}
.timeline .timeline-item .timeline-marker.is-danger.is-outlined.is-icon > * {
  color: #f14668 !important;
}
.timeline .timeline-item .timeline-content {
  padding: 1em 0 0 2em;
}
.timeline .timeline-item .timeline-content .heading {
  font-weight: 600;
}
.timeline .timeline-item.is-white::before {
  background-color: #fff;
}
.timeline .timeline-item.is-black::before {
  background-color: #0a0a0a;
}
.timeline .timeline-item.is-light::before {
  background-color: #f5f5f5;
}
.timeline .timeline-item.is-dark::before {
  background-color: #363636;
}
.timeline .timeline-item.is-primary::before {
  background-color: #00d1b2;
}
.timeline .timeline-item.is-link::before {
  background-color: #485fc7;
}
.timeline .timeline-item.is-info::before {
  background-color: #3e8ed0;
}
.timeline .timeline-item.is-success::before {
  background-color: #48c78e;
}
.timeline .timeline-item.is-warning::before {
  background-color: #ffe08a;
}
.timeline .timeline-item.is-danger::before {
  background-color: #f14668;
}
.timeline.is-centered .timeline-header {
  display: flex;
  width: 100%;
  align-self: center;
}
.timeline.is-centered .timeline-item {
  width: 50%;
  align-self: flex-end;
  flex-direction: row;
}
.timeline.is-centered .timeline-item:nth-of-type(2n) {
  align-self: flex-start;
  flex-direction: row-reverse;
  margin-left: 0;
  margin-right: 2em;
}
.timeline.is-centered .timeline-item:nth-of-type(2n)::before {
  right: -0.5px;
  left: auto;
}
.timeline.is-centered .timeline-item:nth-of-type(2n) .timeline-marker {
  transform: translateX(50%);
}
.timeline.is-centered .timeline-item:nth-of-type(2n) .timeline-content {
  padding: 1em 2em 0 0;
  text-align: right;
  display: flex;
  flex-direction: column;
  align-items: flex-end;
  flex-basis: 100%;
}
.timeline.is-centered .timeline-item:nth-of-type(2n+1)::before {
  content: "";
  background-color: #dbdbdb;
  display: block;
  width: 1px;
  height: 100%;
  position: absolute;
  top: 0;
}
.timeline.is-centered .timeline-item.is-white::before {
  background-color: #fff;
}
.timeline.is-centered .timeline-item.is-black::before {
  background-color: #0a0a0a;
}
.timeline.is-centered .timeline-item.is-light::before {
  background-color: #f5f5f5;
}
.timeline.is-centered .timeline-item.is-dark::before {
  background-color: #363636;
}
.timeline.is-centered .timeline-item.is-primary::before {
  background-color: #00d1b2;
}
.timeline.is-centered .timeline-item.is-link::before {
  background-color: #485fc7;
}
.timeline.is-centered .timeline-item.is-info::before {
  background-color: #3e8ed0;
}
.timeline.is-centered .timeline-item.is-success::before {
  background-color: #48c78e;
}
.timeline.is-centered .timeline-item.is-warning::before {
  background-color: #ffe08a;
}
.timeline.is-centered .timeline-item.is-danger::before {
  background-color: #f14668;
}
.timeline.is-rtl {
  justify-content: flex-end;
  align-items: flex-end;
}
.timeline.is-rtl .timeline-item {
  justify-content: flex-end;
  flex-direction: row;
  border-left: none;
  margin-left: 0;
  margin-right: 2em;
}
.timeline.is-rtl .timeline-item::before {
  right: 0;
  left: auto;
}
.timeline.is-rtl .timeline-item .timeline-marker {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-marker.is-image.is-16x16 {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-marker.is-image.is-24x24 {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-marker.is-image.is-32x32 {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-marker.is-image.is-48x48 {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-marker.is-image.is-64x64 {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-marker.is-image.is-96x96 {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-marker.is-image.is-128x128 {
  transform: translateX(50%);
}
.timeline.is-rtl .timeline-item .timeline-content {
  padding: 1em 2em 0 0;
  text-align: right;
}
.timeline.no-headers .timeline-item.is-first::before {
  height: calc(100% - 1.2rem);
  top: 1.2rem;
}
.timeline.no-headers .timeline-item.is-last::before {
  height: 1.2rem;
}

/* src/scss/main.scss */
.menu-list div {
  border-radius: 0;
  color: #727c8f;
  display: block;
  padding: 0.5rem 0;
}
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.file-cta,
.file-name,
.select select,
.textarea,
.input,
.button {
  -moz-appearance: none;
  -webkit-appearance: none;
  align-items: center;
  border: 1px solid transparent;
  border-radius: 4px;
  box-shadow: none;
  display: inline-flex;
  font-size: 1rem;
  height: 2.5em;
  justify-content: flex-start;
  line-height: 1.5;
  padding-bottom: calc(0.5em - 1px);
  padding-left: calc(0.75em - 1px);
  padding-right: calc(0.75em - 1px);
  padding-top: calc(0.5em - 1px);
  position: relative;
  vertical-align: top;
}
.pagination-previous:focus,
.pagination-next:focus,
.pagination-link:focus,
.pagination-ellipsis:focus,
.file-cta:focus,
.file-name:focus,
.select select:focus,
.textarea:focus,
.input:focus,
.button:focus,
.is-focused.pagination-previous,
.is-focused.pagination-next,
.is-focused.pagination-link,
.is-focused.pagination-ellipsis,
.is-focused.file-cta,
.is-focused.file-name,
.select select.is-focused,
.is-focused.textarea,
.is-focused.input,
.is-focused.button,
.pagination-previous:active,
.pagination-next:active,
.pagination-link:active,
.pagination-ellipsis:active,
.file-cta:active,
.file-name:active,
.select select:active,
.textarea:active,
.input:active,
.button:active,
.is-active.pagination-previous,
.is-active.pagination-next,
.is-active.pagination-link,
.is-active.pagination-ellipsis,
.is-active.file-cta,
.is-active.file-name,
.select select.is-active,
.is-active.textarea,
.is-active.input,
.is-active.button {
  outline: none;
}
[disabled].pagination-previous,
[disabled].pagination-next,
[disabled].pagination-link,
[disabled].pagination-ellipsis,
[disabled].file-cta,
[disabled].file-name,
.select select[disabled],
[disabled].textarea,
[disabled].input,
[disabled].button,
fieldset[disabled] .pagination-previous,
fieldset[disabled] .pagination-next,
fieldset[disabled] .pagination-link,
fieldset[disabled] .pagination-ellipsis,
fieldset[disabled] .file-cta,
fieldset[disabled] .file-name,
fieldset[disabled] .select select,
.select fieldset[disabled] select,
fieldset[disabled] .textarea,
fieldset[disabled] .input,
fieldset[disabled] .button {
  cursor: not-allowed;
}
.is-unselectable,
.tabs,
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.breadcrumb,
.file,
.button,
.switch,
.b-checkbox.checkbox,
.b-radio.radio {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
}
.navbar-link:not(.is-arrowless)::after,
.select:not(.is-multiple):not(.is-loading)::after {
  border: 3px solid transparent;
  border-radius: 2px;
  border-right: 0;
  border-top: 0;
  content: " ";
  display: block;
  height: 0.625em;
  margin-top: -0.4375em;
  pointer-events: none;
  position: absolute;
  top: 50%;
  transform: rotate(-45deg);
  transform-origin: center;
  width: 0.625em;
}
.tabs:not(:last-child),
.pagination:not(:last-child),
.message:not(:last-child),
.level:not(:last-child),
.breadcrumb:not(:last-child),
.block:not(:last-child),
.title:not(:last-child),
.subtitle:not(:last-child),
.table-container:not(:last-child),
.table:not(:last-child),
.progress:not(:last-child),
.notification:not(:last-child),
.content:not(:last-child),
.box:not(:last-child) {
  margin-bottom: 1.5rem;
}
.modal-close,
.delete {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  -moz-appearance: none;
  -webkit-appearance: none;
  background-color: rgba(10, 10, 10, 0.2);
  border: none;
  border-radius: 9999px;
  cursor: pointer;
  pointer-events: auto;
  display: inline-block;
  flex-grow: 0;
  flex-shrink: 0;
  font-size: 0;
  height: 20px;
  max-height: 20px;
  max-width: 20px;
  min-height: 20px;
  min-width: 20px;
  outline: none;
  position: relative;
  vertical-align: top;
  width: 20px;
}
.modal-close::before,
.delete::before,
.modal-close::after,
.delete::after {
  background-color: hsl(0deg, 0%, 100%);
  content: "";
  display: block;
  left: 50%;
  position: absolute;
  top: 50%;
  transform: translateX(-50%) translateY(-50%) rotate(45deg);
  transform-origin: center center;
}
.modal-close::before,
.delete::before {
  height: 2px;
  width: 50%;
}
.modal-close::after,
.delete::after {
  height: 50%;
  width: 2px;
}
.modal-close:hover,
.delete:hover,
.modal-close:focus,
.delete:focus {
  background-color: rgba(10, 10, 10, 0.3);
}
.modal-close:active,
.delete:active {
  background-color: rgba(10, 10, 10, 0.4);
}
.is-small.modal-close,
.is-small.delete {
  height: 16px;
  max-height: 16px;
  max-width: 16px;
  min-height: 16px;
  min-width: 16px;
  width: 16px;
}
.is-medium.modal-close,
.is-medium.delete {
  height: 24px;
  max-height: 24px;
  max-width: 24px;
  min-height: 24px;
  min-width: 24px;
  width: 24px;
}
.is-large.modal-close,
.is-large.delete {
  height: 32px;
  max-height: 32px;
  max-width: 32px;
  min-height: 32px;
  min-width: 32px;
  width: 32px;
}
.control.is-loading::after,
.select.is-loading::after,
.loader,
.button.is-loading::after {
  animation: spinAround 500ms infinite linear;
  border: 2px solid hsl(0deg, 0%, 86%);
  border-radius: 9999px;
  border-right-color: transparent;
  border-top-color: transparent;
  content: "";
  display: block;
  height: 1em;
  position: relative;
  width: 1em;
}
.hero-video,
.is-overlay,
.modal-background,
.modal,
.image.is-square img,
.image.is-square .has-ratio,
.image.is-1by1 img,
.image.is-1by1 .has-ratio,
.image.is-5by4 img,
.image.is-5by4 .has-ratio,
.image.is-4by3 img,
.image.is-4by3 .has-ratio,
.image.is-3by2 img,
.image.is-3by2 .has-ratio,
.image.is-5by3 img,
.image.is-5by3 .has-ratio,
.image.is-16by9 img,
.image.is-16by9 .has-ratio,
.image.is-2by1 img,
.image.is-2by1 .has-ratio,
.image.is-3by1 img,
.image.is-3by1 .has-ratio,
.image.is-4by5 img,
.image.is-4by5 .has-ratio,
.image.is-3by4 img,
.image.is-3by4 .has-ratio,
.image.is-2by3 img,
.image.is-2by3 .has-ratio,
.image.is-3by5 img,
.image.is-3by5 .has-ratio,
.image.is-9by16 img,
.image.is-9by16 .has-ratio,
.image.is-1by2 img,
.image.is-1by2 .has-ratio,
.image.is-1by3 img,
.image.is-1by3 .has-ratio {
  bottom: 0;
  left: 0;
  position: absolute;
  right: 0;
  top: 0;
}
.navbar-burger {
  -moz-appearance: none;
  -webkit-appearance: none;
  appearance: none;
  background: none;
  border: none;
  color: currentColor;
  font-family: inherit;
  font-size: 1em;
  margin: 0;
  padding: 0;
}
.b-radio.radio {
  outline: none;
  display: inline-flex;
  align-items: center;
}
.b-radio.radio:not(.button) {
  margin-right: 0.5em;
}
.b-radio.radio:not(.button) + .radio:last-child {
  margin-right: 0;
}
.b-radio.radio + .radio {
  margin-left: 0;
}
.b-radio.radio input[type=radio] {
  position: absolute;
  left: 0;
  opacity: 0;
  outline: none;
  z-index: -1;
}
.b-radio.radio input[type=radio] + .check {
  display: flex;
  flex-shrink: 0;
  position: relative;
  cursor: pointer;
  width: 1.25em;
  height: 1.25em;
  transition: background 150ms ease-out;
  border-radius: 50%;
  border: 2px solid hsl(0deg, 0%, 48%);
}
.b-radio.radio input[type=radio] + .check:before {
  content: "";
  display: flex;
  position: absolute;
  left: 50%;
  margin-left: calc(-1.25em * 0.5);
  bottom: 50%;
  margin-bottom: calc(-1.25em * 0.5);
  width: 1.25em;
  height: 1.25em;
  transition: transform 150ms ease-out;
  border-radius: 50%;
  transform: scale(0);
  background-color: hsl(171deg, 100%, 41%);
}
.b-radio.radio input[type=radio] + .check.is-white:before {
  background: hsl(0deg, 0%, 100%);
}
.b-radio.radio input[type=radio] + .check.is-black:before {
  background: hsl(0deg, 0%, 4%);
}
.b-radio.radio input[type=radio] + .check.is-light:before {
  background: hsl(0deg, 0%, 96%);
}
.b-radio.radio input[type=radio] + .check.is-dark:before {
  background: hsl(0deg, 0%, 21%);
}
.b-radio.radio input[type=radio] + .check.is-primary:before {
  background: hsl(171deg, 100%, 41%);
}
.b-radio.radio input[type=radio] + .check.is-link:before {
  background: hsl(229deg, 53%, 53%);
}
.b-radio.radio input[type=radio] + .check.is-info:before {
  background: hsl(207deg, 61%, 53%);
}
.b-radio.radio input[type=radio] + .check.is-success:before {
  background: hsl(153deg, 53%, 53%);
}
.b-radio.radio input[type=radio] + .check.is-warning:before {
  background: hsl(44deg, 100%, 77%);
}
.b-radio.radio input[type=radio] + .check.is-danger:before {
  background: hsl(348deg, 86%, 61%);
}
.b-radio.radio input[type=radio]:checked + .check {
  border-color: hsl(171deg, 100%, 41%);
}
.b-radio.radio input[type=radio]:checked + .check.is-white {
  border-color: hsl(0deg, 0%, 100%);
}
.b-radio.radio input[type=radio]:checked + .check.is-black {
  border-color: hsl(0deg, 0%, 4%);
}
.b-radio.radio input[type=radio]:checked + .check.is-light {
  border-color: hsl(0deg, 0%, 96%);
}
.b-radio.radio input[type=radio]:checked + .check.is-dark {
  border-color: hsl(0deg, 0%, 21%);
}
.b-radio.radio input[type=radio]:checked + .check.is-primary {
  border-color: hsl(171deg, 100%, 41%);
}
.b-radio.radio input[type=radio]:checked + .check.is-link {
  border-color: hsl(229deg, 53%, 53%);
}
.b-radio.radio input[type=radio]:checked + .check.is-info {
  border-color: hsl(207deg, 61%, 53%);
}
.b-radio.radio input[type=radio]:checked + .check.is-success {
  border-color: hsl(153deg, 53%, 53%);
}
.b-radio.radio input[type=radio]:checked + .check.is-warning {
  border-color: hsl(44deg, 100%, 77%);
}
.b-radio.radio input[type=radio]:checked + .check.is-danger {
  border-color: hsl(348deg, 86%, 61%);
}
.b-radio.radio input[type=radio]:checked + .check:before {
  transform: scale(0.5);
}
.b-radio.radio input[type=radio]:focus + .check {
  box-shadow: 0 0 0.5em rgba(122, 122, 122, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check {
  box-shadow: 0 0 0.5em rgba(0, 209, 178, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-white {
  box-shadow: 0 0 0.5em rgba(255, 255, 255, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-black {
  box-shadow: 0 0 0.5em rgba(10, 10, 10, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-light {
  box-shadow: 0 0 0.5em rgba(245, 245, 245, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-dark {
  box-shadow: 0 0 0.5em rgba(54, 54, 54, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-primary {
  box-shadow: 0 0 0.5em rgba(0, 209, 178, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-link {
  box-shadow: 0 0 0.5em rgba(72, 95, 199, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-info {
  box-shadow: 0 0 0.5em rgba(62, 142, 208, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-success {
  box-shadow: 0 0 0.5em rgba(72, 199, 142, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-warning {
  box-shadow: 0 0 0.5em rgba(255, 224, 138, 0.8);
}
.b-radio.radio input[type=radio]:focus:checked + .check.is-danger {
  box-shadow: 0 0 0.5em rgba(241, 70, 104, 0.8);
}
.b-radio.radio .control-label {
  padding-left: calc(0.75em - 1px);
}
.b-radio.radio.button {
  display: flex;
}
.b-radio.radio.button.is-selected {
  z-index: 1;
}
.b-radio.radio[disabled] {
  opacity: 0.5;
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check {
  border-color: hsl(171deg, 100%, 41%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-white {
  border-color: hsl(0deg, 0%, 100%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-black {
  border-color: hsl(0deg, 0%, 4%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-light {
  border-color: hsl(0deg, 0%, 96%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-dark {
  border-color: hsl(0deg, 0%, 21%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-primary {
  border-color: hsl(171deg, 100%, 41%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-link {
  border-color: hsl(229deg, 53%, 53%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-info {
  border-color: hsl(207deg, 61%, 53%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-success {
  border-color: hsl(153deg, 53%, 53%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-warning {
  border-color: hsl(44deg, 100%, 77%);
}
.b-radio.radio:hover input[type=radio]:not(:disabled) + .check.is-danger {
  border-color: hsl(348deg, 86%, 61%);
}
.b-radio.radio.is-small {
  border-radius: 2px;
  font-size: 0.75rem;
}
.b-radio.radio.is-medium {
  font-size: 1.25rem;
}
.b-radio.radio.is-large {
  font-size: 1.5rem;
}
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.file-cta,
.file-name,
.select select,
.textarea,
.input,
.button {
  -moz-appearance: none;
  -webkit-appearance: none;
  align-items: center;
  border: 1px solid transparent;
  border-radius: 4px;
  box-shadow: none;
  display: inline-flex;
  font-size: 1rem;
  height: 2.5em;
  justify-content: flex-start;
  line-height: 1.5;
  padding-bottom: calc(0.5em - 1px);
  padding-left: calc(0.75em - 1px);
  padding-right: calc(0.75em - 1px);
  padding-top: calc(0.5em - 1px);
  position: relative;
  vertical-align: top;
}
.pagination-previous:focus,
.pagination-next:focus,
.pagination-link:focus,
.pagination-ellipsis:focus,
.file-cta:focus,
.file-name:focus,
.select select:focus,
.textarea:focus,
.input:focus,
.button:focus,
.is-focused.pagination-previous,
.is-focused.pagination-next,
.is-focused.pagination-link,
.is-focused.pagination-ellipsis,
.is-focused.file-cta,
.is-focused.file-name,
.select select.is-focused,
.is-focused.textarea,
.is-focused.input,
.is-focused.button,
.pagination-previous:active,
.pagination-next:active,
.pagination-link:active,
.pagination-ellipsis:active,
.file-cta:active,
.file-name:active,
.select select:active,
.textarea:active,
.input:active,
.button:active,
.is-active.pagination-previous,
.is-active.pagination-next,
.is-active.pagination-link,
.is-active.pagination-ellipsis,
.is-active.file-cta,
.is-active.file-name,
.select select.is-active,
.is-active.textarea,
.is-active.input,
.is-active.button {
  outline: none;
}
[disabled].pagination-previous,
[disabled].pagination-next,
[disabled].pagination-link,
[disabled].pagination-ellipsis,
[disabled].file-cta,
[disabled].file-name,
.select select[disabled],
[disabled].textarea,
[disabled].input,
[disabled].button,
fieldset[disabled] .pagination-previous,
fieldset[disabled] .pagination-next,
fieldset[disabled] .pagination-link,
fieldset[disabled] .pagination-ellipsis,
fieldset[disabled] .file-cta,
fieldset[disabled] .file-name,
fieldset[disabled] .select select,
.select fieldset[disabled] select,
fieldset[disabled] .textarea,
fieldset[disabled] .input,
fieldset[disabled] .button {
  cursor: not-allowed;
}
.is-unselectable,
.tabs,
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.breadcrumb,
.file,
.button,
.switch,
.b-checkbox.checkbox,
.b-radio.radio {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
}
.navbar-link:not(.is-arrowless)::after,
.select:not(.is-multiple):not(.is-loading)::after {
  border: 3px solid transparent;
  border-radius: 2px;
  border-right: 0;
  border-top: 0;
  content: " ";
  display: block;
  height: 0.625em;
  margin-top: -0.4375em;
  pointer-events: none;
  position: absolute;
  top: 50%;
  transform: rotate(-45deg);
  transform-origin: center;
  width: 0.625em;
}
.tabs:not(:last-child),
.pagination:not(:last-child),
.message:not(:last-child),
.level:not(:last-child),
.breadcrumb:not(:last-child),
.block:not(:last-child),
.title:not(:last-child),
.subtitle:not(:last-child),
.table-container:not(:last-child),
.table:not(:last-child),
.progress:not(:last-child),
.notification:not(:last-child),
.content:not(:last-child),
.box:not(:last-child) {
  margin-bottom: 1.5rem;
}
.modal-close,
.delete {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  -moz-appearance: none;
  -webkit-appearance: none;
  background-color: rgba(10, 10, 10, 0.2);
  border: none;
  border-radius: 9999px;
  cursor: pointer;
  pointer-events: auto;
  display: inline-block;
  flex-grow: 0;
  flex-shrink: 0;
  font-size: 0;
  height: 20px;
  max-height: 20px;
  max-width: 20px;
  min-height: 20px;
  min-width: 20px;
  outline: none;
  position: relative;
  vertical-align: top;
  width: 20px;
}
.modal-close::before,
.delete::before,
.modal-close::after,
.delete::after {
  background-color: hsl(0deg, 0%, 100%);
  content: "";
  display: block;
  left: 50%;
  position: absolute;
  top: 50%;
  transform: translateX(-50%) translateY(-50%) rotate(45deg);
  transform-origin: center center;
}
.modal-close::before,
.delete::before {
  height: 2px;
  width: 50%;
}
.modal-close::after,
.delete::after {
  height: 50%;
  width: 2px;
}
.modal-close:hover,
.delete:hover,
.modal-close:focus,
.delete:focus {
  background-color: rgba(10, 10, 10, 0.3);
}
.modal-close:active,
.delete:active {
  background-color: rgba(10, 10, 10, 0.4);
}
.is-small.modal-close,
.is-small.delete {
  height: 16px;
  max-height: 16px;
  max-width: 16px;
  min-height: 16px;
  min-width: 16px;
  width: 16px;
}
.is-medium.modal-close,
.is-medium.delete {
  height: 24px;
  max-height: 24px;
  max-width: 24px;
  min-height: 24px;
  min-width: 24px;
  width: 24px;
}
.is-large.modal-close,
.is-large.delete {
  height: 32px;
  max-height: 32px;
  max-width: 32px;
  min-height: 32px;
  min-width: 32px;
  width: 32px;
}
.control.is-loading::after,
.select.is-loading::after,
.loader,
.button.is-loading::after {
  animation: spinAround 500ms infinite linear;
  border: 2px solid hsl(0deg, 0%, 86%);
  border-radius: 9999px;
  border-right-color: transparent;
  border-top-color: transparent;
  content: "";
  display: block;
  height: 1em;
  position: relative;
  width: 1em;
}
.hero-video,
.is-overlay,
.modal-background,
.modal,
.image.is-square img,
.image.is-square .has-ratio,
.image.is-1by1 img,
.image.is-1by1 .has-ratio,
.image.is-5by4 img,
.image.is-5by4 .has-ratio,
.image.is-4by3 img,
.image.is-4by3 .has-ratio,
.image.is-3by2 img,
.image.is-3by2 .has-ratio,
.image.is-5by3 img,
.image.is-5by3 .has-ratio,
.image.is-16by9 img,
.image.is-16by9 .has-ratio,
.image.is-2by1 img,
.image.is-2by1 .has-ratio,
.image.is-3by1 img,
.image.is-3by1 .has-ratio,
.image.is-4by5 img,
.image.is-4by5 .has-ratio,
.image.is-3by4 img,
.image.is-3by4 .has-ratio,
.image.is-2by3 img,
.image.is-2by3 .has-ratio,
.image.is-3by5 img,
.image.is-3by5 .has-ratio,
.image.is-9by16 img,
.image.is-9by16 .has-ratio,
.image.is-1by2 img,
.image.is-1by2 .has-ratio,
.image.is-1by3 img,
.image.is-1by3 .has-ratio {
  bottom: 0;
  left: 0;
  position: absolute;
  right: 0;
  top: 0;
}
.navbar-burger {
  -moz-appearance: none;
  -webkit-appearance: none;
  appearance: none;
  background: none;
  border: none;
  color: currentColor;
  font-family: inherit;
  font-size: 1em;
  margin: 0;
  padding: 0;
}
.b-checkbox.checkbox {
  outline: none;
  display: inline-flex;
  align-items: center;
}
.b-checkbox.checkbox:not(.button) {
  margin-right: 0.5em;
}
.b-checkbox.checkbox:not(.button) + .checkbox:last-child {
  margin-right: 0;
}
.b-checkbox.checkbox input[type=checkbox] {
  position: absolute;
  left: 0;
  opacity: 0;
  outline: none;
  z-index: -1;
}
.b-checkbox.checkbox input[type=checkbox] + .check {
  width: 1.25em;
  height: 1.25em;
  flex-shrink: 0;
  border-radius: 4px;
  border: 2px solid hsl(0deg, 0%, 48%);
  transition: background 150ms ease-out;
  background: transparent;
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check {
  background: hsl(171deg, 100%, 41%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:%23fff' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(171deg, 100%, 41%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-white {
  background: hsl(0deg, 0%, 100%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:hsl(0deg, 0%25, 4%25)' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 100%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-black {
  background: hsl(0deg, 0%, 4%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:hsl(0deg, 0%25, 100%25)' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 4%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-light {
  background: hsl(0deg, 0%, 96%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:rgba(0, 0, 0, 0.7)' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 96%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-dark {
  background: hsl(0deg, 0%, 21%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:%23fff' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 21%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-primary {
  background: hsl(171deg, 100%, 41%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:%23fff' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(171deg, 100%, 41%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-link {
  background: hsl(229deg, 53%, 53%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:%23fff' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(229deg, 53%, 53%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-info {
  background: hsl(207deg, 61%, 53%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:%23fff' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(207deg, 61%, 53%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-success {
  background: hsl(153deg, 53%, 53%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:%23fff' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(153deg, 53%, 53%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-warning {
  background: hsl(44deg, 100%, 77%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:rgba(0, 0, 0, 0.7)' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(44deg, 100%, 77%);
}
.b-checkbox.checkbox input[type=checkbox]:checked + .check.is-danger {
  background: hsl(348deg, 86%, 61%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Cpath style='fill:%23fff' d='M 0.04038059,0.6267767 0.14644661,0.52071068 0.42928932,0.80355339 0.3232233,0.90961941 z M 0.21715729,0.80355339 0.85355339,0.16715729 0.95961941,0.2732233 0.3232233,0.90961941 z'%3E%3C/path%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(348deg, 86%, 61%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check {
  background: hsl(171deg, 100%, 41%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:%23fff' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(171deg, 100%, 41%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-white,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-white {
  background: hsl(0deg, 0%, 100%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:hsl(0deg, 0%25, 4%25)' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 100%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-black,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-black {
  background: hsl(0deg, 0%, 4%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:hsl(0deg, 0%25, 100%25)' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 4%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-light,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-light {
  background: hsl(0deg, 0%, 96%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:rgba(0, 0, 0, 0.7)' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 96%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-dark,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-dark {
  background: hsl(0deg, 0%, 21%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:%23fff' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(0deg, 0%, 21%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-primary,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-primary {
  background: hsl(171deg, 100%, 41%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:%23fff' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(171deg, 100%, 41%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-link,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-link {
  background: hsl(229deg, 53%, 53%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:%23fff' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(229deg, 53%, 53%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-info,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-info {
  background: hsl(207deg, 61%, 53%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:%23fff' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(207deg, 61%, 53%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-success,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-success {
  background: hsl(153deg, 53%, 53%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:%23fff' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(153deg, 53%, 53%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-warning,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-warning {
  background: hsl(44deg, 100%, 77%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:rgba(0, 0, 0, 0.7)' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(44deg, 100%, 77%);
}
.b-checkbox.checkbox input[type=checkbox]:indeterminate + .check.is-danger,
.b-checkbox.checkbox input[type=checkbox].is-indeterminate + .check.is-danger {
  background: hsl(348deg, 86%, 61%) url("data:image/svg+xml,%3Csvg xmlns='http://www.w3.org/2000/svg' viewBox='0 0 1 1'%3E%3Crect style='fill:%23fff' width='0.7' height='0.2' x='.15' y='.4'%3E%3C/rect%3E%3C/svg%3E") no-repeat center center;
  border-color: hsl(348deg, 86%, 61%);
}
.b-checkbox.checkbox input[type=checkbox]:focus + .check {
  box-shadow: 0 0 0.5em rgba(122, 122, 122, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check {
  box-shadow: 0 0 0.5em rgba(0, 209, 178, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-white {
  box-shadow: 0 0 0.5em rgba(255, 255, 255, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-black {
  box-shadow: 0 0 0.5em rgba(10, 10, 10, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-light {
  box-shadow: 0 0 0.5em rgba(245, 245, 245, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-dark {
  box-shadow: 0 0 0.5em rgba(54, 54, 54, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-primary {
  box-shadow: 0 0 0.5em rgba(0, 209, 178, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-link {
  box-shadow: 0 0 0.5em rgba(72, 95, 199, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-info {
  box-shadow: 0 0 0.5em rgba(62, 142, 208, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-success {
  box-shadow: 0 0 0.5em rgba(72, 199, 142, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-warning {
  box-shadow: 0 0 0.5em rgba(255, 224, 138, 0.8);
}
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check.is-danger {
  box-shadow: 0 0 0.5em rgba(241, 70, 104, 0.8);
}
.b-checkbox.checkbox .control-label {
  padding-left: calc(0.75em - 1px);
}
.b-checkbox.checkbox.button {
  display: flex;
}
.b-checkbox.checkbox[disabled] {
  opacity: 0.5;
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check {
  border-color: hsl(171deg, 100%, 41%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-white {
  border-color: hsl(0deg, 0%, 100%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-black {
  border-color: hsl(0deg, 0%, 4%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-light {
  border-color: hsl(0deg, 0%, 96%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-dark {
  border-color: hsl(0deg, 0%, 21%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-primary {
  border-color: hsl(171deg, 100%, 41%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-link {
  border-color: hsl(229deg, 53%, 53%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-info {
  border-color: hsl(207deg, 61%, 53%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-success {
  border-color: hsl(153deg, 53%, 53%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-warning {
  border-color: hsl(44deg, 100%, 77%);
}
.b-checkbox.checkbox:hover input[type=checkbox]:not(:disabled) + .check.is-danger {
  border-color: hsl(348deg, 86%, 61%);
}
.b-checkbox.checkbox.is-small {
  border-radius: 2px;
  font-size: 0.75rem;
}
.b-checkbox.checkbox.is-medium {
  font-size: 1.25rem;
}
.b-checkbox.checkbox.is-large {
  font-size: 1.5rem;
}
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.file-cta,
.file-name,
.select select,
.textarea,
.input,
.button {
  -moz-appearance: none;
  -webkit-appearance: none;
  align-items: center;
  border: 1px solid transparent;
  border-radius: 4px;
  box-shadow: none;
  display: inline-flex;
  font-size: 1rem;
  height: 2.5em;
  justify-content: flex-start;
  line-height: 1.5;
  padding-bottom: calc(0.5em - 1px);
  padding-left: calc(0.75em - 1px);
  padding-right: calc(0.75em - 1px);
  padding-top: calc(0.5em - 1px);
  position: relative;
  vertical-align: top;
}
.pagination-previous:focus,
.pagination-next:focus,
.pagination-link:focus,
.pagination-ellipsis:focus,
.file-cta:focus,
.file-name:focus,
.select select:focus,
.textarea:focus,
.input:focus,
.button:focus,
.is-focused.pagination-previous,
.is-focused.pagination-next,
.is-focused.pagination-link,
.is-focused.pagination-ellipsis,
.is-focused.file-cta,
.is-focused.file-name,
.select select.is-focused,
.is-focused.textarea,
.is-focused.input,
.is-focused.button,
.pagination-previous:active,
.pagination-next:active,
.pagination-link:active,
.pagination-ellipsis:active,
.file-cta:active,
.file-name:active,
.select select:active,
.textarea:active,
.input:active,
.button:active,
.is-active.pagination-previous,
.is-active.pagination-next,
.is-active.pagination-link,
.is-active.pagination-ellipsis,
.is-active.file-cta,
.is-active.file-name,
.select select.is-active,
.is-active.textarea,
.is-active.input,
.is-active.button {
  outline: none;
}
[disabled].pagination-previous,
[disabled].pagination-next,
[disabled].pagination-link,
[disabled].pagination-ellipsis,
[disabled].file-cta,
[disabled].file-name,
.select select[disabled],
[disabled].textarea,
[disabled].input,
[disabled].button,
fieldset[disabled] .pagination-previous,
fieldset[disabled] .pagination-next,
fieldset[disabled] .pagination-link,
fieldset[disabled] .pagination-ellipsis,
fieldset[disabled] .file-cta,
fieldset[disabled] .file-name,
fieldset[disabled] .select select,
.select fieldset[disabled] select,
fieldset[disabled] .textarea,
fieldset[disabled] .input,
fieldset[disabled] .button {
  cursor: not-allowed;
}
.is-unselectable,
.tabs,
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.breadcrumb,
.file,
.button,
.switch,
.b-radio.radio,
.b-checkbox.checkbox {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
}
.navbar-link:not(.is-arrowless)::after,
.select:not(.is-multiple):not(.is-loading)::after {
  border: 3px solid transparent;
  border-radius: 2px;
  border-right: 0;
  border-top: 0;
  content: " ";
  display: block;
  height: 0.625em;
  margin-top: -0.4375em;
  pointer-events: none;
  position: absolute;
  top: 50%;
  transform: rotate(-45deg);
  transform-origin: center;
  width: 0.625em;
}
.tabs:not(:last-child),
.pagination:not(:last-child),
.message:not(:last-child),
.level:not(:last-child),
.breadcrumb:not(:last-child),
.block:not(:last-child),
.title:not(:last-child),
.subtitle:not(:last-child),
.table-container:not(:last-child),
.table:not(:last-child),
.progress:not(:last-child),
.notification:not(:last-child),
.content:not(:last-child),
.box:not(:last-child) {
  margin-bottom: 1.5rem;
}
.modal-close,
.delete {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  -moz-appearance: none;
  -webkit-appearance: none;
  background-color: rgba(10, 10, 10, 0.2);
  border: none;
  border-radius: 9999px;
  cursor: pointer;
  pointer-events: auto;
  display: inline-block;
  flex-grow: 0;
  flex-shrink: 0;
  font-size: 0;
  height: 20px;
  max-height: 20px;
  max-width: 20px;
  min-height: 20px;
  min-width: 20px;
  outline: none;
  position: relative;
  vertical-align: top;
  width: 20px;
}
.modal-close::before,
.delete::before,
.modal-close::after,
.delete::after {
  background-color: hsl(0deg, 0%, 100%);
  content: "";
  display: block;
  left: 50%;
  position: absolute;
  top: 50%;
  transform: translateX(-50%) translateY(-50%) rotate(45deg);
  transform-origin: center center;
}
.modal-close::before,
.delete::before {
  height: 2px;
  width: 50%;
}
.modal-close::after,
.delete::after {
  height: 50%;
  width: 2px;
}
.modal-close:hover,
.delete:hover,
.modal-close:focus,
.delete:focus {
  background-color: rgba(10, 10, 10, 0.3);
}
.modal-close:active,
.delete:active {
  background-color: rgba(10, 10, 10, 0.4);
}
.is-small.modal-close,
.is-small.delete {
  height: 16px;
  max-height: 16px;
  max-width: 16px;
  min-height: 16px;
  min-width: 16px;
  width: 16px;
}
.is-medium.modal-close,
.is-medium.delete {
  height: 24px;
  max-height: 24px;
  max-width: 24px;
  min-height: 24px;
  min-width: 24px;
  width: 24px;
}
.is-large.modal-close,
.is-large.delete {
  height: 32px;
  max-height: 32px;
  max-width: 32px;
  min-height: 32px;
  min-width: 32px;
  width: 32px;
}
.control.is-loading::after,
.select.is-loading::after,
.loader,
.button.is-loading::after {
  animation: spinAround 500ms infinite linear;
  border: 2px solid hsl(0deg, 0%, 86%);
  border-radius: 9999px;
  border-right-color: transparent;
  border-top-color: transparent;
  content: "";
  display: block;
  height: 1em;
  position: relative;
  width: 1em;
}
.hero-video,
.is-overlay,
.modal-background,
.modal,
.image.is-square img,
.image.is-square .has-ratio,
.image.is-1by1 img,
.image.is-1by1 .has-ratio,
.image.is-5by4 img,
.image.is-5by4 .has-ratio,
.image.is-4by3 img,
.image.is-4by3 .has-ratio,
.image.is-3by2 img,
.image.is-3by2 .has-ratio,
.image.is-5by3 img,
.image.is-5by3 .has-ratio,
.image.is-16by9 img,
.image.is-16by9 .has-ratio,
.image.is-2by1 img,
.image.is-2by1 .has-ratio,
.image.is-3by1 img,
.image.is-3by1 .has-ratio,
.image.is-4by5 img,
.image.is-4by5 .has-ratio,
.image.is-3by4 img,
.image.is-3by4 .has-ratio,
.image.is-2by3 img,
.image.is-2by3 .has-ratio,
.image.is-3by5 img,
.image.is-3by5 .has-ratio,
.image.is-9by16 img,
.image.is-9by16 .has-ratio,
.image.is-1by2 img,
.image.is-1by2 .has-ratio,
.image.is-1by3 img,
.image.is-1by3 .has-ratio {
  bottom: 0;
  left: 0;
  position: absolute;
  right: 0;
  top: 0;
}
.navbar-burger {
  -moz-appearance: none;
  -webkit-appearance: none;
  appearance: none;
  background: none;
  border: none;
  color: currentColor;
  font-family: inherit;
  font-size: 1em;
  margin: 0;
  padding: 0;
}
.switch {
  cursor: pointer;
  display: inline-flex;
  align-items: center;
  position: relative;
  margin-right: 0.5em;
}
.switch + .switch:last-child {
  margin-right: 0;
}
.switch input[type=checkbox] {
  position: absolute;
  left: 0;
  opacity: 0;
  outline: none;
  z-index: -1;
}
.switch input[type=checkbox] + .check {
  display: flex;
  align-items: center;
  flex-shrink: 0;
  width: 2.75em;
  height: 1.575em;
  padding: 0.2em;
  background: hsl(0deg, 0%, 71%);
  border-radius: 4px;
  transition: background 150ms ease-out, box-shadow 150ms ease-out;
}
.switch input[type=checkbox] + .check.is-white-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(0deg, 0%, 100%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-black-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(0deg, 0%, 4%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-light-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(0deg, 0%, 96%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-dark-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(0deg, 0%, 21%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-primary-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(171deg, 100%, 41%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-link-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(229deg, 53%, 53%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-info-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(207deg, 61%, 53%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-success-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(153deg, 53%, 53%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-warning-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(44deg, 100%, 77%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check.is-danger-passive,
.switch input[type=checkbox] + .check:hover {
  background: hsl(348deg, 86%, 61%);
}
.switch input[type=checkbox] + .check.input[type=checkbox] + .switch input[type=checkbox] + .check.check {
  background: "pink";
}
.switch input[type=checkbox] + .check:before {
  content: "";
  display: block;
  border-radius: 4px;
  width: 1.175em;
  height: 1.175em;
  background: hsl(0deg, 0%, 96%);
  transition: transform 150ms ease-out;
  will-change: transform;
  transform-origin: left;
}
.switch input[type=checkbox] + .check.is-elastic:before {
  transform: scaleX(1.5);
  border-radius: 4px;
}
.switch input[type=checkbox]:checked + .check {
  background: hsl(171deg, 100%, 41%);
}
.switch input[type=checkbox]:checked + .check.is-white {
  background: hsl(0deg, 0%, 100%);
}
.switch input[type=checkbox]:checked + .check.is-black {
  background: hsl(0deg, 0%, 4%);
}
.switch input[type=checkbox]:checked + .check.is-light {
  background: hsl(0deg, 0%, 96%);
}
.switch input[type=checkbox]:checked + .check.is-dark {
  background: hsl(0deg, 0%, 21%);
}
.switch input[type=checkbox]:checked + .check.is-primary {
  background: hsl(171deg, 100%, 41%);
}
.switch input[type=checkbox]:checked + .check.is-link {
  background: hsl(229deg, 53%, 53%);
}
.switch input[type=checkbox]:checked + .check.is-info {
  background: hsl(207deg, 61%, 53%);
}
.switch input[type=checkbox]:checked + .check.is-success {
  background: hsl(153deg, 53%, 53%);
}
.switch input[type=checkbox]:checked + .check.is-warning {
  background: hsl(44deg, 100%, 77%);
}
.switch input[type=checkbox]:checked + .check.is-danger {
  background: hsl(348deg, 86%, 61%);
}
.switch input[type=checkbox]:checked + .check:before {
  transform: translate3d(100%, 0, 0);
}
.switch input[type=checkbox]:checked + .check.is-elastic:before {
  transform: translate3d(50%, 0, 0) scaleX(1.5);
}
.switch input[type=checkbox]:focus,
.switch input[type=checkbox]:active {
  outline: none;
}
.switch .control-label {
  padding-left: 0.5em;
}
.switch:hover input[type=checkbox] + .check {
  background: rgba(181, 181, 181, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-white-passive {
  background: rgba(255, 255, 255, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-black-passive {
  background: rgba(10, 10, 10, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-light-passive {
  background: rgba(245, 245, 245, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-dark-passive {
  background: rgba(54, 54, 54, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-primary-passive {
  background: rgba(0, 209, 178, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-link-passive {
  background: rgba(72, 95, 199, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-info-passive {
  background: rgba(62, 142, 208, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-success-passive {
  background: rgba(72, 199, 142, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-warning-passive {
  background: rgba(255, 224, 138, 0.9);
}
.switch:hover input[type=checkbox] + .check.is-danger-passive {
  background: rgba(241, 70, 104, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check {
  background: rgba(0, 209, 178, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-white {
  background: rgba(255, 255, 255, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-black {
  background: rgba(10, 10, 10, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-light {
  background: rgba(245, 245, 245, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-dark {
  background: rgba(54, 54, 54, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-primary {
  background: rgba(0, 209, 178, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-link {
  background: rgba(72, 95, 199, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-info {
  background: rgba(62, 142, 208, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-success {
  background: rgba(72, 199, 142, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-warning {
  background: rgba(255, 224, 138, 0.9);
}
.switch:hover input[type=checkbox]:checked + .check.is-danger {
  background: rgba(241, 70, 104, 0.9);
}
.switch.is-rounded input[type=checkbox] + .check {
  border-radius: 9999px;
}
.switch.is-rounded input[type=checkbox] + .check:before {
  border-radius: 9999px;
}
.switch.is-rounded input[type=checkbox].is-elastic:before {
  transform: scaleX(1.5);
  border-radius: 9999px;
}
.switch.is-outlined input[type=checkbox] + .check {
  background: transparent;
  border: 0.1rem solid hsl(0deg, 0%, 71%);
  padding: 0.1em;
}
.switch.is-outlined input[type=checkbox] + .check.is-white-passive {
  border: 0.1rem solid rgba(255, 255, 255, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-white-passive:before {
  background: hsl(0deg, 0%, 100%);
}
.switch.is-outlined input[type=checkbox] + .check.is-white-passive:hover {
  border-color: rgba(255, 255, 255, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-black-passive {
  border: 0.1rem solid rgba(10, 10, 10, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-black-passive:before {
  background: hsl(0deg, 0%, 4%);
}
.switch.is-outlined input[type=checkbox] + .check.is-black-passive:hover {
  border-color: rgba(10, 10, 10, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-light-passive {
  border: 0.1rem solid rgba(245, 245, 245, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-light-passive:before {
  background: hsl(0deg, 0%, 96%);
}
.switch.is-outlined input[type=checkbox] + .check.is-light-passive:hover {
  border-color: rgba(245, 245, 245, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-dark-passive {
  border: 0.1rem solid rgba(54, 54, 54, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-dark-passive:before {
  background: hsl(0deg, 0%, 21%);
}
.switch.is-outlined input[type=checkbox] + .check.is-dark-passive:hover {
  border-color: rgba(54, 54, 54, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-primary-passive {
  border: 0.1rem solid rgba(0, 209, 178, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-primary-passive:before {
  background: hsl(171deg, 100%, 41%);
}
.switch.is-outlined input[type=checkbox] + .check.is-primary-passive:hover {
  border-color: rgba(0, 209, 178, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-link-passive {
  border: 0.1rem solid rgba(72, 95, 199, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-link-passive:before {
  background: hsl(229deg, 53%, 53%);
}
.switch.is-outlined input[type=checkbox] + .check.is-link-passive:hover {
  border-color: rgba(72, 95, 199, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-info-passive {
  border: 0.1rem solid rgba(62, 142, 208, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-info-passive:before {
  background: hsl(207deg, 61%, 53%);
}
.switch.is-outlined input[type=checkbox] + .check.is-info-passive:hover {
  border-color: rgba(62, 142, 208, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-success-passive {
  border: 0.1rem solid rgba(72, 199, 142, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-success-passive:before {
  background: hsl(153deg, 53%, 53%);
}
.switch.is-outlined input[type=checkbox] + .check.is-success-passive:hover {
  border-color: rgba(72, 199, 142, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-warning-passive {
  border: 0.1rem solid rgba(255, 224, 138, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-warning-passive:before {
  background: hsl(44deg, 100%, 77%);
}
.switch.is-outlined input[type=checkbox] + .check.is-warning-passive:hover {
  border-color: rgba(255, 224, 138, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-danger-passive {
  border: 0.1rem solid rgba(241, 70, 104, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check.is-danger-passive:before {
  background: hsl(348deg, 86%, 61%);
}
.switch.is-outlined input[type=checkbox] + .check.is-danger-passive:hover {
  border-color: rgba(241, 70, 104, 0.9);
}
.switch.is-outlined input[type=checkbox] + .check:before {
  background: hsl(0deg, 0%, 71%);
}
.switch.is-outlined input[type=checkbox]:checked + .check {
  border-color: hsl(171deg, 100%, 41%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-white {
  background: transparent;
  border-color: hsl(0deg, 0%, 100%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-white:before {
  background: hsl(0deg, 0%, 100%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-black {
  background: transparent;
  border-color: hsl(0deg, 0%, 4%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-black:before {
  background: hsl(0deg, 0%, 4%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-light {
  background: transparent;
  border-color: hsl(0deg, 0%, 96%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-light:before {
  background: hsl(0deg, 0%, 96%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-dark {
  background: transparent;
  border-color: hsl(0deg, 0%, 21%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-dark:before {
  background: hsl(0deg, 0%, 21%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-primary {
  background: transparent;
  border-color: hsl(171deg, 100%, 41%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-primary:before {
  background: hsl(171deg, 100%, 41%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-link {
  background: transparent;
  border-color: hsl(229deg, 53%, 53%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-link:before {
  background: hsl(229deg, 53%, 53%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-info {
  background: transparent;
  border-color: hsl(207deg, 61%, 53%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-info:before {
  background: hsl(207deg, 61%, 53%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-success {
  background: transparent;
  border-color: hsl(153deg, 53%, 53%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-success:before {
  background: hsl(153deg, 53%, 53%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-warning {
  background: transparent;
  border-color: hsl(44deg, 100%, 77%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-warning:before {
  background: hsl(44deg, 100%, 77%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-danger {
  background: transparent;
  border-color: hsl(348deg, 86%, 61%);
}
.switch.is-outlined input[type=checkbox]:checked + .check.is-danger:before {
  background: hsl(348deg, 86%, 61%);
}
.switch.is-outlined input[type=checkbox]:checked + .check:before {
  background: hsl(171deg, 100%, 41%);
}
.switch.is-outlined:hover input[type=checkbox] + .check {
  background: transparent;
  border-color: rgba(181, 181, 181, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check {
  background: transparent;
  border-color: rgba(0, 209, 178, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-white {
  border-color: rgba(255, 255, 255, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-black {
  border-color: rgba(10, 10, 10, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-light {
  border-color: rgba(245, 245, 245, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-dark {
  border-color: rgba(54, 54, 54, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-primary {
  border-color: rgba(0, 209, 178, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-link {
  border-color: rgba(72, 95, 199, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-info {
  border-color: rgba(62, 142, 208, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-success {
  border-color: rgba(72, 199, 142, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-warning {
  border-color: rgba(255, 224, 138, 0.9);
}
.switch.is-outlined:hover input[type=checkbox]:checked + .check.is-danger {
  border-color: rgba(241, 70, 104, 0.9);
}
.switch.is-small {
  border-radius: 2px;
  font-size: 0.75rem;
}
.switch.is-medium {
  font-size: 1.25rem;
}
.switch.is-large {
  font-size: 1.5rem;
}
.switch[disabled] {
  opacity: 0.5;
  cursor: not-allowed;
  color: hsl(0deg, 0%, 48%);
}
.upload {
  position: relative;
  display: inline-flex;
}
.upload input[type=file] {
  position: absolute;
  top: 0;
  left: 0;
  width: 100%;
  height: 100%;
  opacity: 0;
  outline: none;
  cursor: pointer;
  z-index: -1;
}
.upload.is-expanded {
  width: 100%;
}
.field.file.has-addons .button.is-static {
  color: hsl(0deg, 0%, 14%);
  background-color: hsl(0deg, 0%, 100%);
  border-left: 0;
}
@media screen and (-ms-high-contrast: active), screen and (-ms-high-contrast: none) {
  .upload input[type=file] {
    z-index: auto;
  }
}
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.file-cta,
.file-name,
.select select,
.textarea,
.input,
.button {
  -moz-appearance: none;
  -webkit-appearance: none;
  align-items: center;
  border: 1px solid transparent;
  border-radius: 4px;
  box-shadow: none;
  display: inline-flex;
  font-size: 1rem;
  height: 2.5em;
  justify-content: flex-start;
  line-height: 1.5;
  padding-bottom: calc(0.5em - 1px);
  padding-left: calc(0.75em - 1px);
  padding-right: calc(0.75em - 1px);
  padding-top: calc(0.5em - 1px);
  position: relative;
  vertical-align: top;
}
.pagination-previous:focus,
.pagination-next:focus,
.pagination-link:focus,
.pagination-ellipsis:focus,
.file-cta:focus,
.file-name:focus,
.select select:focus,
.textarea:focus,
.input:focus,
.button:focus,
.is-focused.pagination-previous,
.is-focused.pagination-next,
.is-focused.pagination-link,
.is-focused.pagination-ellipsis,
.is-focused.file-cta,
.is-focused.file-name,
.select select.is-focused,
.is-focused.textarea,
.is-focused.input,
.is-focused.button,
.pagination-previous:active,
.pagination-next:active,
.pagination-link:active,
.pagination-ellipsis:active,
.file-cta:active,
.file-name:active,
.select select:active,
.textarea:active,
.input:active,
.button:active,
.is-active.pagination-previous,
.is-active.pagination-next,
.is-active.pagination-link,
.is-active.pagination-ellipsis,
.is-active.file-cta,
.is-active.file-name,
.select select.is-active,
.is-active.textarea,
.is-active.input,
.is-active.button {
  outline: none;
}
[disabled].pagination-previous,
[disabled].pagination-next,
[disabled].pagination-link,
[disabled].pagination-ellipsis,
[disabled].file-cta,
[disabled].file-name,
.select select[disabled],
[disabled].textarea,
[disabled].input,
[disabled].button,
fieldset[disabled] .pagination-previous,
fieldset[disabled] .pagination-next,
fieldset[disabled] .pagination-link,
fieldset[disabled] .pagination-ellipsis,
fieldset[disabled] .file-cta,
fieldset[disabled] .file-name,
fieldset[disabled] .select select,
.select fieldset[disabled] select,
fieldset[disabled] .textarea,
fieldset[disabled] .input,
fieldset[disabled] .button {
  cursor: not-allowed;
}
.is-unselectable,
.tabs,
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis,
.breadcrumb,
.file,
.button,
.b-radio.radio,
.b-checkbox.checkbox,
.switch {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
}
.navbar-link:not(.is-arrowless)::after,
.select:not(.is-multiple):not(.is-loading)::after {
  border: 3px solid transparent;
  border-radius: 2px;
  border-right: 0;
  border-top: 0;
  content: " ";
  display: block;
  height: 0.625em;
  margin-top: -0.4375em;
  pointer-events: none;
  position: absolute;
  top: 50%;
  transform: rotate(-45deg);
  transform-origin: center;
  width: 0.625em;
}
.tabs:not(:last-child),
.pagination:not(:last-child),
.message:not(:last-child),
.level:not(:last-child),
.breadcrumb:not(:last-child),
.block:not(:last-child),
.title:not(:last-child),
.subtitle:not(:last-child),
.table-container:not(:last-child),
.table:not(:last-child),
.progress:not(:last-child),
.notification:not(:last-child),
.content:not(:last-child),
.box:not(:last-child) {
  margin-bottom: 1.5rem;
}
.modal-close,
.delete {
  -webkit-touch-callout: none;
  -webkit-user-select: none;
  -moz-user-select: none;
  -ms-user-select: none;
  user-select: none;
  -moz-appearance: none;
  -webkit-appearance: none;
  background-color: rgba(10, 10, 10, 0.2);
  border: none;
  border-radius: 9999px;
  cursor: pointer;
  pointer-events: auto;
  display: inline-block;
  flex-grow: 0;
  flex-shrink: 0;
  font-size: 0;
  height: 20px;
  max-height: 20px;
  max-width: 20px;
  min-height: 20px;
  min-width: 20px;
  outline: none;
  position: relative;
  vertical-align: top;
  width: 20px;
}
.modal-close::before,
.delete::before,
.modal-close::after,
.delete::after {
  background-color: hsl(0deg, 0%, 100%);
  content: "";
  display: block;
  left: 50%;
  position: absolute;
  top: 50%;
  transform: translateX(-50%) translateY(-50%) rotate(45deg);
  transform-origin: center center;
}
.modal-close::before,
.delete::before {
  height: 2px;
  width: 50%;
}
.modal-close::after,
.delete::after {
  height: 50%;
  width: 2px;
}
.modal-close:hover,
.delete:hover,
.modal-close:focus,
.delete:focus {
  background-color: rgba(10, 10, 10, 0.3);
}
.modal-close:active,
.delete:active {
  background-color: rgba(10, 10, 10, 0.4);
}
.is-small.modal-close,
.is-small.delete {
  height: 16px;
  max-height: 16px;
  max-width: 16px;
  min-height: 16px;
  min-width: 16px;
  width: 16px;
}
.is-medium.modal-close,
.is-medium.delete {
  height: 24px;
  max-height: 24px;
  max-width: 24px;
  min-height: 24px;
  min-width: 24px;
  width: 24px;
}
.is-large.modal-close,
.is-large.delete {
  height: 32px;
  max-height: 32px;
  max-width: 32px;
  min-height: 32px;
  min-width: 32px;
  width: 32px;
}
.control.is-loading::after,
.select.is-loading::after,
.loader,
.button.is-loading::after {
  animation: spinAround 500ms infinite linear;
  border: 2px solid hsl(0deg, 0%, 86%);
  border-radius: 9999px;
  border-right-color: transparent;
  border-top-color: transparent;
  content: "";
  display: block;
  height: 1em;
  position: relative;
  width: 1em;
}
.hero-video,
.is-overlay,
.modal-background,
.modal,
.image.is-square img,
.image.is-square .has-ratio,
.image.is-1by1 img,
.image.is-1by1 .has-ratio,
.image.is-5by4 img,
.image.is-5by4 .has-ratio,
.image.is-4by3 img,
.image.is-4by3 .has-ratio,
.image.is-3by2 img,
.image.is-3by2 .has-ratio,
.image.is-5by3 img,
.image.is-5by3 .has-ratio,
.image.is-16by9 img,
.image.is-16by9 .has-ratio,
.image.is-2by1 img,
.image.is-2by1 .has-ratio,
.image.is-3by1 img,
.image.is-3by1 .has-ratio,
.image.is-4by5 img,
.image.is-4by5 .has-ratio,
.image.is-3by4 img,
.image.is-3by4 .has-ratio,
.image.is-2by3 img,
.image.is-2by3 .has-ratio,
.image.is-3by5 img,
.image.is-3by5 .has-ratio,
.image.is-9by16 img,
.image.is-9by16 .has-ratio,
.image.is-1by2 img,
.image.is-1by2 .has-ratio,
.image.is-1by3 img,
.image.is-1by3 .has-ratio {
  bottom: 0;
  left: 0;
  position: absolute;
  right: 0;
  top: 0;
}
.navbar-burger {
  -moz-appearance: none;
  -webkit-appearance: none;
  appearance: none;
  background: none;
  border: none;
  color: currentColor;
  font-family: inherit;
  font-size: 1em;
  margin: 0;
  padding: 0;
}
html,
body,
p,
ol,
ul,
li,
dl,
dt,
dd,
blockquote,
figure,
fieldset,
legend,
textarea,
pre,
iframe,
hr,
h1,
h2,
h3,
h4,
h5,
h6 {
  margin: 0;
  padding: 0;
}
h1,
h2,
h3,
h4,
h5,
h6 {
  font-size: 100%;
  font-weight: normal;
}
ul {
  list-style: none;
}
button,
input,
select,
textarea {
  margin: 0;
}
html {
  box-sizing: border-box;
}
*,
*::before,
*::after {
  box-sizing: inherit;
}
img,
video {
  height: auto;
  max-width: 100%;
}
iframe {
  border: 0;
}
table {
  border-collapse: collapse;
  border-spacing: 0;
}
td,
th {
  padding: 0;
}
td:not([align]),
th:not([align]) {
  text-align: inherit;
}
html {
  background-color: #f8f8f8;
  font-size: 16px;
  -moz-osx-font-smoothing: grayscale;
  -webkit-font-smoothing: antialiased;
  min-width: 300px;
  overflow-x: hidden;
  overflow-y: scroll;
  text-rendering: optimizeLegibility;
  text-size-adjust: 100%;
}
article,
aside,
figure,
footer,
header,
hgroup,
section {
  display: block;
}
body,
button,
input,
optgroup,
select,
textarea {
  font-family: "Nunito", sans-serif;
}
code,
pre {
  -moz-osx-font-smoothing: auto;
  -webkit-font-smoothing: auto;
  font-family: monospace;
}
body {
  color: hsl(0deg, 0%, 29%);
  font-size: 1em;
  font-weight: 400;
  line-height: 1.5;
}
a {
  color: hsl(229deg, 53%, 53%);
  cursor: pointer;
  text-decoration: none;
}
a strong {
  color: currentColor;
}
a:hover {
  color: hsl(0deg, 0%, 21%);
}
code {
  background-color: hsl(0deg, 0%, 96%);
  color: #da1039;
  font-size: 0.875em;
  font-weight: normal;
  padding: 0.25em 0.5em 0.25em;
}
hr {
  background-color: hsl(0deg, 0%, 96%);
  border: none;
  display: block;
  height: 1px;
  margin: 1.5rem 0;
}
img {
  height: auto;
  max-width: 100%;
}
input[type=checkbox],
input[type=radio] {
  vertical-align: baseline;
}
small {
  font-size: 0.875em;
}
span {
  font-style: inherit;
  font-weight: inherit;
}
strong {
  color: hsl(0deg, 0%, 21%);
  font-weight: 700;
}
fieldset {
  border: none;
}
pre {
  -webkit-overflow-scrolling: touch;
  background-color: hsl(0deg, 0%, 96%);
  color: hsl(0deg, 0%, 29%);
  font-size: 0.875em;
  overflow-x: auto;
  padding: 1.25rem 1.5rem;
  white-space: pre;
  word-wrap: normal;
}
pre code {
  background-color: transparent;
  color: currentColor;
  font-size: 1em;
  padding: 0;
}
table td,
table th {
  vertical-align: top;
}
table td:not([align]),
table th:not([align]) {
  text-align: inherit;
}
table th {
  color: hsl(0deg, 0%, 21%);
}
@keyframes spinAround {
  from {
    transform: rotate(0deg);
  }
  to {
    transform: rotate(359deg);
  }
}
.box {
  background-color: hsl(0deg, 0%, 100%);
  border-radius: 6px;
  box-shadow: 0 0.5em 1em -0.125em rgba(10, 10, 10, 0.1), 0 0px 0 1px rgba(10, 10, 10, 0.02);
  color: hsl(0deg, 0%, 29%);
  display: block;
  padding: 1.25rem;
}
a.box:hover,
a.box:focus {
  box-shadow: 0 0.5em 1em -0.125em rgba(10, 10, 10, 0.1), 0 0 0 1px hsl(229deg, 53%, 53%);
}
a.box:active {
  box-shadow: inset 0 1px 2px rgba(10, 10, 10, 0.2), 0 0 0 1px hsl(229deg, 53%, 53%);
}
.button {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 86%);
  border-width: 1px;
  color: hsl(0deg, 0%, 21%);
  cursor: pointer;
  justify-content: center;
  padding-bottom: calc(0.5em - 1px);
  padding-left: 1em;
  padding-right: 1em;
  padding-top: calc(0.5em - 1px);
  text-align: center;
  white-space: nowrap;
}
.button strong {
  color: inherit;
}
.button .icon,
.button .icon.is-small,
.button .icon.is-medium,
.button .icon.is-large {
  height: 1.5em;
  width: 1.5em;
}
.button .icon:first-child:not(:last-child) {
  margin-left: calc(-0.5em - 1px);
  margin-right: 0.25em;
}
.button .icon:last-child:not(:first-child) {
  margin-left: 0.25em;
  margin-right: calc(-0.5em - 1px);
}
.button .icon:first-child:last-child {
  margin-left: calc(-0.5em - 1px);
  margin-right: calc(-0.5em - 1px);
}
.button:hover,
.button.is-hovered {
  border-color: hsl(0deg, 0%, 71%);
  color: hsl(0deg, 0%, 21%);
}
.button:focus,
.button.is-focused {
  border-color: hsl(229deg, 53%, 53%);
  color: hsl(0deg, 0%, 21%);
}
.button:focus:not(:active),
.button.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(72, 95, 199, 0.25);
}
.button:active,
.button.is-active {
  border-color: hsl(0deg, 0%, 29%);
  color: hsl(0deg, 0%, 21%);
}
.button.is-text {
  background-color: transparent;
  border-color: transparent;
  color: hsl(0deg, 0%, 29%);
  text-decoration: underline;
}
.button.is-text:hover,
.button.is-text.is-hovered,
.button.is-text:focus,
.button.is-text.is-focused {
  background-color: hsl(0deg, 0%, 96%);
  color: hsl(0deg, 0%, 21%);
}
.button.is-text:active,
.button.is-text.is-active {
  background-color: #e8e8e8;
  color: hsl(0deg, 0%, 21%);
}
.button.is-text[disabled],
fieldset[disabled] .button.is-text {
  background-color: transparent;
  border-color: transparent;
  box-shadow: none;
}
.button.is-ghost {
  background: none;
  border-color: transparent;
  color: hsl(229deg, 53%, 53%);
  text-decoration: none;
}
.button.is-ghost:hover,
.button.is-ghost.is-hovered {
  color: hsl(229deg, 53%, 53%);
  text-decoration: underline;
}
.button.is-white {
  background-color: hsl(0deg, 0%, 100%);
  border-color: transparent;
  color: hsl(0deg, 0%, 4%);
}
.button.is-white:hover,
.button.is-white.is-hovered {
  background-color: #f9f9f9;
  border-color: transparent;
  color: hsl(0deg, 0%, 4%);
}
.button.is-white:focus,
.button.is-white.is-focused {
  border-color: transparent;
  color: hsl(0deg, 0%, 4%);
}
.button.is-white:focus:not(:active),
.button.is-white.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(255, 255, 255, 0.25);
}
.button.is-white:active,
.button.is-white.is-active {
  background-color: #f2f2f2;
  border-color: transparent;
  color: hsl(0deg, 0%, 4%);
}
.button.is-white[disabled],
fieldset[disabled] .button.is-white {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 100%);
  box-shadow: none;
}
.button.is-white.is-inverted {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.button.is-white.is-inverted:hover,
.button.is-white.is-inverted.is-hovered {
  background-color: black;
}
.button.is-white.is-inverted[disabled],
fieldset[disabled] .button.is-white.is-inverted {
  background-color: hsl(0deg, 0%, 4%);
  border-color: transparent;
  box-shadow: none;
  color: hsl(0deg, 0%, 100%);
}
.button.is-white.is-loading::after {
  border-color: transparent transparent hsl(0deg, 0%, 4%) hsl(0deg, 0%, 4%) !important;
}
.button.is-white.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 100%);
}
.button.is-white.is-outlined:hover,
.button.is-white.is-outlined.is-hovered,
.button.is-white.is-outlined:focus,
.button.is-white.is-outlined.is-focused {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.button.is-white.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(0deg, 0%, 100%) hsl(0deg, 0%, 100%) !important;
}
.button.is-white.is-outlined.is-loading:hover::after,
.button.is-white.is-outlined.is-loading.is-hovered::after,
.button.is-white.is-outlined.is-loading:focus::after,
.button.is-white.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(0deg, 0%, 4%) hsl(0deg, 0%, 4%) !important;
}
.button.is-white.is-outlined[disabled],
fieldset[disabled] .button.is-white.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 100%);
  box-shadow: none;
  color: hsl(0deg, 0%, 100%);
}
.button.is-white.is-inverted.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 4%);
}
.button.is-white.is-inverted.is-outlined:hover,
.button.is-white.is-inverted.is-outlined.is-hovered,
.button.is-white.is-inverted.is-outlined:focus,
.button.is-white.is-inverted.is-outlined.is-focused {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.button.is-white.is-inverted.is-outlined.is-loading:hover::after,
.button.is-white.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-white.is-inverted.is-outlined.is-loading:focus::after,
.button.is-white.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(0deg, 0%, 100%) hsl(0deg, 0%, 100%) !important;
}
.button.is-white.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-white.is-inverted.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 4%);
  box-shadow: none;
  color: hsl(0deg, 0%, 4%);
}
.button.is-black {
  background-color: hsl(0deg, 0%, 4%);
  border-color: transparent;
  color: hsl(0deg, 0%, 100%);
}
.button.is-black:hover,
.button.is-black.is-hovered {
  background-color: #040404;
  border-color: transparent;
  color: hsl(0deg, 0%, 100%);
}
.button.is-black:focus,
.button.is-black.is-focused {
  border-color: transparent;
  color: hsl(0deg, 0%, 100%);
}
.button.is-black:focus:not(:active),
.button.is-black.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(10, 10, 10, 0.25);
}
.button.is-black:active,
.button.is-black.is-active {
  background-color: black;
  border-color: transparent;
  color: hsl(0deg, 0%, 100%);
}
.button.is-black[disabled],
fieldset[disabled] .button.is-black {
  background-color: hsl(0deg, 0%, 4%);
  border-color: hsl(0deg, 0%, 4%);
  box-shadow: none;
}
.button.is-black.is-inverted {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.button.is-black.is-inverted:hover,
.button.is-black.is-inverted.is-hovered {
  background-color: #f2f2f2;
}
.button.is-black.is-inverted[disabled],
fieldset[disabled] .button.is-black.is-inverted {
  background-color: hsl(0deg, 0%, 100%);
  border-color: transparent;
  box-shadow: none;
  color: hsl(0deg, 0%, 4%);
}
.button.is-black.is-loading::after {
  border-color: transparent transparent hsl(0deg, 0%, 100%) hsl(0deg, 0%, 100%) !important;
}
.button.is-black.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 4%);
}
.button.is-black.is-outlined:hover,
.button.is-black.is-outlined.is-hovered,
.button.is-black.is-outlined:focus,
.button.is-black.is-outlined.is-focused {
  background-color: hsl(0deg, 0%, 4%);
  border-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.button.is-black.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(0deg, 0%, 4%) hsl(0deg, 0%, 4%) !important;
}
.button.is-black.is-outlined.is-loading:hover::after,
.button.is-black.is-outlined.is-loading.is-hovered::after,
.button.is-black.is-outlined.is-loading:focus::after,
.button.is-black.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(0deg, 0%, 100%) hsl(0deg, 0%, 100%) !important;
}
.button.is-black.is-outlined[disabled],
fieldset[disabled] .button.is-black.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 4%);
  box-shadow: none;
  color: hsl(0deg, 0%, 4%);
}
.button.is-black.is-inverted.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 100%);
}
.button.is-black.is-inverted.is-outlined:hover,
.button.is-black.is-inverted.is-outlined.is-hovered,
.button.is-black.is-inverted.is-outlined:focus,
.button.is-black.is-inverted.is-outlined.is-focused {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.button.is-black.is-inverted.is-outlined.is-loading:hover::after,
.button.is-black.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-black.is-inverted.is-outlined.is-loading:focus::after,
.button.is-black.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(0deg, 0%, 4%) hsl(0deg, 0%, 4%) !important;
}
.button.is-black.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-black.is-inverted.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 100%);
  box-shadow: none;
  color: hsl(0deg, 0%, 100%);
}
.button.is-light {
  background-color: hsl(0deg, 0%, 96%);
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-light:hover,
.button.is-light.is-hovered {
  background-color: #eeeeee;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-light:focus,
.button.is-light.is-focused {
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-light:focus:not(:active),
.button.is-light.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(245, 245, 245, 0.25);
}
.button.is-light:active,
.button.is-light.is-active {
  background-color: #e8e8e8;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-light[disabled],
fieldset[disabled] .button.is-light {
  background-color: hsl(0deg, 0%, 96%);
  border-color: hsl(0deg, 0%, 96%);
  box-shadow: none;
}
.button.is-light.is-inverted {
  background-color: rgba(0, 0, 0, 0.7);
  color: hsl(0deg, 0%, 96%);
}
.button.is-light.is-inverted:hover,
.button.is-light.is-inverted.is-hovered {
  background-color: rgba(0, 0, 0, 0.7);
}
.button.is-light.is-inverted[disabled],
fieldset[disabled] .button.is-light.is-inverted {
  background-color: rgba(0, 0, 0, 0.7);
  border-color: transparent;
  box-shadow: none;
  color: hsl(0deg, 0%, 96%);
}
.button.is-light.is-loading::after {
  border-color: transparent transparent rgba(0, 0, 0, 0.7) rgba(0, 0, 0, 0.7) !important;
}
.button.is-light.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 96%);
  color: hsl(0deg, 0%, 96%);
}
.button.is-light.is-outlined:hover,
.button.is-light.is-outlined.is-hovered,
.button.is-light.is-outlined:focus,
.button.is-light.is-outlined.is-focused {
  background-color: hsl(0deg, 0%, 96%);
  border-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.button.is-light.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(0deg, 0%, 96%) hsl(0deg, 0%, 96%) !important;
}
.button.is-light.is-outlined.is-loading:hover::after,
.button.is-light.is-outlined.is-loading.is-hovered::after,
.button.is-light.is-outlined.is-loading:focus::after,
.button.is-light.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent rgba(0, 0, 0, 0.7) rgba(0, 0, 0, 0.7) !important;
}
.button.is-light.is-outlined[disabled],
fieldset[disabled] .button.is-light.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 96%);
  box-shadow: none;
  color: hsl(0deg, 0%, 96%);
}
.button.is-light.is-inverted.is-outlined {
  background-color: transparent;
  border-color: rgba(0, 0, 0, 0.7);
  color: rgba(0, 0, 0, 0.7);
}
.button.is-light.is-inverted.is-outlined:hover,
.button.is-light.is-inverted.is-outlined.is-hovered,
.button.is-light.is-inverted.is-outlined:focus,
.button.is-light.is-inverted.is-outlined.is-focused {
  background-color: rgba(0, 0, 0, 0.7);
  color: hsl(0deg, 0%, 96%);
}
.button.is-light.is-inverted.is-outlined.is-loading:hover::after,
.button.is-light.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-light.is-inverted.is-outlined.is-loading:focus::after,
.button.is-light.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(0deg, 0%, 96%) hsl(0deg, 0%, 96%) !important;
}
.button.is-light.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-light.is-inverted.is-outlined {
  background-color: transparent;
  border-color: rgba(0, 0, 0, 0.7);
  box-shadow: none;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-dark {
  background-color: hsl(0deg, 0%, 21%);
  border-color: transparent;
  color: #fff;
}
.button.is-dark:hover,
.button.is-dark.is-hovered {
  background-color: #2f2f2f;
  border-color: transparent;
  color: #fff;
}
.button.is-dark:focus,
.button.is-dark.is-focused {
  border-color: transparent;
  color: #fff;
}
.button.is-dark:focus:not(:active),
.button.is-dark.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(54, 54, 54, 0.25);
}
.button.is-dark:active,
.button.is-dark.is-active {
  background-color: #292929;
  border-color: transparent;
  color: #fff;
}
.button.is-dark[disabled],
fieldset[disabled] .button.is-dark {
  background-color: hsl(0deg, 0%, 21%);
  border-color: hsl(0deg, 0%, 21%);
  box-shadow: none;
}
.button.is-dark.is-inverted {
  background-color: #fff;
  color: hsl(0deg, 0%, 21%);
}
.button.is-dark.is-inverted:hover,
.button.is-dark.is-inverted.is-hovered {
  background-color: #f2f2f2;
}
.button.is-dark.is-inverted[disabled],
fieldset[disabled] .button.is-dark.is-inverted {
  background-color: #fff;
  border-color: transparent;
  box-shadow: none;
  color: hsl(0deg, 0%, 21%);
}
.button.is-dark.is-loading::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-dark.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 21%);
  color: hsl(0deg, 0%, 21%);
}
.button.is-dark.is-outlined:hover,
.button.is-dark.is-outlined.is-hovered,
.button.is-dark.is-outlined:focus,
.button.is-dark.is-outlined.is-focused {
  background-color: hsl(0deg, 0%, 21%);
  border-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.button.is-dark.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(0deg, 0%, 21%) hsl(0deg, 0%, 21%) !important;
}
.button.is-dark.is-outlined.is-loading:hover::after,
.button.is-dark.is-outlined.is-loading.is-hovered::after,
.button.is-dark.is-outlined.is-loading:focus::after,
.button.is-dark.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-dark.is-outlined[disabled],
fieldset[disabled] .button.is-dark.is-outlined {
  background-color: transparent;
  border-color: hsl(0deg, 0%, 21%);
  box-shadow: none;
  color: hsl(0deg, 0%, 21%);
}
.button.is-dark.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  color: #fff;
}
.button.is-dark.is-inverted.is-outlined:hover,
.button.is-dark.is-inverted.is-outlined.is-hovered,
.button.is-dark.is-inverted.is-outlined:focus,
.button.is-dark.is-inverted.is-outlined.is-focused {
  background-color: #fff;
  color: hsl(0deg, 0%, 21%);
}
.button.is-dark.is-inverted.is-outlined.is-loading:hover::after,
.button.is-dark.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-dark.is-inverted.is-outlined.is-loading:focus::after,
.button.is-dark.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(0deg, 0%, 21%) hsl(0deg, 0%, 21%) !important;
}
.button.is-dark.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-dark.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  box-shadow: none;
  color: #fff;
}
.button.is-primary {
  background-color: hsl(171deg, 100%, 41%);
  border-color: transparent;
  color: #fff;
}
.button.is-primary:hover,
.button.is-primary.is-hovered {
  background-color: #00c4a7;
  border-color: transparent;
  color: #fff;
}
.button.is-primary:focus,
.button.is-primary.is-focused {
  border-color: transparent;
  color: #fff;
}
.button.is-primary:focus:not(:active),
.button.is-primary.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(0, 209, 178, 0.25);
}
.button.is-primary:active,
.button.is-primary.is-active {
  background-color: #00b89c;
  border-color: transparent;
  color: #fff;
}
.button.is-primary[disabled],
fieldset[disabled] .button.is-primary {
  background-color: hsl(171deg, 100%, 41%);
  border-color: hsl(171deg, 100%, 41%);
  box-shadow: none;
}
.button.is-primary.is-inverted {
  background-color: #fff;
  color: hsl(171deg, 100%, 41%);
}
.button.is-primary.is-inverted:hover,
.button.is-primary.is-inverted.is-hovered {
  background-color: #f2f2f2;
}
.button.is-primary.is-inverted[disabled],
fieldset[disabled] .button.is-primary.is-inverted {
  background-color: #fff;
  border-color: transparent;
  box-shadow: none;
  color: hsl(171deg, 100%, 41%);
}
.button.is-primary.is-loading::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-primary.is-outlined {
  background-color: transparent;
  border-color: hsl(171deg, 100%, 41%);
  color: hsl(171deg, 100%, 41%);
}
.button.is-primary.is-outlined:hover,
.button.is-primary.is-outlined.is-hovered,
.button.is-primary.is-outlined:focus,
.button.is-primary.is-outlined.is-focused {
  background-color: hsl(171deg, 100%, 41%);
  border-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.button.is-primary.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(171deg, 100%, 41%) hsl(171deg, 100%, 41%) !important;
}
.button.is-primary.is-outlined.is-loading:hover::after,
.button.is-primary.is-outlined.is-loading.is-hovered::after,
.button.is-primary.is-outlined.is-loading:focus::after,
.button.is-primary.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-primary.is-outlined[disabled],
fieldset[disabled] .button.is-primary.is-outlined {
  background-color: transparent;
  border-color: hsl(171deg, 100%, 41%);
  box-shadow: none;
  color: hsl(171deg, 100%, 41%);
}
.button.is-primary.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  color: #fff;
}
.button.is-primary.is-inverted.is-outlined:hover,
.button.is-primary.is-inverted.is-outlined.is-hovered,
.button.is-primary.is-inverted.is-outlined:focus,
.button.is-primary.is-inverted.is-outlined.is-focused {
  background-color: #fff;
  color: hsl(171deg, 100%, 41%);
}
.button.is-primary.is-inverted.is-outlined.is-loading:hover::after,
.button.is-primary.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-primary.is-inverted.is-outlined.is-loading:focus::after,
.button.is-primary.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(171deg, 100%, 41%) hsl(171deg, 100%, 41%) !important;
}
.button.is-primary.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-primary.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  box-shadow: none;
  color: #fff;
}
.button.is-primary.is-light {
  background-color: #ebfffc;
  color: #00947e;
}
.button.is-primary.is-light:hover,
.button.is-primary.is-light.is-hovered {
  background-color: #defffa;
  border-color: transparent;
  color: #00947e;
}
.button.is-primary.is-light:active,
.button.is-primary.is-light.is-active {
  background-color: #d1fff8;
  border-color: transparent;
  color: #00947e;
}
.button.is-link {
  background-color: hsl(229deg, 53%, 53%);
  border-color: transparent;
  color: #fff;
}
.button.is-link:hover,
.button.is-link.is-hovered {
  background-color: #3e56c4;
  border-color: transparent;
  color: #fff;
}
.button.is-link:focus,
.button.is-link.is-focused {
  border-color: transparent;
  color: #fff;
}
.button.is-link:focus:not(:active),
.button.is-link.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(72, 95, 199, 0.25);
}
.button.is-link:active,
.button.is-link.is-active {
  background-color: #3a51bb;
  border-color: transparent;
  color: #fff;
}
.button.is-link[disabled],
fieldset[disabled] .button.is-link {
  background-color: hsl(229deg, 53%, 53%);
  border-color: hsl(229deg, 53%, 53%);
  box-shadow: none;
}
.button.is-link.is-inverted {
  background-color: #fff;
  color: hsl(229deg, 53%, 53%);
}
.button.is-link.is-inverted:hover,
.button.is-link.is-inverted.is-hovered {
  background-color: #f2f2f2;
}
.button.is-link.is-inverted[disabled],
fieldset[disabled] .button.is-link.is-inverted {
  background-color: #fff;
  border-color: transparent;
  box-shadow: none;
  color: hsl(229deg, 53%, 53%);
}
.button.is-link.is-loading::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-link.is-outlined {
  background-color: transparent;
  border-color: hsl(229deg, 53%, 53%);
  color: hsl(229deg, 53%, 53%);
}
.button.is-link.is-outlined:hover,
.button.is-link.is-outlined.is-hovered,
.button.is-link.is-outlined:focus,
.button.is-link.is-outlined.is-focused {
  background-color: hsl(229deg, 53%, 53%);
  border-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.button.is-link.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(229deg, 53%, 53%) hsl(229deg, 53%, 53%) !important;
}
.button.is-link.is-outlined.is-loading:hover::after,
.button.is-link.is-outlined.is-loading.is-hovered::after,
.button.is-link.is-outlined.is-loading:focus::after,
.button.is-link.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-link.is-outlined[disabled],
fieldset[disabled] .button.is-link.is-outlined {
  background-color: transparent;
  border-color: hsl(229deg, 53%, 53%);
  box-shadow: none;
  color: hsl(229deg, 53%, 53%);
}
.button.is-link.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  color: #fff;
}
.button.is-link.is-inverted.is-outlined:hover,
.button.is-link.is-inverted.is-outlined.is-hovered,
.button.is-link.is-inverted.is-outlined:focus,
.button.is-link.is-inverted.is-outlined.is-focused {
  background-color: #fff;
  color: hsl(229deg, 53%, 53%);
}
.button.is-link.is-inverted.is-outlined.is-loading:hover::after,
.button.is-link.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-link.is-inverted.is-outlined.is-loading:focus::after,
.button.is-link.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(229deg, 53%, 53%) hsl(229deg, 53%, 53%) !important;
}
.button.is-link.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-link.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  box-shadow: none;
  color: #fff;
}
.button.is-link.is-light {
  background-color: #eff1fa;
  color: #3850b7;
}
.button.is-link.is-light:hover,
.button.is-link.is-light.is-hovered {
  background-color: #e6e9f7;
  border-color: transparent;
  color: #3850b7;
}
.button.is-link.is-light:active,
.button.is-link.is-light.is-active {
  background-color: #dce0f4;
  border-color: transparent;
  color: #3850b7;
}
.button.is-info {
  background-color: hsl(207deg, 61%, 53%);
  border-color: transparent;
  color: #fff;
}
.button.is-info:hover,
.button.is-info.is-hovered {
  background-color: #3488ce;
  border-color: transparent;
  color: #fff;
}
.button.is-info:focus,
.button.is-info.is-focused {
  border-color: transparent;
  color: #fff;
}
.button.is-info:focus:not(:active),
.button.is-info.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(62, 142, 208, 0.25);
}
.button.is-info:active,
.button.is-info.is-active {
  background-color: #3082c5;
  border-color: transparent;
  color: #fff;
}
.button.is-info[disabled],
fieldset[disabled] .button.is-info {
  background-color: hsl(207deg, 61%, 53%);
  border-color: hsl(207deg, 61%, 53%);
  box-shadow: none;
}
.button.is-info.is-inverted {
  background-color: #fff;
  color: hsl(207deg, 61%, 53%);
}
.button.is-info.is-inverted:hover,
.button.is-info.is-inverted.is-hovered {
  background-color: #f2f2f2;
}
.button.is-info.is-inverted[disabled],
fieldset[disabled] .button.is-info.is-inverted {
  background-color: #fff;
  border-color: transparent;
  box-shadow: none;
  color: hsl(207deg, 61%, 53%);
}
.button.is-info.is-loading::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-info.is-outlined {
  background-color: transparent;
  border-color: hsl(207deg, 61%, 53%);
  color: hsl(207deg, 61%, 53%);
}
.button.is-info.is-outlined:hover,
.button.is-info.is-outlined.is-hovered,
.button.is-info.is-outlined:focus,
.button.is-info.is-outlined.is-focused {
  background-color: hsl(207deg, 61%, 53%);
  border-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.button.is-info.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(207deg, 61%, 53%) hsl(207deg, 61%, 53%) !important;
}
.button.is-info.is-outlined.is-loading:hover::after,
.button.is-info.is-outlined.is-loading.is-hovered::after,
.button.is-info.is-outlined.is-loading:focus::after,
.button.is-info.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-info.is-outlined[disabled],
fieldset[disabled] .button.is-info.is-outlined {
  background-color: transparent;
  border-color: hsl(207deg, 61%, 53%);
  box-shadow: none;
  color: hsl(207deg, 61%, 53%);
}
.button.is-info.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  color: #fff;
}
.button.is-info.is-inverted.is-outlined:hover,
.button.is-info.is-inverted.is-outlined.is-hovered,
.button.is-info.is-inverted.is-outlined:focus,
.button.is-info.is-inverted.is-outlined.is-focused {
  background-color: #fff;
  color: hsl(207deg, 61%, 53%);
}
.button.is-info.is-inverted.is-outlined.is-loading:hover::after,
.button.is-info.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-info.is-inverted.is-outlined.is-loading:focus::after,
.button.is-info.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(207deg, 61%, 53%) hsl(207deg, 61%, 53%) !important;
}
.button.is-info.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-info.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  box-shadow: none;
  color: #fff;
}
.button.is-info.is-light {
  background-color: #eff5fb;
  color: #296fa8;
}
.button.is-info.is-light:hover,
.button.is-info.is-light.is-hovered {
  background-color: #e4eff9;
  border-color: transparent;
  color: #296fa8;
}
.button.is-info.is-light:active,
.button.is-info.is-light.is-active {
  background-color: #dae9f6;
  border-color: transparent;
  color: #296fa8;
}
.button.is-success {
  background-color: hsl(153deg, 53%, 53%);
  border-color: transparent;
  color: #fff;
}
.button.is-success:hover,
.button.is-success.is-hovered {
  background-color: #3ec487;
  border-color: transparent;
  color: #fff;
}
.button.is-success:focus,
.button.is-success.is-focused {
  border-color: transparent;
  color: #fff;
}
.button.is-success:focus:not(:active),
.button.is-success.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(72, 199, 142, 0.25);
}
.button.is-success:active,
.button.is-success.is-active {
  background-color: #3abb81;
  border-color: transparent;
  color: #fff;
}
.button.is-success[disabled],
fieldset[disabled] .button.is-success {
  background-color: hsl(153deg, 53%, 53%);
  border-color: hsl(153deg, 53%, 53%);
  box-shadow: none;
}
.button.is-success.is-inverted {
  background-color: #fff;
  color: hsl(153deg, 53%, 53%);
}
.button.is-success.is-inverted:hover,
.button.is-success.is-inverted.is-hovered {
  background-color: #f2f2f2;
}
.button.is-success.is-inverted[disabled],
fieldset[disabled] .button.is-success.is-inverted {
  background-color: #fff;
  border-color: transparent;
  box-shadow: none;
  color: hsl(153deg, 53%, 53%);
}
.button.is-success.is-loading::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-success.is-outlined {
  background-color: transparent;
  border-color: hsl(153deg, 53%, 53%);
  color: hsl(153deg, 53%, 53%);
}
.button.is-success.is-outlined:hover,
.button.is-success.is-outlined.is-hovered,
.button.is-success.is-outlined:focus,
.button.is-success.is-outlined.is-focused {
  background-color: hsl(153deg, 53%, 53%);
  border-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.button.is-success.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(153deg, 53%, 53%) hsl(153deg, 53%, 53%) !important;
}
.button.is-success.is-outlined.is-loading:hover::after,
.button.is-success.is-outlined.is-loading.is-hovered::after,
.button.is-success.is-outlined.is-loading:focus::after,
.button.is-success.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-success.is-outlined[disabled],
fieldset[disabled] .button.is-success.is-outlined {
  background-color: transparent;
  border-color: hsl(153deg, 53%, 53%);
  box-shadow: none;
  color: hsl(153deg, 53%, 53%);
}
.button.is-success.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  color: #fff;
}
.button.is-success.is-inverted.is-outlined:hover,
.button.is-success.is-inverted.is-outlined.is-hovered,
.button.is-success.is-inverted.is-outlined:focus,
.button.is-success.is-inverted.is-outlined.is-focused {
  background-color: #fff;
  color: hsl(153deg, 53%, 53%);
}
.button.is-success.is-inverted.is-outlined.is-loading:hover::after,
.button.is-success.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-success.is-inverted.is-outlined.is-loading:focus::after,
.button.is-success.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(153deg, 53%, 53%) hsl(153deg, 53%, 53%) !important;
}
.button.is-success.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-success.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  box-shadow: none;
  color: #fff;
}
.button.is-success.is-light {
  background-color: #effaf5;
  color: #257953;
}
.button.is-success.is-light:hover,
.button.is-success.is-light.is-hovered {
  background-color: #e6f7ef;
  border-color: transparent;
  color: #257953;
}
.button.is-success.is-light:active,
.button.is-success.is-light.is-active {
  background-color: #dcf4e9;
  border-color: transparent;
  color: #257953;
}
.button.is-warning {
  background-color: hsl(44deg, 100%, 77%);
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-warning:hover,
.button.is-warning.is-hovered {
  background-color: #ffdc7d;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-warning:focus,
.button.is-warning.is-focused {
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-warning:focus:not(:active),
.button.is-warning.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(255, 224, 138, 0.25);
}
.button.is-warning:active,
.button.is-warning.is-active {
  background-color: #ffd970;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-warning[disabled],
fieldset[disabled] .button.is-warning {
  background-color: hsl(44deg, 100%, 77%);
  border-color: hsl(44deg, 100%, 77%);
  box-shadow: none;
}
.button.is-warning.is-inverted {
  background-color: rgba(0, 0, 0, 0.7);
  color: hsl(44deg, 100%, 77%);
}
.button.is-warning.is-inverted:hover,
.button.is-warning.is-inverted.is-hovered {
  background-color: rgba(0, 0, 0, 0.7);
}
.button.is-warning.is-inverted[disabled],
fieldset[disabled] .button.is-warning.is-inverted {
  background-color: rgba(0, 0, 0, 0.7);
  border-color: transparent;
  box-shadow: none;
  color: hsl(44deg, 100%, 77%);
}
.button.is-warning.is-loading::after {
  border-color: transparent transparent rgba(0, 0, 0, 0.7) rgba(0, 0, 0, 0.7) !important;
}
.button.is-warning.is-outlined {
  background-color: transparent;
  border-color: hsl(44deg, 100%, 77%);
  color: hsl(44deg, 100%, 77%);
}
.button.is-warning.is-outlined:hover,
.button.is-warning.is-outlined.is-hovered,
.button.is-warning.is-outlined:focus,
.button.is-warning.is-outlined.is-focused {
  background-color: hsl(44deg, 100%, 77%);
  border-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.button.is-warning.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(44deg, 100%, 77%) hsl(44deg, 100%, 77%) !important;
}
.button.is-warning.is-outlined.is-loading:hover::after,
.button.is-warning.is-outlined.is-loading.is-hovered::after,
.button.is-warning.is-outlined.is-loading:focus::after,
.button.is-warning.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent rgba(0, 0, 0, 0.7) rgba(0, 0, 0, 0.7) !important;
}
.button.is-warning.is-outlined[disabled],
fieldset[disabled] .button.is-warning.is-outlined {
  background-color: transparent;
  border-color: hsl(44deg, 100%, 77%);
  box-shadow: none;
  color: hsl(44deg, 100%, 77%);
}
.button.is-warning.is-inverted.is-outlined {
  background-color: transparent;
  border-color: rgba(0, 0, 0, 0.7);
  color: rgba(0, 0, 0, 0.7);
}
.button.is-warning.is-inverted.is-outlined:hover,
.button.is-warning.is-inverted.is-outlined.is-hovered,
.button.is-warning.is-inverted.is-outlined:focus,
.button.is-warning.is-inverted.is-outlined.is-focused {
  background-color: rgba(0, 0, 0, 0.7);
  color: hsl(44deg, 100%, 77%);
}
.button.is-warning.is-inverted.is-outlined.is-loading:hover::after,
.button.is-warning.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-warning.is-inverted.is-outlined.is-loading:focus::after,
.button.is-warning.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(44deg, 100%, 77%) hsl(44deg, 100%, 77%) !important;
}
.button.is-warning.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-warning.is-inverted.is-outlined {
  background-color: transparent;
  border-color: rgba(0, 0, 0, 0.7);
  box-shadow: none;
  color: rgba(0, 0, 0, 0.7);
}
.button.is-warning.is-light {
  background-color: #fffaeb;
  color: #946c00;
}
.button.is-warning.is-light:hover,
.button.is-warning.is-light.is-hovered {
  background-color: #fff6de;
  border-color: transparent;
  color: #946c00;
}
.button.is-warning.is-light:active,
.button.is-warning.is-light.is-active {
  background-color: #fff3d1;
  border-color: transparent;
  color: #946c00;
}
.button.is-danger {
  background-color: hsl(348deg, 86%, 61%);
  border-color: transparent;
  color: #fff;
}
.button.is-danger:hover,
.button.is-danger.is-hovered {
  background-color: #f03a5f;
  border-color: transparent;
  color: #fff;
}
.button.is-danger:focus,
.button.is-danger.is-focused {
  border-color: transparent;
  color: #fff;
}
.button.is-danger:focus:not(:active),
.button.is-danger.is-focused:not(:active) {
  box-shadow: 0 0 0 0.125em rgba(241, 70, 104, 0.25);
}
.button.is-danger:active,
.button.is-danger.is-active {
  background-color: #ef2e55;
  border-color: transparent;
  color: #fff;
}
.button.is-danger[disabled],
fieldset[disabled] .button.is-danger {
  background-color: hsl(348deg, 86%, 61%);
  border-color: hsl(348deg, 86%, 61%);
  box-shadow: none;
}
.button.is-danger.is-inverted {
  background-color: #fff;
  color: hsl(348deg, 86%, 61%);
}
.button.is-danger.is-inverted:hover,
.button.is-danger.is-inverted.is-hovered {
  background-color: #f2f2f2;
}
.button.is-danger.is-inverted[disabled],
fieldset[disabled] .button.is-danger.is-inverted {
  background-color: #fff;
  border-color: transparent;
  box-shadow: none;
  color: hsl(348deg, 86%, 61%);
}
.button.is-danger.is-loading::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-danger.is-outlined {
  background-color: transparent;
  border-color: hsl(348deg, 86%, 61%);
  color: hsl(348deg, 86%, 61%);
}
.button.is-danger.is-outlined:hover,
.button.is-danger.is-outlined.is-hovered,
.button.is-danger.is-outlined:focus,
.button.is-danger.is-outlined.is-focused {
  background-color: hsl(348deg, 86%, 61%);
  border-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.button.is-danger.is-outlined.is-loading::after {
  border-color: transparent transparent hsl(348deg, 86%, 61%) hsl(348deg, 86%, 61%) !important;
}
.button.is-danger.is-outlined.is-loading:hover::after,
.button.is-danger.is-outlined.is-loading.is-hovered::after,
.button.is-danger.is-outlined.is-loading:focus::after,
.button.is-danger.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent #fff #fff !important;
}
.button.is-danger.is-outlined[disabled],
fieldset[disabled] .button.is-danger.is-outlined {
  background-color: transparent;
  border-color: hsl(348deg, 86%, 61%);
  box-shadow: none;
  color: hsl(348deg, 86%, 61%);
}
.button.is-danger.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  color: #fff;
}
.button.is-danger.is-inverted.is-outlined:hover,
.button.is-danger.is-inverted.is-outlined.is-hovered,
.button.is-danger.is-inverted.is-outlined:focus,
.button.is-danger.is-inverted.is-outlined.is-focused {
  background-color: #fff;
  color: hsl(348deg, 86%, 61%);
}
.button.is-danger.is-inverted.is-outlined.is-loading:hover::after,
.button.is-danger.is-inverted.is-outlined.is-loading.is-hovered::after,
.button.is-danger.is-inverted.is-outlined.is-loading:focus::after,
.button.is-danger.is-inverted.is-outlined.is-loading.is-focused::after {
  border-color: transparent transparent hsl(348deg, 86%, 61%) hsl(348deg, 86%, 61%) !important;
}
.button.is-danger.is-inverted.is-outlined[disabled],
fieldset[disabled] .button.is-danger.is-inverted.is-outlined {
  background-color: transparent;
  border-color: #fff;
  box-shadow: none;
  color: #fff;
}
.button.is-danger.is-light {
  background-color: #feecf0;
  color: #cc0f35;
}
.button.is-danger.is-light:hover,
.button.is-danger.is-light.is-hovered {
  background-color: #fde0e6;
  border-color: transparent;
  color: #cc0f35;
}
.button.is-danger.is-light:active,
.button.is-danger.is-light.is-active {
  background-color: #fcd4dc;
  border-color: transparent;
  color: #cc0f35;
}
.button.is-small {
  font-size: 0.75rem;
}
.button.is-small:not(.is-rounded) {
  border-radius: 2px;
}
.button.is-normal {
  font-size: 1rem;
}
.button.is-medium {
  font-size: 1.25rem;
}
.button.is-large {
  font-size: 1.5rem;
}
.button[disabled],
fieldset[disabled] .button {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 86%);
  box-shadow: none;
  opacity: 0.5;
}
.button.is-fullwidth {
  display: flex;
  width: 100%;
}
.button.is-loading {
  color: transparent !important;
  pointer-events: none;
}
.button.is-loading::after {
  position: absolute;
  left: calc(50% - (1em * 0.5));
  top: calc(50% - (1em * 0.5));
  position: absolute !important;
}
.button.is-static {
  background-color: hsl(0deg, 0%, 96%);
  border-color: hsl(0deg, 0%, 86%);
  color: hsl(0deg, 0%, 48%);
  box-shadow: none;
  pointer-events: none;
}
.button.is-rounded {
  border-radius: 9999px;
  padding-left: calc(1em + 0.25em);
  padding-right: calc(1em + 0.25em);
}
.buttons {
  align-items: center;
  display: flex;
  flex-wrap: wrap;
  justify-content: flex-start;
}
.buttons .button {
  margin-bottom: 0.5rem;
}
.buttons .button:not(:last-child):not(.is-fullwidth) {
  margin-right: 0.5rem;
}
.buttons:last-child {
  margin-bottom: -0.5rem;
}
.buttons:not(:last-child) {
  margin-bottom: 1rem;
}
.buttons.are-small .button:not(.is-normal):not(.is-medium):not(.is-large) {
  font-size: 0.75rem;
}
.buttons.are-small .button:not(.is-normal):not(.is-medium):not(.is-large):not(.is-rounded) {
  border-radius: 2px;
}
.buttons.are-medium .button:not(.is-small):not(.is-normal):not(.is-large) {
  font-size: 1.25rem;
}
.buttons.are-large .button:not(.is-small):not(.is-normal):not(.is-medium) {
  font-size: 1.5rem;
}
.buttons.has-addons .button:not(:first-child) {
  border-bottom-left-radius: 0;
  border-top-left-radius: 0;
}
.buttons.has-addons .button:not(:last-child) {
  border-bottom-right-radius: 0;
  border-top-right-radius: 0;
  margin-right: -1px;
}
.buttons.has-addons .button:last-child {
  margin-right: 0;
}
.buttons.has-addons .button:hover,
.buttons.has-addons .button.is-hovered {
  z-index: 2;
}
.buttons.has-addons .button:focus,
.buttons.has-addons .button.is-focused,
.buttons.has-addons .button:active,
.buttons.has-addons .button.is-active,
.buttons.has-addons .button.is-selected {
  z-index: 3;
}
.buttons.has-addons .button:focus:hover,
.buttons.has-addons .button.is-focused:hover,
.buttons.has-addons .button:active:hover,
.buttons.has-addons .button.is-active:hover,
.buttons.has-addons .button.is-selected:hover {
  z-index: 4;
}
.buttons.has-addons .button.is-expanded {
  flex-grow: 1;
  flex-shrink: 1;
}
.buttons.is-centered {
  justify-content: center;
}
.buttons.is-centered:not(.has-addons) .button:not(.is-fullwidth) {
  margin-left: 0.25rem;
  margin-right: 0.25rem;
}
.buttons.is-right {
  justify-content: flex-end;
}
.buttons.is-right:not(.has-addons) .button:not(.is-fullwidth) {
  margin-left: 0.25rem;
  margin-right: 0.25rem;
}
@media screen and (max-width: 768px) {
  .button.is-responsive.is-small {
    font-size: 0.5625rem;
  }
  .button.is-responsive,
  .button.is-responsive.is-normal {
    font-size: 0.65625rem;
  }
  .button.is-responsive.is-medium {
    font-size: 0.75rem;
  }
  .button.is-responsive.is-large {
    font-size: 1rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .button.is-responsive.is-small {
    font-size: 0.65625rem;
  }
  .button.is-responsive,
  .button.is-responsive.is-normal {
    font-size: 0.75rem;
  }
  .button.is-responsive.is-medium {
    font-size: 1rem;
  }
  .button.is-responsive.is-large {
    font-size: 1.25rem;
  }
}
.container {
  flex-grow: 1;
  margin: 0 auto;
  position: relative;
  width: auto;
}
.container.is-fluid {
  max-width: none !important;
  padding-left: 32px;
  padding-right: 32px;
  width: 100%;
}
@media screen and (min-width: 1024px) {
  .container {
    max-width: 960px;
  }
}
@media screen and (max-width: 1215px) {
  .container.is-widescreen:not(.is-max-desktop) {
    max-width: 1152px;
  }
}
@media screen and (max-width: 1407px) {
  .container.is-fullhd:not(.is-max-desktop):not(.is-max-widescreen) {
    max-width: 1344px;
  }
}
@media screen and (min-width: 1216px) {
  .container:not(.is-max-desktop) {
    max-width: 1152px;
  }
}
@media screen and (min-width: 1408px) {
  .container:not(.is-max-desktop):not(.is-max-widescreen) {
    max-width: 1344px;
  }
}
.content li + li {
  margin-top: 0.25em;
}
.content p:not(:last-child),
.content dl:not(:last-child),
.content ol:not(:last-child),
.content ul:not(:last-child),
.content blockquote:not(:last-child),
.content pre:not(:last-child),
.content table:not(:last-child) {
  margin-bottom: 1em;
}
.content h1,
.content h2,
.content h3,
.content h4,
.content h5,
.content h6 {
  color: hsl(0deg, 0%, 21%);
  font-weight: 600;
  line-height: 1.125;
}
.content h1 {
  font-size: 2em;
  margin-bottom: 0.5em;
}
.content h1:not(:first-child) {
  margin-top: 1em;
}
.content h2 {
  font-size: 1.75em;
  margin-bottom: 0.5714em;
}
.content h2:not(:first-child) {
  margin-top: 1.1428em;
}
.content h3 {
  font-size: 1.5em;
  margin-bottom: 0.6666em;
}
.content h3:not(:first-child) {
  margin-top: 1.3333em;
}
.content h4 {
  font-size: 1.25em;
  margin-bottom: 0.8em;
}
.content h5 {
  font-size: 1.125em;
  margin-bottom: 0.8888em;
}
.content h6 {
  font-size: 1em;
  margin-bottom: 1em;
}
.content blockquote {
  background-color: hsl(0deg, 0%, 96%);
  border-left: 5px solid hsl(0deg, 0%, 86%);
  padding: 1.25em 1.5em;
}
.content ol {
  list-style-position: outside;
  margin-left: 2em;
  margin-top: 1em;
}
.content ol:not([type]) {
  list-style-type: decimal;
}
.content ol:not([type]).is-lower-alpha {
  list-style-type: lower-alpha;
}
.content ol:not([type]).is-lower-roman {
  list-style-type: lower-roman;
}
.content ol:not([type]).is-upper-alpha {
  list-style-type: upper-alpha;
}
.content ol:not([type]).is-upper-roman {
  list-style-type: upper-roman;
}
.content ul {
  list-style: disc outside;
  margin-left: 2em;
  margin-top: 1em;
}
.content ul ul {
  list-style-type: circle;
  margin-top: 0.5em;
}
.content ul ul ul {
  list-style-type: square;
}
.content dd {
  margin-left: 2em;
}
.content figure {
  margin-left: 2em;
  margin-right: 2em;
  text-align: center;
}
.content figure:not(:first-child) {
  margin-top: 2em;
}
.content figure:not(:last-child) {
  margin-bottom: 2em;
}
.content figure img {
  display: inline-block;
}
.content figure figcaption {
  font-style: italic;
}
.content pre {
  -webkit-overflow-scrolling: touch;
  overflow-x: auto;
  padding: 1.25em 1.5em;
  white-space: pre;
  word-wrap: normal;
}
.content sup,
.content sub {
  font-size: 75%;
}
.content table {
  width: 100%;
}
.content table td,
.content table th {
  border: 1px solid hsl(0deg, 0%, 86%);
  border-width: 0 0 1px;
  padding: 0.5em 0.75em;
  vertical-align: top;
}
.content table th {
  color: hsl(0deg, 0%, 21%);
}
.content table th:not([align]) {
  text-align: inherit;
}
.content table thead td,
.content table thead th {
  border-width: 0 0 2px;
  color: hsl(0deg, 0%, 21%);
}
.content table tfoot td,
.content table tfoot th {
  border-width: 2px 0 0;
  color: hsl(0deg, 0%, 21%);
}
.content table tbody tr:last-child td,
.content table tbody tr:last-child th {
  border-bottom-width: 0;
}
.content .tabs li + li {
  margin-top: 0;
}
.content.is-small {
  font-size: 0.75rem;
}
.content.is-normal {
  font-size: 1rem;
}
.content.is-medium {
  font-size: 1.25rem;
}
.content.is-large {
  font-size: 1.5rem;
}
.icon {
  align-items: center;
  display: inline-flex;
  justify-content: center;
  height: 1.5rem;
  width: 1.5rem;
}
.icon.is-small {
  height: 1rem;
  width: 1rem;
}
.icon.is-medium {
  height: 2rem;
  width: 2rem;
}
.icon.is-large {
  height: 3rem;
  width: 3rem;
}
.icon-text {
  align-items: flex-start;
  color: inherit;
  display: inline-flex;
  flex-wrap: wrap;
  line-height: 1.5rem;
  vertical-align: top;
}
.icon-text .icon {
  flex-grow: 0;
  flex-shrink: 0;
}
.icon-text .icon:not(:last-child) {
  margin-right: 0.25em;
}
.icon-text .icon:not(:first-child) {
  margin-left: 0.25em;
}
div.icon-text {
  display: flex;
}
.image {
  display: block;
  position: relative;
}
.image img {
  display: block;
  height: auto;
  width: 100%;
}
.image img.is-rounded {
  border-radius: 9999px;
}
.image.is-fullwidth {
  width: 100%;
}
.image.is-square img,
.image.is-square .has-ratio,
.image.is-1by1 img,
.image.is-1by1 .has-ratio,
.image.is-5by4 img,
.image.is-5by4 .has-ratio,
.image.is-4by3 img,
.image.is-4by3 .has-ratio,
.image.is-3by2 img,
.image.is-3by2 .has-ratio,
.image.is-5by3 img,
.image.is-5by3 .has-ratio,
.image.is-16by9 img,
.image.is-16by9 .has-ratio,
.image.is-2by1 img,
.image.is-2by1 .has-ratio,
.image.is-3by1 img,
.image.is-3by1 .has-ratio,
.image.is-4by5 img,
.image.is-4by5 .has-ratio,
.image.is-3by4 img,
.image.is-3by4 .has-ratio,
.image.is-2by3 img,
.image.is-2by3 .has-ratio,
.image.is-3by5 img,
.image.is-3by5 .has-ratio,
.image.is-9by16 img,
.image.is-9by16 .has-ratio,
.image.is-1by2 img,
.image.is-1by2 .has-ratio,
.image.is-1by3 img,
.image.is-1by3 .has-ratio {
  height: 100%;
  width: 100%;
}
.image.is-square,
.image.is-1by1 {
  padding-top: 100%;
}
.image.is-5by4 {
  padding-top: 80%;
}
.image.is-4by3 {
  padding-top: 75%;
}
.image.is-3by2 {
  padding-top: 66.6666%;
}
.image.is-5by3 {
  padding-top: 60%;
}
.image.is-16by9 {
  padding-top: 56.25%;
}
.image.is-2by1 {
  padding-top: 50%;
}
.image.is-3by1 {
  padding-top: 33.3333%;
}
.image.is-4by5 {
  padding-top: 125%;
}
.image.is-3by4 {
  padding-top: 133.3333%;
}
.image.is-2by3 {
  padding-top: 150%;
}
.image.is-3by5 {
  padding-top: 166.6666%;
}
.image.is-9by16 {
  padding-top: 177.7777%;
}
.image.is-1by2 {
  padding-top: 200%;
}
.image.is-1by3 {
  padding-top: 300%;
}
.image.is-16x16 {
  height: 16px;
  width: 16px;
}
.image.is-24x24 {
  height: 24px;
  width: 24px;
}
.image.is-32x32 {
  height: 32px;
  width: 32px;
}
.image.is-48x48 {
  height: 48px;
  width: 48px;
}
.image.is-64x64 {
  height: 64px;
  width: 64px;
}
.image.is-96x96 {
  height: 96px;
  width: 96px;
}
.image.is-128x128 {
  height: 128px;
  width: 128px;
}
.notification {
  background-color: hsl(0deg, 0%, 96%);
  border-radius: 4px;
  position: relative;
  padding: 1.25rem 2.5rem 1.25rem 1.5rem;
}
.notification a:not(.button):not(.dropdown-item) {
  color: currentColor;
  text-decoration: underline;
}
.notification strong {
  color: currentColor;
}
.notification code,
.notification pre {
  background: hsl(0deg, 0%, 100%);
}
.notification pre code {
  background: transparent;
}
.notification > .delete {
  right: 0.5rem;
  position: absolute;
  top: 0.5rem;
}
.notification .title,
.notification .subtitle,
.notification .content {
  color: currentColor;
}
.notification.is-white {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.notification.is-black {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.notification.is-light {
  background-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.notification.is-dark {
  background-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.notification.is-primary {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.notification.is-primary.is-light {
  background-color: #ebfffc;
  color: #00947e;
}
.notification.is-link {
  background-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.notification.is-link.is-light {
  background-color: #eff1fa;
  color: #3850b7;
}
.notification.is-info {
  background-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.notification.is-info.is-light {
  background-color: #eff5fb;
  color: #296fa8;
}
.notification.is-success {
  background-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.notification.is-success.is-light {
  background-color: #effaf5;
  color: #257953;
}
.notification.is-warning {
  background-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.notification.is-warning.is-light {
  background-color: #fffaeb;
  color: #946c00;
}
.notification.is-danger {
  background-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.notification.is-danger.is-light {
  background-color: #feecf0;
  color: #cc0f35;
}
.progress {
  -moz-appearance: none;
  -webkit-appearance: none;
  border: none;
  border-radius: 9999px;
  display: block;
  height: 1rem;
  overflow: hidden;
  padding: 0;
  width: 100%;
}
.progress::-webkit-progress-bar {
  background-color: hsl(0deg, 0%, 86%);
}
.progress::-webkit-progress-value {
  background-color: hsl(0deg, 0%, 29%);
}
.progress::-moz-progress-bar {
  background-color: hsl(0deg, 0%, 29%);
}
.progress::-ms-fill {
  background-color: hsl(0deg, 0%, 29%);
  border: none;
}
.progress.is-white::-webkit-progress-value {
  background-color: hsl(0deg, 0%, 100%);
}
.progress.is-white::-moz-progress-bar {
  background-color: hsl(0deg, 0%, 100%);
}
.progress.is-white::-ms-fill {
  background-color: hsl(0deg, 0%, 100%);
}
.progress.is-white:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(0deg, 0%, 100%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-black::-webkit-progress-value {
  background-color: hsl(0deg, 0%, 4%);
}
.progress.is-black::-moz-progress-bar {
  background-color: hsl(0deg, 0%, 4%);
}
.progress.is-black::-ms-fill {
  background-color: hsl(0deg, 0%, 4%);
}
.progress.is-black:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(0deg, 0%, 4%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-light::-webkit-progress-value {
  background-color: hsl(0deg, 0%, 96%);
}
.progress.is-light::-moz-progress-bar {
  background-color: hsl(0deg, 0%, 96%);
}
.progress.is-light::-ms-fill {
  background-color: hsl(0deg, 0%, 96%);
}
.progress.is-light:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(0deg, 0%, 96%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-dark::-webkit-progress-value {
  background-color: hsl(0deg, 0%, 21%);
}
.progress.is-dark::-moz-progress-bar {
  background-color: hsl(0deg, 0%, 21%);
}
.progress.is-dark::-ms-fill {
  background-color: hsl(0deg, 0%, 21%);
}
.progress.is-dark:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(0deg, 0%, 21%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-primary::-webkit-progress-value {
  background-color: hsl(171deg, 100%, 41%);
}
.progress.is-primary::-moz-progress-bar {
  background-color: hsl(171deg, 100%, 41%);
}
.progress.is-primary::-ms-fill {
  background-color: hsl(171deg, 100%, 41%);
}
.progress.is-primary:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(171deg, 100%, 41%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-link::-webkit-progress-value {
  background-color: hsl(229deg, 53%, 53%);
}
.progress.is-link::-moz-progress-bar {
  background-color: hsl(229deg, 53%, 53%);
}
.progress.is-link::-ms-fill {
  background-color: hsl(229deg, 53%, 53%);
}
.progress.is-link:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(229deg, 53%, 53%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-info::-webkit-progress-value {
  background-color: hsl(207deg, 61%, 53%);
}
.progress.is-info::-moz-progress-bar {
  background-color: hsl(207deg, 61%, 53%);
}
.progress.is-info::-ms-fill {
  background-color: hsl(207deg, 61%, 53%);
}
.progress.is-info:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(207deg, 61%, 53%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-success::-webkit-progress-value {
  background-color: hsl(153deg, 53%, 53%);
}
.progress.is-success::-moz-progress-bar {
  background-color: hsl(153deg, 53%, 53%);
}
.progress.is-success::-ms-fill {
  background-color: hsl(153deg, 53%, 53%);
}
.progress.is-success:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(153deg, 53%, 53%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-warning::-webkit-progress-value {
  background-color: hsl(44deg, 100%, 77%);
}
.progress.is-warning::-moz-progress-bar {
  background-color: hsl(44deg, 100%, 77%);
}
.progress.is-warning::-ms-fill {
  background-color: hsl(44deg, 100%, 77%);
}
.progress.is-warning:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(44deg, 100%, 77%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress.is-danger::-webkit-progress-value {
  background-color: hsl(348deg, 86%, 61%);
}
.progress.is-danger::-moz-progress-bar {
  background-color: hsl(348deg, 86%, 61%);
}
.progress.is-danger::-ms-fill {
  background-color: hsl(348deg, 86%, 61%);
}
.progress.is-danger:indeterminate {
  background-image:
    linear-gradient(
      to right,
      hsl(348deg, 86%, 61%) 30%,
      hsl(0deg, 0%, 86%) 30%);
}
.progress:indeterminate {
  animation-duration: 1.5s;
  animation-iteration-count: infinite;
  animation-name: moveIndeterminate;
  animation-timing-function: linear;
  background-color: hsl(0deg, 0%, 86%);
  background-image:
    linear-gradient(
      to right,
      hsl(0deg, 0%, 29%) 30%,
      hsl(0deg, 0%, 86%) 30%);
  background-position: top left;
  background-repeat: no-repeat;
  background-size: 150% 150%;
}
.progress:indeterminate::-webkit-progress-bar {
  background-color: transparent;
}
.progress:indeterminate::-moz-progress-bar {
  background-color: transparent;
}
.progress:indeterminate::-ms-fill {
  animation-name: none;
}
.progress.is-small {
  height: 0.75rem;
}
.progress.is-medium {
  height: 1.25rem;
}
.progress.is-large {
  height: 1.5rem;
}
@keyframes moveIndeterminate {
  from {
    background-position: 200% 0;
  }
  to {
    background-position: -200% 0;
  }
}
.table {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 21%);
}
.table td,
.table th {
  border: 1px solid hsl(0deg, 0%, 98%);
  border-width: 0 0 1px;
  padding: 0.5em 0.75em;
  vertical-align: top;
}
.table td.is-white,
.table th.is-white {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.table td.is-black,
.table th.is-black {
  background-color: hsl(0deg, 0%, 4%);
  border-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.table td.is-light,
.table th.is-light {
  background-color: hsl(0deg, 0%, 96%);
  border-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.table td.is-dark,
.table th.is-dark {
  background-color: hsl(0deg, 0%, 21%);
  border-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.table td.is-primary,
.table th.is-primary {
  background-color: hsl(171deg, 100%, 41%);
  border-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.table td.is-link,
.table th.is-link {
  background-color: hsl(229deg, 53%, 53%);
  border-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.table td.is-info,
.table th.is-info {
  background-color: hsl(207deg, 61%, 53%);
  border-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.table td.is-success,
.table th.is-success {
  background-color: hsl(153deg, 53%, 53%);
  border-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.table td.is-warning,
.table th.is-warning {
  background-color: hsl(44deg, 100%, 77%);
  border-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.table td.is-danger,
.table th.is-danger {
  background-color: hsl(348deg, 86%, 61%);
  border-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.table td.is-narrow,
.table th.is-narrow {
  white-space: nowrap;
  width: 1%;
}
.table td.is-selected,
.table th.is-selected {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.table td.is-selected a,
.table td.is-selected strong,
.table th.is-selected a,
.table th.is-selected strong {
  color: currentColor;
}
.table td.is-vcentered,
.table th.is-vcentered {
  vertical-align: middle;
}
.table th {
  color: hsl(0deg, 0%, 21%);
}
.table th:not([align]) {
  text-align: left;
}
.table tr.is-selected {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.table tr.is-selected a,
.table tr.is-selected strong {
  color: currentColor;
}
.table tr.is-selected td,
.table tr.is-selected th {
  border-color: #fff;
  color: currentColor;
}
.table thead {
  background-color: transparent;
}
.table thead td,
.table thead th {
  border-width: 0 0 2px;
  color: hsl(0deg, 0%, 21%);
}
.table tfoot {
  background-color: transparent;
}
.table tfoot td,
.table tfoot th {
  border-width: 2px 0 0;
  color: hsl(0deg, 0%, 21%);
}
.table tbody {
  background-color: transparent;
}
.table tbody tr:last-child td,
.table tbody tr:last-child th {
  border-bottom-width: 0;
}
.table.is-bordered td,
.table.is-bordered th {
  border-width: 1px;
}
.table.is-bordered tr:last-child td,
.table.is-bordered tr:last-child th {
  border-bottom-width: 1px;
}
.table.is-fullwidth {
  width: 100%;
}
.table.is-hoverable tbody tr:not(.is-selected):hover {
  background-color: hsl(0deg, 0%, 80%);
}
.table.is-hoverable.is-striped tbody tr:not(.is-selected):hover {
  background-color: hsl(0deg, 0%, 80%);
}
.table.is-hoverable.is-striped tbody tr:not(.is-selected):hover:nth-child(even) {
  background-color: hsl(0deg, 0%, 96%);
}
.table.is-narrow td,
.table.is-narrow th {
  padding: 0.25em 0.5em;
}
.table.is-striped tbody tr:not(.is-selected):nth-child(even) {
  background-color: hsl(0deg, 0%, 98%);
}
.table-container {
  -webkit-overflow-scrolling: touch;
  overflow: auto;
  overflow-y: hidden;
  max-width: 100%;
}
.tags {
  align-items: center;
  display: flex;
  flex-wrap: wrap;
  justify-content: flex-start;
}
.tags .tag {
  margin-bottom: 0.5rem;
}
.tags .tag:not(:last-child) {
  margin-right: 0.5rem;
}
.tags:last-child {
  margin-bottom: -0.5rem;
}
.tags:not(:last-child) {
  margin-bottom: 1rem;
}
.tags.are-medium .tag:not(.is-normal):not(.is-large) {
  font-size: 1rem;
}
.tags.are-large .tag:not(.is-normal):not(.is-medium) {
  font-size: 1.25rem;
}
.tags.is-centered {
  justify-content: center;
}
.tags.is-centered .tag {
  margin-right: 0.25rem;
  margin-left: 0.25rem;
}
.tags.is-right {
  justify-content: flex-end;
}
.tags.is-right .tag:not(:first-child) {
  margin-left: 0.5rem;
}
.tags.is-right .tag:not(:last-child) {
  margin-right: 0;
}
.tags.has-addons .tag {
  margin-right: 0;
}
.tags.has-addons .tag:not(:first-child) {
  margin-left: 0;
  border-top-left-radius: 0;
  border-bottom-left-radius: 0;
}
.tags.has-addons .tag:not(:last-child) {
  border-top-right-radius: 0;
  border-bottom-right-radius: 0;
}
.tag:not(body) {
  align-items: center;
  background-color: hsl(0deg, 0%, 96%);
  border-radius: 4px;
  color: hsl(0deg, 0%, 29%);
  display: inline-flex;
  font-size: 0.75rem;
  height: 2em;
  justify-content: center;
  line-height: 1.5;
  padding-left: 0.75em;
  padding-right: 0.75em;
  white-space: nowrap;
}
.tag:not(body) .delete {
  margin-left: 0.25rem;
  margin-right: -0.375rem;
}
.tag:not(body).is-white {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.tag:not(body).is-black {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.tag:not(body).is-light {
  background-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.tag:not(body).is-dark {
  background-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.tag:not(body).is-primary {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.tag:not(body).is-primary.is-light {
  background-color: #ebfffc;
  color: #00947e;
}
.tag:not(body).is-link {
  background-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.tag:not(body).is-link.is-light {
  background-color: #eff1fa;
  color: #3850b7;
}
.tag:not(body).is-info {
  background-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.tag:not(body).is-info.is-light {
  background-color: #eff5fb;
  color: #296fa8;
}
.tag:not(body).is-success {
  background-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.tag:not(body).is-success.is-light {
  background-color: #effaf5;
  color: #257953;
}
.tag:not(body).is-warning {
  background-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.tag:not(body).is-warning.is-light {
  background-color: #fffaeb;
  color: #946c00;
}
.tag:not(body).is-danger {
  background-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.tag:not(body).is-danger.is-light {
  background-color: #feecf0;
  color: #cc0f35;
}
.tag:not(body).is-normal {
  font-size: 0.75rem;
}
.tag:not(body).is-medium {
  font-size: 1rem;
}
.tag:not(body).is-large {
  font-size: 1.25rem;
}
.tag:not(body) .icon:first-child:not(:last-child) {
  margin-left: -0.375em;
  margin-right: 0.1875em;
}
.tag:not(body) .icon:last-child:not(:first-child) {
  margin-left: 0.1875em;
  margin-right: -0.375em;
}
.tag:not(body) .icon:first-child:last-child {
  margin-left: -0.375em;
  margin-right: -0.375em;
}
.tag:not(body).is-delete {
  margin-left: 1px;
  padding: 0;
  position: relative;
  width: 2em;
}
.tag:not(body).is-delete::before,
.tag:not(body).is-delete::after {
  background-color: currentColor;
  content: "";
  display: block;
  left: 50%;
  position: absolute;
  top: 50%;
  transform: translateX(-50%) translateY(-50%) rotate(45deg);
  transform-origin: center center;
}
.tag:not(body).is-delete::before {
  height: 1px;
  width: 50%;
}
.tag:not(body).is-delete::after {
  height: 50%;
  width: 1px;
}
.tag:not(body).is-delete:hover,
.tag:not(body).is-delete:focus {
  background-color: #e8e8e8;
}
.tag:not(body).is-delete:active {
  background-color: #dbdbdb;
}
.tag:not(body).is-rounded {
  border-radius: 9999px;
}
a.tag:hover {
  text-decoration: underline;
}
.title,
.subtitle {
  word-break: break-word;
}
.title em,
.title span,
.subtitle em,
.subtitle span {
  font-weight: inherit;
}
.title sub,
.subtitle sub {
  font-size: 0.75em;
}
.title sup,
.subtitle sup {
  font-size: 0.75em;
}
.title .tag,
.subtitle .tag {
  vertical-align: middle;
}
.title {
  color: hsl(0deg, 0%, 21%);
  font-size: 2rem;
  font-weight: 600;
  line-height: 1.125;
}
.title strong {
  color: inherit;
  font-weight: inherit;
}
.title:not(.is-spaced) + .subtitle {
  margin-top: -1.25rem;
}
.title.is-1 {
  font-size: 3rem;
}
.title.is-2 {
  font-size: 2.5rem;
}
.title.is-3 {
  font-size: 2rem;
}
.title.is-4 {
  font-size: 1.5rem;
}
.title.is-5 {
  font-size: 1.25rem;
}
.title.is-6 {
  font-size: 1rem;
}
.title.is-7 {
  font-size: 0.75rem;
}
.subtitle {
  color: hsl(0deg, 0%, 29%);
  font-size: 1.25rem;
  font-weight: 400;
  line-height: 1.25;
}
.subtitle strong {
  color: hsl(0deg, 0%, 21%);
  font-weight: 600;
}
.subtitle:not(.is-spaced) + .title {
  margin-top: -1.25rem;
}
.subtitle.is-1 {
  font-size: 3rem;
}
.subtitle.is-2 {
  font-size: 2.5rem;
}
.subtitle.is-3 {
  font-size: 2rem;
}
.subtitle.is-4 {
  font-size: 1.5rem;
}
.subtitle.is-5 {
  font-size: 1.25rem;
}
.subtitle.is-6 {
  font-size: 1rem;
}
.subtitle.is-7 {
  font-size: 0.75rem;
}
.heading {
  display: block;
  font-size: 11px;
  letter-spacing: 1px;
  margin-bottom: 5px;
  text-transform: uppercase;
}
.number {
  align-items: center;
  background-color: hsl(0deg, 0%, 96%);
  border-radius: 9999px;
  display: inline-flex;
  font-size: 1.25rem;
  height: 2em;
  justify-content: center;
  margin-right: 1.5rem;
  min-width: 2.5em;
  padding: 0.25rem 0.5rem;
  text-align: center;
  vertical-align: top;
}
.select select,
.textarea,
.input {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 86%);
  border-radius: 4px;
  color: hsl(0deg, 0%, 21%);
}
.select select::-moz-placeholder,
.textarea::-moz-placeholder,
.input::-moz-placeholder {
  color: rgba(54, 54, 54, 0.3);
}
.select select::-webkit-input-placeholder,
.textarea::-webkit-input-placeholder,
.input::-webkit-input-placeholder {
  color: rgba(54, 54, 54, 0.3);
}
.select select:-moz-placeholder,
.textarea:-moz-placeholder,
.input:-moz-placeholder {
  color: rgba(54, 54, 54, 0.3);
}
.select select:-ms-input-placeholder,
.textarea:-ms-input-placeholder,
.input:-ms-input-placeholder {
  color: rgba(54, 54, 54, 0.3);
}
.select select:hover,
.textarea:hover,
.input:hover,
.select select.is-hovered,
.is-hovered.textarea,
.is-hovered.input {
  border-color: hsl(0deg, 0%, 71%);
}
.select select:focus,
.textarea:focus,
.input:focus,
.select select.is-focused,
.is-focused.textarea,
.is-focused.input,
.select select:active,
.textarea:active,
.input:active,
.select select.is-active,
.is-active.textarea,
.is-active.input {
  border-color: hsl(229deg, 53%, 53%);
  box-shadow: 0 0 0 0.125em rgba(72, 95, 199, 0.25);
}
.select select[disabled],
[disabled].textarea,
[disabled].input,
fieldset[disabled] .select select,
.select fieldset[disabled] select,
fieldset[disabled] .textarea,
fieldset[disabled] .input {
  background-color: hsl(0deg, 0%, 96%);
  border-color: hsl(0deg, 0%, 86%);
  box-shadow: none;
  color: hsl(0deg, 0%, 48%);
}
.select select[disabled]::-moz-placeholder,
[disabled].textarea::-moz-placeholder,
[disabled].input::-moz-placeholder,
fieldset[disabled] .select select::-moz-placeholder,
.select fieldset[disabled] select::-moz-placeholder,
fieldset[disabled] .textarea::-moz-placeholder,
fieldset[disabled] .input::-moz-placeholder {
  color: rgba(122, 122, 122, 0.3);
}
.select select[disabled]::-webkit-input-placeholder,
[disabled].textarea::-webkit-input-placeholder,
[disabled].input::-webkit-input-placeholder,
fieldset[disabled] .select select::-webkit-input-placeholder,
.select fieldset[disabled] select::-webkit-input-placeholder,
fieldset[disabled] .textarea::-webkit-input-placeholder,
fieldset[disabled] .input::-webkit-input-placeholder {
  color: rgba(122, 122, 122, 0.3);
}
.select select[disabled]:-moz-placeholder,
[disabled].textarea:-moz-placeholder,
[disabled].input:-moz-placeholder,
fieldset[disabled] .select select:-moz-placeholder,
.select fieldset[disabled] select:-moz-placeholder,
fieldset[disabled] .textarea:-moz-placeholder,
fieldset[disabled] .input:-moz-placeholder {
  color: rgba(122, 122, 122, 0.3);
}
.select select[disabled]:-ms-input-placeholder,
[disabled].textarea:-ms-input-placeholder,
[disabled].input:-ms-input-placeholder,
fieldset[disabled] .select select:-ms-input-placeholder,
.select fieldset[disabled] select:-ms-input-placeholder,
fieldset[disabled] .textarea:-ms-input-placeholder,
fieldset[disabled] .input:-ms-input-placeholder {
  color: rgba(122, 122, 122, 0.3);
}
.textarea,
.input {
  box-shadow: inset 0 0.0625em 0.125em rgba(10, 10, 10, 0.05);
  max-width: 100%;
  width: 100%;
}
[readonly].textarea,
[readonly].input {
  box-shadow: none;
}
.is-white.textarea,
.is-white.input {
  border-color: hsl(0deg, 0%, 100%);
}
.is-white.textarea:focus,
.is-white.input:focus,
.is-white.is-focused.textarea,
.is-white.is-focused.input,
.is-white.textarea:active,
.is-white.input:active,
.is-white.is-active.textarea,
.is-white.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(255, 255, 255, 0.25);
}
.is-black.textarea,
.is-black.input {
  border-color: hsl(0deg, 0%, 4%);
}
.is-black.textarea:focus,
.is-black.input:focus,
.is-black.is-focused.textarea,
.is-black.is-focused.input,
.is-black.textarea:active,
.is-black.input:active,
.is-black.is-active.textarea,
.is-black.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(10, 10, 10, 0.25);
}
.is-light.textarea,
.is-light.input {
  border-color: hsl(0deg, 0%, 96%);
}
.is-light.textarea:focus,
.is-light.input:focus,
.is-light.is-focused.textarea,
.is-light.is-focused.input,
.is-light.textarea:active,
.is-light.input:active,
.is-light.is-active.textarea,
.is-light.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(245, 245, 245, 0.25);
}
.is-dark.textarea,
.is-dark.input {
  border-color: hsl(0deg, 0%, 21%);
}
.is-dark.textarea:focus,
.is-dark.input:focus,
.is-dark.is-focused.textarea,
.is-dark.is-focused.input,
.is-dark.textarea:active,
.is-dark.input:active,
.is-dark.is-active.textarea,
.is-dark.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(54, 54, 54, 0.25);
}
.is-primary.textarea,
.is-primary.input {
  border-color: hsl(171deg, 100%, 41%);
}
.is-primary.textarea:focus,
.is-primary.input:focus,
.is-primary.is-focused.textarea,
.is-primary.is-focused.input,
.is-primary.textarea:active,
.is-primary.input:active,
.is-primary.is-active.textarea,
.is-primary.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(0, 209, 178, 0.25);
}
.is-link.textarea,
.is-link.input {
  border-color: hsl(229deg, 53%, 53%);
}
.is-link.textarea:focus,
.is-link.input:focus,
.is-link.is-focused.textarea,
.is-link.is-focused.input,
.is-link.textarea:active,
.is-link.input:active,
.is-link.is-active.textarea,
.is-link.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(72, 95, 199, 0.25);
}
.is-info.textarea,
.is-info.input {
  border-color: hsl(207deg, 61%, 53%);
}
.is-info.textarea:focus,
.is-info.input:focus,
.is-info.is-focused.textarea,
.is-info.is-focused.input,
.is-info.textarea:active,
.is-info.input:active,
.is-info.is-active.textarea,
.is-info.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(62, 142, 208, 0.25);
}
.is-success.textarea,
.is-success.input {
  border-color: hsl(153deg, 53%, 53%);
}
.is-success.textarea:focus,
.is-success.input:focus,
.is-success.is-focused.textarea,
.is-success.is-focused.input,
.is-success.textarea:active,
.is-success.input:active,
.is-success.is-active.textarea,
.is-success.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(72, 199, 142, 0.25);
}
.is-warning.textarea,
.is-warning.input {
  border-color: hsl(44deg, 100%, 77%);
}
.is-warning.textarea:focus,
.is-warning.input:focus,
.is-warning.is-focused.textarea,
.is-warning.is-focused.input,
.is-warning.textarea:active,
.is-warning.input:active,
.is-warning.is-active.textarea,
.is-warning.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(255, 224, 138, 0.25);
}
.is-danger.textarea,
.is-danger.input {
  border-color: hsl(348deg, 86%, 61%);
}
.is-danger.textarea:focus,
.is-danger.input:focus,
.is-danger.is-focused.textarea,
.is-danger.is-focused.input,
.is-danger.textarea:active,
.is-danger.input:active,
.is-danger.is-active.textarea,
.is-danger.is-active.input {
  box-shadow: 0 0 0 0.125em rgba(241, 70, 104, 0.25);
}
.is-small.textarea,
.is-small.input {
  border-radius: 2px;
  font-size: 0.75rem;
}
.is-medium.textarea,
.is-medium.input {
  font-size: 1.25rem;
}
.is-large.textarea,
.is-large.input {
  font-size: 1.5rem;
}
.is-fullwidth.textarea,
.is-fullwidth.input {
  display: block;
  width: 100%;
}
.is-inline.textarea,
.is-inline.input {
  display: inline;
  width: auto;
}
.input.is-rounded {
  border-radius: 9999px;
  padding-left: calc(calc(0.75em - 1px) + 0.375em);
  padding-right: calc(calc(0.75em - 1px) + 0.375em);
}
.input.is-static {
  background-color: transparent;
  border-color: transparent;
  box-shadow: none;
  padding-left: 0;
  padding-right: 0;
}
.textarea {
  display: block;
  max-width: 100%;
  min-width: 100%;
  padding: calc(0.75em - 1px);
  resize: vertical;
}
.textarea:not([rows]) {
  max-height: 40em;
  min-height: 8em;
}
.textarea[rows] {
  height: initial;
}
.textarea.has-fixed-size {
  resize: none;
}
.radio,
.checkbox {
  cursor: pointer;
  display: inline-block;
  line-height: 1.25;
  position: relative;
}
.radio input,
.checkbox input {
  cursor: pointer;
}
.radio:hover,
.checkbox:hover {
  color: hsl(0deg, 0%, 21%);
}
[disabled].radio,
[disabled].checkbox,
fieldset[disabled] .radio,
fieldset[disabled] .checkbox,
.radio input[disabled],
.checkbox input[disabled] {
  color: hsl(0deg, 0%, 48%);
  cursor: not-allowed;
}
.radio + .radio {
  margin-left: 0.5em;
}
.select {
  display: inline-block;
  max-width: 100%;
  position: relative;
  vertical-align: top;
}
.select:not(.is-multiple) {
  height: 2.5em;
}
.select:not(.is-multiple):not(.is-loading)::after {
  border-color: hsl(229deg, 53%, 53%);
  right: 1.125em;
  z-index: 4;
}
.select.is-rounded select {
  border-radius: 9999px;
  padding-left: 1em;
}
.select select {
  cursor: pointer;
  display: block;
  font-size: 1em;
  max-width: 100%;
  outline: none;
}
.select select::-ms-expand {
  display: none;
}
.select select[disabled]:hover,
fieldset[disabled] .select select:hover {
  border-color: hsl(0deg, 0%, 86%);
}
.select select:not([multiple]) {
  padding-right: 2.5em;
}
.select select[multiple] {
  height: auto;
  padding: 0;
}
.select select[multiple] option {
  padding: 0.5em 1em;
}
.select:not(.is-multiple):not(.is-loading):hover::after {
  border-color: hsl(0deg, 0%, 21%);
}
.select.is-white:not(:hover)::after {
  border-color: hsl(0deg, 0%, 100%);
}
.select.is-white select {
  border-color: hsl(0deg, 0%, 100%);
}
.select.is-white select:hover,
.select.is-white select.is-hovered {
  border-color: #f2f2f2;
}
.select.is-white select:focus,
.select.is-white select.is-focused,
.select.is-white select:active,
.select.is-white select.is-active {
  box-shadow: 0 0 0 0.125em rgba(255, 255, 255, 0.25);
}
.select.is-black:not(:hover)::after {
  border-color: hsl(0deg, 0%, 4%);
}
.select.is-black select {
  border-color: hsl(0deg, 0%, 4%);
}
.select.is-black select:hover,
.select.is-black select.is-hovered {
  border-color: black;
}
.select.is-black select:focus,
.select.is-black select.is-focused,
.select.is-black select:active,
.select.is-black select.is-active {
  box-shadow: 0 0 0 0.125em rgba(10, 10, 10, 0.25);
}
.select.is-light:not(:hover)::after {
  border-color: hsl(0deg, 0%, 96%);
}
.select.is-light select {
  border-color: hsl(0deg, 0%, 96%);
}
.select.is-light select:hover,
.select.is-light select.is-hovered {
  border-color: #e8e8e8;
}
.select.is-light select:focus,
.select.is-light select.is-focused,
.select.is-light select:active,
.select.is-light select.is-active {
  box-shadow: 0 0 0 0.125em rgba(245, 245, 245, 0.25);
}
.select.is-dark:not(:hover)::after {
  border-color: hsl(0deg, 0%, 21%);
}
.select.is-dark select {
  border-color: hsl(0deg, 0%, 21%);
}
.select.is-dark select:hover,
.select.is-dark select.is-hovered {
  border-color: #292929;
}
.select.is-dark select:focus,
.select.is-dark select.is-focused,
.select.is-dark select:active,
.select.is-dark select.is-active {
  box-shadow: 0 0 0 0.125em rgba(54, 54, 54, 0.25);
}
.select.is-primary:not(:hover)::after {
  border-color: hsl(171deg, 100%, 41%);
}
.select.is-primary select {
  border-color: hsl(171deg, 100%, 41%);
}
.select.is-primary select:hover,
.select.is-primary select.is-hovered {
  border-color: #00b89c;
}
.select.is-primary select:focus,
.select.is-primary select.is-focused,
.select.is-primary select:active,
.select.is-primary select.is-active {
  box-shadow: 0 0 0 0.125em rgba(0, 209, 178, 0.25);
}
.select.is-link:not(:hover)::after {
  border-color: hsl(229deg, 53%, 53%);
}
.select.is-link select {
  border-color: hsl(229deg, 53%, 53%);
}
.select.is-link select:hover,
.select.is-link select.is-hovered {
  border-color: #3a51bb;
}
.select.is-link select:focus,
.select.is-link select.is-focused,
.select.is-link select:active,
.select.is-link select.is-active {
  box-shadow: 0 0 0 0.125em rgba(72, 95, 199, 0.25);
}
.select.is-info:not(:hover)::after {
  border-color: hsl(207deg, 61%, 53%);
}
.select.is-info select {
  border-color: hsl(207deg, 61%, 53%);
}
.select.is-info select:hover,
.select.is-info select.is-hovered {
  border-color: #3082c5;
}
.select.is-info select:focus,
.select.is-info select.is-focused,
.select.is-info select:active,
.select.is-info select.is-active {
  box-shadow: 0 0 0 0.125em rgba(62, 142, 208, 0.25);
}
.select.is-success:not(:hover)::after {
  border-color: hsl(153deg, 53%, 53%);
}
.select.is-success select {
  border-color: hsl(153deg, 53%, 53%);
}
.select.is-success select:hover,
.select.is-success select.is-hovered {
  border-color: #3abb81;
}
.select.is-success select:focus,
.select.is-success select.is-focused,
.select.is-success select:active,
.select.is-success select.is-active {
  box-shadow: 0 0 0 0.125em rgba(72, 199, 142, 0.25);
}
.select.is-warning:not(:hover)::after {
  border-color: hsl(44deg, 100%, 77%);
}
.select.is-warning select {
  border-color: hsl(44deg, 100%, 77%);
}
.select.is-warning select:hover,
.select.is-warning select.is-hovered {
  border-color: #ffd970;
}
.select.is-warning select:focus,
.select.is-warning select.is-focused,
.select.is-warning select:active,
.select.is-warning select.is-active {
  box-shadow: 0 0 0 0.125em rgba(255, 224, 138, 0.25);
}
.select.is-danger:not(:hover)::after {
  border-color: hsl(348deg, 86%, 61%);
}
.select.is-danger select {
  border-color: hsl(348deg, 86%, 61%);
}
.select.is-danger select:hover,
.select.is-danger select.is-hovered {
  border-color: #ef2e55;
}
.select.is-danger select:focus,
.select.is-danger select.is-focused,
.select.is-danger select:active,
.select.is-danger select.is-active {
  box-shadow: 0 0 0 0.125em rgba(241, 70, 104, 0.25);
}
.select.is-small {
  border-radius: 2px;
  font-size: 0.75rem;
}
.select.is-medium {
  font-size: 1.25rem;
}
.select.is-large {
  font-size: 1.5rem;
}
.select.is-disabled::after {
  border-color: hsl(0deg, 0%, 48%) !important;
  opacity: 0.5;
}
.select.is-fullwidth {
  width: 100%;
}
.select.is-fullwidth select {
  width: 100%;
}
.select.is-loading::after {
  margin-top: 0;
  position: absolute;
  right: 0.625em;
  top: 0.625em;
  transform: none;
}
.select.is-loading.is-small:after {
  font-size: 0.75rem;
}
.select.is-loading.is-medium:after {
  font-size: 1.25rem;
}
.select.is-loading.is-large:after {
  font-size: 1.5rem;
}
.file {
  align-items: stretch;
  display: flex;
  justify-content: flex-start;
  position: relative;
}
.file.is-white .file-cta {
  background-color: hsl(0deg, 0%, 100%);
  border-color: transparent;
  color: hsl(0deg, 0%, 4%);
}
.file.is-white:hover .file-cta,
.file.is-white.is-hovered .file-cta {
  background-color: #f9f9f9;
  border-color: transparent;
  color: hsl(0deg, 0%, 4%);
}
.file.is-white:focus .file-cta,
.file.is-white.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(255, 255, 255, 0.25);
  color: hsl(0deg, 0%, 4%);
}
.file.is-white:active .file-cta,
.file.is-white.is-active .file-cta {
  background-color: #f2f2f2;
  border-color: transparent;
  color: hsl(0deg, 0%, 4%);
}
.file.is-black .file-cta {
  background-color: hsl(0deg, 0%, 4%);
  border-color: transparent;
  color: hsl(0deg, 0%, 100%);
}
.file.is-black:hover .file-cta,
.file.is-black.is-hovered .file-cta {
  background-color: #040404;
  border-color: transparent;
  color: hsl(0deg, 0%, 100%);
}
.file.is-black:focus .file-cta,
.file.is-black.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(10, 10, 10, 0.25);
  color: hsl(0deg, 0%, 100%);
}
.file.is-black:active .file-cta,
.file.is-black.is-active .file-cta {
  background-color: black;
  border-color: transparent;
  color: hsl(0deg, 0%, 100%);
}
.file.is-light .file-cta {
  background-color: hsl(0deg, 0%, 96%);
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.file.is-light:hover .file-cta,
.file.is-light.is-hovered .file-cta {
  background-color: #eeeeee;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.file.is-light:focus .file-cta,
.file.is-light.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(245, 245, 245, 0.25);
  color: rgba(0, 0, 0, 0.7);
}
.file.is-light:active .file-cta,
.file.is-light.is-active .file-cta {
  background-color: #e8e8e8;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.file.is-dark .file-cta {
  background-color: hsl(0deg, 0%, 21%);
  border-color: transparent;
  color: #fff;
}
.file.is-dark:hover .file-cta,
.file.is-dark.is-hovered .file-cta {
  background-color: #2f2f2f;
  border-color: transparent;
  color: #fff;
}
.file.is-dark:focus .file-cta,
.file.is-dark.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(54, 54, 54, 0.25);
  color: #fff;
}
.file.is-dark:active .file-cta,
.file.is-dark.is-active .file-cta {
  background-color: #292929;
  border-color: transparent;
  color: #fff;
}
.file.is-primary .file-cta {
  background-color: hsl(171deg, 100%, 41%);
  border-color: transparent;
  color: #fff;
}
.file.is-primary:hover .file-cta,
.file.is-primary.is-hovered .file-cta {
  background-color: #00c4a7;
  border-color: transparent;
  color: #fff;
}
.file.is-primary:focus .file-cta,
.file.is-primary.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(0, 209, 178, 0.25);
  color: #fff;
}
.file.is-primary:active .file-cta,
.file.is-primary.is-active .file-cta {
  background-color: #00b89c;
  border-color: transparent;
  color: #fff;
}
.file.is-link .file-cta {
  background-color: hsl(229deg, 53%, 53%);
  border-color: transparent;
  color: #fff;
}
.file.is-link:hover .file-cta,
.file.is-link.is-hovered .file-cta {
  background-color: #3e56c4;
  border-color: transparent;
  color: #fff;
}
.file.is-link:focus .file-cta,
.file.is-link.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(72, 95, 199, 0.25);
  color: #fff;
}
.file.is-link:active .file-cta,
.file.is-link.is-active .file-cta {
  background-color: #3a51bb;
  border-color: transparent;
  color: #fff;
}
.file.is-info .file-cta {
  background-color: hsl(207deg, 61%, 53%);
  border-color: transparent;
  color: #fff;
}
.file.is-info:hover .file-cta,
.file.is-info.is-hovered .file-cta {
  background-color: #3488ce;
  border-color: transparent;
  color: #fff;
}
.file.is-info:focus .file-cta,
.file.is-info.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(62, 142, 208, 0.25);
  color: #fff;
}
.file.is-info:active .file-cta,
.file.is-info.is-active .file-cta {
  background-color: #3082c5;
  border-color: transparent;
  color: #fff;
}
.file.is-success .file-cta {
  background-color: hsl(153deg, 53%, 53%);
  border-color: transparent;
  color: #fff;
}
.file.is-success:hover .file-cta,
.file.is-success.is-hovered .file-cta {
  background-color: #3ec487;
  border-color: transparent;
  color: #fff;
}
.file.is-success:focus .file-cta,
.file.is-success.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(72, 199, 142, 0.25);
  color: #fff;
}
.file.is-success:active .file-cta,
.file.is-success.is-active .file-cta {
  background-color: #3abb81;
  border-color: transparent;
  color: #fff;
}
.file.is-warning .file-cta {
  background-color: hsl(44deg, 100%, 77%);
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.file.is-warning:hover .file-cta,
.file.is-warning.is-hovered .file-cta {
  background-color: #ffdc7d;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.file.is-warning:focus .file-cta,
.file.is-warning.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(255, 224, 138, 0.25);
  color: rgba(0, 0, 0, 0.7);
}
.file.is-warning:active .file-cta,
.file.is-warning.is-active .file-cta {
  background-color: #ffd970;
  border-color: transparent;
  color: rgba(0, 0, 0, 0.7);
}
.file.is-danger .file-cta {
  background-color: hsl(348deg, 86%, 61%);
  border-color: transparent;
  color: #fff;
}
.file.is-danger:hover .file-cta,
.file.is-danger.is-hovered .file-cta {
  background-color: #f03a5f;
  border-color: transparent;
  color: #fff;
}
.file.is-danger:focus .file-cta,
.file.is-danger.is-focused .file-cta {
  border-color: transparent;
  box-shadow: 0 0 0.5em rgba(241, 70, 104, 0.25);
  color: #fff;
}
.file.is-danger:active .file-cta,
.file.is-danger.is-active .file-cta {
  background-color: #ef2e55;
  border-color: transparent;
  color: #fff;
}
.file.is-small {
  font-size: 0.75rem;
}
.file.is-normal {
  font-size: 1rem;
}
.file.is-medium {
  font-size: 1.25rem;
}
.file.is-medium .file-icon .fa {
  font-size: 21px;
}
.file.is-large {
  font-size: 1.5rem;
}
.file.is-large .file-icon .fa {
  font-size: 28px;
}
.file.has-name .file-cta {
  border-bottom-right-radius: 0;
  border-top-right-radius: 0;
}
.file.has-name .file-name {
  border-bottom-left-radius: 0;
  border-top-left-radius: 0;
}
.file.has-name.is-empty .file-cta {
  border-radius: 4px;
}
.file.has-name.is-empty .file-name {
  display: none;
}
.file.is-boxed .file-label {
  flex-direction: column;
}
.file.is-boxed .file-cta {
  flex-direction: column;
  height: auto;
  padding: 1em 3em;
}
.file.is-boxed .file-name {
  border-width: 0 1px 1px;
}
.file.is-boxed .file-icon {
  height: 1.5em;
  width: 1.5em;
}
.file.is-boxed .file-icon .fa {
  font-size: 21px;
}
.file.is-boxed.is-small .file-icon .fa {
  font-size: 14px;
}
.file.is-boxed.is-medium .file-icon .fa {
  font-size: 28px;
}
.file.is-boxed.is-large .file-icon .fa {
  font-size: 35px;
}
.file.is-boxed.has-name .file-cta {
  border-radius: 4px 4px 0 0;
}
.file.is-boxed.has-name .file-name {
  border-radius: 0 0 4px 4px;
  border-width: 0 1px 1px;
}
.file.is-centered {
  justify-content: center;
}
.file.is-fullwidth .file-label {
  width: 100%;
}
.file.is-fullwidth .file-name {
  flex-grow: 1;
  max-width: none;
}
.file.is-right {
  justify-content: flex-end;
}
.file.is-right .file-cta {
  border-radius: 0 4px 4px 0;
}
.file.is-right .file-name {
  border-radius: 4px 0 0 4px;
  border-width: 1px 0 1px 1px;
  order: -1;
}
.file-label {
  align-items: stretch;
  display: flex;
  cursor: pointer;
  justify-content: flex-start;
  overflow: hidden;
  position: relative;
}
.file-label:hover .file-cta {
  background-color: #eeeeee;
  color: hsl(0deg, 0%, 21%);
}
.file-label:hover .file-name {
  border-color: #d5d5d5;
}
.file-label:active .file-cta {
  background-color: #e8e8e8;
  color: hsl(0deg, 0%, 21%);
}
.file-label:active .file-name {
  border-color: #cfcfcf;
}
.file-input {
  height: 100%;
  left: 0;
  opacity: 0;
  outline: none;
  position: absolute;
  top: 0;
  width: 100%;
}
.file-cta,
.file-name {
  border-color: hsl(0deg, 0%, 86%);
  border-radius: 4px;
  font-size: 1em;
  padding-left: 1em;
  padding-right: 1em;
  white-space: nowrap;
}
.file-cta {
  background-color: hsl(0deg, 0%, 96%);
  color: hsl(0deg, 0%, 29%);
}
.file-name {
  border-color: hsl(0deg, 0%, 86%);
  border-style: solid;
  border-width: 1px 1px 1px 0;
  display: block;
  max-width: 16em;
  overflow: hidden;
  text-align: inherit;
  text-overflow: ellipsis;
}
.file-icon {
  align-items: center;
  display: flex;
  height: 1em;
  justify-content: center;
  margin-right: 0.5em;
  width: 1em;
}
.file-icon .fa {
  font-size: 14px;
}
.label {
  color: hsl(0deg, 0%, 21%);
  display: block;
  font-size: 1rem;
  font-weight: 700;
}
.label:not(:last-child) {
  margin-bottom: 0.5em;
}
.label.is-small {
  font-size: 0.75rem;
}
.label.is-medium {
  font-size: 1.25rem;
}
.label.is-large {
  font-size: 1.5rem;
}
.help {
  display: block;
  font-size: 0.75rem;
  margin-top: 0.25rem;
}
.help.is-white {
  color: hsl(0deg, 0%, 100%);
}
.help.is-black {
  color: hsl(0deg, 0%, 4%);
}
.help.is-light {
  color: hsl(0deg, 0%, 96%);
}
.help.is-dark {
  color: hsl(0deg, 0%, 21%);
}
.help.is-primary {
  color: hsl(171deg, 100%, 41%);
}
.help.is-link {
  color: hsl(229deg, 53%, 53%);
}
.help.is-info {
  color: hsl(207deg, 61%, 53%);
}
.help.is-success {
  color: hsl(153deg, 53%, 53%);
}
.help.is-warning {
  color: hsl(44deg, 100%, 77%);
}
.help.is-danger {
  color: hsl(348deg, 86%, 61%);
}
.field:not(:last-child) {
  margin-bottom: 0.75rem;
}
.field.has-addons {
  display: flex;
  justify-content: flex-start;
}
.field.has-addons .control:not(:last-child) {
  margin-right: -1px;
}
.field.has-addons .control:not(:first-child):not(:last-child) .button,
.field.has-addons .control:not(:first-child):not(:last-child) .input,
.field.has-addons .control:not(:first-child):not(:last-child) .select select {
  border-radius: 0;
}
.field.has-addons .control:first-child:not(:only-child) .button,
.field.has-addons .control:first-child:not(:only-child) .input,
.field.has-addons .control:first-child:not(:only-child) .select select {
  border-bottom-right-radius: 0;
  border-top-right-radius: 0;
}
.field.has-addons .control:last-child:not(:only-child) .button,
.field.has-addons .control:last-child:not(:only-child) .input,
.field.has-addons .control:last-child:not(:only-child) .select select {
  border-bottom-left-radius: 0;
  border-top-left-radius: 0;
}
.field.has-addons .control .button:not([disabled]):hover,
.field.has-addons .control .button:not([disabled]).is-hovered,
.field.has-addons .control .input:not([disabled]):hover,
.field.has-addons .control .input:not([disabled]).is-hovered,
.field.has-addons .control .select select:not([disabled]):hover,
.field.has-addons .control .select select:not([disabled]).is-hovered {
  z-index: 2;
}
.field.has-addons .control .button:not([disabled]):focus,
.field.has-addons .control .button:not([disabled]).is-focused,
.field.has-addons .control .button:not([disabled]):active,
.field.has-addons .control .button:not([disabled]).is-active,
.field.has-addons .control .input:not([disabled]):focus,
.field.has-addons .control .input:not([disabled]).is-focused,
.field.has-addons .control .input:not([disabled]):active,
.field.has-addons .control .input:not([disabled]).is-active,
.field.has-addons .control .select select:not([disabled]):focus,
.field.has-addons .control .select select:not([disabled]).is-focused,
.field.has-addons .control .select select:not([disabled]):active,
.field.has-addons .control .select select:not([disabled]).is-active {
  z-index: 3;
}
.field.has-addons .control .button:not([disabled]):focus:hover,
.field.has-addons .control .button:not([disabled]).is-focused:hover,
.field.has-addons .control .button:not([disabled]):active:hover,
.field.has-addons .control .button:not([disabled]).is-active:hover,
.field.has-addons .control .input:not([disabled]):focus:hover,
.field.has-addons .control .input:not([disabled]).is-focused:hover,
.field.has-addons .control .input:not([disabled]):active:hover,
.field.has-addons .control .input:not([disabled]).is-active:hover,
.field.has-addons .control .select select:not([disabled]):focus:hover,
.field.has-addons .control .select select:not([disabled]).is-focused:hover,
.field.has-addons .control .select select:not([disabled]):active:hover,
.field.has-addons .control .select select:not([disabled]).is-active:hover {
  z-index: 4;
}
.field.has-addons .control.is-expanded {
  flex-grow: 1;
  flex-shrink: 1;
}
.field.has-addons.has-addons-centered {
  justify-content: center;
}
.field.has-addons.has-addons-right {
  justify-content: flex-end;
}
.field.has-addons.has-addons-fullwidth .control {
  flex-grow: 1;
  flex-shrink: 0;
}
.field.is-grouped {
  display: flex;
  justify-content: flex-start;
}
.field.is-grouped > .control {
  flex-shrink: 0;
}
.field.is-grouped > .control:not(:last-child) {
  margin-bottom: 0;
  margin-right: 0.75rem;
}
.field.is-grouped > .control.is-expanded {
  flex-grow: 1;
  flex-shrink: 1;
}
.field.is-grouped.is-grouped-centered {
  justify-content: center;
}
.field.is-grouped.is-grouped-right {
  justify-content: flex-end;
}
.field.is-grouped.is-grouped-multiline {
  flex-wrap: wrap;
}
.field.is-grouped.is-grouped-multiline > .control:last-child,
.field.is-grouped.is-grouped-multiline > .control:not(:last-child) {
  margin-bottom: 0.75rem;
}
.field.is-grouped.is-grouped-multiline:last-child {
  margin-bottom: -0.75rem;
}
.field.is-grouped.is-grouped-multiline:not(:last-child) {
  margin-bottom: 0;
}
@media screen and (min-width: 769px), print {
  .field.is-horizontal {
    display: flex;
  }
}
.field-label .label {
  font-size: inherit;
}
@media screen and (max-width: 768px) {
  .field-label {
    margin-bottom: 0.5rem;
  }
}
@media screen and (min-width: 769px), print {
  .field-label {
    flex-basis: 0;
    flex-grow: 1;
    flex-shrink: 0;
    margin-right: 1.5rem;
    text-align: right;
  }
  .field-label.is-small {
    font-size: 0.75rem;
    padding-top: 0.375em;
  }
  .field-label.is-normal {
    padding-top: 0.375em;
  }
  .field-label.is-medium {
    font-size: 1.25rem;
    padding-top: 0.375em;
  }
  .field-label.is-large {
    font-size: 1.5rem;
    padding-top: 0.375em;
  }
}
.field-body .field .field {
  margin-bottom: 0;
}
@media screen and (min-width: 769px), print {
  .field-body {
    display: flex;
    flex-basis: 0;
    flex-grow: 5;
    flex-shrink: 1;
  }
  .field-body .field {
    margin-bottom: 0;
  }
  .field-body > .field {
    flex-shrink: 1;
  }
  .field-body > .field:not(.is-narrow) {
    flex-grow: 1;
  }
  .field-body > .field:not(:last-child) {
    margin-right: 0.75rem;
  }
}
.control {
  box-sizing: border-box;
  clear: both;
  font-size: 1rem;
  position: relative;
  text-align: inherit;
}
.control.has-icons-left .input:focus ~ .icon,
.control.has-icons-left .select:focus ~ .icon,
.control.has-icons-right .input:focus ~ .icon,
.control.has-icons-right .select:focus ~ .icon {
  color: hsl(0deg, 0%, 29%);
}
.control.has-icons-left .input.is-small ~ .icon,
.control.has-icons-left .select.is-small ~ .icon,
.control.has-icons-right .input.is-small ~ .icon,
.control.has-icons-right .select.is-small ~ .icon {
  font-size: 0.75rem;
}
.control.has-icons-left .input.is-medium ~ .icon,
.control.has-icons-left .select.is-medium ~ .icon,
.control.has-icons-right .input.is-medium ~ .icon,
.control.has-icons-right .select.is-medium ~ .icon {
  font-size: 1.25rem;
}
.control.has-icons-left .input.is-large ~ .icon,
.control.has-icons-left .select.is-large ~ .icon,
.control.has-icons-right .input.is-large ~ .icon,
.control.has-icons-right .select.is-large ~ .icon {
  font-size: 1.5rem;
}
.control.has-icons-left .icon,
.control.has-icons-right .icon {
  color: hsl(0deg, 0%, 86%);
  height: 2.5em;
  pointer-events: none;
  position: absolute;
  top: 0;
  width: 2.5em;
  z-index: 4;
}
.control.has-icons-left .input,
.control.has-icons-left .select select {
  padding-left: 2.5em;
}
.control.has-icons-left .icon.is-left {
  left: 0;
}
.control.has-icons-right .input,
.control.has-icons-right .select select {
  padding-right: 2.5em;
}
.control.has-icons-right .icon.is-right {
  right: 0;
}
.control.is-loading::after {
  position: absolute !important;
  right: 0.625em;
  top: 0.625em;
  z-index: 4;
}
.control.is-loading.is-small:after {
  font-size: 0.75rem;
}
.control.is-loading.is-medium:after {
  font-size: 1.25rem;
}
.control.is-loading.is-large:after {
  font-size: 1.5rem;
}
.breadcrumb {
  font-size: 1rem;
  white-space: nowrap;
}
.breadcrumb a {
  align-items: center;
  color: hsl(229deg, 53%, 53%);
  display: flex;
  justify-content: center;
  padding: 0 0.75em;
}
.breadcrumb a:hover {
  color: hsl(0deg, 0%, 21%);
}
.breadcrumb li {
  align-items: center;
  display: flex;
}
.breadcrumb li:first-child a {
  padding-left: 0;
}
.breadcrumb li.is-active a {
  color: hsl(0deg, 0%, 21%);
  cursor: default;
  pointer-events: none;
}
.breadcrumb li + li::before {
  color: hsl(0deg, 0%, 71%);
  content: "/";
}
.breadcrumb ul,
.breadcrumb ol {
  align-items: flex-start;
  display: flex;
  flex-wrap: wrap;
  justify-content: flex-start;
}
.breadcrumb .icon:first-child {
  margin-right: 0.5em;
}
.breadcrumb .icon:last-child {
  margin-left: 0.5em;
}
.breadcrumb.is-centered ol,
.breadcrumb.is-centered ul {
  justify-content: center;
}
.breadcrumb.is-right ol,
.breadcrumb.is-right ul {
  justify-content: flex-end;
}
.breadcrumb.is-small {
  font-size: 0.75rem;
}
.breadcrumb.is-medium {
  font-size: 1.25rem;
}
.breadcrumb.is-large {
  font-size: 1.5rem;
}
.breadcrumb.has-arrow-separator li + li::before {
  content: "\2192";
}
.breadcrumb.has-bullet-separator li + li::before {
  content: "\2022";
}
.breadcrumb.has-dot-separator li + li::before {
  content: "\b7";
}
.breadcrumb.has-succeeds-separator li + li::before {
  content: "\227b";
}
.card {
  background-color: hsl(0deg, 0%, 100%);
  border-radius: 0.25rem;
  box-shadow: none;
  color: hsl(0deg, 0%, 29%);
  max-width: 100%;
  position: relative;
}
.card-footer:first-child,
.card-content:first-child,
.card-header:first-child {
  border-top-left-radius: 0.25rem;
  border-top-right-radius: 0.25rem;
}
.card-footer:last-child,
.card-content:last-child,
.card-header:last-child {
  border-bottom-left-radius: 0.25rem;
  border-bottom-right-radius: 0.25rem;
}
.card-header {
  background-color: transparent;
  align-items: stretch;
  box-shadow: none;
  display: flex;
}
.card-header-title {
  align-items: center;
  color: hsl(0deg, 0%, 21%);
  display: flex;
  flex-grow: 1;
  font-weight: 700;
  padding: 0.75rem 1rem;
}
.card-header-title.is-centered {
  justify-content: center;
}
.card-header-icon {
  -moz-appearance: none;
  -webkit-appearance: none;
  appearance: none;
  background: none;
  border: none;
  color: currentColor;
  font-family: inherit;
  font-size: 1em;
  margin: 0;
  padding: 0;
  align-items: center;
  cursor: pointer;
  display: flex;
  justify-content: center;
  padding: 0.75rem 1rem;
}
.card-image {
  display: block;
  position: relative;
}
.card-image:first-child img {
  border-top-left-radius: 0.25rem;
  border-top-right-radius: 0.25rem;
}
.card-image:last-child img {
  border-bottom-left-radius: 0.25rem;
  border-bottom-right-radius: 0.25rem;
}
.card-content {
  background-color: transparent;
  padding: 1.5rem;
}
.card-footer {
  background-color: transparent;
  border-top: 1px solid hsl(0deg, 0%, 93%);
  align-items: stretch;
  display: flex;
}
.card-footer-item {
  align-items: center;
  display: flex;
  flex-basis: 0;
  flex-grow: 1;
  flex-shrink: 0;
  justify-content: center;
  padding: 0.75rem;
}
.card-footer-item:not(:last-child) {
  border-right: 1px solid hsl(0deg, 0%, 93%);
}
.card .media:not(:last-child) {
  margin-bottom: 1.5rem;
}
.dropdown {
  display: inline-flex;
  position: relative;
  vertical-align: top;
}
.dropdown.is-active .dropdown-menu,
.dropdown.is-hoverable:hover .dropdown-menu {
  display: block;
}
.dropdown.is-right .dropdown-menu {
  left: auto;
  right: 0;
}
.dropdown.is-up .dropdown-menu {
  bottom: 100%;
  padding-bottom: 4px;
  padding-top: initial;
  top: auto;
}
.dropdown-menu {
  display: none;
  left: 0;
  min-width: 12rem;
  padding-top: 4px;
  position: absolute;
  top: 100%;
  z-index: 20;
}
.dropdown-content {
  background-color: hsl(0deg, 0%, 100%);
  border-radius: 4px;
  box-shadow: 0 0.5em 1em -0.125em rgba(10, 10, 10, 0.1), 0 0px 0 1px rgba(10, 10, 10, 0.02);
  padding-bottom: 0.5rem;
  padding-top: 0.5rem;
}
.dropdown-item {
  color: hsl(0deg, 0%, 29%);
  display: block;
  font-size: 0.875rem;
  line-height: 1.5;
  padding: 0.375rem 1rem;
  position: relative;
}
a.dropdown-item,
button.dropdown-item {
  padding-right: 3rem;
  text-align: inherit;
  white-space: nowrap;
  width: 100%;
}
a.dropdown-item:hover,
button.dropdown-item:hover {
  background-color: hsl(0deg, 0%, 96%);
  color: hsl(0deg, 0%, 4%);
}
a.dropdown-item.is-active,
button.dropdown-item.is-active {
  background-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.dropdown-divider {
  background-color: hsl(0deg, 0%, 93%);
  border: none;
  display: block;
  height: 1px;
  margin: 0.5rem 0;
}
.level {
  align-items: center;
  justify-content: space-between;
}
.level code {
  border-radius: 4px;
}
.level img {
  display: inline-block;
  vertical-align: top;
}
.level.is-mobile {
  display: flex;
}
.level.is-mobile .level-left,
.level.is-mobile .level-right {
  display: flex;
}
.level.is-mobile .level-left + .level-right {
  margin-top: 0;
}
.level.is-mobile .level-item:not(:last-child) {
  margin-bottom: 0;
  margin-right: 0.75rem;
}
.level.is-mobile .level-item:not(.is-narrow) {
  flex-grow: 1;
}
@media screen and (min-width: 769px), print {
  .level {
    display: flex;
  }
  .level > .level-item:not(.is-narrow) {
    flex-grow: 1;
  }
}
.level-item {
  align-items: center;
  display: flex;
  flex-basis: auto;
  flex-grow: 0;
  flex-shrink: 0;
  justify-content: center;
}
.level-item .title,
.level-item .subtitle {
  margin-bottom: 0;
}
@media screen and (max-width: 768px) {
  .level-item:not(:last-child) {
    margin-bottom: 0.75rem;
  }
}
.level-left,
.level-right {
  flex-basis: auto;
  flex-grow: 0;
  flex-shrink: 0;
}
.level-left .level-item.is-flexible,
.level-right .level-item.is-flexible {
  flex-grow: 1;
}
@media screen and (min-width: 769px), print {
  .level-left .level-item:not(:last-child),
  .level-right .level-item:not(:last-child) {
    margin-right: 0.75rem;
  }
}
.level-left {
  align-items: center;
  justify-content: flex-start;
}
@media screen and (max-width: 768px) {
  .level-left + .level-right {
    margin-top: 1.5rem;
  }
}
@media screen and (min-width: 769px), print {
  .level-left {
    display: flex;
  }
}
.level-right {
  align-items: center;
  justify-content: flex-end;
}
@media screen and (min-width: 769px), print {
  .level-right {
    display: flex;
  }
}
.media {
  align-items: flex-start;
  display: flex;
  text-align: inherit;
}
.media .content:not(:last-child) {
  margin-bottom: 0.75rem;
}
.media .media {
  border-top: 1px solid rgba(219, 219, 219, 0.5);
  display: flex;
  padding-top: 0.75rem;
}
.media .media .content:not(:last-child),
.media .media .control:not(:last-child) {
  margin-bottom: 0.5rem;
}
.media .media .media {
  padding-top: 0.5rem;
}
.media .media .media + .media {
  margin-top: 0.5rem;
}
.media + .media {
  border-top: 1px solid rgba(219, 219, 219, 0.5);
  margin-top: 1rem;
  padding-top: 1rem;
}
.media.is-large + .media {
  margin-top: 1.5rem;
  padding-top: 1.5rem;
}
.media-left,
.media-right {
  flex-basis: auto;
  flex-grow: 0;
  flex-shrink: 0;
}
.media-left {
  margin-right: 1rem;
}
.media-right {
  margin-left: 1rem;
}
.media-content {
  flex-basis: auto;
  flex-grow: 1;
  flex-shrink: 1;
  text-align: inherit;
}
@media screen and (max-width: 768px) {
  .media-content {
    overflow-x: auto;
  }
}
.menu {
  font-size: 1rem;
}
.menu.is-small {
  font-size: 0.75rem;
}
.menu.is-medium {
  font-size: 1.25rem;
}
.menu.is-large {
  font-size: 1.5rem;
}
.menu-list {
  line-height: 1.25;
}
.menu-list a {
  border-radius: 0;
  color: #727c8f;
  display: block;
  padding: 0.5rem 0;
}
.menu-list a:hover {
  background-color: #262930;
  color: hsl(0deg, 0%, 100%);
}
.menu-list a.is-active {
  background-color: #282c33;
  color: hsl(0deg, 0%, 100%);
}
.menu-list li ul {
  border-left: 1px solid hsl(0deg, 0%, 86%);
  margin: 0.75em;
  padding-left: 0.75em;
}
.menu-label {
  color: #666f81;
  font-size: 0.75em;
  letter-spacing: 0.1em;
  text-transform: uppercase;
}
.menu-label:not(:first-child) {
  margin-top: 1em;
}
.menu-label:not(:last-child) {
  margin-bottom: 1em;
}
.message {
  background-color: hsl(0deg, 0%, 96%);
  border-radius: 4px;
  font-size: 1rem;
}
.message strong {
  color: currentColor;
}
.message a:not(.button):not(.tag):not(.dropdown-item) {
  color: currentColor;
  text-decoration: underline;
}
.message.is-small {
  font-size: 0.75rem;
}
.message.is-medium {
  font-size: 1.25rem;
}
.message.is-large {
  font-size: 1.5rem;
}
.message.is-white {
  background-color: white;
}
.message.is-white .message-header {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.message.is-white .message-body {
  border-color: hsl(0deg, 0%, 100%);
}
.message.is-black {
  background-color: #fafafa;
}
.message.is-black .message-header {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.message.is-black .message-body {
  border-color: hsl(0deg, 0%, 4%);
}
.message.is-light {
  background-color: #fafafa;
}
.message.is-light .message-header {
  background-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.message.is-light .message-body {
  border-color: hsl(0deg, 0%, 96%);
}
.message.is-dark {
  background-color: #fafafa;
}
.message.is-dark .message-header {
  background-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.message.is-dark .message-body {
  border-color: hsl(0deg, 0%, 21%);
}
.message.is-primary {
  background-color: #ebfffc;
}
.message.is-primary .message-header {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.message.is-primary .message-body {
  border-color: hsl(171deg, 100%, 41%);
  color: #00947e;
}
.message.is-link {
  background-color: #eff1fa;
}
.message.is-link .message-header {
  background-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.message.is-link .message-body {
  border-color: hsl(229deg, 53%, 53%);
  color: #3850b7;
}
.message.is-info {
  background-color: #eff5fb;
}
.message.is-info .message-header {
  background-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.message.is-info .message-body {
  border-color: hsl(207deg, 61%, 53%);
  color: #296fa8;
}
.message.is-success {
  background-color: #effaf5;
}
.message.is-success .message-header {
  background-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.message.is-success .message-body {
  border-color: hsl(153deg, 53%, 53%);
  color: #257953;
}
.message.is-warning {
  background-color: #fffaeb;
}
.message.is-warning .message-header {
  background-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.message.is-warning .message-body {
  border-color: hsl(44deg, 100%, 77%);
  color: #946c00;
}
.message.is-danger {
  background-color: #feecf0;
}
.message.is-danger .message-header {
  background-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.message.is-danger .message-body {
  border-color: hsl(348deg, 86%, 61%);
  color: #cc0f35;
}
.message-header {
  align-items: center;
  background-color: hsl(0deg, 0%, 29%);
  border-radius: 4px 4px 0 0;
  color: #fff;
  display: flex;
  font-weight: 700;
  justify-content: space-between;
  line-height: 1.25;
  padding: 0.75em 1em;
  position: relative;
}
.message-header .delete {
  flex-grow: 0;
  flex-shrink: 0;
  margin-left: 0.75em;
}
.message-header + .message-body {
  border-width: 0;
  border-top-left-radius: 0;
  border-top-right-radius: 0;
}
.message-body {
  border-color: hsl(0deg, 0%, 86%);
  border-radius: 4px;
  border-style: solid;
  border-width: 0 0 0 4px;
  color: hsl(0deg, 0%, 29%);
  padding: 1.25em 1.5em;
}
.message-body code,
.message-body pre {
  background-color: hsl(0deg, 0%, 100%);
}
.message-body pre code {
  background-color: transparent;
}
.modal {
  align-items: center;
  display: none;
  flex-direction: column;
  justify-content: center;
  overflow: hidden;
  position: fixed;
  z-index: 40;
}
.modal.is-active {
  display: flex;
}
.modal-background {
  background-color: rgba(10, 10, 10, 0.86);
}
.modal-content,
.modal-card {
  margin: 0 20px;
  max-height: calc(100vh - 160px);
  overflow: auto;
  position: relative;
  width: 100%;
}
@media screen and (min-width: 769px) {
  .modal-content,
  .modal-card {
    margin: 0 auto;
    max-height: calc(100vh - 40px);
    width: 640px;
  }
}
.modal-close {
  background: none;
  height: 40px;
  position: fixed;
  right: 20px;
  top: 20px;
  width: 40px;
}
.modal-card {
  display: flex;
  flex-direction: column;
  max-height: calc(100vh - 40px);
  overflow: hidden;
  -ms-overflow-y: visible;
}
.modal-card-head,
.modal-card-foot {
  align-items: center;
  background-color: hsl(0deg, 0%, 96%);
  display: flex;
  flex-shrink: 0;
  justify-content: flex-start;
  padding: 20px;
  position: relative;
}
.modal-card-head {
  border-bottom: 1px solid hsl(0deg, 0%, 96%);
  border-top-left-radius: 6px;
  border-top-right-radius: 6px;
}
.modal-card-title {
  color: hsl(0deg, 0%, 21%);
  flex-grow: 1;
  flex-shrink: 0;
  font-size: 1rem;
  line-height: 1;
}
.modal-card-foot {
  border-bottom-left-radius: 6px;
  border-bottom-right-radius: 6px;
  border-top: 0;
}
.modal-card-foot .button:not(:last-child) {
  margin-right: 0.5em;
}
.modal-card-body {
  -webkit-overflow-scrolling: touch;
  background-color: hsl(0deg, 0%, 100%);
  flex-grow: 1;
  flex-shrink: 1;
  overflow: auto;
  padding: 1.5rem 20px;
}
.navbar {
  background-color: hsl(0deg, 0%, 100%);
  min-height: 3.25rem;
  position: relative;
  z-index: 30;
}
.navbar.is-white {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.navbar.is-white .navbar-brand > .navbar-item,
.navbar.is-white .navbar-brand .navbar-link {
  color: hsl(0deg, 0%, 4%);
}
.navbar.is-white .navbar-brand > a.navbar-item:focus,
.navbar.is-white .navbar-brand > a.navbar-item:hover,
.navbar.is-white .navbar-brand > a.navbar-item.is-active,
.navbar.is-white .navbar-brand .navbar-link:focus,
.navbar.is-white .navbar-brand .navbar-link:hover,
.navbar.is-white .navbar-brand .navbar-link.is-active {
  background-color: #f2f2f2;
  color: hsl(0deg, 0%, 4%);
}
.navbar.is-white .navbar-brand .navbar-link::after {
  border-color: hsl(0deg, 0%, 4%);
}
.navbar.is-white .navbar-burger {
  color: hsl(0deg, 0%, 4%);
}
@media screen and (min-width: 1024px) {
  .navbar.is-white .navbar-start > .navbar-item,
  .navbar.is-white .navbar-start .navbar-link,
  .navbar.is-white .navbar-end > .navbar-item,
  .navbar.is-white .navbar-end .navbar-link {
    color: hsl(0deg, 0%, 4%);
  }
  .navbar.is-white .navbar-start > a.navbar-item:focus,
  .navbar.is-white .navbar-start > a.navbar-item:hover,
  .navbar.is-white .navbar-start > a.navbar-item.is-active,
  .navbar.is-white .navbar-start .navbar-link:focus,
  .navbar.is-white .navbar-start .navbar-link:hover,
  .navbar.is-white .navbar-start .navbar-link.is-active,
  .navbar.is-white .navbar-end > a.navbar-item:focus,
  .navbar.is-white .navbar-end > a.navbar-item:hover,
  .navbar.is-white .navbar-end > a.navbar-item.is-active,
  .navbar.is-white .navbar-end .navbar-link:focus,
  .navbar.is-white .navbar-end .navbar-link:hover,
  .navbar.is-white .navbar-end .navbar-link.is-active {
    background-color: #f2f2f2;
    color: hsl(0deg, 0%, 4%);
  }
  .navbar.is-white .navbar-start .navbar-link::after,
  .navbar.is-white .navbar-end .navbar-link::after {
    border-color: hsl(0deg, 0%, 4%);
  }
  .navbar.is-white .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-white .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-white .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #f2f2f2;
    color: hsl(0deg, 0%, 4%);
  }
  .navbar.is-white .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(0deg, 0%, 100%);
    color: hsl(0deg, 0%, 4%);
  }
}
.navbar.is-black {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.navbar.is-black .navbar-brand > .navbar-item,
.navbar.is-black .navbar-brand .navbar-link {
  color: hsl(0deg, 0%, 100%);
}
.navbar.is-black .navbar-brand > a.navbar-item:focus,
.navbar.is-black .navbar-brand > a.navbar-item:hover,
.navbar.is-black .navbar-brand > a.navbar-item.is-active,
.navbar.is-black .navbar-brand .navbar-link:focus,
.navbar.is-black .navbar-brand .navbar-link:hover,
.navbar.is-black .navbar-brand .navbar-link.is-active {
  background-color: black;
  color: hsl(0deg, 0%, 100%);
}
.navbar.is-black .navbar-brand .navbar-link::after {
  border-color: hsl(0deg, 0%, 100%);
}
.navbar.is-black .navbar-burger {
  color: hsl(0deg, 0%, 100%);
}
@media screen and (min-width: 1024px) {
  .navbar.is-black .navbar-start > .navbar-item,
  .navbar.is-black .navbar-start .navbar-link,
  .navbar.is-black .navbar-end > .navbar-item,
  .navbar.is-black .navbar-end .navbar-link {
    color: hsl(0deg, 0%, 100%);
  }
  .navbar.is-black .navbar-start > a.navbar-item:focus,
  .navbar.is-black .navbar-start > a.navbar-item:hover,
  .navbar.is-black .navbar-start > a.navbar-item.is-active,
  .navbar.is-black .navbar-start .navbar-link:focus,
  .navbar.is-black .navbar-start .navbar-link:hover,
  .navbar.is-black .navbar-start .navbar-link.is-active,
  .navbar.is-black .navbar-end > a.navbar-item:focus,
  .navbar.is-black .navbar-end > a.navbar-item:hover,
  .navbar.is-black .navbar-end > a.navbar-item.is-active,
  .navbar.is-black .navbar-end .navbar-link:focus,
  .navbar.is-black .navbar-end .navbar-link:hover,
  .navbar.is-black .navbar-end .navbar-link.is-active {
    background-color: black;
    color: hsl(0deg, 0%, 100%);
  }
  .navbar.is-black .navbar-start .navbar-link::after,
  .navbar.is-black .navbar-end .navbar-link::after {
    border-color: hsl(0deg, 0%, 100%);
  }
  .navbar.is-black .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-black .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-black .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: black;
    color: hsl(0deg, 0%, 100%);
  }
  .navbar.is-black .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(0deg, 0%, 4%);
    color: hsl(0deg, 0%, 100%);
  }
}
.navbar.is-light {
  background-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.navbar.is-light .navbar-brand > .navbar-item,
.navbar.is-light .navbar-brand .navbar-link {
  color: rgba(0, 0, 0, 0.7);
}
.navbar.is-light .navbar-brand > a.navbar-item:focus,
.navbar.is-light .navbar-brand > a.navbar-item:hover,
.navbar.is-light .navbar-brand > a.navbar-item.is-active,
.navbar.is-light .navbar-brand .navbar-link:focus,
.navbar.is-light .navbar-brand .navbar-link:hover,
.navbar.is-light .navbar-brand .navbar-link.is-active {
  background-color: #e8e8e8;
  color: rgba(0, 0, 0, 0.7);
}
.navbar.is-light .navbar-brand .navbar-link::after {
  border-color: rgba(0, 0, 0, 0.7);
}
.navbar.is-light .navbar-burger {
  color: rgba(0, 0, 0, 0.7);
}
@media screen and (min-width: 1024px) {
  .navbar.is-light .navbar-start > .navbar-item,
  .navbar.is-light .navbar-start .navbar-link,
  .navbar.is-light .navbar-end > .navbar-item,
  .navbar.is-light .navbar-end .navbar-link {
    color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-light .navbar-start > a.navbar-item:focus,
  .navbar.is-light .navbar-start > a.navbar-item:hover,
  .navbar.is-light .navbar-start > a.navbar-item.is-active,
  .navbar.is-light .navbar-start .navbar-link:focus,
  .navbar.is-light .navbar-start .navbar-link:hover,
  .navbar.is-light .navbar-start .navbar-link.is-active,
  .navbar.is-light .navbar-end > a.navbar-item:focus,
  .navbar.is-light .navbar-end > a.navbar-item:hover,
  .navbar.is-light .navbar-end > a.navbar-item.is-active,
  .navbar.is-light .navbar-end .navbar-link:focus,
  .navbar.is-light .navbar-end .navbar-link:hover,
  .navbar.is-light .navbar-end .navbar-link.is-active {
    background-color: #e8e8e8;
    color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-light .navbar-start .navbar-link::after,
  .navbar.is-light .navbar-end .navbar-link::after {
    border-color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-light .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-light .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-light .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #e8e8e8;
    color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-light .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(0deg, 0%, 96%);
    color: rgba(0, 0, 0, 0.7);
  }
}
.navbar.is-dark {
  background-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.navbar.is-dark .navbar-brand > .navbar-item,
.navbar.is-dark .navbar-brand .navbar-link {
  color: #fff;
}
.navbar.is-dark .navbar-brand > a.navbar-item:focus,
.navbar.is-dark .navbar-brand > a.navbar-item:hover,
.navbar.is-dark .navbar-brand > a.navbar-item.is-active,
.navbar.is-dark .navbar-brand .navbar-link:focus,
.navbar.is-dark .navbar-brand .navbar-link:hover,
.navbar.is-dark .navbar-brand .navbar-link.is-active {
  background-color: #292929;
  color: #fff;
}
.navbar.is-dark .navbar-brand .navbar-link::after {
  border-color: #fff;
}
.navbar.is-dark .navbar-burger {
  color: #fff;
}
@media screen and (min-width: 1024px) {
  .navbar.is-dark .navbar-start > .navbar-item,
  .navbar.is-dark .navbar-start .navbar-link,
  .navbar.is-dark .navbar-end > .navbar-item,
  .navbar.is-dark .navbar-end .navbar-link {
    color: #fff;
  }
  .navbar.is-dark .navbar-start > a.navbar-item:focus,
  .navbar.is-dark .navbar-start > a.navbar-item:hover,
  .navbar.is-dark .navbar-start > a.navbar-item.is-active,
  .navbar.is-dark .navbar-start .navbar-link:focus,
  .navbar.is-dark .navbar-start .navbar-link:hover,
  .navbar.is-dark .navbar-start .navbar-link.is-active,
  .navbar.is-dark .navbar-end > a.navbar-item:focus,
  .navbar.is-dark .navbar-end > a.navbar-item:hover,
  .navbar.is-dark .navbar-end > a.navbar-item.is-active,
  .navbar.is-dark .navbar-end .navbar-link:focus,
  .navbar.is-dark .navbar-end .navbar-link:hover,
  .navbar.is-dark .navbar-end .navbar-link.is-active {
    background-color: #292929;
    color: #fff;
  }
  .navbar.is-dark .navbar-start .navbar-link::after,
  .navbar.is-dark .navbar-end .navbar-link::after {
    border-color: #fff;
  }
  .navbar.is-dark .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-dark .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-dark .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #292929;
    color: #fff;
  }
  .navbar.is-dark .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(0deg, 0%, 21%);
    color: #fff;
  }
}
.navbar.is-primary {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.navbar.is-primary .navbar-brand > .navbar-item,
.navbar.is-primary .navbar-brand .navbar-link {
  color: #fff;
}
.navbar.is-primary .navbar-brand > a.navbar-item:focus,
.navbar.is-primary .navbar-brand > a.navbar-item:hover,
.navbar.is-primary .navbar-brand > a.navbar-item.is-active,
.navbar.is-primary .navbar-brand .navbar-link:focus,
.navbar.is-primary .navbar-brand .navbar-link:hover,
.navbar.is-primary .navbar-brand .navbar-link.is-active {
  background-color: #00b89c;
  color: #fff;
}
.navbar.is-primary .navbar-brand .navbar-link::after {
  border-color: #fff;
}
.navbar.is-primary .navbar-burger {
  color: #fff;
}
@media screen and (min-width: 1024px) {
  .navbar.is-primary .navbar-start > .navbar-item,
  .navbar.is-primary .navbar-start .navbar-link,
  .navbar.is-primary .navbar-end > .navbar-item,
  .navbar.is-primary .navbar-end .navbar-link {
    color: #fff;
  }
  .navbar.is-primary .navbar-start > a.navbar-item:focus,
  .navbar.is-primary .navbar-start > a.navbar-item:hover,
  .navbar.is-primary .navbar-start > a.navbar-item.is-active,
  .navbar.is-primary .navbar-start .navbar-link:focus,
  .navbar.is-primary .navbar-start .navbar-link:hover,
  .navbar.is-primary .navbar-start .navbar-link.is-active,
  .navbar.is-primary .navbar-end > a.navbar-item:focus,
  .navbar.is-primary .navbar-end > a.navbar-item:hover,
  .navbar.is-primary .navbar-end > a.navbar-item.is-active,
  .navbar.is-primary .navbar-end .navbar-link:focus,
  .navbar.is-primary .navbar-end .navbar-link:hover,
  .navbar.is-primary .navbar-end .navbar-link.is-active {
    background-color: #00b89c;
    color: #fff;
  }
  .navbar.is-primary .navbar-start .navbar-link::after,
  .navbar.is-primary .navbar-end .navbar-link::after {
    border-color: #fff;
  }
  .navbar.is-primary .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-primary .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-primary .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #00b89c;
    color: #fff;
  }
  .navbar.is-primary .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(171deg, 100%, 41%);
    color: #fff;
  }
}
.navbar.is-link {
  background-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.navbar.is-link .navbar-brand > .navbar-item,
.navbar.is-link .navbar-brand .navbar-link {
  color: #fff;
}
.navbar.is-link .navbar-brand > a.navbar-item:focus,
.navbar.is-link .navbar-brand > a.navbar-item:hover,
.navbar.is-link .navbar-brand > a.navbar-item.is-active,
.navbar.is-link .navbar-brand .navbar-link:focus,
.navbar.is-link .navbar-brand .navbar-link:hover,
.navbar.is-link .navbar-brand .navbar-link.is-active {
  background-color: #3a51bb;
  color: #fff;
}
.navbar.is-link .navbar-brand .navbar-link::after {
  border-color: #fff;
}
.navbar.is-link .navbar-burger {
  color: #fff;
}
@media screen and (min-width: 1024px) {
  .navbar.is-link .navbar-start > .navbar-item,
  .navbar.is-link .navbar-start .navbar-link,
  .navbar.is-link .navbar-end > .navbar-item,
  .navbar.is-link .navbar-end .navbar-link {
    color: #fff;
  }
  .navbar.is-link .navbar-start > a.navbar-item:focus,
  .navbar.is-link .navbar-start > a.navbar-item:hover,
  .navbar.is-link .navbar-start > a.navbar-item.is-active,
  .navbar.is-link .navbar-start .navbar-link:focus,
  .navbar.is-link .navbar-start .navbar-link:hover,
  .navbar.is-link .navbar-start .navbar-link.is-active,
  .navbar.is-link .navbar-end > a.navbar-item:focus,
  .navbar.is-link .navbar-end > a.navbar-item:hover,
  .navbar.is-link .navbar-end > a.navbar-item.is-active,
  .navbar.is-link .navbar-end .navbar-link:focus,
  .navbar.is-link .navbar-end .navbar-link:hover,
  .navbar.is-link .navbar-end .navbar-link.is-active {
    background-color: #3a51bb;
    color: #fff;
  }
  .navbar.is-link .navbar-start .navbar-link::after,
  .navbar.is-link .navbar-end .navbar-link::after {
    border-color: #fff;
  }
  .navbar.is-link .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-link .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-link .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #3a51bb;
    color: #fff;
  }
  .navbar.is-link .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(229deg, 53%, 53%);
    color: #fff;
  }
}
.navbar.is-info {
  background-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.navbar.is-info .navbar-brand > .navbar-item,
.navbar.is-info .navbar-brand .navbar-link {
  color: #fff;
}
.navbar.is-info .navbar-brand > a.navbar-item:focus,
.navbar.is-info .navbar-brand > a.navbar-item:hover,
.navbar.is-info .navbar-brand > a.navbar-item.is-active,
.navbar.is-info .navbar-brand .navbar-link:focus,
.navbar.is-info .navbar-brand .navbar-link:hover,
.navbar.is-info .navbar-brand .navbar-link.is-active {
  background-color: #3082c5;
  color: #fff;
}
.navbar.is-info .navbar-brand .navbar-link::after {
  border-color: #fff;
}
.navbar.is-info .navbar-burger {
  color: #fff;
}
@media screen and (min-width: 1024px) {
  .navbar.is-info .navbar-start > .navbar-item,
  .navbar.is-info .navbar-start .navbar-link,
  .navbar.is-info .navbar-end > .navbar-item,
  .navbar.is-info .navbar-end .navbar-link {
    color: #fff;
  }
  .navbar.is-info .navbar-start > a.navbar-item:focus,
  .navbar.is-info .navbar-start > a.navbar-item:hover,
  .navbar.is-info .navbar-start > a.navbar-item.is-active,
  .navbar.is-info .navbar-start .navbar-link:focus,
  .navbar.is-info .navbar-start .navbar-link:hover,
  .navbar.is-info .navbar-start .navbar-link.is-active,
  .navbar.is-info .navbar-end > a.navbar-item:focus,
  .navbar.is-info .navbar-end > a.navbar-item:hover,
  .navbar.is-info .navbar-end > a.navbar-item.is-active,
  .navbar.is-info .navbar-end .navbar-link:focus,
  .navbar.is-info .navbar-end .navbar-link:hover,
  .navbar.is-info .navbar-end .navbar-link.is-active {
    background-color: #3082c5;
    color: #fff;
  }
  .navbar.is-info .navbar-start .navbar-link::after,
  .navbar.is-info .navbar-end .navbar-link::after {
    border-color: #fff;
  }
  .navbar.is-info .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-info .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-info .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #3082c5;
    color: #fff;
  }
  .navbar.is-info .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(207deg, 61%, 53%);
    color: #fff;
  }
}
.navbar.is-success {
  background-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.navbar.is-success .navbar-brand > .navbar-item,
.navbar.is-success .navbar-brand .navbar-link {
  color: #fff;
}
.navbar.is-success .navbar-brand > a.navbar-item:focus,
.navbar.is-success .navbar-brand > a.navbar-item:hover,
.navbar.is-success .navbar-brand > a.navbar-item.is-active,
.navbar.is-success .navbar-brand .navbar-link:focus,
.navbar.is-success .navbar-brand .navbar-link:hover,
.navbar.is-success .navbar-brand .navbar-link.is-active {
  background-color: #3abb81;
  color: #fff;
}
.navbar.is-success .navbar-brand .navbar-link::after {
  border-color: #fff;
}
.navbar.is-success .navbar-burger {
  color: #fff;
}
@media screen and (min-width: 1024px) {
  .navbar.is-success .navbar-start > .navbar-item,
  .navbar.is-success .navbar-start .navbar-link,
  .navbar.is-success .navbar-end > .navbar-item,
  .navbar.is-success .navbar-end .navbar-link {
    color: #fff;
  }
  .navbar.is-success .navbar-start > a.navbar-item:focus,
  .navbar.is-success .navbar-start > a.navbar-item:hover,
  .navbar.is-success .navbar-start > a.navbar-item.is-active,
  .navbar.is-success .navbar-start .navbar-link:focus,
  .navbar.is-success .navbar-start .navbar-link:hover,
  .navbar.is-success .navbar-start .navbar-link.is-active,
  .navbar.is-success .navbar-end > a.navbar-item:focus,
  .navbar.is-success .navbar-end > a.navbar-item:hover,
  .navbar.is-success .navbar-end > a.navbar-item.is-active,
  .navbar.is-success .navbar-end .navbar-link:focus,
  .navbar.is-success .navbar-end .navbar-link:hover,
  .navbar.is-success .navbar-end .navbar-link.is-active {
    background-color: #3abb81;
    color: #fff;
  }
  .navbar.is-success .navbar-start .navbar-link::after,
  .navbar.is-success .navbar-end .navbar-link::after {
    border-color: #fff;
  }
  .navbar.is-success .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-success .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-success .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #3abb81;
    color: #fff;
  }
  .navbar.is-success .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(153deg, 53%, 53%);
    color: #fff;
  }
}
.navbar.is-warning {
  background-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.navbar.is-warning .navbar-brand > .navbar-item,
.navbar.is-warning .navbar-brand .navbar-link {
  color: rgba(0, 0, 0, 0.7);
}
.navbar.is-warning .navbar-brand > a.navbar-item:focus,
.navbar.is-warning .navbar-brand > a.navbar-item:hover,
.navbar.is-warning .navbar-brand > a.navbar-item.is-active,
.navbar.is-warning .navbar-brand .navbar-link:focus,
.navbar.is-warning .navbar-brand .navbar-link:hover,
.navbar.is-warning .navbar-brand .navbar-link.is-active {
  background-color: #ffd970;
  color: rgba(0, 0, 0, 0.7);
}
.navbar.is-warning .navbar-brand .navbar-link::after {
  border-color: rgba(0, 0, 0, 0.7);
}
.navbar.is-warning .navbar-burger {
  color: rgba(0, 0, 0, 0.7);
}
@media screen and (min-width: 1024px) {
  .navbar.is-warning .navbar-start > .navbar-item,
  .navbar.is-warning .navbar-start .navbar-link,
  .navbar.is-warning .navbar-end > .navbar-item,
  .navbar.is-warning .navbar-end .navbar-link {
    color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-warning .navbar-start > a.navbar-item:focus,
  .navbar.is-warning .navbar-start > a.navbar-item:hover,
  .navbar.is-warning .navbar-start > a.navbar-item.is-active,
  .navbar.is-warning .navbar-start .navbar-link:focus,
  .navbar.is-warning .navbar-start .navbar-link:hover,
  .navbar.is-warning .navbar-start .navbar-link.is-active,
  .navbar.is-warning .navbar-end > a.navbar-item:focus,
  .navbar.is-warning .navbar-end > a.navbar-item:hover,
  .navbar.is-warning .navbar-end > a.navbar-item.is-active,
  .navbar.is-warning .navbar-end .navbar-link:focus,
  .navbar.is-warning .navbar-end .navbar-link:hover,
  .navbar.is-warning .navbar-end .navbar-link.is-active {
    background-color: #ffd970;
    color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-warning .navbar-start .navbar-link::after,
  .navbar.is-warning .navbar-end .navbar-link::after {
    border-color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-warning .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-warning .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-warning .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #ffd970;
    color: rgba(0, 0, 0, 0.7);
  }
  .navbar.is-warning .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(44deg, 100%, 77%);
    color: rgba(0, 0, 0, 0.7);
  }
}
.navbar.is-danger {
  background-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.navbar.is-danger .navbar-brand > .navbar-item,
.navbar.is-danger .navbar-brand .navbar-link {
  color: #fff;
}
.navbar.is-danger .navbar-brand > a.navbar-item:focus,
.navbar.is-danger .navbar-brand > a.navbar-item:hover,
.navbar.is-danger .navbar-brand > a.navbar-item.is-active,
.navbar.is-danger .navbar-brand .navbar-link:focus,
.navbar.is-danger .navbar-brand .navbar-link:hover,
.navbar.is-danger .navbar-brand .navbar-link.is-active {
  background-color: #ef2e55;
  color: #fff;
}
.navbar.is-danger .navbar-brand .navbar-link::after {
  border-color: #fff;
}
.navbar.is-danger .navbar-burger {
  color: #fff;
}
@media screen and (min-width: 1024px) {
  .navbar.is-danger .navbar-start > .navbar-item,
  .navbar.is-danger .navbar-start .navbar-link,
  .navbar.is-danger .navbar-end > .navbar-item,
  .navbar.is-danger .navbar-end .navbar-link {
    color: #fff;
  }
  .navbar.is-danger .navbar-start > a.navbar-item:focus,
  .navbar.is-danger .navbar-start > a.navbar-item:hover,
  .navbar.is-danger .navbar-start > a.navbar-item.is-active,
  .navbar.is-danger .navbar-start .navbar-link:focus,
  .navbar.is-danger .navbar-start .navbar-link:hover,
  .navbar.is-danger .navbar-start .navbar-link.is-active,
  .navbar.is-danger .navbar-end > a.navbar-item:focus,
  .navbar.is-danger .navbar-end > a.navbar-item:hover,
  .navbar.is-danger .navbar-end > a.navbar-item.is-active,
  .navbar.is-danger .navbar-end .navbar-link:focus,
  .navbar.is-danger .navbar-end .navbar-link:hover,
  .navbar.is-danger .navbar-end .navbar-link.is-active {
    background-color: #ef2e55;
    color: #fff;
  }
  .navbar.is-danger .navbar-start .navbar-link::after,
  .navbar.is-danger .navbar-end .navbar-link::after {
    border-color: #fff;
  }
  .navbar.is-danger .navbar-item.has-dropdown:focus .navbar-link,
  .navbar.is-danger .navbar-item.has-dropdown:hover .navbar-link,
  .navbar.is-danger .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: #ef2e55;
    color: #fff;
  }
  .navbar.is-danger .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(348deg, 86%, 61%);
    color: #fff;
  }
}
.navbar > .container {
  align-items: stretch;
  display: flex;
  min-height: 3.25rem;
  width: 100%;
}
.navbar.has-shadow {
  box-shadow: 0 2px 0 0 hsl(0deg, 0%, 96%);
}
.navbar.is-fixed-bottom,
.navbar.is-fixed-top {
  left: 0;
  position: fixed;
  right: 0;
  z-index: 30;
}
.navbar.is-fixed-bottom {
  bottom: 0;
}
.navbar.is-fixed-bottom.has-shadow {
  box-shadow: 0 -2px 0 0 hsl(0deg, 0%, 96%);
}
.navbar.is-fixed-top {
  top: 0;
}
html.has-navbar-fixed-top,
body.has-navbar-fixed-top {
  padding-top: 3.25rem;
}
html.has-navbar-fixed-bottom,
body.has-navbar-fixed-bottom {
  padding-bottom: 3.25rem;
}
.navbar-brand,
.navbar-tabs {
  align-items: stretch;
  display: flex;
  flex-shrink: 0;
  min-height: 3.25rem;
}
.navbar-brand a.navbar-item:focus,
.navbar-brand a.navbar-item:hover {
  background-color: transparent;
}
.navbar-tabs {
  -webkit-overflow-scrolling: touch;
  max-width: 100vw;
  overflow-x: auto;
  overflow-y: hidden;
}
.navbar-burger {
  color: hsl(0deg, 0%, 29%);
  -moz-appearance: none;
  -webkit-appearance: none;
  appearance: none;
  background: none;
  border: none;
  cursor: pointer;
  display: block;
  height: 3.25rem;
  position: relative;
  width: 3.25rem;
  margin-left: auto;
}
.navbar-burger span {
  background-color: currentColor;
  display: block;
  height: 1px;
  left: calc(50% - 8px);
  position: absolute;
  transform-origin: center;
  transition-duration: 86ms;
  transition-property:
    background-color,
    opacity,
    transform;
  transition-timing-function: ease-out;
  width: 16px;
}
.navbar-burger span:nth-child(1) {
  top: calc(50% - 6px);
}
.navbar-burger span:nth-child(2) {
  top: calc(50% - 1px);
}
.navbar-burger span:nth-child(3) {
  top: calc(50% + 4px);
}
.navbar-burger:hover {
  background-color: rgba(0, 0, 0, 0.05);
}
.navbar-burger.is-active span:nth-child(1) {
  transform: translateY(5px) rotate(45deg);
}
.navbar-burger.is-active span:nth-child(2) {
  opacity: 0;
}
.navbar-burger.is-active span:nth-child(3) {
  transform: translateY(-5px) rotate(-45deg);
}
.navbar-menu {
  display: none;
}
.navbar-item,
.navbar-link {
  color: hsl(0deg, 0%, 29%);
  display: block;
  line-height: 1.5;
  padding: 0.5rem 0.75rem;
  position: relative;
}
.navbar-item .icon:only-child,
.navbar-link .icon:only-child {
  margin-left: -0.25rem;
  margin-right: -0.25rem;
}
a.navbar-item,
.navbar-link {
  cursor: pointer;
}
a.navbar-item:focus,
a.navbar-item:focus-within,
a.navbar-item:hover,
a.navbar-item.is-active,
.navbar-link:focus,
.navbar-link:focus-within,
.navbar-link:hover,
.navbar-link.is-active {
  background-color: hsl(0deg, 0%, 98%);
  color: hsl(229deg, 53%, 53%);
}
.navbar-item {
  flex-grow: 0;
  flex-shrink: 0;
}
.navbar-item img {
  max-height: 1.75rem;
}
.navbar-item.has-dropdown {
  padding: 0;
}
.navbar-item.is-expanded {
  flex-grow: 1;
  flex-shrink: 1;
}
.navbar-item.is-tab {
  border-bottom: 1px solid transparent;
  min-height: 3.25rem;
  padding-bottom: calc(0.5rem - 1px);
}
.navbar-item.is-tab:focus,
.navbar-item.is-tab:hover {
  background-color: transparent;
  border-bottom-color: hsl(229deg, 53%, 53%);
}
.navbar-item.is-tab.is-active {
  background-color: transparent;
  border-bottom-color: hsl(229deg, 53%, 53%);
  border-bottom-style: solid;
  border-bottom-width: 3px;
  color: hsl(229deg, 53%, 53%);
  padding-bottom: calc(0.5rem - 3px);
}
.navbar-content {
  flex-grow: 1;
  flex-shrink: 1;
}
.navbar-link:not(.is-arrowless) {
  padding-right: 2.5em;
}
.navbar-link:not(.is-arrowless)::after {
  border-color: hsl(229deg, 53%, 53%);
  margin-top: -0.375em;
  right: 1.125em;
}
.navbar-dropdown {
  font-size: 0.875rem;
  padding-bottom: 0.5rem;
  padding-top: 0.5rem;
}
.navbar-dropdown .navbar-item {
  padding-left: 1.5rem;
  padding-right: 1.5rem;
}
.navbar-divider {
  background-color: hsl(0deg, 0%, 96%);
  border: none;
  display: none;
  height: 2px;
  margin: 0.5rem 0;
}
@media screen and (max-width: 1023px) {
  .navbar > .container {
    display: block;
  }
  .navbar-brand .navbar-item,
  .navbar-tabs .navbar-item {
    align-items: center;
    display: flex;
  }
  .navbar-link::after {
    display: none;
  }
  .navbar-menu {
    background-color: hsl(0deg, 0%, 100%);
    box-shadow: 0 8px 16px rgba(10, 10, 10, 0.1);
    padding: 0.5rem 0;
  }
  .navbar-menu.is-active {
    display: block;
  }
  .navbar.is-fixed-bottom-touch,
  .navbar.is-fixed-top-touch {
    left: 0;
    position: fixed;
    right: 0;
    z-index: 30;
  }
  .navbar.is-fixed-bottom-touch {
    bottom: 0;
  }
  .navbar.is-fixed-bottom-touch.has-shadow {
    box-shadow: 0 -2px 3px rgba(10, 10, 10, 0.1);
  }
  .navbar.is-fixed-top-touch {
    top: 0;
  }
  .navbar.is-fixed-top .navbar-menu,
  .navbar.is-fixed-top-touch .navbar-menu {
    -webkit-overflow-scrolling: touch;
    max-height: calc(100vh - 3.25rem);
    overflow: auto;
  }
  html.has-navbar-fixed-top-touch,
  body.has-navbar-fixed-top-touch {
    padding-top: 3.25rem;
  }
  html.has-navbar-fixed-bottom-touch,
  body.has-navbar-fixed-bottom-touch {
    padding-bottom: 3.25rem;
  }
}
@media screen and (min-width: 1024px) {
  .navbar,
  .navbar-menu,
  .navbar-start,
  .navbar-end {
    align-items: stretch;
    display: flex;
  }
  .navbar {
    min-height: 3.25rem;
  }
  .navbar.is-spaced {
    padding: 1rem 2rem;
  }
  .navbar.is-spaced .navbar-start,
  .navbar.is-spaced .navbar-end {
    align-items: center;
  }
  .navbar.is-spaced a.navbar-item,
  .navbar.is-spaced .navbar-link {
    border-radius: 4px;
  }
  .navbar.is-transparent a.navbar-item:focus,
  .navbar.is-transparent a.navbar-item:hover,
  .navbar.is-transparent a.navbar-item.is-active,
  .navbar.is-transparent .navbar-link:focus,
  .navbar.is-transparent .navbar-link:hover,
  .navbar.is-transparent .navbar-link.is-active {
    background-color: transparent !important;
  }
  .navbar.is-transparent .navbar-item.has-dropdown.is-active .navbar-link,
  .navbar.is-transparent .navbar-item.has-dropdown.is-hoverable:focus .navbar-link,
  .navbar.is-transparent .navbar-item.has-dropdown.is-hoverable:focus-within .navbar-link,
  .navbar.is-transparent .navbar-item.has-dropdown.is-hoverable:hover .navbar-link {
    background-color: transparent !important;
  }
  .navbar.is-transparent .navbar-dropdown a.navbar-item:focus,
  .navbar.is-transparent .navbar-dropdown a.navbar-item:hover {
    background-color: hsl(0deg, 0%, 96%);
    color: hsl(0deg, 0%, 4%);
  }
  .navbar.is-transparent .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(0deg, 0%, 96%);
    color: hsl(229deg, 53%, 53%);
  }
  .navbar-burger {
    display: none;
  }
  .navbar-item,
  .navbar-link {
    align-items: center;
    display: flex;
  }
  .navbar-item.has-dropdown {
    align-items: stretch;
  }
  .navbar-item.has-dropdown-up .navbar-link::after {
    transform: rotate(135deg) translate(0.25em, -0.25em);
  }
  .navbar-item.has-dropdown-up .navbar-dropdown {
    border-bottom: 2px solid hsl(0deg, 0%, 86%);
    border-radius: 6px 6px 0 0;
    border-top: none;
    bottom: 100%;
    box-shadow: 0 -8px 8px rgba(10, 10, 10, 0.1);
    top: auto;
  }
  .navbar-item.is-active .navbar-dropdown,
  .navbar-item.is-hoverable:focus .navbar-dropdown,
  .navbar-item.is-hoverable:focus-within .navbar-dropdown,
  .navbar-item.is-hoverable:hover .navbar-dropdown {
    display: block;
  }
  .navbar.is-spaced .navbar-item.is-active .navbar-dropdown,
  .navbar-item.is-active .navbar-dropdown.is-boxed,
  .navbar.is-spaced .navbar-item.is-hoverable:focus .navbar-dropdown,
  .navbar-item.is-hoverable:focus .navbar-dropdown.is-boxed,
  .navbar.is-spaced .navbar-item.is-hoverable:focus-within .navbar-dropdown,
  .navbar-item.is-hoverable:focus-within .navbar-dropdown.is-boxed,
  .navbar.is-spaced .navbar-item.is-hoverable:hover .navbar-dropdown,
  .navbar-item.is-hoverable:hover .navbar-dropdown.is-boxed {
    opacity: 1;
    pointer-events: auto;
    transform: translateY(0);
  }
  .navbar-menu {
    flex-grow: 1;
    flex-shrink: 0;
  }
  .navbar-start {
    justify-content: flex-start;
    margin-right: auto;
  }
  .navbar-end {
    justify-content: flex-end;
    margin-left: auto;
  }
  .navbar-dropdown {
    background-color: hsl(0deg, 0%, 100%);
    border-bottom-left-radius: 6px;
    border-bottom-right-radius: 6px;
    border-top: 2px solid hsl(0deg, 0%, 86%);
    box-shadow: 0 8px 8px rgba(10, 10, 10, 0.1);
    display: none;
    font-size: 0.875rem;
    left: 0;
    min-width: 100%;
    position: absolute;
    top: 100%;
    z-index: 20;
  }
  .navbar-dropdown .navbar-item {
    padding: 0.375rem 1rem;
    white-space: nowrap;
  }
  .navbar-dropdown a.navbar-item {
    padding-right: 3rem;
  }
  .navbar-dropdown a.navbar-item:focus,
  .navbar-dropdown a.navbar-item:hover {
    background-color: hsl(0deg, 0%, 96%);
    color: hsl(0deg, 0%, 4%);
  }
  .navbar-dropdown a.navbar-item.is-active {
    background-color: hsl(0deg, 0%, 96%);
    color: hsl(229deg, 53%, 53%);
  }
  .navbar.is-spaced .navbar-dropdown,
  .navbar-dropdown.is-boxed {
    border-radius: 6px;
    border-top: none;
    box-shadow: 0 8px 8px rgba(10, 10, 10, 0.1), 0 0 0 1px rgba(10, 10, 10, 0.1);
    display: block;
    opacity: 0;
    pointer-events: none;
    top: calc(100% + (-4px));
    transform: translateY(-5px);
    transition-duration: 86ms;
    transition-property: opacity, transform;
  }
  .navbar-dropdown.is-right {
    left: auto;
    right: 0;
  }
  .navbar-divider {
    display: block;
  }
  .navbar > .container .navbar-brand,
  .container > .navbar .navbar-brand {
    margin-left: -0.75rem;
  }
  .navbar > .container .navbar-menu,
  .container > .navbar .navbar-menu {
    margin-right: -0.75rem;
  }
  .navbar.is-fixed-bottom-desktop,
  .navbar.is-fixed-top-desktop {
    left: 0;
    position: fixed;
    right: 0;
    z-index: 30;
  }
  .navbar.is-fixed-bottom-desktop {
    bottom: 0;
  }
  .navbar.is-fixed-bottom-desktop.has-shadow {
    box-shadow: 0 -2px 3px rgba(10, 10, 10, 0.1);
  }
  .navbar.is-fixed-top-desktop {
    top: 0;
  }
  html.has-navbar-fixed-top-desktop,
  body.has-navbar-fixed-top-desktop {
    padding-top: 3.25rem;
  }
  html.has-navbar-fixed-bottom-desktop,
  body.has-navbar-fixed-bottom-desktop {
    padding-bottom: 3.25rem;
  }
  html.has-spaced-navbar-fixed-top,
  body.has-spaced-navbar-fixed-top {
    padding-top: 5.25rem;
  }
  html.has-spaced-navbar-fixed-bottom,
  body.has-spaced-navbar-fixed-bottom {
    padding-bottom: 5.25rem;
  }
  a.navbar-item.is-active,
  .navbar-link.is-active {
    color: hsl(0deg, 0%, 4%);
  }
  a.navbar-item.is-active:not(:focus):not(:hover),
  .navbar-link.is-active:not(:focus):not(:hover) {
    background-color: transparent;
  }
  .navbar-item.has-dropdown:focus .navbar-link,
  .navbar-item.has-dropdown:hover .navbar-link,
  .navbar-item.has-dropdown.is-active .navbar-link {
    background-color: hsl(0deg, 0%, 98%);
  }
}
.hero.is-fullheight-with-navbar {
  min-height: calc(100vh - 3.25rem);
}
.pagination {
  font-size: 1rem;
  margin: -0.25rem;
}
.pagination.is-small {
  font-size: 0.75rem;
}
.pagination.is-medium {
  font-size: 1.25rem;
}
.pagination.is-large {
  font-size: 1.5rem;
}
.pagination.is-rounded .pagination-previous,
.pagination.is-rounded .pagination-next {
  padding-left: 1em;
  padding-right: 1em;
  border-radius: 9999px;
}
.pagination.is-rounded .pagination-link {
  border-radius: 9999px;
}
.pagination,
.pagination-list {
  align-items: center;
  display: flex;
  justify-content: center;
  text-align: center;
}
.pagination-previous,
.pagination-next,
.pagination-link,
.pagination-ellipsis {
  font-size: 1em;
  justify-content: center;
  margin: 0.25rem;
  padding-left: 0.5em;
  padding-right: 0.5em;
  text-align: center;
}
.pagination-previous,
.pagination-next,
.pagination-link {
  border-color: hsl(0deg, 0%, 86%);
  color: hsl(0deg, 0%, 21%);
  min-width: 2.5em;
}
.pagination-previous:hover,
.pagination-next:hover,
.pagination-link:hover {
  border-color: hsl(0deg, 0%, 71%);
  color: hsl(0deg, 0%, 21%);
}
.pagination-previous:focus,
.pagination-next:focus,
.pagination-link:focus {
  border-color: hsl(229deg, 53%, 53%);
}
.pagination-previous:active,
.pagination-next:active,
.pagination-link:active {
  box-shadow: inset 0 1px 2px rgba(10, 10, 10, 0.2);
}
.pagination-previous[disabled],
.pagination-previous.is-disabled,
.pagination-next[disabled],
.pagination-next.is-disabled,
.pagination-link[disabled],
.pagination-link.is-disabled {
  background-color: hsl(0deg, 0%, 86%);
  border-color: hsl(0deg, 0%, 86%);
  box-shadow: none;
  color: hsl(0deg, 0%, 48%);
  opacity: 0.5;
}
.pagination-previous,
.pagination-next {
  padding-left: 0.75em;
  padding-right: 0.75em;
  white-space: nowrap;
}
.pagination-link.is-current {
  background-color: hsl(229deg, 53%, 53%);
  border-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.pagination-ellipsis {
  color: hsl(0deg, 0%, 71%);
  pointer-events: none;
}
.pagination-list {
  flex-wrap: wrap;
}
.pagination-list li {
  list-style: none;
}
@media screen and (max-width: 768px) {
  .pagination {
    flex-wrap: wrap;
  }
  .pagination-previous,
  .pagination-next {
    flex-grow: 1;
    flex-shrink: 1;
  }
  .pagination-list li {
    flex-grow: 1;
    flex-shrink: 1;
  }
}
@media screen and (min-width: 769px), print {
  .pagination-list {
    flex-grow: 1;
    flex-shrink: 1;
    justify-content: flex-start;
    order: 1;
  }
  .pagination-previous,
  .pagination-next,
  .pagination-link,
  .pagination-ellipsis {
    margin-bottom: 0;
    margin-top: 0;
  }
  .pagination-previous {
    order: 2;
  }
  .pagination-next {
    order: 3;
  }
  .pagination {
    justify-content: space-between;
    margin-bottom: 0;
    margin-top: 0;
  }
  .pagination.is-centered .pagination-previous {
    order: 1;
  }
  .pagination.is-centered .pagination-list {
    justify-content: center;
    order: 2;
  }
  .pagination.is-centered .pagination-next {
    order: 3;
  }
  .pagination.is-right .pagination-previous {
    order: 1;
  }
  .pagination.is-right .pagination-next {
    order: 2;
  }
  .pagination.is-right .pagination-list {
    justify-content: flex-end;
    order: 3;
  }
}
.panel {
  border-radius: 6px;
  box-shadow: 0 0.5em 1em -0.125em rgba(10, 10, 10, 0.1), 0 0px 0 1px rgba(10, 10, 10, 0.02);
  font-size: 1rem;
}
.panel:not(:last-child) {
  margin-bottom: 1.5rem;
}
.panel.is-white .panel-heading {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.panel.is-white .panel-tabs a.is-active {
  border-bottom-color: hsl(0deg, 0%, 100%);
}
.panel.is-white .panel-block.is-active .panel-icon {
  color: hsl(0deg, 0%, 100%);
}
.panel.is-black .panel-heading {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.panel.is-black .panel-tabs a.is-active {
  border-bottom-color: hsl(0deg, 0%, 4%);
}
.panel.is-black .panel-block.is-active .panel-icon {
  color: hsl(0deg, 0%, 4%);
}
.panel.is-light .panel-heading {
  background-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.panel.is-light .panel-tabs a.is-active {
  border-bottom-color: hsl(0deg, 0%, 96%);
}
.panel.is-light .panel-block.is-active .panel-icon {
  color: hsl(0deg, 0%, 96%);
}
.panel.is-dark .panel-heading {
  background-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.panel.is-dark .panel-tabs a.is-active {
  border-bottom-color: hsl(0deg, 0%, 21%);
}
.panel.is-dark .panel-block.is-active .panel-icon {
  color: hsl(0deg, 0%, 21%);
}
.panel.is-primary .panel-heading {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.panel.is-primary .panel-tabs a.is-active {
  border-bottom-color: hsl(171deg, 100%, 41%);
}
.panel.is-primary .panel-block.is-active .panel-icon {
  color: hsl(171deg, 100%, 41%);
}
.panel.is-link .panel-heading {
  background-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.panel.is-link .panel-tabs a.is-active {
  border-bottom-color: hsl(229deg, 53%, 53%);
}
.panel.is-link .panel-block.is-active .panel-icon {
  color: hsl(229deg, 53%, 53%);
}
.panel.is-info .panel-heading {
  background-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.panel.is-info .panel-tabs a.is-active {
  border-bottom-color: hsl(207deg, 61%, 53%);
}
.panel.is-info .panel-block.is-active .panel-icon {
  color: hsl(207deg, 61%, 53%);
}
.panel.is-success .panel-heading {
  background-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.panel.is-success .panel-tabs a.is-active {
  border-bottom-color: hsl(153deg, 53%, 53%);
}
.panel.is-success .panel-block.is-active .panel-icon {
  color: hsl(153deg, 53%, 53%);
}
.panel.is-warning .panel-heading {
  background-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.panel.is-warning .panel-tabs a.is-active {
  border-bottom-color: hsl(44deg, 100%, 77%);
}
.panel.is-warning .panel-block.is-active .panel-icon {
  color: hsl(44deg, 100%, 77%);
}
.panel.is-danger .panel-heading {
  background-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.panel.is-danger .panel-tabs a.is-active {
  border-bottom-color: hsl(348deg, 86%, 61%);
}
.panel.is-danger .panel-block.is-active .panel-icon {
  color: hsl(348deg, 86%, 61%);
}
.panel-tabs:not(:last-child),
.panel-block:not(:last-child) {
  border-bottom: 1px solid hsl(0deg, 0%, 93%);
}
.panel-heading {
  background-color: hsl(0deg, 0%, 93%);
  border-radius: 6px 6px 0 0;
  color: hsl(0deg, 0%, 21%);
  font-size: 1.25em;
  font-weight: 700;
  line-height: 1.25;
  padding: 0.75em 1em;
}
.panel-tabs {
  align-items: flex-end;
  display: flex;
  font-size: 0.875em;
  justify-content: center;
}
.panel-tabs a {
  border-bottom: 1px solid hsl(0deg, 0%, 86%);
  margin-bottom: -1px;
  padding: 0.5em;
}
.panel-tabs a.is-active {
  border-bottom-color: hsl(0deg, 0%, 29%);
  color: hsl(0deg, 0%, 21%);
}
.panel-list a {
  color: hsl(0deg, 0%, 29%);
}
.panel-list a:hover {
  color: hsl(229deg, 53%, 53%);
}
.panel-block {
  align-items: center;
  color: hsl(0deg, 0%, 21%);
  display: flex;
  justify-content: flex-start;
  padding: 0.5em 0.75em;
}
.panel-block input[type=checkbox] {
  margin-right: 0.75em;
}
.panel-block > .control {
  flex-grow: 1;
  flex-shrink: 1;
  width: 100%;
}
.panel-block.is-wrapped {
  flex-wrap: wrap;
}
.panel-block.is-active {
  border-left-color: hsl(229deg, 53%, 53%);
  color: hsl(0deg, 0%, 21%);
}
.panel-block.is-active .panel-icon {
  color: hsl(229deg, 53%, 53%);
}
.panel-block:last-child {
  border-bottom-left-radius: 6px;
  border-bottom-right-radius: 6px;
}
a.panel-block,
label.panel-block {
  cursor: pointer;
}
a.panel-block:hover,
label.panel-block:hover {
  background-color: hsl(0deg, 0%, 96%);
}
.panel-icon {
  display: inline-block;
  font-size: 14px;
  height: 1em;
  line-height: 1em;
  text-align: center;
  vertical-align: top;
  width: 1em;
  color: hsl(0deg, 0%, 48%);
  margin-right: 0.75em;
}
.panel-icon .fa {
  font-size: inherit;
  line-height: inherit;
}
.tabs {
  -webkit-overflow-scrolling: touch;
  align-items: stretch;
  display: flex;
  font-size: 1rem;
  justify-content: space-between;
  overflow: hidden;
  overflow-x: auto;
  white-space: nowrap;
}
.tabs a {
  align-items: center;
  border-bottom-color: hsl(0deg, 0%, 86%);
  border-bottom-style: solid;
  border-bottom-width: 1px;
  color: hsl(0deg, 0%, 29%);
  display: flex;
  justify-content: center;
  margin-bottom: -1px;
  padding: 0.5em 1em;
  vertical-align: top;
}
.tabs a:hover {
  border-bottom-color: hsl(0deg, 0%, 21%);
  color: hsl(0deg, 0%, 21%);
}
.tabs li {
  display: block;
}
.tabs li.is-active a {
  border-bottom-color: hsl(229deg, 53%, 53%);
  color: hsl(229deg, 53%, 53%);
}
.tabs ul {
  align-items: center;
  border-bottom-color: hsl(0deg, 0%, 86%);
  border-bottom-style: solid;
  border-bottom-width: 1px;
  display: flex;
  flex-grow: 1;
  flex-shrink: 0;
  justify-content: flex-start;
}
.tabs ul.is-left {
  padding-right: 0.75em;
}
.tabs ul.is-center {
  flex: none;
  justify-content: center;
  padding-left: 0.75em;
  padding-right: 0.75em;
}
.tabs ul.is-right {
  justify-content: flex-end;
  padding-left: 0.75em;
}
.tabs .icon:first-child {
  margin-right: 0.5em;
}
.tabs .icon:last-child {
  margin-left: 0.5em;
}
.tabs.is-centered ul {
  justify-content: center;
}
.tabs.is-right ul {
  justify-content: flex-end;
}
.tabs.is-boxed a {
  border: 1px solid transparent;
  border-radius: 4px 4px 0 0;
}
.tabs.is-boxed a:hover {
  background-color: hsl(0deg, 0%, 96%);
  border-bottom-color: hsl(0deg, 0%, 86%);
}
.tabs.is-boxed li.is-active a {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 86%);
  border-bottom-color: transparent !important;
}
.tabs.is-fullwidth li {
  flex-grow: 1;
  flex-shrink: 0;
}
.tabs.is-toggle a {
  border-color: hsl(0deg, 0%, 86%);
  border-style: solid;
  border-width: 1px;
  margin-bottom: 0;
  position: relative;
}
.tabs.is-toggle a:hover {
  background-color: hsl(0deg, 0%, 96%);
  border-color: hsl(0deg, 0%, 71%);
  z-index: 2;
}
.tabs.is-toggle li + li {
  margin-left: -1px;
}
.tabs.is-toggle li:first-child a {
  border-top-left-radius: 4px;
  border-bottom-left-radius: 4px;
}
.tabs.is-toggle li:last-child a {
  border-top-right-radius: 4px;
  border-bottom-right-radius: 4px;
}
.tabs.is-toggle li.is-active a {
  background-color: hsl(229deg, 53%, 53%);
  border-color: hsl(229deg, 53%, 53%);
  color: #fff;
  z-index: 1;
}
.tabs.is-toggle ul {
  border-bottom: none;
}
.tabs.is-toggle.is-toggle-rounded li:first-child a {
  border-bottom-left-radius: 9999px;
  border-top-left-radius: 9999px;
  padding-left: 1.25em;
}
.tabs.is-toggle.is-toggle-rounded li:last-child a {
  border-bottom-right-radius: 9999px;
  border-top-right-radius: 9999px;
  padding-right: 1.25em;
}
.tabs.is-small {
  font-size: 0.75rem;
}
.tabs.is-medium {
  font-size: 1.25rem;
}
.tabs.is-large {
  font-size: 1.5rem;
}
.column {
  display: block;
  flex-basis: 0;
  flex-grow: 1;
  flex-shrink: 1;
  padding: 0.75rem;
}
.columns.is-mobile > .column.is-narrow {
  flex: none;
  width: unset;
}
.columns.is-mobile > .column.is-full {
  flex: none;
  width: 100%;
}
.columns.is-mobile > .column.is-three-quarters {
  flex: none;
  width: 75%;
}
.columns.is-mobile > .column.is-two-thirds {
  flex: none;
  width: 66.6666%;
}
.columns.is-mobile > .column.is-half {
  flex: none;
  width: 50%;
}
.columns.is-mobile > .column.is-one-third {
  flex: none;
  width: 33.3333%;
}
.columns.is-mobile > .column.is-one-quarter {
  flex: none;
  width: 25%;
}
.columns.is-mobile > .column.is-one-fifth {
  flex: none;
  width: 20%;
}
.columns.is-mobile > .column.is-two-fifths {
  flex: none;
  width: 40%;
}
.columns.is-mobile > .column.is-three-fifths {
  flex: none;
  width: 60%;
}
.columns.is-mobile > .column.is-four-fifths {
  flex: none;
  width: 80%;
}
.columns.is-mobile > .column.is-offset-three-quarters {
  margin-left: 75%;
}
.columns.is-mobile > .column.is-offset-two-thirds {
  margin-left: 66.6666%;
}
.columns.is-mobile > .column.is-offset-half {
  margin-left: 50%;
}
.columns.is-mobile > .column.is-offset-one-third {
  margin-left: 33.3333%;
}
.columns.is-mobile > .column.is-offset-one-quarter {
  margin-left: 25%;
}
.columns.is-mobile > .column.is-offset-one-fifth {
  margin-left: 20%;
}
.columns.is-mobile > .column.is-offset-two-fifths {
  margin-left: 40%;
}
.columns.is-mobile > .column.is-offset-three-fifths {
  margin-left: 60%;
}
.columns.is-mobile > .column.is-offset-four-fifths {
  margin-left: 80%;
}
.columns.is-mobile > .column.is-0 {
  flex: none;
  width: 0%;
}
.columns.is-mobile > .column.is-offset-0 {
  margin-left: 0%;
}
.columns.is-mobile > .column.is-1 {
  flex: none;
  width: 8.33333337%;
}
.columns.is-mobile > .column.is-offset-1 {
  margin-left: 8.33333337%;
}
.columns.is-mobile > .column.is-2 {
  flex: none;
  width: 16.66666674%;
}
.columns.is-mobile > .column.is-offset-2 {
  margin-left: 16.66666674%;
}
.columns.is-mobile > .column.is-3 {
  flex: none;
  width: 25%;
}
.columns.is-mobile > .column.is-offset-3 {
  margin-left: 25%;
}
.columns.is-mobile > .column.is-4 {
  flex: none;
  width: 33.33333337%;
}
.columns.is-mobile > .column.is-offset-4 {
  margin-left: 33.33333337%;
}
.columns.is-mobile > .column.is-5 {
  flex: none;
  width: 41.66666674%;
}
.columns.is-mobile > .column.is-offset-5 {
  margin-left: 41.66666674%;
}
.columns.is-mobile > .column.is-6 {
  flex: none;
  width: 50%;
}
.columns.is-mobile > .column.is-offset-6 {
  margin-left: 50%;
}
.columns.is-mobile > .column.is-7 {
  flex: none;
  width: 58.33333337%;
}
.columns.is-mobile > .column.is-offset-7 {
  margin-left: 58.33333337%;
}
.columns.is-mobile > .column.is-8 {
  flex: none;
  width: 66.66666674%;
}
.columns.is-mobile > .column.is-offset-8 {
  margin-left: 66.66666674%;
}
.columns.is-mobile > .column.is-9 {
  flex: none;
  width: 75%;
}
.columns.is-mobile > .column.is-offset-9 {
  margin-left: 75%;
}
.columns.is-mobile > .column.is-10 {
  flex: none;
  width: 83.33333337%;
}
.columns.is-mobile > .column.is-offset-10 {
  margin-left: 83.33333337%;
}
.columns.is-mobile > .column.is-11 {
  flex: none;
  width: 91.66666674%;
}
.columns.is-mobile > .column.is-offset-11 {
  margin-left: 91.66666674%;
}
.columns.is-mobile > .column.is-12 {
  flex: none;
  width: 100%;
}
.columns.is-mobile > .column.is-offset-12 {
  margin-left: 100%;
}
@media screen and (max-width: 768px) {
  .column.is-narrow-mobile {
    flex: none;
    width: unset;
  }
  .column.is-full-mobile {
    flex: none;
    width: 100%;
  }
  .column.is-three-quarters-mobile {
    flex: none;
    width: 75%;
  }
  .column.is-two-thirds-mobile {
    flex: none;
    width: 66.6666%;
  }
  .column.is-half-mobile {
    flex: none;
    width: 50%;
  }
  .column.is-one-third-mobile {
    flex: none;
    width: 33.3333%;
  }
  .column.is-one-quarter-mobile {
    flex: none;
    width: 25%;
  }
  .column.is-one-fifth-mobile {
    flex: none;
    width: 20%;
  }
  .column.is-two-fifths-mobile {
    flex: none;
    width: 40%;
  }
  .column.is-three-fifths-mobile {
    flex: none;
    width: 60%;
  }
  .column.is-four-fifths-mobile {
    flex: none;
    width: 80%;
  }
  .column.is-offset-three-quarters-mobile {
    margin-left: 75%;
  }
  .column.is-offset-two-thirds-mobile {
    margin-left: 66.6666%;
  }
  .column.is-offset-half-mobile {
    margin-left: 50%;
  }
  .column.is-offset-one-third-mobile {
    margin-left: 33.3333%;
  }
  .column.is-offset-one-quarter-mobile {
    margin-left: 25%;
  }
  .column.is-offset-one-fifth-mobile {
    margin-left: 20%;
  }
  .column.is-offset-two-fifths-mobile {
    margin-left: 40%;
  }
  .column.is-offset-three-fifths-mobile {
    margin-left: 60%;
  }
  .column.is-offset-four-fifths-mobile {
    margin-left: 80%;
  }
  .column.is-0-mobile {
    flex: none;
    width: 0%;
  }
  .column.is-offset-0-mobile {
    margin-left: 0%;
  }
  .column.is-1-mobile {
    flex: none;
    width: 8.33333337%;
  }
  .column.is-offset-1-mobile {
    margin-left: 8.33333337%;
  }
  .column.is-2-mobile {
    flex: none;
    width: 16.66666674%;
  }
  .column.is-offset-2-mobile {
    margin-left: 16.66666674%;
  }
  .column.is-3-mobile {
    flex: none;
    width: 25%;
  }
  .column.is-offset-3-mobile {
    margin-left: 25%;
  }
  .column.is-4-mobile {
    flex: none;
    width: 33.33333337%;
  }
  .column.is-offset-4-mobile {
    margin-left: 33.33333337%;
  }
  .column.is-5-mobile {
    flex: none;
    width: 41.66666674%;
  }
  .column.is-offset-5-mobile {
    margin-left: 41.66666674%;
  }
  .column.is-6-mobile {
    flex: none;
    width: 50%;
  }
  .column.is-offset-6-mobile {
    margin-left: 50%;
  }
  .column.is-7-mobile {
    flex: none;
    width: 58.33333337%;
  }
  .column.is-offset-7-mobile {
    margin-left: 58.33333337%;
  }
  .column.is-8-mobile {
    flex: none;
    width: 66.66666674%;
  }
  .column.is-offset-8-mobile {
    margin-left: 66.66666674%;
  }
  .column.is-9-mobile {
    flex: none;
    width: 75%;
  }
  .column.is-offset-9-mobile {
    margin-left: 75%;
  }
  .column.is-10-mobile {
    flex: none;
    width: 83.33333337%;
  }
  .column.is-offset-10-mobile {
    margin-left: 83.33333337%;
  }
  .column.is-11-mobile {
    flex: none;
    width: 91.66666674%;
  }
  .column.is-offset-11-mobile {
    margin-left: 91.66666674%;
  }
  .column.is-12-mobile {
    flex: none;
    width: 100%;
  }
  .column.is-offset-12-mobile {
    margin-left: 100%;
  }
}
@media screen and (min-width: 769px), print {
  .column.is-narrow,
  .column.is-narrow-tablet {
    flex: none;
    width: unset;
  }
  .column.is-full,
  .column.is-full-tablet {
    flex: none;
    width: 100%;
  }
  .column.is-three-quarters,
  .column.is-three-quarters-tablet {
    flex: none;
    width: 75%;
  }
  .column.is-two-thirds,
  .column.is-two-thirds-tablet {
    flex: none;
    width: 66.6666%;
  }
  .column.is-half,
  .column.is-half-tablet {
    flex: none;
    width: 50%;
  }
  .column.is-one-third,
  .column.is-one-third-tablet {
    flex: none;
    width: 33.3333%;
  }
  .column.is-one-quarter,
  .column.is-one-quarter-tablet {
    flex: none;
    width: 25%;
  }
  .column.is-one-fifth,
  .column.is-one-fifth-tablet {
    flex: none;
    width: 20%;
  }
  .column.is-two-fifths,
  .column.is-two-fifths-tablet {
    flex: none;
    width: 40%;
  }
  .column.is-three-fifths,
  .column.is-three-fifths-tablet {
    flex: none;
    width: 60%;
  }
  .column.is-four-fifths,
  .column.is-four-fifths-tablet {
    flex: none;
    width: 80%;
  }
  .column.is-offset-three-quarters,
  .column.is-offset-three-quarters-tablet {
    margin-left: 75%;
  }
  .column.is-offset-two-thirds,
  .column.is-offset-two-thirds-tablet {
    margin-left: 66.6666%;
  }
  .column.is-offset-half,
  .column.is-offset-half-tablet {
    margin-left: 50%;
  }
  .column.is-offset-one-third,
  .column.is-offset-one-third-tablet {
    margin-left: 33.3333%;
  }
  .column.is-offset-one-quarter,
  .column.is-offset-one-quarter-tablet {
    margin-left: 25%;
  }
  .column.is-offset-one-fifth,
  .column.is-offset-one-fifth-tablet {
    margin-left: 20%;
  }
  .column.is-offset-two-fifths,
  .column.is-offset-two-fifths-tablet {
    margin-left: 40%;
  }
  .column.is-offset-three-fifths,
  .column.is-offset-three-fifths-tablet {
    margin-left: 60%;
  }
  .column.is-offset-four-fifths,
  .column.is-offset-four-fifths-tablet {
    margin-left: 80%;
  }
  .column.is-0,
  .column.is-0-tablet {
    flex: none;
    width: 0%;
  }
  .column.is-offset-0,
  .column.is-offset-0-tablet {
    margin-left: 0%;
  }
  .column.is-1,
  .column.is-1-tablet {
    flex: none;
    width: 8.33333337%;
  }
  .column.is-offset-1,
  .column.is-offset-1-tablet {
    margin-left: 8.33333337%;
  }
  .column.is-2,
  .column.is-2-tablet {
    flex: none;
    width: 16.66666674%;
  }
  .column.is-offset-2,
  .column.is-offset-2-tablet {
    margin-left: 16.66666674%;
  }
  .column.is-3,
  .column.is-3-tablet {
    flex: none;
    width: 25%;
  }
  .column.is-offset-3,
  .column.is-offset-3-tablet {
    margin-left: 25%;
  }
  .column.is-4,
  .column.is-4-tablet {
    flex: none;
    width: 33.33333337%;
  }
  .column.is-offset-4,
  .column.is-offset-4-tablet {
    margin-left: 33.33333337%;
  }
  .column.is-5,
  .column.is-5-tablet {
    flex: none;
    width: 41.66666674%;
  }
  .column.is-offset-5,
  .column.is-offset-5-tablet {
    margin-left: 41.66666674%;
  }
  .column.is-6,
  .column.is-6-tablet {
    flex: none;
    width: 50%;
  }
  .column.is-offset-6,
  .column.is-offset-6-tablet {
    margin-left: 50%;
  }
  .column.is-7,
  .column.is-7-tablet {
    flex: none;
    width: 58.33333337%;
  }
  .column.is-offset-7,
  .column.is-offset-7-tablet {
    margin-left: 58.33333337%;
  }
  .column.is-8,
  .column.is-8-tablet {
    flex: none;
    width: 66.66666674%;
  }
  .column.is-offset-8,
  .column.is-offset-8-tablet {
    margin-left: 66.66666674%;
  }
  .column.is-9,
  .column.is-9-tablet {
    flex: none;
    width: 75%;
  }
  .column.is-offset-9,
  .column.is-offset-9-tablet {
    margin-left: 75%;
  }
  .column.is-10,
  .column.is-10-tablet {
    flex: none;
    width: 83.33333337%;
  }
  .column.is-offset-10,
  .column.is-offset-10-tablet {
    margin-left: 83.33333337%;
  }
  .column.is-11,
  .column.is-11-tablet {
    flex: none;
    width: 91.66666674%;
  }
  .column.is-offset-11,
  .column.is-offset-11-tablet {
    margin-left: 91.66666674%;
  }
  .column.is-12,
  .column.is-12-tablet {
    flex: none;
    width: 100%;
  }
  .column.is-offset-12,
  .column.is-offset-12-tablet {
    margin-left: 100%;
  }
}
@media screen and (max-width: 1023px) {
  .column.is-narrow-touch {
    flex: none;
    width: unset;
  }
  .column.is-full-touch {
    flex: none;
    width: 100%;
  }
  .column.is-three-quarters-touch {
    flex: none;
    width: 75%;
  }
  .column.is-two-thirds-touch {
    flex: none;
    width: 66.6666%;
  }
  .column.is-half-touch {
    flex: none;
    width: 50%;
  }
  .column.is-one-third-touch {
    flex: none;
    width: 33.3333%;
  }
  .column.is-one-quarter-touch {
    flex: none;
    width: 25%;
  }
  .column.is-one-fifth-touch {
    flex: none;
    width: 20%;
  }
  .column.is-two-fifths-touch {
    flex: none;
    width: 40%;
  }
  .column.is-three-fifths-touch {
    flex: none;
    width: 60%;
  }
  .column.is-four-fifths-touch {
    flex: none;
    width: 80%;
  }
  .column.is-offset-three-quarters-touch {
    margin-left: 75%;
  }
  .column.is-offset-two-thirds-touch {
    margin-left: 66.6666%;
  }
  .column.is-offset-half-touch {
    margin-left: 50%;
  }
  .column.is-offset-one-third-touch {
    margin-left: 33.3333%;
  }
  .column.is-offset-one-quarter-touch {
    margin-left: 25%;
  }
  .column.is-offset-one-fifth-touch {
    margin-left: 20%;
  }
  .column.is-offset-two-fifths-touch {
    margin-left: 40%;
  }
  .column.is-offset-three-fifths-touch {
    margin-left: 60%;
  }
  .column.is-offset-four-fifths-touch {
    margin-left: 80%;
  }
  .column.is-0-touch {
    flex: none;
    width: 0%;
  }
  .column.is-offset-0-touch {
    margin-left: 0%;
  }
  .column.is-1-touch {
    flex: none;
    width: 8.33333337%;
  }
  .column.is-offset-1-touch {
    margin-left: 8.33333337%;
  }
  .column.is-2-touch {
    flex: none;
    width: 16.66666674%;
  }
  .column.is-offset-2-touch {
    margin-left: 16.66666674%;
  }
  .column.is-3-touch {
    flex: none;
    width: 25%;
  }
  .column.is-offset-3-touch {
    margin-left: 25%;
  }
  .column.is-4-touch {
    flex: none;
    width: 33.33333337%;
  }
  .column.is-offset-4-touch {
    margin-left: 33.33333337%;
  }
  .column.is-5-touch {
    flex: none;
    width: 41.66666674%;
  }
  .column.is-offset-5-touch {
    margin-left: 41.66666674%;
  }
  .column.is-6-touch {
    flex: none;
    width: 50%;
  }
  .column.is-offset-6-touch {
    margin-left: 50%;
  }
  .column.is-7-touch {
    flex: none;
    width: 58.33333337%;
  }
  .column.is-offset-7-touch {
    margin-left: 58.33333337%;
  }
  .column.is-8-touch {
    flex: none;
    width: 66.66666674%;
  }
  .column.is-offset-8-touch {
    margin-left: 66.66666674%;
  }
  .column.is-9-touch {
    flex: none;
    width: 75%;
  }
  .column.is-offset-9-touch {
    margin-left: 75%;
  }
  .column.is-10-touch {
    flex: none;
    width: 83.33333337%;
  }
  .column.is-offset-10-touch {
    margin-left: 83.33333337%;
  }
  .column.is-11-touch {
    flex: none;
    width: 91.66666674%;
  }
  .column.is-offset-11-touch {
    margin-left: 91.66666674%;
  }
  .column.is-12-touch {
    flex: none;
    width: 100%;
  }
  .column.is-offset-12-touch {
    margin-left: 100%;
  }
}
@media screen and (min-width: 1024px) {
  .column.is-narrow-desktop {
    flex: none;
    width: unset;
  }
  .column.is-full-desktop {
    flex: none;
    width: 100%;
  }
  .column.is-three-quarters-desktop {
    flex: none;
    width: 75%;
  }
  .column.is-two-thirds-desktop {
    flex: none;
    width: 66.6666%;
  }
  .column.is-half-desktop {
    flex: none;
    width: 50%;
  }
  .column.is-one-third-desktop {
    flex: none;
    width: 33.3333%;
  }
  .column.is-one-quarter-desktop {
    flex: none;
    width: 25%;
  }
  .column.is-one-fifth-desktop {
    flex: none;
    width: 20%;
  }
  .column.is-two-fifths-desktop {
    flex: none;
    width: 40%;
  }
  .column.is-three-fifths-desktop {
    flex: none;
    width: 60%;
  }
  .column.is-four-fifths-desktop {
    flex: none;
    width: 80%;
  }
  .column.is-offset-three-quarters-desktop {
    margin-left: 75%;
  }
  .column.is-offset-two-thirds-desktop {
    margin-left: 66.6666%;
  }
  .column.is-offset-half-desktop {
    margin-left: 50%;
  }
  .column.is-offset-one-third-desktop {
    margin-left: 33.3333%;
  }
  .column.is-offset-one-quarter-desktop {
    margin-left: 25%;
  }
  .column.is-offset-one-fifth-desktop {
    margin-left: 20%;
  }
  .column.is-offset-two-fifths-desktop {
    margin-left: 40%;
  }
  .column.is-offset-three-fifths-desktop {
    margin-left: 60%;
  }
  .column.is-offset-four-fifths-desktop {
    margin-left: 80%;
  }
  .column.is-0-desktop {
    flex: none;
    width: 0%;
  }
  .column.is-offset-0-desktop {
    margin-left: 0%;
  }
  .column.is-1-desktop {
    flex: none;
    width: 8.33333337%;
  }
  .column.is-offset-1-desktop {
    margin-left: 8.33333337%;
  }
  .column.is-2-desktop {
    flex: none;
    width: 16.66666674%;
  }
  .column.is-offset-2-desktop {
    margin-left: 16.66666674%;
  }
  .column.is-3-desktop {
    flex: none;
    width: 25%;
  }
  .column.is-offset-3-desktop {
    margin-left: 25%;
  }
  .column.is-4-desktop {
    flex: none;
    width: 33.33333337%;
  }
  .column.is-offset-4-desktop {
    margin-left: 33.33333337%;
  }
  .column.is-5-desktop {
    flex: none;
    width: 41.66666674%;
  }
  .column.is-offset-5-desktop {
    margin-left: 41.66666674%;
  }
  .column.is-6-desktop {
    flex: none;
    width: 50%;
  }
  .column.is-offset-6-desktop {
    margin-left: 50%;
  }
  .column.is-7-desktop {
    flex: none;
    width: 58.33333337%;
  }
  .column.is-offset-7-desktop {
    margin-left: 58.33333337%;
  }
  .column.is-8-desktop {
    flex: none;
    width: 66.66666674%;
  }
  .column.is-offset-8-desktop {
    margin-left: 66.66666674%;
  }
  .column.is-9-desktop {
    flex: none;
    width: 75%;
  }
  .column.is-offset-9-desktop {
    margin-left: 75%;
  }
  .column.is-10-desktop {
    flex: none;
    width: 83.33333337%;
  }
  .column.is-offset-10-desktop {
    margin-left: 83.33333337%;
  }
  .column.is-11-desktop {
    flex: none;
    width: 91.66666674%;
  }
  .column.is-offset-11-desktop {
    margin-left: 91.66666674%;
  }
  .column.is-12-desktop {
    flex: none;
    width: 100%;
  }
  .column.is-offset-12-desktop {
    margin-left: 100%;
  }
}
@media screen and (min-width: 1216px) {
  .column.is-narrow-widescreen {
    flex: none;
    width: unset;
  }
  .column.is-full-widescreen {
    flex: none;
    width: 100%;
  }
  .column.is-three-quarters-widescreen {
    flex: none;
    width: 75%;
  }
  .column.is-two-thirds-widescreen {
    flex: none;
    width: 66.6666%;
  }
  .column.is-half-widescreen {
    flex: none;
    width: 50%;
  }
  .column.is-one-third-widescreen {
    flex: none;
    width: 33.3333%;
  }
  .column.is-one-quarter-widescreen {
    flex: none;
    width: 25%;
  }
  .column.is-one-fifth-widescreen {
    flex: none;
    width: 20%;
  }
  .column.is-two-fifths-widescreen {
    flex: none;
    width: 40%;
  }
  .column.is-three-fifths-widescreen {
    flex: none;
    width: 60%;
  }
  .column.is-four-fifths-widescreen {
    flex: none;
    width: 80%;
  }
  .column.is-offset-three-quarters-widescreen {
    margin-left: 75%;
  }
  .column.is-offset-two-thirds-widescreen {
    margin-left: 66.6666%;
  }
  .column.is-offset-half-widescreen {
    margin-left: 50%;
  }
  .column.is-offset-one-third-widescreen {
    margin-left: 33.3333%;
  }
  .column.is-offset-one-quarter-widescreen {
    margin-left: 25%;
  }
  .column.is-offset-one-fifth-widescreen {
    margin-left: 20%;
  }
  .column.is-offset-two-fifths-widescreen {
    margin-left: 40%;
  }
  .column.is-offset-three-fifths-widescreen {
    margin-left: 60%;
  }
  .column.is-offset-four-fifths-widescreen {
    margin-left: 80%;
  }
  .column.is-0-widescreen {
    flex: none;
    width: 0%;
  }
  .column.is-offset-0-widescreen {
    margin-left: 0%;
  }
  .column.is-1-widescreen {
    flex: none;
    width: 8.33333337%;
  }
  .column.is-offset-1-widescreen {
    margin-left: 8.33333337%;
  }
  .column.is-2-widescreen {
    flex: none;
    width: 16.66666674%;
  }
  .column.is-offset-2-widescreen {
    margin-left: 16.66666674%;
  }
  .column.is-3-widescreen {
    flex: none;
    width: 25%;
  }
  .column.is-offset-3-widescreen {
    margin-left: 25%;
  }
  .column.is-4-widescreen {
    flex: none;
    width: 33.33333337%;
  }
  .column.is-offset-4-widescreen {
    margin-left: 33.33333337%;
  }
  .column.is-5-widescreen {
    flex: none;
    width: 41.66666674%;
  }
  .column.is-offset-5-widescreen {
    margin-left: 41.66666674%;
  }
  .column.is-6-widescreen {
    flex: none;
    width: 50%;
  }
  .column.is-offset-6-widescreen {
    margin-left: 50%;
  }
  .column.is-7-widescreen {
    flex: none;
    width: 58.33333337%;
  }
  .column.is-offset-7-widescreen {
    margin-left: 58.33333337%;
  }
  .column.is-8-widescreen {
    flex: none;
    width: 66.66666674%;
  }
  .column.is-offset-8-widescreen {
    margin-left: 66.66666674%;
  }
  .column.is-9-widescreen {
    flex: none;
    width: 75%;
  }
  .column.is-offset-9-widescreen {
    margin-left: 75%;
  }
  .column.is-10-widescreen {
    flex: none;
    width: 83.33333337%;
  }
  .column.is-offset-10-widescreen {
    margin-left: 83.33333337%;
  }
  .column.is-11-widescreen {
    flex: none;
    width: 91.66666674%;
  }
  .column.is-offset-11-widescreen {
    margin-left: 91.66666674%;
  }
  .column.is-12-widescreen {
    flex: none;
    width: 100%;
  }
  .column.is-offset-12-widescreen {
    margin-left: 100%;
  }
}
@media screen and (min-width: 1408px) {
  .column.is-narrow-fullhd {
    flex: none;
    width: unset;
  }
  .column.is-full-fullhd {
    flex: none;
    width: 100%;
  }
  .column.is-three-quarters-fullhd {
    flex: none;
    width: 75%;
  }
  .column.is-two-thirds-fullhd {
    flex: none;
    width: 66.6666%;
  }
  .column.is-half-fullhd {
    flex: none;
    width: 50%;
  }
  .column.is-one-third-fullhd {
    flex: none;
    width: 33.3333%;
  }
  .column.is-one-quarter-fullhd {
    flex: none;
    width: 25%;
  }
  .column.is-one-fifth-fullhd {
    flex: none;
    width: 20%;
  }
  .column.is-two-fifths-fullhd {
    flex: none;
    width: 40%;
  }
  .column.is-three-fifths-fullhd {
    flex: none;
    width: 60%;
  }
  .column.is-four-fifths-fullhd {
    flex: none;
    width: 80%;
  }
  .column.is-offset-three-quarters-fullhd {
    margin-left: 75%;
  }
  .column.is-offset-two-thirds-fullhd {
    margin-left: 66.6666%;
  }
  .column.is-offset-half-fullhd {
    margin-left: 50%;
  }
  .column.is-offset-one-third-fullhd {
    margin-left: 33.3333%;
  }
  .column.is-offset-one-quarter-fullhd {
    margin-left: 25%;
  }
  .column.is-offset-one-fifth-fullhd {
    margin-left: 20%;
  }
  .column.is-offset-two-fifths-fullhd {
    margin-left: 40%;
  }
  .column.is-offset-three-fifths-fullhd {
    margin-left: 60%;
  }
  .column.is-offset-four-fifths-fullhd {
    margin-left: 80%;
  }
  .column.is-0-fullhd {
    flex: none;
    width: 0%;
  }
  .column.is-offset-0-fullhd {
    margin-left: 0%;
  }
  .column.is-1-fullhd {
    flex: none;
    width: 8.33333337%;
  }
  .column.is-offset-1-fullhd {
    margin-left: 8.33333337%;
  }
  .column.is-2-fullhd {
    flex: none;
    width: 16.66666674%;
  }
  .column.is-offset-2-fullhd {
    margin-left: 16.66666674%;
  }
  .column.is-3-fullhd {
    flex: none;
    width: 25%;
  }
  .column.is-offset-3-fullhd {
    margin-left: 25%;
  }
  .column.is-4-fullhd {
    flex: none;
    width: 33.33333337%;
  }
  .column.is-offset-4-fullhd {
    margin-left: 33.33333337%;
  }
  .column.is-5-fullhd {
    flex: none;
    width: 41.66666674%;
  }
  .column.is-offset-5-fullhd {
    margin-left: 41.66666674%;
  }
  .column.is-6-fullhd {
    flex: none;
    width: 50%;
  }
  .column.is-offset-6-fullhd {
    margin-left: 50%;
  }
  .column.is-7-fullhd {
    flex: none;
    width: 58.33333337%;
  }
  .column.is-offset-7-fullhd {
    margin-left: 58.33333337%;
  }
  .column.is-8-fullhd {
    flex: none;
    width: 66.66666674%;
  }
  .column.is-offset-8-fullhd {
    margin-left: 66.66666674%;
  }
  .column.is-9-fullhd {
    flex: none;
    width: 75%;
  }
  .column.is-offset-9-fullhd {
    margin-left: 75%;
  }
  .column.is-10-fullhd {
    flex: none;
    width: 83.33333337%;
  }
  .column.is-offset-10-fullhd {
    margin-left: 83.33333337%;
  }
  .column.is-11-fullhd {
    flex: none;
    width: 91.66666674%;
  }
  .column.is-offset-11-fullhd {
    margin-left: 91.66666674%;
  }
  .column.is-12-fullhd {
    flex: none;
    width: 100%;
  }
  .column.is-offset-12-fullhd {
    margin-left: 100%;
  }
}
.columns {
  margin-left: -0.75rem;
  margin-right: -0.75rem;
  margin-top: -0.75rem;
}
.columns:last-child {
  margin-bottom: -0.75rem;
}
.columns:not(:last-child) {
  margin-bottom: calc(1.5rem - 0.75rem);
}
.columns.is-centered {
  justify-content: center;
}
.columns.is-gapless {
  margin-left: 0;
  margin-right: 0;
  margin-top: 0;
}
.columns.is-gapless > .column {
  margin: 0;
  padding: 0 !important;
}
.columns.is-gapless:not(:last-child) {
  margin-bottom: 1.5rem;
}
.columns.is-gapless:last-child {
  margin-bottom: 0;
}
.columns.is-mobile {
  display: flex;
}
.columns.is-multiline {
  flex-wrap: wrap;
}
.columns.is-vcentered {
  align-items: center;
}
@media screen and (min-width: 769px), print {
  .columns:not(.is-desktop) {
    display: flex;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-desktop {
    display: flex;
  }
}
.columns.is-variable {
  --columnGap: 0.75rem;
  margin-left: calc(-1 * var(--columnGap));
  margin-right: calc(-1 * var(--columnGap));
}
.columns.is-variable > .column {
  padding-left: var(--columnGap);
  padding-right: var(--columnGap);
}
.columns.is-variable.is-0 {
  --columnGap: 0rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-0-mobile {
    --columnGap: 0rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-0-tablet {
    --columnGap: 0rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-0-tablet-only {
    --columnGap: 0rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-0-touch {
    --columnGap: 0rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-0-desktop {
    --columnGap: 0rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-0-desktop-only {
    --columnGap: 0rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-0-widescreen {
    --columnGap: 0rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-0-widescreen-only {
    --columnGap: 0rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-0-fullhd {
    --columnGap: 0rem;
  }
}
.columns.is-variable.is-1 {
  --columnGap: 0.25rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-1-mobile {
    --columnGap: 0.25rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-1-tablet {
    --columnGap: 0.25rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-1-tablet-only {
    --columnGap: 0.25rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-1-touch {
    --columnGap: 0.25rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-1-desktop {
    --columnGap: 0.25rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-1-desktop-only {
    --columnGap: 0.25rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-1-widescreen {
    --columnGap: 0.25rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-1-widescreen-only {
    --columnGap: 0.25rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-1-fullhd {
    --columnGap: 0.25rem;
  }
}
.columns.is-variable.is-2 {
  --columnGap: 0.5rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-2-mobile {
    --columnGap: 0.5rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-2-tablet {
    --columnGap: 0.5rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-2-tablet-only {
    --columnGap: 0.5rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-2-touch {
    --columnGap: 0.5rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-2-desktop {
    --columnGap: 0.5rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-2-desktop-only {
    --columnGap: 0.5rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-2-widescreen {
    --columnGap: 0.5rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-2-widescreen-only {
    --columnGap: 0.5rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-2-fullhd {
    --columnGap: 0.5rem;
  }
}
.columns.is-variable.is-3 {
  --columnGap: 0.75rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-3-mobile {
    --columnGap: 0.75rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-3-tablet {
    --columnGap: 0.75rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-3-tablet-only {
    --columnGap: 0.75rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-3-touch {
    --columnGap: 0.75rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-3-desktop {
    --columnGap: 0.75rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-3-desktop-only {
    --columnGap: 0.75rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-3-widescreen {
    --columnGap: 0.75rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-3-widescreen-only {
    --columnGap: 0.75rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-3-fullhd {
    --columnGap: 0.75rem;
  }
}
.columns.is-variable.is-4 {
  --columnGap: 1rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-4-mobile {
    --columnGap: 1rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-4-tablet {
    --columnGap: 1rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-4-tablet-only {
    --columnGap: 1rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-4-touch {
    --columnGap: 1rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-4-desktop {
    --columnGap: 1rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-4-desktop-only {
    --columnGap: 1rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-4-widescreen {
    --columnGap: 1rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-4-widescreen-only {
    --columnGap: 1rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-4-fullhd {
    --columnGap: 1rem;
  }
}
.columns.is-variable.is-5 {
  --columnGap: 1.25rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-5-mobile {
    --columnGap: 1.25rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-5-tablet {
    --columnGap: 1.25rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-5-tablet-only {
    --columnGap: 1.25rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-5-touch {
    --columnGap: 1.25rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-5-desktop {
    --columnGap: 1.25rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-5-desktop-only {
    --columnGap: 1.25rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-5-widescreen {
    --columnGap: 1.25rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-5-widescreen-only {
    --columnGap: 1.25rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-5-fullhd {
    --columnGap: 1.25rem;
  }
}
.columns.is-variable.is-6 {
  --columnGap: 1.5rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-6-mobile {
    --columnGap: 1.5rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-6-tablet {
    --columnGap: 1.5rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-6-tablet-only {
    --columnGap: 1.5rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-6-touch {
    --columnGap: 1.5rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-6-desktop {
    --columnGap: 1.5rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-6-desktop-only {
    --columnGap: 1.5rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-6-widescreen {
    --columnGap: 1.5rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-6-widescreen-only {
    --columnGap: 1.5rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-6-fullhd {
    --columnGap: 1.5rem;
  }
}
.columns.is-variable.is-7 {
  --columnGap: 1.75rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-7-mobile {
    --columnGap: 1.75rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-7-tablet {
    --columnGap: 1.75rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-7-tablet-only {
    --columnGap: 1.75rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-7-touch {
    --columnGap: 1.75rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-7-desktop {
    --columnGap: 1.75rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-7-desktop-only {
    --columnGap: 1.75rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-7-widescreen {
    --columnGap: 1.75rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-7-widescreen-only {
    --columnGap: 1.75rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-7-fullhd {
    --columnGap: 1.75rem;
  }
}
.columns.is-variable.is-8 {
  --columnGap: 2rem;
}
@media screen and (max-width: 768px) {
  .columns.is-variable.is-8-mobile {
    --columnGap: 2rem;
  }
}
@media screen and (min-width: 769px), print {
  .columns.is-variable.is-8-tablet {
    --columnGap: 2rem;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .columns.is-variable.is-8-tablet-only {
    --columnGap: 2rem;
  }
}
@media screen and (max-width: 1023px) {
  .columns.is-variable.is-8-touch {
    --columnGap: 2rem;
  }
}
@media screen and (min-width: 1024px) {
  .columns.is-variable.is-8-desktop {
    --columnGap: 2rem;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .columns.is-variable.is-8-desktop-only {
    --columnGap: 2rem;
  }
}
@media screen and (min-width: 1216px) {
  .columns.is-variable.is-8-widescreen {
    --columnGap: 2rem;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .columns.is-variable.is-8-widescreen-only {
    --columnGap: 2rem;
  }
}
@media screen and (min-width: 1408px) {
  .columns.is-variable.is-8-fullhd {
    --columnGap: 2rem;
  }
}
.tile {
  align-items: stretch;
  display: block;
  flex-basis: 0;
  flex-grow: 1;
  flex-shrink: 1;
  min-height: min-content;
}
.tile.is-ancestor {
  margin-left: -0.75rem;
  margin-right: -0.75rem;
  margin-top: -0.75rem;
}
.tile.is-ancestor:last-child {
  margin-bottom: -0.75rem;
}
.tile.is-ancestor:not(:last-child) {
  margin-bottom: 0.75rem;
}
.tile.is-child {
  margin: 0 !important;
}
.tile.is-parent {
  padding: 0.75rem;
}
.tile.is-vertical {
  flex-direction: column;
}
.tile.is-vertical > .tile.is-child:not(:last-child) {
  margin-bottom: 1.5rem !important;
}
@media screen and (min-width: 769px), print {
  .tile:not(.is-child) {
    display: flex;
  }
  .tile.is-1 {
    flex: none;
    width: 8.33333337%;
  }
  .tile.is-2 {
    flex: none;
    width: 16.66666674%;
  }
  .tile.is-3 {
    flex: none;
    width: 25%;
  }
  .tile.is-4 {
    flex: none;
    width: 33.33333337%;
  }
  .tile.is-5 {
    flex: none;
    width: 41.66666674%;
  }
  .tile.is-6 {
    flex: none;
    width: 50%;
  }
  .tile.is-7 {
    flex: none;
    width: 58.33333337%;
  }
  .tile.is-8 {
    flex: none;
    width: 66.66666674%;
  }
  .tile.is-9 {
    flex: none;
    width: 75%;
  }
  .tile.is-10 {
    flex: none;
    width: 83.33333337%;
  }
  .tile.is-11 {
    flex: none;
    width: 91.66666674%;
  }
  .tile.is-12 {
    flex: none;
    width: 100%;
  }
}
.has-text-white {
  color: hsl(0deg, 0%, 100%) !important;
}
a.has-text-white:hover,
a.has-text-white:focus {
  color: #e6e6e6 !important;
}
.has-background-white {
  background-color: hsl(0deg, 0%, 100%) !important;
}
.has-text-black {
  color: hsl(0deg, 0%, 4%) !important;
}
a.has-text-black:hover,
a.has-text-black:focus {
  color: black !important;
}
.has-background-black {
  background-color: hsl(0deg, 0%, 4%) !important;
}
.has-text-light {
  color: hsl(0deg, 0%, 96%) !important;
}
a.has-text-light:hover,
a.has-text-light:focus {
  color: #dbdbdb !important;
}
.has-background-light {
  background-color: hsl(0deg, 0%, 96%) !important;
}
.has-text-dark {
  color: hsl(0deg, 0%, 21%) !important;
}
a.has-text-dark:hover,
a.has-text-dark:focus {
  color: #1c1c1c !important;
}
.has-background-dark {
  background-color: hsl(0deg, 0%, 21%) !important;
}
.has-text-primary {
  color: hsl(171deg, 100%, 41%) !important;
}
a.has-text-primary:hover,
a.has-text-primary:focus {
  color: #009e86 !important;
}
.has-background-primary {
  background-color: hsl(171deg, 100%, 41%) !important;
}
.has-text-primary-light {
  color: #ebfffc !important;
}
a.has-text-primary-light:hover,
a.has-text-primary-light:focus {
  color: #b8fff4 !important;
}
.has-background-primary-light {
  background-color: #ebfffc !important;
}
.has-text-primary-dark {
  color: #00947e !important;
}
a.has-text-primary-dark:hover,
a.has-text-primary-dark:focus {
  color: #00c7a9 !important;
}
.has-background-primary-dark {
  background-color: #00947e !important;
}
.has-text-link {
  color: hsl(229deg, 53%, 53%) !important;
}
a.has-text-link:hover,
a.has-text-link:focus {
  color: #3449a8 !important;
}
.has-background-link {
  background-color: hsl(229deg, 53%, 53%) !important;
}
.has-text-link-light {
  color: #eff1fa !important;
}
a.has-text-link-light:hover,
a.has-text-link-light:focus {
  color: #c8cfee !important;
}
.has-background-link-light {
  background-color: #eff1fa !important;
}
.has-text-link-dark {
  color: #3850b7 !important;
}
a.has-text-link-dark:hover,
a.has-text-link-dark:focus {
  color: #576dcb !important;
}
.has-background-link-dark {
  background-color: #3850b7 !important;
}
.has-text-info {
  color: hsl(207deg, 61%, 53%) !important;
}
a.has-text-info:hover,
a.has-text-info:focus {
  color: #2b74b1 !important;
}
.has-background-info {
  background-color: hsl(207deg, 61%, 53%) !important;
}
.has-text-info-light {
  color: #eff5fb !important;
}
a.has-text-info-light:hover,
a.has-text-info-light:focus {
  color: #c6ddf1 !important;
}
.has-background-info-light {
  background-color: #eff5fb !important;
}
.has-text-info-dark {
  color: #296fa8 !important;
}
a.has-text-info-dark:hover,
a.has-text-info-dark:focus {
  color: #368ace !important;
}
.has-background-info-dark {
  background-color: #296fa8 !important;
}
.has-text-success {
  color: hsl(153deg, 53%, 53%) !important;
}
a.has-text-success:hover,
a.has-text-success:focus {
  color: #34a873 !important;
}
.has-background-success {
  background-color: hsl(153deg, 53%, 53%) !important;
}
.has-text-success-light {
  color: #effaf5 !important;
}
a.has-text-success-light:hover,
a.has-text-success-light:focus {
  color: #c8eedd !important;
}
.has-background-success-light {
  background-color: #effaf5 !important;
}
.has-text-success-dark {
  color: #257953 !important;
}
a.has-text-success-dark:hover,
a.has-text-success-dark:focus {
  color: #31a06e !important;
}
.has-background-success-dark {
  background-color: #257953 !important;
}
.has-text-warning {
  color: hsl(44deg, 100%, 77%) !important;
}
a.has-text-warning:hover,
a.has-text-warning:focus {
  color: #ffd257 !important;
}
.has-background-warning {
  background-color: hsl(44deg, 100%, 77%) !important;
}
.has-text-warning-light {
  color: #fffaeb !important;
}
a.has-text-warning-light:hover,
a.has-text-warning-light:focus {
  color: #ffecb8 !important;
}
.has-background-warning-light {
  background-color: #fffaeb !important;
}
.has-text-warning-dark {
  color: #946c00 !important;
}
a.has-text-warning-dark:hover,
a.has-text-warning-dark:focus {
  color: #c79200 !important;
}
.has-background-warning-dark {
  background-color: #946c00 !important;
}
.has-text-danger {
  color: hsl(348deg, 86%, 61%) !important;
}
a.has-text-danger:hover,
a.has-text-danger:focus {
  color: #ee1742 !important;
}
.has-background-danger {
  background-color: hsl(348deg, 86%, 61%) !important;
}
.has-text-danger-light {
  color: #feecf0 !important;
}
a.has-text-danger-light:hover,
a.has-text-danger-light:focus {
  color: #fabdc9 !important;
}
.has-background-danger-light {
  background-color: #feecf0 !important;
}
.has-text-danger-dark {
  color: #cc0f35 !important;
}
a.has-text-danger-dark:hover,
a.has-text-danger-dark:focus {
  color: #ee2049 !important;
}
.has-background-danger-dark {
  background-color: #cc0f35 !important;
}
.has-text-black-bis {
  color: hsl(0deg, 0%, 7%) !important;
}
.has-background-black-bis {
  background-color: hsl(0deg, 0%, 7%) !important;
}
.has-text-black-ter {
  color: hsl(0deg, 0%, 14%) !important;
}
.has-background-black-ter {
  background-color: hsl(0deg, 0%, 14%) !important;
}
.has-text-grey-darker {
  color: hsl(0deg, 0%, 21%) !important;
}
.has-background-grey-darker {
  background-color: hsl(0deg, 0%, 21%) !important;
}
.has-text-grey-dark {
  color: hsl(0deg, 0%, 29%) !important;
}
.has-background-grey-dark {
  background-color: hsl(0deg, 0%, 29%) !important;
}
.has-text-grey {
  color: hsl(0deg, 0%, 48%) !important;
}
.has-background-grey {
  background-color: hsl(0deg, 0%, 48%) !important;
}
.has-text-grey-light {
  color: hsl(0deg, 0%, 71%) !important;
}
.has-background-grey-light {
  background-color: hsl(0deg, 0%, 71%) !important;
}
.has-text-grey-lighter {
  color: hsl(0deg, 0%, 86%) !important;
}
.has-background-grey-lighter {
  background-color: hsl(0deg, 0%, 86%) !important;
}
.has-text-white-ter {
  color: hsl(0deg, 0%, 96%) !important;
}
.has-background-white-ter {
  background-color: hsl(0deg, 0%, 96%) !important;
}
.has-text-white-bis {
  color: hsl(0deg, 0%, 98%) !important;
}
.has-background-white-bis {
  background-color: hsl(0deg, 0%, 98%) !important;
}
.is-flex-direction-row {
  flex-direction: row !important;
}
.is-flex-direction-row-reverse {
  flex-direction: row-reverse !important;
}
.is-flex-direction-column {
  flex-direction: column !important;
}
.is-flex-direction-column-reverse {
  flex-direction: column-reverse !important;
}
.is-flex-wrap-nowrap {
  flex-wrap: nowrap !important;
}
.is-flex-wrap-wrap {
  flex-wrap: wrap !important;
}
.is-flex-wrap-wrap-reverse {
  flex-wrap: wrap-reverse !important;
}
.is-justify-content-flex-start {
  justify-content: flex-start !important;
}
.is-justify-content-flex-end {
  justify-content: flex-end !important;
}
.is-justify-content-center {
  justify-content: center !important;
}
.is-justify-content-space-between {
  justify-content: space-between !important;
}
.is-justify-content-space-around {
  justify-content: space-around !important;
}
.is-justify-content-space-evenly {
  justify-content: space-evenly !important;
}
.is-justify-content-start {
  justify-content: start !important;
}
.is-justify-content-end {
  justify-content: end !important;
}
.is-justify-content-left {
  justify-content: left !important;
}
.is-justify-content-right {
  justify-content: right !important;
}
.is-align-content-flex-start {
  align-content: flex-start !important;
}
.is-align-content-flex-end {
  align-content: flex-end !important;
}
.is-align-content-center {
  align-content: center !important;
}
.is-align-content-space-between {
  align-content: space-between !important;
}
.is-align-content-space-around {
  align-content: space-around !important;
}
.is-align-content-space-evenly {
  align-content: space-evenly !important;
}
.is-align-content-stretch {
  align-content: stretch !important;
}
.is-align-content-start {
  align-content: start !important;
}
.is-align-content-end {
  align-content: end !important;
}
.is-align-content-baseline {
  align-content: baseline !important;
}
.is-align-items-stretch {
  align-items: stretch !important;
}
.is-align-items-flex-start {
  align-items: flex-start !important;
}
.is-align-items-flex-end {
  align-items: flex-end !important;
}
.is-align-items-center {
  align-items: center !important;
}
.is-align-items-baseline {
  align-items: baseline !important;
}
.is-align-items-start {
  align-items: start !important;
}
.is-align-items-end {
  align-items: end !important;
}
.is-align-items-self-start {
  align-items: self-start !important;
}
.is-align-items-self-end {
  align-items: self-end !important;
}
.is-align-self-auto {
  align-self: auto !important;
}
.is-align-self-flex-start {
  align-self: flex-start !important;
}
.is-align-self-flex-end {
  align-self: flex-end !important;
}
.is-align-self-center {
  align-self: center !important;
}
.is-align-self-baseline {
  align-self: baseline !important;
}
.is-align-self-stretch {
  align-self: stretch !important;
}
.is-flex-grow-0 {
  flex-grow: 0 !important;
}
.is-flex-grow-1 {
  flex-grow: 1 !important;
}
.is-flex-grow-2 {
  flex-grow: 2 !important;
}
.is-flex-grow-3 {
  flex-grow: 3 !important;
}
.is-flex-grow-4 {
  flex-grow: 4 !important;
}
.is-flex-grow-5 {
  flex-grow: 5 !important;
}
.is-flex-shrink-0 {
  flex-shrink: 0 !important;
}
.is-flex-shrink-1 {
  flex-shrink: 1 !important;
}
.is-flex-shrink-2 {
  flex-shrink: 2 !important;
}
.is-flex-shrink-3 {
  flex-shrink: 3 !important;
}
.is-flex-shrink-4 {
  flex-shrink: 4 !important;
}
.is-flex-shrink-5 {
  flex-shrink: 5 !important;
}
.is-clearfix::after {
  clear: both;
  content: " ";
  display: table;
}
.is-pulled-left {
  float: left !important;
}
.is-pulled-right {
  float: right !important;
}
.is-radiusless {
  border-radius: 0 !important;
}
.is-shadowless {
  box-shadow: none !important;
}
.is-clickable {
  cursor: pointer !important;
  pointer-events: all !important;
}
.is-clipped {
  overflow: hidden !important;
}
.is-relative {
  position: relative !important;
}
.is-marginless {
  margin: 0 !important;
}
.is-paddingless {
  padding: 0 !important;
}
.m-0 {
  margin: 0 !important;
}
.mt-0 {
  margin-top: 0 !important;
}
.mr-0 {
  margin-right: 0 !important;
}
.mb-0 {
  margin-bottom: 0 !important;
}
.ml-0 {
  margin-left: 0 !important;
}
.mx-0 {
  margin-left: 0 !important;
  margin-right: 0 !important;
}
.my-0 {
  margin-top: 0 !important;
  margin-bottom: 0 !important;
}
.m-1 {
  margin: 0.25rem !important;
}
.mt-1 {
  margin-top: 0.25rem !important;
}
.mr-1 {
  margin-right: 0.25rem !important;
}
.mb-1 {
  margin-bottom: 0.25rem !important;
}
.ml-1 {
  margin-left: 0.25rem !important;
}
.mx-1 {
  margin-left: 0.25rem !important;
  margin-right: 0.25rem !important;
}
.my-1 {
  margin-top: 0.25rem !important;
  margin-bottom: 0.25rem !important;
}
.m-2 {
  margin: 0.5rem !important;
}
.mt-2 {
  margin-top: 0.5rem !important;
}
.mr-2 {
  margin-right: 0.5rem !important;
}
.mb-2 {
  margin-bottom: 0.5rem !important;
}
.ml-2 {
  margin-left: 0.5rem !important;
}
.mx-2 {
  margin-left: 0.5rem !important;
  margin-right: 0.5rem !important;
}
.my-2 {
  margin-top: 0.5rem !important;
  margin-bottom: 0.5rem !important;
}
.m-3 {
  margin: 0.75rem !important;
}
.mt-3 {
  margin-top: 0.75rem !important;
}
.mr-3 {
  margin-right: 0.75rem !important;
}
.mb-3 {
  margin-bottom: 0.75rem !important;
}
.ml-3 {
  margin-left: 0.75rem !important;
}
.mx-3 {
  margin-left: 0.75rem !important;
  margin-right: 0.75rem !important;
}
.my-3 {
  margin-top: 0.75rem !important;
  margin-bottom: 0.75rem !important;
}
.m-4 {
  margin: 1rem !important;
}
.mt-4 {
  margin-top: 1rem !important;
}
.mr-4 {
  margin-right: 1rem !important;
}
.mb-4 {
  margin-bottom: 1rem !important;
}
.ml-4 {
  margin-left: 1rem !important;
}
.mx-4 {
  margin-left: 1rem !important;
  margin-right: 1rem !important;
}
.my-4 {
  margin-top: 1rem !important;
  margin-bottom: 1rem !important;
}
.m-5 {
  margin: 1.5rem !important;
}
.mt-5 {
  margin-top: 1.5rem !important;
}
.mr-5 {
  margin-right: 1.5rem !important;
}
.mb-5 {
  margin-bottom: 1.5rem !important;
}
.ml-5 {
  margin-left: 1.5rem !important;
}
.mx-5 {
  margin-left: 1.5rem !important;
  margin-right: 1.5rem !important;
}
.my-5 {
  margin-top: 1.5rem !important;
  margin-bottom: 1.5rem !important;
}
.m-6 {
  margin: 3rem !important;
}
.mt-6 {
  margin-top: 3rem !important;
}
.mr-6 {
  margin-right: 3rem !important;
}
.mb-6 {
  margin-bottom: 3rem !important;
}
.ml-6 {
  margin-left: 3rem !important;
}
.mx-6 {
  margin-left: 3rem !important;
  margin-right: 3rem !important;
}
.my-6 {
  margin-top: 3rem !important;
  margin-bottom: 3rem !important;
}
.m-auto {
  margin: auto !important;
}
.mt-auto {
  margin-top: auto !important;
}
.mr-auto {
  margin-right: auto !important;
}
.mb-auto {
  margin-bottom: auto !important;
}
.ml-auto {
  margin-left: auto !important;
}
.mx-auto {
  margin-left: auto !important;
  margin-right: auto !important;
}
.my-auto {
  margin-top: auto !important;
  margin-bottom: auto !important;
}
.p-0 {
  padding: 0 !important;
}
.pt-0 {
  padding-top: 0 !important;
}
.pr-0 {
  padding-right: 0 !important;
}
.pb-0 {
  padding-bottom: 0 !important;
}
.pl-0 {
  padding-left: 0 !important;
}
.px-0 {
  padding-left: 0 !important;
  padding-right: 0 !important;
}
.py-0 {
  padding-top: 0 !important;
  padding-bottom: 0 !important;
}
.p-1 {
  padding: 0.25rem !important;
}
.pt-1 {
  padding-top: 0.25rem !important;
}
.pr-1 {
  padding-right: 0.25rem !important;
}
.pb-1 {
  padding-bottom: 0.25rem !important;
}
.pl-1 {
  padding-left: 0.25rem !important;
}
.px-1 {
  padding-left: 0.25rem !important;
  padding-right: 0.25rem !important;
}
.py-1 {
  padding-top: 0.25rem !important;
  padding-bottom: 0.25rem !important;
}
.p-2 {
  padding: 0.5rem !important;
}
.pt-2 {
  padding-top: 0.5rem !important;
}
.pr-2 {
  padding-right: 0.5rem !important;
}
.pb-2 {
  padding-bottom: 0.5rem !important;
}
.pl-2 {
  padding-left: 0.5rem !important;
}
.px-2 {
  padding-left: 0.5rem !important;
  padding-right: 0.5rem !important;
}
.py-2 {
  padding-top: 0.5rem !important;
  padding-bottom: 0.5rem !important;
}
.p-3 {
  padding: 0.75rem !important;
}
.pt-3 {
  padding-top: 0.75rem !important;
}
.pr-3 {
  padding-right: 0.75rem !important;
}
.pb-3 {
  padding-bottom: 0.75rem !important;
}
.pl-3 {
  padding-left: 0.75rem !important;
}
.px-3 {
  padding-left: 0.75rem !important;
  padding-right: 0.75rem !important;
}
.py-3 {
  padding-top: 0.75rem !important;
  padding-bottom: 0.75rem !important;
}
.p-4 {
  padding: 1rem !important;
}
.pt-4 {
  padding-top: 1rem !important;
}
.pr-4 {
  padding-right: 1rem !important;
}
.pb-4 {
  padding-bottom: 1rem !important;
}
.pl-4 {
  padding-left: 1rem !important;
}
.px-4 {
  padding-left: 1rem !important;
  padding-right: 1rem !important;
}
.py-4 {
  padding-top: 1rem !important;
  padding-bottom: 1rem !important;
}
.p-5 {
  padding: 1.5rem !important;
}
.pt-5 {
  padding-top: 1.5rem !important;
}
.pr-5 {
  padding-right: 1.5rem !important;
}
.pb-5 {
  padding-bottom: 1.5rem !important;
}
.pl-5 {
  padding-left: 1.5rem !important;
}
.px-5 {
  padding-left: 1.5rem !important;
  padding-right: 1.5rem !important;
}
.py-5 {
  padding-top: 1.5rem !important;
  padding-bottom: 1.5rem !important;
}
.p-6 {
  padding: 3rem !important;
}
.pt-6 {
  padding-top: 3rem !important;
}
.pr-6 {
  padding-right: 3rem !important;
}
.pb-6 {
  padding-bottom: 3rem !important;
}
.pl-6 {
  padding-left: 3rem !important;
}
.px-6 {
  padding-left: 3rem !important;
  padding-right: 3rem !important;
}
.py-6 {
  padding-top: 3rem !important;
  padding-bottom: 3rem !important;
}
.p-auto {
  padding: auto !important;
}
.pt-auto {
  padding-top: auto !important;
}
.pr-auto {
  padding-right: auto !important;
}
.pb-auto {
  padding-bottom: auto !important;
}
.pl-auto {
  padding-left: auto !important;
}
.px-auto {
  padding-left: auto !important;
  padding-right: auto !important;
}
.py-auto {
  padding-top: auto !important;
  padding-bottom: auto !important;
}
.is-size-1 {
  font-size: 3rem !important;
}
.is-size-2 {
  font-size: 2.5rem !important;
}
.is-size-3 {
  font-size: 2rem !important;
}
.is-size-4 {
  font-size: 1.5rem !important;
}
.is-size-5 {
  font-size: 1.25rem !important;
}
.is-size-6 {
  font-size: 1rem !important;
}
.is-size-7 {
  font-size: 0.75rem !important;
}
@media screen and (max-width: 768px) {
  .is-size-1-mobile {
    font-size: 3rem !important;
  }
  .is-size-2-mobile {
    font-size: 2.5rem !important;
  }
  .is-size-3-mobile {
    font-size: 2rem !important;
  }
  .is-size-4-mobile {
    font-size: 1.5rem !important;
  }
  .is-size-5-mobile {
    font-size: 1.25rem !important;
  }
  .is-size-6-mobile {
    font-size: 1rem !important;
  }
  .is-size-7-mobile {
    font-size: 0.75rem !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-size-1-tablet {
    font-size: 3rem !important;
  }
  .is-size-2-tablet {
    font-size: 2.5rem !important;
  }
  .is-size-3-tablet {
    font-size: 2rem !important;
  }
  .is-size-4-tablet {
    font-size: 1.5rem !important;
  }
  .is-size-5-tablet {
    font-size: 1.25rem !important;
  }
  .is-size-6-tablet {
    font-size: 1rem !important;
  }
  .is-size-7-tablet {
    font-size: 0.75rem !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-size-1-touch {
    font-size: 3rem !important;
  }
  .is-size-2-touch {
    font-size: 2.5rem !important;
  }
  .is-size-3-touch {
    font-size: 2rem !important;
  }
  .is-size-4-touch {
    font-size: 1.5rem !important;
  }
  .is-size-5-touch {
    font-size: 1.25rem !important;
  }
  .is-size-6-touch {
    font-size: 1rem !important;
  }
  .is-size-7-touch {
    font-size: 0.75rem !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-size-1-desktop {
    font-size: 3rem !important;
  }
  .is-size-2-desktop {
    font-size: 2.5rem !important;
  }
  .is-size-3-desktop {
    font-size: 2rem !important;
  }
  .is-size-4-desktop {
    font-size: 1.5rem !important;
  }
  .is-size-5-desktop {
    font-size: 1.25rem !important;
  }
  .is-size-6-desktop {
    font-size: 1rem !important;
  }
  .is-size-7-desktop {
    font-size: 0.75rem !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-size-1-widescreen {
    font-size: 3rem !important;
  }
  .is-size-2-widescreen {
    font-size: 2.5rem !important;
  }
  .is-size-3-widescreen {
    font-size: 2rem !important;
  }
  .is-size-4-widescreen {
    font-size: 1.5rem !important;
  }
  .is-size-5-widescreen {
    font-size: 1.25rem !important;
  }
  .is-size-6-widescreen {
    font-size: 1rem !important;
  }
  .is-size-7-widescreen {
    font-size: 0.75rem !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-size-1-fullhd {
    font-size: 3rem !important;
  }
  .is-size-2-fullhd {
    font-size: 2.5rem !important;
  }
  .is-size-3-fullhd {
    font-size: 2rem !important;
  }
  .is-size-4-fullhd {
    font-size: 1.5rem !important;
  }
  .is-size-5-fullhd {
    font-size: 1.25rem !important;
  }
  .is-size-6-fullhd {
    font-size: 1rem !important;
  }
  .is-size-7-fullhd {
    font-size: 0.75rem !important;
  }
}
.has-text-centered {
  text-align: center !important;
}
.has-text-justified {
  text-align: justify !important;
}
.has-text-left {
  text-align: left !important;
}
.has-text-right {
  text-align: right !important;
}
@media screen and (max-width: 768px) {
  .has-text-centered-mobile {
    text-align: center !important;
  }
}
@media screen and (min-width: 769px), print {
  .has-text-centered-tablet {
    text-align: center !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .has-text-centered-tablet-only {
    text-align: center !important;
  }
}
@media screen and (max-width: 1023px) {
  .has-text-centered-touch {
    text-align: center !important;
  }
}
@media screen and (min-width: 1024px) {
  .has-text-centered-desktop {
    text-align: center !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .has-text-centered-desktop-only {
    text-align: center !important;
  }
}
@media screen and (min-width: 1216px) {
  .has-text-centered-widescreen {
    text-align: center !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .has-text-centered-widescreen-only {
    text-align: center !important;
  }
}
@media screen and (min-width: 1408px) {
  .has-text-centered-fullhd {
    text-align: center !important;
  }
}
@media screen and (max-width: 768px) {
  .has-text-justified-mobile {
    text-align: justify !important;
  }
}
@media screen and (min-width: 769px), print {
  .has-text-justified-tablet {
    text-align: justify !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .has-text-justified-tablet-only {
    text-align: justify !important;
  }
}
@media screen and (max-width: 1023px) {
  .has-text-justified-touch {
    text-align: justify !important;
  }
}
@media screen and (min-width: 1024px) {
  .has-text-justified-desktop {
    text-align: justify !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .has-text-justified-desktop-only {
    text-align: justify !important;
  }
}
@media screen and (min-width: 1216px) {
  .has-text-justified-widescreen {
    text-align: justify !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .has-text-justified-widescreen-only {
    text-align: justify !important;
  }
}
@media screen and (min-width: 1408px) {
  .has-text-justified-fullhd {
    text-align: justify !important;
  }
}
@media screen and (max-width: 768px) {
  .has-text-left-mobile {
    text-align: left !important;
  }
}
@media screen and (min-width: 769px), print {
  .has-text-left-tablet {
    text-align: left !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .has-text-left-tablet-only {
    text-align: left !important;
  }
}
@media screen and (max-width: 1023px) {
  .has-text-left-touch {
    text-align: left !important;
  }
}
@media screen and (min-width: 1024px) {
  .has-text-left-desktop {
    text-align: left !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .has-text-left-desktop-only {
    text-align: left !important;
  }
}
@media screen and (min-width: 1216px) {
  .has-text-left-widescreen {
    text-align: left !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .has-text-left-widescreen-only {
    text-align: left !important;
  }
}
@media screen and (min-width: 1408px) {
  .has-text-left-fullhd {
    text-align: left !important;
  }
}
@media screen and (max-width: 768px) {
  .has-text-right-mobile {
    text-align: right !important;
  }
}
@media screen and (min-width: 769px), print {
  .has-text-right-tablet {
    text-align: right !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .has-text-right-tablet-only {
    text-align: right !important;
  }
}
@media screen and (max-width: 1023px) {
  .has-text-right-touch {
    text-align: right !important;
  }
}
@media screen and (min-width: 1024px) {
  .has-text-right-desktop {
    text-align: right !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .has-text-right-desktop-only {
    text-align: right !important;
  }
}
@media screen and (min-width: 1216px) {
  .has-text-right-widescreen {
    text-align: right !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .has-text-right-widescreen-only {
    text-align: right !important;
  }
}
@media screen and (min-width: 1408px) {
  .has-text-right-fullhd {
    text-align: right !important;
  }
}
.is-capitalized {
  text-transform: capitalize !important;
}
.is-lowercase {
  text-transform: lowercase !important;
}
.is-uppercase {
  text-transform: uppercase !important;
}
.is-italic {
  font-style: italic !important;
}
.is-underlined {
  text-decoration: underline !important;
}
.has-text-weight-light {
  font-weight: 300 !important;
}
.has-text-weight-normal {
  font-weight: 400 !important;
}
.has-text-weight-medium {
  font-weight: 500 !important;
}
.has-text-weight-semibold {
  font-weight: 600 !important;
}
.has-text-weight-bold {
  font-weight: 700 !important;
}
.is-family-primary {
  font-family: "Nunito", sans-serif !important;
}
.is-family-secondary {
  font-family: "Nunito", sans-serif !important;
}
.is-family-sans-serif {
  font-family: "Nunito", sans-serif !important;
}
.is-family-monospace {
  font-family: monospace !important;
}
.is-family-code {
  font-family: monospace !important;
}
.is-block {
  display: block !important;
}
@media screen and (max-width: 768px) {
  .is-block-mobile {
    display: block !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-block-tablet {
    display: block !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .is-block-tablet-only {
    display: block !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-block-touch {
    display: block !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-block-desktop {
    display: block !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .is-block-desktop-only {
    display: block !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-block-widescreen {
    display: block !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .is-block-widescreen-only {
    display: block !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-block-fullhd {
    display: block !important;
  }
}
.is-flex {
  display: flex !important;
}
@media screen and (max-width: 768px) {
  .is-flex-mobile {
    display: flex !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-flex-tablet {
    display: flex !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .is-flex-tablet-only {
    display: flex !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-flex-touch {
    display: flex !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-flex-desktop {
    display: flex !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .is-flex-desktop-only {
    display: flex !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-flex-widescreen {
    display: flex !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .is-flex-widescreen-only {
    display: flex !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-flex-fullhd {
    display: flex !important;
  }
}
.is-inline {
  display: inline !important;
}
@media screen and (max-width: 768px) {
  .is-inline-mobile {
    display: inline !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-inline-tablet {
    display: inline !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .is-inline-tablet-only {
    display: inline !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-inline-touch {
    display: inline !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-inline-desktop {
    display: inline !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .is-inline-desktop-only {
    display: inline !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-inline-widescreen {
    display: inline !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .is-inline-widescreen-only {
    display: inline !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-inline-fullhd {
    display: inline !important;
  }
}
.is-inline-block {
  display: inline-block !important;
}
@media screen and (max-width: 768px) {
  .is-inline-block-mobile {
    display: inline-block !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-inline-block-tablet {
    display: inline-block !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .is-inline-block-tablet-only {
    display: inline-block !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-inline-block-touch {
    display: inline-block !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-inline-block-desktop {
    display: inline-block !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .is-inline-block-desktop-only {
    display: inline-block !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-inline-block-widescreen {
    display: inline-block !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .is-inline-block-widescreen-only {
    display: inline-block !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-inline-block-fullhd {
    display: inline-block !important;
  }
}
.is-inline-flex {
  display: inline-flex !important;
}
@media screen and (max-width: 768px) {
  .is-inline-flex-mobile {
    display: inline-flex !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-inline-flex-tablet {
    display: inline-flex !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .is-inline-flex-tablet-only {
    display: inline-flex !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-inline-flex-touch {
    display: inline-flex !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-inline-flex-desktop {
    display: inline-flex !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .is-inline-flex-desktop-only {
    display: inline-flex !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-inline-flex-widescreen {
    display: inline-flex !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .is-inline-flex-widescreen-only {
    display: inline-flex !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-inline-flex-fullhd {
    display: inline-flex !important;
  }
}
.is-hidden {
  display: none !important;
}
.is-sr-only {
  border: none !important;
  clip: rect(0, 0, 0, 0) !important;
  height: 0.01em !important;
  overflow: hidden !important;
  padding: 0 !important;
  position: absolute !important;
  white-space: nowrap !important;
  width: 0.01em !important;
}
@media screen and (max-width: 768px) {
  .is-hidden-mobile {
    display: none !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-hidden-tablet {
    display: none !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .is-hidden-tablet-only {
    display: none !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-hidden-touch {
    display: none !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-hidden-desktop {
    display: none !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .is-hidden-desktop-only {
    display: none !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-hidden-widescreen {
    display: none !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .is-hidden-widescreen-only {
    display: none !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-hidden-fullhd {
    display: none !important;
  }
}
.is-invisible {
  visibility: hidden !important;
}
@media screen and (max-width: 768px) {
  .is-invisible-mobile {
    visibility: hidden !important;
  }
}
@media screen and (min-width: 769px), print {
  .is-invisible-tablet {
    visibility: hidden !important;
  }
}
@media screen and (min-width: 769px) and (max-width: 1023px) {
  .is-invisible-tablet-only {
    visibility: hidden !important;
  }
}
@media screen and (max-width: 1023px) {
  .is-invisible-touch {
    visibility: hidden !important;
  }
}
@media screen and (min-width: 1024px) {
  .is-invisible-desktop {
    visibility: hidden !important;
  }
}
@media screen and (min-width: 1024px) and (max-width: 1215px) {
  .is-invisible-desktop-only {
    visibility: hidden !important;
  }
}
@media screen and (min-width: 1216px) {
  .is-invisible-widescreen {
    visibility: hidden !important;
  }
}
@media screen and (min-width: 1216px) and (max-width: 1407px) {
  .is-invisible-widescreen-only {
    visibility: hidden !important;
  }
}
@media screen and (min-width: 1408px) {
  .is-invisible-fullhd {
    visibility: hidden !important;
  }
}
.hero {
  align-items: stretch;
  display: flex;
  flex-direction: column;
  justify-content: space-between;
}
.hero .navbar {
  background: none;
}
.hero .tabs ul {
  border-bottom: none;
}
.hero.is-white {
  background-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.hero.is-white a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-white strong {
  color: inherit;
}
.hero.is-white .title {
  color: hsl(0deg, 0%, 4%);
}
.hero.is-white .subtitle {
  color: rgba(10, 10, 10, 0.9);
}
.hero.is-white .subtitle a:not(.button),
.hero.is-white .subtitle strong {
  color: hsl(0deg, 0%, 4%);
}
@media screen and (max-width: 1023px) {
  .hero.is-white .navbar-menu {
    background-color: hsl(0deg, 0%, 100%);
  }
}
.hero.is-white .navbar-item,
.hero.is-white .navbar-link {
  color: rgba(10, 10, 10, 0.7);
}
.hero.is-white a.navbar-item:hover,
.hero.is-white a.navbar-item.is-active,
.hero.is-white .navbar-link:hover,
.hero.is-white .navbar-link.is-active {
  background-color: #f2f2f2;
  color: hsl(0deg, 0%, 4%);
}
.hero.is-white .tabs a {
  color: hsl(0deg, 0%, 4%);
  opacity: 0.9;
}
.hero.is-white .tabs a:hover {
  opacity: 1;
}
.hero.is-white .tabs li.is-active a {
  color: hsl(0deg, 0%, 100%) !important;
  opacity: 1;
}
.hero.is-white .tabs.is-boxed a,
.hero.is-white .tabs.is-toggle a {
  color: hsl(0deg, 0%, 4%);
}
.hero.is-white .tabs.is-boxed a:hover,
.hero.is-white .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-white .tabs.is-boxed li.is-active a,
.hero.is-white .tabs.is-boxed li.is-active a:hover,
.hero.is-white .tabs.is-toggle li.is-active a,
.hero.is-white .tabs.is-toggle li.is-active a:hover {
  background-color: hsl(0deg, 0%, 4%);
  border-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.hero.is-white.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #e8e3e4 0%,
      hsl(0deg, 0%, 100%) 71%,
      white 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-white.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #e8e3e4 0%,
        hsl(0deg, 0%, 100%) 71%,
        white 100%);
  }
}
.hero.is-black {
  background-color: hsl(0deg, 0%, 4%);
  color: hsl(0deg, 0%, 100%);
}
.hero.is-black a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-black strong {
  color: inherit;
}
.hero.is-black .title {
  color: hsl(0deg, 0%, 100%);
}
.hero.is-black .subtitle {
  color: rgba(255, 255, 255, 0.9);
}
.hero.is-black .subtitle a:not(.button),
.hero.is-black .subtitle strong {
  color: hsl(0deg, 0%, 100%);
}
@media screen and (max-width: 1023px) {
  .hero.is-black .navbar-menu {
    background-color: hsl(0deg, 0%, 4%);
  }
}
.hero.is-black .navbar-item,
.hero.is-black .navbar-link {
  color: rgba(255, 255, 255, 0.7);
}
.hero.is-black a.navbar-item:hover,
.hero.is-black a.navbar-item.is-active,
.hero.is-black .navbar-link:hover,
.hero.is-black .navbar-link.is-active {
  background-color: black;
  color: hsl(0deg, 0%, 100%);
}
.hero.is-black .tabs a {
  color: hsl(0deg, 0%, 100%);
  opacity: 0.9;
}
.hero.is-black .tabs a:hover {
  opacity: 1;
}
.hero.is-black .tabs li.is-active a {
  color: hsl(0deg, 0%, 4%) !important;
  opacity: 1;
}
.hero.is-black .tabs.is-boxed a,
.hero.is-black .tabs.is-toggle a {
  color: hsl(0deg, 0%, 100%);
}
.hero.is-black .tabs.is-boxed a:hover,
.hero.is-black .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-black .tabs.is-boxed li.is-active a,
.hero.is-black .tabs.is-boxed li.is-active a:hover,
.hero.is-black .tabs.is-toggle li.is-active a,
.hero.is-black .tabs.is-toggle li.is-active a:hover {
  background-color: hsl(0deg, 0%, 100%);
  border-color: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 4%);
}
.hero.is-black.is-bold {
  background-image:
    linear-gradient(
      141deg,
      black 0%,
      hsl(0deg, 0%, 4%) 71%,
      #181616 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-black.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        black 0%,
        hsl(0deg, 0%, 4%) 71%,
        #181616 100%);
  }
}
.hero.is-light {
  background-color: hsl(0deg, 0%, 96%);
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-light a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-light strong {
  color: inherit;
}
.hero.is-light .title {
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-light .subtitle {
  color: rgba(0, 0, 0, 0.9);
}
.hero.is-light .subtitle a:not(.button),
.hero.is-light .subtitle strong {
  color: rgba(0, 0, 0, 0.7);
}
@media screen and (max-width: 1023px) {
  .hero.is-light .navbar-menu {
    background-color: hsl(0deg, 0%, 96%);
  }
}
.hero.is-light .navbar-item,
.hero.is-light .navbar-link {
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-light a.navbar-item:hover,
.hero.is-light a.navbar-item.is-active,
.hero.is-light .navbar-link:hover,
.hero.is-light .navbar-link.is-active {
  background-color: #e8e8e8;
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-light .tabs a {
  color: rgba(0, 0, 0, 0.7);
  opacity: 0.9;
}
.hero.is-light .tabs a:hover {
  opacity: 1;
}
.hero.is-light .tabs li.is-active a {
  color: hsl(0deg, 0%, 96%) !important;
  opacity: 1;
}
.hero.is-light .tabs.is-boxed a,
.hero.is-light .tabs.is-toggle a {
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-light .tabs.is-boxed a:hover,
.hero.is-light .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-light .tabs.is-boxed li.is-active a,
.hero.is-light .tabs.is-boxed li.is-active a:hover,
.hero.is-light .tabs.is-toggle li.is-active a,
.hero.is-light .tabs.is-toggle li.is-active a:hover {
  background-color: rgba(0, 0, 0, 0.7);
  border-color: rgba(0, 0, 0, 0.7);
  color: hsl(0deg, 0%, 96%);
}
.hero.is-light.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #dfd8d9 0%,
      hsl(0deg, 0%, 96%) 71%,
      white 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-light.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #dfd8d9 0%,
        hsl(0deg, 0%, 96%) 71%,
        white 100%);
  }
}
.hero.is-dark {
  background-color: hsl(0deg, 0%, 21%);
  color: #fff;
}
.hero.is-dark a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-dark strong {
  color: inherit;
}
.hero.is-dark .title {
  color: #fff;
}
.hero.is-dark .subtitle {
  color: rgba(255, 255, 255, 0.9);
}
.hero.is-dark .subtitle a:not(.button),
.hero.is-dark .subtitle strong {
  color: #fff;
}
@media screen and (max-width: 1023px) {
  .hero.is-dark .navbar-menu {
    background-color: hsl(0deg, 0%, 21%);
  }
}
.hero.is-dark .navbar-item,
.hero.is-dark .navbar-link {
  color: rgba(255, 255, 255, 0.7);
}
.hero.is-dark a.navbar-item:hover,
.hero.is-dark a.navbar-item.is-active,
.hero.is-dark .navbar-link:hover,
.hero.is-dark .navbar-link.is-active {
  background-color: #292929;
  color: #fff;
}
.hero.is-dark .tabs a {
  color: #fff;
  opacity: 0.9;
}
.hero.is-dark .tabs a:hover {
  opacity: 1;
}
.hero.is-dark .tabs li.is-active a {
  color: hsl(0deg, 0%, 21%) !important;
  opacity: 1;
}
.hero.is-dark .tabs.is-boxed a,
.hero.is-dark .tabs.is-toggle a {
  color: #fff;
}
.hero.is-dark .tabs.is-boxed a:hover,
.hero.is-dark .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-dark .tabs.is-boxed li.is-active a,
.hero.is-dark .tabs.is-boxed li.is-active a:hover,
.hero.is-dark .tabs.is-toggle li.is-active a,
.hero.is-dark .tabs.is-toggle li.is-active a:hover {
  background-color: #fff;
  border-color: #fff;
  color: hsl(0deg, 0%, 21%);
}
.hero.is-dark.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #1f191a 0%,
      hsl(0deg, 0%, 21%) 71%,
      #46403f 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-dark.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #1f191a 0%,
        hsl(0deg, 0%, 21%) 71%,
        #46403f 100%);
  }
}
.hero.is-primary {
  background-color: hsl(171deg, 100%, 41%);
  color: #fff;
}
.hero.is-primary a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-primary strong {
  color: inherit;
}
.hero.is-primary .title {
  color: #fff;
}
.hero.is-primary .subtitle {
  color: rgba(255, 255, 255, 0.9);
}
.hero.is-primary .subtitle a:not(.button),
.hero.is-primary .subtitle strong {
  color: #fff;
}
@media screen and (max-width: 1023px) {
  .hero.is-primary .navbar-menu {
    background-color: hsl(171deg, 100%, 41%);
  }
}
.hero.is-primary .navbar-item,
.hero.is-primary .navbar-link {
  color: rgba(255, 255, 255, 0.7);
}
.hero.is-primary a.navbar-item:hover,
.hero.is-primary a.navbar-item.is-active,
.hero.is-primary .navbar-link:hover,
.hero.is-primary .navbar-link.is-active {
  background-color: #00b89c;
  color: #fff;
}
.hero.is-primary .tabs a {
  color: #fff;
  opacity: 0.9;
}
.hero.is-primary .tabs a:hover {
  opacity: 1;
}
.hero.is-primary .tabs li.is-active a {
  color: hsl(171deg, 100%, 41%) !important;
  opacity: 1;
}
.hero.is-primary .tabs.is-boxed a,
.hero.is-primary .tabs.is-toggle a {
  color: #fff;
}
.hero.is-primary .tabs.is-boxed a:hover,
.hero.is-primary .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-primary .tabs.is-boxed li.is-active a,
.hero.is-primary .tabs.is-boxed li.is-active a:hover,
.hero.is-primary .tabs.is-toggle li.is-active a,
.hero.is-primary .tabs.is-toggle li.is-active a:hover {
  background-color: #fff;
  border-color: #fff;
  color: hsl(171deg, 100%, 41%);
}
.hero.is-primary.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #009e6c 0%,
      hsl(171deg, 100%, 41%) 71%,
      #00e7eb 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-primary.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #009e6c 0%,
        hsl(171deg, 100%, 41%) 71%,
        #00e7eb 100%);
  }
}
.hero.is-link {
  background-color: hsl(229deg, 53%, 53%);
  color: #fff;
}
.hero.is-link a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-link strong {
  color: inherit;
}
.hero.is-link .title {
  color: #fff;
}
.hero.is-link .subtitle {
  color: rgba(255, 255, 255, 0.9);
}
.hero.is-link .subtitle a:not(.button),
.hero.is-link .subtitle strong {
  color: #fff;
}
@media screen and (max-width: 1023px) {
  .hero.is-link .navbar-menu {
    background-color: hsl(229deg, 53%, 53%);
  }
}
.hero.is-link .navbar-item,
.hero.is-link .navbar-link {
  color: rgba(255, 255, 255, 0.7);
}
.hero.is-link a.navbar-item:hover,
.hero.is-link a.navbar-item.is-active,
.hero.is-link .navbar-link:hover,
.hero.is-link .navbar-link.is-active {
  background-color: #3a51bb;
  color: #fff;
}
.hero.is-link .tabs a {
  color: #fff;
  opacity: 0.9;
}
.hero.is-link .tabs a:hover {
  opacity: 1;
}
.hero.is-link .tabs li.is-active a {
  color: hsl(229deg, 53%, 53%) !important;
  opacity: 1;
}
.hero.is-link .tabs.is-boxed a,
.hero.is-link .tabs.is-toggle a {
  color: #fff;
}
.hero.is-link .tabs.is-boxed a:hover,
.hero.is-link .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-link .tabs.is-boxed li.is-active a,
.hero.is-link .tabs.is-boxed li.is-active a:hover,
.hero.is-link .tabs.is-toggle li.is-active a,
.hero.is-link .tabs.is-toggle li.is-active a:hover {
  background-color: #fff;
  border-color: #fff;
  color: hsl(229deg, 53%, 53%);
}
.hero.is-link.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #2959b3 0%,
      hsl(229deg, 53%, 53%) 71%,
      #5658d2 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-link.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #2959b3 0%,
        hsl(229deg, 53%, 53%) 71%,
        #5658d2 100%);
  }
}
.hero.is-info {
  background-color: hsl(207deg, 61%, 53%);
  color: #fff;
}
.hero.is-info a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-info strong {
  color: inherit;
}
.hero.is-info .title {
  color: #fff;
}
.hero.is-info .subtitle {
  color: rgba(255, 255, 255, 0.9);
}
.hero.is-info .subtitle a:not(.button),
.hero.is-info .subtitle strong {
  color: #fff;
}
@media screen and (max-width: 1023px) {
  .hero.is-info .navbar-menu {
    background-color: hsl(207deg, 61%, 53%);
  }
}
.hero.is-info .navbar-item,
.hero.is-info .navbar-link {
  color: rgba(255, 255, 255, 0.7);
}
.hero.is-info a.navbar-item:hover,
.hero.is-info a.navbar-item.is-active,
.hero.is-info .navbar-link:hover,
.hero.is-info .navbar-link.is-active {
  background-color: #3082c5;
  color: #fff;
}
.hero.is-info .tabs a {
  color: #fff;
  opacity: 0.9;
}
.hero.is-info .tabs a:hover {
  opacity: 1;
}
.hero.is-info .tabs li.is-active a {
  color: hsl(207deg, 61%, 53%) !important;
  opacity: 1;
}
.hero.is-info .tabs.is-boxed a,
.hero.is-info .tabs.is-toggle a {
  color: #fff;
}
.hero.is-info .tabs.is-boxed a:hover,
.hero.is-info .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-info .tabs.is-boxed li.is-active a,
.hero.is-info .tabs.is-boxed li.is-active a:hover,
.hero.is-info .tabs.is-toggle li.is-active a,
.hero.is-info .tabs.is-toggle li.is-active a:hover {
  background-color: #fff;
  border-color: #fff;
  color: hsl(207deg, 61%, 53%);
}
.hero.is-info.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #208fbc 0%,
      hsl(207deg, 61%, 53%) 71%,
      #4d83db 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-info.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #208fbc 0%,
        hsl(207deg, 61%, 53%) 71%,
        #4d83db 100%);
  }
}
.hero.is-success {
  background-color: hsl(153deg, 53%, 53%);
  color: #fff;
}
.hero.is-success a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-success strong {
  color: inherit;
}
.hero.is-success .title {
  color: #fff;
}
.hero.is-success .subtitle {
  color: rgba(255, 255, 255, 0.9);
}
.hero.is-success .subtitle a:not(.button),
.hero.is-success .subtitle strong {
  color: #fff;
}
@media screen and (max-width: 1023px) {
  .hero.is-success .navbar-menu {
    background-color: hsl(153deg, 53%, 53%);
  }
}
.hero.is-success .navbar-item,
.hero.is-success .navbar-link {
  color: rgba(255, 255, 255, 0.7);
}
.hero.is-success a.navbar-item:hover,
.hero.is-success a.navbar-item.is-active,
.hero.is-success .navbar-link:hover,
.hero.is-success .navbar-link.is-active {
  background-color: #3abb81;
  color: #fff;
}
.hero.is-success .tabs a {
  color: #fff;
  opacity: 0.9;
}
.hero.is-success .tabs a:hover {
  opacity: 1;
}
.hero.is-success .tabs li.is-active a {
  color: hsl(153deg, 53%, 53%) !important;
  opacity: 1;
}
.hero.is-success .tabs.is-boxed a,
.hero.is-success .tabs.is-toggle a {
  color: #fff;
}
.hero.is-success .tabs.is-boxed a:hover,
.hero.is-success .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-success .tabs.is-boxed li.is-active a,
.hero.is-success .tabs.is-boxed li.is-active a:hover,
.hero.is-success .tabs.is-toggle li.is-active a,
.hero.is-success .tabs.is-toggle li.is-active a:hover {
  background-color: #fff;
  border-color: #fff;
  color: hsl(153deg, 53%, 53%);
}
.hero.is-success.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #29b35e 0%,
      hsl(153deg, 53%, 53%) 71%,
      #56d2af 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-success.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #29b35e 0%,
        hsl(153deg, 53%, 53%) 71%,
        #56d2af 100%);
  }
}
.hero.is-warning {
  background-color: hsl(44deg, 100%, 77%);
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-warning a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-warning strong {
  color: inherit;
}
.hero.is-warning .title {
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-warning .subtitle {
  color: rgba(0, 0, 0, 0.9);
}
.hero.is-warning .subtitle a:not(.button),
.hero.is-warning .subtitle strong {
  color: rgba(0, 0, 0, 0.7);
}
@media screen and (max-width: 1023px) {
  .hero.is-warning .navbar-menu {
    background-color: hsl(44deg, 100%, 77%);
  }
}
.hero.is-warning .navbar-item,
.hero.is-warning .navbar-link {
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-warning a.navbar-item:hover,
.hero.is-warning a.navbar-item.is-active,
.hero.is-warning .navbar-link:hover,
.hero.is-warning .navbar-link.is-active {
  background-color: #ffd970;
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-warning .tabs a {
  color: rgba(0, 0, 0, 0.7);
  opacity: 0.9;
}
.hero.is-warning .tabs a:hover {
  opacity: 1;
}
.hero.is-warning .tabs li.is-active a {
  color: hsl(44deg, 100%, 77%) !important;
  opacity: 1;
}
.hero.is-warning .tabs.is-boxed a,
.hero.is-warning .tabs.is-toggle a {
  color: rgba(0, 0, 0, 0.7);
}
.hero.is-warning .tabs.is-boxed a:hover,
.hero.is-warning .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-warning .tabs.is-boxed li.is-active a,
.hero.is-warning .tabs.is-boxed li.is-active a:hover,
.hero.is-warning .tabs.is-toggle li.is-active a,
.hero.is-warning .tabs.is-toggle li.is-active a:hover {
  background-color: rgba(0, 0, 0, 0.7);
  border-color: rgba(0, 0, 0, 0.7);
  color: hsl(44deg, 100%, 77%);
}
.hero.is-warning.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #ffb657 0%,
      hsl(44deg, 100%, 77%) 71%,
      #fff6a3 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-warning.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #ffb657 0%,
        hsl(44deg, 100%, 77%) 71%,
        #fff6a3 100%);
  }
}
.hero.is-danger {
  background-color: hsl(348deg, 86%, 61%);
  color: #fff;
}
.hero.is-danger a:not(.button):not(.dropdown-item):not(.tag):not(.pagination-link.is-current),
.hero.is-danger strong {
  color: inherit;
}
.hero.is-danger .title {
  color: #fff;
}
.hero.is-danger .subtitle {
  color: rgba(255, 255, 255, 0.9);
}
.hero.is-danger .subtitle a:not(.button),
.hero.is-danger .subtitle strong {
  color: #fff;
}
@media screen and (max-width: 1023px) {
  .hero.is-danger .navbar-menu {
    background-color: hsl(348deg, 86%, 61%);
  }
}
.hero.is-danger .navbar-item,
.hero.is-danger .navbar-link {
  color: rgba(255, 255, 255, 0.7);
}
.hero.is-danger a.navbar-item:hover,
.hero.is-danger a.navbar-item.is-active,
.hero.is-danger .navbar-link:hover,
.hero.is-danger .navbar-link.is-active {
  background-color: #ef2e55;
  color: #fff;
}
.hero.is-danger .tabs a {
  color: #fff;
  opacity: 0.9;
}
.hero.is-danger .tabs a:hover {
  opacity: 1;
}
.hero.is-danger .tabs li.is-active a {
  color: hsl(348deg, 86%, 61%) !important;
  opacity: 1;
}
.hero.is-danger .tabs.is-boxed a,
.hero.is-danger .tabs.is-toggle a {
  color: #fff;
}
.hero.is-danger .tabs.is-boxed a:hover,
.hero.is-danger .tabs.is-toggle a:hover {
  background-color: rgba(10, 10, 10, 0.1);
}
.hero.is-danger .tabs.is-boxed li.is-active a,
.hero.is-danger .tabs.is-boxed li.is-active a:hover,
.hero.is-danger .tabs.is-toggle li.is-active a,
.hero.is-danger .tabs.is-toggle li.is-active a:hover {
  background-color: #fff;
  border-color: #fff;
  color: hsl(348deg, 86%, 61%);
}
.hero.is-danger.is-bold {
  background-image:
    linear-gradient(
      141deg,
      #fa0a62 0%,
      hsl(348deg, 86%, 61%) 71%,
      #f7595f 100%);
}
@media screen and (max-width: 768px) {
  .hero.is-danger.is-bold .navbar-menu {
    background-image:
      linear-gradient(
        141deg,
        #fa0a62 0%,
        hsl(348deg, 86%, 61%) 71%,
        #f7595f 100%);
  }
}
.hero.is-small .hero-body {
  padding: 1.5rem;
}
@media screen and (min-width: 769px), print {
  .hero.is-medium .hero-body {
    padding: 9rem 4.5rem;
  }
}
@media screen and (min-width: 769px), print {
  .hero.is-large .hero-body {
    padding: 18rem 6rem;
  }
}
.hero.is-halfheight .hero-body,
.hero.is-fullheight .hero-body,
.hero.is-fullheight-with-navbar .hero-body {
  align-items: center;
  display: flex;
}
.hero.is-halfheight .hero-body > .container,
.hero.is-fullheight .hero-body > .container,
.hero.is-fullheight-with-navbar .hero-body > .container {
  flex-grow: 1;
  flex-shrink: 1;
}
.hero.is-halfheight {
  min-height: 50vh;
}
.hero.is-fullheight {
  min-height: 100vh;
}
.hero-video {
  overflow: hidden;
}
.hero-video video {
  left: 50%;
  min-height: 100%;
  min-width: 100%;
  position: absolute;
  top: 50%;
  transform: translate3d(-50%, -50%, 0);
}
.hero-video.is-transparent {
  opacity: 0.3;
}
@media screen and (max-width: 768px) {
  .hero-video {
    display: none;
  }
}
.hero-buttons {
  margin-top: 1.5rem;
}
@media screen and (max-width: 768px) {
  .hero-buttons .button {
    display: flex;
  }
  .hero-buttons .button:not(:last-child) {
    margin-bottom: 0.75rem;
  }
}
@media screen and (min-width: 769px), print {
  .hero-buttons {
    display: flex;
    justify-content: center;
  }
  .hero-buttons .button:not(:last-child) {
    margin-right: 1.5rem;
  }
}
.hero-head,
.hero-foot {
  flex-grow: 0;
  flex-shrink: 0;
}
.hero-body {
  flex-grow: 1;
  flex-shrink: 0;
  padding: 3rem 1.5rem;
}
@media screen and (min-width: 769px), print {
  .hero-body {
    padding: 3rem 3rem;
  }
}
.section {
  padding: 3rem 1.5rem;
}
@media screen and (min-width: 1024px) {
  .section {
    padding: 3rem 3rem;
  }
  .section.is-medium {
    padding: 9rem 4.5rem;
  }
  .section.is-large {
    padding: 18rem 6rem;
  }
}
.footer {
  background-color: hsl(0deg, 0%, 100%);
  padding: 0.495rem 1.5rem;
}
nav.navbar {
  box-shadow: 0 1px 0 rgba(24, 28, 33, 0.04);
}
nav.navbar .navbar-item.has-user-avatar .is-user-avatar {
  margin-right: 0.75rem;
  display: inline-flex;
  width: 1.75rem;
  height: 1.75rem;
}
nav.navbar .navbar-item.has-divider {
  border-right: 1px solid rgba(219, 219, 219, 0.25);
}
nav.navbar .navbar-item.no-left-space {
  padding-left: 0;
}
nav.navbar .navbar-item.has-dropdown {
  padding-right: 0;
  padding-left: 0;
}
nav.navbar .navbar-item.has-dropdown .navbar-link {
  padding-right: 1.125rem;
  padding-left: 1.125rem;
}
nav.navbar .navbar-item.has-control {
  padding-top: 0;
  padding-bottom: 0;
}
nav.navbar .navbar-item .control .input {
  color: hsl(0deg, 0%, 21%);
  border: 0;
  box-shadow: none;
  background: transparent;
}
nav.navbar .navbar-item .control .input::placeholder {
  color: hsl(0deg, 0%, 86%);
}
@media screen and (max-width: 1023px) {
  nav.navbar {
    display: flex;
    padding-right: 0;
  }
  nav.navbar .navbar-brand {
    flex: 1;
  }
  nav.navbar .navbar-brand.is-right {
    flex: none;
  }
  nav.navbar .navbar-item.no-left-space-touch {
    padding-left: 0;
  }
  nav.navbar .navbar-menu {
    position: absolute;
    width: 100vw;
    padding-top: 0;
    top: 3.25rem;
    left: 0;
  }
  nav.navbar .navbar-menu .navbar-item .icon:first-child {
    margin-right: 0.75rem;
  }
  nav.navbar .navbar-menu .navbar-item.has-dropdown > .navbar-link {
    background-color: hsl(0deg, 0%, 96%);
  }
  nav.navbar .navbar-menu .navbar-item.has-dropdown > .navbar-link .icon:last-child {
    display: none;
  }
  nav.navbar .navbar-menu .navbar-item.has-user-avatar > .navbar-link {
    display: flex;
    align-items: center;
    padding-top: 0.75rem;
    padding-bottom: 0.75rem;
  }
}
@media screen and (min-width: 1024px) {
  nav.navbar .navbar-item {
    padding-right: 1.125rem;
    padding-left: 1.125rem;
  }
  nav.navbar .navbar-item:not(.is-desktop-icon-only) .icon:first-child {
    margin-right: 0.75rem;
  }
  nav.navbar .navbar-item.is-desktop-icon-only span:not(.icon) {
    display: none;
  }
}
@media screen and (min-width: 1024px) {
  html.has-aside-left.has-aside-expanded nav.navbar,
  html.has-aside-left.has-aside-expanded body {
    padding-left: 14rem;
  }
  html.has-aside-left aside.is-placed-left {
    display: block;
  }
  aside.aside.is-expanded {
    width: 14rem;
  }
  aside.aside.is-expanded .menu-list .icon {
    width: 3rem;
  }
  aside.aside.is-expanded .menu-list .icon.has-update-mark:after {
    right: 0.65rem;
  }
  aside.aside.is-expanded .menu-list span.menu-item-label {
    display: inline-block;
  }
  aside.aside.is-expanded .menu-list li.is-active ul {
    display: block;
  }
}
aside.aside {
  display: none;
  position: fixed;
  top: 0;
  left: 0;
  z-index: 40;
  height: 100vh;
  padding: 0;
  box-shadow: none;
  background: #2e323a;
}
aside.aside .aside-tools {
  display: flex;
  flex-direction: row;
  width: 100%;
  background-color: #17191e;
  color: hsl(0deg, 0%, 100%);
  line-height: 3.25rem;
  height: 3.25rem;
  padding-left: 0.75rem;
  flex: 1;
}
aside.aside .aside-tools .icon {
  margin-right: 0.75rem;
}
aside.aside .menu-list li a.has-dropdown-icon {
  position: relative;
  padding-right: 3rem;
}
aside.aside .menu-list li a.has-dropdown-icon .dropdown-icon {
  position: absolute;
  top: 0.5rem;
  right: 0;
}
aside.aside .menu-list li ul {
  display: none;
  border-left: 0;
  background-color: #282c33;
  padding-left: 0;
  margin: 0 0 0.75rem;
}
aside.aside .menu-list li ul li a {
  padding: 0.75rem 0 0.75rem 0.75rem;
  font-size: 0.95rem;
}
aside.aside .menu-list li ul li a.has-icon {
  padding-left: 0;
}
aside.aside .menu-list li ul li a.is-active:not(:hover) {
  background: transparent;
}
aside.aside .menu-label {
  padding: 0 0.75rem;
  margin-top: 0.75rem;
  margin-bottom: 0.75rem;
}
@media screen and (max-width: 1023px) {
  nav.navbar {
    transition: margin-left 250ms ease-in-out 50ms, width 250ms ease-in-out 50ms;
  }
  aside.aside {
    transition: left 250ms ease-in-out 50ms;
  }
  html.has-aside-mobile-transition body {
    overflow-x: hidden;
  }
  html.has-aside-mobile-transition body,
  html.has-aside-mobile-transition nav.navbar {
    width: 100vw;
  }
  html.has-aside-mobile-transition aside.aside {
    width: 15rem;
    display: block;
    left: -15rem;
  }
  html.has-aside-mobile-transition aside.aside .image img {
    max-width: 4.95rem;
  }
  html.has-aside-mobile-transition aside.aside .menu-list li.is-active ul {
    display: block;
  }
  html.has-aside-mobile-transition aside.aside .menu-list a .icon {
    width: 3rem;
  }
  html.has-aside-mobile-transition aside.aside .menu-list a .icon.has-update-mark:after {
    right: 0.65rem;
  }
  html.has-aside-mobile-transition aside.aside .menu-list a span.menu-item-label {
    display: inline-block;
  }
  div.has-aside-mobile-expanded nav.navbar {
    margin-left: 15rem;
    width: calc(100vw - 15rem);
  }
  div.has-aside-mobile-expanded aside.aside {
    left: 0;
  }
}
section.section.is-title-bar {
  padding: 1.5rem;
  border-bottom: 1px solid rgba(24, 28, 33, 0.06);
}
section.section.is-title-bar ul li {
  display: inline-block;
  padding: 0 0.75rem 0 0;
  font-size: 1.5rem;
  color: hsl(0deg, 0%, 48%);
}
section.section.is-title-bar ul li:after {
  display: inline-block;
  content: "/";
  padding-left: 0.75rem;
}
section.section.is-title-bar ul li:last-child {
  padding-right: 0;
  font-weight: 900;
  color: hsl(0deg, 0%, 14%);
}
section.section.is-title-bar ul li:last-child:after {
  display: none;
}
section.hero.is-hero-bar {
  background-color: hsl(0deg, 0%, 100%);
  border-bottom: 1px solid rgba(24, 28, 33, 0.06);
}
section.hero.is-hero-bar .hero-body {
  padding: 1.5rem;
}
section.hero.is-hero-bar .hero-body .level-item.is-hero-avatar-item {
  margin-right: 1.5rem;
}
section.hero.is-hero-bar .hero-body .level-item > div > .level {
  margin-bottom: 0.75rem;
}
section.hero.is-hero-bar .hero-body .level-item .subtitle + p {
  margin-top: 0.75rem;
}
section.hero.is-hero-bar .hero-body .button.is-hero-button {
  background-color: rgba(255, 255, 255, 0.5);
  font-weight: 300;
  transition: background-color 250ms ease-in-out 50ms;
}
section.hero.is-hero-bar .hero-body .button.is-hero-button:hover {
  background-color: hsl(0deg, 0%, 100%);
}
.card:not(:last-child) {
  margin-bottom: 1.5rem;
}
.card {
  border-radius: 6px;
  border: 1px solid rgba(24, 28, 33, 0.06);
}
.card.has-table .card-content {
  padding: 0;
}
.card.has-table .b-table {
  border-radius: 6px;
  overflow: hidden;
}
.card.is-card-widget .card-content {
  padding: 0.75rem;
}
.card .card-header {
  border-bottom: 1px solid rgba(24, 28, 33, 0.06);
}
.card .card-content hr {
  margin-left: -1.5rem;
  margin-right: -1.5rem;
}
.card .is-widget-icon .icon {
  width: 5rem;
  height: 5rem;
}
.card .is-widget-label .subtitle {
  color: hsl(0deg, 0%, 48%);
}
table.table thead th {
  border-bottom-width: 1px;
}
table.table td.checkbox-cell .b-checkbox.checkbox:not(.button),
table.table th.checkbox-cell .b-checkbox.checkbox:not(.button) {
  margin-right: 0;
  width: 20px;
}
table.table td.checkbox-cell .b-checkbox.checkbox:not(.button) .control-label,
table.table th.checkbox-cell .b-checkbox.checkbox:not(.button) .control-label {
  display: none;
  padding: 0;
}
table.table td .image {
  margin: 0 auto;
  width: 1.5rem;
  height: 1.5rem;
}
table.table td.is-progress-col {
  min-width: 5rem;
  vertical-align: middle;
}
.b-table {
}
.b-table .table {
  border: 0;
  border-radius: 0;
}
.b-table .table-wrapper {
  margin-bottom: 0;
}
.b-table .table-wrapper + .level {
  padding: 1.125rem 1.5rem;
  padding-left: 1.5rem;
  padding-right: 1.5rem;
  margin: 0;
  border-top: rgba(24, 28, 33, 0.06);
  background: hsl(0deg, 0%, 96%);
}
.b-table .table-wrapper + .level .pagination-link {
  background: hsl(0deg, 0%, 100%);
  color: hsl(0deg, 0%, 21%);
  border-color: hsl(0deg, 0%, 86%);
}
.b-table .table-wrapper + .level .pagination-link.is-current {
  border-color: hsl(0deg, 0%, 29%);
}
.b-table .table-wrapper + .level .pagination-previous,
.b-table .table-wrapper + .level .pagination-next,
.b-table .table-wrapper + .level .pagination-link {
  border-color: hsl(0deg, 0%, 86%);
  color: #2e323a;
}
.b-table .table-wrapper + .level .pagination-previous[disabled],
.b-table .table-wrapper + .level .pagination-next[disabled],
.b-table .table-wrapper + .level .pagination-link[disabled] {
  background-color: transparent;
}
@media screen and (max-width: 768px) {
  .card.has-table .b-table .table-wrapper + .level .level-left + .level-right {
    margin-top: 0;
  }
  .card.has-mobile-sort-spaced .b-table .field.table-mobile-sort {
    padding-top: 0.75rem;
  }
  .b-table .field.table-mobile-sort {
    padding: 0 0.75rem;
  }
  .b-table .table-wrapper.has-mobile-cards tr {
    box-shadow: 0 2px 3px rgba(10, 10, 10, 0.1);
    margin-bottom: 3px !important;
  }
  .b-table .table-wrapper.has-mobile-cards td.is-progress-col span,
  .b-table .table-wrapper.has-mobile-cards td.is-progress-col progress {
    display: flex;
    width: 45%;
    align-items: center;
    align-self: center;
  }
  .b-table .table-wrapper.has-mobile-cards td.checkbox-cell,
  .b-table .table-wrapper.has-mobile-cards td.is-image-cell {
    border-bottom: 0 !important;
  }
  .b-table .table-wrapper.has-mobile-cards td.checkbox-cell:before,
  .b-table .table-wrapper.has-mobile-cards td.is-actions-cell:before {
    display: none;
  }
  .b-table .table-wrapper.has-mobile-cards td.has-no-head-mobile:before {
    display: none;
  }
  .b-table .table-wrapper.has-mobile-cards td.has-no-head-mobile span {
    display: block;
    width: 100%;
  }
  .b-table .table-wrapper.has-mobile-cards td.has-no-head-mobile.is-progress-col progress {
    width: 100%;
  }
  .b-table .table-wrapper.has-mobile-cards td.has-no-head-mobile.is-image-cell .image {
    width: 25vw;
    height: auto;
    margin: 0 auto 0.375rem;
  }
}
.is-tiles-wrapper {
  margin-bottom: 1.5rem;
}
.field.has-check .field-body {
  margin-top: 0.1875rem;
}
.field .control .mdi-24px.mdi-set,
.field .control .mdi-24px.mdi:before {
  font-size: inherit;
}
.upload .upload-draggable {
  display: block;
}
.input,
.textarea,
select {
  box-shadow: none;
}
.input:focus,
.input:active,
.textarea:focus,
.textarea:active,
select:focus,
select:active {
  box-shadow: none !important;
}
.switch input[type=checkbox] + .check:before {
  box-shadow: none;
}
.switch input[type=checkbox]:focus + .check,
.switch input[type=checkbox]:focus:checked + .check,
.b-checkbox.checkbox input[type=checkbox]:focus + .check,
.b-checkbox.checkbox input[type=checkbox]:focus:checked + .check {
  box-shadow: none !important;
}
.b-checkbox.checkbox input[type=checkbox] + .check,
.b-radio.radio input[type=radio] + .check {
  border: 1px solid #2e323a;
}
section.section.is-main-section {
  padding-top: 1.5rem;
}
.modal-card {
  width: 80vw;
}
.modal-card-foot {
  background-color: hsl(0deg, 0%, 96%);
}
@media screen and (max-width: 768px) {
  .modal .animation-content .modal-card {
    width: 90vw;
    margin: 0 auto;
  }
}
footer.footer .logo img {
  width: auto;
  height: 2rem;
}
@media screen and (max-width: 768px) {
  .footer-copyright {
    text-align: center;
  }
}
.is-user-avatar.has-max-width {
  max-width: 7rem;
}
.is-user-avatar.is-aligned-center {
  margin: 0 auto;
}
.is-user-avatar img {
  margin: 0 auto;
  border-radius: 9999px;
}
.icon.has-update-mark {
  position: relative;
}
.icon.has-update-mark:after {
  content: "";
  width: 0.5rem;
  height: 0.5rem;
  position: absolute;
  top: 1px;
  right: 1px;
  background-color: hsl(44deg, 100%, 77%);
  border-radius: 9999px;
}
:root {
  --primary-color: #3298dc;
  --primary-text-color-dark: rgba(0, 0, 0, 0.87);
  --secondary-text-color-dark: rgba(0, 0, 0, 0.57);
  --disabled-text-color-dark: rgba(0, 0, 0, 0.13);
  --primary-text-color-light: rgba(255, 255, 255, 0.87);
  --secondary-text-color-light: rgba(255, 255, 255, 0.57);
  --disabled-text-color-light: rgba(255, 255, 255, 0.13);
  --font-stack:
    "Roboto",
    "Helvetica Neue",
    Helvetica,
    Arial,
    sans-serif;
  --primary-card-color: #fff;
  --primary-background-color: #f2f2f2;
  --box-shadow-lvl-1: 0 1px 3px rgba(0, 0, 0, 0.12), 0 1px 2px rgba(0, 0, 0, 0.24);
  --box-shadow-lvl-2: 0 3px 6px rgba(0, 0, 0, 0.16), 0 3px 6px rgba(0, 0, 0, 0.23);
  --box-shadow-lvl-3: 0 10px 20px rgba(0, 0, 0, 0.19), 0 6px 6px rgba(0, 0, 0, 0.23);
  --box-shadow-lvl-4: 0 14px 28px rgba(0, 0, 0, 0.25), 0 10px 10px rgba(0, 0, 0, 0.22);
}
.datePicker {
  text-align: left;
  background: var(--primary-card-color);
  border-radius: 3px;
  z-index: 200;
  position: fixed;
  height: auto;
  max-height: 90vh;
  width: 90vw;
  max-width: 448px;
  transform-origin: top left;
  transition: transform 0.22s ease-in-out, opacity 0.22s ease-in-out;
  top: 50%;
  left: 50%;
  opacity: 0;
  transform: scale(0) translate(-50%, -50%);
  user-select: none;
}
.datePicker.datePicker--opened {
  opacity: 1;
  transform: scale(1) translate(-50%, -50%);
}
.datePicker .datePicker--titles {
  border-top-left-radius: 3px;
  border-top-right-radius: 3px;
  padding: 24px;
  height: 100px;
  background: var(--primary-color);
}
.datePicker .datePicker--titles h2,
.datePicker .datePicker--titles h3 {
  cursor: pointer;
  color: #fff;
  line-height: 1;
  padding: 0;
  margin: 0;
  font-size: 32px;
}
.datePicker .datePicker--titles h3 {
  color: rgba(255, 255, 255, 0.57);
  font-size: 18px;
  padding-bottom: 2px;
}
.datePicker nav {
  padding: 20px;
  height: 56px;
}
.datePicker nav h4 {
  width: calc(100% - 60px);
  text-align: center;
  display: inline-block;
  padding: 0;
  font-size: 14px;
  line-height: 24px;
  margin: 0;
  position: relative;
  top: -9px;
  color: var(--primary-text-color);
}
.datePicker nav i {
  cursor: pointer;
  color: var(--secondary-text-color);
  font-size: 26px;
  user-select: none;
  border-radius: 50%;
}
.datePicker nav i:hover {
  background: var(--disabled-text-color-dark);
}
.datePicker .datePicker--scroll {
  overflow-y: auto;
  max-height: calc(90vh - 56px - 100px);
}
.datePicker .datePicker--calendar {
  padding: 0 20px;
}
.datePicker .datePicker--calendar .datePicker--dayNames {
  width: 100%;
  display: grid;
  text-align: center;
  grid-template-columns: 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857%;
}
.datePicker .datePicker--calendar .datePicker--dayNames span {
  color: var(--secondary-text-color-dark);
  font-size: 14px;
  line-height: 42px;
  display: inline-grid;
}
.datePicker .datePicker--calendar .datePicker--days {
  width: 100%;
  display: grid;
  text-align: center;
  grid-template-columns: 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857% 14.2857142857%;
}
.datePicker .datePicker--calendar .datePicker--days span {
  color: var(--primary-text-color-dark);
  line-height: 42px;
  font-size: 14px;
  display: inline-grid;
  transition: color 0.22s;
  height: 42px;
  position: relative;
  cursor: pointer;
  user-select: none;
  border-radius: 50%;
}
.datePicker .datePicker--calendar .datePicker--days span::before {
  content: "";
  position: absolute;
  z-index: -1;
  height: 42px;
  width: 42px;
  left: calc(50% - 21px);
  background: var(--primary-color);
  border-radius: 50%;
  transition: transform 0.22s, opacity 0.22s;
  transform: scale(0);
  opacity: 0;
}
.datePicker .datePicker--calendar .datePicker--days span[disabled=true] {
  cursor: unset;
}
.datePicker .datePicker--calendar .datePicker--days span.datePicker--today {
  font-weight: 700;
}
.datePicker .datePicker--calendar .datePicker--days span.datePicker--selected {
  color: rgba(255, 255, 255, 0.87);
}
.datePicker .datePicker--calendar .datePicker--days span.datePicker--selected:before {
  transform: scale(1);
  opacity: 1;
}
.datePicker .datePicker--selectYear {
  padding: 0 20px;
  display: block;
  width: 100%;
  text-align: center;
  max-height: 362px;
}
.datePicker .datePicker--selectYear span {
  display: block;
  width: 100%;
  font-size: 24px;
  margin: 20px auto;
  cursor: pointer;
}
.datePicker .datePicker--selectYear span.selected {
  font-size: 42px;
  color: var(--primary-color);
}
.datePicker div.datePicker--actions {
  width: 100%;
  padding: 8px;
  text-align: right;
}
.datePicker div.datePicker--actions button {
  margin-bottom: 0;
  font-size: 15px;
  cursor: pointer;
  color: var(--primary-text-color);
  border: none;
  margin-left: 8px;
  min-width: 64px;
  line-height: 36px;
  background-color: transparent;
  appearance: none;
  padding: 0 16px;
  border-radius: 3px;
  transition: background-color 0.13s;
}
.datePicker div.datePicker--actions button:hover,
.datePicker div.datePicker--actions button:focus {
  outline: none;
  background-color: var(--disabled-text-color-dark);
}
.datePicker--background {
  z-index: 199;
  position: fixed;
  top: 0;
  left: 0;
  bottom: 0;
  right: 0;
  background: rgba(0, 0, 0, 0.52);
  animation: fadeIn 0.22s forwards;
}
.lds-ring {
  display: inline-block;
  position: relative;
  width: 80px;
  height: 80px;
}
.lds-ring div {
  box-sizing: border-box;
  display: block;
  position: absolute;
  width: 64px;
  height: 64px;
  margin: 8px;
  border: 8px solid black;
  border-radius: 50%;
  animation: lds-ring 1.2s cubic-bezier(0.5, 0, 0.5, 1) infinite;
  border-color: black transparent transparent transparent;
}
.lds-ring div:nth-child(1) {
  animation-delay: -0.45s;
}
.lds-ring div:nth-child(2) {
  animation-delay: -0.3s;
}
.lds-ring div:nth-child(3) {
  animation-delay: -0.15s;
}
@keyframes lds-ring {
  0% {
    transform: rotate(0deg);
  }
  100% {
    transform: rotate(360deg);
  }
}
.toggle {
  cursor: pointer;
  display: inline-block;
}
.toggle-switch {
  display: inline-block;
  background: #ccc;
  border-radius: 16px;
  width: 58px;
  height: 32px;
  position: relative;
  vertical-align: middle;
  transition: background 0.25s;
}
.toggle-switch:before,
.toggle-switch:after {
  content: "";
}
.toggle-switch:before {
  display: block;
  background:
    linear-gradient(
      to bottom,
      #fff 0%,
      #eee 100%);
  border-radius: 50%;
  box-shadow: 0 0 0 1px rgba(0, 0, 0, 0.25);
  width: 24px;
  height: 24px;
  position: absolute;
  top: 4px;
  left: 4px;
  transition: left 0.25s;
}
.toggle:hover .toggle-switch:before {
  background:
    linear-gradient(
      to bottom,
      #fff 0%,
      #fff 100%);
  box-shadow: 0 0 0 1px rgba(0, 0, 0, 0.5);
}
.toggle-switch.disabled:before {
  background:
    linear-gradient(
      to bottom,
      #ccc 0%,
      #bbb 100%);
}
.toggle:hover .toggle-switch.disabled:before {
  background:
    linear-gradient(
      to bottom,
      #ccc 0%,
      #bbb 100%);
}
.toggle-checkbox:checked + .toggle-switch {
  background: #56c080;
}
.toggle-checkbox:checked + .toggle-switch:before {
  left: 30px;
}
.toggle-checkbox:not(checked) + .toggle-switch {
  background: hsl(0deg, 0%, 86%);
}
.toggle-checkbox:not(checked) + .toggle-switch:before {
  left: 4px;
}
.toggle-checkbox:indeterminate + .toggle-switch {
  background: rgba(0, 0, 0, 0.301);
}
.toggle-checkbox:indeterminate + .toggle-switch:before {
  left: 16px;
  background: rgba(0, 0, 0, 0.301);
}
.toggle-checkbox {
  position: absolute;
  visibility: hidden;
}
.toggle-label {
  margin-left: 5px;
  position: relative;
  top: 2px;
}
.notification {
  background-color: transparent;
}
.timeline .timeline-item .timeline-content {
  padding-top: 0;
}
.timeline .timeline-item:last-child::before {
  display: none;
}
.timeline .timeline-item .timeline-marker {
  top: 0;
}
.toast {
  position: absolute;
  width: 60%;
  margin-left: 10%;
  margin-right: 10%;
  z-index: 999;
  display: flex;
  flex-direction: column;
  padding: 15px;
  text-align: center;
  pointer-events: none;
}
.toast > .message {
  white-space: pre-wrap;
  opacity: 80%;
}
div.is-loading {
  position: relative;
  pointer-events: none;
  opacity: 0.5;
}
div.is-loading:after {
  position: absolute;
  top: calc(50% - 2.5em);
  left: calc(50% - 2.5em);
  width: 5em;
  height: 5em;
  border-width: 0.25em;
}
input[type=checkbox]:indeterminate + .check {
  background: red !important;
}
.right-sticky {
  position: sticky;
  right: 0px;
  background-color: hsl(0deg, 0%, 100%);
}
.right-sticky .buttons {
  flex-wrap: nowrap;
}
.table.is-striped tbody tr:not(.is-selected):nth-child(even) .right-sticky {
  background-color: #fafafa;
}
tr:hover .right-sticky {
  background-color: hsl(0deg, 0%, 80%);
}
.table.is-striped tbody tr:nth-child(even):hover .right-sticky {
  background-color: hsl(0deg, 0%, 95%);
}
.content-full-size {
  height: calc(100% - 3rem);
  position: absolute;
  width: calc(100% - 14rem);
  display: flex;
}
.content-full-size .column .card {
  min-width: 200px;
}
@media screen and (max-width: 1023px) {
  .content-full-size {
    height: 100%;
    position: absolute;
    width: 100%;
  }
}
.column.is-half {
  flex: none;
  width: 50%;
}
input:read-only {
  cursor: initial;
}
[data-tooltip]:before {
  max-width: 15rem;
  width: max-content;
  text-align: left;
  transition: opacity 0.1s linear 1s;
  white-space: pre-wrap !important;
  font-weight: normal;
}
.icon[data-tooltip]:before {
  transition: none;
  z-index: 5;
}
span[data-tooltip] {
  border-bottom: none;
}
div[data-tooltip]::before {
  position: absolute;
}
.modal-card-body > p {
  padding: 1em;
}
.modal-card-body > p.warning {
  background-color: #fffbdd;
  border: solid 1px #f2e9bf;
}
/*! bulma.io v0.9.4 | MIT License | github.com/jgthms/bulma */
/*! minireset.css v0.0.6 | MIT License | github.com/jgthms/minireset.css */
/*! Bundled license information:

@creativebulma/bulma-tooltip/dist/bulma-tooltip.min.css:
  (*! @creativebulma/bulma-tooltip v1.2.0 | (c) 2020 Gaetan | MIT License | https://github.com/CreativeBulma/bulma-tooltip *)
*/
/*# sourceMappingURL=index.css.map */
